/* Generated by Cython 0.29.15 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "name": "validr._validator_c",
        "sources": [
            "src/validr/_validator_c.pyx"
        ]
    },
    "module_name": "validr._validator_c"
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_15"
#define CYTHON_HEX_VERSION 0x001D0FF0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
#if PY_VERSION_HEX >= 0x030800A4 && PY_VERSION_HEX < 0x030800B2
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, 0, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
  #define PyObject_Unicode             PyObject_Str
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : (Py_INCREF(func), func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif


#define __PYX_ERR(f_index, lineno, Ln_error) \
{ \
  __pyx_filename = __pyx_f[f_index]; __pyx_lineno = lineno; __pyx_clineno = __LINE__; goto Ln_error; \
}

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__validr___validator_c
#define __PYX_HAVE_API__validr___validator_c
/* Early includes */
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "src/validr/_validator_c.pyx",
};

/*--- Type declarations ---*/
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct__validator;
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_1_decorator;
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_2__m_validator;
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_3__key_func_of_schema;
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_4_key_of;
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_5_genexpr;
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_6_key_of;
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_7_genexpr;
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_8_list_validator;
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_9_dict_validator;
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_10_model_validator;
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_11_enum_validator;
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_12__union_list_validator;
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_13__union_dict_validator;
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_14_int_validator;
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_15_float_validator;
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_16_str_validator;
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_17_bytes_validator;
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_18_date_validator;
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_19_time_validator;
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_20_datetime_validator;
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_21_timedelta_validator;
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_22_ipv4_validator;
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_23_ipv6_validator;
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_24_email_validator;
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_25_url_validator;
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_26_uuid_validator;
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_27_create_re_validator;
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_28_re_validator;
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_29_create_enum_validator;
struct __pyx_defaults;
typedef struct __pyx_defaults __pyx_defaults;
struct __pyx_defaults {
  int __pyx_arg_maxlen;
};

/* "validr/_validator_c.pyx":91
 * 
 * 
 * def validator(string=None, *, accept=None, output=None):             # <<<<<<<<<<<<<<
 *     """Decorator for create validator
 * 
 */
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct__validator {
  PyObject_HEAD
  PyObject *__pyx_v_accept_hints;
  int __pyx_v_accept_object;
  int __pyx_v_accept_string;
  PyObject *__pyx_v_output_hints;
  int __pyx_v_output_object;
  int __pyx_v_output_string;
};


/* "validr/_validator_c.pyx":145
 *     del string, accept, output
 * 
 *     def decorator(f):             # <<<<<<<<<<<<<<
 * 
 *         def _m_validator(compiler, schema):
 */
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_1_decorator {
  PyObject_HEAD
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct__validator *__pyx_outer_scope;
  PyObject *__pyx_v__m_validator;
  PyObject *__pyx_v_f;
};


/* "validr/_validator_c.pyx":147
 *     def decorator(f):
 * 
 *         def _m_validator(compiler, schema):             # <<<<<<<<<<<<<<
 *             params = schema.params.copy()
 *             if schema.items is not None:
 */
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_2__m_validator {
  PyObject_HEAD
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_1_decorator *__pyx_outer_scope;
  PyObject *__pyx_v__m_validate;
  PyObject *__pyx_v_default;
  int __pyx_v_has_default;
  PyObject *__pyx_v_invalid_to;
  PyObject *__pyx_v_null_output;
  int __pyx_v_optional;
  int __pyx_v_supress_invalid;
  PyObject *__pyx_v_validate;
};


/* "validr/_validator_c.pyx":285
 * 
 * 
 * def _key_func_of_schema(schema):             # <<<<<<<<<<<<<<
 *     if schema is None:
 *         raise SchemaError(_UNIQUE_CHECK_ERROR_MESSAGE)
 */
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_3__key_func_of_schema {
  PyObject_HEAD
  PyObject *__pyx_v_key_of_value;
  PyObject *__pyx_v_keys;
};


/* "validr/_validator_c.pyx":296
 *             keys.append((k, _key_func_of_schema(v)))
 * 
 *         def key_of(dict v):             # <<<<<<<<<<<<<<
 *             cdef str k
 *             return tuple(key_of_value(v[k]) for k, key_of_value in keys)
 */
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_4_key_of {
  PyObject_HEAD
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_3__key_func_of_schema *__pyx_outer_scope;
  PyObject *__pyx_v_v;
};


/* "validr/_validator_c.pyx":298
 *         def key_of(dict v):
 *             cdef str k
 *             return tuple(key_of_value(v[k]) for k, key_of_value in keys)             # <<<<<<<<<<<<<<
 * 
 *     elif schema.validator == 'list':
 */
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_5_genexpr {
  PyObject_HEAD
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_4_key_of *__pyx_outer_scope;
  PyObject *__pyx_v_k;
  PyObject *__pyx_v_key_of_value;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
};


/* "validr/_validator_c.pyx":305
 *         key_of_value = _key_func_of_schema(schema.items)
 * 
 *         def key_of(list v):             # <<<<<<<<<<<<<<
 *             return tuple(key_of_value(x) for x in v)
 * 
 */
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_6_key_of {
  PyObject_HEAD
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_3__key_func_of_schema *__pyx_outer_scope;
  PyObject *__pyx_v_v;
};


/* "validr/_validator_c.pyx":306
 * 
 *         def key_of(list v):
 *             return tuple(key_of_value(x) for x in v)             # <<<<<<<<<<<<<<
 * 
 *     else:
 */
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_7_genexpr {
  PyObject_HEAD
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_6_key_of *__pyx_outer_scope;
  PyObject *__pyx_v_x;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
};


/* "validr/_validator_c.pyx":315
 * 
 * @validator(accept=typing.Iterable, output=typing.List)
 * def list_validator(compiler, items=None, int minlen=0, int maxlen=1024,             # <<<<<<<<<<<<<<
 *                    bint unique=False):
 *     if items is None:
 */
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_8_list_validator {
  PyObject_HEAD
  PyObject *__pyx_v_inner;
  PyObject *__pyx_v_key_of;
  int __pyx_v_maxlen;
  int __pyx_v_minlen;
  int __pyx_v_unique;
};


/* "validr/_validator_c.pyx":357
 * 
 * @validator(accept=(typing.Mapping, typing.Any), output=dict)
 * def dict_validator(compiler, items=None, key=None, value=None,             # <<<<<<<<<<<<<<
 *                    int minlen=0, int maxlen=1024, bint slim=False):
 *     if items is None:
 */
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_9_dict_validator {
  PyObject_HEAD
  PyObject *__pyx_v_inners;
  int __pyx_v_is_dynamic;
  int __pyx_v_maxlen;
  int __pyx_v_minlen;
  int __pyx_v_slim;
  PyObject *__pyx_v_validate_extra_key;
  PyObject *__pyx_v_validate_extra_value;
};


/* "validr/_validator_c.pyx":424
 * 
 * @validator(accept=(typing.Mapping, typing.Any), output=object)
 * def model_validator(compiler, items=None):             # <<<<<<<<<<<<<<
 *     if items is None:
 *         raise SchemaError('model class not provided')
 */
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_10_model_validator {
  PyObject_HEAD
  PyObject *__pyx_v_items;
};


/* "validr/_validator_c.pyx":448
 * 
 * @validator(output=object)
 * def enum_validator(compiler, items):             # <<<<<<<<<<<<<<
 *     if not items:
 *         raise SchemaError("enum items not provided")
 */
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_11_enum_validator {
  PyObject_HEAD
  PyObject *__pyx_v_expects;
  PyObject *__pyx_v_items;
};


/* "validr/_validator_c.pyx":491
 * 
 * 
 * def _union_list_validator(compiler, schema):             # <<<<<<<<<<<<<<
 *     scalar_inner = None
 *     list_inner = None
 */
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_12__union_list_validator {
  PyObject_HEAD
  PyObject *__pyx_v_dict_inner;
  PyObject *__pyx_v_list_inner;
  PyObject *__pyx_v_scalar_inner;
};


/* "validr/_validator_c.pyx":539
 * 
 * @validator(accept=object, output=object)
 * def _union_dict_validator(compiler, items, str by):             # <<<<<<<<<<<<<<
 *     inners = {}
 *     for key, schema in items.items():
 */
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_13__union_dict_validator {
  PyObject_HEAD
  PyObject *__pyx_v_by;
  PyObject *__pyx_v_expect_bys;
  PyObject *__pyx_v_inners;
};


/* "validr/_validator_c.pyx":586
 * 
 * @validator(accept=(int, float, str), output=int)
 * def int_validator(compiler, min=-sys.maxsize, max=sys.maxsize):             # <<<<<<<<<<<<<<
 *     """Validate int or convert string to int
 * 
 */
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_14_int_validator {
  PyObject_HEAD
  PyObject *__pyx_v_max;
  PyObject *__pyx_v_min;
};


/* "validr/_validator_c.pyx":636
 * 
 * @validator(accept=(int, float, str), output=float)
 * def float_validator(compiler, min=-sys.float_info.max, max=sys.float_info.max,             # <<<<<<<<<<<<<<
 *                     exmin=False, exmax=False):
 *     """Validate float string
 */
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_15_float_validator {
  PyObject_HEAD
  PyObject *__pyx_v_exmax;
  PyObject *__pyx_v_exmin;
  PyObject *__pyx_v_max;
  PyObject *__pyx_v_min;
};


/* "validr/_validator_c.pyx":680
 * 
 * @validator(accept=(str, object), output=str)
 * def str_validator(compiler, int minlen=0, int maxlen=1024 * 1024,             # <<<<<<<<<<<<<<
 *                   bint strip=False, bint escape=False, str match=None,
 *                   bint accept_object=False):
 */
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_16_str_validator {
  PyObject_HEAD
  int __pyx_v_accept_object;
  int __pyx_v_escape;
  PyObject *__pyx_v_match;
  int __pyx_v_maxlen;
  int __pyx_v_minlen;
  PyObject *__pyx_v_re_match;
  int __pyx_v_strip;
};


/* "validr/_validator_c.pyx":728
 * 
 * @validator(accept=bytes, output=bytes)
 * def bytes_validator(compiler, int minlen=0, int maxlen=-1):             # <<<<<<<<<<<<<<
 *     """Validate bytes
 * 
 */
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_17_bytes_validator {
  PyObject_HEAD
  int __pyx_v_maxlen;
  int __pyx_v_minlen;
};


/* "validr/_validator_c.pyx":750
 * 
 * @validator(accept=(str, datetime.date), output=(str, datetime.date))
 * def date_validator(compiler, str format='%Y-%m-%d', bint output_object=False):             # <<<<<<<<<<<<<<
 *     """Validate date string or convert date to string
 * 
 */
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_18_date_validator {
  PyObject_HEAD
  PyObject *__pyx_v_format;
  int __pyx_v_output_object;
};


/* "validr/_validator_c.pyx":772
 * 
 * @validator(accept=(str, datetime.time), output=(str, datetime.time))
 * def time_validator(compiler, str format='%H:%M:%S', bint output_object=False):             # <<<<<<<<<<<<<<
 *     """Validate time string or convert time to string
 * 
 */
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_19_time_validator {
  PyObject_HEAD
  PyObject *__pyx_v_format;
  int __pyx_v_output_object;
};


/* "validr/_validator_c.pyx":794
 * 
 * @validator(accept=(str, datetime.datetime), output=(str, datetime.datetime))
 * def datetime_validator(compiler, str format='%Y-%m-%dT%H:%M:%S.%fZ', bint output_object=False):             # <<<<<<<<<<<<<<
 *     """Validate datetime string or convert datetime to string
 * 
 */
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_20_datetime_validator {
  PyObject_HEAD
  PyObject *__pyx_v_format;
  int __pyx_v_output_object;
};


/* "validr/_validator_c.pyx":827
 * 
 * @validator(accept=(int, float, str, datetime.timedelta), output=(str, float, datetime.timedelta))
 * def timedelta_validator(compiler, min=None, max=None, bint string=False,             # <<<<<<<<<<<<<<
 *                         bint extended=False, bint output_object=False):
 *     """Validate timedelta string or convert timedelta to string
 */
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_21_timedelta_validator {
  PyObject_HEAD
  int __pyx_v_extended;
  PyObject *__pyx_v_max_repr;
  PyObject *__pyx_v_max_value;
  PyObject *__pyx_v_min_repr;
  PyObject *__pyx_v_min_value;
  int __pyx_v_output_object;
  int __pyx_v_string;
};


/* "validr/_validator_c.pyx":881
 * 
 * @validator(accept=(str, ipaddress.IPv4Address), output=(str, ipaddress.IPv4Address))
 * def ipv4_validator(compiler, bint output_object=False):             # <<<<<<<<<<<<<<
 *     def validate(value):
 *         try:
 */
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_22_ipv4_validator {
  PyObject_HEAD
  int __pyx_v_output_object;
};


/* "validr/_validator_c.pyx":897
 * 
 * @validator(accept=(str, ipaddress.IPv6Address), output=(str, ipaddress.IPv6Address))
 * def ipv6_validator(compiler, bint output_object=False):             # <<<<<<<<<<<<<<
 *     def validate(value):
 *         try:
 */
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_23_ipv6_validator {
  PyObject_HEAD
  int __pyx_v_output_object;
};


/* "validr/_validator_c.pyx":913
 * 
 * @validator(accept=str, output=str)
 * def email_validator(compiler):             # <<<<<<<<<<<<<<
 *     # https://stackoverflow.com/questions/201323/using-a-regular-expression-to-validate-an-email-address
 *     # http://emailregex.com/
 */
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_24_email_validator {
  PyObject_HEAD
  PyObject *__pyx_v__validate;
};


/* "validr/_validator_c.pyx":936
 * 
 * @validator(output=(str, object))
 * def url_validator(compiler, str scheme='http https', int maxlen=255, bint output_object=False):             # <<<<<<<<<<<<<<
 *     # https://stackoverflow.com/questions/7160737/python-how-to-validate-a-url-in-python-malformed-or-not
 *     # https://stackoverflow.com/questions/827557/how-do-you-validate-a-url-with-a-regular-expression-in-python
 */
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_25_url_validator {
  PyObject_HEAD
  PyObject *__pyx_v_allow_scheme;
  int __pyx_v_maxlen;
  int __pyx_v_output_object;
};


/* "validr/_validator_c.pyx":979
 * 
 * @validator(output=(str, uuid.UUID))
 * def uuid_validator(compiler, version=None, bint output_object=False):             # <<<<<<<<<<<<<<
 *     if version is None:
 *         msg = 'invalid uuid'
 */
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_26_uuid_validator {
  PyObject_HEAD
  PyObject *__pyx_v_msg;
  int __pyx_v_output_object;
  PyObject *__pyx_v_version;
};


/* "validr/_validator_c.pyx":1002
 * 
 * 
 * def create_re_validator(str name, str r, int default_maxlen=255):             # <<<<<<<<<<<<<<
 *     """Create validator by regex string
 * 
 */
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_27_create_re_validator {
  PyObject_HEAD
  PyObject *__pyx_v_match;
  PyObject *__pyx_v_message;
  PyObject *__pyx_v_name;
};


/* "validr/_validator_c.pyx":1016
 *     message = 'invalid %s' % name
 * 
 *     def re_validator(compiler, int minlen=0, int maxlen=default_maxlen, bint strip=False):             # <<<<<<<<<<<<<<
 *         def validate(value):
 *             if not isinstance(value, str):
 */
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_28_re_validator {
  PyObject_HEAD
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_27_create_re_validator *__pyx_outer_scope;
  int __pyx_v_maxlen;
  int __pyx_v_minlen;
  int __pyx_v_strip;
};


/* "validr/_validator_c.pyx":1070
 * 
 * 
 * def create_enum_validator(str name, items, bint string=True):             # <<<<<<<<<<<<<<
 *     """Create validator by enum items
 * 
 */
struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_29_create_enum_validator {
  PyObject_HEAD
  PyObject *__pyx_v_items;
  PyObject *__pyx_v_message;
};


/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* Profile.proto */
#ifndef CYTHON_PROFILE
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
  #define CYTHON_PROFILE 0
#else
  #define CYTHON_PROFILE 1
#endif
#endif
#ifndef CYTHON_TRACE_NOGIL
  #define CYTHON_TRACE_NOGIL 0
#else
  #if CYTHON_TRACE_NOGIL && !defined(CYTHON_TRACE)
    #define CYTHON_TRACE 1
  #endif
#endif
#ifndef CYTHON_TRACE
  #define CYTHON_TRACE 0
#endif
#if CYTHON_TRACE
  #undef CYTHON_PROFILE_REUSE_FRAME
#endif
#ifndef CYTHON_PROFILE_REUSE_FRAME
  #define CYTHON_PROFILE_REUSE_FRAME 0
#endif
#if CYTHON_PROFILE || CYTHON_TRACE
  #include "compile.h"
  #include "frameobject.h"
  #include "traceback.h"
  #if CYTHON_PROFILE_REUSE_FRAME
    #define CYTHON_FRAME_MODIFIER static
    #define CYTHON_FRAME_DEL(frame)
  #else
    #define CYTHON_FRAME_MODIFIER
    #define CYTHON_FRAME_DEL(frame) Py_CLEAR(frame)
  #endif
  #define __Pyx_TraceDeclarations\
  static PyCodeObject *__pyx_frame_code = NULL;\
  CYTHON_FRAME_MODIFIER PyFrameObject *__pyx_frame = NULL;\
  int __Pyx_use_tracing = 0;
  #define __Pyx_TraceFrameInit(codeobj)\
  if (codeobj) __pyx_frame_code = (PyCodeObject*) codeobj;
  #ifdef WITH_THREAD
  #define __Pyx_TraceCall(funcname, srcfile, firstlineno, nogil, goto_error)\
  if (nogil) {\
      if (CYTHON_TRACE_NOGIL) {\
          PyThreadState *tstate;\
          PyGILState_STATE state = PyGILState_Ensure();\
          tstate = __Pyx_PyThreadState_Current;\
          if (unlikely(tstate->use_tracing) && !tstate->tracing &&\
                  (tstate->c_profilefunc || (CYTHON_TRACE && tstate->c_tracefunc))) {\
              __Pyx_use_tracing = __Pyx_TraceSetupAndCall(&__pyx_frame_code, &__pyx_frame, tstate, funcname, srcfile, firstlineno);\
          }\
          PyGILState_Release(state);\
          if (unlikely(__Pyx_use_tracing < 0)) goto_error;\
      }\
  } else {\
      PyThreadState* tstate = PyThreadState_GET();\
      if (unlikely(tstate->use_tracing) && !tstate->tracing &&\
              (tstate->c_profilefunc || (CYTHON_TRACE && tstate->c_tracefunc))) {\
          __Pyx_use_tracing = __Pyx_TraceSetupAndCall(&__pyx_frame_code, &__pyx_frame, tstate, funcname, srcfile, firstlineno);\
          if (unlikely(__Pyx_use_tracing < 0)) goto_error;\
      }\
  }
  #else
  #define __Pyx_TraceCall(funcname, srcfile, firstlineno, nogil, goto_error)\
  {   PyThreadState* tstate = PyThreadState_GET();\
      if (unlikely(tstate->use_tracing) && !tstate->tracing &&\
              (tstate->c_profilefunc || (CYTHON_TRACE && tstate->c_tracefunc))) {\
          __Pyx_use_tracing = __Pyx_TraceSetupAndCall(&__pyx_frame_code, &__pyx_frame, tstate, funcname, srcfile, firstlineno);\
          if (unlikely(__Pyx_use_tracing < 0)) goto_error;\
      }\
  }
  #endif
  #define __Pyx_TraceException()\
  if (likely(!__Pyx_use_tracing)); else {\
      PyThreadState* tstate = __Pyx_PyThreadState_Current;\
      if (tstate->use_tracing &&\
              (tstate->c_profilefunc || (CYTHON_TRACE && tstate->c_tracefunc))) {\
          tstate->tracing++;\
          tstate->use_tracing = 0;\
          PyObject *exc_info = __Pyx_GetExceptionTuple(tstate);\
          if (exc_info) {\
              if (CYTHON_TRACE && tstate->c_tracefunc)\
                  tstate->c_tracefunc(\
                      tstate->c_traceobj, __pyx_frame, PyTrace_EXCEPTION, exc_info);\
              tstate->c_profilefunc(\
                  tstate->c_profileobj, __pyx_frame, PyTrace_EXCEPTION, exc_info);\
              Py_DECREF(exc_info);\
          }\
          tstate->use_tracing = 1;\
          tstate->tracing--;\
      }\
  }
  static void __Pyx_call_return_trace_func(PyThreadState *tstate, PyFrameObject *frame, PyObject *result) {
      PyObject *type, *value, *traceback;
      __Pyx_ErrFetchInState(tstate, &type, &value, &traceback);
      tstate->tracing++;
      tstate->use_tracing = 0;
      if (CYTHON_TRACE && tstate->c_tracefunc)
          tstate->c_tracefunc(tstate->c_traceobj, frame, PyTrace_RETURN, result);
      if (tstate->c_profilefunc)
          tstate->c_profilefunc(tstate->c_profileobj, frame, PyTrace_RETURN, result);
      CYTHON_FRAME_DEL(frame);
      tstate->use_tracing = 1;
      tstate->tracing--;
      __Pyx_ErrRestoreInState(tstate, type, value, traceback);
  }
  #ifdef WITH_THREAD
  #define __Pyx_TraceReturn(result, nogil)\
  if (likely(!__Pyx_use_tracing)); else {\
      if (nogil) {\
          if (CYTHON_TRACE_NOGIL) {\
              PyThreadState *tstate;\
              PyGILState_STATE state = PyGILState_Ensure();\
              tstate = __Pyx_PyThreadState_Current;\
              if (tstate->use_tracing) {\
                  __Pyx_call_return_trace_func(tstate, __pyx_frame, (PyObject*)result);\
              }\
              PyGILState_Release(state);\
          }\
      } else {\
          PyThreadState* tstate = __Pyx_PyThreadState_Current;\
          if (tstate->use_tracing) {\
              __Pyx_call_return_trace_func(tstate, __pyx_frame, (PyObject*)result);\
          }\
      }\
  }
  #else
  #define __Pyx_TraceReturn(result, nogil)\
  if (likely(!__Pyx_use_tracing)); else {\
      PyThreadState* tstate = __Pyx_PyThreadState_Current;\
      if (tstate->use_tracing) {\
          __Pyx_call_return_trace_func(tstate, __pyx_frame, (PyObject*)result);\
      }\
  }
  #endif
  static PyCodeObject *__Pyx_createFrameCodeObject(const char *funcname, const char *srcfile, int firstlineno);
  static int __Pyx_TraceSetupAndCall(PyCodeObject** code, PyFrameObject** frame, PyThreadState* tstate, const char *funcname, const char *srcfile, int firstlineno);
#else
  #define __Pyx_TraceDeclarations
  #define __Pyx_TraceFrameInit(codeobj)
  #define __Pyx_TraceCall(funcname, srcfile, firstlineno, nogil, goto_error)   if ((1)); else goto_error;
  #define __Pyx_TraceException()
  #define __Pyx_TraceReturn(result, nogil)
#endif
#if CYTHON_TRACE
  static int __Pyx_call_line_trace_func(PyThreadState *tstate, PyFrameObject *frame, int lineno) {
      int ret;
      PyObject *type, *value, *traceback;
      __Pyx_ErrFetchInState(tstate, &type, &value, &traceback);
      __Pyx_PyFrame_SetLineNumber(frame, lineno);
      tstate->tracing++;
      tstate->use_tracing = 0;
      ret = tstate->c_tracefunc(tstate->c_traceobj, frame, PyTrace_LINE, NULL);
      tstate->use_tracing = 1;
      tstate->tracing--;
      if (likely(!ret)) {
          __Pyx_ErrRestoreInState(tstate, type, value, traceback);
      } else {
          Py_XDECREF(type);
          Py_XDECREF(value);
          Py_XDECREF(traceback);
      }
      return ret;
  }
  #ifdef WITH_THREAD
  #define __Pyx_TraceLine(lineno, nogil, goto_error)\
  if (likely(!__Pyx_use_tracing)); else {\
      if (nogil) {\
          if (CYTHON_TRACE_NOGIL) {\
              int ret = 0;\
              PyThreadState *tstate;\
              PyGILState_STATE state = PyGILState_Ensure();\
              tstate = __Pyx_PyThreadState_Current;\
              if (unlikely(tstate->use_tracing && tstate->c_tracefunc && __pyx_frame->f_trace)) {\
                  ret = __Pyx_call_line_trace_func(tstate, __pyx_frame, lineno);\
              }\
              PyGILState_Release(state);\
              if (unlikely(ret)) goto_error;\
          }\
      } else {\
          PyThreadState* tstate = __Pyx_PyThreadState_Current;\
          if (unlikely(tstate->use_tracing && tstate->c_tracefunc && __pyx_frame->f_trace)) {\
              int ret = __Pyx_call_line_trace_func(tstate, __pyx_frame, lineno);\
              if (unlikely(ret)) goto_error;\
          }\
      }\
  }
  #else
  #define __Pyx_TraceLine(lineno, nogil, goto_error)\
  if (likely(!__Pyx_use_tracing)); else {\
      PyThreadState* tstate = __Pyx_PyThreadState_Current;\
      if (unlikely(tstate->use_tracing && tstate->c_tracefunc && __pyx_frame->f_trace)) {\
          int ret = __Pyx_call_line_trace_func(tstate, __pyx_frame, lineno);\
          if (unlikely(ret)) goto_error;\
      }\
  }
  #endif
#else
  #define __Pyx_TraceLine(lineno, nogil, goto_error)   if ((1)); else goto_error;
#endif

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        Py_SIZE(list) = len+1;
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname);

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* CythonFunction.proto */
#define __Pyx_CyFunction_USED 1
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#define __Pyx_CyFunction_GetClassObj(f)\
    (((__pyx_CyFunctionObject *) (f))->func_classobj)
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
    PyCFunctionObject func;
#if PY_VERSION_HEX < 0x030500A0
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
    PyObject *func_classobj;
    void *defaults;
    int defaults_pyobjects;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
} __pyx_CyFunctionObject;
static PyTypeObject *__pyx_CyFunctionType = 0;
#define __Pyx_CyFunction_Check(obj)  (__Pyx_TypeCheck(obj, __pyx_CyFunctionType))
#define __Pyx_CyFunction_NewEx(ml, flags, qualname, self, module, globals, code)\
    __Pyx_CyFunction_New(__pyx_CyFunctionType, ml, flags, qualname, self, module, globals, code)
static PyObject *__Pyx_CyFunction_New(PyTypeObject *, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *self,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(void);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* UnpackTupleError.proto */
static void __Pyx_UnpackTupleError(PyObject *, Py_ssize_t index);

/* UnpackTuple2.proto */
#define __Pyx_unpack_tuple2(tuple, value1, value2, is_tuple, has_known_size, decref_tuple)\
    (likely(is_tuple || PyTuple_Check(tuple)) ?\
        (likely(has_known_size || PyTuple_GET_SIZE(tuple) == 2) ?\
            __Pyx_unpack_tuple2_exact(tuple, value1, value2, decref_tuple) :\
            (__Pyx_UnpackTupleError(tuple, 2), -1)) :\
        __Pyx_unpack_tuple2_generic(tuple, value1, value2, has_known_size, decref_tuple))
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
    PyObject* tuple, PyObject** value1, PyObject** value2, int decref_tuple);
static int __Pyx_unpack_tuple2_generic(
    PyObject* tuple, PyObject** value1, PyObject** value2, int has_known_size, int decref_tuple);

/* dict_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* dict, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_is_dict);
static CYTHON_INLINE int __Pyx_dict_iter_next(PyObject* dict_or_iter, Py_ssize_t orig_length, Py_ssize_t* ppos,
                                              PyObject** pkey, PyObject** pvalue, PyObject** pitem, int is_dict);

/* PyObjectLookupSpecial.proto */
#if CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_LookupSpecial(PyObject* obj, PyObject* attr_name) {
    PyObject *res;
    PyTypeObject *tp = Py_TYPE(obj);
#if PY_MAJOR_VERSION < 3
    if (unlikely(PyInstance_Check(obj)))
        return __Pyx_PyObject_GetAttrStr(obj, attr_name);
#endif
    res = _PyType_Lookup(tp, attr_name);
    if (likely(res)) {
        descrgetfunc f = Py_TYPE(res)->tp_descr_get;
        if (!f) {
            Py_INCREF(res);
        } else {
            res = f(res, obj, (PyObject *)tp);
        }
    } else {
        PyErr_SetObject(PyExc_AttributeError, attr_name);
    }
    return res;
}
#else
#define __Pyx_PyObject_LookupSpecial(o,n) __Pyx_PyObject_GetAttrStr(o,n)
#endif

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname);

/* pyfrozenset_new.proto */
static CYTHON_INLINE PyObject* __Pyx_PyFrozenSet_New(PyObject* it);

/* PySetContains.proto */
static CYTHON_INLINE int __Pyx_PySet_ContainsTF(PyObject* key, PyObject* set, int eq);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* dict_getitem_default.proto */
static PyObject* __Pyx_PyDict_GetItemDefault(PyObject* d, PyObject* key, PyObject* default_value);

/* UnpackUnboundCMethod.proto */
typedef struct {
    PyObject *type;
    PyObject **method_name;
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;

/* CallUnboundCMethod1.proto */
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#else
#define __Pyx_CallUnboundCMethod1(cfunc, self, arg)  __Pyx__CallUnboundCMethod1(cfunc, self, arg)
#endif

/* CallUnboundCMethod2.proto */
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030600B1
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2);
#else
#define __Pyx_CallUnboundCMethod2(cfunc, self, arg1, arg2)  __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2)
#endif

/* py_dict_keys.proto */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Keys(PyObject* d);

/* CallUnboundCMethod0.proto */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CallUnboundCMethod0(cfunc, self)\
    (likely((cfunc)->func) ?\
        (likely((cfunc)->flag == METH_NOARGS) ?  (*((cfunc)->func))(self, NULL) :\
         (PY_VERSION_HEX >= 0x030600B1 && likely((cfunc)->flag == METH_FASTCALL) ?\
            (PY_VERSION_HEX >= 0x030700A0 ?\
                (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0) :\
                (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL)) :\
          (PY_VERSION_HEX >= 0x030700A0 && (cfunc)->flag == (METH_FASTCALL | METH_KEYWORDS) ?\
            (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL) :\
            (likely((cfunc)->flag == (METH_VARARGS | METH_KEYWORDS)) ?  ((*(PyCFunctionWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, __pyx_empty_tuple, NULL)) :\
               ((cfunc)->flag == METH_VARARGS ?  (*((cfunc)->func))(self, __pyx_empty_tuple) :\
               __Pyx__CallUnboundCMethod0(cfunc, self)))))) :\
        __Pyx__CallUnboundCMethod0(cfunc, self))
#else
#define __Pyx_CallUnboundCMethod0(cfunc, self)  __Pyx__CallUnboundCMethod0(cfunc, self)
#endif

/* pyobject_as_double.proto */
static double __Pyx__PyObject_AsDouble(PyObject* obj);
#if CYTHON_COMPILING_IN_PYPY
#define __Pyx_PyObject_AsDouble(obj)\
(likely(PyFloat_CheckExact(obj)) ? PyFloat_AS_DOUBLE(obj) :\
 likely(PyInt_CheckExact(obj)) ?\
 PyFloat_AsDouble(obj) : __Pyx__PyObject_AsDouble(obj))
#else
#define __Pyx_PyObject_AsDouble(obj)\
((likely(PyFloat_CheckExact(obj))) ?\
 PyFloat_AS_DOUBLE(obj) : __Pyx__PyObject_AsDouble(obj))
#endif

/* PyIntCompare.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_NeObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* PyUnicode_Unicode.proto */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_Unicode(PyObject *obj);

/* BuildPyUnicode.proto */
static PyObject* __Pyx_PyUnicode_BuildFromAscii(Py_ssize_t ulength, char* chars, int clength,
                                                int prepend_sign, char padding_char);

/* CIntToPyUnicode.proto */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_From_int(int value, Py_ssize_t width, char padding_char, char format_char);

/* JoinPyUnicode.proto */
static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* CoroutineBase.proto */
typedef PyObject *(*__pyx_coroutine_body_t)(PyObject *, PyThreadState *, PyObject *);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_ExcInfoStruct  _PyErr_StackItem
#else
typedef struct {
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
} __Pyx_ExcInfoStruct;
#endif
typedef struct {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    __Pyx_ExcInfoStruct gi_exc_state;
    PyObject *gi_weakreflist;
    PyObject *classobj;
    PyObject *yieldfrom;
    PyObject *gi_name;
    PyObject *gi_qualname;
    PyObject *gi_modulename;
    PyObject *gi_code;
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
    PyObject *name, PyObject *qualname, PyObject *module_name);
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name);
static CYTHON_INLINE void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *self);
static int __Pyx_Coroutine_clear(PyObject *self);
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static PyObject *__Pyx_Coroutine_Close(PyObject *self);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_Coroutine_SwapException(self)
#define __Pyx_Coroutine_ResetAndClearException(self)  __Pyx_Coroutine_ExceptionClear(&(self)->gi_exc_state)
#else
#define __Pyx_Coroutine_SwapException(self) {\
    __Pyx_ExceptionSwap(&(self)->gi_exc_state.exc_type, &(self)->gi_exc_state.exc_value, &(self)->gi_exc_state.exc_traceback);\
    __Pyx_Coroutine_ResetFrameBackpointer(&(self)->gi_exc_state);\
    }
#define __Pyx_Coroutine_ResetAndClearException(self) {\
    __Pyx_ExceptionReset((self)->gi_exc_state.exc_type, (self)->gi_exc_state.exc_value, (self)->gi_exc_state.exc_traceback);\
    (self)->gi_exc_state.exc_type = (self)->gi_exc_state.exc_value = (self)->gi_exc_state.exc_traceback = NULL;\
    }
#endif
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__pyx_tstate, pvalue)
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, pvalue)
#endif
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *tstate, PyObject **pvalue);
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state);

/* PatchModuleWithCoroutine.proto */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code);

/* PatchGeneratorABC.proto */
static int __Pyx_patch_abc(void);

/* Generator.proto */
#define __Pyx_Generator_USED
static PyTypeObject *__pyx_GeneratorType = 0;
#define __Pyx_Generator_CheckExact(obj) (Py_TYPE(obj) == __pyx_GeneratorType)
#define __Pyx_Generator_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_GeneratorType, body, code, closure, name, qualname, module_name)
static PyObject *__Pyx_Generator_Next(PyObject *self);
static int __pyx_Generator_init(void);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);


/* Module declarations from 'validr._validator_c' */
static PyTypeObject *__pyx_ptype_6validr_12_validator_c___pyx_scope_struct__validator = 0;
static PyTypeObject *__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_1_decorator = 0;
static PyTypeObject *__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_2__m_validator = 0;
static PyTypeObject *__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_3__key_func_of_schema = 0;
static PyTypeObject *__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_4_key_of = 0;
static PyTypeObject *__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_5_genexpr = 0;
static PyTypeObject *__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_6_key_of = 0;
static PyTypeObject *__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_7_genexpr = 0;
static PyTypeObject *__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_8_list_validator = 0;
static PyTypeObject *__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_9_dict_validator = 0;
static PyTypeObject *__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_10_model_validator = 0;
static PyTypeObject *__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_11_enum_validator = 0;
static PyTypeObject *__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_12__union_list_validator = 0;
static PyTypeObject *__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_13__union_dict_validator = 0;
static PyTypeObject *__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_14_int_validator = 0;
static PyTypeObject *__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_15_float_validator = 0;
static PyTypeObject *__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_16_str_validator = 0;
static PyTypeObject *__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_17_bytes_validator = 0;
static PyTypeObject *__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_18_date_validator = 0;
static PyTypeObject *__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_19_time_validator = 0;
static PyTypeObject *__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_20_datetime_validator = 0;
static PyTypeObject *__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_21_timedelta_validator = 0;
static PyTypeObject *__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_22_ipv4_validator = 0;
static PyTypeObject *__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_23_ipv6_validator = 0;
static PyTypeObject *__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_24_email_validator = 0;
static PyTypeObject *__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_25_url_validator = 0;
static PyTypeObject *__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_26_uuid_validator = 0;
static PyTypeObject *__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_27_create_re_validator = 0;
static PyTypeObject *__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_28_re_validator = 0;
static PyTypeObject *__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_29_create_enum_validator = 0;
static PyObject *__pyx_v_6validr_12_validator_c__UNIQUE_CHECK_ERROR_MESSAGE = 0;
static int __pyx_f_6validr_12_validator_c_is_dict(PyObject *, int __pyx_skip_dispatch); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_6validr_12_validator_c_get_dict_value(PyObject *, PyObject *, int __pyx_skip_dispatch); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_6validr_12_validator_c_get_object_value(PyObject *, PyObject *, int __pyx_skip_dispatch); /*proto*/
static CYTHON_INLINE int __pyx_f_6validr_12_validator_c__is_empty(PyObject *, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_6validr_12_validator_c__update_validate_func_info(PyObject *, PyObject *, PyObject *, int __pyx_skip_dispatch); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_6validr_12_validator_c__key_of_scalar(PyObject *, int __pyx_skip_dispatch); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_6validr_12_validator_c__slim_dict(PyObject *, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_6validr_12_validator_c__dump_enum_value(PyObject *, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_6validr_12_validator_c_union_validator(PyObject *, PyObject *, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_6validr_12_validator_c__optional_or_has_default(PyObject *, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_6validr_12_validator_c_any_validate(PyObject *, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_6validr_12_validator_c__parse_timedelta(PyObject *, int __pyx_skip_dispatch); /*proto*/
#define __Pyx_MODULE_NAME "validr._validator_c"
extern int __pyx_module_is_main_validr___validator_c;
int __pyx_module_is_main_validr___validator_c = 0;

/* Implementation of 'validr._validator_c' */
static PyObject *__pyx_builtin_object;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_enumerate;
static PyObject *__pyx_builtin_map;
static PyObject *__pyx_builtin_AttributeError;
static const char __pyx_k_[] = "";
static const char __pyx_k_0[] = "0";
static const char __pyx_k_1[] = "1";
static const char __pyx_k_N[] = "N";
static const char __pyx_k_Y[] = "Y";
static const char __pyx_k_d[] = "d";
static const char __pyx_k_e[] = "e";
static const char __pyx_k_f[] = "f";
static const char __pyx_k_i[] = "i";
static const char __pyx_k_k[] = "k";
static const char __pyx_k_n[] = "n";
static const char __pyx_k_r[] = "r";
static const char __pyx_k_v[] = "v";
static const char __pyx_k_x[] = "x";
static const char __pyx_k_y[] = "y";
static const char __pyx_k_34[] = "&#34;";
static const char __pyx_k_39[] = "&#39;";
static const char __pyx_k_NO[] = "NO";
static const char __pyx_k_No[] = "No";
static const char __pyx_k_ON[] = "ON";
static const char __pyx_k_On[] = "On";
static const char __pyx_k__2[] = "{}<{}>";
static const char __pyx_k_by[] = "by";
static const char __pyx_k_ex[] = "ex";
static const char __pyx_k_gt[] = "&gt;";
static const char __pyx_k_lt[] = "&lt;";
static const char __pyx_k_no[] = "no";
static const char __pyx_k_on[] = "on";
static const char __pyx_k_re[] = "re";
static const char __pyx_k_Any[] = "Any";
static const char __pyx_k_OFF[] = "OFF";
static const char __pyx_k_Off[] = "Off";
static const char __pyx_k_YES[] = "YES";
static const char __pyx_k_Yes[] = "Yes";
static const char __pyx_k__39[] = "{";
  static const char __pyx_k__40[] = ", ";
static const char __pyx_k__41[] = "}";
static const char __pyx_k__65[] = "&";
static const char __pyx_k__67[] = ">";
static const char __pyx_k__69[] = "<";
static const char __pyx_k__71[] = "'";
static const char __pyx_k__73[] = "\"";
static const char __pyx_k_amp[] = "&amp;";
static const char __pyx_k_any[] = "any";
static const char __pyx_k_doc[] = "__doc__";
static const char __pyx_k_get[] = "get";
static const char __pyx_k_int[] = "int";
static const char __pyx_k_key[] = "key";
static const char __pyx_k_map[] = "map";
static const char __pyx_k_max[] = "max";
static const char __pyx_k_min[] = "min";
static const char __pyx_k_msg[] = "msg";
static const char __pyx_k_obj[] = "obj";
static const char __pyx_k_off[] = "off";
static const char __pyx_k_pop[] = "pop";
static const char __pyx_k_s_Z[] = "(?:%s)\\Z";
static const char __pyx_k_str[] = "str";
static const char __pyx_k_sys[] = "sys";
static const char __pyx_k_url[] = "url";
static const char __pyx_k_yes[] = "yes";
static const char __pyx_k_FQDN[] = "FQDN";
static const char __pyx_k_List[] = "List";
static const char __pyx_k_None[] = "None";
static const char __pyx_k_TRUE[] = "TRUE";
static const char __pyx_k_True[] = "True";
static const char __pyx_k_UUID[] = "UUID";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_bool[] = "bool";
static const char __pyx_k_copy[] = "copy";
static const char __pyx_k_date[] = "date";
static const char __pyx_k_desc[] = "desc";
static const char __pyx_k_dict[] = "dict";
static const char __pyx_k_enum[] = "enum";
static const char __pyx_k_exit[] = "__exit__";
static const char __pyx_k_fqdn[] = "fqdn";
static const char __pyx_k_hint[] = "hint";
static const char __pyx_k_ipv4[] = "ipv4";
static const char __pyx_k_ipv6[] = "ipv6";
static const char __pyx_k_keys[] = "keys";
static const char __pyx_k_list[] = "list";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_null[] = "null";
static const char __pyx_k_repr[] = "repr";
static const char __pyx_k_send[] = "send";
static const char __pyx_k_slim[] = "slim";
static const char __pyx_k_slug[] = "slug";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_time[] = "time";
static const char __pyx_k_true[] = "true";
static const char __pyx_k_uuid[] = "uuid";
static const char __pyx_k_FALSE[] = "FALSE";
static const char __pyx_k_False[] = "False";
static const char __pyx_k_H_M_S[] = "%H:%M:%S";
static const char __pyx_k_Union[] = "Union";
static const char __pyx_k_Y_m_d[] = "%Y-%m-%d";
static const char __pyx_k_bytes[] = "bytes";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_email[] = "email";
static const char __pyx_k_enter[] = "__enter__";
static const char __pyx_k_exmax[] = "exmax";
static const char __pyx_k_exmin[] = "exmin";
static const char __pyx_k_false[] = "false";
static const char __pyx_k_float[] = "float";
static const char __pyx_k_hints[] = "hints";
static const char __pyx_k_inner[] = "inner";
static const char __pyx_k_items[] = "items";
static const char __pyx_k_match[] = "match";
static const char __pyx_k_model[] = "model";
static const char __pyx_k_phone[] = "phone";
static const char __pyx_k_strip[] = "strip";
static const char __pyx_k_throw[] = "throw";
static const char __pyx_k_types[] = "types";
static const char __pyx_k_union[] = "union";
static const char __pyx_k_value[] = "value";
static const char __pyx_k_accept[] = "accept";
static const char __pyx_k_asdict[] = "__asdict__";
static const char __pyx_k_escape[] = "escape";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_getter[] = "getter";
static const char __pyx_k_idcard[] = "idcard";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_inners[] = "inners";
static const char __pyx_k_key_of[] = "key_of";
static const char __pyx_k_length[] = "length";
static const char __pyx_k_maxlen[] = "maxlen";
static const char __pyx_k_minlen[] = "minlen";
static const char __pyx_k_mktime[] = "mktime";
static const char __pyx_k_module[] = "__module__";
static const char __pyx_k_name_2[] = "name";
static const char __pyx_k_object[] = "object";
static const char __pyx_k_output[] = "output";
static const char __pyx_k_params[] = "params";
static const char __pyx_k_parsed[] = "parsed";
static const char __pyx_k_prefix[] = "prefix";
static const char __pyx_k_regexs[] = "regexs";
static const char __pyx_k_result[] = "result";
static const char __pyx_k_return[] = "return";
static const char __pyx_k_schema[] = "__schema__";
static const char __pyx_k_scheme[] = "scheme";
static const char __pyx_k_string[] = "string";
static const char __pyx_k_to_str[] = "to_str";
static const char __pyx_k_typing[] = "typing";
static const char __pyx_k_unique[] = "unique";
static const char __pyx_k_vendor[] = "_vendor";
static const char __pyx_k_Invalid[] = "Invalid";
static const char __pyx_k_Mapping[] = "Mapping";
static const char __pyx_k_by_name[] = "by_name";
static const char __pyx_k_compile[] = "compile";
static const char __pyx_k_default[] = "default";
static const char __pyx_k_err_msg[] = "err_msg";
static const char __pyx_k_expects[] = "expects";
static const char __pyx_k_genexpr[] = "genexpr";
static const char __pyx_k_getitem[] = "__getitem__";
static const char __pyx_k_is_null[] = "is_null";
static const char __pyx_k_maxsize[] = "maxsize";
static const char __pyx_k_message[] = "message";
static const char __pyx_k_options[] = "options";
static const char __pyx_k_partial[] = "partial";
static const char __pyx_k_replace[] = "replace";
static const char __pyx_k_seconds[] = "seconds";
static const char __pyx_k_version[] = "version";
static const char __pyx_k_Iterable[] = "Iterable";
static const char __pyx_k_Optional[] = "Optional";
static const char __pyx_k_compiler[] = "compiler";
static const char __pyx_k_datetime[] = "datetime";
static const char __pyx_k_extended[] = "extended";
static const char __pyx_k_fqdn_obj[] = "fqdn_obj";
static const char __pyx_k_from_str[] = "from_str";
static const char __pyx_k_is_model[] = "is_model";
static const char __pyx_k_is_valid[] = "is_valid";
static const char __pyx_k_mark_key[] = "mark_key";
static const char __pyx_k_max_repr[] = "max_repr";
static const char __pyx_k_min_repr[] = "min_repr";
static const char __pyx_k_not_list[] = "not list";
static const char __pyx_k_optional[] = "optional";
static const char __pyx_k_qualname[] = "__qualname__";
static const char __pyx_k_re_match[] = "re_match";
static const char __pyx_k_relative[] = "relative";
static const char __pyx_k_required[] = "required";
static const char __pyx_k_schema_2[] = "schema";
static const char __pyx_k_self_key[] = "$self_key";
static const char __pyx_k_strftime[] = "strftime";
static const char __pyx_k_strptime[] = "strptime";
static const char __pyx_k_urlparse[] = "urlparse";
static const char __pyx_k_validate[] = "validate";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_decorator[] = "decorator";
static const char __pyx_k_enumerate[] = "enumerate";
static const char __pyx_k_exception[] = "exception";
static const char __pyx_k_functools[] = "functools";
static const char __pyx_k_invalid_s[] = "invalid %s";
static const char __pyx_k_ipaddress[] = "ipaddress";
static const char __pyx_k_is_object[] = "is_object";
static const char __pyx_k_is_string[] = "is_string";
static const char __pyx_k_max_value[] = "max_value";
static const char __pyx_k_min_value[] = "min_value";
static const char __pyx_k_set_value[] = "set_value";
static const char __pyx_k_timedelta[] = "timedelta";
static const char __pyx_k_validator[] = "validator";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_compressed[] = "compressed";
static const char __pyx_k_dict_inner[] = "dict_inner";
static const char __pyx_k_durationpy[] = "durationpy";
static const char __pyx_k_expect_bys[] = "expect_bys";
static const char __pyx_k_extra_keys[] = "extra_keys";
static const char __pyx_k_float_info[] = "float_info";
static const char __pyx_k_http_https[] = "http https";
static const char __pyx_k_inner_info[] = "inner_info";
static const char __pyx_k_invalid_to[] = "invalid_to";
static const char __pyx_k_is_dynamic[] = "is_dynamic";
static const char __pyx_k_list_inner[] = "list_inner";
static const char __pyx_k_m_validate[] = "_m_validate";
static const char __pyx_k_mark_index[] = "mark_index";
static const char __pyx_k_not_unique[] = "not unique";
static const char __pyx_k_self_value[] = "$self_value";
static const char __pyx_k_urlunparse[] = "urlunparse";
static const char __pyx_k_validate_2[] = "_validate";
static const char __pyx_k_value_must[] = "value must >= {}";
static const char __pyx_k_IPv4Address[] = "IPv4Address";
static const char __pyx_k_IPv6Address[] = "IPv6Address";
static const char __pyx_k_SchemaError[] = "SchemaError";
static const char __pyx_k_TRUE_VALUES[] = "_TRUE_VALUES";
static const char __pyx_k_annotations[] = "__annotations__";
static const char __pyx_k_has_default[] = "has_default";
static const char __pyx_k_invalid_int[] = "invalid int";
static const char __pyx_k_invalid_url[] = "invalid url";
static const char __pyx_k_length_must[] = " length must >= ";
static const char __pyx_k_m_validator[] = "_m_validator";
static const char __pyx_k_null_output[] = "null_output";
static const char __pyx_k_origin_func[] = "origin_func";
static const char __pyx_k_parse_hints[] = "_parse_hints";
static const char __pyx_k_validator_2[] = "_validator";
static const char __pyx_k_vendor_fqdn[] = "_vendor.fqdn";
static const char __pyx_k_FALSE_VALUES[] = "_FALSE_VALUES";
static const char __pyx_k_accept_hints[] = "accept_hints";
static const char __pyx_k_allow_scheme[] = "allow_scheme";
static const char __pyx_k_any_validate[] = "any_validate";
static const char __pyx_k_inner_schema[] = "inner_schema";
static const char __pyx_k_invalid_bool[] = "invalid bool";
static const char __pyx_k_invalid_date[] = "invalid date";
static const char __pyx_k_invalid_fqdn[] = "invalid fqdn";
static const char __pyx_k_invalid_time[] = "invalid time";
static const char __pyx_k_invalid_uuid[] = "invalid uuid";
static const char __pyx_k_key_of_value[] = "key_of_value";
static const char __pyx_k_m_validate_2[] = "m_validate";
static const char __pyx_k_must_be_dict[] = "must be dict";
static const char __pyx_k_output_hints[] = "output_hints";
static const char __pyx_k_re_validator[] = "re_validator";
static const char __pyx_k_scalar_inner[] = "scalar_inner";
static const char __pyx_k_urllib_parse[] = "urllib.parse";
static const char __pyx_k_value_must_2[] = "value must <= {}";
static const char __pyx_k_value_must_d[] = "value must >= %d";
static const char __pyx_k_value_typing[] = "value_typing";
static const char __pyx_k_DurationError[] = "DurationError";
static const char __pyx_k_a_z0_9_a_z0_9[] = "[a-z0-9]+(?:-[a-z0-9]+)*";
static const char __pyx_k_accept_object[] = "accept_object";
static const char __pyx_k_accept_string[] = "accept_string";
static const char __pyx_k_annotations_2[] = "annotations";
static const char __pyx_k_any_validator[] = "any_validator";
static const char __pyx_k_expect_one_of[] = "expect one of {}";
static const char __pyx_k_fromtimestamp[] = "fromtimestamp";
static const char __pyx_k_ignore_kwargs[] = "ignore_kwargs";
static const char __pyx_k_int_validator[] = "int_validator";
static const char __pyx_k_invalid_bytes[] = "invalid bytes";
static const char __pyx_k_invalid_float[] = "invalid float";
static const char __pyx_k_key_of_scalar[] = "_key_of_scalar";
static const char __pyx_k_length_must_2[] = " length must <= ";
static const char __pyx_k_m_validator_2[] = "m_validator";
static const char __pyx_k_output_object[] = "output_object";
static const char __pyx_k_output_string[] = "output_string";
static const char __pyx_k_return_typing[] = "return_typing";
static const char __pyx_k_str_validator[] = "str_validator";
static const char __pyx_k_total_seconds[] = "total_seconds";
static const char __pyx_k_url_validator[] = "url_validator";
static const char __pyx_k_validate_func[] = "validate_func";
static const char __pyx_k_AttributeError[] = "AttributeError";
static const char __pyx_k_allow_smtputf8[] = "allow_smtputf8";
static const char __pyx_k_bool_validator[] = "bool_validator";
static const char __pyx_k_date_validator[] = "date_validator";
static const char __pyx_k_default_maxlen[] = "default_maxlen";
static const char __pyx_k_dict_validator[] = "dict_validator";
static const char __pyx_k_enum_validator[] = "enum_validator";
static const char __pyx_k_fqdn_validator[] = "fqdn_validator";
static const char __pyx_k_has_invalid_to[] = "has_invalid_to";
static const char __pyx_k_invalid_string[] = "invalid string";
static const char __pyx_k_invalid_uuid_2[] = "invalid uuid{}";
static const char __pyx_k_ipv4_validator[] = "ipv4_validator";
static const char __pyx_k_ipv6_validator[] = "ipv6_validator";
static const char __pyx_k_list_validator[] = "list_validator";
static const char __pyx_k_time_validator[] = "time_validator";
static const char __pyx_k_uuid_validator[] = "uuid_validator";
static const char __pyx_k_validate_email[] = "validate_email";
static const char __pyx_k_value_must_d_2[] = "value must <= %d";
static const char __pyx_k_value_must_d_3[] = "value must > %d";
static const char __pyx_k_value_must_d_4[] = "value must < %d";
static const char __pyx_k_Y_m_dT_H_M_S_fZ[] = "%Y-%m-%dT%H:%M:%S.%fZ";
static const char __pyx_k_bytes_validator[] = "bytes_validator";
static const char __pyx_k_d_17_d_x_X_d_15[] = "(\\d{17}[\\d|x|X])|(\\d{15})";
static const char __pyx_k_dump_enum_value[] = "_dump_enum_value";
static const char __pyx_k_email_validator[] = "email_validator";
static const char __pyx_k_float_validator[] = "float_validator";
static const char __pyx_k_model_validator[] = "model_validator";
static const char __pyx_k_no_accept_hints[] = "no accept_hints";
static const char __pyx_k_no_output_hints[] = "no output_hints";
static const char __pyx_k_supress_invalid[] = "supress_invalid";
static const char __pyx_k_union_validator[] = "union_validator";
static const char __pyx_k_url_length_must[] = "url length must <= {}";
static const char __pyx_k_invalid_datetime[] = "invalid datetime";
static const char __pyx_k_not_allowed_dict[] = "not allowed dict";
static const char __pyx_k_not_allowed_list[] = "not allowed list";
static const char __pyx_k_AddressValueError[] = "AddressValueError";
static const char __pyx_k_allow_empty_local[] = "allow_empty_local";
static const char __pyx_k_invalid_timedelta[] = "invalid timedelta";
static const char __pyx_k_EmailNotValidError[] = "EmailNotValidError";
static const char __pyx_k_builtin_validators[] = "builtin_validators";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_d_2_s_d_2_s_1_d_10[] = "((\\+\\d{2}\\s?)|(\\d{2}\\s?))?1\\d{10}";
static const char __pyx_k_datetime_validator[] = "datetime_validator";
static const char __pyx_k_dict_length_must_d[] = "dict length must <= %d";
static const char __pyx_k_invalid_to_default[] = "invalid_to_default";
static const char __pyx_k_key_func_of_schema[] = "_key_func_of_schema";
static const char __pyx_k_list_length_must_d[] = "list length must <= %d";
static const char __pyx_k_validate_extra_key[] = "validate_extra_key";
static const char __pyx_k_bytes_length_must_d[] = "bytes length must >= %d";
static const char __pyx_k_create_re_validator[] = "create_re_validator";
static const char __pyx_k_local_output_object[] = "local_output_object";
static const char __pyx_k_timedelta_validator[] = "timedelta_validator";
static const char __pyx_k_validr__validator_c[] = "validr._validator_c";
static const char __pyx_k_check_deliverability[] = "check_deliverability";
static const char __pyx_k_dict_length_must_d_2[] = "dict length must >= %d";
static const char __pyx_k_invalid_ipv4_address[] = "invalid ipv4 address";
static const char __pyx_k_invalid_ipv6_address[] = "invalid ipv6 address";
static const char __pyx_k_list_length_must_d_2[] = "list length must >= %d";
static const char __pyx_k_require_string_value[] = "require string value";
static const char __pyx_k_string_length_must_d[] = "string length must >= %d";
static const char __pyx_k_union_dict_validator[] = "_union_dict_validator";
static const char __pyx_k_union_list_validator[] = "_union_list_validator";
static const char __pyx_k_validate_extra_value[] = "validate_extra_value";
static const char __pyx_k_value_must_be_string[] = "value must be string";
static const char __pyx_k_bytes_length_must_d_2[] = "bytes length must <= %d";
static const char __pyx_k_create_enum_validator[] = "create_enum_validator";
static const char __pyx_k_invalid_email_address[] = "invalid email address";
static const char __pyx_k_invalid_expect_one_of[] = "invalid {}, expect one of {}";
static const char __pyx_k_invalid_max_timedelta[] = "invalid max timedelta";
static const char __pyx_k_invalid_min_timedelta[] = "invalid min timedelta";
static const char __pyx_k_ambiguous_union_schema[] = "ambiguous union schema";
static const char __pyx_k_illegal_version_number[] = "illegal version number";
static const char __pyx_k_string_length_must_d_2[] = "string length must <= %d";
static const char __pyx_k_vendor_email_validator[] = "_vendor.email_validator";
static const char __pyx_k_enum_items_not_provided[] = "enum items not provided";
static const char __pyx_k_invalid_accept_argument[] = "invalid accept argument {}";
static const char __pyx_k_invalid_default_value_r[] = "invalid default value {!r}";
static const char __pyx_k_invalid_output_argument[] = "invalid output argument {}";
static const char __pyx_k_model_class_not_provided[] = "model class not provided";
static const char __pyx_k_not_allowed_scalar_value[] = "not allowed scalar value";
static const char __pyx_k_string_not_match_regex_s[] = "string not match regex %s";
static const char __pyx_k_invalid_url_scheme_expect[] = "invalid url scheme, expect {}";
static const char __pyx_k_invalid_invalid_to_value_r[] = "invalid invalid_to value {!r}";
static const char __pyx_k_union_schemas_not_provided[] = "union schemas not provided";
static const char __pyx_k_union_schemas_type_invalid[] = "union schemas type invalid";
static const char __pyx_k_validator_locals_decorator[] = "validator.<locals>.decorator";
static const char __pyx_k_src_validr__validator_c_pyx[] = "src/validr/_validator_c.pyx";
static const char __pyx_k_match_regex_s_compile_failed[] = "match regex %s compile failed";
static const char __pyx_k_must_be_dict_or_model_schema[] = "must be dict or model schema";
static const char __pyx_k_int_validator_locals_validate[] = "int_validator.<locals>.validate";
static const char __pyx_k_str_validator_locals_validate[] = "str_validator.<locals>.validate";
static const char __pyx_k_url_validator_locals_validate[] = "url_validator.<locals>.validate";
static const char __pyx_k_bool_validator_locals_validate[] = "bool_validator.<locals>.validate";
static const char __pyx_k_date_validator_locals_validate[] = "date_validator.<locals>.validate";
static const char __pyx_k_dict_validator_locals_validate[] = "dict_validator.<locals>.validate";
static const char __pyx_k_enum_validator_locals_validate[] = "enum_validator.<locals>.validate";
static const char __pyx_k_fqdn_validator_locals_validate[] = "fqdn_validator.<locals>.validate";
static const char __pyx_k_ipv4_validator_locals_validate[] = "ipv4_validator.<locals>.validate";
static const char __pyx_k_ipv6_validator_locals_validate[] = "ipv6_validator.<locals>.validate";
static const char __pyx_k_list_validator_locals_validate[] = "list_validator.<locals>.validate";
static const char __pyx_k_time_validator_locals_validate[] = "time_validator.<locals>.validate";
static const char __pyx_k_uuid_validator_locals_validate[] = "uuid_validator.<locals>.validate";
static const char __pyx_k_by_argument_must_be_str_type_fo[] = "'by' argument must be str type for union schema";
static const char __pyx_k_bytes_validator_locals_validate[] = "bytes_validator.<locals>.validate";
static const char __pyx_k_can_not_set_both_invalid_to_and[] = "can not set both invalid_to and invalid_to_default";
static const char __pyx_k_default_or_optional_must_be_set[] = "default or optional must be set when set invalid_to_default";
static const char __pyx_k_email_validator_locals_validate[] = "email_validator.<locals>.validate";
static const char __pyx_k_float_validator_locals_validate[] = "float_validator.<locals>.validate";
static const char __pyx_k_key_func_of_schema_locals_key_o[] = "_key_func_of_schema.<locals>.key_of.<locals>.genexpr";
static const char __pyx_k_model_validator_locals_validate[] = "model_validator.<locals>.validate";
static const char __pyx_k_not_allowed_optional_or_default[] = "not allowed optional or default for union schemas";
static const char __pyx_k_union_dict_validator_locals_val[] = "_union_dict_validator.<locals>.validate";
static const char __pyx_k_union_list_validator_locals_val[] = "_union_list_validator.<locals>.validate";
static const char __pyx_k_update_validate_func_type_hints[] = "_update_validate_func_type_hints";
static const char __pyx_k_can_not_output_both_string_and_o[] = "can not output both string and object";
static const char __pyx_k_create_enum_validator_locals_enu[] = "create_enum_validator.<locals>.enum_validator.<locals>.validate";
static const char __pyx_k_create_re_validator_locals_re_va[] = "create_re_validator.<locals>.re_validator.<locals>.validate";
static const char __pyx_k_datetime_validator_locals_valida[] = "datetime_validator.<locals>.validate";
static const char __pyx_k_dynamic_dict_not_allowed_non_dic[] = "dynamic dict not allowed non-dict value";
static const char __pyx_k_not_allowed_by_argument_for_unio[] = "not allowed 'by' argument for union list schema";
static const char __pyx_k_not_allowed_default_for_union_sc[] = "not allowed default for union schema";
static const char __pyx_k_required_by_argument_for_union_d[] = "required 'by' argument for union dict schema";
static const char __pyx_k_timedelta_validator_locals_valid[] = "timedelta_validator.<locals>.validate";
static const char __pyx_k_unable_to_check_unique_for_non_h[] = "unable to check unique for non-hashable types";
static const char __pyx_k_validator_locals_decorator_local[] = "validator.<locals>.decorator.<locals>._m_validator.<locals>._m_validate";
static const char __pyx_k_key_func_of_schema_locals_key_o_2[] = "_key_func_of_schema.<locals>.key_of";
static const char __pyx_k_create_enum_validator_locals_enu_2[] = "create_enum_validator.<locals>.enum_validator";
static const char __pyx_k_create_re_validator_locals_re_va_2[] = "create_re_validator.<locals>.re_validator";
static const char __pyx_k_validator_locals_decorator_local_2[] = "validator.<locals>.decorator.<locals>._m_validator.<locals>.m_validate";
static const char __pyx_k_validator_locals_decorator_local_3[] = "validator.<locals>.decorator.<locals>._m_validator";
static const char __pyx_k_validator_locals_decorator_local_4[] = "validator.<locals>.decorator.<locals>.m_validator";
static PyObject *__pyx_kp_u_;
static PyObject *__pyx_kp_u_0;
static PyObject *__pyx_kp_u_1;
static PyObject *__pyx_kp_u_34;
static PyObject *__pyx_kp_u_39;
static PyObject *__pyx_n_s_AddressValueError;
static PyObject *__pyx_n_s_Any;
static PyObject *__pyx_n_s_AttributeError;
static PyObject *__pyx_n_s_DurationError;
static PyObject *__pyx_n_s_EmailNotValidError;
static PyObject *__pyx_n_u_FALSE;
static PyObject *__pyx_n_s_FALSE_VALUES;
static PyObject *__pyx_n_s_FQDN;
static PyObject *__pyx_n_u_False;
static PyObject *__pyx_kp_u_H_M_S;
static PyObject *__pyx_n_s_IPv4Address;
static PyObject *__pyx_n_s_IPv6Address;
static PyObject *__pyx_n_s_Invalid;
static PyObject *__pyx_n_s_Iterable;
static PyObject *__pyx_n_s_List;
static PyObject *__pyx_n_s_Mapping;
static PyObject *__pyx_n_u_N;
static PyObject *__pyx_n_u_NO;
static PyObject *__pyx_n_u_No;
static PyObject *__pyx_kp_u_None;
static PyObject *__pyx_n_u_OFF;
static PyObject *__pyx_n_u_ON;
static PyObject *__pyx_n_u_Off;
static PyObject *__pyx_n_u_On;
static PyObject *__pyx_n_s_Optional;
static PyObject *__pyx_n_s_SchemaError;
static PyObject *__pyx_n_u_TRUE;
static PyObject *__pyx_n_s_TRUE_VALUES;
static PyObject *__pyx_n_u_True;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_n_s_UUID;
static PyObject *__pyx_n_s_Union;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_n_u_Y;
static PyObject *__pyx_n_u_YES;
static PyObject *__pyx_kp_u_Y_m_d;
static PyObject *__pyx_kp_u_Y_m_dT_H_M_S_fZ;
static PyObject *__pyx_n_u_Yes;
static PyObject *__pyx_kp_u__2;
static PyObject *__pyx_kp_u__39;
static PyObject *__pyx_kp_u__40;
static PyObject *__pyx_kp_u__41;
static PyObject *__pyx_kp_u__65;
static PyObject *__pyx_kp_u__67;
static PyObject *__pyx_kp_u__69;
static PyObject *__pyx_kp_u__71;
static PyObject *__pyx_kp_u__73;
static PyObject *__pyx_kp_u_a_z0_9_a_z0_9;
static PyObject *__pyx_n_s_accept;
static PyObject *__pyx_n_s_accept_hints;
static PyObject *__pyx_n_s_accept_object;
static PyObject *__pyx_n_s_accept_string;
static PyObject *__pyx_n_s_allow_empty_local;
static PyObject *__pyx_n_s_allow_scheme;
static PyObject *__pyx_n_s_allow_smtputf8;
static PyObject *__pyx_kp_u_ambiguous_union_schema;
static PyObject *__pyx_kp_u_amp;
static PyObject *__pyx_n_s_annotations;
static PyObject *__pyx_n_s_annotations_2;
static PyObject *__pyx_n_u_any;
static PyObject *__pyx_n_s_any_validate;
static PyObject *__pyx_n_s_any_validator;
static PyObject *__pyx_n_s_args;
static PyObject *__pyx_n_u_asdict;
static PyObject *__pyx_n_u_bool;
static PyObject *__pyx_n_s_bool_validator;
static PyObject *__pyx_n_s_bool_validator_locals_validate;
static PyObject *__pyx_n_s_builtin_validators;
static PyObject *__pyx_n_s_by;
static PyObject *__pyx_n_u_by;
static PyObject *__pyx_kp_u_by_argument_must_be_str_type_fo;
static PyObject *__pyx_n_s_by_name;
static PyObject *__pyx_n_u_bytes;
static PyObject *__pyx_kp_u_bytes_length_must_d;
static PyObject *__pyx_kp_u_bytes_length_must_d_2;
static PyObject *__pyx_n_s_bytes_validator;
static PyObject *__pyx_n_s_bytes_validator_locals_validate;
static PyObject *__pyx_kp_u_can_not_output_both_string_and_o;
static PyObject *__pyx_kp_u_can_not_set_both_invalid_to_and;
static PyObject *__pyx_n_s_check_deliverability;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_close;
static PyObject *__pyx_n_s_compile;
static PyObject *__pyx_n_s_compiler;
static PyObject *__pyx_n_s_compressed;
static PyObject *__pyx_n_s_copy;
static PyObject *__pyx_n_s_create_enum_validator;
static PyObject *__pyx_n_s_create_enum_validator_locals_enu;
static PyObject *__pyx_n_s_create_enum_validator_locals_enu_2;
static PyObject *__pyx_n_s_create_re_validator;
static PyObject *__pyx_n_s_create_re_validator_locals_re_va;
static PyObject *__pyx_n_s_create_re_validator_locals_re_va_2;
static PyObject *__pyx_n_u_d;
static PyObject *__pyx_kp_u_d_17_d_x_X_d_15;
static PyObject *__pyx_kp_u_d_2_s_d_2_s_1_d_10;
static PyObject *__pyx_n_s_date;
static PyObject *__pyx_n_u_date;
static PyObject *__pyx_n_s_date_validator;
static PyObject *__pyx_n_s_date_validator_locals_validate;
static PyObject *__pyx_n_s_datetime;
static PyObject *__pyx_n_u_datetime;
static PyObject *__pyx_n_s_datetime_validator;
static PyObject *__pyx_n_s_datetime_validator_locals_valida;
static PyObject *__pyx_n_s_decorator;
static PyObject *__pyx_n_s_default;
static PyObject *__pyx_n_u_default;
static PyObject *__pyx_n_s_default_maxlen;
static PyObject *__pyx_kp_u_default_or_optional_must_be_set;
static PyObject *__pyx_n_s_desc;
static PyObject *__pyx_n_u_desc;
static PyObject *__pyx_n_u_dict;
static PyObject *__pyx_n_s_dict_inner;
static PyObject *__pyx_kp_u_dict_length_must_d;
static PyObject *__pyx_kp_u_dict_length_must_d_2;
static PyObject *__pyx_n_s_dict_validator;
static PyObject *__pyx_n_s_dict_validator_locals_validate;
static PyObject *__pyx_n_s_doc;
static PyObject *__pyx_n_s_dump_enum_value;
static PyObject *__pyx_n_s_durationpy;
static PyObject *__pyx_kp_u_dynamic_dict_not_allowed_non_dic;
static PyObject *__pyx_n_s_e;
static PyObject *__pyx_n_u_email;
static PyObject *__pyx_n_s_email_validator;
static PyObject *__pyx_n_s_email_validator_locals_validate;
static PyObject *__pyx_n_s_enter;
static PyObject *__pyx_n_u_enum;
static PyObject *__pyx_kp_u_enum_items_not_provided;
static PyObject *__pyx_n_s_enum_validator;
static PyObject *__pyx_n_s_enum_validator_locals_validate;
static PyObject *__pyx_n_s_enumerate;
static PyObject *__pyx_n_s_err_msg;
static PyObject *__pyx_n_s_escape;
static PyObject *__pyx_n_s_ex;
static PyObject *__pyx_n_s_exception;
static PyObject *__pyx_n_s_exit;
static PyObject *__pyx_n_s_exmax;
static PyObject *__pyx_n_s_exmin;
static PyObject *__pyx_n_s_expect_bys;
static PyObject *__pyx_kp_u_expect_one_of;
static PyObject *__pyx_n_s_expects;
static PyObject *__pyx_n_s_extended;
static PyObject *__pyx_n_s_extra_keys;
static PyObject *__pyx_n_s_f;
static PyObject *__pyx_n_u_false;
static PyObject *__pyx_n_u_float;
static PyObject *__pyx_n_s_float_info;
static PyObject *__pyx_n_s_float_validator;
static PyObject *__pyx_n_s_float_validator_locals_validate;
static PyObject *__pyx_n_s_format;
static PyObject *__pyx_n_u_fqdn;
static PyObject *__pyx_n_s_fqdn_obj;
static PyObject *__pyx_n_s_fqdn_validator;
static PyObject *__pyx_n_s_fqdn_validator_locals_validate;
static PyObject *__pyx_n_s_from_str;
static PyObject *__pyx_n_s_fromtimestamp;
static PyObject *__pyx_n_s_functools;
static PyObject *__pyx_n_s_genexpr;
static PyObject *__pyx_n_s_get;
static PyObject *__pyx_n_u_get;
static PyObject *__pyx_n_u_getitem;
static PyObject *__pyx_n_s_getter;
static PyObject *__pyx_kp_u_gt;
static PyObject *__pyx_n_s_has_default;
static PyObject *__pyx_n_s_has_invalid_to;
static PyObject *__pyx_n_s_hint;
static PyObject *__pyx_n_s_hints;
static PyObject *__pyx_kp_u_http_https;
static PyObject *__pyx_n_s_i;
static PyObject *__pyx_n_u_idcard;
static PyObject *__pyx_n_s_ignore_kwargs;
static PyObject *__pyx_kp_u_illegal_version_number;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_inner;
static PyObject *__pyx_n_s_inner_info;
static PyObject *__pyx_n_s_inner_schema;
static PyObject *__pyx_n_s_inners;
static PyObject *__pyx_n_u_int;
static PyObject *__pyx_n_s_int_validator;
static PyObject *__pyx_n_s_int_validator_locals_validate;
static PyObject *__pyx_kp_u_invalid_accept_argument;
static PyObject *__pyx_kp_u_invalid_bool;
static PyObject *__pyx_kp_u_invalid_bytes;
static PyObject *__pyx_kp_u_invalid_date;
static PyObject *__pyx_kp_u_invalid_datetime;
static PyObject *__pyx_kp_u_invalid_default_value_r;
static PyObject *__pyx_kp_u_invalid_email_address;
static PyObject *__pyx_kp_u_invalid_expect_one_of;
static PyObject *__pyx_kp_u_invalid_float;
static PyObject *__pyx_kp_u_invalid_fqdn;
static PyObject *__pyx_kp_u_invalid_int;
static PyObject *__pyx_kp_u_invalid_invalid_to_value_r;
static PyObject *__pyx_kp_u_invalid_ipv4_address;
static PyObject *__pyx_kp_u_invalid_ipv6_address;
static PyObject *__pyx_kp_u_invalid_max_timedelta;
static PyObject *__pyx_kp_u_invalid_min_timedelta;
static PyObject *__pyx_kp_u_invalid_output_argument;
static PyObject *__pyx_kp_u_invalid_s;
static PyObject *__pyx_kp_u_invalid_string;
static PyObject *__pyx_kp_u_invalid_time;
static PyObject *__pyx_kp_u_invalid_timedelta;
static PyObject *__pyx_n_s_invalid_to;
static PyObject *__pyx_n_u_invalid_to;
static PyObject *__pyx_n_s_invalid_to_default;
static PyObject *__pyx_n_u_invalid_to_default;
static PyObject *__pyx_kp_u_invalid_url;
static PyObject *__pyx_kp_u_invalid_url_scheme_expect;
static PyObject *__pyx_kp_u_invalid_uuid;
static PyObject *__pyx_kp_u_invalid_uuid_2;
static PyObject *__pyx_n_s_ipaddress;
static PyObject *__pyx_n_u_ipv4;
static PyObject *__pyx_n_s_ipv4_validator;
static PyObject *__pyx_n_s_ipv4_validator_locals_validate;
static PyObject *__pyx_n_u_ipv6;
static PyObject *__pyx_n_s_ipv6_validator;
static PyObject *__pyx_n_s_ipv6_validator_locals_validate;
static PyObject *__pyx_n_s_is_dynamic;
static PyObject *__pyx_n_s_is_model;
static PyObject *__pyx_n_s_is_null;
static PyObject *__pyx_n_s_is_object;
static PyObject *__pyx_n_s_is_string;
static PyObject *__pyx_n_s_is_valid;
static PyObject *__pyx_n_s_items;
static PyObject *__pyx_n_u_items;
static PyObject *__pyx_n_s_k;
static PyObject *__pyx_n_s_key;
static PyObject *__pyx_n_s_key_func_of_schema;
static PyObject *__pyx_n_s_key_func_of_schema_locals_key_o;
static PyObject *__pyx_n_s_key_func_of_schema_locals_key_o_2;
static PyObject *__pyx_n_s_key_of;
static PyObject *__pyx_n_s_key_of_scalar;
static PyObject *__pyx_n_s_key_of_value;
static PyObject *__pyx_n_s_keys;
static PyObject *__pyx_n_s_length;
static PyObject *__pyx_kp_u_length_must;
static PyObject *__pyx_kp_u_length_must_2;
static PyObject *__pyx_n_u_list;
static PyObject *__pyx_n_s_list_inner;
static PyObject *__pyx_kp_u_list_length_must_d;
static PyObject *__pyx_kp_u_list_length_must_d_2;
static PyObject *__pyx_n_s_list_validator;
static PyObject *__pyx_n_s_list_validator_locals_validate;
static PyObject *__pyx_n_s_local_output_object;
static PyObject *__pyx_kp_u_lt;
static PyObject *__pyx_n_s_m_validate;
static PyObject *__pyx_n_s_m_validate_2;
static PyObject *__pyx_n_s_m_validator;
static PyObject *__pyx_n_s_m_validator_2;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_map;
static PyObject *__pyx_n_s_mark_index;
static PyObject *__pyx_n_s_mark_key;
static PyObject *__pyx_n_s_match;
static PyObject *__pyx_kp_u_match_regex_s_compile_failed;
static PyObject *__pyx_n_s_max;
static PyObject *__pyx_n_s_max_repr;
static PyObject *__pyx_n_s_max_value;
static PyObject *__pyx_n_s_maxlen;
static PyObject *__pyx_n_s_maxsize;
static PyObject *__pyx_n_s_message;
static PyObject *__pyx_n_s_min;
static PyObject *__pyx_n_s_min_repr;
static PyObject *__pyx_n_s_min_value;
static PyObject *__pyx_n_s_minlen;
static PyObject *__pyx_n_s_mktime;
static PyObject *__pyx_n_u_model;
static PyObject *__pyx_kp_u_model_class_not_provided;
static PyObject *__pyx_n_s_model_validator;
static PyObject *__pyx_n_s_model_validator_locals_validate;
static PyObject *__pyx_n_s_module;
static PyObject *__pyx_n_s_msg;
static PyObject *__pyx_kp_u_must_be_dict;
static PyObject *__pyx_kp_u_must_be_dict_or_model_schema;
static PyObject *__pyx_n_u_n;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_name_2;
static PyObject *__pyx_n_u_no;
static PyObject *__pyx_kp_u_no_accept_hints;
static PyObject *__pyx_kp_u_no_output_hints;
static PyObject *__pyx_kp_u_not_allowed_by_argument_for_unio;
static PyObject *__pyx_kp_u_not_allowed_default_for_union_sc;
static PyObject *__pyx_kp_u_not_allowed_dict;
static PyObject *__pyx_kp_u_not_allowed_list;
static PyObject *__pyx_kp_u_not_allowed_optional_or_default;
static PyObject *__pyx_kp_u_not_allowed_scalar_value;
static PyObject *__pyx_kp_u_not_list;
static PyObject *__pyx_kp_u_not_unique;
static PyObject *__pyx_n_u_null;
static PyObject *__pyx_n_s_null_output;
static PyObject *__pyx_n_s_obj;
static PyObject *__pyx_n_s_object;
static PyObject *__pyx_n_u_object;
static PyObject *__pyx_n_u_off;
static PyObject *__pyx_n_u_on;
static PyObject *__pyx_n_s_optional;
static PyObject *__pyx_n_u_optional;
static PyObject *__pyx_n_s_options;
static PyObject *__pyx_n_s_origin_func;
static PyObject *__pyx_n_s_output;
static PyObject *__pyx_n_s_output_hints;
static PyObject *__pyx_n_s_output_object;
static PyObject *__pyx_n_u_output_object;
static PyObject *__pyx_n_s_output_string;
static PyObject *__pyx_n_s_params;
static PyObject *__pyx_n_s_parse_hints;
static PyObject *__pyx_n_s_parsed;
static PyObject *__pyx_n_s_partial;
static PyObject *__pyx_n_u_phone;
static PyObject *__pyx_n_s_pop;
static PyObject *__pyx_n_s_prefix;
static PyObject *__pyx_n_s_qualname;
static PyObject *__pyx_n_u_qualname;
static PyObject *__pyx_n_s_r;
static PyObject *__pyx_n_s_re;
static PyObject *__pyx_n_s_re_match;
static PyObject *__pyx_n_s_re_validator;
static PyObject *__pyx_n_s_regexs;
static PyObject *__pyx_n_s_relative;
static PyObject *__pyx_n_s_replace;
static PyObject *__pyx_n_s_repr;
static PyObject *__pyx_kp_u_require_string_value;
static PyObject *__pyx_n_u_required;
static PyObject *__pyx_kp_u_required_by_argument_for_union_d;
static PyObject *__pyx_n_s_result;
static PyObject *__pyx_n_u_return;
static PyObject *__pyx_n_s_return_typing;
static PyObject *__pyx_kp_u_s_Z;
static PyObject *__pyx_n_s_scalar_inner;
static PyObject *__pyx_n_s_schema;
static PyObject *__pyx_n_s_schema_2;
static PyObject *__pyx_n_s_scheme;
static PyObject *__pyx_n_s_seconds;
static PyObject *__pyx_kp_u_self_key;
static PyObject *__pyx_kp_u_self_value;
static PyObject *__pyx_n_s_send;
static PyObject *__pyx_n_s_set_value;
static PyObject *__pyx_n_s_slim;
static PyObject *__pyx_n_u_slug;
static PyObject *__pyx_kp_s_src_validr__validator_c_pyx;
static PyObject *__pyx_n_u_str;
static PyObject *__pyx_n_s_str_validator;
static PyObject *__pyx_n_s_str_validator_locals_validate;
static PyObject *__pyx_n_s_strftime;
static PyObject *__pyx_n_s_string;
static PyObject *__pyx_n_u_string;
static PyObject *__pyx_kp_u_string_length_must_d;
static PyObject *__pyx_kp_u_string_length_must_d_2;
static PyObject *__pyx_kp_u_string_not_match_regex_s;
static PyObject *__pyx_n_s_strip;
static PyObject *__pyx_n_s_strptime;
static PyObject *__pyx_n_s_supress_invalid;
static PyObject *__pyx_n_s_sys;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_throw;
static PyObject *__pyx_n_s_time;
static PyObject *__pyx_n_u_time;
static PyObject *__pyx_n_s_time_validator;
static PyObject *__pyx_n_s_time_validator_locals_validate;
static PyObject *__pyx_n_s_timedelta;
static PyObject *__pyx_n_u_timedelta;
static PyObject *__pyx_n_s_timedelta_validator;
static PyObject *__pyx_n_s_timedelta_validator_locals_valid;
static PyObject *__pyx_n_s_to_str;
static PyObject *__pyx_n_s_total_seconds;
static PyObject *__pyx_n_u_true;
static PyObject *__pyx_n_s_types;
static PyObject *__pyx_n_s_typing;
static PyObject *__pyx_kp_u_unable_to_check_unique_for_non_h;
static PyObject *__pyx_n_u_union;
static PyObject *__pyx_n_s_union_dict_validator;
static PyObject *__pyx_n_s_union_dict_validator_locals_val;
static PyObject *__pyx_n_s_union_list_validator;
static PyObject *__pyx_n_s_union_list_validator_locals_val;
static PyObject *__pyx_kp_u_union_schemas_not_provided;
static PyObject *__pyx_kp_u_union_schemas_type_invalid;
static PyObject *__pyx_n_s_union_validator;
static PyObject *__pyx_n_s_unique;
static PyObject *__pyx_n_s_update_validate_func_type_hints;
static PyObject *__pyx_n_u_url;
static PyObject *__pyx_kp_u_url_length_must;
static PyObject *__pyx_n_s_url_validator;
static PyObject *__pyx_n_s_url_validator_locals_validate;
static PyObject *__pyx_n_s_urllib_parse;
static PyObject *__pyx_n_s_urlparse;
static PyObject *__pyx_n_s_urlunparse;
static PyObject *__pyx_n_s_uuid;
static PyObject *__pyx_n_u_uuid;
static PyObject *__pyx_n_s_uuid_validator;
static PyObject *__pyx_n_s_uuid_validator_locals_validate;
static PyObject *__pyx_n_s_v;
static PyObject *__pyx_n_s_validate;
static PyObject *__pyx_n_s_validate_2;
static PyObject *__pyx_n_s_validate_email;
static PyObject *__pyx_n_s_validate_extra_key;
static PyObject *__pyx_n_s_validate_extra_value;
static PyObject *__pyx_n_s_validate_func;
static PyObject *__pyx_n_s_validator;
static PyObject *__pyx_n_u_validator_2;
static PyObject *__pyx_n_s_validator_locals_decorator;
static PyObject *__pyx_n_s_validator_locals_decorator_local;
static PyObject *__pyx_n_s_validator_locals_decorator_local_2;
static PyObject *__pyx_n_s_validator_locals_decorator_local_3;
static PyObject *__pyx_n_s_validator_locals_decorator_local_4;
static PyObject *__pyx_n_s_validr__validator_c;
static PyObject *__pyx_n_s_value;
static PyObject *__pyx_n_u_value;
static PyObject *__pyx_kp_u_value_must;
static PyObject *__pyx_kp_u_value_must_2;
static PyObject *__pyx_kp_u_value_must_be_string;
static PyObject *__pyx_kp_u_value_must_d;
static PyObject *__pyx_kp_u_value_must_d_2;
static PyObject *__pyx_kp_u_value_must_d_3;
static PyObject *__pyx_kp_u_value_must_d_4;
static PyObject *__pyx_n_s_value_typing;
static PyObject *__pyx_n_s_vendor;
static PyObject *__pyx_n_s_vendor_email_validator;
static PyObject *__pyx_n_s_vendor_fqdn;
static PyObject *__pyx_n_s_version;
static PyObject *__pyx_n_s_x;
static PyObject *__pyx_n_u_y;
static PyObject *__pyx_n_u_yes;
static PyObject *__pyx_pf_6validr_12_validator_c_is_dict(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_obj); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_2get_dict_value(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_obj, PyObject *__pyx_v_key); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_4get_object_value(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_obj, PyObject *__pyx_v_key); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_6_is_empty(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_8_update_validate_func_info(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_validate_func, PyObject *__pyx_v_origin_func, PyObject *__pyx_v_schema); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_10_parse_hints(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_hints); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_12_update_validate_func_type_hints(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_validate_func, PyObject *__pyx_v_optional, PyObject *__pyx_v_has_default, PyObject *__pyx_v_accept_hints, PyObject *__pyx_v_output_hints); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_9validator_9decorator_12_m_validator__m_validate(PyObject *__pyx_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_9validator_9decorator_12_m_validator_2m_validate(PyObject *__pyx_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_9validator_9decorator__m_validator(PyObject *__pyx_self, PyObject *__pyx_v_compiler, PyObject *__pyx_v_schema); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_9validator_9decorator_2m_validator(PyObject *__pyx_self, PyObject *__pyx_v_compiler, PyObject *__pyx_v_schema); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_9validator_decorator(PyObject *__pyx_self, PyObject *__pyx_v_f); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_14validator(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_string, PyObject *__pyx_v_accept, PyObject *__pyx_v_output); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_16_key_of_scalar(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_v); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_19_key_func_of_schema_6key_of_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_19_key_func_of_schema_key_of(PyObject *__pyx_self, PyObject *__pyx_v_v); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_19_key_func_of_schema_6key_of_3genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_19_key_func_of_schema_2key_of(PyObject *__pyx_self, PyObject *__pyx_v_v); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_18_key_func_of_schema(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_schema); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_14list_validator_validate(PyObject *__pyx_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_20list_validator(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_compiler, PyObject *__pyx_v_items, int __pyx_v_minlen, int __pyx_v_maxlen, int __pyx_v_unique); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_22_slim_dict(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_14dict_validator_validate(PyObject *__pyx_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_24dict_validator(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_compiler, PyObject *__pyx_v_items, PyObject *__pyx_v_key, PyObject *__pyx_v_value, int __pyx_v_minlen, int __pyx_v_maxlen, int __pyx_v_slim); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_15model_validator_validate(PyObject *__pyx_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_26model_validator(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_compiler, PyObject *__pyx_v_items); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_28_dump_enum_value(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_14enum_validator_validate(PyObject *__pyx_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_30enum_validator(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_compiler, PyObject *__pyx_v_items); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_32union_validator(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_compiler, PyObject *__pyx_v_schema); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_34_optional_or_has_default(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_schema); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_21_union_list_validator_validate(PyObject *__pyx_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_36_union_list_validator(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_compiler, PyObject *__pyx_v_schema); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_21_union_dict_validator_validate(PyObject *__pyx_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_38_union_dict_validator(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_compiler, PyObject *__pyx_v_items, PyObject *__pyx_v_by); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_40any_validate(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_42any_validator(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_compiler, CYTHON_UNUSED PyObject *__pyx_v_ignore_kwargs); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_13int_validator_validate(PyObject *__pyx_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_44int_validator(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_compiler, PyObject *__pyx_v_min, PyObject *__pyx_v_max); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_14bool_validator_validate(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_46bool_validator(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_compiler); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_15float_validator_validate(PyObject *__pyx_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_48float_validator(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_compiler, PyObject *__pyx_v_min, PyObject *__pyx_v_max, PyObject *__pyx_v_exmin, PyObject *__pyx_v_exmax); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_13str_validator_validate(PyObject *__pyx_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_50str_validator(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_compiler, int __pyx_v_minlen, int __pyx_v_maxlen, int __pyx_v_strip, int __pyx_v_escape, PyObject *__pyx_v_match, int __pyx_v_accept_object); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_15bytes_validator_validate(PyObject *__pyx_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_52bytes_validator(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_compiler, int __pyx_v_minlen, int __pyx_v_maxlen); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_14date_validator_validate(PyObject *__pyx_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_54date_validator(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_compiler, PyObject *__pyx_v_format, int __pyx_v_output_object); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_14time_validator_validate(PyObject *__pyx_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_56time_validator(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_compiler, PyObject *__pyx_v_format, int __pyx_v_output_object); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_18datetime_validator_validate(PyObject *__pyx_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_58datetime_validator(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_compiler, PyObject *__pyx_v_format, int __pyx_v_output_object); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_60_parse_timedelta(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_19timedelta_validator_validate(PyObject *__pyx_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_62timedelta_validator(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_compiler, PyObject *__pyx_v_min, PyObject *__pyx_v_max, int __pyx_v_string, int __pyx_v_extended, int __pyx_v_output_object); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_14ipv4_validator_validate(PyObject *__pyx_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_64ipv4_validator(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_compiler, int __pyx_v_output_object); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_14ipv6_validator_validate(PyObject *__pyx_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_66ipv6_validator(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_compiler, int __pyx_v_output_object); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_15email_validator_validate(PyObject *__pyx_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_68email_validator(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_compiler); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_13url_validator_validate(PyObject *__pyx_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_70url_validator(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_compiler, PyObject *__pyx_v_scheme, int __pyx_v_maxlen, int __pyx_v_output_object); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_14fqdn_validator_validate(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_72fqdn_validator(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_compiler); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_14uuid_validator_validate(PyObject *__pyx_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_74uuid_validator(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_compiler, PyObject *__pyx_v_version, int __pyx_v_output_object); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_80__defaults__(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_19create_re_validator_12re_validator_validate(PyObject *__pyx_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_19create_re_validator_re_validator(PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_compiler, int __pyx_v_minlen, int __pyx_v_maxlen, int __pyx_v_strip); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_76create_re_validator(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_name, PyObject *__pyx_v_r, int __pyx_v_default_maxlen); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_21create_enum_validator_14enum_validator_validate(PyObject *__pyx_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_21create_enum_validator_enum_validator(PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_compiler); /* proto */
static PyObject *__pyx_pf_6validr_12_validator_c_78create_enum_validator(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_name, PyObject *__pyx_v_items, int __pyx_v_string); /* proto */
static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct__validator(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_1_decorator(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_2__m_validator(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_3__key_func_of_schema(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_4_key_of(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_5_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_6_key_of(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_7_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_8_list_validator(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_9_dict_validator(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_10_model_validator(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_11_enum_validator(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_12__union_list_validator(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_13__union_dict_validator(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_14_int_validator(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_15_float_validator(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_16_str_validator(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_17_bytes_validator(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_18_date_validator(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_19_time_validator(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_20_datetime_validator(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_21_timedelta_validator(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_22_ipv4_validator(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_23_ipv6_validator(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_24_email_validator(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_25_url_validator(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_26_uuid_validator(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_27_create_re_validator(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_28_re_validator(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_29_create_enum_validator(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_get = {0, &__pyx_n_s_get, 0, 0, 0};
static __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_keys = {0, &__pyx_n_s_keys, 0, 0, 0};
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_3;
static PyObject *__pyx_int_4;
static PyObject *__pyx_int_5;
static PyObject *__pyx_int_15;
static PyObject *__pyx_int_18;
static PyObject *__pyx_int_255;
static PyObject *__pyx_k__52;
static PyObject *__pyx_k__53;
static PyObject *__pyx_k__60;
static PyObject *__pyx_k__61;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__11;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__16;
static PyObject *__pyx_tuple__18;
static PyObject *__pyx_tuple__21;
static PyObject *__pyx_tuple__24;
static PyObject *__pyx_tuple__26;
static PyObject *__pyx_tuple__28;
static PyObject *__pyx_tuple__30;
static PyObject *__pyx_tuple__33;
static PyObject *__pyx_tuple__36;
static PyObject *__pyx_tuple__42;
static PyObject *__pyx_tuple__45;
static PyObject *__pyx_tuple__47;
static PyObject *__pyx_tuple__49;
static PyObject *__pyx_tuple__55;
static PyObject *__pyx_tuple__58;
static PyObject *__pyx_tuple__63;
static PyObject *__pyx_tuple__66;
static PyObject *__pyx_tuple__68;
static PyObject *__pyx_tuple__70;
static PyObject *__pyx_tuple__72;
static PyObject *__pyx_tuple__74;
static PyObject *__pyx_tuple__76;
static PyObject *__pyx_tuple__79;
static PyObject *__pyx_tuple__82;
static PyObject *__pyx_tuple__85;
static PyObject *__pyx_tuple__88;
static PyObject *__pyx_tuple__90;
static PyObject *__pyx_tuple__92;
static PyObject *__pyx_tuple__95;
static PyObject *__pyx_tuple__98;
static PyObject *__pyx_codeobj__3;
static PyObject *__pyx_codeobj__4;
static PyObject *__pyx_tuple__101;
static PyObject *__pyx_tuple__104;
static PyObject *__pyx_tuple__107;
static PyObject *__pyx_tuple__110;
static PyObject *__pyx_tuple__112;
static PyObject *__pyx_tuple__115;
static PyObject *__pyx_tuple__117;
static PyObject *__pyx_tuple__120;
static PyObject *__pyx_tuple__122;
static PyObject *__pyx_tuple__123;
static PyObject *__pyx_tuple__124;
static PyObject *__pyx_tuple__125;
static PyObject *__pyx_tuple__126;
static PyObject *__pyx_tuple__127;
static PyObject *__pyx_tuple__128;
static PyObject *__pyx_tuple__129;
static PyObject *__pyx_tuple__130;
static PyObject *__pyx_tuple__131;
static PyObject *__pyx_tuple__132;
static PyObject *__pyx_tuple__133;
static PyObject *__pyx_tuple__134;
static PyObject *__pyx_tuple__135;
static PyObject *__pyx_tuple__136;
static PyObject *__pyx_tuple__137;
static PyObject *__pyx_tuple__138;
static PyObject *__pyx_tuple__139;
static PyObject *__pyx_tuple__140;
static PyObject *__pyx_tuple__141;
static PyObject *__pyx_tuple__142;
static PyObject *__pyx_tuple__143;
static PyObject *__pyx_tuple__144;
static PyObject *__pyx_tuple__145;
static PyObject *__pyx_tuple__146;
static PyObject *__pyx_tuple__147;
static PyObject *__pyx_tuple__148;
static PyObject *__pyx_tuple__149;
static PyObject *__pyx_tuple__150;
static PyObject *__pyx_tuple__151;
static PyObject *__pyx_tuple__152;
static PyObject *__pyx_codeobj__13;
static PyObject *__pyx_codeobj__15;
static PyObject *__pyx_codeobj__17;
static PyObject *__pyx_codeobj__19;
static PyObject *__pyx_codeobj__20;
static PyObject *__pyx_codeobj__22;
static PyObject *__pyx_codeobj__23;
static PyObject *__pyx_codeobj__25;
static PyObject *__pyx_codeobj__27;
static PyObject *__pyx_codeobj__29;
static PyObject *__pyx_codeobj__31;
static PyObject *__pyx_codeobj__32;
static PyObject *__pyx_codeobj__34;
static PyObject *__pyx_codeobj__35;
static PyObject *__pyx_codeobj__37;
static PyObject *__pyx_codeobj__38;
static PyObject *__pyx_codeobj__43;
static PyObject *__pyx_codeobj__44;
static PyObject *__pyx_codeobj__46;
static PyObject *__pyx_codeobj__48;
static PyObject *__pyx_codeobj__50;
static PyObject *__pyx_codeobj__51;
static PyObject *__pyx_codeobj__54;
static PyObject *__pyx_codeobj__56;
static PyObject *__pyx_codeobj__57;
static PyObject *__pyx_codeobj__59;
static PyObject *__pyx_codeobj__62;
static PyObject *__pyx_codeobj__64;
static PyObject *__pyx_codeobj__75;
static PyObject *__pyx_codeobj__77;
static PyObject *__pyx_codeobj__78;
static PyObject *__pyx_codeobj__80;
static PyObject *__pyx_codeobj__81;
static PyObject *__pyx_codeobj__83;
static PyObject *__pyx_codeobj__84;
static PyObject *__pyx_codeobj__86;
static PyObject *__pyx_codeobj__87;
static PyObject *__pyx_codeobj__89;
static PyObject *__pyx_codeobj__91;
static PyObject *__pyx_codeobj__93;
static PyObject *__pyx_codeobj__94;
static PyObject *__pyx_codeobj__96;
static PyObject *__pyx_codeobj__97;
static PyObject *__pyx_codeobj__99;
static PyObject *__pyx_codeobj__100;
static PyObject *__pyx_codeobj__102;
static PyObject *__pyx_codeobj__103;
static PyObject *__pyx_codeobj__105;
static PyObject *__pyx_codeobj__106;
static PyObject *__pyx_codeobj__108;
static PyObject *__pyx_codeobj__109;
static PyObject *__pyx_codeobj__111;
static PyObject *__pyx_codeobj__113;
static PyObject *__pyx_codeobj__114;
static PyObject *__pyx_codeobj__116;
static PyObject *__pyx_codeobj__118;
static PyObject *__pyx_codeobj__119;
static PyObject *__pyx_codeobj__121;
/* Late includes */

/* "validr/_validator_c.pyx":18
 * 
 * 
 * cpdef bint is_dict(obj):             # <<<<<<<<<<<<<<
 *     # use isinstance(obj, Mapping) is slow,
 *     # hasattr check can speed up about 30%
 */

static PyObject *__pyx_pw_6validr_12_validator_c_1is_dict(PyObject *__pyx_self, PyObject *__pyx_v_obj); /*proto*/
static int __pyx_f_6validr_12_validator_c_is_dict(PyObject *__pyx_v_obj, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_r;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  __Pyx_RefNannySetupContext("is_dict", 0);
  __Pyx_TraceCall("is_dict", __pyx_f[0], 18, 0, __PYX_ERR(0, 18, __pyx_L1_error));

  /* "validr/_validator_c.pyx":21
 *     # use isinstance(obj, Mapping) is slow,
 *     # hasattr check can speed up about 30%
 *     return hasattr(obj, '__getitem__') and hasattr(obj, 'get')             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(21,0,__PYX_ERR(0, 21, __pyx_L1_error))
  __pyx_t_2 = __Pyx_HasAttr(__pyx_v_obj, __pyx_n_u_getitem); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 21, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_HasAttr(__pyx_v_obj, __pyx_n_u_get); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 21, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_3 != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L3_bool_binop_done:;
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "validr/_validator_c.pyx":18
 * 
 * 
 * cpdef bint is_dict(obj):             # <<<<<<<<<<<<<<
 *     # use isinstance(obj, Mapping) is slow,
 *     # hasattr check can speed up about 30%
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("validr._validator_c.is_dict", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_1is_dict(PyObject *__pyx_self, PyObject *__pyx_v_obj); /*proto*/
static PyObject *__pyx_pw_6validr_12_validator_c_1is_dict(PyObject *__pyx_self, PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_dict (wrapper)", 0);
  __pyx_r = __pyx_pf_6validr_12_validator_c_is_dict(__pyx_self, ((PyObject *)__pyx_v_obj));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6validr_12_validator_c_is_dict(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("is_dict", 0);
  __Pyx_TraceCall("is_dict (wrapper)", __pyx_f[0], 18, 0, __PYX_ERR(0, 18, __pyx_L1_error));
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_6validr_12_validator_c_is_dict(__pyx_v_obj, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("validr._validator_c.is_dict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":24
 * 
 * 
 * cpdef inline get_dict_value(obj, str key):             # <<<<<<<<<<<<<<
 *     return obj.get(key, None)
 * 
 */

static PyObject *__pyx_pw_6validr_12_validator_c_3get_dict_value(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_6validr_12_validator_c_get_dict_value(PyObject *__pyx_v_obj, PyObject *__pyx_v_key, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("get_dict_value", 0);
  __Pyx_TraceCall("get_dict_value", __pyx_f[0], 24, 0, __PYX_ERR(0, 24, __pyx_L1_error));

  /* "validr/_validator_c.pyx":25
 * 
 * cpdef inline get_dict_value(obj, str key):
 *     return obj.get(key, None)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(25,0,__PYX_ERR(0, 25, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_get); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_key, Py_None};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 25, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_key, Py_None};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 25, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 25, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_INCREF(__pyx_v_key);
    __Pyx_GIVEREF(__pyx_v_key);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_key);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, Py_None);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 25, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "validr/_validator_c.pyx":24
 * 
 * 
 * cpdef inline get_dict_value(obj, str key):             # <<<<<<<<<<<<<<
 *     return obj.get(key, None)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("validr._validator_c.get_dict_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_3get_dict_value(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6validr_12_validator_c_3get_dict_value(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_obj = 0;
  PyObject *__pyx_v_key = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_dict_value (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_obj,&__pyx_n_s_key,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_obj)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_key)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("get_dict_value", 1, 2, 2, 1); __PYX_ERR(0, 24, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "get_dict_value") < 0)) __PYX_ERR(0, 24, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_obj = values[0];
    __pyx_v_key = ((PyObject*)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_dict_value", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 24, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("validr._validator_c.get_dict_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_key), (&PyUnicode_Type), 1, "key", 1))) __PYX_ERR(0, 24, __pyx_L1_error)
  __pyx_r = __pyx_pf_6validr_12_validator_c_2get_dict_value(__pyx_self, __pyx_v_obj, __pyx_v_key);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6validr_12_validator_c_2get_dict_value(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_obj, PyObject *__pyx_v_key) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("get_dict_value", 0);
  __Pyx_TraceCall("get_dict_value (wrapper)", __pyx_f[0], 24, 0, __PYX_ERR(0, 24, __pyx_L1_error));
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6validr_12_validator_c_get_dict_value(__pyx_v_obj, __pyx_v_key, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("validr._validator_c.get_dict_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":28
 * 
 * 
 * cpdef inline get_object_value(obj, str key):             # <<<<<<<<<<<<<<
 *     return getattr(obj, key, None)
 * 
 */

static PyObject *__pyx_pw_6validr_12_validator_c_5get_object_value(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_6validr_12_validator_c_get_object_value(PyObject *__pyx_v_obj, PyObject *__pyx_v_key, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("get_object_value", 0);
  __Pyx_TraceCall("get_object_value", __pyx_f[0], 28, 0, __PYX_ERR(0, 28, __pyx_L1_error));

  /* "validr/_validator_c.pyx":29
 * 
 * cpdef inline get_object_value(obj, str key):
 *     return getattr(obj, key, None)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(29,0,__PYX_ERR(0, 29, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetAttr3(__pyx_v_obj, __pyx_v_key, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "validr/_validator_c.pyx":28
 * 
 * 
 * cpdef inline get_object_value(obj, str key):             # <<<<<<<<<<<<<<
 *     return getattr(obj, key, None)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("validr._validator_c.get_object_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_5get_object_value(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6validr_12_validator_c_5get_object_value(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_obj = 0;
  PyObject *__pyx_v_key = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_object_value (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_obj,&__pyx_n_s_key,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_obj)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_key)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("get_object_value", 1, 2, 2, 1); __PYX_ERR(0, 28, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "get_object_value") < 0)) __PYX_ERR(0, 28, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_obj = values[0];
    __pyx_v_key = ((PyObject*)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_object_value", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 28, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("validr._validator_c.get_object_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_key), (&PyUnicode_Type), 1, "key", 1))) __PYX_ERR(0, 28, __pyx_L1_error)
  __pyx_r = __pyx_pf_6validr_12_validator_c_4get_object_value(__pyx_self, __pyx_v_obj, __pyx_v_key);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6validr_12_validator_c_4get_object_value(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_obj, PyObject *__pyx_v_key) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("get_object_value", 0);
  __Pyx_TraceCall("get_object_value (wrapper)", __pyx_f[0], 28, 0, __PYX_ERR(0, 28, __pyx_L1_error));
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6validr_12_validator_c_get_object_value(__pyx_v_obj, __pyx_v_key, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("validr._validator_c.get_object_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":32
 * 
 * 
 * cpdef inline bint _is_empty(value):             # <<<<<<<<<<<<<<
 *     return value is None or value == ''
 * 
 */

static PyObject *__pyx_pw_6validr_12_validator_c_7_is_empty(PyObject *__pyx_self, PyObject *__pyx_v_value); /*proto*/
static CYTHON_INLINE int __pyx_f_6validr_12_validator_c__is_empty(PyObject *__pyx_v_value, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_r;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  __Pyx_RefNannySetupContext("_is_empty", 0);
  __Pyx_TraceCall("_is_empty", __pyx_f[0], 32, 0, __PYX_ERR(0, 32, __pyx_L1_error));

  /* "validr/_validator_c.pyx":33
 * 
 * cpdef inline bint _is_empty(value):
 *     return value is None or value == ''             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(33,0,__PYX_ERR(0, 33, __pyx_L1_error))
  __pyx_t_2 = (__pyx_v_value == Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (!__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_3 = (__Pyx_PyUnicode_Equals(__pyx_v_value, __pyx_kp_u_, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 33, __pyx_L1_error)
  __pyx_t_1 = __pyx_t_3;
  __pyx_L3_bool_binop_done:;
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "validr/_validator_c.pyx":32
 * 
 * 
 * cpdef inline bint _is_empty(value):             # <<<<<<<<<<<<<<
 *     return value is None or value == ''
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("validr._validator_c._is_empty", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_TraceReturn(Py_None, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_7_is_empty(PyObject *__pyx_self, PyObject *__pyx_v_value); /*proto*/
static PyObject *__pyx_pw_6validr_12_validator_c_7_is_empty(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_is_empty (wrapper)", 0);
  __pyx_r = __pyx_pf_6validr_12_validator_c_6_is_empty(__pyx_self, ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6validr_12_validator_c_6_is_empty(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_is_empty", 0);
  __Pyx_TraceCall("_is_empty (wrapper)", __pyx_f[0], 32, 0, __PYX_ERR(0, 32, __pyx_L1_error));
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_6validr_12_validator_c__is_empty(__pyx_v_value, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("validr._validator_c._is_empty", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":36
 * 
 * 
 * cpdef _update_validate_func_info(validate_func, origin_func, schema):             # <<<<<<<<<<<<<<
 *     # make friendly validate func representation
 *     m_repr = schema.repr(prefix=False, desc=False)
 */

static PyObject *__pyx_pw_6validr_12_validator_c_9_update_validate_func_info(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_6validr_12_validator_c__update_validate_func_info(PyObject *__pyx_v_validate_func, PyObject *__pyx_v_origin_func, PyObject *__pyx_v_schema, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_v_m_repr = NULL;
  PyObject *__pyx_v_qualname = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("_update_validate_func_info", 0);
  __Pyx_TraceCall("_update_validate_func_info", __pyx_f[0], 36, 0, __PYX_ERR(0, 36, __pyx_L1_error));

  /* "validr/_validator_c.pyx":38
 * cpdef _update_validate_func_info(validate_func, origin_func, schema):
 *     # make friendly validate func representation
 *     m_repr = schema.repr(prefix=False, desc=False)             # <<<<<<<<<<<<<<
 *     validate_func.__schema__ = schema
 *     validate_func.__module__ = origin_func.__module__
 */
  __Pyx_TraceLine(38,0,__PYX_ERR(0, 38, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_schema, __pyx_n_s_repr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_prefix, Py_False) < 0) __PYX_ERR(0, 38, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_desc, Py_False) < 0) __PYX_ERR(0, 38, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_m_repr = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "validr/_validator_c.pyx":39
 *     # make friendly validate func representation
 *     m_repr = schema.repr(prefix=False, desc=False)
 *     validate_func.__schema__ = schema             # <<<<<<<<<<<<<<
 *     validate_func.__module__ = origin_func.__module__
 *     validate_func.__name__ = '{}<{}>'.format(origin_func.__name__, m_repr)
 */
  __Pyx_TraceLine(39,0,__PYX_ERR(0, 39, __pyx_L1_error))
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_validate_func, __pyx_n_s_schema, __pyx_v_schema) < 0) __PYX_ERR(0, 39, __pyx_L1_error)

  /* "validr/_validator_c.pyx":40
 *     m_repr = schema.repr(prefix=False, desc=False)
 *     validate_func.__schema__ = schema
 *     validate_func.__module__ = origin_func.__module__             # <<<<<<<<<<<<<<
 *     validate_func.__name__ = '{}<{}>'.format(origin_func.__name__, m_repr)
 *     if hasattr(origin_func, '__qualname__'):
 */
  __Pyx_TraceLine(40,0,__PYX_ERR(0, 40, __pyx_L1_error))
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_origin_func, __pyx_n_s_module); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_validate_func, __pyx_n_s_module, __pyx_t_3) < 0) __PYX_ERR(0, 40, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "validr/_validator_c.pyx":41
 *     validate_func.__schema__ = schema
 *     validate_func.__module__ = origin_func.__module__
 *     validate_func.__name__ = '{}<{}>'.format(origin_func.__name__, m_repr)             # <<<<<<<<<<<<<<
 *     if hasattr(origin_func, '__qualname__'):
 *         qualname = '{}<{}>'.format(origin_func.__qualname__, m_repr)
 */
  __Pyx_TraceLine(41,0,__PYX_ERR(0, 41, __pyx_L1_error))
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u__2, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_origin_func, __pyx_n_s_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_t_1, __pyx_v_m_repr};
    __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 41, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_t_1, __pyx_v_m_repr};
    __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 41, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 41, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_1);
    __Pyx_INCREF(__pyx_v_m_repr);
    __Pyx_GIVEREF(__pyx_v_m_repr);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v_m_repr);
    __pyx_t_1 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 41, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_validate_func, __pyx_n_s_name, __pyx_t_3) < 0) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "validr/_validator_c.pyx":42
 *     validate_func.__module__ = origin_func.__module__
 *     validate_func.__name__ = '{}<{}>'.format(origin_func.__name__, m_repr)
 *     if hasattr(origin_func, '__qualname__'):             # <<<<<<<<<<<<<<
 *         qualname = '{}<{}>'.format(origin_func.__qualname__, m_repr)
 *         validate_func.__qualname__ = qualname
 */
  __Pyx_TraceLine(42,0,__PYX_ERR(0, 42, __pyx_L1_error))
  __pyx_t_7 = __Pyx_HasAttr(__pyx_v_origin_func, __pyx_n_u_qualname); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 42, __pyx_L1_error)
  __pyx_t_8 = (__pyx_t_7 != 0);
  if (__pyx_t_8) {

    /* "validr/_validator_c.pyx":43
 *     validate_func.__name__ = '{}<{}>'.format(origin_func.__name__, m_repr)
 *     if hasattr(origin_func, '__qualname__'):
 *         qualname = '{}<{}>'.format(origin_func.__qualname__, m_repr)             # <<<<<<<<<<<<<<
 *         validate_func.__qualname__ = qualname
 *     if origin_func.__doc__:
 */
    __Pyx_TraceLine(43,0,__PYX_ERR(0, 43, __pyx_L1_error))
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u__2, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 43, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_origin_func, __pyx_n_s_qualname); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 43, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = NULL;
    __pyx_t_5 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_5 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_t_6, __pyx_v_m_repr};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 43, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_t_6, __pyx_v_m_repr};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 43, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 43, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (__pyx_t_1) {
        __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1); __pyx_t_1 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_5, __pyx_t_6);
      __Pyx_INCREF(__pyx_v_m_repr);
      __Pyx_GIVEREF(__pyx_v_m_repr);
      PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_5, __pyx_v_m_repr);
      __pyx_t_6 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 43, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_qualname = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "validr/_validator_c.pyx":44
 *     if hasattr(origin_func, '__qualname__'):
 *         qualname = '{}<{}>'.format(origin_func.__qualname__, m_repr)
 *         validate_func.__qualname__ = qualname             # <<<<<<<<<<<<<<
 *     if origin_func.__doc__:
 *         validate_func.__doc__ = origin_func.__doc__
 */
    __Pyx_TraceLine(44,0,__PYX_ERR(0, 44, __pyx_L1_error))
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_validate_func, __pyx_n_s_qualname, __pyx_v_qualname) < 0) __PYX_ERR(0, 44, __pyx_L1_error)

    /* "validr/_validator_c.pyx":42
 *     validate_func.__module__ = origin_func.__module__
 *     validate_func.__name__ = '{}<{}>'.format(origin_func.__name__, m_repr)
 *     if hasattr(origin_func, '__qualname__'):             # <<<<<<<<<<<<<<
 *         qualname = '{}<{}>'.format(origin_func.__qualname__, m_repr)
 *         validate_func.__qualname__ = qualname
 */
  }

  /* "validr/_validator_c.pyx":45
 *         qualname = '{}<{}>'.format(origin_func.__qualname__, m_repr)
 *         validate_func.__qualname__ = qualname
 *     if origin_func.__doc__:             # <<<<<<<<<<<<<<
 *         validate_func.__doc__ = origin_func.__doc__
 *     else:
 */
  __Pyx_TraceLine(45,0,__PYX_ERR(0, 45, __pyx_L1_error))
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_origin_func, __pyx_n_s_doc); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_8) {

    /* "validr/_validator_c.pyx":46
 *         validate_func.__qualname__ = qualname
 *     if origin_func.__doc__:
 *         validate_func.__doc__ = origin_func.__doc__             # <<<<<<<<<<<<<<
 *     else:
 *         validate_func.__doc__ = schema.params.get('desc')
 */
    __Pyx_TraceLine(46,0,__PYX_ERR(0, 46, __pyx_L1_error))
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_origin_func, __pyx_n_s_doc); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 46, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_validate_func, __pyx_n_s_doc, __pyx_t_3) < 0) __PYX_ERR(0, 46, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "validr/_validator_c.pyx":45
 *         qualname = '{}<{}>'.format(origin_func.__qualname__, m_repr)
 *         validate_func.__qualname__ = qualname
 *     if origin_func.__doc__:             # <<<<<<<<<<<<<<
 *         validate_func.__doc__ = origin_func.__doc__
 *     else:
 */
    goto __pyx_L4;
  }

  /* "validr/_validator_c.pyx":48
 *         validate_func.__doc__ = origin_func.__doc__
 *     else:
 *         validate_func.__doc__ = schema.params.get('desc')             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(48,0,__PYX_ERR(0, 48, __pyx_L1_error))
  /*else*/ {
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_schema, __pyx_n_s_params); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 48, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_get); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 48, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_2, __pyx_n_u_desc) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_n_u_desc);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 48, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_validate_func, __pyx_n_s_doc, __pyx_t_3) < 0) __PYX_ERR(0, 48, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_L4:;

  /* "validr/_validator_c.pyx":36
 * 
 * 
 * cpdef _update_validate_func_info(validate_func, origin_func, schema):             # <<<<<<<<<<<<<<
 *     # make friendly validate func representation
 *     m_repr = schema.repr(prefix=False, desc=False)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("validr._validator_c._update_validate_func_info", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_m_repr);
  __Pyx_XDECREF(__pyx_v_qualname);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_9_update_validate_func_info(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6validr_12_validator_c_9_update_validate_func_info(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_validate_func = 0;
  PyObject *__pyx_v_origin_func = 0;
  PyObject *__pyx_v_schema = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_update_validate_func_info (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_validate_func,&__pyx_n_s_origin_func,&__pyx_n_s_schema_2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_validate_func)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_origin_func)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_update_validate_func_info", 1, 3, 3, 1); __PYX_ERR(0, 36, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_schema_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_update_validate_func_info", 1, 3, 3, 2); __PYX_ERR(0, 36, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_update_validate_func_info") < 0)) __PYX_ERR(0, 36, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_validate_func = values[0];
    __pyx_v_origin_func = values[1];
    __pyx_v_schema = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_update_validate_func_info", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 36, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("validr._validator_c._update_validate_func_info", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6validr_12_validator_c_8_update_validate_func_info(__pyx_self, __pyx_v_validate_func, __pyx_v_origin_func, __pyx_v_schema);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6validr_12_validator_c_8_update_validate_func_info(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_validate_func, PyObject *__pyx_v_origin_func, PyObject *__pyx_v_schema) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_update_validate_func_info", 0);
  __Pyx_TraceCall("_update_validate_func_info (wrapper)", __pyx_f[0], 36, 0, __PYX_ERR(0, 36, __pyx_L1_error));
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6validr_12_validator_c__update_validate_func_info(__pyx_v_validate_func, __pyx_v_origin_func, __pyx_v_schema, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("validr._validator_c._update_validate_func_info", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":51
 * 
 * 
 * def _parse_hints(hints):             # <<<<<<<<<<<<<<
 *     if not isinstance(hints, (tuple, set, list)):
 *         hints = [hints]
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_11_parse_hints(PyObject *__pyx_self, PyObject *__pyx_v_hints); /*proto*/
static PyMethodDef __pyx_mdef_6validr_12_validator_c_11_parse_hints = {"_parse_hints", (PyCFunction)__pyx_pw_6validr_12_validator_c_11_parse_hints, METH_O, 0};
static PyObject *__pyx_pw_6validr_12_validator_c_11_parse_hints(PyObject *__pyx_self, PyObject *__pyx_v_hints) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_parse_hints (wrapper)", 0);
  __pyx_r = __pyx_pf_6validr_12_validator_c_10_parse_hints(__pyx_self, ((PyObject *)__pyx_v_hints));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6validr_12_validator_c_10_parse_hints(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_hints) {
  int __pyx_v_is_string;
  int __pyx_v_is_object;
  PyObject *__pyx_v_types = NULL;
  PyObject *__pyx_v_hint = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_TraceFrameInit(__pyx_codeobj__3)
  __Pyx_RefNannySetupContext("_parse_hints", 0);
  __Pyx_TraceCall("_parse_hints", __pyx_f[0], 51, 0, __PYX_ERR(0, 51, __pyx_L1_error));
  __Pyx_INCREF(__pyx_v_hints);

  /* "validr/_validator_c.pyx":52
 * 
 * def _parse_hints(hints):
 *     if not isinstance(hints, (tuple, set, list)):             # <<<<<<<<<<<<<<
 *         hints = [hints]
 *     is_string = False
 */
  __Pyx_TraceLine(52,0,__PYX_ERR(0, 52, __pyx_L1_error))
  __pyx_t_2 = PyTuple_Check(__pyx_v_hints); 
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (!__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = PySet_Check(__pyx_v_hints); 
  __pyx_t_2 = (__pyx_t_3 != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = PyList_Check(__pyx_v_hints); 
  __pyx_t_3 = (__pyx_t_2 != 0);
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  __pyx_t_3 = ((!(__pyx_t_1 != 0)) != 0);
  if (__pyx_t_3) {

    /* "validr/_validator_c.pyx":53
 * def _parse_hints(hints):
 *     if not isinstance(hints, (tuple, set, list)):
 *         hints = [hints]             # <<<<<<<<<<<<<<
 *     is_string = False
 *     is_object = False
 */
    __Pyx_TraceLine(53,0,__PYX_ERR(0, 53, __pyx_L1_error))
    __pyx_t_4 = PyList_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_hints);
    __Pyx_GIVEREF(__pyx_v_hints);
    PyList_SET_ITEM(__pyx_t_4, 0, __pyx_v_hints);
    __Pyx_DECREF_SET(__pyx_v_hints, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "validr/_validator_c.pyx":52
 * 
 * def _parse_hints(hints):
 *     if not isinstance(hints, (tuple, set, list)):             # <<<<<<<<<<<<<<
 *         hints = [hints]
 *     is_string = False
 */
  }

  /* "validr/_validator_c.pyx":54
 *     if not isinstance(hints, (tuple, set, list)):
 *         hints = [hints]
 *     is_string = False             # <<<<<<<<<<<<<<
 *     is_object = False
 *     types = []
 */
  __Pyx_TraceLine(54,0,__PYX_ERR(0, 54, __pyx_L1_error))
  __pyx_v_is_string = 0;

  /* "validr/_validator_c.pyx":55
 *         hints = [hints]
 *     is_string = False
 *     is_object = False             # <<<<<<<<<<<<<<
 *     types = []
 *     for hint in hints:
 */
  __Pyx_TraceLine(55,0,__PYX_ERR(0, 55, __pyx_L1_error))
  __pyx_v_is_object = 0;

  /* "validr/_validator_c.pyx":56
 *     is_string = False
 *     is_object = False
 *     types = []             # <<<<<<<<<<<<<<
 *     for hint in hints:
 *         if hint is str:
 */
  __Pyx_TraceLine(56,0,__PYX_ERR(0, 56, __pyx_L1_error))
  __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_types = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "validr/_validator_c.pyx":57
 *     is_object = False
 *     types = []
 *     for hint in hints:             # <<<<<<<<<<<<<<
 *         if hint is str:
 *             is_string = True
 */
  __Pyx_TraceLine(57,0,__PYX_ERR(0, 57, __pyx_L1_error))
  if (likely(PyList_CheckExact(__pyx_v_hints)) || PyTuple_CheckExact(__pyx_v_hints)) {
    __pyx_t_4 = __pyx_v_hints; __Pyx_INCREF(__pyx_t_4); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
  } else {
    __pyx_t_5 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_v_hints); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 57, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 57, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_6)) {
      if (likely(PyList_CheckExact(__pyx_t_4))) {
        if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_7 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_7); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 57, __pyx_L1_error)
        #else
        __pyx_t_7 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 57, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
      } else {
        if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_7); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 57, __pyx_L1_error)
        #else
        __pyx_t_7 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 57, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
      }
    } else {
      __pyx_t_7 = __pyx_t_6(__pyx_t_4);
      if (unlikely(!__pyx_t_7)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 57, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_7);
    }
    __Pyx_XDECREF_SET(__pyx_v_hint, __pyx_t_7);
    __pyx_t_7 = 0;

    /* "validr/_validator_c.pyx":58
 *     types = []
 *     for hint in hints:
 *         if hint is str:             # <<<<<<<<<<<<<<
 *             is_string = True
 *         else:
 */
    __Pyx_TraceLine(58,0,__PYX_ERR(0, 58, __pyx_L1_error))
    __pyx_t_3 = (__pyx_v_hint == ((PyObject *)(&PyUnicode_Type)));
    __pyx_t_1 = (__pyx_t_3 != 0);
    if (__pyx_t_1) {

      /* "validr/_validator_c.pyx":59
 *     for hint in hints:
 *         if hint is str:
 *             is_string = True             # <<<<<<<<<<<<<<
 *         else:
 *             is_object = True
 */
      __Pyx_TraceLine(59,0,__PYX_ERR(0, 59, __pyx_L1_error))
      __pyx_v_is_string = 1;

      /* "validr/_validator_c.pyx":58
 *     types = []
 *     for hint in hints:
 *         if hint is str:             # <<<<<<<<<<<<<<
 *             is_string = True
 *         else:
 */
      goto __pyx_L9;
    }

    /* "validr/_validator_c.pyx":61
 *             is_string = True
 *         else:
 *             is_object = True             # <<<<<<<<<<<<<<
 *         if hint is object:
 *             hint = typing.Any
 */
    __Pyx_TraceLine(61,0,__PYX_ERR(0, 61, __pyx_L1_error))
    /*else*/ {
      __pyx_v_is_object = 1;
    }
    __pyx_L9:;

    /* "validr/_validator_c.pyx":62
 *         else:
 *             is_object = True
 *         if hint is object:             # <<<<<<<<<<<<<<
 *             hint = typing.Any
 *         types.append(hint)
 */
    __Pyx_TraceLine(62,0,__PYX_ERR(0, 62, __pyx_L1_error))
    __pyx_t_1 = (__pyx_v_hint == __pyx_builtin_object);
    __pyx_t_3 = (__pyx_t_1 != 0);
    if (__pyx_t_3) {

      /* "validr/_validator_c.pyx":63
 *             is_object = True
 *         if hint is object:
 *             hint = typing.Any             # <<<<<<<<<<<<<<
 *         types.append(hint)
 *     return is_string, is_object, tuple(types)
 */
      __Pyx_TraceLine(63,0,__PYX_ERR(0, 63, __pyx_L1_error))
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_typing); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 63, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Any); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 63, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF_SET(__pyx_v_hint, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "validr/_validator_c.pyx":62
 *         else:
 *             is_object = True
 *         if hint is object:             # <<<<<<<<<<<<<<
 *             hint = typing.Any
 *         types.append(hint)
 */
    }

    /* "validr/_validator_c.pyx":64
 *         if hint is object:
 *             hint = typing.Any
 *         types.append(hint)             # <<<<<<<<<<<<<<
 *     return is_string, is_object, tuple(types)
 * 
 */
    __Pyx_TraceLine(64,0,__PYX_ERR(0, 64, __pyx_L1_error))
    __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_types, __pyx_v_hint); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 64, __pyx_L1_error)

    /* "validr/_validator_c.pyx":57
 *     is_object = False
 *     types = []
 *     for hint in hints:             # <<<<<<<<<<<<<<
 *         if hint is str:
 *             is_string = True
 */
    __Pyx_TraceLine(57,0,__PYX_ERR(0, 57, __pyx_L1_error))
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "validr/_validator_c.pyx":65
 *             hint = typing.Any
 *         types.append(hint)
 *     return is_string, is_object, tuple(types)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(65,0,__PYX_ERR(0, 65, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_is_string); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_8 = __Pyx_PyBool_FromLong(__pyx_v_is_object); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_7 = PyList_AsTuple(__pyx_v_types); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_10 = PyTuple_New(3); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_t_7);
  __pyx_t_4 = 0;
  __pyx_t_8 = 0;
  __pyx_t_7 = 0;
  __pyx_r = __pyx_t_10;
  __pyx_t_10 = 0;
  goto __pyx_L0;

  /* "validr/_validator_c.pyx":51
 * 
 * 
 * def _parse_hints(hints):             # <<<<<<<<<<<<<<
 *     if not isinstance(hints, (tuple, set, list)):
 *         hints = [hints]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("validr._validator_c._parse_hints", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_types);
  __Pyx_XDECREF(__pyx_v_hint);
  __Pyx_XDECREF(__pyx_v_hints);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":68
 * 
 * 
 * def _update_validate_func_type_hints(             # <<<<<<<<<<<<<<
 *         validate_func,
 *         optional, has_default,
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_13_update_validate_func_type_hints(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6validr_12_validator_c_13_update_validate_func_type_hints = {"_update_validate_func_type_hints", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6validr_12_validator_c_13_update_validate_func_type_hints, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6validr_12_validator_c_13_update_validate_func_type_hints(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_validate_func = 0;
  PyObject *__pyx_v_optional = 0;
  PyObject *__pyx_v_has_default = 0;
  PyObject *__pyx_v_accept_hints = 0;
  PyObject *__pyx_v_output_hints = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_update_validate_func_type_hints (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_validate_func,&__pyx_n_s_optional,&__pyx_n_s_has_default,&__pyx_n_s_accept_hints,&__pyx_n_s_output_hints,0};
    PyObject* values[5] = {0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_validate_func)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_optional)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_update_validate_func_type_hints", 1, 5, 5, 1); __PYX_ERR(0, 68, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_has_default)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_update_validate_func_type_hints", 1, 5, 5, 2); __PYX_ERR(0, 68, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_accept_hints)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_update_validate_func_type_hints", 1, 5, 5, 3); __PYX_ERR(0, 68, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_output_hints)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_update_validate_func_type_hints", 1, 5, 5, 4); __PYX_ERR(0, 68, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_update_validate_func_type_hints") < 0)) __PYX_ERR(0, 68, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 5) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
    }
    __pyx_v_validate_func = values[0];
    __pyx_v_optional = values[1];
    __pyx_v_has_default = values[2];
    __pyx_v_accept_hints = values[3];
    __pyx_v_output_hints = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_update_validate_func_type_hints", 1, 5, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 68, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("validr._validator_c._update_validate_func_type_hints", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6validr_12_validator_c_12_update_validate_func_type_hints(__pyx_self, __pyx_v_validate_func, __pyx_v_optional, __pyx_v_has_default, __pyx_v_accept_hints, __pyx_v_output_hints);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6validr_12_validator_c_12_update_validate_func_type_hints(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_validate_func, PyObject *__pyx_v_optional, PyObject *__pyx_v_has_default, PyObject *__pyx_v_accept_hints, PyObject *__pyx_v_output_hints) {
  PyObject *__pyx_v_value_typing = NULL;
  PyObject *__pyx_v_return_typing = NULL;
  PyObject *__pyx_v_annotations = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  __Pyx_TraceFrameInit(__pyx_codeobj__4)
  __Pyx_RefNannySetupContext("_update_validate_func_type_hints", 0);
  __Pyx_TraceCall("_update_validate_func_type_hints", __pyx_f[0], 68, 0, __PYX_ERR(0, 68, __pyx_L1_error));

  /* "validr/_validator_c.pyx":73
 *         accept_hints, output_hints,
 * ):
 *     assert accept_hints, 'no accept_hints'             # <<<<<<<<<<<<<<
 *     assert output_hints, 'no output_hints'
 *     if len(accept_hints) == 1:
 */
  __Pyx_TraceLine(73,0,__PYX_ERR(0, 73, __pyx_L1_error))
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_accept_hints); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 73, __pyx_L1_error)
    if (unlikely(!__pyx_t_1)) {
      PyErr_SetObject(PyExc_AssertionError, __pyx_kp_u_no_accept_hints);
      __PYX_ERR(0, 73, __pyx_L1_error)
    }
  }
  #endif

  /* "validr/_validator_c.pyx":74
 * ):
 *     assert accept_hints, 'no accept_hints'
 *     assert output_hints, 'no output_hints'             # <<<<<<<<<<<<<<
 *     if len(accept_hints) == 1:
 *         value_typing = accept_hints[0]
 */
  __Pyx_TraceLine(74,0,__PYX_ERR(0, 74, __pyx_L1_error))
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_output_hints); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 74, __pyx_L1_error)
    if (unlikely(!__pyx_t_1)) {
      PyErr_SetObject(PyExc_AssertionError, __pyx_kp_u_no_output_hints);
      __PYX_ERR(0, 74, __pyx_L1_error)
    }
  }
  #endif

  /* "validr/_validator_c.pyx":75
 *     assert accept_hints, 'no accept_hints'
 *     assert output_hints, 'no output_hints'
 *     if len(accept_hints) == 1:             # <<<<<<<<<<<<<<
 *         value_typing = accept_hints[0]
 *     else:
 */
  __Pyx_TraceLine(75,0,__PYX_ERR(0, 75, __pyx_L1_error))
  __pyx_t_2 = PyObject_Length(__pyx_v_accept_hints); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 75, __pyx_L1_error)
  __pyx_t_1 = ((__pyx_t_2 == 1) != 0);
  if (__pyx_t_1) {

    /* "validr/_validator_c.pyx":76
 *     assert output_hints, 'no output_hints'
 *     if len(accept_hints) == 1:
 *         value_typing = accept_hints[0]             # <<<<<<<<<<<<<<
 *     else:
 *         value_typing = typing.Union[accept_hints]
 */
    __Pyx_TraceLine(76,0,__PYX_ERR(0, 76, __pyx_L1_error))
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_accept_hints, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_value_typing = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "validr/_validator_c.pyx":75
 *     assert accept_hints, 'no accept_hints'
 *     assert output_hints, 'no output_hints'
 *     if len(accept_hints) == 1:             # <<<<<<<<<<<<<<
 *         value_typing = accept_hints[0]
 *     else:
 */
    goto __pyx_L3;
  }

  /* "validr/_validator_c.pyx":78
 *         value_typing = accept_hints[0]
 *     else:
 *         value_typing = typing.Union[accept_hints]             # <<<<<<<<<<<<<<
 *     if optional:
 *         value_typing = typing.Optional[value_typing]
 */
  __Pyx_TraceLine(78,0,__PYX_ERR(0, 78, __pyx_L1_error))
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_typing); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 78, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Union); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 78, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_t_4, __pyx_v_accept_hints); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 78, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_value_typing = __pyx_t_3;
    __pyx_t_3 = 0;
  }
  __pyx_L3:;

  /* "validr/_validator_c.pyx":79
 *     else:
 *         value_typing = typing.Union[accept_hints]
 *     if optional:             # <<<<<<<<<<<<<<
 *         value_typing = typing.Optional[value_typing]
 *     if len(output_hints) == 1:
 */
  __Pyx_TraceLine(79,0,__PYX_ERR(0, 79, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_optional); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 79, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "validr/_validator_c.pyx":80
 *         value_typing = typing.Union[accept_hints]
 *     if optional:
 *         value_typing = typing.Optional[value_typing]             # <<<<<<<<<<<<<<
 *     if len(output_hints) == 1:
 *         return_typing = output_hints[0]
 */
    __Pyx_TraceLine(80,0,__PYX_ERR(0, 80, __pyx_L1_error))
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_typing); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 80, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Optional); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 80, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_t_4, __pyx_v_value_typing); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 80, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_value_typing, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "validr/_validator_c.pyx":79
 *     else:
 *         value_typing = typing.Union[accept_hints]
 *     if optional:             # <<<<<<<<<<<<<<
 *         value_typing = typing.Optional[value_typing]
 *     if len(output_hints) == 1:
 */
  }

  /* "validr/_validator_c.pyx":81
 *     if optional:
 *         value_typing = typing.Optional[value_typing]
 *     if len(output_hints) == 1:             # <<<<<<<<<<<<<<
 *         return_typing = output_hints[0]
 *     else:
 */
  __Pyx_TraceLine(81,0,__PYX_ERR(0, 81, __pyx_L1_error))
  __pyx_t_2 = PyObject_Length(__pyx_v_output_hints); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 81, __pyx_L1_error)
  __pyx_t_1 = ((__pyx_t_2 == 1) != 0);
  if (__pyx_t_1) {

    /* "validr/_validator_c.pyx":82
 *         value_typing = typing.Optional[value_typing]
 *     if len(output_hints) == 1:
 *         return_typing = output_hints[0]             # <<<<<<<<<<<<<<
 *     else:
 *         return_typing = typing.Union[output_hints]
 */
    __Pyx_TraceLine(82,0,__PYX_ERR(0, 82, __pyx_L1_error))
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_output_hints, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_return_typing = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "validr/_validator_c.pyx":81
 *     if optional:
 *         value_typing = typing.Optional[value_typing]
 *     if len(output_hints) == 1:             # <<<<<<<<<<<<<<
 *         return_typing = output_hints[0]
 *     else:
 */
    goto __pyx_L5;
  }

  /* "validr/_validator_c.pyx":84
 *         return_typing = output_hints[0]
 *     else:
 *         return_typing = typing.Union[output_hints]             # <<<<<<<<<<<<<<
 *     if optional and not has_default:
 *         return_typing = typing.Optional[return_typing]
 */
  __Pyx_TraceLine(84,0,__PYX_ERR(0, 84, __pyx_L1_error))
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_typing); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 84, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Union); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 84, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_t_4, __pyx_v_output_hints); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 84, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_return_typing = __pyx_t_3;
    __pyx_t_3 = 0;
  }
  __pyx_L5:;

  /* "validr/_validator_c.pyx":85
 *     else:
 *         return_typing = typing.Union[output_hints]
 *     if optional and not has_default:             # <<<<<<<<<<<<<<
 *         return_typing = typing.Optional[return_typing]
 *     annotations = {'value': value_typing, 'return': return_typing}
 */
  __Pyx_TraceLine(85,0,__PYX_ERR(0, 85, __pyx_L1_error))
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_optional); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 85, __pyx_L1_error)
  if (__pyx_t_5) {
  } else {
    __pyx_t_1 = __pyx_t_5;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_has_default); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 85, __pyx_L1_error)
  __pyx_t_6 = ((!__pyx_t_5) != 0);
  __pyx_t_1 = __pyx_t_6;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_1) {

    /* "validr/_validator_c.pyx":86
 *         return_typing = typing.Union[output_hints]
 *     if optional and not has_default:
 *         return_typing = typing.Optional[return_typing]             # <<<<<<<<<<<<<<
 *     annotations = {'value': value_typing, 'return': return_typing}
 *     validate_func.__annotations__ = annotations
 */
    __Pyx_TraceLine(86,0,__PYX_ERR(0, 86, __pyx_L1_error))
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_typing); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Optional); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_t_4, __pyx_v_return_typing); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_return_typing, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "validr/_validator_c.pyx":85
 *     else:
 *         return_typing = typing.Union[output_hints]
 *     if optional and not has_default:             # <<<<<<<<<<<<<<
 *         return_typing = typing.Optional[return_typing]
 *     annotations = {'value': value_typing, 'return': return_typing}
 */
  }

  /* "validr/_validator_c.pyx":87
 *     if optional and not has_default:
 *         return_typing = typing.Optional[return_typing]
 *     annotations = {'value': value_typing, 'return': return_typing}             # <<<<<<<<<<<<<<
 *     validate_func.__annotations__ = annotations
 * 
 */
  __Pyx_TraceLine(87,0,__PYX_ERR(0, 87, __pyx_L1_error))
  __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 87, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_value, __pyx_v_value_typing) < 0) __PYX_ERR(0, 87, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_return, __pyx_v_return_typing) < 0) __PYX_ERR(0, 87, __pyx_L1_error)
  __pyx_v_annotations = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "validr/_validator_c.pyx":88
 *         return_typing = typing.Optional[return_typing]
 *     annotations = {'value': value_typing, 'return': return_typing}
 *     validate_func.__annotations__ = annotations             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(88,0,__PYX_ERR(0, 88, __pyx_L1_error))
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_validate_func, __pyx_n_s_annotations, __pyx_v_annotations) < 0) __PYX_ERR(0, 88, __pyx_L1_error)

  /* "validr/_validator_c.pyx":68
 * 
 * 
 * def _update_validate_func_type_hints(             # <<<<<<<<<<<<<<
 *         validate_func,
 *         optional, has_default,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("validr._validator_c._update_validate_func_type_hints", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value_typing);
  __Pyx_XDECREF(__pyx_v_return_typing);
  __Pyx_XDECREF(__pyx_v_annotations);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":91
 * 
 * 
 * def validator(string=None, *, accept=None, output=None):             # <<<<<<<<<<<<<<
 *     """Decorator for create validator
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_15validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6validr_12_validator_c_14validator[] = "Decorator for create validator\n\n    It will handle params default,optional,desc automatically.\n\n    Usage:\n\n        @validator(accept=(str,object), output=(str, object))\n        def xxx_validator(compiler, output_object, **params):\n            def validate(value):\n                try:\n                    # validate/convert the value\n                except Exception:\n                    # raise Invalid('invalid xxx')\n                if output_object:\n                    # return python object\n                else:\n                    # return string\n            return validate\n\n    Args:\n        accept (str | object | (str,object)):\n            str: the validator accept only string, treat both None and empty string as None\n            object: the validator accept only object\n            (str,object): (default) the validator accept both string and object,\n                treat both None and empty string as None\n        output (str | object | (str,object)):\n            str: (default) the validator always output string, convert None to empty string\n            object: the validator always output object\n            (str, object): the validator can output both string and object,\n                and has an `object` parameter to control which to output\n        string (bool): deprecated in v1.1.0.\n            string=True equal to accept=(str, object), output=str\n            string=False equal to accept=(str, object), output=object\n    ";
static PyMethodDef __pyx_mdef_6validr_12_validator_c_15validator = {"validator", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6validr_12_validator_c_15validator, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6validr_12_validator_c_14validator};
static PyObject *__pyx_pw_6validr_12_validator_c_15validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_string = 0;
  PyObject *__pyx_v_accept = 0;
  PyObject *__pyx_v_output = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("validator (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_string,&__pyx_n_s_accept,&__pyx_n_s_output,0};
    PyObject* values[3] = {0,0,0};
    values[0] = ((PyObject *)Py_None);
    values[1] = ((PyObject *)Py_None);
    values[2] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_string);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (kw_args > 0 && likely(kw_args <= 2)) {
        Py_ssize_t index;
        for (index = 1; index < 3 && kw_args > 0; index++) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, *__pyx_pyargnames[index]);
          if (value) { values[index] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "validator") < 0)) __PYX_ERR(0, 91, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_string = values[0];
    __pyx_v_accept = values[1];
    __pyx_v_output = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("validator", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 91, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("validr._validator_c.validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6validr_12_validator_c_14validator(__pyx_self, __pyx_v_string, __pyx_v_accept, __pyx_v_output);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":145
 *     del string, accept, output
 * 
 *     def decorator(f):             # <<<<<<<<<<<<<<
 * 
 *         def _m_validator(compiler, schema):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_9validator_1decorator(PyObject *__pyx_self, PyObject *__pyx_v_f); /*proto*/
static PyMethodDef __pyx_mdef_6validr_12_validator_c_9validator_1decorator = {"decorator", (PyCFunction)__pyx_pw_6validr_12_validator_c_9validator_1decorator, METH_O, 0};
static PyObject *__pyx_pw_6validr_12_validator_c_9validator_1decorator(PyObject *__pyx_self, PyObject *__pyx_v_f) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("decorator (wrapper)", 0);
  __pyx_r = __pyx_pf_6validr_12_validator_c_9validator_decorator(__pyx_self, ((PyObject *)__pyx_v_f));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":147
 *     def decorator(f):
 * 
 *         def _m_validator(compiler, schema):             # <<<<<<<<<<<<<<
 *             params = schema.params.copy()
 *             if schema.items is not None:
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_9validator_9decorator_1_m_validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6validr_12_validator_c_9validator_9decorator_1_m_validator = {"_m_validator", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6validr_12_validator_c_9validator_9decorator_1_m_validator, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6validr_12_validator_c_9validator_9decorator_1_m_validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_compiler = 0;
  PyObject *__pyx_v_schema = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_m_validator (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_compiler,&__pyx_n_s_schema_2,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_compiler)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_schema_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_m_validator", 1, 2, 2, 1); __PYX_ERR(0, 147, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_m_validator") < 0)) __PYX_ERR(0, 147, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_compiler = values[0];
    __pyx_v_schema = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_m_validator", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 147, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("validr._validator_c.validator.decorator._m_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6validr_12_validator_c_9validator_9decorator__m_validator(__pyx_self, __pyx_v_compiler, __pyx_v_schema);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":203
 * 
 *             # check null, empty string and default
 *             def _m_validate(value):             # <<<<<<<<<<<<<<
 *                 cdef bint is_null
 *                 if accept_string:
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_9validator_9decorator_12_m_validator_1_m_validate(PyObject *__pyx_self, PyObject *__pyx_v_value); /*proto*/
static PyMethodDef __pyx_mdef_6validr_12_validator_c_9validator_9decorator_12_m_validator_1_m_validate = {"_m_validate", (PyCFunction)__pyx_pw_6validr_12_validator_c_9validator_9decorator_12_m_validator_1_m_validate, METH_O, 0};
static PyObject *__pyx_pw_6validr_12_validator_c_9validator_9decorator_12_m_validator_1_m_validate(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_m_validate (wrapper)", 0);
  __pyx_r = __pyx_pf_6validr_12_validator_c_9validator_9decorator_12_m_validator__m_validate(__pyx_self, ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6validr_12_validator_c_9validator_9decorator_12_m_validator__m_validate(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_2__m_validator *__pyx_cur_scope;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_2__m_validator *__pyx_outer_scope;
  int __pyx_v_is_null;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("_m_validate", 0);
  __pyx_outer_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_2__m_validator *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_TraceCall("_m_validate", __pyx_f[0], 203, 0, __PYX_ERR(0, 203, __pyx_L1_error));
  __Pyx_INCREF(__pyx_v_value);

  /* "validr/_validator_c.pyx":205
 *             def _m_validate(value):
 *                 cdef bint is_null
 *                 if accept_string:             # <<<<<<<<<<<<<<
 *                     is_null = _is_empty(value)
 *                 else:
 */
  __Pyx_TraceLine(205,0,__PYX_ERR(0, 205, __pyx_L1_error))
  __pyx_t_1 = (__pyx_cur_scope->__pyx_outer_scope->__pyx_outer_scope->__pyx_v_accept_string != 0);
  if (__pyx_t_1) {

    /* "validr/_validator_c.pyx":206
 *                 cdef bint is_null
 *                 if accept_string:
 *                     is_null = _is_empty(value)             # <<<<<<<<<<<<<<
 *                 else:
 *                     is_null = value is None
 */
    __Pyx_TraceLine(206,0,__PYX_ERR(0, 206, __pyx_L1_error))
    __pyx_v_is_null = __pyx_f_6validr_12_validator_c__is_empty(__pyx_v_value, 0);

    /* "validr/_validator_c.pyx":205
 *             def _m_validate(value):
 *                 cdef bint is_null
 *                 if accept_string:             # <<<<<<<<<<<<<<
 *                     is_null = _is_empty(value)
 *                 else:
 */
    goto __pyx_L3;
  }

  /* "validr/_validator_c.pyx":208
 *                     is_null = _is_empty(value)
 *                 else:
 *                     is_null = value is None             # <<<<<<<<<<<<<<
 *                 if is_null:
 *                     if has_default:
 */
  __Pyx_TraceLine(208,0,__PYX_ERR(0, 208, __pyx_L1_error))
  /*else*/ {
    __pyx_t_1 = (__pyx_v_value == Py_None);
    __pyx_v_is_null = __pyx_t_1;
  }
  __pyx_L3:;

  /* "validr/_validator_c.pyx":209
 *                 else:
 *                     is_null = value is None
 *                 if is_null:             # <<<<<<<<<<<<<<
 *                     if has_default:
 *                         return default
 */
  __Pyx_TraceLine(209,0,__PYX_ERR(0, 209, __pyx_L1_error))
  __pyx_t_1 = (__pyx_v_is_null != 0);
  if (__pyx_t_1) {

    /* "validr/_validator_c.pyx":210
 *                     is_null = value is None
 *                 if is_null:
 *                     if has_default:             # <<<<<<<<<<<<<<
 *                         return default
 *                     elif optional:
 */
    __Pyx_TraceLine(210,0,__PYX_ERR(0, 210, __pyx_L1_error))
    __pyx_t_1 = (__pyx_cur_scope->__pyx_v_has_default != 0);
    if (__pyx_t_1) {

      /* "validr/_validator_c.pyx":211
 *                 if is_null:
 *                     if has_default:
 *                         return default             # <<<<<<<<<<<<<<
 *                     elif optional:
 *                         return null_output
 */
      __Pyx_TraceLine(211,0,__PYX_ERR(0, 211, __pyx_L1_error))
      __Pyx_XDECREF(__pyx_r);
      if (unlikely(!__pyx_cur_scope->__pyx_v_default)) { __Pyx_RaiseClosureNameError("default"); __PYX_ERR(0, 211, __pyx_L1_error) }
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_default);
      __pyx_r = __pyx_cur_scope->__pyx_v_default;
      goto __pyx_L0;

      /* "validr/_validator_c.pyx":210
 *                     is_null = value is None
 *                 if is_null:
 *                     if has_default:             # <<<<<<<<<<<<<<
 *                         return default
 *                     elif optional:
 */
    }

    /* "validr/_validator_c.pyx":212
 *                     if has_default:
 *                         return default
 *                     elif optional:             # <<<<<<<<<<<<<<
 *                         return null_output
 *                     else:
 */
    __Pyx_TraceLine(212,0,__PYX_ERR(0, 212, __pyx_L1_error))
    __pyx_t_1 = (__pyx_cur_scope->__pyx_v_optional != 0);
    if (likely(__pyx_t_1)) {

      /* "validr/_validator_c.pyx":213
 *                         return default
 *                     elif optional:
 *                         return null_output             # <<<<<<<<<<<<<<
 *                     else:
 *                         raise Invalid('required')
 */
      __Pyx_TraceLine(213,0,__PYX_ERR(0, 213, __pyx_L1_error))
      __Pyx_XDECREF(__pyx_r);
      if (unlikely(!__pyx_cur_scope->__pyx_v_null_output)) { __Pyx_RaiseClosureNameError("null_output"); __PYX_ERR(0, 213, __pyx_L1_error) }
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_null_output);
      __pyx_r = __pyx_cur_scope->__pyx_v_null_output;
      goto __pyx_L0;

      /* "validr/_validator_c.pyx":212
 *                     if has_default:
 *                         return default
 *                     elif optional:             # <<<<<<<<<<<<<<
 *                         return null_output
 *                     else:
 */
    }

    /* "validr/_validator_c.pyx":215
 *                         return null_output
 *                     else:
 *                         raise Invalid('required')             # <<<<<<<<<<<<<<
 *                 if not accept_object and not isinstance(value, str):
 *                     raise Invalid('require string value')
 */
    __Pyx_TraceLine(215,0,__PYX_ERR(0, 215, __pyx_L1_error))
    /*else*/ {
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 215, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_n_u_required) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_n_u_required);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 215, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 215, __pyx_L1_error)
    }

    /* "validr/_validator_c.pyx":209
 *                 else:
 *                     is_null = value is None
 *                 if is_null:             # <<<<<<<<<<<<<<
 *                     if has_default:
 *                         return default
 */
  }

  /* "validr/_validator_c.pyx":216
 *                     else:
 *                         raise Invalid('required')
 *                 if not accept_object and not isinstance(value, str):             # <<<<<<<<<<<<<<
 *                     raise Invalid('require string value')
 *                 value = validate(value)
 */
  __Pyx_TraceLine(216,0,__PYX_ERR(0, 216, __pyx_L1_error))
  __pyx_t_5 = ((!(__pyx_cur_scope->__pyx_outer_scope->__pyx_outer_scope->__pyx_v_accept_object != 0)) != 0);
  if (__pyx_t_5) {
  } else {
    __pyx_t_1 = __pyx_t_5;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_5 = PyUnicode_Check(__pyx_v_value); 
  __pyx_t_6 = ((!(__pyx_t_5 != 0)) != 0);
  __pyx_t_1 = __pyx_t_6;
  __pyx_L7_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "validr/_validator_c.pyx":217
 *                         raise Invalid('required')
 *                 if not accept_object and not isinstance(value, str):
 *                     raise Invalid('require string value')             # <<<<<<<<<<<<<<
 *                 value = validate(value)
 *                 # check again after validate
 */
    __Pyx_TraceLine(217,0,__PYX_ERR(0, 217, __pyx_L1_error))
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 217, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_kp_u_require_string_value) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_kp_u_require_string_value);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 217, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 217, __pyx_L1_error)

    /* "validr/_validator_c.pyx":216
 *                     else:
 *                         raise Invalid('required')
 *                 if not accept_object and not isinstance(value, str):             # <<<<<<<<<<<<<<
 *                     raise Invalid('require string value')
 *                 value = validate(value)
 */
  }

  /* "validr/_validator_c.pyx":218
 *                 if not accept_object and not isinstance(value, str):
 *                     raise Invalid('require string value')
 *                 value = validate(value)             # <<<<<<<<<<<<<<
 *                 # check again after validate
 *                 if accept_string:
 */
  __Pyx_TraceLine(218,0,__PYX_ERR(0, 218, __pyx_L1_error))
  if (unlikely(!__pyx_cur_scope->__pyx_v_validate)) { __Pyx_RaiseClosureNameError("validate"); __PYX_ERR(0, 218, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_validate);
  __pyx_t_3 = __pyx_cur_scope->__pyx_v_validate; __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_value) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_value);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 218, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_2);
  __pyx_t_2 = 0;

  /* "validr/_validator_c.pyx":220
 *                 value = validate(value)
 *                 # check again after validate
 *                 if accept_string:             # <<<<<<<<<<<<<<
 *                     is_null = _is_empty(value)
 *                 else:
 */
  __Pyx_TraceLine(220,0,__PYX_ERR(0, 220, __pyx_L1_error))
  __pyx_t_1 = (__pyx_cur_scope->__pyx_outer_scope->__pyx_outer_scope->__pyx_v_accept_string != 0);
  if (__pyx_t_1) {

    /* "validr/_validator_c.pyx":221
 *                 # check again after validate
 *                 if accept_string:
 *                     is_null = _is_empty(value)             # <<<<<<<<<<<<<<
 *                 else:
 *                     is_null = value is None
 */
    __Pyx_TraceLine(221,0,__PYX_ERR(0, 221, __pyx_L1_error))
    __pyx_v_is_null = __pyx_f_6validr_12_validator_c__is_empty(__pyx_v_value, 0);

    /* "validr/_validator_c.pyx":220
 *                 value = validate(value)
 *                 # check again after validate
 *                 if accept_string:             # <<<<<<<<<<<<<<
 *                     is_null = _is_empty(value)
 *                 else:
 */
    goto __pyx_L9;
  }

  /* "validr/_validator_c.pyx":223
 *                     is_null = _is_empty(value)
 *                 else:
 *                     is_null = value is None             # <<<<<<<<<<<<<<
 *                 if is_null:
 *                     if has_default:
 */
  __Pyx_TraceLine(223,0,__PYX_ERR(0, 223, __pyx_L1_error))
  /*else*/ {
    __pyx_t_1 = (__pyx_v_value == Py_None);
    __pyx_v_is_null = __pyx_t_1;
  }
  __pyx_L9:;

  /* "validr/_validator_c.pyx":224
 *                 else:
 *                     is_null = value is None
 *                 if is_null:             # <<<<<<<<<<<<<<
 *                     if has_default:
 *                         return default
 */
  __Pyx_TraceLine(224,0,__PYX_ERR(0, 224, __pyx_L1_error))
  __pyx_t_1 = (__pyx_v_is_null != 0);
  if (__pyx_t_1) {

    /* "validr/_validator_c.pyx":225
 *                     is_null = value is None
 *                 if is_null:
 *                     if has_default:             # <<<<<<<<<<<<<<
 *                         return default
 *                     elif optional:
 */
    __Pyx_TraceLine(225,0,__PYX_ERR(0, 225, __pyx_L1_error))
    __pyx_t_1 = (__pyx_cur_scope->__pyx_v_has_default != 0);
    if (__pyx_t_1) {

      /* "validr/_validator_c.pyx":226
 *                 if is_null:
 *                     if has_default:
 *                         return default             # <<<<<<<<<<<<<<
 *                     elif optional:
 *                         return null_output
 */
      __Pyx_TraceLine(226,0,__PYX_ERR(0, 226, __pyx_L1_error))
      __Pyx_XDECREF(__pyx_r);
      if (unlikely(!__pyx_cur_scope->__pyx_v_default)) { __Pyx_RaiseClosureNameError("default"); __PYX_ERR(0, 226, __pyx_L1_error) }
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_default);
      __pyx_r = __pyx_cur_scope->__pyx_v_default;
      goto __pyx_L0;

      /* "validr/_validator_c.pyx":225
 *                     is_null = value is None
 *                 if is_null:
 *                     if has_default:             # <<<<<<<<<<<<<<
 *                         return default
 *                     elif optional:
 */
    }

    /* "validr/_validator_c.pyx":227
 *                     if has_default:
 *                         return default
 *                     elif optional:             # <<<<<<<<<<<<<<
 *                         return null_output
 *                     else:
 */
    __Pyx_TraceLine(227,0,__PYX_ERR(0, 227, __pyx_L1_error))
    __pyx_t_1 = (__pyx_cur_scope->__pyx_v_optional != 0);
    if (likely(__pyx_t_1)) {

      /* "validr/_validator_c.pyx":228
 *                         return default
 *                     elif optional:
 *                         return null_output             # <<<<<<<<<<<<<<
 *                     else:
 *                         raise Invalid('required')
 */
      __Pyx_TraceLine(228,0,__PYX_ERR(0, 228, __pyx_L1_error))
      __Pyx_XDECREF(__pyx_r);
      if (unlikely(!__pyx_cur_scope->__pyx_v_null_output)) { __Pyx_RaiseClosureNameError("null_output"); __PYX_ERR(0, 228, __pyx_L1_error) }
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_null_output);
      __pyx_r = __pyx_cur_scope->__pyx_v_null_output;
      goto __pyx_L0;

      /* "validr/_validator_c.pyx":227
 *                     if has_default:
 *                         return default
 *                     elif optional:             # <<<<<<<<<<<<<<
 *                         return null_output
 *                     else:
 */
    }

    /* "validr/_validator_c.pyx":230
 *                         return null_output
 *                     else:
 *                         raise Invalid('required')             # <<<<<<<<<<<<<<
 *                 return value
 * 
 */
    __Pyx_TraceLine(230,0,__PYX_ERR(0, 230, __pyx_L1_error))
    /*else*/ {
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 230, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_n_u_required) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_n_u_required);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 230, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 230, __pyx_L1_error)
    }

    /* "validr/_validator_c.pyx":224
 *                 else:
 *                     is_null = value is None
 *                 if is_null:             # <<<<<<<<<<<<<<
 *                     if has_default:
 *                         return default
 */
  }

  /* "validr/_validator_c.pyx":231
 *                     else:
 *                         raise Invalid('required')
 *                 return value             # <<<<<<<<<<<<<<
 * 
 *             supress_invalid = has_invalid_to or invalid_to_default
 */
  __Pyx_TraceLine(231,0,__PYX_ERR(0, 231, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_value);
  __pyx_r = __pyx_v_value;
  goto __pyx_L0;

  /* "validr/_validator_c.pyx":203
 * 
 *             # check null, empty string and default
 *             def _m_validate(value):             # <<<<<<<<<<<<<<
 *                 cdef bint is_null
 *                 if accept_string:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("validr._validator_c.validator.decorator._m_validator._m_validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":235
 *             supress_invalid = has_invalid_to or invalid_to_default
 * 
 *             def m_validate(value):             # <<<<<<<<<<<<<<
 *                 try:
 *                     return _m_validate(value)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_9validator_9decorator_12_m_validator_3m_validate(PyObject *__pyx_self, PyObject *__pyx_v_value); /*proto*/
static PyMethodDef __pyx_mdef_6validr_12_validator_c_9validator_9decorator_12_m_validator_3m_validate = {"m_validate", (PyCFunction)__pyx_pw_6validr_12_validator_c_9validator_9decorator_12_m_validator_3m_validate, METH_O, 0};
static PyObject *__pyx_pw_6validr_12_validator_c_9validator_9decorator_12_m_validator_3m_validate(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("m_validate (wrapper)", 0);
  __pyx_r = __pyx_pf_6validr_12_validator_c_9validator_9decorator_12_m_validator_2m_validate(__pyx_self, ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6validr_12_validator_c_9validator_9decorator_12_m_validator_2m_validate(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_2__m_validator *__pyx_cur_scope;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_2__m_validator *__pyx_outer_scope;
  PyObject *__pyx_v_ex = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  char const *__pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  __Pyx_RefNannySetupContext("m_validate", 0);
  __pyx_outer_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_2__m_validator *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_TraceCall("m_validate", __pyx_f[0], 235, 0, __PYX_ERR(0, 235, __pyx_L1_error));

  /* "validr/_validator_c.pyx":236
 * 
 *             def m_validate(value):
 *                 try:             # <<<<<<<<<<<<<<
 *                     return _m_validate(value)
 *                 except Invalid as ex:
 */
  __Pyx_TraceLine(236,0,__PYX_ERR(0, 236, __pyx_L1_error))
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "validr/_validator_c.pyx":237
 *             def m_validate(value):
 *                 try:
 *                     return _m_validate(value)             # <<<<<<<<<<<<<<
 *                 except Invalid as ex:
 *                     ex.set_value(value)
 */
      __Pyx_TraceLine(237,0,__PYX_ERR(0, 237, __pyx_L3_error))
      __Pyx_XDECREF(__pyx_r);
      if (unlikely(!__pyx_cur_scope->__pyx_v__m_validate)) { __Pyx_RaiseClosureNameError("_m_validate"); __PYX_ERR(0, 237, __pyx_L3_error) }
      __pyx_t_4 = __pyx_pf_6validr_12_validator_c_9validator_9decorator_12_m_validator__m_validate(__pyx_cur_scope->__pyx_v__m_validate, __pyx_v_value); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 237, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_r = __pyx_t_4;
      __pyx_t_4 = 0;
      goto __pyx_L7_try_return;

      /* "validr/_validator_c.pyx":236
 * 
 *             def m_validate(value):
 *                 try:             # <<<<<<<<<<<<<<
 *                     return _m_validate(value)
 *                 except Invalid as ex:
 */
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "validr/_validator_c.pyx":238
 *                 try:
 *                     return _m_validate(value)
 *                 except Invalid as ex:             # <<<<<<<<<<<<<<
 *                     ex.set_value(value)
 *                     if supress_invalid:
 */
    __Pyx_TraceLine(238,0,__PYX_ERR(0, 238, __pyx_L5_except_error))
    __Pyx_ErrFetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 238, __pyx_L5_except_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_4, __pyx_t_7);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_ErrRestore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
    __pyx_t_4 = 0; __pyx_t_5 = 0; __pyx_t_6 = 0;
    if (__pyx_t_8) {
      __Pyx_AddTraceback("validr._validator_c.validator.decorator._m_validator.m_validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_5, &__pyx_t_4) < 0) __PYX_ERR(0, 238, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __pyx_v_ex = __pyx_t_5;
      /*try:*/ {

        /* "validr/_validator_c.pyx":239
 *                     return _m_validate(value)
 *                 except Invalid as ex:
 *                     ex.set_value(value)             # <<<<<<<<<<<<<<
 *                     if supress_invalid:
 *                         return invalid_to
 */
        __Pyx_TraceLine(239,0,__PYX_ERR(0, 239, __pyx_L14_error))
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_ex, __pyx_n_s_set_value); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 239, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_10 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
          __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
          if (likely(__pyx_t_10)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
            __Pyx_INCREF(__pyx_t_10);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_9, function);
          }
        }
        __pyx_t_7 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_10, __pyx_v_value) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_v_value);
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 239, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "validr/_validator_c.pyx":240
 *                 except Invalid as ex:
 *                     ex.set_value(value)
 *                     if supress_invalid:             # <<<<<<<<<<<<<<
 *                         return invalid_to
 *                     else:
 */
        __Pyx_TraceLine(240,0,__PYX_ERR(0, 240, __pyx_L14_error))
        __pyx_t_11 = (__pyx_cur_scope->__pyx_v_supress_invalid != 0);
        if (likely(__pyx_t_11)) {

          /* "validr/_validator_c.pyx":241
 *                     ex.set_value(value)
 *                     if supress_invalid:
 *                         return invalid_to             # <<<<<<<<<<<<<<
 *                     else:
 *                         raise
 */
          __Pyx_TraceLine(241,0,__PYX_ERR(0, 241, __pyx_L14_error))
          __Pyx_XDECREF(__pyx_r);
          if (unlikely(!__pyx_cur_scope->__pyx_v_invalid_to)) { __Pyx_RaiseClosureNameError("invalid_to"); __PYX_ERR(0, 241, __pyx_L14_error) }
          __Pyx_INCREF(__pyx_cur_scope->__pyx_v_invalid_to);
          __pyx_r = __pyx_cur_scope->__pyx_v_invalid_to;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          goto __pyx_L13_return;

          /* "validr/_validator_c.pyx":240
 *                 except Invalid as ex:
 *                     ex.set_value(value)
 *                     if supress_invalid:             # <<<<<<<<<<<<<<
 *                         return invalid_to
 *                     else:
 */
        }

        /* "validr/_validator_c.pyx":243
 *                         return invalid_to
 *                     else:
 *                         raise             # <<<<<<<<<<<<<<
 * 
 *             _update_validate_func_info(m_validate, f, schema)
 */
        __Pyx_TraceLine(243,0,__PYX_ERR(0, 243, __pyx_L14_error))
        /*else*/ {
          __Pyx_GIVEREF(__pyx_t_6);
          __Pyx_GIVEREF(__pyx_t_5);
          __Pyx_XGIVEREF(__pyx_t_4);
          __Pyx_ErrRestoreWithState(__pyx_t_6, __pyx_t_5, __pyx_t_4);
          __pyx_t_6 = 0; __pyx_t_5 = 0; __pyx_t_4 = 0; 
          __PYX_ERR(0, 243, __pyx_L14_error)
        }
      }

      /* "validr/_validator_c.pyx":238
 *                 try:
 *                     return _m_validate(value)
 *                 except Invalid as ex:             # <<<<<<<<<<<<<<
 *                     ex.set_value(value)
 *                     if supress_invalid:
 */
      __Pyx_TraceLine(238,0,__PYX_ERR(0, 238, __pyx_L14_error))
      /*finally:*/ {
        __pyx_L14_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16) < 0)) __Pyx_ErrFetch(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_18);
          __Pyx_XGOTREF(__pyx_t_19);
          __pyx_t_8 = __pyx_lineno; __pyx_t_12 = __pyx_clineno; __pyx_t_13 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_ex);
            __pyx_v_ex = NULL;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_XGIVEREF(__pyx_t_19);
            __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
          }
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_XGIVEREF(__pyx_t_16);
          __Pyx_ErrRestore(__pyx_t_14, __pyx_t_15, __pyx_t_16);
          __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
          __pyx_lineno = __pyx_t_8; __pyx_clineno = __pyx_t_12; __pyx_filename = __pyx_t_13;
          goto __pyx_L5_except_error;
        }
        __pyx_L13_return: {
          __pyx_t_19 = __pyx_r;
          __pyx_r = 0;
          __Pyx_DECREF(__pyx_v_ex);
          __pyx_v_ex = NULL;
          __pyx_r = __pyx_t_19;
          __pyx_t_19 = 0;
          goto __pyx_L6_except_return;
        }
      }
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "validr/_validator_c.pyx":236
 * 
 *             def m_validate(value):
 *                 try:             # <<<<<<<<<<<<<<
 *                     return _m_validate(value)
 *                 except Invalid as ex:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L7_try_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
  }

  /* "validr/_validator_c.pyx":235
 *             supress_invalid = has_invalid_to or invalid_to_default
 * 
 *             def m_validate(value):             # <<<<<<<<<<<<<<
 *                 try:
 *                     return _m_validate(value)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("validr._validator_c.validator.decorator._m_validator.m_validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ex);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":147
 *     def decorator(f):
 * 
 *         def _m_validator(compiler, schema):             # <<<<<<<<<<<<<<
 *             params = schema.params.copy()
 *             if schema.items is not None:
 */

static PyObject *__pyx_pf_6validr_12_validator_c_9validator_9decorator__m_validator(PyObject *__pyx_self, PyObject *__pyx_v_compiler, PyObject *__pyx_v_schema) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_2__m_validator *__pyx_cur_scope;
  PyObject *__pyx_v_params = NULL;
  int __pyx_v_local_output_object;
  int __pyx_v_invalid_to_default;
  int __pyx_v_has_invalid_to;
  PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_v_msg = NULL;
  PyObject *__pyx_v_m_validate = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_t_15;
  char const *__pyx_t_16;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  __Pyx_RefNannySetupContext("_m_validator", 0);
  __pyx_cur_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_2__m_validator *)__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_2__m_validator(__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_2__m_validator, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_2__m_validator *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 147, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_1_decorator *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  __Pyx_TraceCall("_m_validator", __pyx_f[0], 147, 0, __PYX_ERR(0, 147, __pyx_L1_error));

  /* "validr/_validator_c.pyx":148
 * 
 *         def _m_validator(compiler, schema):
 *             params = schema.params.copy()             # <<<<<<<<<<<<<<
 *             if schema.items is not None:
 *                 params['items'] = schema.items
 */
  __Pyx_TraceLine(148,0,__PYX_ERR(0, 148, __pyx_L1_error))
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_schema, __pyx_n_s_params); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_copy); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_params = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":149
 *         def _m_validator(compiler, schema):
 *             params = schema.params.copy()
 *             if schema.items is not None:             # <<<<<<<<<<<<<<
 *                 params['items'] = schema.items
 *             cdef bint local_output_object = output_object
 */
  __Pyx_TraceLine(149,0,__PYX_ERR(0, 149, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_schema, __pyx_n_s_items); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = (__pyx_t_1 != Py_None);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {

    /* "validr/_validator_c.pyx":150
 *             params = schema.params.copy()
 *             if schema.items is not None:
 *                 params['items'] = schema.items             # <<<<<<<<<<<<<<
 *             cdef bint local_output_object = output_object
 *             if output_string and output_object:
 */
    __Pyx_TraceLine(150,0,__PYX_ERR(0, 150, __pyx_L1_error))
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_schema, __pyx_n_s_items); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 150, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(PyObject_SetItem(__pyx_v_params, __pyx_n_u_items, __pyx_t_1) < 0)) __PYX_ERR(0, 150, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "validr/_validator_c.pyx":149
 *         def _m_validator(compiler, schema):
 *             params = schema.params.copy()
 *             if schema.items is not None:             # <<<<<<<<<<<<<<
 *                 params['items'] = schema.items
 *             cdef bint local_output_object = output_object
 */
  }

  /* "validr/_validator_c.pyx":151
 *             if schema.items is not None:
 *                 params['items'] = schema.items
 *             cdef bint local_output_object = output_object             # <<<<<<<<<<<<<<
 *             if output_string and output_object:
 *                 local_output_object = bool(params.get('object', None))
 */
  __Pyx_TraceLine(151,0,__PYX_ERR(0, 151, __pyx_L1_error))
  __pyx_v_local_output_object = __pyx_cur_scope->__pyx_outer_scope->__pyx_outer_scope->__pyx_v_output_object;

  /* "validr/_validator_c.pyx":152
 *                 params['items'] = schema.items
 *             cdef bint local_output_object = output_object
 *             if output_string and output_object:             # <<<<<<<<<<<<<<
 *                 local_output_object = bool(params.get('object', None))
 *                 # TODO: fix special case of timedelta
 */
  __Pyx_TraceLine(152,0,__PYX_ERR(0, 152, __pyx_L1_error))
  __pyx_t_4 = (__pyx_cur_scope->__pyx_outer_scope->__pyx_outer_scope->__pyx_v_output_string != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_5 = __pyx_t_4;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_4 = (__pyx_cur_scope->__pyx_outer_scope->__pyx_outer_scope->__pyx_v_output_object != 0);
  __pyx_t_5 = __pyx_t_4;
  __pyx_L5_bool_binop_done:;
  if (__pyx_t_5) {

    /* "validr/_validator_c.pyx":153
 *             cdef bint local_output_object = output_object
 *             if output_string and output_object:
 *                 local_output_object = bool(params.get('object', None))             # <<<<<<<<<<<<<<
 *                 # TODO: fix special case of timedelta
 *                 if schema.validator == 'timedelta':
 */
    __Pyx_TraceLine(153,0,__PYX_ERR(0, 153, __pyx_L1_error))
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_params, __pyx_n_s_get); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 153, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 153, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 153, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_local_output_object = (!(!__pyx_t_5));

    /* "validr/_validator_c.pyx":155
 *                 local_output_object = bool(params.get('object', None))
 *                 # TODO: fix special case of timedelta
 *                 if schema.validator == 'timedelta':             # <<<<<<<<<<<<<<
 *                     local_output_object = not bool(params.get('string', None))
 *             if output_object and 'object' in params:
 */
    __Pyx_TraceLine(155,0,__PYX_ERR(0, 155, __pyx_L1_error))
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_schema, __pyx_n_s_validator); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 155, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_t_3, __pyx_n_u_timedelta, Py_EQ)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 155, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_5) {

      /* "validr/_validator_c.pyx":156
 *                 # TODO: fix special case of timedelta
 *                 if schema.validator == 'timedelta':
 *                     local_output_object = not bool(params.get('string', None))             # <<<<<<<<<<<<<<
 *             if output_object and 'object' in params:
 *                 params['output_object'] = bool(params.pop('object', None))
 */
      __Pyx_TraceLine(156,0,__PYX_ERR(0, 156, __pyx_L1_error))
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_params, __pyx_n_s_get); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 156, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 156, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 156, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_local_output_object = (!((!(!__pyx_t_5)) != 0));

      /* "validr/_validator_c.pyx":155
 *                 local_output_object = bool(params.get('object', None))
 *                 # TODO: fix special case of timedelta
 *                 if schema.validator == 'timedelta':             # <<<<<<<<<<<<<<
 *                     local_output_object = not bool(params.get('string', None))
 *             if output_object and 'object' in params:
 */
    }

    /* "validr/_validator_c.pyx":152
 *                 params['items'] = schema.items
 *             cdef bint local_output_object = output_object
 *             if output_string and output_object:             # <<<<<<<<<<<<<<
 *                 local_output_object = bool(params.get('object', None))
 *                 # TODO: fix special case of timedelta
 */
  }

  /* "validr/_validator_c.pyx":157
 *                 if schema.validator == 'timedelta':
 *                     local_output_object = not bool(params.get('string', None))
 *             if output_object and 'object' in params:             # <<<<<<<<<<<<<<
 *                 params['output_object'] = bool(params.pop('object', None))
 *             if local_output_object:
 */
  __Pyx_TraceLine(157,0,__PYX_ERR(0, 157, __pyx_L1_error))
  __pyx_t_4 = (__pyx_cur_scope->__pyx_outer_scope->__pyx_outer_scope->__pyx_v_output_object != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_5 = __pyx_t_4;
    goto __pyx_L9_bool_binop_done;
  }
  __pyx_t_4 = (__Pyx_PySequence_ContainsTF(__pyx_n_u_object, __pyx_v_params, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 157, __pyx_L1_error)
  __pyx_t_6 = (__pyx_t_4 != 0);
  __pyx_t_5 = __pyx_t_6;
  __pyx_L9_bool_binop_done:;
  if (__pyx_t_5) {

    /* "validr/_validator_c.pyx":158
 *                     local_output_object = not bool(params.get('string', None))
 *             if output_object and 'object' in params:
 *                 params['output_object'] = bool(params.pop('object', None))             # <<<<<<<<<<<<<<
 *             if local_output_object:
 *                 null_output = None
 */
    __Pyx_TraceLine(158,0,__PYX_ERR(0, 158, __pyx_L1_error))
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_params, __pyx_n_s_pop); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 158, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 158, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 158, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyBool_FromLong((!(!__pyx_t_5))); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 158, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (unlikely(PyObject_SetItem(__pyx_v_params, __pyx_n_u_output_object, __pyx_t_3) < 0)) __PYX_ERR(0, 158, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "validr/_validator_c.pyx":157
 *                 if schema.validator == 'timedelta':
 *                     local_output_object = not bool(params.get('string', None))
 *             if output_object and 'object' in params:             # <<<<<<<<<<<<<<
 *                 params['output_object'] = bool(params.pop('object', None))
 *             if local_output_object:
 */
  }

  /* "validr/_validator_c.pyx":159
 *             if output_object and 'object' in params:
 *                 params['output_object'] = bool(params.pop('object', None))
 *             if local_output_object:             # <<<<<<<<<<<<<<
 *                 null_output = None
 *             else:
 */
  __Pyx_TraceLine(159,0,__PYX_ERR(0, 159, __pyx_L1_error))
  __pyx_t_5 = (__pyx_v_local_output_object != 0);
  if (__pyx_t_5) {

    /* "validr/_validator_c.pyx":160
 *                 params['output_object'] = bool(params.pop('object', None))
 *             if local_output_object:
 *                 null_output = None             # <<<<<<<<<<<<<<
 *             else:
 *                 null_output = ''
 */
    __Pyx_TraceLine(160,0,__PYX_ERR(0, 160, __pyx_L1_error))
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __pyx_cur_scope->__pyx_v_null_output = ((PyObject*)Py_None);

    /* "validr/_validator_c.pyx":159
 *             if output_object and 'object' in params:
 *                 params['output_object'] = bool(params.pop('object', None))
 *             if local_output_object:             # <<<<<<<<<<<<<<
 *                 null_output = None
 *             else:
 */
    goto __pyx_L11;
  }

  /* "validr/_validator_c.pyx":162
 *                 null_output = None
 *             else:
 *                 null_output = ''             # <<<<<<<<<<<<<<
 *             cdef bint optional = params.pop('optional', False)
 *             default = params.pop('default', None)
 */
  __Pyx_TraceLine(162,0,__PYX_ERR(0, 162, __pyx_L1_error))
  /*else*/ {
    __Pyx_INCREF(__pyx_kp_u_);
    __Pyx_GIVEREF(__pyx_kp_u_);
    __pyx_cur_scope->__pyx_v_null_output = __pyx_kp_u_;
  }
  __pyx_L11:;

  /* "validr/_validator_c.pyx":163
 *             else:
 *                 null_output = ''
 *             cdef bint optional = params.pop('optional', False)             # <<<<<<<<<<<<<<
 *             default = params.pop('default', None)
 *             params.pop('desc', None)
 */
  __Pyx_TraceLine(163,0,__PYX_ERR(0, 163, __pyx_L1_error))
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_params, __pyx_n_s_pop); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 163, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_cur_scope->__pyx_v_optional = __pyx_t_5;

  /* "validr/_validator_c.pyx":164
 *                 null_output = ''
 *             cdef bint optional = params.pop('optional', False)
 *             default = params.pop('default', None)             # <<<<<<<<<<<<<<
 *             params.pop('desc', None)
 *             cdef bint invalid_to_default = params.pop('invalid_to_default', False)
 */
  __Pyx_TraceLine(164,0,__PYX_ERR(0, 164, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_params, __pyx_n_s_pop); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__8, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_cur_scope->__pyx_v_default = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "validr/_validator_c.pyx":165
 *             cdef bint optional = params.pop('optional', False)
 *             default = params.pop('default', None)
 *             params.pop('desc', None)             # <<<<<<<<<<<<<<
 *             cdef bint invalid_to_default = params.pop('invalid_to_default', False)
 *             cdef bint has_invalid_to = 'invalid_to' in params
 */
  __Pyx_TraceLine(165,0,__PYX_ERR(0, 165, __pyx_L1_error))
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_params, __pyx_n_s_pop); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":166
 *             default = params.pop('default', None)
 *             params.pop('desc', None)
 *             cdef bint invalid_to_default = params.pop('invalid_to_default', False)             # <<<<<<<<<<<<<<
 *             cdef bint has_invalid_to = 'invalid_to' in params
 *             invalid_to = params.pop('invalid_to', None)
 */
  __Pyx_TraceLine(166,0,__PYX_ERR(0, 166, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_params, __pyx_n_s_pop); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__10, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 166, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_invalid_to_default = __pyx_t_5;

  /* "validr/_validator_c.pyx":167
 *             params.pop('desc', None)
 *             cdef bint invalid_to_default = params.pop('invalid_to_default', False)
 *             cdef bint has_invalid_to = 'invalid_to' in params             # <<<<<<<<<<<<<<
 *             invalid_to = params.pop('invalid_to', None)
 *             cdef bint has_default
 */
  __Pyx_TraceLine(167,0,__PYX_ERR(0, 167, __pyx_L1_error))
  __pyx_t_5 = (__Pyx_PySequence_ContainsTF(__pyx_n_u_invalid_to, __pyx_v_params, Py_EQ)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 167, __pyx_L1_error)
  __pyx_v_has_invalid_to = __pyx_t_5;

  /* "validr/_validator_c.pyx":168
 *             cdef bint invalid_to_default = params.pop('invalid_to_default', False)
 *             cdef bint has_invalid_to = 'invalid_to' in params
 *             invalid_to = params.pop('invalid_to', None)             # <<<<<<<<<<<<<<
 *             cdef bint has_default
 *             if accept_string:
 */
  __Pyx_TraceLine(168,0,__PYX_ERR(0, 168, __pyx_L1_error))
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_params, __pyx_n_s_pop); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__11, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_invalid_to = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":170
 *             invalid_to = params.pop('invalid_to', None)
 *             cdef bint has_default
 *             if accept_string:             # <<<<<<<<<<<<<<
 *                 has_default = not (default is None or default == '')
 *             else:
 */
  __Pyx_TraceLine(170,0,__PYX_ERR(0, 170, __pyx_L1_error))
  __pyx_t_5 = (__pyx_cur_scope->__pyx_outer_scope->__pyx_outer_scope->__pyx_v_accept_string != 0);
  if (__pyx_t_5) {

    /* "validr/_validator_c.pyx":171
 *             cdef bint has_default
 *             if accept_string:
 *                 has_default = not (default is None or default == '')             # <<<<<<<<<<<<<<
 *             else:
 *                 has_default = not (default is None)
 */
    __Pyx_TraceLine(171,0,__PYX_ERR(0, 171, __pyx_L1_error))
    __pyx_t_6 = (__pyx_cur_scope->__pyx_v_default == Py_None);
    __pyx_t_4 = (__pyx_t_6 != 0);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_5 = __pyx_t_4;
      goto __pyx_L13_bool_binop_done;
    }
    __pyx_t_4 = (__Pyx_PyUnicode_Equals(__pyx_cur_scope->__pyx_v_default, __pyx_kp_u_, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 171, __pyx_L1_error)
    __pyx_t_5 = __pyx_t_4;
    __pyx_L13_bool_binop_done:;
    __pyx_cur_scope->__pyx_v_has_default = (!__pyx_t_5);

    /* "validr/_validator_c.pyx":170
 *             invalid_to = params.pop('invalid_to', None)
 *             cdef bint has_default
 *             if accept_string:             # <<<<<<<<<<<<<<
 *                 has_default = not (default is None or default == '')
 *             else:
 */
    goto __pyx_L12;
  }

  /* "validr/_validator_c.pyx":173
 *                 has_default = not (default is None or default == '')
 *             else:
 *                 has_default = not (default is None)             # <<<<<<<<<<<<<<
 *             if has_invalid_to and invalid_to_default:
 *                 raise SchemaError('can not set both invalid_to and invalid_to_default')
 */
  __Pyx_TraceLine(173,0,__PYX_ERR(0, 173, __pyx_L1_error))
  /*else*/ {
    __pyx_t_5 = (__pyx_cur_scope->__pyx_v_default != Py_None);
    __pyx_cur_scope->__pyx_v_has_default = __pyx_t_5;
  }
  __pyx_L12:;

  /* "validr/_validator_c.pyx":174
 *             else:
 *                 has_default = not (default is None)
 *             if has_invalid_to and invalid_to_default:             # <<<<<<<<<<<<<<
 *                 raise SchemaError('can not set both invalid_to and invalid_to_default')
 *             if invalid_to_default and (not has_default) and (not optional):
 */
  __Pyx_TraceLine(174,0,__PYX_ERR(0, 174, __pyx_L1_error))
  __pyx_t_4 = (__pyx_v_has_invalid_to != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_5 = __pyx_t_4;
    goto __pyx_L16_bool_binop_done;
  }
  __pyx_t_4 = (__pyx_v_invalid_to_default != 0);
  __pyx_t_5 = __pyx_t_4;
  __pyx_L16_bool_binop_done:;
  if (unlikely(__pyx_t_5)) {

    /* "validr/_validator_c.pyx":175
 *                 has_default = not (default is None)
 *             if has_invalid_to and invalid_to_default:
 *                 raise SchemaError('can not set both invalid_to and invalid_to_default')             # <<<<<<<<<<<<<<
 *             if invalid_to_default and (not has_default) and (not optional):
 *                 raise SchemaError('default or optional must be set when set invalid_to_default')
 */
    __Pyx_TraceLine(175,0,__PYX_ERR(0, 175, __pyx_L1_error))
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_SchemaError); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 175, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_kp_u_can_not_set_both_invalid_to_and) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_kp_u_can_not_set_both_invalid_to_and);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 175, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 175, __pyx_L1_error)

    /* "validr/_validator_c.pyx":174
 *             else:
 *                 has_default = not (default is None)
 *             if has_invalid_to and invalid_to_default:             # <<<<<<<<<<<<<<
 *                 raise SchemaError('can not set both invalid_to and invalid_to_default')
 *             if invalid_to_default and (not has_default) and (not optional):
 */
  }

  /* "validr/_validator_c.pyx":176
 *             if has_invalid_to and invalid_to_default:
 *                 raise SchemaError('can not set both invalid_to and invalid_to_default')
 *             if invalid_to_default and (not has_default) and (not optional):             # <<<<<<<<<<<<<<
 *                 raise SchemaError('default or optional must be set when set invalid_to_default')
 *             try:
 */
  __Pyx_TraceLine(176,0,__PYX_ERR(0, 176, __pyx_L1_error))
  __pyx_t_4 = (__pyx_v_invalid_to_default != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_5 = __pyx_t_4;
    goto __pyx_L19_bool_binop_done;
  }
  __pyx_t_4 = ((!(__pyx_cur_scope->__pyx_v_has_default != 0)) != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_5 = __pyx_t_4;
    goto __pyx_L19_bool_binop_done;
  }
  __pyx_t_4 = ((!(__pyx_cur_scope->__pyx_v_optional != 0)) != 0);
  __pyx_t_5 = __pyx_t_4;
  __pyx_L19_bool_binop_done:;
  if (unlikely(__pyx_t_5)) {

    /* "validr/_validator_c.pyx":177
 *                 raise SchemaError('can not set both invalid_to and invalid_to_default')
 *             if invalid_to_default and (not has_default) and (not optional):
 *                 raise SchemaError('default or optional must be set when set invalid_to_default')             # <<<<<<<<<<<<<<
 *             try:
 *                 validate = f(compiler, **params)
 */
    __Pyx_TraceLine(177,0,__PYX_ERR(0, 177, __pyx_L1_error))
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_SchemaError); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_kp_u_default_or_optional_must_be_set) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_kp_u_default_or_optional_must_be_set);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 177, __pyx_L1_error)

    /* "validr/_validator_c.pyx":176
 *             if has_invalid_to and invalid_to_default:
 *                 raise SchemaError('can not set both invalid_to and invalid_to_default')
 *             if invalid_to_default and (not has_default) and (not optional):             # <<<<<<<<<<<<<<
 *                 raise SchemaError('default or optional must be set when set invalid_to_default')
 *             try:
 */
  }

  /* "validr/_validator_c.pyx":178
 *             if invalid_to_default and (not has_default) and (not optional):
 *                 raise SchemaError('default or optional must be set when set invalid_to_default')
 *             try:             # <<<<<<<<<<<<<<
 *                 validate = f(compiler, **params)
 *             except TypeError as e:
 */
  __Pyx_TraceLine(178,0,__PYX_ERR(0, 178, __pyx_L1_error))
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
    __Pyx_XGOTREF(__pyx_t_7);
    __Pyx_XGOTREF(__pyx_t_8);
    __Pyx_XGOTREF(__pyx_t_9);
    /*try:*/ {

      /* "validr/_validator_c.pyx":179
 *                 raise SchemaError('default or optional must be set when set invalid_to_default')
 *             try:
 *                 validate = f(compiler, **params)             # <<<<<<<<<<<<<<
 *             except TypeError as e:
 *                 raise SchemaError(str(e)) from None
 */
      __Pyx_TraceLine(179,0,__PYX_ERR(0, 179, __pyx_L22_error))
      if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_f)) { __Pyx_RaiseClosureNameError("f"); __PYX_ERR(0, 179, __pyx_L22_error) }
      __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 179, __pyx_L22_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_v_compiler);
      __Pyx_GIVEREF(__pyx_v_compiler);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_compiler);
      if (unlikely(__pyx_v_params == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "argument after ** must be a mapping, not NoneType");
        __PYX_ERR(0, 179, __pyx_L22_error)
      }
      if (likely(PyDict_CheckExact(__pyx_v_params))) {
        __pyx_t_3 = PyDict_Copy(__pyx_v_params); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 179, __pyx_L22_error)
        __Pyx_GOTREF(__pyx_t_3);
      } else {
        __pyx_t_3 = PyObject_CallFunctionObjArgs((PyObject*)&PyDict_Type, __pyx_v_params, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 179, __pyx_L22_error)
        __Pyx_GOTREF(__pyx_t_3);
      }
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_f, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 179, __pyx_L22_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GIVEREF(__pyx_t_2);
      __pyx_cur_scope->__pyx_v_validate = __pyx_t_2;
      __pyx_t_2 = 0;

      /* "validr/_validator_c.pyx":178
 *             if invalid_to_default and (not has_default) and (not optional):
 *                 raise SchemaError('default or optional must be set when set invalid_to_default')
 *             try:             # <<<<<<<<<<<<<<
 *                 validate = f(compiler, **params)
 *             except TypeError as e:
 */
    }
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    goto __pyx_L27_try_end;
    __pyx_L22_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "validr/_validator_c.pyx":180
 *             try:
 *                 validate = f(compiler, **params)
 *             except TypeError as e:             # <<<<<<<<<<<<<<
 *                 raise SchemaError(str(e)) from None
 *             # check default value
 */
    __Pyx_TraceLine(180,0,__PYX_ERR(0, 180, __pyx_L24_except_error))
    __pyx_t_10 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError);
    if (__pyx_t_10) {
      __Pyx_AddTraceback("validr._validator_c.validator.decorator._m_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_3, &__pyx_t_1) < 0) __PYX_ERR(0, 180, __pyx_L24_except_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __pyx_v_e = __pyx_t_3;
      /*try:*/ {

        /* "validr/_validator_c.pyx":181
 *                 validate = f(compiler, **params)
 *             except TypeError as e:
 *                 raise SchemaError(str(e)) from None             # <<<<<<<<<<<<<<
 *             # check default value
 *             if has_default:
 */
        __Pyx_TraceLine(181,0,__PYX_ERR(0, 181, __pyx_L33_error))
        __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_SchemaError); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 181, __pyx_L33_error)
        __Pyx_GOTREF(__pyx_t_12);
        __pyx_t_13 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyUnicode_Type)), __pyx_v_e); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 181, __pyx_L33_error)
        __Pyx_GOTREF(__pyx_t_13);
        __pyx_t_14 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_12))) {
          __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_12);
          if (likely(__pyx_t_14)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
            __Pyx_INCREF(__pyx_t_14);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_12, function);
          }
        }
        __pyx_t_11 = (__pyx_t_14) ? __Pyx_PyObject_Call2Args(__pyx_t_12, __pyx_t_14, __pyx_t_13) : __Pyx_PyObject_CallOneArg(__pyx_t_12, __pyx_t_13);
        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 181, __pyx_L33_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_Raise(__pyx_t_11, 0, 0, Py_None);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __PYX_ERR(0, 181, __pyx_L33_error)
      }

      /* "validr/_validator_c.pyx":180
 *             try:
 *                 validate = f(compiler, **params)
 *             except TypeError as e:             # <<<<<<<<<<<<<<
 *                 raise SchemaError(str(e)) from None
 *             # check default value
 */
      __Pyx_TraceLine(180,0,__PYX_ERR(0, 180, __pyx_L33_error))
      /*finally:*/ {
        __pyx_L33_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0;
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_20, &__pyx_t_21, &__pyx_t_22);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19) < 0)) __Pyx_ErrFetch(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_18);
          __Pyx_XGOTREF(__pyx_t_19);
          __Pyx_XGOTREF(__pyx_t_20);
          __Pyx_XGOTREF(__pyx_t_21);
          __Pyx_XGOTREF(__pyx_t_22);
          __pyx_t_10 = __pyx_lineno; __pyx_t_15 = __pyx_clineno; __pyx_t_16 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_e);
            __pyx_v_e = NULL;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_20);
            __Pyx_XGIVEREF(__pyx_t_21);
            __Pyx_XGIVEREF(__pyx_t_22);
            __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_21, __pyx_t_22);
          }
          __Pyx_XGIVEREF(__pyx_t_17);
          __Pyx_XGIVEREF(__pyx_t_18);
          __Pyx_XGIVEREF(__pyx_t_19);
          __Pyx_ErrRestore(__pyx_t_17, __pyx_t_18, __pyx_t_19);
          __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0;
          __pyx_lineno = __pyx_t_10; __pyx_clineno = __pyx_t_15; __pyx_filename = __pyx_t_16;
          goto __pyx_L24_except_error;
        }
      }
    }
    goto __pyx_L24_except_error;
    __pyx_L24_except_error:;

    /* "validr/_validator_c.pyx":178
 *             if invalid_to_default and (not has_default) and (not optional):
 *                 raise SchemaError('default or optional must be set when set invalid_to_default')
 *             try:             # <<<<<<<<<<<<<<
 *                 validate = f(compiler, **params)
 *             except TypeError as e:
 */
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
    goto __pyx_L1_error;
    __pyx_L27_try_end:;
  }

  /* "validr/_validator_c.pyx":183
 *                 raise SchemaError(str(e)) from None
 *             # check default value
 *             if has_default:             # <<<<<<<<<<<<<<
 *                 try:
 *                     default = validate(default)
 */
  __Pyx_TraceLine(183,0,__PYX_ERR(0, 183, __pyx_L1_error))
  __pyx_t_5 = (__pyx_cur_scope->__pyx_v_has_default != 0);
  if (__pyx_t_5) {

    /* "validr/_validator_c.pyx":184
 *             # check default value
 *             if has_default:
 *                 try:             # <<<<<<<<<<<<<<
 *                     default = validate(default)
 *                 except Invalid:
 */
    __Pyx_TraceLine(184,0,__PYX_ERR(0, 184, __pyx_L1_error))
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_9, &__pyx_t_8, &__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_7);
      /*try:*/ {

        /* "validr/_validator_c.pyx":185
 *             if has_default:
 *                 try:
 *                     default = validate(default)             # <<<<<<<<<<<<<<
 *                 except Invalid:
 *                     msg = 'invalid default value {!r}'.format(default)
 */
        __Pyx_TraceLine(185,0,__PYX_ERR(0, 185, __pyx_L40_error))
        __Pyx_INCREF(__pyx_cur_scope->__pyx_v_validate);
        __pyx_t_3 = __pyx_cur_scope->__pyx_v_validate; __pyx_t_2 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_cur_scope->__pyx_v_default) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_cur_scope->__pyx_v_default);
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 185, __pyx_L40_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_default);
        __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_default, __pyx_t_1);
        __Pyx_GIVEREF(__pyx_t_1);
        __pyx_t_1 = 0;

        /* "validr/_validator_c.pyx":184
 *             # check default value
 *             if has_default:
 *                 try:             # <<<<<<<<<<<<<<
 *                     default = validate(default)
 *                 except Invalid:
 */
      }
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L45_try_end;
      __pyx_L40_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "validr/_validator_c.pyx":186
 *                 try:
 *                     default = validate(default)
 *                 except Invalid:             # <<<<<<<<<<<<<<
 *                     msg = 'invalid default value {!r}'.format(default)
 *                     raise SchemaError(msg) from None
 */
      __Pyx_TraceLine(186,0,__PYX_ERR(0, 186, __pyx_L42_except_error))
      __Pyx_ErrFetch(&__pyx_t_1, &__pyx_t_3, &__pyx_t_2);
      __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 186, __pyx_L42_except_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_15 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_1, __pyx_t_11);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_ErrRestore(__pyx_t_1, __pyx_t_3, __pyx_t_2);
      __pyx_t_1 = 0; __pyx_t_3 = 0; __pyx_t_2 = 0;
      if (__pyx_t_15) {
        __Pyx_AddTraceback("validr._validator_c.validator.decorator._m_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_3, &__pyx_t_1) < 0) __PYX_ERR(0, 186, __pyx_L42_except_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_1);

        /* "validr/_validator_c.pyx":187
 *                     default = validate(default)
 *                 except Invalid:
 *                     msg = 'invalid default value {!r}'.format(default)             # <<<<<<<<<<<<<<
 *                     raise SchemaError(msg) from None
 *                 if invalid_to_default:
 */
        __Pyx_TraceLine(187,0,__PYX_ERR(0, 187, __pyx_L42_except_error))
        __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_invalid_default_value_r, __pyx_n_s_format); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 187, __pyx_L42_except_error)
        __Pyx_GOTREF(__pyx_t_12);
        __pyx_t_13 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
          __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_12);
          if (likely(__pyx_t_13)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
            __Pyx_INCREF(__pyx_t_13);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_12, function);
          }
        }
        __pyx_t_11 = (__pyx_t_13) ? __Pyx_PyObject_Call2Args(__pyx_t_12, __pyx_t_13, __pyx_cur_scope->__pyx_v_default) : __Pyx_PyObject_CallOneArg(__pyx_t_12, __pyx_cur_scope->__pyx_v_default);
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 187, __pyx_L42_except_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __pyx_v_msg = __pyx_t_11;
        __pyx_t_11 = 0;

        /* "validr/_validator_c.pyx":188
 *                 except Invalid:
 *                     msg = 'invalid default value {!r}'.format(default)
 *                     raise SchemaError(msg) from None             # <<<<<<<<<<<<<<
 *                 if invalid_to_default:
 *                     invalid_to = default
 */
        __Pyx_TraceLine(188,0,__PYX_ERR(0, 188, __pyx_L42_except_error))
        __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_SchemaError); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 188, __pyx_L42_except_error)
        __Pyx_GOTREF(__pyx_t_12);
        __pyx_t_13 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_12))) {
          __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_12);
          if (likely(__pyx_t_13)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
            __Pyx_INCREF(__pyx_t_13);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_12, function);
          }
        }
        __pyx_t_11 = (__pyx_t_13) ? __Pyx_PyObject_Call2Args(__pyx_t_12, __pyx_t_13, __pyx_v_msg) : __Pyx_PyObject_CallOneArg(__pyx_t_12, __pyx_v_msg);
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 188, __pyx_L42_except_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_Raise(__pyx_t_11, 0, 0, Py_None);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __PYX_ERR(0, 188, __pyx_L42_except_error)
      }
      goto __pyx_L42_except_error;
      __pyx_L42_except_error:;

      /* "validr/_validator_c.pyx":184
 *             # check default value
 *             if has_default:
 *                 try:             # <<<<<<<<<<<<<<
 *                     default = validate(default)
 *                 except Invalid:
 */
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_8, __pyx_t_7);
      goto __pyx_L1_error;
      __pyx_L45_try_end:;
    }

    /* "validr/_validator_c.pyx":189
 *                     msg = 'invalid default value {!r}'.format(default)
 *                     raise SchemaError(msg) from None
 *                 if invalid_to_default:             # <<<<<<<<<<<<<<
 *                     invalid_to = default
 *             else:
 */
    __Pyx_TraceLine(189,0,__PYX_ERR(0, 189, __pyx_L1_error))
    __pyx_t_5 = (__pyx_v_invalid_to_default != 0);
    if (__pyx_t_5) {

      /* "validr/_validator_c.pyx":190
 *                     raise SchemaError(msg) from None
 *                 if invalid_to_default:
 *                     invalid_to = default             # <<<<<<<<<<<<<<
 *             else:
 *                 if invalid_to_default:
 */
      __Pyx_TraceLine(190,0,__PYX_ERR(0, 190, __pyx_L1_error))
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_default);
      __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_invalid_to);
      __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_invalid_to, __pyx_cur_scope->__pyx_v_default);
      __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_default);

      /* "validr/_validator_c.pyx":189
 *                     msg = 'invalid default value {!r}'.format(default)
 *                     raise SchemaError(msg) from None
 *                 if invalid_to_default:             # <<<<<<<<<<<<<<
 *                     invalid_to = default
 *             else:
 */
    }

    /* "validr/_validator_c.pyx":183
 *                 raise SchemaError(str(e)) from None
 *             # check default value
 *             if has_default:             # <<<<<<<<<<<<<<
 *                 try:
 *                     default = validate(default)
 */
    goto __pyx_L39;
  }

  /* "validr/_validator_c.pyx":192
 *                     invalid_to = default
 *             else:
 *                 if invalid_to_default:             # <<<<<<<<<<<<<<
 *                     invalid_to = null_output
 *             # check invalid_to value
 */
  __Pyx_TraceLine(192,0,__PYX_ERR(0, 192, __pyx_L1_error))
  /*else*/ {
    __pyx_t_5 = (__pyx_v_invalid_to_default != 0);
    if (__pyx_t_5) {

      /* "validr/_validator_c.pyx":193
 *             else:
 *                 if invalid_to_default:
 *                     invalid_to = null_output             # <<<<<<<<<<<<<<
 *             # check invalid_to value
 *             if has_invalid_to:
 */
      __Pyx_TraceLine(193,0,__PYX_ERR(0, 193, __pyx_L1_error))
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_null_output);
      __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_invalid_to);
      __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_invalid_to, __pyx_cur_scope->__pyx_v_null_output);
      __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_null_output);

      /* "validr/_validator_c.pyx":192
 *                     invalid_to = default
 *             else:
 *                 if invalid_to_default:             # <<<<<<<<<<<<<<
 *                     invalid_to = null_output
 *             # check invalid_to value
 */
    }
  }
  __pyx_L39:;

  /* "validr/_validator_c.pyx":195
 *                     invalid_to = null_output
 *             # check invalid_to value
 *             if has_invalid_to:             # <<<<<<<<<<<<<<
 *                 try:
 *                     invalid_to = validate(invalid_to)
 */
  __Pyx_TraceLine(195,0,__PYX_ERR(0, 195, __pyx_L1_error))
  __pyx_t_5 = (__pyx_v_has_invalid_to != 0);
  if (__pyx_t_5) {

    /* "validr/_validator_c.pyx":196
 *             # check invalid_to value
 *             if has_invalid_to:
 *                 try:             # <<<<<<<<<<<<<<
 *                     invalid_to = validate(invalid_to)
 *                 except Invalid:
 */
    __Pyx_TraceLine(196,0,__PYX_ERR(0, 196, __pyx_L1_error))
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      /*try:*/ {

        /* "validr/_validator_c.pyx":197
 *             if has_invalid_to:
 *                 try:
 *                     invalid_to = validate(invalid_to)             # <<<<<<<<<<<<<<
 *                 except Invalid:
 *                     msg = 'invalid invalid_to value {!r}'.format(invalid_to)
 */
        __Pyx_TraceLine(197,0,__PYX_ERR(0, 197, __pyx_L51_error))
        __Pyx_INCREF(__pyx_cur_scope->__pyx_v_validate);
        __pyx_t_3 = __pyx_cur_scope->__pyx_v_validate; __pyx_t_2 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_cur_scope->__pyx_v_invalid_to) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_cur_scope->__pyx_v_invalid_to);
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 197, __pyx_L51_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_invalid_to);
        __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_invalid_to, __pyx_t_1);
        __Pyx_GIVEREF(__pyx_t_1);
        __pyx_t_1 = 0;

        /* "validr/_validator_c.pyx":196
 *             # check invalid_to value
 *             if has_invalid_to:
 *                 try:             # <<<<<<<<<<<<<<
 *                     invalid_to = validate(invalid_to)
 *                 except Invalid:
 */
      }
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      goto __pyx_L56_try_end;
      __pyx_L51_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "validr/_validator_c.pyx":198
 *                 try:
 *                     invalid_to = validate(invalid_to)
 *                 except Invalid:             # <<<<<<<<<<<<<<
 *                     msg = 'invalid invalid_to value {!r}'.format(invalid_to)
 *                     raise SchemaError(msg) from None
 */
      __Pyx_TraceLine(198,0,__PYX_ERR(0, 198, __pyx_L53_except_error))
      __Pyx_ErrFetch(&__pyx_t_1, &__pyx_t_3, &__pyx_t_2);
      __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 198, __pyx_L53_except_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_15 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_1, __pyx_t_11);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_ErrRestore(__pyx_t_1, __pyx_t_3, __pyx_t_2);
      __pyx_t_1 = 0; __pyx_t_3 = 0; __pyx_t_2 = 0;
      if (__pyx_t_15) {
        __Pyx_AddTraceback("validr._validator_c.validator.decorator._m_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_3, &__pyx_t_1) < 0) __PYX_ERR(0, 198, __pyx_L53_except_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_1);

        /* "validr/_validator_c.pyx":199
 *                     invalid_to = validate(invalid_to)
 *                 except Invalid:
 *                     msg = 'invalid invalid_to value {!r}'.format(invalid_to)             # <<<<<<<<<<<<<<
 *                     raise SchemaError(msg) from None
 * 
 */
        __Pyx_TraceLine(199,0,__PYX_ERR(0, 199, __pyx_L53_except_error))
        __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_invalid_invalid_to_value_r, __pyx_n_s_format); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 199, __pyx_L53_except_error)
        __Pyx_GOTREF(__pyx_t_12);
        __pyx_t_13 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
          __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_12);
          if (likely(__pyx_t_13)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
            __Pyx_INCREF(__pyx_t_13);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_12, function);
          }
        }
        __pyx_t_11 = (__pyx_t_13) ? __Pyx_PyObject_Call2Args(__pyx_t_12, __pyx_t_13, __pyx_cur_scope->__pyx_v_invalid_to) : __Pyx_PyObject_CallOneArg(__pyx_t_12, __pyx_cur_scope->__pyx_v_invalid_to);
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 199, __pyx_L53_except_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __pyx_v_msg = __pyx_t_11;
        __pyx_t_11 = 0;

        /* "validr/_validator_c.pyx":200
 *                 except Invalid:
 *                     msg = 'invalid invalid_to value {!r}'.format(invalid_to)
 *                     raise SchemaError(msg) from None             # <<<<<<<<<<<<<<
 * 
 *             # check null, empty string and default
 */
        __Pyx_TraceLine(200,0,__PYX_ERR(0, 200, __pyx_L53_except_error))
        __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_SchemaError); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 200, __pyx_L53_except_error)
        __Pyx_GOTREF(__pyx_t_12);
        __pyx_t_13 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_12))) {
          __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_12);
          if (likely(__pyx_t_13)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
            __Pyx_INCREF(__pyx_t_13);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_12, function);
          }
        }
        __pyx_t_11 = (__pyx_t_13) ? __Pyx_PyObject_Call2Args(__pyx_t_12, __pyx_t_13, __pyx_v_msg) : __Pyx_PyObject_CallOneArg(__pyx_t_12, __pyx_v_msg);
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 200, __pyx_L53_except_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_Raise(__pyx_t_11, 0, 0, Py_None);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __PYX_ERR(0, 200, __pyx_L53_except_error)
      }
      goto __pyx_L53_except_error;
      __pyx_L53_except_error:;

      /* "validr/_validator_c.pyx":196
 *             # check invalid_to value
 *             if has_invalid_to:
 *                 try:             # <<<<<<<<<<<<<<
 *                     invalid_to = validate(invalid_to)
 *                 except Invalid:
 */
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      goto __pyx_L1_error;
      __pyx_L56_try_end:;
    }

    /* "validr/_validator_c.pyx":195
 *                     invalid_to = null_output
 *             # check invalid_to value
 *             if has_invalid_to:             # <<<<<<<<<<<<<<
 *                 try:
 *                     invalid_to = validate(invalid_to)
 */
  }

  /* "validr/_validator_c.pyx":203
 * 
 *             # check null, empty string and default
 *             def _m_validate(value):             # <<<<<<<<<<<<<<
 *                 cdef bint is_null
 *                 if accept_string:
 */
  __Pyx_TraceLine(203,0,__PYX_ERR(0, 203, __pyx_L1_error))
  __pyx_t_1 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_9validator_9decorator_12_m_validator_1_m_validate, 0, __pyx_n_s_validator_locals_decorator_local, ((PyObject*)__pyx_cur_scope), __pyx_n_s_validr__validator_c, __pyx_d, ((PyObject *)__pyx_codeobj__13)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 203, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v__m_validate = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":233
 *                 return value
 * 
 *             supress_invalid = has_invalid_to or invalid_to_default             # <<<<<<<<<<<<<<
 * 
 *             def m_validate(value):
 */
  __Pyx_TraceLine(233,0,__PYX_ERR(0, 233, __pyx_L1_error))
  __pyx_t_4 = (__pyx_v_has_invalid_to != 0);
  if (!__pyx_t_4) {
  } else {
    __pyx_t_5 = __pyx_t_4;
    goto __pyx_L59_bool_binop_done;
  }
  __pyx_t_4 = (__pyx_v_invalid_to_default != 0);
  __pyx_t_5 = __pyx_t_4;
  __pyx_L59_bool_binop_done:;
  __pyx_cur_scope->__pyx_v_supress_invalid = __pyx_t_5;

  /* "validr/_validator_c.pyx":235
 *             supress_invalid = has_invalid_to or invalid_to_default
 * 
 *             def m_validate(value):             # <<<<<<<<<<<<<<
 *                 try:
 *                     return _m_validate(value)
 */
  __Pyx_TraceLine(235,0,__PYX_ERR(0, 235, __pyx_L1_error))
  __pyx_t_1 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_9validator_9decorator_12_m_validator_3m_validate, 0, __pyx_n_s_validator_locals_decorator_local_2, ((PyObject*)__pyx_cur_scope), __pyx_n_s_validr__validator_c, __pyx_d, ((PyObject *)__pyx_codeobj__15)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_m_validate = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":245
 *                         raise
 * 
 *             _update_validate_func_info(m_validate, f, schema)             # <<<<<<<<<<<<<<
 * 
 *             _update_validate_func_type_hints(
 */
  __Pyx_TraceLine(245,0,__PYX_ERR(0, 245, __pyx_L1_error))
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_f)) { __Pyx_RaiseClosureNameError("f"); __PYX_ERR(0, 245, __pyx_L1_error) }
  __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_f;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_3 = __pyx_f_6validr_12_validator_c__update_validate_func_info(__pyx_v_m_validate, __pyx_t_1, __pyx_v_schema, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "validr/_validator_c.pyx":247
 *             _update_validate_func_info(m_validate, f, schema)
 * 
 *             _update_validate_func_type_hints(             # <<<<<<<<<<<<<<
 *                 m_validate, optional=optional, has_default=has_default,
 *                 accept_hints=accept_hints, output_hints=output_hints)
 */
  __Pyx_TraceLine(247,0,__PYX_ERR(0, 247, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_update_validate_func_type_hints); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "validr/_validator_c.pyx":248
 * 
 *             _update_validate_func_type_hints(
 *                 m_validate, optional=optional, has_default=has_default,             # <<<<<<<<<<<<<<
 *                 accept_hints=accept_hints, output_hints=output_hints)
 * 
 */
  __Pyx_TraceLine(248,0,__PYX_ERR(0, 248, __pyx_L1_error))
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_m_validate);
  __Pyx_GIVEREF(__pyx_v_m_validate);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_m_validate);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 248, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_11 = __Pyx_PyBool_FromLong(__pyx_cur_scope->__pyx_v_optional); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 248, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_optional, __pyx_t_11) < 0) __PYX_ERR(0, 248, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __pyx_t_11 = __Pyx_PyBool_FromLong(__pyx_cur_scope->__pyx_v_has_default); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 248, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_has_default, __pyx_t_11) < 0) __PYX_ERR(0, 248, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "validr/_validator_c.pyx":249
 *             _update_validate_func_type_hints(
 *                 m_validate, optional=optional, has_default=has_default,
 *                 accept_hints=accept_hints, output_hints=output_hints)             # <<<<<<<<<<<<<<
 * 
 *             return m_validate
 */
  __Pyx_TraceLine(249,0,__PYX_ERR(0, 249, __pyx_L1_error))
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_outer_scope->__pyx_v_accept_hints)) { __Pyx_RaiseClosureNameError("accept_hints"); __PYX_ERR(0, 249, __pyx_L1_error) }
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_accept_hints, __pyx_cur_scope->__pyx_outer_scope->__pyx_outer_scope->__pyx_v_accept_hints) < 0) __PYX_ERR(0, 248, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_outer_scope->__pyx_v_output_hints)) { __Pyx_RaiseClosureNameError("output_hints"); __PYX_ERR(0, 249, __pyx_L1_error) }
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_output_hints, __pyx_cur_scope->__pyx_outer_scope->__pyx_outer_scope->__pyx_v_output_hints) < 0) __PYX_ERR(0, 248, __pyx_L1_error)

  /* "validr/_validator_c.pyx":247
 *             _update_validate_func_info(m_validate, f, schema)
 * 
 *             _update_validate_func_type_hints(             # <<<<<<<<<<<<<<
 *                 m_validate, optional=optional, has_default=has_default,
 *                 accept_hints=accept_hints, output_hints=output_hints)
 */
  __Pyx_TraceLine(247,0,__PYX_ERR(0, 247, __pyx_L1_error))
  __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "validr/_validator_c.pyx":251
 *                 accept_hints=accept_hints, output_hints=output_hints)
 * 
 *             return m_validate             # <<<<<<<<<<<<<<
 * 
 *         def m_validator(compiler, schema):
 */
  __Pyx_TraceLine(251,0,__PYX_ERR(0, 251, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_m_validate);
  __pyx_r = __pyx_v_m_validate;
  goto __pyx_L0;

  /* "validr/_validator_c.pyx":147
 *     def decorator(f):
 * 
 *         def _m_validator(compiler, schema):             # <<<<<<<<<<<<<<
 *             params = schema.params.copy()
 *             if schema.items is not None:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("validr._validator_c.validator.decorator._m_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_params);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XDECREF(__pyx_v_msg);
  __Pyx_XDECREF(__pyx_v_m_validate);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":253
 *             return m_validate
 * 
 *         def m_validator(compiler, schema):             # <<<<<<<<<<<<<<
 *             try:
 *                 return _m_validator(compiler, schema)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_9validator_9decorator_3m_validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6validr_12_validator_c_9validator_9decorator_3m_validator = {"m_validator", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6validr_12_validator_c_9validator_9decorator_3m_validator, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6validr_12_validator_c_9validator_9decorator_3m_validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_compiler = 0;
  PyObject *__pyx_v_schema = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("m_validator (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_compiler,&__pyx_n_s_schema_2,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_compiler)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_schema_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("m_validator", 1, 2, 2, 1); __PYX_ERR(0, 253, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "m_validator") < 0)) __PYX_ERR(0, 253, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_compiler = values[0];
    __pyx_v_schema = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("m_validator", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 253, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("validr._validator_c.validator.decorator.m_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6validr_12_validator_c_9validator_9decorator_2m_validator(__pyx_self, __pyx_v_compiler, __pyx_v_schema);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6validr_12_validator_c_9validator_9decorator_2m_validator(PyObject *__pyx_self, PyObject *__pyx_v_compiler, PyObject *__pyx_v_schema) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_1_decorator *__pyx_cur_scope;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_1_decorator *__pyx_outer_scope;
  PyObject *__pyx_v_ex = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  char const *__pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  __Pyx_RefNannySetupContext("m_validator", 0);
  __pyx_outer_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_1_decorator *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_TraceCall("m_validator", __pyx_f[0], 253, 0, __PYX_ERR(0, 253, __pyx_L1_error));

  /* "validr/_validator_c.pyx":254
 * 
 *         def m_validator(compiler, schema):
 *             try:             # <<<<<<<<<<<<<<
 *                 return _m_validator(compiler, schema)
 *             except SchemaError as ex:
 */
  __Pyx_TraceLine(254,0,__PYX_ERR(0, 254, __pyx_L1_error))
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "validr/_validator_c.pyx":255
 *         def m_validator(compiler, schema):
 *             try:
 *                 return _m_validator(compiler, schema)             # <<<<<<<<<<<<<<
 *             except SchemaError as ex:
 *                 ex.set_value(schema)
 */
      __Pyx_TraceLine(255,0,__PYX_ERR(0, 255, __pyx_L3_error))
      __Pyx_XDECREF(__pyx_r);
      if (unlikely(!__pyx_cur_scope->__pyx_v__m_validator)) { __Pyx_RaiseClosureNameError("_m_validator"); __PYX_ERR(0, 255, __pyx_L3_error) }
      __pyx_t_4 = __pyx_pf_6validr_12_validator_c_9validator_9decorator__m_validator(__pyx_cur_scope->__pyx_v__m_validator, __pyx_v_compiler, __pyx_v_schema); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 255, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_r = __pyx_t_4;
      __pyx_t_4 = 0;
      goto __pyx_L7_try_return;

      /* "validr/_validator_c.pyx":254
 * 
 *         def m_validator(compiler, schema):
 *             try:             # <<<<<<<<<<<<<<
 *                 return _m_validator(compiler, schema)
 *             except SchemaError as ex:
 */
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "validr/_validator_c.pyx":256
 *             try:
 *                 return _m_validator(compiler, schema)
 *             except SchemaError as ex:             # <<<<<<<<<<<<<<
 *                 ex.set_value(schema)
 *                 raise
 */
    __Pyx_TraceLine(256,0,__PYX_ERR(0, 256, __pyx_L5_except_error))
    __Pyx_ErrFetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_SchemaError); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 256, __pyx_L5_except_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_4, __pyx_t_7);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_ErrRestore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
    __pyx_t_4 = 0; __pyx_t_5 = 0; __pyx_t_6 = 0;
    if (__pyx_t_8) {
      __Pyx_AddTraceback("validr._validator_c.validator.decorator.m_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_5, &__pyx_t_4) < 0) __PYX_ERR(0, 256, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __pyx_v_ex = __pyx_t_5;
      /*try:*/ {

        /* "validr/_validator_c.pyx":257
 *                 return _m_validator(compiler, schema)
 *             except SchemaError as ex:
 *                 ex.set_value(schema)             # <<<<<<<<<<<<<<
 *                 raise
 * 
 */
        __Pyx_TraceLine(257,0,__PYX_ERR(0, 257, __pyx_L14_error))
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_ex, __pyx_n_s_set_value); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 257, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_10 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
          __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
          if (likely(__pyx_t_10)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
            __Pyx_INCREF(__pyx_t_10);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_9, function);
          }
        }
        __pyx_t_7 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_10, __pyx_v_schema) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_v_schema);
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 257, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "validr/_validator_c.pyx":258
 *             except SchemaError as ex:
 *                 ex.set_value(schema)
 *                 raise             # <<<<<<<<<<<<<<
 * 
 *         # TODO: deprecate below attributes because they are implement details
 */
        __Pyx_TraceLine(258,0,__PYX_ERR(0, 258, __pyx_L14_error))
        __Pyx_GIVEREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5);
        __Pyx_XGIVEREF(__pyx_t_4);
        __Pyx_ErrRestoreWithState(__pyx_t_6, __pyx_t_5, __pyx_t_4);
        __pyx_t_6 = 0; __pyx_t_5 = 0; __pyx_t_4 = 0; 
        __PYX_ERR(0, 258, __pyx_L14_error)
      }

      /* "validr/_validator_c.pyx":256
 *             try:
 *                 return _m_validator(compiler, schema)
 *             except SchemaError as ex:             # <<<<<<<<<<<<<<
 *                 ex.set_value(schema)
 *                 raise
 */
      __Pyx_TraceLine(256,0,__PYX_ERR(0, 256, __pyx_L14_error))
      /*finally:*/ {
        __pyx_L14_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0;
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15) < 0)) __Pyx_ErrFetch(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_13);
          __Pyx_XGOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_18);
          __pyx_t_8 = __pyx_lineno; __pyx_t_11 = __pyx_clineno; __pyx_t_12 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_ex);
            __pyx_v_ex = NULL;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ExceptionReset(__pyx_t_16, __pyx_t_17, __pyx_t_18);
          }
          __Pyx_XGIVEREF(__pyx_t_13);
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_ErrRestore(__pyx_t_13, __pyx_t_14, __pyx_t_15);
          __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0;
          __pyx_lineno = __pyx_t_8; __pyx_clineno = __pyx_t_11; __pyx_filename = __pyx_t_12;
          goto __pyx_L5_except_error;
        }
      }
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "validr/_validator_c.pyx":254
 * 
 *         def m_validator(compiler, schema):
 *             try:             # <<<<<<<<<<<<<<
 *                 return _m_validator(compiler, schema)
 *             except SchemaError as ex:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L7_try_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
  }

  /* "validr/_validator_c.pyx":253
 *             return m_validate
 * 
 *         def m_validator(compiler, schema):             # <<<<<<<<<<<<<<
 *             try:
 *                 return _m_validator(compiler, schema)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("validr._validator_c.validator.decorator.m_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ex);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":145
 *     del string, accept, output
 * 
 *     def decorator(f):             # <<<<<<<<<<<<<<
 * 
 *         def _m_validator(compiler, schema):
 */

static PyObject *__pyx_pf_6validr_12_validator_c_9validator_decorator(PyObject *__pyx_self, PyObject *__pyx_v_f) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_1_decorator *__pyx_cur_scope;
  PyObject *__pyx_v_m_validator = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  __Pyx_RefNannySetupContext("decorator", 0);
  __pyx_cur_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_1_decorator *)__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_1_decorator(__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_1_decorator, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_1_decorator *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 145, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct__validator *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  __Pyx_TraceCall("decorator", __pyx_f[0], 145, 0, __PYX_ERR(0, 145, __pyx_L1_error));
  __pyx_cur_scope->__pyx_v_f = __pyx_v_f;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_f);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_f);

  /* "validr/_validator_c.pyx":147
 *     def decorator(f):
 * 
 *         def _m_validator(compiler, schema):             # <<<<<<<<<<<<<<
 *             params = schema.params.copy()
 *             if schema.items is not None:
 */
  __Pyx_TraceLine(147,0,__PYX_ERR(0, 147, __pyx_L1_error))
  __pyx_t_1 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_9validator_9decorator_1_m_validator, 0, __pyx_n_s_validator_locals_decorator_local_3, ((PyObject*)__pyx_cur_scope), __pyx_n_s_validr__validator_c, __pyx_d, ((PyObject *)__pyx_codeobj__17)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v__m_validator = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":253
 *             return m_validate
 * 
 *         def m_validator(compiler, schema):             # <<<<<<<<<<<<<<
 *             try:
 *                 return _m_validator(compiler, schema)
 */
  __Pyx_TraceLine(253,0,__PYX_ERR(0, 253, __pyx_L1_error))
  __pyx_t_1 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_9validator_9decorator_3m_validator, 0, __pyx_n_s_validator_locals_decorator_local_4, ((PyObject*)__pyx_cur_scope), __pyx_n_s_validr__validator_c, __pyx_d, ((PyObject *)__pyx_codeobj__19)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_m_validator = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":261
 * 
 *         # TODO: deprecate below attributes because they are implement details
 *         m_validator.is_string = output_string             # <<<<<<<<<<<<<<
 *         m_validator.accept_string = accept_string
 *         m_validator.accept_object = accept_object
 */
  __Pyx_TraceLine(261,0,__PYX_ERR(0, 261, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_output_string); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_m_validator, __pyx_n_s_is_string, __pyx_t_1) < 0) __PYX_ERR(0, 261, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":262
 *         # TODO: deprecate below attributes because they are implement details
 *         m_validator.is_string = output_string
 *         m_validator.accept_string = accept_string             # <<<<<<<<<<<<<<
 *         m_validator.accept_object = accept_object
 *         m_validator.output_string = output_string
 */
  __Pyx_TraceLine(262,0,__PYX_ERR(0, 262, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_accept_string); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_m_validator, __pyx_n_s_accept_string, __pyx_t_1) < 0) __PYX_ERR(0, 262, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":263
 *         m_validator.is_string = output_string
 *         m_validator.accept_string = accept_string
 *         m_validator.accept_object = accept_object             # <<<<<<<<<<<<<<
 *         m_validator.output_string = output_string
 *         m_validator.output_object = output_object
 */
  __Pyx_TraceLine(263,0,__PYX_ERR(0, 263, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_accept_object); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 263, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_m_validator, __pyx_n_s_accept_object, __pyx_t_1) < 0) __PYX_ERR(0, 263, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":264
 *         m_validator.accept_string = accept_string
 *         m_validator.accept_object = accept_object
 *         m_validator.output_string = output_string             # <<<<<<<<<<<<<<
 *         m_validator.output_object = output_object
 *         m_validator.validator = f
 */
  __Pyx_TraceLine(264,0,__PYX_ERR(0, 264, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_output_string); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_m_validator, __pyx_n_s_output_string, __pyx_t_1) < 0) __PYX_ERR(0, 264, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":265
 *         m_validator.accept_object = accept_object
 *         m_validator.output_string = output_string
 *         m_validator.output_object = output_object             # <<<<<<<<<<<<<<
 *         m_validator.validator = f
 *         # end deprecate
 */
  __Pyx_TraceLine(265,0,__PYX_ERR(0, 265, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_output_object); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_m_validator, __pyx_n_s_output_object, __pyx_t_1) < 0) __PYX_ERR(0, 265, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":266
 *         m_validator.output_string = output_string
 *         m_validator.output_object = output_object
 *         m_validator.validator = f             # <<<<<<<<<<<<<<
 *         # end deprecate
 * 
 */
  __Pyx_TraceLine(266,0,__PYX_ERR(0, 266, __pyx_L1_error))
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_m_validator, __pyx_n_s_validator, __pyx_cur_scope->__pyx_v_f) < 0) __PYX_ERR(0, 266, __pyx_L1_error)

  /* "validr/_validator_c.pyx":269
 *         # end deprecate
 * 
 *         m_validator.__module__ = f.__module__             # <<<<<<<<<<<<<<
 *         m_validator.__name__ = f.__name__
 *         if hasattr(f, '__qualname__'):
 */
  __Pyx_TraceLine(269,0,__PYX_ERR(0, 269, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_f, __pyx_n_s_module); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_m_validator, __pyx_n_s_module, __pyx_t_1) < 0) __PYX_ERR(0, 269, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":270
 * 
 *         m_validator.__module__ = f.__module__
 *         m_validator.__name__ = f.__name__             # <<<<<<<<<<<<<<
 *         if hasattr(f, '__qualname__'):
 *             m_validator.__qualname__ = f.__qualname__
 */
  __Pyx_TraceLine(270,0,__PYX_ERR(0, 270, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_f, __pyx_n_s_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_m_validator, __pyx_n_s_name, __pyx_t_1) < 0) __PYX_ERR(0, 270, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":271
 *         m_validator.__module__ = f.__module__
 *         m_validator.__name__ = f.__name__
 *         if hasattr(f, '__qualname__'):             # <<<<<<<<<<<<<<
 *             m_validator.__qualname__ = f.__qualname__
 *         m_validator.__doc__ = f.__doc__
 */
  __Pyx_TraceLine(271,0,__PYX_ERR(0, 271, __pyx_L1_error))
  __pyx_t_1 = __pyx_cur_scope->__pyx_v_f;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_HasAttr(__pyx_t_1, __pyx_n_u_qualname); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 271, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "validr/_validator_c.pyx":272
 *         m_validator.__name__ = f.__name__
 *         if hasattr(f, '__qualname__'):
 *             m_validator.__qualname__ = f.__qualname__             # <<<<<<<<<<<<<<
 *         m_validator.__doc__ = f.__doc__
 *         return m_validator
 */
    __Pyx_TraceLine(272,0,__PYX_ERR(0, 272, __pyx_L1_error))
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_f, __pyx_n_s_qualname); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 272, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_m_validator, __pyx_n_s_qualname, __pyx_t_1) < 0) __PYX_ERR(0, 272, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "validr/_validator_c.pyx":271
 *         m_validator.__module__ = f.__module__
 *         m_validator.__name__ = f.__name__
 *         if hasattr(f, '__qualname__'):             # <<<<<<<<<<<<<<
 *             m_validator.__qualname__ = f.__qualname__
 *         m_validator.__doc__ = f.__doc__
 */
  }

  /* "validr/_validator_c.pyx":273
 *         if hasattr(f, '__qualname__'):
 *             m_validator.__qualname__ = f.__qualname__
 *         m_validator.__doc__ = f.__doc__             # <<<<<<<<<<<<<<
 *         return m_validator
 *     return decorator
 */
  __Pyx_TraceLine(273,0,__PYX_ERR(0, 273, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_f, __pyx_n_s_doc); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_m_validator, __pyx_n_s_doc, __pyx_t_1) < 0) __PYX_ERR(0, 273, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":274
 *             m_validator.__qualname__ = f.__qualname__
 *         m_validator.__doc__ = f.__doc__
 *         return m_validator             # <<<<<<<<<<<<<<
 *     return decorator
 * 
 */
  __Pyx_TraceLine(274,0,__PYX_ERR(0, 274, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_m_validator);
  __pyx_r = __pyx_v_m_validator;
  goto __pyx_L0;

  /* "validr/_validator_c.pyx":145
 *     del string, accept, output
 * 
 *     def decorator(f):             # <<<<<<<<<<<<<<
 * 
 *         def _m_validator(compiler, schema):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("validr._validator_c.validator.decorator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_m_validator);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":91
 * 
 * 
 * def validator(string=None, *, accept=None, output=None):             # <<<<<<<<<<<<<<
 *     """Decorator for create validator
 * 
 */

static PyObject *__pyx_pf_6validr_12_validator_c_14validator(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_string, PyObject *__pyx_v_accept, PyObject *__pyx_v_output) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct__validator *__pyx_cur_scope;
  PyObject *__pyx_v_decorator = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *(*__pyx_t_9)(PyObject *);
  __Pyx_TraceFrameInit(__pyx_codeobj__20)
  __Pyx_RefNannySetupContext("validator", 0);
  __pyx_cur_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct__validator *)__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct__validator(__pyx_ptype_6validr_12_validator_c___pyx_scope_struct__validator, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct__validator *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 91, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __Pyx_TraceCall("validator", __pyx_f[0], 91, 0, __PYX_ERR(0, 91, __pyx_L1_error));
  __Pyx_INCREF(__pyx_v_string);
  __Pyx_INCREF(__pyx_v_accept);
  __Pyx_INCREF(__pyx_v_output);

  /* "validr/_validator_c.pyx":127
 *     """
 *     cdef bint accept_string, accept_object, output_string, output_object
 *     if accept is None or not accept:             # <<<<<<<<<<<<<<
 *         accept_string = True
 *         accept_object = True
 */
  __Pyx_TraceLine(127,0,__PYX_ERR(0, 127, __pyx_L1_error))
  __pyx_t_2 = (__pyx_v_accept == Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (!__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_accept); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 127, __pyx_L1_error)
  __pyx_t_2 = ((!__pyx_t_3) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "validr/_validator_c.pyx":128
 *     cdef bint accept_string, accept_object, output_string, output_object
 *     if accept is None or not accept:
 *         accept_string = True             # <<<<<<<<<<<<<<
 *         accept_object = True
 *         accept_hints = (str, typing.Any)
 */
    __Pyx_TraceLine(128,0,__PYX_ERR(0, 128, __pyx_L1_error))
    __pyx_cur_scope->__pyx_v_accept_string = 1;

    /* "validr/_validator_c.pyx":129
 *     if accept is None or not accept:
 *         accept_string = True
 *         accept_object = True             # <<<<<<<<<<<<<<
 *         accept_hints = (str, typing.Any)
 *     else:
 */
    __Pyx_TraceLine(129,0,__PYX_ERR(0, 129, __pyx_L1_error))
    __pyx_cur_scope->__pyx_v_accept_object = 1;

    /* "validr/_validator_c.pyx":130
 *         accept_string = True
 *         accept_object = True
 *         accept_hints = (str, typing.Any)             # <<<<<<<<<<<<<<
 *     else:
 *         accept_string, accept_object, accept_hints = _parse_hints(accept)
 */
    __Pyx_TraceLine(130,0,__PYX_ERR(0, 130, __pyx_L1_error))
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_typing); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 130, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_Any); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 130, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 130, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(((PyObject *)(&PyUnicode_Type)));
    __Pyx_GIVEREF(((PyObject *)(&PyUnicode_Type)));
    PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)(&PyUnicode_Type)));
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_cur_scope->__pyx_v_accept_hints = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "validr/_validator_c.pyx":127
 *     """
 *     cdef bint accept_string, accept_object, output_string, output_object
 *     if accept is None or not accept:             # <<<<<<<<<<<<<<
 *         accept_string = True
 *         accept_object = True
 */
    goto __pyx_L3;
  }

  /* "validr/_validator_c.pyx":132
 *         accept_hints = (str, typing.Any)
 *     else:
 *         accept_string, accept_object, accept_hints = _parse_hints(accept)             # <<<<<<<<<<<<<<
 *     if not (accept_string or accept_object):
 *         raise ValueError('invalid accept argument {}'.format(accept))
 */
  __Pyx_TraceLine(132,0,__PYX_ERR(0, 132, __pyx_L1_error))
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_parse_hints); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 132, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_v_accept) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_accept);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 132, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
      PyObject* sequence = __pyx_t_4;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 3)) {
        if (size > 3) __Pyx_RaiseTooManyValuesError(3);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 132, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
        __pyx_t_7 = PyTuple_GET_ITEM(sequence, 2); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
        __pyx_t_7 = PyList_GET_ITEM(sequence, 2); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_7);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 132, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 132, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 132, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      #endif
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_8 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 132, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_9 = Py_TYPE(__pyx_t_8)->tp_iternext;
      index = 0; __pyx_t_5 = __pyx_t_9(__pyx_t_8); if (unlikely(!__pyx_t_5)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_6 = __pyx_t_9(__pyx_t_8); if (unlikely(!__pyx_t_6)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_6);
      index = 2; __pyx_t_7 = __pyx_t_9(__pyx_t_8); if (unlikely(!__pyx_t_7)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_7);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_9(__pyx_t_8), 3) < 0) __PYX_ERR(0, 132, __pyx_L1_error)
      __pyx_t_9 = NULL;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      goto __pyx_L7_unpacking_done;
      __pyx_L6_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_9 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 132, __pyx_L1_error)
      __pyx_L7_unpacking_done:;
    }
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 132, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 132, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_cur_scope->__pyx_v_accept_string = __pyx_t_1;
    __pyx_cur_scope->__pyx_v_accept_object = __pyx_t_2;
    __Pyx_GIVEREF(__pyx_t_7);
    __pyx_cur_scope->__pyx_v_accept_hints = __pyx_t_7;
    __pyx_t_7 = 0;
  }
  __pyx_L3:;

  /* "validr/_validator_c.pyx":133
 *     else:
 *         accept_string, accept_object, accept_hints = _parse_hints(accept)
 *     if not (accept_string or accept_object):             # <<<<<<<<<<<<<<
 *         raise ValueError('invalid accept argument {}'.format(accept))
 *     if output is None or not output:
 */
  __Pyx_TraceLine(133,0,__PYX_ERR(0, 133, __pyx_L1_error))
  __pyx_t_1 = (__pyx_cur_scope->__pyx_v_accept_string != 0);
  if (!__pyx_t_1) {
  } else {
    __pyx_t_2 = __pyx_t_1;
    goto __pyx_L9_bool_binop_done;
  }
  __pyx_t_1 = (__pyx_cur_scope->__pyx_v_accept_object != 0);
  __pyx_t_2 = __pyx_t_1;
  __pyx_L9_bool_binop_done:;
  __pyx_t_1 = ((!__pyx_t_2) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "validr/_validator_c.pyx":134
 *         accept_string, accept_object, accept_hints = _parse_hints(accept)
 *     if not (accept_string or accept_object):
 *         raise ValueError('invalid accept argument {}'.format(accept))             # <<<<<<<<<<<<<<
 *     if output is None or not output:
 *         output_string = string
 */
    __Pyx_TraceLine(134,0,__PYX_ERR(0, 134, __pyx_L1_error))
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_invalid_accept_argument, __pyx_n_s_format); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 134, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_6, __pyx_v_accept) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_accept);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 134, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 134, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_7, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __PYX_ERR(0, 134, __pyx_L1_error)

    /* "validr/_validator_c.pyx":133
 *     else:
 *         accept_string, accept_object, accept_hints = _parse_hints(accept)
 *     if not (accept_string or accept_object):             # <<<<<<<<<<<<<<
 *         raise ValueError('invalid accept argument {}'.format(accept))
 *     if output is None or not output:
 */
  }

  /* "validr/_validator_c.pyx":135
 *     if not (accept_string or accept_object):
 *         raise ValueError('invalid accept argument {}'.format(accept))
 *     if output is None or not output:             # <<<<<<<<<<<<<<
 *         output_string = string
 *         output_object = not string
 */
  __Pyx_TraceLine(135,0,__PYX_ERR(0, 135, __pyx_L1_error))
  __pyx_t_2 = (__pyx_v_output == Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (!__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L12_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_output); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 135, __pyx_L1_error)
  __pyx_t_2 = ((!__pyx_t_3) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L12_bool_binop_done:;
  if (__pyx_t_1) {

    /* "validr/_validator_c.pyx":136
 *         raise ValueError('invalid accept argument {}'.format(accept))
 *     if output is None or not output:
 *         output_string = string             # <<<<<<<<<<<<<<
 *         output_object = not string
 *         output_hints = (str,) if string else (typing.Any,)
 */
    __Pyx_TraceLine(136,0,__PYX_ERR(0, 136, __pyx_L1_error))
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_string); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 136, __pyx_L1_error)
    __pyx_cur_scope->__pyx_v_output_string = __pyx_t_1;

    /* "validr/_validator_c.pyx":137
 *     if output is None or not output:
 *         output_string = string
 *         output_object = not string             # <<<<<<<<<<<<<<
 *         output_hints = (str,) if string else (typing.Any,)
 *     else:
 */
    __Pyx_TraceLine(137,0,__PYX_ERR(0, 137, __pyx_L1_error))
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_string); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 137, __pyx_L1_error)
    __pyx_cur_scope->__pyx_v_output_object = (!__pyx_t_1);

    /* "validr/_validator_c.pyx":138
 *         output_string = string
 *         output_object = not string
 *         output_hints = (str,) if string else (typing.Any,)             # <<<<<<<<<<<<<<
 *     else:
 *         output_string, output_object, output_hints = _parse_hints(output)
 */
    __Pyx_TraceLine(138,0,__PYX_ERR(0, 138, __pyx_L1_error))
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_string); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 138, __pyx_L1_error)
    if (__pyx_t_1) {
      __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 138, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(((PyObject *)(&PyUnicode_Type)));
      __Pyx_GIVEREF(((PyObject *)(&PyUnicode_Type)));
      PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)(&PyUnicode_Type)));
      __pyx_t_7 = __pyx_t_4;
      __pyx_t_4 = 0;
    } else {
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_typing); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 138, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_Any); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 138, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 138, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_6);
      __pyx_t_6 = 0;
      __pyx_t_7 = __pyx_t_4;
      __pyx_t_4 = 0;
    }
    __Pyx_GIVEREF(__pyx_t_7);
    __pyx_cur_scope->__pyx_v_output_hints = __pyx_t_7;
    __pyx_t_7 = 0;

    /* "validr/_validator_c.pyx":135
 *     if not (accept_string or accept_object):
 *         raise ValueError('invalid accept argument {}'.format(accept))
 *     if output is None or not output:             # <<<<<<<<<<<<<<
 *         output_string = string
 *         output_object = not string
 */
    goto __pyx_L11;
  }

  /* "validr/_validator_c.pyx":140
 *         output_hints = (str,) if string else (typing.Any,)
 *     else:
 *         output_string, output_object, output_hints = _parse_hints(output)             # <<<<<<<<<<<<<<
 *     if not (output_string or output_object):
 *         raise ValueError('invalid output argument {}'.format(output))
 */
  __Pyx_TraceLine(140,0,__PYX_ERR(0, 140, __pyx_L1_error))
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_parse_hints); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 140, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_7 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_6, __pyx_v_output) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_output);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 140, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if ((likely(PyTuple_CheckExact(__pyx_t_7))) || (PyList_CheckExact(__pyx_t_7))) {
      PyObject* sequence = __pyx_t_7;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 3)) {
        if (size > 3) __Pyx_RaiseTooManyValuesError(3);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 140, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 2); 
      } else {
        __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
        __pyx_t_5 = PyList_GET_ITEM(sequence, 2); 
      }
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_5);
      #else
      __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 140, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 140, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_5 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 140, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      #endif
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_8 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 140, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_9 = Py_TYPE(__pyx_t_8)->tp_iternext;
      index = 0; __pyx_t_4 = __pyx_t_9(__pyx_t_8); if (unlikely(!__pyx_t_4)) goto __pyx_L14_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_4);
      index = 1; __pyx_t_6 = __pyx_t_9(__pyx_t_8); if (unlikely(!__pyx_t_6)) goto __pyx_L14_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_6);
      index = 2; __pyx_t_5 = __pyx_t_9(__pyx_t_8); if (unlikely(!__pyx_t_5)) goto __pyx_L14_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_9(__pyx_t_8), 3) < 0) __PYX_ERR(0, 140, __pyx_L1_error)
      __pyx_t_9 = NULL;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      goto __pyx_L15_unpacking_done;
      __pyx_L14_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_9 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 140, __pyx_L1_error)
      __pyx_L15_unpacking_done:;
    }
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 140, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 140, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_cur_scope->__pyx_v_output_string = __pyx_t_1;
    __pyx_cur_scope->__pyx_v_output_object = __pyx_t_2;
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_cur_scope->__pyx_v_output_hints = __pyx_t_5;
    __pyx_t_5 = 0;
  }
  __pyx_L11:;

  /* "validr/_validator_c.pyx":141
 *     else:
 *         output_string, output_object, output_hints = _parse_hints(output)
 *     if not (output_string or output_object):             # <<<<<<<<<<<<<<
 *         raise ValueError('invalid output argument {}'.format(output))
 *     del string, accept, output
 */
  __Pyx_TraceLine(141,0,__PYX_ERR(0, 141, __pyx_L1_error))
  __pyx_t_1 = (__pyx_cur_scope->__pyx_v_output_string != 0);
  if (!__pyx_t_1) {
  } else {
    __pyx_t_2 = __pyx_t_1;
    goto __pyx_L17_bool_binop_done;
  }
  __pyx_t_1 = (__pyx_cur_scope->__pyx_v_output_object != 0);
  __pyx_t_2 = __pyx_t_1;
  __pyx_L17_bool_binop_done:;
  __pyx_t_1 = ((!__pyx_t_2) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "validr/_validator_c.pyx":142
 *         output_string, output_object, output_hints = _parse_hints(output)
 *     if not (output_string or output_object):
 *         raise ValueError('invalid output argument {}'.format(output))             # <<<<<<<<<<<<<<
 *     del string, accept, output
 * 
 */
    __Pyx_TraceLine(142,0,__PYX_ERR(0, 142, __pyx_L1_error))
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_invalid_output_argument, __pyx_n_s_format); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 142, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_7 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_v_output) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_output);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 142, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 142, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(0, 142, __pyx_L1_error)

    /* "validr/_validator_c.pyx":141
 *     else:
 *         output_string, output_object, output_hints = _parse_hints(output)
 *     if not (output_string or output_object):             # <<<<<<<<<<<<<<
 *         raise ValueError('invalid output argument {}'.format(output))
 *     del string, accept, output
 */
  }

  /* "validr/_validator_c.pyx":143
 *     if not (output_string or output_object):
 *         raise ValueError('invalid output argument {}'.format(output))
 *     del string, accept, output             # <<<<<<<<<<<<<<
 * 
 *     def decorator(f):
 */
  __Pyx_TraceLine(143,0,__PYX_ERR(0, 143, __pyx_L1_error))
  __Pyx_DECREF(__pyx_v_string);
  __pyx_v_string = NULL;
  __Pyx_DECREF(__pyx_v_accept);
  __pyx_v_accept = NULL;
  __Pyx_DECREF(__pyx_v_output);
  __pyx_v_output = NULL;

  /* "validr/_validator_c.pyx":145
 *     del string, accept, output
 * 
 *     def decorator(f):             # <<<<<<<<<<<<<<
 * 
 *         def _m_validator(compiler, schema):
 */
  __Pyx_TraceLine(145,0,__PYX_ERR(0, 145, __pyx_L1_error))
  __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_9validator_1decorator, 0, __pyx_n_s_validator_locals_decorator, ((PyObject*)__pyx_cur_scope), __pyx_n_s_validr__validator_c, __pyx_d, ((PyObject *)__pyx_codeobj__22)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_v_decorator = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "validr/_validator_c.pyx":275
 *         m_validator.__doc__ = f.__doc__
 *         return m_validator
 *     return decorator             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(275,0,__PYX_ERR(0, 275, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_decorator);
  __pyx_r = __pyx_v_decorator;
  goto __pyx_L0;

  /* "validr/_validator_c.pyx":91
 * 
 * 
 * def validator(string=None, *, accept=None, output=None):             # <<<<<<<<<<<<<<
 *     """Decorator for create validator
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("validr._validator_c.validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_decorator);
  __Pyx_XDECREF(__pyx_v_string);
  __Pyx_XDECREF(__pyx_v_accept);
  __Pyx_XDECREF(__pyx_v_output);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":281
 * 
 * 
 * cpdef inline _key_of_scalar(v):             # <<<<<<<<<<<<<<
 *     return v
 * 
 */

static PyObject *__pyx_pw_6validr_12_validator_c_17_key_of_scalar(PyObject *__pyx_self, PyObject *__pyx_v_v); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_6validr_12_validator_c__key_of_scalar(PyObject *__pyx_v_v, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_key_of_scalar", 0);
  __Pyx_TraceCall("_key_of_scalar", __pyx_f[0], 281, 0, __PYX_ERR(0, 281, __pyx_L1_error));

  /* "validr/_validator_c.pyx":282
 * 
 * cpdef inline _key_of_scalar(v):
 *     return v             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(282,0,__PYX_ERR(0, 282, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_v);
  __pyx_r = __pyx_v_v;
  goto __pyx_L0;

  /* "validr/_validator_c.pyx":281
 * 
 * 
 * cpdef inline _key_of_scalar(v):             # <<<<<<<<<<<<<<
 *     return v
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("validr._validator_c._key_of_scalar", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_17_key_of_scalar(PyObject *__pyx_self, PyObject *__pyx_v_v); /*proto*/
static PyObject *__pyx_pw_6validr_12_validator_c_17_key_of_scalar(PyObject *__pyx_self, PyObject *__pyx_v_v) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_key_of_scalar (wrapper)", 0);
  __pyx_r = __pyx_pf_6validr_12_validator_c_16_key_of_scalar(__pyx_self, ((PyObject *)__pyx_v_v));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6validr_12_validator_c_16_key_of_scalar(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_v) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_key_of_scalar", 0);
  __Pyx_TraceCall("_key_of_scalar (wrapper)", __pyx_f[0], 281, 0, __PYX_ERR(0, 281, __pyx_L1_error));
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6validr_12_validator_c__key_of_scalar(__pyx_v_v, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("validr._validator_c._key_of_scalar", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":285
 * 
 * 
 * def _key_func_of_schema(schema):             # <<<<<<<<<<<<<<
 *     if schema is None:
 *         raise SchemaError(_UNIQUE_CHECK_ERROR_MESSAGE)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_19_key_func_of_schema(PyObject *__pyx_self, PyObject *__pyx_v_schema); /*proto*/
static PyMethodDef __pyx_mdef_6validr_12_validator_c_19_key_func_of_schema = {"_key_func_of_schema", (PyCFunction)__pyx_pw_6validr_12_validator_c_19_key_func_of_schema, METH_O, 0};
static PyObject *__pyx_pw_6validr_12_validator_c_19_key_func_of_schema(PyObject *__pyx_self, PyObject *__pyx_v_schema) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_key_func_of_schema (wrapper)", 0);
  __pyx_r = __pyx_pf_6validr_12_validator_c_18_key_func_of_schema(__pyx_self, ((PyObject *)__pyx_v_schema));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":296
 *             keys.append((k, _key_func_of_schema(v)))
 * 
 *         def key_of(dict v):             # <<<<<<<<<<<<<<
 *             cdef str k
 *             return tuple(key_of_value(v[k]) for k, key_of_value in keys)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_19_key_func_of_schema_1key_of(PyObject *__pyx_self, PyObject *__pyx_v_v); /*proto*/
static PyMethodDef __pyx_mdef_6validr_12_validator_c_19_key_func_of_schema_1key_of = {"key_of", (PyCFunction)__pyx_pw_6validr_12_validator_c_19_key_func_of_schema_1key_of, METH_O, 0};
static PyObject *__pyx_pw_6validr_12_validator_c_19_key_func_of_schema_1key_of(PyObject *__pyx_self, PyObject *__pyx_v_v) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("key_of (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_v), (&PyDict_Type), 1, "v", 1))) __PYX_ERR(0, 296, __pyx_L1_error)
  __pyx_r = __pyx_pf_6validr_12_validator_c_19_key_func_of_schema_key_of(__pyx_self, ((PyObject*)__pyx_v_v));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6validr_12_validator_c_19_key_func_of_schema_6key_of_2generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "validr/_validator_c.pyx":298
 *         def key_of(dict v):
 *             cdef str k
 *             return tuple(key_of_value(v[k]) for k, key_of_value in keys)             # <<<<<<<<<<<<<<
 * 
 *     elif schema.validator == 'list':
 */

static PyObject *__pyx_pf_6validr_12_validator_c_19_key_func_of_schema_6key_of_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_5_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_5_genexpr *)__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_5_genexpr(__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_5_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_5_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 298, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_4_key_of *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_6validr_12_validator_c_19_key_func_of_schema_6key_of_2generator, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_key_func_of_schema_locals_key_o, __pyx_n_s_validr__validator_c); if (unlikely(!gen)) __PYX_ERR(0, 298, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("validr._validator_c._key_func_of_schema.key_of.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6validr_12_validator_c_19_key_func_of_schema_6key_of_2generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_5_genexpr *__pyx_cur_scope = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_5_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *(*__pyx_t_7)(PyObject *);
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  __Pyx_TraceCall("genexpr", __pyx_f[0], 298, 0, __PYX_ERR(0, 298, __pyx_L1_error));
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L8_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_TraceReturn(Py_None, 0);
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 298, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_outer_scope->__pyx_v_keys)) { __Pyx_RaiseClosureNameError("keys"); __PYX_ERR(0, 298, __pyx_L1_error) }
  if (unlikely(__pyx_cur_scope->__pyx_outer_scope->__pyx_outer_scope->__pyx_v_keys == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 298, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_outer_scope->__pyx_v_keys; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 298, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 298, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 298, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      #else
      __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 298, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 298, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_6 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 298, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_7 = Py_TYPE(__pyx_t_6)->tp_iternext;
      index = 0; __pyx_t_4 = __pyx_t_7(__pyx_t_6); if (unlikely(!__pyx_t_4)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_4);
      index = 1; __pyx_t_5 = __pyx_t_7(__pyx_t_6); if (unlikely(!__pyx_t_5)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_7(__pyx_t_6), 2) < 0) __PYX_ERR(0, 298, __pyx_L1_error)
      __pyx_t_7 = NULL;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      goto __pyx_L7_unpacking_done;
      __pyx_L6_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_7 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 298, __pyx_L1_error)
      __pyx_L7_unpacking_done:;
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_k);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_k, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_key_of_value);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_key_of_value, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_t_5 = 0;
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_v)) { __Pyx_RaiseClosureNameError("v"); __PYX_ERR(0, 298, __pyx_L1_error) }
    if (unlikely(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_v == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 298, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_PyDict_GetItem(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_v, __pyx_cur_scope->__pyx_v_k); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 298, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_key_of_value);
    __pyx_t_4 = __pyx_cur_scope->__pyx_v_key_of_value; __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 298, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_TraceReturn(__pyx_r, 0);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L8_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 298, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":296
 *             keys.append((k, _key_func_of_schema(v)))
 * 
 *         def key_of(dict v):             # <<<<<<<<<<<<<<
 *             cdef str k
 *             return tuple(key_of_value(v[k]) for k, key_of_value in keys)
 */

static PyObject *__pyx_pf_6validr_12_validator_c_19_key_func_of_schema_key_of(PyObject *__pyx_self, PyObject *__pyx_v_v) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_4_key_of *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("key_of", 0);
  __pyx_cur_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_4_key_of *)__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_4_key_of(__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_4_key_of, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_4_key_of *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 296, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_3__key_func_of_schema *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  __Pyx_TraceCall("key_of", __pyx_f[0], 296, 0, __PYX_ERR(0, 296, __pyx_L1_error));
  __pyx_cur_scope->__pyx_v_v = __pyx_v_v;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_v);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_v);

  /* "validr/_validator_c.pyx":298
 *         def key_of(dict v):
 *             cdef str k
 *             return tuple(key_of_value(v[k]) for k, key_of_value in keys)             # <<<<<<<<<<<<<<
 * 
 *     elif schema.validator == 'list':
 */
  __Pyx_TraceLine(298,0,__PYX_ERR(0, 298, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_pf_6validr_12_validator_c_19_key_func_of_schema_6key_of_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 298, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PySequence_Tuple(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 298, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "validr/_validator_c.pyx":296
 *             keys.append((k, _key_func_of_schema(v)))
 * 
 *         def key_of(dict v):             # <<<<<<<<<<<<<<
 *             cdef str k
 *             return tuple(key_of_value(v[k]) for k, key_of_value in keys)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("validr._validator_c._key_func_of_schema.key_of", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":305
 *         key_of_value = _key_func_of_schema(schema.items)
 * 
 *         def key_of(list v):             # <<<<<<<<<<<<<<
 *             return tuple(key_of_value(x) for x in v)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_19_key_func_of_schema_3key_of(PyObject *__pyx_self, PyObject *__pyx_v_v); /*proto*/
static PyMethodDef __pyx_mdef_6validr_12_validator_c_19_key_func_of_schema_3key_of = {"key_of", (PyCFunction)__pyx_pw_6validr_12_validator_c_19_key_func_of_schema_3key_of, METH_O, 0};
static PyObject *__pyx_pw_6validr_12_validator_c_19_key_func_of_schema_3key_of(PyObject *__pyx_self, PyObject *__pyx_v_v) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("key_of (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_v), (&PyList_Type), 1, "v", 1))) __PYX_ERR(0, 305, __pyx_L1_error)
  __pyx_r = __pyx_pf_6validr_12_validator_c_19_key_func_of_schema_2key_of(__pyx_self, ((PyObject*)__pyx_v_v));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6validr_12_validator_c_19_key_func_of_schema_6key_of_5generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "validr/_validator_c.pyx":306
 * 
 *         def key_of(list v):
 *             return tuple(key_of_value(x) for x in v)             # <<<<<<<<<<<<<<
 * 
 *     else:
 */

static PyObject *__pyx_pf_6validr_12_validator_c_19_key_func_of_schema_6key_of_3genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_7_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_7_genexpr *)__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_7_genexpr(__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_7_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_7_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 306, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_6_key_of *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_6validr_12_validator_c_19_key_func_of_schema_6key_of_5generator1, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_key_func_of_schema_locals_key_o, __pyx_n_s_validr__validator_c); if (unlikely(!gen)) __PYX_ERR(0, 306, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("validr._validator_c._key_func_of_schema.key_of.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6validr_12_validator_c_19_key_func_of_schema_6key_of_5generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_7_genexpr *__pyx_cur_scope = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_7_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  __Pyx_TraceCall("genexpr", __pyx_f[0], 306, 0, __PYX_ERR(0, 306, __pyx_L1_error));
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_TraceReturn(Py_None, 0);
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 306, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_v)) { __Pyx_RaiseClosureNameError("v"); __PYX_ERR(0, 306, __pyx_L1_error) }
  if (unlikely(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_v == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 306, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_v; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 306, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 306, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_x);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_x, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_outer_scope->__pyx_v_key_of_value)) { __Pyx_RaiseClosureNameError("key_of_value"); __PYX_ERR(0, 306, __pyx_L1_error) }
    __Pyx_INCREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_outer_scope->__pyx_v_key_of_value);
    __pyx_t_4 = __pyx_cur_scope->__pyx_outer_scope->__pyx_outer_scope->__pyx_v_key_of_value; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_cur_scope->__pyx_v_x) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_cur_scope->__pyx_v_x);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 306, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_TraceReturn(__pyx_r, 0);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 306, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":305
 *         key_of_value = _key_func_of_schema(schema.items)
 * 
 *         def key_of(list v):             # <<<<<<<<<<<<<<
 *             return tuple(key_of_value(x) for x in v)
 * 
 */

static PyObject *__pyx_pf_6validr_12_validator_c_19_key_func_of_schema_2key_of(PyObject *__pyx_self, PyObject *__pyx_v_v) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_6_key_of *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("key_of", 0);
  __pyx_cur_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_6_key_of *)__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_6_key_of(__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_6_key_of, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_6_key_of *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 305, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_3__key_func_of_schema *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  __Pyx_TraceCall("key_of", __pyx_f[0], 305, 0, __PYX_ERR(0, 305, __pyx_L1_error));
  __pyx_cur_scope->__pyx_v_v = __pyx_v_v;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_v);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_v);

  /* "validr/_validator_c.pyx":306
 * 
 *         def key_of(list v):
 *             return tuple(key_of_value(x) for x in v)             # <<<<<<<<<<<<<<
 * 
 *     else:
 */
  __Pyx_TraceLine(306,0,__PYX_ERR(0, 306, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_pf_6validr_12_validator_c_19_key_func_of_schema_6key_of_3genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PySequence_Tuple(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "validr/_validator_c.pyx":305
 *         key_of_value = _key_func_of_schema(schema.items)
 * 
 *         def key_of(list v):             # <<<<<<<<<<<<<<
 *             return tuple(key_of_value(x) for x in v)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("validr._validator_c._key_func_of_schema.key_of", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":285
 * 
 * 
 * def _key_func_of_schema(schema):             # <<<<<<<<<<<<<<
 *     if schema is None:
 *         raise SchemaError(_UNIQUE_CHECK_ERROR_MESSAGE)
 */

static PyObject *__pyx_pf_6validr_12_validator_c_18_key_func_of_schema(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_schema) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_3__key_func_of_schema *__pyx_cur_scope;
  PyObject *__pyx_v_k = NULL;
  PyObject *__pyx_v_v = NULL;
  PyObject *__pyx_v_key_of = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  __Pyx_TraceFrameInit(__pyx_codeobj__23)
  __Pyx_RefNannySetupContext("_key_func_of_schema", 0);
  __pyx_cur_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_3__key_func_of_schema *)__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_3__key_func_of_schema(__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_3__key_func_of_schema, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_3__key_func_of_schema *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 285, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __Pyx_TraceCall("_key_func_of_schema", __pyx_f[0], 285, 0, __PYX_ERR(0, 285, __pyx_L1_error));

  /* "validr/_validator_c.pyx":286
 * 
 * def _key_func_of_schema(schema):
 *     if schema is None:             # <<<<<<<<<<<<<<
 *         raise SchemaError(_UNIQUE_CHECK_ERROR_MESSAGE)
 * 
 */
  __Pyx_TraceLine(286,0,__PYX_ERR(0, 286, __pyx_L1_error))
  __pyx_t_1 = (__pyx_v_schema == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (unlikely(__pyx_t_2)) {

    /* "validr/_validator_c.pyx":287
 * def _key_func_of_schema(schema):
 *     if schema is None:
 *         raise SchemaError(_UNIQUE_CHECK_ERROR_MESSAGE)             # <<<<<<<<<<<<<<
 * 
 *     if schema.validator == 'dict':
 */
    __Pyx_TraceLine(287,0,__PYX_ERR(0, 287, __pyx_L1_error))
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_SchemaError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 287, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_6validr_12_validator_c__UNIQUE_CHECK_ERROR_MESSAGE) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_6validr_12_validator_c__UNIQUE_CHECK_ERROR_MESSAGE);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 287, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 287, __pyx_L1_error)

    /* "validr/_validator_c.pyx":286
 * 
 * def _key_func_of_schema(schema):
 *     if schema is None:             # <<<<<<<<<<<<<<
 *         raise SchemaError(_UNIQUE_CHECK_ERROR_MESSAGE)
 * 
 */
  }

  /* "validr/_validator_c.pyx":289
 *         raise SchemaError(_UNIQUE_CHECK_ERROR_MESSAGE)
 * 
 *     if schema.validator == 'dict':             # <<<<<<<<<<<<<<
 *         if schema.items is None:
 *             raise SchemaError(_UNIQUE_CHECK_ERROR_MESSAGE)
 */
  __Pyx_TraceLine(289,0,__PYX_ERR(0, 289, __pyx_L1_error))
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_schema, __pyx_n_s_validator); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_t_3, __pyx_n_u_dict, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 289, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_2) {

    /* "validr/_validator_c.pyx":290
 * 
 *     if schema.validator == 'dict':
 *         if schema.items is None:             # <<<<<<<<<<<<<<
 *             raise SchemaError(_UNIQUE_CHECK_ERROR_MESSAGE)
 *         keys = []
 */
    __Pyx_TraceLine(290,0,__PYX_ERR(0, 290, __pyx_L1_error))
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_schema, __pyx_n_s_items); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 290, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = (__pyx_t_3 == Py_None);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_1 = (__pyx_t_2 != 0);
    if (unlikely(__pyx_t_1)) {

      /* "validr/_validator_c.pyx":291
 *     if schema.validator == 'dict':
 *         if schema.items is None:
 *             raise SchemaError(_UNIQUE_CHECK_ERROR_MESSAGE)             # <<<<<<<<<<<<<<
 *         keys = []
 *         for k, v in schema.items.items():
 */
      __Pyx_TraceLine(291,0,__PYX_ERR(0, 291, __pyx_L1_error))
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_SchemaError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 291, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_6validr_12_validator_c__UNIQUE_CHECK_ERROR_MESSAGE) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_6validr_12_validator_c__UNIQUE_CHECK_ERROR_MESSAGE);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 291, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 291, __pyx_L1_error)

      /* "validr/_validator_c.pyx":290
 * 
 *     if schema.validator == 'dict':
 *         if schema.items is None:             # <<<<<<<<<<<<<<
 *             raise SchemaError(_UNIQUE_CHECK_ERROR_MESSAGE)
 *         keys = []
 */
    }

    /* "validr/_validator_c.pyx":292
 *         if schema.items is None:
 *             raise SchemaError(_UNIQUE_CHECK_ERROR_MESSAGE)
 *         keys = []             # <<<<<<<<<<<<<<
 *         for k, v in schema.items.items():
 *             keys.append((k, _key_func_of_schema(v)))
 */
    __Pyx_TraceLine(292,0,__PYX_ERR(0, 292, __pyx_L1_error))
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 292, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_cur_scope->__pyx_v_keys = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "validr/_validator_c.pyx":293
 *             raise SchemaError(_UNIQUE_CHECK_ERROR_MESSAGE)
 *         keys = []
 *         for k, v in schema.items.items():             # <<<<<<<<<<<<<<
 *             keys.append((k, _key_func_of_schema(v)))
 * 
 */
    __Pyx_TraceLine(293,0,__PYX_ERR(0, 293, __pyx_L1_error))
    __pyx_t_6 = 0;
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_schema, __pyx_n_s_items); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 293, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (unlikely(__pyx_t_4 == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
      __PYX_ERR(0, 293, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_dict_iterator(__pyx_t_4, 0, __pyx_n_s_items, (&__pyx_t_7), (&__pyx_t_8)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 293, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_3);
    __pyx_t_3 = __pyx_t_5;
    __pyx_t_5 = 0;
    while (1) {
      __pyx_t_9 = __Pyx_dict_iter_next(__pyx_t_3, __pyx_t_7, &__pyx_t_6, &__pyx_t_5, &__pyx_t_4, NULL, __pyx_t_8);
      if (unlikely(__pyx_t_9 == 0)) break;
      if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(0, 293, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_XDECREF_SET(__pyx_v_k, __pyx_t_5);
      __pyx_t_5 = 0;
      __Pyx_XDECREF_SET(__pyx_v_v, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "validr/_validator_c.pyx":294
 *         keys = []
 *         for k, v in schema.items.items():
 *             keys.append((k, _key_func_of_schema(v)))             # <<<<<<<<<<<<<<
 * 
 *         def key_of(dict v):
 */
      __Pyx_TraceLine(294,0,__PYX_ERR(0, 294, __pyx_L1_error))
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_key_func_of_schema); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 294, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_10 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_4 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_10, __pyx_v_v) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_v);
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 294, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 294, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_v_k);
      __Pyx_GIVEREF(__pyx_v_k);
      PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_k);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
      __pyx_t_4 = 0;
      __pyx_t_11 = __Pyx_PyList_Append(__pyx_cur_scope->__pyx_v_keys, __pyx_t_5); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 294, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "validr/_validator_c.pyx":296
 *             keys.append((k, _key_func_of_schema(v)))
 * 
 *         def key_of(dict v):             # <<<<<<<<<<<<<<
 *             cdef str k
 *             return tuple(key_of_value(v[k]) for k, key_of_value in keys)
 */
    __Pyx_TraceLine(296,0,__PYX_ERR(0, 296, __pyx_L1_error))
    __pyx_t_3 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_19_key_func_of_schema_1key_of, 0, __pyx_n_s_key_func_of_schema_locals_key_o_2, ((PyObject*)__pyx_cur_scope), __pyx_n_s_validr__validator_c, __pyx_d, ((PyObject *)__pyx_codeobj__25)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 296, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_key_of = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "validr/_validator_c.pyx":289
 *         raise SchemaError(_UNIQUE_CHECK_ERROR_MESSAGE)
 * 
 *     if schema.validator == 'dict':             # <<<<<<<<<<<<<<
 *         if schema.items is None:
 *             raise SchemaError(_UNIQUE_CHECK_ERROR_MESSAGE)
 */
    goto __pyx_L4;
  }

  /* "validr/_validator_c.pyx":300
 *             return tuple(key_of_value(v[k]) for k, key_of_value in keys)
 * 
 *     elif schema.validator == 'list':             # <<<<<<<<<<<<<<
 *         if schema.items is None:
 *             raise SchemaError(_UNIQUE_CHECK_ERROR_MESSAGE)
 */
  __Pyx_TraceLine(300,0,__PYX_ERR(0, 300, __pyx_L1_error))
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_schema, __pyx_n_s_validator); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_t_3, __pyx_n_u_list, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 300, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_1) {

    /* "validr/_validator_c.pyx":301
 * 
 *     elif schema.validator == 'list':
 *         if schema.items is None:             # <<<<<<<<<<<<<<
 *             raise SchemaError(_UNIQUE_CHECK_ERROR_MESSAGE)
 *         key_of_value = _key_func_of_schema(schema.items)
 */
    __Pyx_TraceLine(301,0,__PYX_ERR(0, 301, __pyx_L1_error))
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_schema, __pyx_n_s_items); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 301, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = (__pyx_t_3 == Py_None);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_2 = (__pyx_t_1 != 0);
    if (unlikely(__pyx_t_2)) {

      /* "validr/_validator_c.pyx":302
 *     elif schema.validator == 'list':
 *         if schema.items is None:
 *             raise SchemaError(_UNIQUE_CHECK_ERROR_MESSAGE)             # <<<<<<<<<<<<<<
 *         key_of_value = _key_func_of_schema(schema.items)
 * 
 */
      __Pyx_TraceLine(302,0,__PYX_ERR(0, 302, __pyx_L1_error))
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_SchemaError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 302, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_4, __pyx_v_6validr_12_validator_c__UNIQUE_CHECK_ERROR_MESSAGE) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_6validr_12_validator_c__UNIQUE_CHECK_ERROR_MESSAGE);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 302, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 302, __pyx_L1_error)

      /* "validr/_validator_c.pyx":301
 * 
 *     elif schema.validator == 'list':
 *         if schema.items is None:             # <<<<<<<<<<<<<<
 *             raise SchemaError(_UNIQUE_CHECK_ERROR_MESSAGE)
 *         key_of_value = _key_func_of_schema(schema.items)
 */
    }

    /* "validr/_validator_c.pyx":303
 *         if schema.items is None:
 *             raise SchemaError(_UNIQUE_CHECK_ERROR_MESSAGE)
 *         key_of_value = _key_func_of_schema(schema.items)             # <<<<<<<<<<<<<<
 * 
 *         def key_of(list v):
 */
    __Pyx_TraceLine(303,0,__PYX_ERR(0, 303, __pyx_L1_error))
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_key_func_of_schema); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 303, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_schema, __pyx_n_s_items); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 303, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_3 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_10, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 303, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_cur_scope->__pyx_v_key_of_value = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "validr/_validator_c.pyx":305
 *         key_of_value = _key_func_of_schema(schema.items)
 * 
 *         def key_of(list v):             # <<<<<<<<<<<<<<
 *             return tuple(key_of_value(x) for x in v)
 * 
 */
    __Pyx_TraceLine(305,0,__PYX_ERR(0, 305, __pyx_L1_error))
    __pyx_t_3 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_19_key_func_of_schema_3key_of, 0, __pyx_n_s_key_func_of_schema_locals_key_o_2, ((PyObject*)__pyx_cur_scope), __pyx_n_s_validr__validator_c, __pyx_d, ((PyObject *)__pyx_codeobj__27)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 305, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_key_of = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "validr/_validator_c.pyx":300
 *             return tuple(key_of_value(v[k]) for k, key_of_value in keys)
 * 
 *     elif schema.validator == 'list':             # <<<<<<<<<<<<<<
 *         if schema.items is None:
 *             raise SchemaError(_UNIQUE_CHECK_ERROR_MESSAGE)
 */
    goto __pyx_L4;
  }

  /* "validr/_validator_c.pyx":309
 * 
 *     else:
 *         key_of = _key_of_scalar             # <<<<<<<<<<<<<<
 * 
 *     return key_of
 */
  __Pyx_TraceLine(309,0,__PYX_ERR(0, 309, __pyx_L1_error))
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_key_of_scalar); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 309, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_key_of = __pyx_t_3;
    __pyx_t_3 = 0;
  }
  __pyx_L4:;

  /* "validr/_validator_c.pyx":311
 *         key_of = _key_of_scalar
 * 
 *     return key_of             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(311,0,__PYX_ERR(0, 311, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_key_of);
  __pyx_r = __pyx_v_key_of;
  goto __pyx_L0;

  /* "validr/_validator_c.pyx":285
 * 
 * 
 * def _key_func_of_schema(schema):             # <<<<<<<<<<<<<<
 *     if schema is None:
 *         raise SchemaError(_UNIQUE_CHECK_ERROR_MESSAGE)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("validr._validator_c._key_func_of_schema", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_k);
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_XDECREF(__pyx_v_key_of);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":315
 * 
 * @validator(accept=typing.Iterable, output=typing.List)
 * def list_validator(compiler, items=None, int minlen=0, int maxlen=1024,             # <<<<<<<<<<<<<<
 *                    bint unique=False):
 *     if items is None:
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_21list_validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6validr_12_validator_c_21list_validator = {"list_validator", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6validr_12_validator_c_21list_validator, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6validr_12_validator_c_21list_validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_compiler = 0;
  PyObject *__pyx_v_items = 0;
  int __pyx_v_minlen;
  int __pyx_v_maxlen;
  int __pyx_v_unique;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("list_validator (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_compiler,&__pyx_n_s_items,&__pyx_n_s_minlen,&__pyx_n_s_maxlen,&__pyx_n_s_unique,0};
    PyObject* values[5] = {0,0,0,0,0};
    values[1] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_compiler)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_items);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_minlen);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_maxlen);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_unique);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "list_validator") < 0)) __PYX_ERR(0, 315, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_compiler = values[0];
    __pyx_v_items = values[1];
    if (values[2]) {
      __pyx_v_minlen = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_minlen == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 315, __pyx_L3_error)
    } else {
      __pyx_v_minlen = ((int)0);
    }
    if (values[3]) {
      __pyx_v_maxlen = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_maxlen == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 315, __pyx_L3_error)
    } else {
      __pyx_v_maxlen = ((int)0x400);
    }
    if (values[4]) {
      __pyx_v_unique = __Pyx_PyObject_IsTrue(values[4]); if (unlikely((__pyx_v_unique == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 316, __pyx_L3_error)
    } else {

      /* "validr/_validator_c.pyx":316
 * @validator(accept=typing.Iterable, output=typing.List)
 * def list_validator(compiler, items=None, int minlen=0, int maxlen=1024,
 *                    bint unique=False):             # <<<<<<<<<<<<<<
 *     if items is None:
 *         inner = None
 */
      __pyx_v_unique = ((int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("list_validator", 0, 1, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 315, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("validr._validator_c.list_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6validr_12_validator_c_20list_validator(__pyx_self, __pyx_v_compiler, __pyx_v_items, __pyx_v_minlen, __pyx_v_maxlen, __pyx_v_unique);

  /* "validr/_validator_c.pyx":315
 * 
 * @validator(accept=typing.Iterable, output=typing.List)
 * def list_validator(compiler, items=None, int minlen=0, int maxlen=1024,             # <<<<<<<<<<<<<<
 *                    bint unique=False):
 *     if items is None:
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":326
 *     del compiler, items
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         try:
 *             value = enumerate(value)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_14list_validator_1validate(PyObject *__pyx_self, PyObject *__pyx_v_value); /*proto*/
static PyMethodDef __pyx_mdef_6validr_12_validator_c_14list_validator_1validate = {"validate", (PyCFunction)__pyx_pw_6validr_12_validator_c_14list_validator_1validate, METH_O, 0};
static PyObject *__pyx_pw_6validr_12_validator_c_14list_validator_1validate(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("validate (wrapper)", 0);
  __pyx_r = __pyx_pf_6validr_12_validator_c_14list_validator_validate(__pyx_self, ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6validr_12_validator_c_14list_validator_validate(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_8_list_validator *__pyx_cur_scope;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_8_list_validator *__pyx_outer_scope;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_v_keys = NULL;
  int __pyx_v_i;
  PyObject *__pyx_v_x = NULL;
  PyObject *__pyx_v_v = NULL;
  PyObject *__pyx_v_k = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  PyObject *(*__pyx_t_13)(PyObject *);
  PyObject *(*__pyx_t_14)(PyObject *);
  PyObject *__pyx_t_15 = NULL;
  int __pyx_t_16;
  int __pyx_t_17;
  PyObject *__pyx_t_18 = NULL;
  __Pyx_RefNannySetupContext("validate", 0);
  __pyx_outer_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_8_list_validator *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_TraceCall("validate", __pyx_f[0], 326, 0, __PYX_ERR(0, 326, __pyx_L1_error));
  __Pyx_INCREF(__pyx_v_value);

  /* "validr/_validator_c.pyx":327
 * 
 *     def validate(value):
 *         try:             # <<<<<<<<<<<<<<
 *             value = enumerate(value)
 *         except TypeError:
 */
  __Pyx_TraceLine(327,0,__PYX_ERR(0, 327, __pyx_L1_error))
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "validr/_validator_c.pyx":328
 *     def validate(value):
 *         try:
 *             value = enumerate(value)             # <<<<<<<<<<<<<<
 *         except TypeError:
 *             raise Invalid('not list')
 */
      __Pyx_TraceLine(328,0,__PYX_ERR(0, 328, __pyx_L3_error))
      __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_enumerate, __pyx_v_value); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 328, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "validr/_validator_c.pyx":327
 * 
 *     def validate(value):
 *         try:             # <<<<<<<<<<<<<<
 *             value = enumerate(value)
 *         except TypeError:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "validr/_validator_c.pyx":329
 *         try:
 *             value = enumerate(value)
 *         except TypeError:             # <<<<<<<<<<<<<<
 *             raise Invalid('not list')
 *         result = []
 */
    __Pyx_TraceLine(329,0,__PYX_ERR(0, 329, __pyx_L5_except_error))
    __pyx_t_5 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError);
    if (__pyx_t_5) {
      __Pyx_AddTraceback("validr._validator_c.list_validator.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(0, 329, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "validr/_validator_c.pyx":330
 *             value = enumerate(value)
 *         except TypeError:
 *             raise Invalid('not list')             # <<<<<<<<<<<<<<
 *         result = []
 *         if unique:
 */
      __Pyx_TraceLine(330,0,__PYX_ERR(0, 330, __pyx_L5_except_error))
      __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 330, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_9, function);
        }
      }
      __pyx_t_8 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_10, __pyx_kp_u_not_list) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_kp_u_not_list);
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 330, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(0, 330, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "validr/_validator_c.pyx":327
 * 
 *     def validate(value):
 *         try:             # <<<<<<<<<<<<<<
 *             value = enumerate(value)
 *         except TypeError:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "validr/_validator_c.pyx":331
 *         except TypeError:
 *             raise Invalid('not list')
 *         result = []             # <<<<<<<<<<<<<<
 *         if unique:
 *             keys = set()
 */
  __Pyx_TraceLine(331,0,__PYX_ERR(0, 331, __pyx_L1_error))
  __pyx_t_7 = PyList_New(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 331, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_v_result = ((PyObject*)__pyx_t_7);
  __pyx_t_7 = 0;

  /* "validr/_validator_c.pyx":332
 *             raise Invalid('not list')
 *         result = []
 *         if unique:             # <<<<<<<<<<<<<<
 *             keys = set()
 *         cdef int i = -1
 */
  __Pyx_TraceLine(332,0,__PYX_ERR(0, 332, __pyx_L1_error))
  __pyx_t_11 = (__pyx_cur_scope->__pyx_v_unique != 0);
  if (__pyx_t_11) {

    /* "validr/_validator_c.pyx":333
 *         result = []
 *         if unique:
 *             keys = set()             # <<<<<<<<<<<<<<
 *         cdef int i = -1
 *         for i, x in value:
 */
    __Pyx_TraceLine(333,0,__PYX_ERR(0, 333, __pyx_L1_error))
    __pyx_t_7 = PySet_New(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 333, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_v_keys = ((PyObject*)__pyx_t_7);
    __pyx_t_7 = 0;

    /* "validr/_validator_c.pyx":332
 *             raise Invalid('not list')
 *         result = []
 *         if unique:             # <<<<<<<<<<<<<<
 *             keys = set()
 *         cdef int i = -1
 */
  }

  /* "validr/_validator_c.pyx":334
 *         if unique:
 *             keys = set()
 *         cdef int i = -1             # <<<<<<<<<<<<<<
 *         for i, x in value:
 *             if i >= maxlen:
 */
  __Pyx_TraceLine(334,0,__PYX_ERR(0, 334, __pyx_L1_error))
  __pyx_v_i = -1;

  /* "validr/_validator_c.pyx":335
 *             keys = set()
 *         cdef int i = -1
 *         for i, x in value:             # <<<<<<<<<<<<<<
 *             if i >= maxlen:
 *                 raise Invalid('list length must <= %d' % maxlen)
 */
  __Pyx_TraceLine(335,0,__PYX_ERR(0, 335, __pyx_L1_error))
  if (likely(PyList_CheckExact(__pyx_v_value)) || PyTuple_CheckExact(__pyx_v_value)) {
    __pyx_t_7 = __pyx_v_value; __Pyx_INCREF(__pyx_t_7); __pyx_t_12 = 0;
    __pyx_t_13 = NULL;
  } else {
    __pyx_t_12 = -1; __pyx_t_7 = PyObject_GetIter(__pyx_v_value); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 335, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_13 = Py_TYPE(__pyx_t_7)->tp_iternext; if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 335, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_13)) {
      if (likely(PyList_CheckExact(__pyx_t_7))) {
        if (__pyx_t_12 >= PyList_GET_SIZE(__pyx_t_7)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_6 = PyList_GET_ITEM(__pyx_t_7, __pyx_t_12); __Pyx_INCREF(__pyx_t_6); __pyx_t_12++; if (unlikely(0 < 0)) __PYX_ERR(0, 335, __pyx_L1_error)
        #else
        __pyx_t_6 = PySequence_ITEM(__pyx_t_7, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 335, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        #endif
      } else {
        if (__pyx_t_12 >= PyTuple_GET_SIZE(__pyx_t_7)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_7, __pyx_t_12); __Pyx_INCREF(__pyx_t_6); __pyx_t_12++; if (unlikely(0 < 0)) __PYX_ERR(0, 335, __pyx_L1_error)
        #else
        __pyx_t_6 = PySequence_ITEM(__pyx_t_7, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 335, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        #endif
      }
    } else {
      __pyx_t_6 = __pyx_t_13(__pyx_t_7);
      if (unlikely(!__pyx_t_6)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 335, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_6);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_6))) || (PyList_CheckExact(__pyx_t_6))) {
      PyObject* sequence = __pyx_t_6;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 335, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_8 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_8 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_8);
      #else
      __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 335, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_8 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 335, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      #endif
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_9 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 335, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_14 = Py_TYPE(__pyx_t_9)->tp_iternext;
      index = 0; __pyx_t_4 = __pyx_t_14(__pyx_t_9); if (unlikely(!__pyx_t_4)) goto __pyx_L14_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_4);
      index = 1; __pyx_t_8 = __pyx_t_14(__pyx_t_9); if (unlikely(!__pyx_t_8)) goto __pyx_L14_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_8);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_14(__pyx_t_9), 2) < 0) __PYX_ERR(0, 335, __pyx_L1_error)
      __pyx_t_14 = NULL;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      goto __pyx_L15_unpacking_done;
      __pyx_L14_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_14 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 335, __pyx_L1_error)
      __pyx_L15_unpacking_done:;
    }
    __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_4); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 335, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_i = __pyx_t_5;
    __Pyx_XDECREF_SET(__pyx_v_x, __pyx_t_8);
    __pyx_t_8 = 0;

    /* "validr/_validator_c.pyx":336
 *         cdef int i = -1
 *         for i, x in value:
 *             if i >= maxlen:             # <<<<<<<<<<<<<<
 *                 raise Invalid('list length must <= %d' % maxlen)
 *             with mark_index(i):
 */
    __Pyx_TraceLine(336,0,__PYX_ERR(0, 336, __pyx_L1_error))
    __pyx_t_11 = ((__pyx_v_i >= __pyx_cur_scope->__pyx_v_maxlen) != 0);
    if (unlikely(__pyx_t_11)) {

      /* "validr/_validator_c.pyx":337
 *         for i, x in value:
 *             if i >= maxlen:
 *                 raise Invalid('list length must <= %d' % maxlen)             # <<<<<<<<<<<<<<
 *             with mark_index(i):
 *                 v = inner(x) if inner is not None else copy(x)
 */
      __Pyx_TraceLine(337,0,__PYX_ERR(0, 337, __pyx_L1_error))
      __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 337, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_cur_scope->__pyx_v_maxlen); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 337, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_9 = PyUnicode_Format(__pyx_kp_u_list_length_must_d, __pyx_t_4); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 337, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
        }
      }
      __pyx_t_6 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_4, __pyx_t_9) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_9);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 337, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(0, 337, __pyx_L1_error)

      /* "validr/_validator_c.pyx":336
 *         cdef int i = -1
 *         for i, x in value:
 *             if i >= maxlen:             # <<<<<<<<<<<<<<
 *                 raise Invalid('list length must <= %d' % maxlen)
 *             with mark_index(i):
 */
    }

    /* "validr/_validator_c.pyx":338
 *             if i >= maxlen:
 *                 raise Invalid('list length must <= %d' % maxlen)
 *             with mark_index(i):             # <<<<<<<<<<<<<<
 *                 v = inner(x) if inner is not None else copy(x)
 *                 if unique:
 */
    __Pyx_TraceLine(338,0,__PYX_ERR(0, 338, __pyx_L1_error))
    /*with:*/ {
      __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_mark_index); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 338, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 338, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
        }
      }
      __pyx_t_6 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_4, __pyx_t_9) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_9);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 338, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_3 = __Pyx_PyObject_LookupSpecial(__pyx_t_6, __pyx_n_s_exit); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 338, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_9 = __Pyx_PyObject_LookupSpecial(__pyx_t_6, __pyx_n_s_enter); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 338, __pyx_L17_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_9);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_9, function);
        }
      }
      __pyx_t_8 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 338, __pyx_L17_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      /*try:*/ {
        {
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_1, &__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_2);
          __Pyx_XGOTREF(__pyx_t_1);
          __Pyx_XGOTREF(__pyx_t_15);
          /*try:*/ {

            /* "validr/_validator_c.pyx":339
 *                 raise Invalid('list length must <= %d' % maxlen)
 *             with mark_index(i):
 *                 v = inner(x) if inner is not None else copy(x)             # <<<<<<<<<<<<<<
 *                 if unique:
 *                     k = key_of(v)
 */
            __Pyx_TraceLine(339,0,__PYX_ERR(0, 339, __pyx_L23_error))
            if (unlikely(!__pyx_cur_scope->__pyx_v_inner)) { __Pyx_RaiseClosureNameError("inner"); __PYX_ERR(0, 339, __pyx_L23_error) }
            __pyx_t_11 = (__pyx_cur_scope->__pyx_v_inner != Py_None);
            if ((__pyx_t_11 != 0)) {
              if (unlikely(!__pyx_cur_scope->__pyx_v_inner)) { __Pyx_RaiseClosureNameError("inner"); __PYX_ERR(0, 339, __pyx_L23_error) }
              __Pyx_INCREF(__pyx_cur_scope->__pyx_v_inner);
              __pyx_t_9 = __pyx_cur_scope->__pyx_v_inner; __pyx_t_4 = NULL;
              if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
                __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_9);
                if (likely(__pyx_t_4)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
                  __Pyx_INCREF(__pyx_t_4);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_9, function);
                }
              }
              __pyx_t_8 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_4, __pyx_v_x) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_v_x);
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 339, __pyx_L23_error)
              __Pyx_GOTREF(__pyx_t_8);
              __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
              __pyx_t_6 = __pyx_t_8;
              __pyx_t_8 = 0;
            } else {
              __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_copy); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 339, __pyx_L23_error)
              __Pyx_GOTREF(__pyx_t_9);
              __pyx_t_4 = NULL;
              if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
                __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_9);
                if (likely(__pyx_t_4)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
                  __Pyx_INCREF(__pyx_t_4);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_9, function);
                }
              }
              __pyx_t_8 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_4, __pyx_v_x) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_v_x);
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 339, __pyx_L23_error)
              __Pyx_GOTREF(__pyx_t_8);
              __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
              __pyx_t_6 = __pyx_t_8;
              __pyx_t_8 = 0;
            }
            __Pyx_XDECREF_SET(__pyx_v_v, __pyx_t_6);
            __pyx_t_6 = 0;

            /* "validr/_validator_c.pyx":340
 *             with mark_index(i):
 *                 v = inner(x) if inner is not None else copy(x)
 *                 if unique:             # <<<<<<<<<<<<<<
 *                     k = key_of(v)
 *                     if k in keys:
 */
            __Pyx_TraceLine(340,0,__PYX_ERR(0, 340, __pyx_L23_error))
            __pyx_t_11 = (__pyx_cur_scope->__pyx_v_unique != 0);
            if (__pyx_t_11) {

              /* "validr/_validator_c.pyx":341
 *                 v = inner(x) if inner is not None else copy(x)
 *                 if unique:
 *                     k = key_of(v)             # <<<<<<<<<<<<<<
 *                     if k in keys:
 *                         raise Invalid('not unique')
 */
              __Pyx_TraceLine(341,0,__PYX_ERR(0, 341, __pyx_L23_error))
              if (unlikely(!__pyx_cur_scope->__pyx_v_key_of)) { __Pyx_RaiseClosureNameError("key_of"); __PYX_ERR(0, 341, __pyx_L23_error) }
              __Pyx_INCREF(__pyx_cur_scope->__pyx_v_key_of);
              __pyx_t_8 = __pyx_cur_scope->__pyx_v_key_of; __pyx_t_9 = NULL;
              if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
                __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
                if (likely(__pyx_t_9)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
                  __Pyx_INCREF(__pyx_t_9);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_8, function);
                }
              }
              __pyx_t_6 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_9, __pyx_v_v) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_v_v);
              __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
              if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 341, __pyx_L23_error)
              __Pyx_GOTREF(__pyx_t_6);
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
              __Pyx_XDECREF_SET(__pyx_v_k, __pyx_t_6);
              __pyx_t_6 = 0;

              /* "validr/_validator_c.pyx":342
 *                 if unique:
 *                     k = key_of(v)
 *                     if k in keys:             # <<<<<<<<<<<<<<
 *                         raise Invalid('not unique')
 *                     keys.add(k)
 */
              __Pyx_TraceLine(342,0,__PYX_ERR(0, 342, __pyx_L23_error))
              if (unlikely(!__pyx_v_keys)) { __Pyx_RaiseUnboundLocalError("keys"); __PYX_ERR(0, 342, __pyx_L23_error) }
              __pyx_t_11 = (__Pyx_PySet_ContainsTF(__pyx_v_k, __pyx_v_keys, Py_EQ)); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 342, __pyx_L23_error)
              __pyx_t_16 = (__pyx_t_11 != 0);
              if (unlikely(__pyx_t_16)) {

                /* "validr/_validator_c.pyx":343
 *                     k = key_of(v)
 *                     if k in keys:
 *                         raise Invalid('not unique')             # <<<<<<<<<<<<<<
 *                     keys.add(k)
 *             result.append(v)
 */
                __Pyx_TraceLine(343,0,__PYX_ERR(0, 343, __pyx_L23_error))
                __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 343, __pyx_L23_error)
                __Pyx_GOTREF(__pyx_t_8);
                __pyx_t_9 = NULL;
                if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
                  __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
                  if (likely(__pyx_t_9)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
                    __Pyx_INCREF(__pyx_t_9);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_8, function);
                  }
                }
                __pyx_t_6 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_9, __pyx_kp_u_not_unique) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_kp_u_not_unique);
                __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
                if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 343, __pyx_L23_error)
                __Pyx_GOTREF(__pyx_t_6);
                __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
                __Pyx_Raise(__pyx_t_6, 0, 0, 0);
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                __PYX_ERR(0, 343, __pyx_L23_error)

                /* "validr/_validator_c.pyx":342
 *                 if unique:
 *                     k = key_of(v)
 *                     if k in keys:             # <<<<<<<<<<<<<<
 *                         raise Invalid('not unique')
 *                     keys.add(k)
 */
              }

              /* "validr/_validator_c.pyx":344
 *                     if k in keys:
 *                         raise Invalid('not unique')
 *                     keys.add(k)             # <<<<<<<<<<<<<<
 *             result.append(v)
 *         if minlen > 0 and i + 1 < minlen:
 */
              __Pyx_TraceLine(344,0,__PYX_ERR(0, 344, __pyx_L23_error))
              if (unlikely(!__pyx_v_keys)) { __Pyx_RaiseUnboundLocalError("keys"); __PYX_ERR(0, 344, __pyx_L23_error) }
              __pyx_t_17 = PySet_Add(__pyx_v_keys, __pyx_v_k); if (unlikely(__pyx_t_17 == ((int)-1))) __PYX_ERR(0, 344, __pyx_L23_error)

              /* "validr/_validator_c.pyx":340
 *             with mark_index(i):
 *                 v = inner(x) if inner is not None else copy(x)
 *                 if unique:             # <<<<<<<<<<<<<<
 *                     k = key_of(v)
 *                     if k in keys:
 */
            }

            /* "validr/_validator_c.pyx":338
 *             if i >= maxlen:
 *                 raise Invalid('list length must <= %d' % maxlen)
 *             with mark_index(i):             # <<<<<<<<<<<<<<
 *                 v = inner(x) if inner is not None else copy(x)
 *                 if unique:
 */
          }
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
          goto __pyx_L30_try_end;
          __pyx_L23_error:;
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          /*except:*/ {
            __Pyx_AddTraceback("validr._validator_c.list_validator.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
            if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_8, &__pyx_t_9) < 0) __PYX_ERR(0, 338, __pyx_L25_except_error)
            __Pyx_GOTREF(__pyx_t_6);
            __Pyx_GOTREF(__pyx_t_8);
            __Pyx_GOTREF(__pyx_t_9);
            __pyx_t_4 = PyTuple_Pack(3, __pyx_t_6, __pyx_t_8, __pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 338, __pyx_L25_except_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_18 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 338, __pyx_L25_except_error)
            __Pyx_GOTREF(__pyx_t_18);
            __pyx_t_16 = __Pyx_PyObject_IsTrue(__pyx_t_18);
            __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
            if (__pyx_t_16 < 0) __PYX_ERR(0, 338, __pyx_L25_except_error)
            __pyx_t_11 = ((!(__pyx_t_16 != 0)) != 0);
            if (__pyx_t_11) {
              __Pyx_GIVEREF(__pyx_t_6);
              __Pyx_GIVEREF(__pyx_t_8);
              __Pyx_XGIVEREF(__pyx_t_9);
              __Pyx_ErrRestoreWithState(__pyx_t_6, __pyx_t_8, __pyx_t_9);
              __pyx_t_6 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; 
              __PYX_ERR(0, 338, __pyx_L25_except_error)
            }
            __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
            goto __pyx_L24_exception_handled;
          }
          __pyx_L25_except_error:;
          __Pyx_XGIVEREF(__pyx_t_2);
          __Pyx_XGIVEREF(__pyx_t_1);
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_1, __pyx_t_15);
          goto __pyx_L1_error;
          __pyx_L24_exception_handled:;
          __Pyx_XGIVEREF(__pyx_t_2);
          __Pyx_XGIVEREF(__pyx_t_1);
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_1, __pyx_t_15);
          __pyx_L30_try_end:;
        }
      }
      /*finally:*/ {
        /*normal exit:*/{
          if (__pyx_t_3) {
            __pyx_t_15 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__28, NULL);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 338, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_15);
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          }
          goto __pyx_L22;
        }
        __pyx_L22:;
      }
      goto __pyx_L36;
      __pyx_L17_error:;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L1_error;
      __pyx_L36:;
    }

    /* "validr/_validator_c.pyx":345
 *                         raise Invalid('not unique')
 *                     keys.add(k)
 *             result.append(v)             # <<<<<<<<<<<<<<
 *         if minlen > 0 and i + 1 < minlen:
 *             raise Invalid('list length must >= %d' % minlen)
 */
    __Pyx_TraceLine(345,0,__PYX_ERR(0, 345, __pyx_L1_error))
    if (unlikely(!__pyx_v_v)) { __Pyx_RaiseUnboundLocalError("v"); __PYX_ERR(0, 345, __pyx_L1_error) }
    __pyx_t_17 = __Pyx_PyList_Append(__pyx_v_result, __pyx_v_v); if (unlikely(__pyx_t_17 == ((int)-1))) __PYX_ERR(0, 345, __pyx_L1_error)

    /* "validr/_validator_c.pyx":335
 *             keys = set()
 *         cdef int i = -1
 *         for i, x in value:             # <<<<<<<<<<<<<<
 *             if i >= maxlen:
 *                 raise Invalid('list length must <= %d' % maxlen)
 */
    __Pyx_TraceLine(335,0,__PYX_ERR(0, 335, __pyx_L1_error))
  }
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "validr/_validator_c.pyx":346
 *                     keys.add(k)
 *             result.append(v)
 *         if minlen > 0 and i + 1 < minlen:             # <<<<<<<<<<<<<<
 *             raise Invalid('list length must >= %d' % minlen)
 *         return result
 */
  __Pyx_TraceLine(346,0,__PYX_ERR(0, 346, __pyx_L1_error))
  __pyx_t_16 = ((__pyx_cur_scope->__pyx_v_minlen > 0) != 0);
  if (__pyx_t_16) {
  } else {
    __pyx_t_11 = __pyx_t_16;
    goto __pyx_L38_bool_binop_done;
  }
  __pyx_t_16 = (((__pyx_v_i + 1) < __pyx_cur_scope->__pyx_v_minlen) != 0);
  __pyx_t_11 = __pyx_t_16;
  __pyx_L38_bool_binop_done:;
  if (unlikely(__pyx_t_11)) {

    /* "validr/_validator_c.pyx":347
 *             result.append(v)
 *         if minlen > 0 and i + 1 < minlen:
 *             raise Invalid('list length must >= %d' % minlen)             # <<<<<<<<<<<<<<
 *         return result
 *     return validate
 */
    __Pyx_TraceLine(347,0,__PYX_ERR(0, 347, __pyx_L1_error))
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 347, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_cur_scope->__pyx_v_minlen); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 347, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_6 = PyUnicode_Format(__pyx_kp_u_list_length_must_d_2, __pyx_t_8); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 347, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
      }
    }
    __pyx_t_7 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 347, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_Raise(__pyx_t_7, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __PYX_ERR(0, 347, __pyx_L1_error)

    /* "validr/_validator_c.pyx":346
 *                     keys.add(k)
 *             result.append(v)
 *         if minlen > 0 and i + 1 < minlen:             # <<<<<<<<<<<<<<
 *             raise Invalid('list length must >= %d' % minlen)
 *         return result
 */
  }

  /* "validr/_validator_c.pyx":348
 *         if minlen > 0 and i + 1 < minlen:
 *             raise Invalid('list length must >= %d' % minlen)
 *         return result             # <<<<<<<<<<<<<<
 *     return validate
 * 
 */
  __Pyx_TraceLine(348,0,__PYX_ERR(0, 348, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "validr/_validator_c.pyx":326
 *     del compiler, items
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         try:
 *             value = enumerate(value)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("validr._validator_c.list_validator.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_keys);
  __Pyx_XDECREF(__pyx_v_x);
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_XDECREF(__pyx_v_k);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":315
 * 
 * @validator(accept=typing.Iterable, output=typing.List)
 * def list_validator(compiler, items=None, int minlen=0, int maxlen=1024,             # <<<<<<<<<<<<<<
 *                    bint unique=False):
 *     if items is None:
 */

static PyObject *__pyx_pf_6validr_12_validator_c_20list_validator(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_compiler, PyObject *__pyx_v_items, int __pyx_v_minlen, int __pyx_v_maxlen, int __pyx_v_unique) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_8_list_validator *__pyx_cur_scope;
  PyObject *__pyx_v_validate = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  __Pyx_TraceFrameInit(__pyx_codeobj__29)
  __Pyx_RefNannySetupContext("list_validator", 0);
  __pyx_cur_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_8_list_validator *)__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_8_list_validator(__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_8_list_validator, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_8_list_validator *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 315, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __Pyx_TraceCall("list_validator", __pyx_f[0], 315, 0, __PYX_ERR(0, 315, __pyx_L1_error));
  __pyx_cur_scope->__pyx_v_minlen = __pyx_v_minlen;
  __pyx_cur_scope->__pyx_v_maxlen = __pyx_v_maxlen;
  __pyx_cur_scope->__pyx_v_unique = __pyx_v_unique;
  __Pyx_INCREF(__pyx_v_compiler);
  __Pyx_INCREF(__pyx_v_items);

  /* "validr/_validator_c.pyx":317
 * def list_validator(compiler, items=None, int minlen=0, int maxlen=1024,
 *                    bint unique=False):
 *     if items is None:             # <<<<<<<<<<<<<<
 *         inner = None
 *     else:
 */
  __Pyx_TraceLine(317,0,__PYX_ERR(0, 317, __pyx_L1_error))
  __pyx_t_1 = (__pyx_v_items == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "validr/_validator_c.pyx":318
 *                    bint unique=False):
 *     if items is None:
 *         inner = None             # <<<<<<<<<<<<<<
 *     else:
 *         with mark_index():
 */
    __Pyx_TraceLine(318,0,__PYX_ERR(0, 318, __pyx_L1_error))
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __pyx_cur_scope->__pyx_v_inner = Py_None;

    /* "validr/_validator_c.pyx":317
 * def list_validator(compiler, items=None, int minlen=0, int maxlen=1024,
 *                    bint unique=False):
 *     if items is None:             # <<<<<<<<<<<<<<
 *         inner = None
 *     else:
 */
    goto __pyx_L3;
  }

  /* "validr/_validator_c.pyx":320
 *         inner = None
 *     else:
 *         with mark_index():             # <<<<<<<<<<<<<<
 *             inner = compiler.compile(items)
 *     if unique:
 */
  __Pyx_TraceLine(320,0,__PYX_ERR(0, 320, __pyx_L1_error))
  /*else*/ {
    /*with:*/ {
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_mark_index); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 320, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 320, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_LookupSpecial(__pyx_t_3, __pyx_n_s_exit); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 320, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_5 = __Pyx_PyObject_LookupSpecial(__pyx_t_3, __pyx_n_s_enter); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 320, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_7 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_4 = (__pyx_t_7) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_7) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 320, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      /*try:*/ {
        {
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
          __Pyx_XGOTREF(__pyx_t_8);
          __Pyx_XGOTREF(__pyx_t_9);
          __Pyx_XGOTREF(__pyx_t_10);
          /*try:*/ {

            /* "validr/_validator_c.pyx":321
 *     else:
 *         with mark_index():
 *             inner = compiler.compile(items)             # <<<<<<<<<<<<<<
 *     if unique:
 *         key_of = _key_func_of_schema(items)
 */
            __Pyx_TraceLine(321,0,__PYX_ERR(0, 321, __pyx_L8_error))
            __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_compiler, __pyx_n_s_compile); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 321, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_5 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
              __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
              if (likely(__pyx_t_5)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                __Pyx_INCREF(__pyx_t_5);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_4, function);
              }
            }
            __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_items) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_items);
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 321, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GIVEREF(__pyx_t_3);
            __pyx_cur_scope->__pyx_v_inner = __pyx_t_3;
            __pyx_t_3 = 0;

            /* "validr/_validator_c.pyx":320
 *         inner = None
 *     else:
 *         with mark_index():             # <<<<<<<<<<<<<<
 *             inner = compiler.compile(items)
 *     if unique:
 */
          }
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          goto __pyx_L13_try_end;
          __pyx_L8_error:;
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          /*except:*/ {
            __Pyx_AddTraceback("validr._validator_c.list_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
            if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5) < 0) __PYX_ERR(0, 320, __pyx_L10_except_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_7 = PyTuple_Pack(3, __pyx_t_3, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 320, __pyx_L10_except_error)
            __Pyx_GOTREF(__pyx_t_7);
            __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_7, NULL);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 320, __pyx_L10_except_error)
            __Pyx_GOTREF(__pyx_t_11);
            __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_11);
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            if (__pyx_t_2 < 0) __PYX_ERR(0, 320, __pyx_L10_except_error)
            __pyx_t_1 = ((!(__pyx_t_2 != 0)) != 0);
            if (__pyx_t_1) {
              __Pyx_GIVEREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_4);
              __Pyx_XGIVEREF(__pyx_t_5);
              __Pyx_ErrRestoreWithState(__pyx_t_3, __pyx_t_4, __pyx_t_5);
              __pyx_t_3 = 0; __pyx_t_4 = 0; __pyx_t_5 = 0; 
              __PYX_ERR(0, 320, __pyx_L10_except_error)
            }
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            goto __pyx_L9_exception_handled;
          }
          __pyx_L10_except_error:;
          __Pyx_XGIVEREF(__pyx_t_8);
          __Pyx_XGIVEREF(__pyx_t_9);
          __Pyx_XGIVEREF(__pyx_t_10);
          __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
          goto __pyx_L1_error;
          __pyx_L9_exception_handled:;
          __Pyx_XGIVEREF(__pyx_t_8);
          __Pyx_XGIVEREF(__pyx_t_9);
          __Pyx_XGIVEREF(__pyx_t_10);
          __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
          __pyx_L13_try_end:;
        }
      }
      /*finally:*/ {
        /*normal exit:*/{
          if (__pyx_t_6) {
            __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_tuple__28, NULL);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 320, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_10);
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          }
          goto __pyx_L7;
        }
        __pyx_L7:;
      }
      goto __pyx_L17;
      __pyx_L4_error:;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      goto __pyx_L1_error;
      __pyx_L17:;
    }
  }
  __pyx_L3:;

  /* "validr/_validator_c.pyx":322
 *         with mark_index():
 *             inner = compiler.compile(items)
 *     if unique:             # <<<<<<<<<<<<<<
 *         key_of = _key_func_of_schema(items)
 *     del compiler, items
 */
  __Pyx_TraceLine(322,0,__PYX_ERR(0, 322, __pyx_L1_error))
  __pyx_t_1 = (__pyx_cur_scope->__pyx_v_unique != 0);
  if (__pyx_t_1) {

    /* "validr/_validator_c.pyx":323
 *             inner = compiler.compile(items)
 *     if unique:
 *         key_of = _key_func_of_schema(items)             # <<<<<<<<<<<<<<
 *     del compiler, items
 * 
 */
    __Pyx_TraceLine(323,0,__PYX_ERR(0, 323, __pyx_L1_error))
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_key_func_of_schema); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 323, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_5 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_v_items) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_items);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 323, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_cur_scope->__pyx_v_key_of = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "validr/_validator_c.pyx":322
 *         with mark_index():
 *             inner = compiler.compile(items)
 *     if unique:             # <<<<<<<<<<<<<<
 *         key_of = _key_func_of_schema(items)
 *     del compiler, items
 */
  }

  /* "validr/_validator_c.pyx":324
 *     if unique:
 *         key_of = _key_func_of_schema(items)
 *     del compiler, items             # <<<<<<<<<<<<<<
 * 
 *     def validate(value):
 */
  __Pyx_TraceLine(324,0,__PYX_ERR(0, 324, __pyx_L1_error))
  __Pyx_DECREF(__pyx_v_compiler);
  __pyx_v_compiler = NULL;
  __Pyx_DECREF(__pyx_v_items);
  __pyx_v_items = NULL;

  /* "validr/_validator_c.pyx":326
 *     del compiler, items
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         try:
 *             value = enumerate(value)
 */
  __Pyx_TraceLine(326,0,__PYX_ERR(0, 326, __pyx_L1_error))
  __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_14list_validator_1validate, 0, __pyx_n_s_list_validator_locals_validate, ((PyObject*)__pyx_cur_scope), __pyx_n_s_validr__validator_c, __pyx_d, ((PyObject *)__pyx_codeobj__31)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_v_validate = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "validr/_validator_c.pyx":349
 *             raise Invalid('list length must >= %d' % minlen)
 *         return result
 *     return validate             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(349,0,__PYX_ERR(0, 349, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_validate);
  __pyx_r = __pyx_v_validate;
  goto __pyx_L0;

  /* "validr/_validator_c.pyx":315
 * 
 * @validator(accept=typing.Iterable, output=typing.List)
 * def list_validator(compiler, items=None, int minlen=0, int maxlen=1024,             # <<<<<<<<<<<<<<
 *                    bint unique=False):
 *     if items is None:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("validr._validator_c.list_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_validate);
  __Pyx_XDECREF(__pyx_v_compiler);
  __Pyx_XDECREF(__pyx_v_items);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":352
 * 
 * 
 * cpdef inline dict _slim_dict(dict value):             # <<<<<<<<<<<<<<
 *     return {k: v for k, v in value.items() if not _is_empty(v)}
 * 
 */

static PyObject *__pyx_pw_6validr_12_validator_c_23_slim_dict(PyObject *__pyx_self, PyObject *__pyx_v_value); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_6validr_12_validator_c__slim_dict(PyObject *__pyx_v_value, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_8genexpr2__pyx_v_k = NULL;
  PyObject *__pyx_8genexpr2__pyx_v_v = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_t_9;
  __Pyx_RefNannySetupContext("_slim_dict", 0);
  __Pyx_TraceCall("_slim_dict", __pyx_f[0], 352, 0, __PYX_ERR(0, 352, __pyx_L1_error));

  /* "validr/_validator_c.pyx":353
 * 
 * cpdef inline dict _slim_dict(dict value):
 *     return {k: v for k, v in value.items() if not _is_empty(v)}             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(353,0,__PYX_ERR(0, 353, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 353, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = 0;
    if (unlikely(__pyx_v_value == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
      __PYX_ERR(0, 353, __pyx_L5_error)
    }
    __pyx_t_6 = __Pyx_dict_iterator(__pyx_v_value, 1, __pyx_n_s_items, (&__pyx_t_4), (&__pyx_t_5)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 353, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_2);
    __pyx_t_2 = __pyx_t_6;
    __pyx_t_6 = 0;
    while (1) {
      __pyx_t_8 = __Pyx_dict_iter_next(__pyx_t_2, __pyx_t_4, &__pyx_t_3, &__pyx_t_6, &__pyx_t_7, NULL, __pyx_t_5);
      if (unlikely(__pyx_t_8 == 0)) break;
      if (unlikely(__pyx_t_8 == -1)) __PYX_ERR(0, 353, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_XDECREF_SET(__pyx_8genexpr2__pyx_v_k, __pyx_t_6);
      __pyx_t_6 = 0;
      __Pyx_XDECREF_SET(__pyx_8genexpr2__pyx_v_v, __pyx_t_7);
      __pyx_t_7 = 0;
      __pyx_t_9 = ((!(__pyx_f_6validr_12_validator_c__is_empty(__pyx_8genexpr2__pyx_v_v, 0) != 0)) != 0);
      if (__pyx_t_9) {
        if (unlikely(PyDict_SetItem(__pyx_t_1, (PyObject*)__pyx_8genexpr2__pyx_v_k, (PyObject*)__pyx_8genexpr2__pyx_v_v))) __PYX_ERR(0, 353, __pyx_L5_error)
      }
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_k); __pyx_8genexpr2__pyx_v_k = 0;
    __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_v); __pyx_8genexpr2__pyx_v_v = 0;
    goto __pyx_L9_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_k); __pyx_8genexpr2__pyx_v_k = 0;
    __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_v); __pyx_8genexpr2__pyx_v_v = 0;
    goto __pyx_L1_error;
    __pyx_L9_exit_scope:;
  } /* exit inner scope */
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "validr/_validator_c.pyx":352
 * 
 * 
 * cpdef inline dict _slim_dict(dict value):             # <<<<<<<<<<<<<<
 *     return {k: v for k, v in value.items() if not _is_empty(v)}
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("validr._validator_c._slim_dict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_k);
  __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_v);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_23_slim_dict(PyObject *__pyx_self, PyObject *__pyx_v_value); /*proto*/
static PyObject *__pyx_pw_6validr_12_validator_c_23_slim_dict(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_slim_dict (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_value), (&PyDict_Type), 1, "value", 1))) __PYX_ERR(0, 352, __pyx_L1_error)
  __pyx_r = __pyx_pf_6validr_12_validator_c_22_slim_dict(__pyx_self, ((PyObject*)__pyx_v_value));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6validr_12_validator_c_22_slim_dict(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_slim_dict", 0);
  __Pyx_TraceCall("_slim_dict (wrapper)", __pyx_f[0], 352, 0, __PYX_ERR(0, 352, __pyx_L1_error));
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6validr_12_validator_c__slim_dict(__pyx_v_value, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 352, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("validr._validator_c._slim_dict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":357
 * 
 * @validator(accept=(typing.Mapping, typing.Any), output=dict)
 * def dict_validator(compiler, items=None, key=None, value=None,             # <<<<<<<<<<<<<<
 *                    int minlen=0, int maxlen=1024, bint slim=False):
 *     if items is None:
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_25dict_validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6validr_12_validator_c_25dict_validator = {"dict_validator", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6validr_12_validator_c_25dict_validator, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6validr_12_validator_c_25dict_validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_compiler = 0;
  PyObject *__pyx_v_items = 0;
  PyObject *__pyx_v_key = 0;
  PyObject *__pyx_v_value = 0;
  int __pyx_v_minlen;
  int __pyx_v_maxlen;
  int __pyx_v_slim;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dict_validator (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_compiler,&__pyx_n_s_items,&__pyx_n_s_key,&__pyx_n_s_value,&__pyx_n_s_minlen,&__pyx_n_s_maxlen,&__pyx_n_s_slim,0};
    PyObject* values[7] = {0,0,0,0,0,0,0};
    values[1] = ((PyObject *)Py_None);
    values[2] = ((PyObject *)Py_None);
    values[3] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_compiler)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_items);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_key);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_value);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_minlen);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_maxlen);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_slim);
          if (value) { values[6] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dict_validator") < 0)) __PYX_ERR(0, 357, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_compiler = values[0];
    __pyx_v_items = values[1];
    __pyx_v_key = values[2];
    __pyx_v_value = values[3];
    if (values[4]) {
      __pyx_v_minlen = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_minlen == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 358, __pyx_L3_error)
    } else {
      __pyx_v_minlen = ((int)0);
    }
    if (values[5]) {
      __pyx_v_maxlen = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_maxlen == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 358, __pyx_L3_error)
    } else {
      __pyx_v_maxlen = ((int)0x400);
    }
    if (values[6]) {
      __pyx_v_slim = __Pyx_PyObject_IsTrue(values[6]); if (unlikely((__pyx_v_slim == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 358, __pyx_L3_error)
    } else {

      /* "validr/_validator_c.pyx":358
 * @validator(accept=(typing.Mapping, typing.Any), output=dict)
 * def dict_validator(compiler, items=None, key=None, value=None,
 *                    int minlen=0, int maxlen=1024, bint slim=False):             # <<<<<<<<<<<<<<
 *     if items is None:
 *         inners = None
 */
      __pyx_v_slim = ((int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dict_validator", 0, 1, 7, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 357, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("validr._validator_c.dict_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6validr_12_validator_c_24dict_validator(__pyx_self, __pyx_v_compiler, __pyx_v_items, __pyx_v_key, __pyx_v_value, __pyx_v_minlen, __pyx_v_maxlen, __pyx_v_slim);

  /* "validr/_validator_c.pyx":357
 * 
 * @validator(accept=(typing.Mapping, typing.Any), output=dict)
 * def dict_validator(compiler, items=None, key=None, value=None,             # <<<<<<<<<<<<<<
 *                    int minlen=0, int maxlen=1024, bint slim=False):
 *     if items is None:
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":377
 *     del compiler, items, key, value
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         if inners is None and not is_dynamic:
 *             if not is_dict(value):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_14dict_validator_1validate(PyObject *__pyx_self, PyObject *__pyx_v_value); /*proto*/
static PyMethodDef __pyx_mdef_6validr_12_validator_c_14dict_validator_1validate = {"validate", (PyCFunction)__pyx_pw_6validr_12_validator_c_14dict_validator_1validate, METH_O, 0};
static PyObject *__pyx_pw_6validr_12_validator_c_14dict_validator_1validate(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("validate (wrapper)", 0);
  __pyx_r = __pyx_pf_6validr_12_validator_c_14dict_validator_validate(__pyx_self, ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6validr_12_validator_c_14dict_validator_validate(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_9_dict_validator *__pyx_cur_scope;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_9_dict_validator *__pyx_outer_scope;
  PyObject *(*__pyx_v_getter)(PyObject *, PyObject *, int __pyx_skip_dispatch);
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_v_k = 0;
  PyObject *__pyx_v_inner = NULL;
  PyObject *__pyx_v_extra_keys = NULL;
  PyObject *__pyx_v_v = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *(*__pyx_t_10)(PyObject *);
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *(*__pyx_t_16)(PyObject *);
  __Pyx_RefNannySetupContext("validate", 0);
  __pyx_outer_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_9_dict_validator *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_TraceCall("validate", __pyx_f[0], 377, 0, __PYX_ERR(0, 377, __pyx_L1_error));
  __Pyx_INCREF(__pyx_v_value);

  /* "validr/_validator_c.pyx":378
 * 
 *     def validate(value):
 *         if inners is None and not is_dynamic:             # <<<<<<<<<<<<<<
 *             if not is_dict(value):
 *                 raise Invalid('must be dict')
 */
  __Pyx_TraceLine(378,0,__PYX_ERR(0, 378, __pyx_L1_error))
  if (unlikely(!__pyx_cur_scope->__pyx_v_inners)) { __Pyx_RaiseClosureNameError("inners"); __PYX_ERR(0, 378, __pyx_L1_error) }
  __pyx_t_2 = (__pyx_cur_scope->__pyx_v_inners == ((PyObject*)Py_None));
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = ((!(__pyx_cur_scope->__pyx_v_is_dynamic != 0)) != 0);
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "validr/_validator_c.pyx":379
 *     def validate(value):
 *         if inners is None and not is_dynamic:
 *             if not is_dict(value):             # <<<<<<<<<<<<<<
 *                 raise Invalid('must be dict')
 *             if len(value) > maxlen:
 */
    __Pyx_TraceLine(379,0,__PYX_ERR(0, 379, __pyx_L1_error))
    __pyx_t_1 = ((!(__pyx_f_6validr_12_validator_c_is_dict(__pyx_v_value, 0) != 0)) != 0);
    if (unlikely(__pyx_t_1)) {

      /* "validr/_validator_c.pyx":380
 *         if inners is None and not is_dynamic:
 *             if not is_dict(value):
 *                 raise Invalid('must be dict')             # <<<<<<<<<<<<<<
 *             if len(value) > maxlen:
 *                 raise Invalid('dict length must <= %d' % maxlen)
 */
      __Pyx_TraceLine(380,0,__PYX_ERR(0, 380, __pyx_L1_error))
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 380, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_kp_u_must_be_dict) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_kp_u_must_be_dict);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 380, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_Raise(__pyx_t_4, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __PYX_ERR(0, 380, __pyx_L1_error)

      /* "validr/_validator_c.pyx":379
 *     def validate(value):
 *         if inners is None and not is_dynamic:
 *             if not is_dict(value):             # <<<<<<<<<<<<<<
 *                 raise Invalid('must be dict')
 *             if len(value) > maxlen:
 */
    }

    /* "validr/_validator_c.pyx":381
 *             if not is_dict(value):
 *                 raise Invalid('must be dict')
 *             if len(value) > maxlen:             # <<<<<<<<<<<<<<
 *                 raise Invalid('dict length must <= %d' % maxlen)
 *             elif minlen > 0 and len(value) < minlen:
 */
    __Pyx_TraceLine(381,0,__PYX_ERR(0, 381, __pyx_L1_error))
    __pyx_t_7 = PyObject_Length(__pyx_v_value); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 381, __pyx_L1_error)
    __pyx_t_1 = ((__pyx_t_7 > __pyx_cur_scope->__pyx_v_maxlen) != 0);
    if (unlikely(__pyx_t_1)) {

      /* "validr/_validator_c.pyx":382
 *                 raise Invalid('must be dict')
 *             if len(value) > maxlen:
 *                 raise Invalid('dict length must <= %d' % maxlen)             # <<<<<<<<<<<<<<
 *             elif minlen > 0 and len(value) < minlen:
 *                 raise Invalid('dict length must >= %d' % minlen)
 */
      __Pyx_TraceLine(382,0,__PYX_ERR(0, 382, __pyx_L1_error))
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 382, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_cur_scope->__pyx_v_maxlen); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 382, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_8 = PyUnicode_Format(__pyx_kp_u_dict_length_must_d, __pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 382, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_8);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 382, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_Raise(__pyx_t_4, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __PYX_ERR(0, 382, __pyx_L1_error)

      /* "validr/_validator_c.pyx":381
 *             if not is_dict(value):
 *                 raise Invalid('must be dict')
 *             if len(value) > maxlen:             # <<<<<<<<<<<<<<
 *                 raise Invalid('dict length must <= %d' % maxlen)
 *             elif minlen > 0 and len(value) < minlen:
 */
    }

    /* "validr/_validator_c.pyx":383
 *             if len(value) > maxlen:
 *                 raise Invalid('dict length must <= %d' % maxlen)
 *             elif minlen > 0 and len(value) < minlen:             # <<<<<<<<<<<<<<
 *                 raise Invalid('dict length must >= %d' % minlen)
 *             if slim:
 */
    __Pyx_TraceLine(383,0,__PYX_ERR(0, 383, __pyx_L1_error))
    __pyx_t_3 = ((__pyx_cur_scope->__pyx_v_minlen > 0) != 0);
    if (__pyx_t_3) {
    } else {
      __pyx_t_1 = __pyx_t_3;
      goto __pyx_L8_bool_binop_done;
    }
    __pyx_t_7 = PyObject_Length(__pyx_v_value); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 383, __pyx_L1_error)
    __pyx_t_3 = ((__pyx_t_7 < __pyx_cur_scope->__pyx_v_minlen) != 0);
    __pyx_t_1 = __pyx_t_3;
    __pyx_L8_bool_binop_done:;
    if (unlikely(__pyx_t_1)) {

      /* "validr/_validator_c.pyx":384
 *                 raise Invalid('dict length must <= %d' % maxlen)
 *             elif minlen > 0 and len(value) < minlen:
 *                 raise Invalid('dict length must >= %d' % minlen)             # <<<<<<<<<<<<<<
 *             if slim:
 *                 value = _slim_dict(value)
 */
      __Pyx_TraceLine(384,0,__PYX_ERR(0, 384, __pyx_L1_error))
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 384, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_cur_scope->__pyx_v_minlen); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 384, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_6 = PyUnicode_Format(__pyx_kp_u_dict_length_must_d_2, __pyx_t_8); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 384, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_4 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 384, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_Raise(__pyx_t_4, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __PYX_ERR(0, 384, __pyx_L1_error)

      /* "validr/_validator_c.pyx":383
 *             if len(value) > maxlen:
 *                 raise Invalid('dict length must <= %d' % maxlen)
 *             elif minlen > 0 and len(value) < minlen:             # <<<<<<<<<<<<<<
 *                 raise Invalid('dict length must >= %d' % minlen)
 *             if slim:
 */
    }

    /* "validr/_validator_c.pyx":385
 *             elif minlen > 0 and len(value) < minlen:
 *                 raise Invalid('dict length must >= %d' % minlen)
 *             if slim:             # <<<<<<<<<<<<<<
 *                 value = _slim_dict(value)
 *             return copy(value)
 */
    __Pyx_TraceLine(385,0,__PYX_ERR(0, 385, __pyx_L1_error))
    __pyx_t_1 = (__pyx_cur_scope->__pyx_v_slim != 0);
    if (__pyx_t_1) {

      /* "validr/_validator_c.pyx":386
 *                 raise Invalid('dict length must >= %d' % minlen)
 *             if slim:
 *                 value = _slim_dict(value)             # <<<<<<<<<<<<<<
 *             return copy(value)
 *         if is_dict(value):
 */
      __Pyx_TraceLine(386,0,__PYX_ERR(0, 386, __pyx_L1_error))
      if (!(likely(PyDict_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(0, 386, __pyx_L1_error)
      __pyx_t_4 = __pyx_f_6validr_12_validator_c__slim_dict(((PyObject*)__pyx_v_value), 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 386, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "validr/_validator_c.pyx":385
 *             elif minlen > 0 and len(value) < minlen:
 *                 raise Invalid('dict length must >= %d' % minlen)
 *             if slim:             # <<<<<<<<<<<<<<
 *                 value = _slim_dict(value)
 *             return copy(value)
 */
    }

    /* "validr/_validator_c.pyx":387
 *             if slim:
 *                 value = _slim_dict(value)
 *             return copy(value)             # <<<<<<<<<<<<<<
 *         if is_dict(value):
 *             getter = get_dict_value
 */
    __Pyx_TraceLine(387,0,__PYX_ERR(0, 387, __pyx_L1_error))
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_copy); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 387, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_v_value) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_value);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 387, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "validr/_validator_c.pyx":378
 * 
 *     def validate(value):
 *         if inners is None and not is_dynamic:             # <<<<<<<<<<<<<<
 *             if not is_dict(value):
 *                 raise Invalid('must be dict')
 */
  }

  /* "validr/_validator_c.pyx":388
 *                 value = _slim_dict(value)
 *             return copy(value)
 *         if is_dict(value):             # <<<<<<<<<<<<<<
 *             getter = get_dict_value
 *             if is_dynamic:
 */
  __Pyx_TraceLine(388,0,__PYX_ERR(0, 388, __pyx_L1_error))
  __pyx_t_1 = (__pyx_f_6validr_12_validator_c_is_dict(__pyx_v_value, 0) != 0);
  if (__pyx_t_1) {

    /* "validr/_validator_c.pyx":389
 *             return copy(value)
 *         if is_dict(value):
 *             getter = get_dict_value             # <<<<<<<<<<<<<<
 *             if is_dynamic:
 *                 if len(value) > maxlen:
 */
    __Pyx_TraceLine(389,0,__PYX_ERR(0, 389, __pyx_L1_error))
    __pyx_v_getter = __pyx_f_6validr_12_validator_c_get_dict_value;

    /* "validr/_validator_c.pyx":390
 *         if is_dict(value):
 *             getter = get_dict_value
 *             if is_dynamic:             # <<<<<<<<<<<<<<
 *                 if len(value) > maxlen:
 *                     raise Invalid('dict length must <= %d' % maxlen)
 */
    __Pyx_TraceLine(390,0,__PYX_ERR(0, 390, __pyx_L1_error))
    __pyx_t_1 = (__pyx_cur_scope->__pyx_v_is_dynamic != 0);
    if (__pyx_t_1) {

      /* "validr/_validator_c.pyx":391
 *             getter = get_dict_value
 *             if is_dynamic:
 *                 if len(value) > maxlen:             # <<<<<<<<<<<<<<
 *                     raise Invalid('dict length must <= %d' % maxlen)
 *                 elif minlen > 0 and len(value) < minlen:
 */
      __Pyx_TraceLine(391,0,__PYX_ERR(0, 391, __pyx_L1_error))
      __pyx_t_7 = PyObject_Length(__pyx_v_value); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 391, __pyx_L1_error)
      __pyx_t_1 = ((__pyx_t_7 > __pyx_cur_scope->__pyx_v_maxlen) != 0);
      if (unlikely(__pyx_t_1)) {

        /* "validr/_validator_c.pyx":392
 *             if is_dynamic:
 *                 if len(value) > maxlen:
 *                     raise Invalid('dict length must <= %d' % maxlen)             # <<<<<<<<<<<<<<
 *                 elif minlen > 0 and len(value) < minlen:
 *                     raise Invalid('dict length must >= %d' % minlen)
 */
        __Pyx_TraceLine(392,0,__PYX_ERR(0, 392, __pyx_L1_error))
        __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 392, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_cur_scope->__pyx_v_maxlen); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 392, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_8 = PyUnicode_Format(__pyx_kp_u_dict_length_must_d, __pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 392, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_6 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
          }
        }
        __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_8);
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 392, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(0, 392, __pyx_L1_error)

        /* "validr/_validator_c.pyx":391
 *             getter = get_dict_value
 *             if is_dynamic:
 *                 if len(value) > maxlen:             # <<<<<<<<<<<<<<
 *                     raise Invalid('dict length must <= %d' % maxlen)
 *                 elif minlen > 0 and len(value) < minlen:
 */
      }

      /* "validr/_validator_c.pyx":393
 *                 if len(value) > maxlen:
 *                     raise Invalid('dict length must <= %d' % maxlen)
 *                 elif minlen > 0 and len(value) < minlen:             # <<<<<<<<<<<<<<
 *                     raise Invalid('dict length must >= %d' % minlen)
 *         else:
 */
      __Pyx_TraceLine(393,0,__PYX_ERR(0, 393, __pyx_L1_error))
      __pyx_t_3 = ((__pyx_cur_scope->__pyx_v_minlen > 0) != 0);
      if (__pyx_t_3) {
      } else {
        __pyx_t_1 = __pyx_t_3;
        goto __pyx_L14_bool_binop_done;
      }
      __pyx_t_7 = PyObject_Length(__pyx_v_value); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 393, __pyx_L1_error)
      __pyx_t_3 = ((__pyx_t_7 < __pyx_cur_scope->__pyx_v_minlen) != 0);
      __pyx_t_1 = __pyx_t_3;
      __pyx_L14_bool_binop_done:;
      if (unlikely(__pyx_t_1)) {

        /* "validr/_validator_c.pyx":394
 *                     raise Invalid('dict length must <= %d' % maxlen)
 *                 elif minlen > 0 and len(value) < minlen:
 *                     raise Invalid('dict length must >= %d' % minlen)             # <<<<<<<<<<<<<<
 *         else:
 *             getter = get_object_value
 */
        __Pyx_TraceLine(394,0,__PYX_ERR(0, 394, __pyx_L1_error))
        __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 394, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_cur_scope->__pyx_v_minlen); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 394, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_6 = PyUnicode_Format(__pyx_kp_u_dict_length_must_d_2, __pyx_t_8); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 394, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_8 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_8)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_8);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
          }
        }
        __pyx_t_4 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6);
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 394, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(0, 394, __pyx_L1_error)

        /* "validr/_validator_c.pyx":393
 *                 if len(value) > maxlen:
 *                     raise Invalid('dict length must <= %d' % maxlen)
 *                 elif minlen > 0 and len(value) < minlen:             # <<<<<<<<<<<<<<
 *                     raise Invalid('dict length must >= %d' % minlen)
 *         else:
 */
      }

      /* "validr/_validator_c.pyx":390
 *         if is_dict(value):
 *             getter = get_dict_value
 *             if is_dynamic:             # <<<<<<<<<<<<<<
 *                 if len(value) > maxlen:
 *                     raise Invalid('dict length must <= %d' % maxlen)
 */
    }

    /* "validr/_validator_c.pyx":388
 *                 value = _slim_dict(value)
 *             return copy(value)
 *         if is_dict(value):             # <<<<<<<<<<<<<<
 *             getter = get_dict_value
 *             if is_dynamic:
 */
    goto __pyx_L11;
  }

  /* "validr/_validator_c.pyx":396
 *                     raise Invalid('dict length must >= %d' % minlen)
 *         else:
 *             getter = get_object_value             # <<<<<<<<<<<<<<
 *             if is_dynamic:
 *                 raise Invalid("dynamic dict not allowed non-dict value")
 */
  __Pyx_TraceLine(396,0,__PYX_ERR(0, 396, __pyx_L1_error))
  /*else*/ {
    __pyx_v_getter = __pyx_f_6validr_12_validator_c_get_object_value;

    /* "validr/_validator_c.pyx":397
 *         else:
 *             getter = get_object_value
 *             if is_dynamic:             # <<<<<<<<<<<<<<
 *                 raise Invalid("dynamic dict not allowed non-dict value")
 *         result = {}
 */
    __Pyx_TraceLine(397,0,__PYX_ERR(0, 397, __pyx_L1_error))
    __pyx_t_1 = (__pyx_cur_scope->__pyx_v_is_dynamic != 0);
    if (unlikely(__pyx_t_1)) {

      /* "validr/_validator_c.pyx":398
 *             getter = get_object_value
 *             if is_dynamic:
 *                 raise Invalid("dynamic dict not allowed non-dict value")             # <<<<<<<<<<<<<<
 *         result = {}
 *         cdef str k
 */
      __Pyx_TraceLine(398,0,__PYX_ERR(0, 398, __pyx_L1_error))
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 398, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_kp_u_dynamic_dict_not_allowed_non_dic) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_kp_u_dynamic_dict_not_allowed_non_dic);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 398, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_Raise(__pyx_t_4, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __PYX_ERR(0, 398, __pyx_L1_error)

      /* "validr/_validator_c.pyx":397
 *         else:
 *             getter = get_object_value
 *             if is_dynamic:             # <<<<<<<<<<<<<<
 *                 raise Invalid("dynamic dict not allowed non-dict value")
 *         result = {}
 */
    }
  }
  __pyx_L11:;

  /* "validr/_validator_c.pyx":399
 *             if is_dynamic:
 *                 raise Invalid("dynamic dict not allowed non-dict value")
 *         result = {}             # <<<<<<<<<<<<<<
 *         cdef str k
 *         if inners is not None:
 */
  __Pyx_TraceLine(399,0,__PYX_ERR(0, 399, __pyx_L1_error))
  __pyx_t_4 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_result = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "validr/_validator_c.pyx":401
 *         result = {}
 *         cdef str k
 *         if inners is not None:             # <<<<<<<<<<<<<<
 *             for k, inner in inners:
 *                 with mark_key(k):
 */
  __Pyx_TraceLine(401,0,__PYX_ERR(0, 401, __pyx_L1_error))
  if (unlikely(!__pyx_cur_scope->__pyx_v_inners)) { __Pyx_RaiseClosureNameError("inners"); __PYX_ERR(0, 401, __pyx_L1_error) }
  __pyx_t_1 = (__pyx_cur_scope->__pyx_v_inners != ((PyObject*)Py_None));
  __pyx_t_3 = (__pyx_t_1 != 0);
  if (__pyx_t_3) {

    /* "validr/_validator_c.pyx":402
 *         cdef str k
 *         if inners is not None:
 *             for k, inner in inners:             # <<<<<<<<<<<<<<
 *                 with mark_key(k):
 *                     result[k] = inner(getter(value, k))
 */
    __Pyx_TraceLine(402,0,__PYX_ERR(0, 402, __pyx_L1_error))
    if (unlikely(!__pyx_cur_scope->__pyx_v_inners)) { __Pyx_RaiseClosureNameError("inners"); __PYX_ERR(0, 402, __pyx_L1_error) }
    if (unlikely(__pyx_cur_scope->__pyx_v_inners == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 402, __pyx_L1_error)
    }
    __pyx_t_4 = __pyx_cur_scope->__pyx_v_inners; __Pyx_INCREF(__pyx_t_4); __pyx_t_7 = 0;
    for (;;) {
      if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_4)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_5 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_7); __Pyx_INCREF(__pyx_t_5); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 402, __pyx_L1_error)
      #else
      __pyx_t_5 = PySequence_ITEM(__pyx_t_4, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 402, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      #endif
      if ((likely(PyTuple_CheckExact(__pyx_t_5))) || (PyList_CheckExact(__pyx_t_5))) {
        PyObject* sequence = __pyx_t_5;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 402, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_6 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_8 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_6 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_8 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_8);
        #else
        __pyx_t_6 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 402, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_8 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 402, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        #endif
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_9 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 402, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_10 = Py_TYPE(__pyx_t_9)->tp_iternext;
        index = 0; __pyx_t_6 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_6)) goto __pyx_L20_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_6);
        index = 1; __pyx_t_8 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_8)) goto __pyx_L20_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_8);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_10(__pyx_t_9), 2) < 0) __PYX_ERR(0, 402, __pyx_L1_error)
        __pyx_t_10 = NULL;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L21_unpacking_done;
        __pyx_L20_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_10 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 402, __pyx_L1_error)
        __pyx_L21_unpacking_done:;
      }
      if (!(likely(PyUnicode_CheckExact(__pyx_t_6))||((__pyx_t_6) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_6)->tp_name), 0))) __PYX_ERR(0, 402, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_k, ((PyObject*)__pyx_t_6));
      __pyx_t_6 = 0;
      __Pyx_XDECREF_SET(__pyx_v_inner, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "validr/_validator_c.pyx":403
 *         if inners is not None:
 *             for k, inner in inners:
 *                 with mark_key(k):             # <<<<<<<<<<<<<<
 *                     result[k] = inner(getter(value, k))
 *         if is_dynamic:
 */
      __Pyx_TraceLine(403,0,__PYX_ERR(0, 403, __pyx_L1_error))
      /*with:*/ {
        __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_mark_key); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 403, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_6 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_8);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_8, function);
          }
        }
        __pyx_t_5 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_6, __pyx_v_k) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_v_k);
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 403, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_11 = __Pyx_PyObject_LookupSpecial(__pyx_t_5, __pyx_n_s_exit); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 403, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_6 = __Pyx_PyObject_LookupSpecial(__pyx_t_5, __pyx_n_s_enter); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 403, __pyx_L22_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_9 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_9)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_9);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
          }
        }
        __pyx_t_8 = (__pyx_t_9) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_9) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 403, __pyx_L22_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        /*try:*/ {
          {
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __Pyx_ExceptionSave(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14);
            __Pyx_XGOTREF(__pyx_t_12);
            __Pyx_XGOTREF(__pyx_t_13);
            __Pyx_XGOTREF(__pyx_t_14);
            /*try:*/ {

              /* "validr/_validator_c.pyx":404
 *             for k, inner in inners:
 *                 with mark_key(k):
 *                     result[k] = inner(getter(value, k))             # <<<<<<<<<<<<<<
 *         if is_dynamic:
 *             extra_keys = map(str, set(value) - set(result))
 */
              __Pyx_TraceLine(404,0,__PYX_ERR(0, 404, __pyx_L28_error))
              __pyx_t_8 = __pyx_v_getter(__pyx_v_value, __pyx_v_k, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 404, __pyx_L28_error)
              __Pyx_GOTREF(__pyx_t_8);
              __Pyx_INCREF(__pyx_v_inner);
              __pyx_t_6 = __pyx_v_inner; __pyx_t_9 = NULL;
              if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
                __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_6);
                if (likely(__pyx_t_9)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
                  __Pyx_INCREF(__pyx_t_9);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_6, function);
                }
              }
              __pyx_t_5 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_9, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_8);
              __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
              if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 404, __pyx_L28_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              if (unlikely(PyDict_SetItem(__pyx_v_result, __pyx_v_k, __pyx_t_5) < 0)) __PYX_ERR(0, 404, __pyx_L28_error)
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

              /* "validr/_validator_c.pyx":403
 *         if inners is not None:
 *             for k, inner in inners:
 *                 with mark_key(k):             # <<<<<<<<<<<<<<
 *                     result[k] = inner(getter(value, k))
 *         if is_dynamic:
 */
            }
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
            goto __pyx_L35_try_end;
            __pyx_L28_error:;
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
            /*except:*/ {
              __Pyx_AddTraceback("validr._validator_c.dict_validator.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
              if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_8) < 0) __PYX_ERR(0, 403, __pyx_L30_except_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_GOTREF(__pyx_t_6);
              __Pyx_GOTREF(__pyx_t_8);
              __pyx_t_9 = PyTuple_Pack(3, __pyx_t_5, __pyx_t_6, __pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 403, __pyx_L30_except_error)
              __Pyx_GOTREF(__pyx_t_9);
              __pyx_t_15 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_9, NULL);
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
              __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
              if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 403, __pyx_L30_except_error)
              __Pyx_GOTREF(__pyx_t_15);
              __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_15);
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
              if (__pyx_t_3 < 0) __PYX_ERR(0, 403, __pyx_L30_except_error)
              __pyx_t_1 = ((!(__pyx_t_3 != 0)) != 0);
              if (__pyx_t_1) {
                __Pyx_GIVEREF(__pyx_t_5);
                __Pyx_GIVEREF(__pyx_t_6);
                __Pyx_XGIVEREF(__pyx_t_8);
                __Pyx_ErrRestoreWithState(__pyx_t_5, __pyx_t_6, __pyx_t_8);
                __pyx_t_5 = 0; __pyx_t_6 = 0; __pyx_t_8 = 0; 
                __PYX_ERR(0, 403, __pyx_L30_except_error)
              }
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
              goto __pyx_L29_exception_handled;
            }
            __pyx_L30_except_error:;
            __Pyx_XGIVEREF(__pyx_t_12);
            __Pyx_XGIVEREF(__pyx_t_13);
            __Pyx_XGIVEREF(__pyx_t_14);
            __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
            goto __pyx_L1_error;
            __pyx_L29_exception_handled:;
            __Pyx_XGIVEREF(__pyx_t_12);
            __Pyx_XGIVEREF(__pyx_t_13);
            __Pyx_XGIVEREF(__pyx_t_14);
            __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
            __pyx_L35_try_end:;
          }
        }
        /*finally:*/ {
          /*normal exit:*/{
            if (__pyx_t_11) {
              __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_tuple__28, NULL);
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
              if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 403, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_14);
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
            }
            goto __pyx_L27;
          }
          __pyx_L27:;
        }
        goto __pyx_L39;
        __pyx_L22_error:;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        goto __pyx_L1_error;
        __pyx_L39:;
      }

      /* "validr/_validator_c.pyx":402
 *         cdef str k
 *         if inners is not None:
 *             for k, inner in inners:             # <<<<<<<<<<<<<<
 *                 with mark_key(k):
 *                     result[k] = inner(getter(value, k))
 */
      __Pyx_TraceLine(402,0,__PYX_ERR(0, 402, __pyx_L1_error))
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "validr/_validator_c.pyx":401
 *         result = {}
 *         cdef str k
 *         if inners is not None:             # <<<<<<<<<<<<<<
 *             for k, inner in inners:
 *                 with mark_key(k):
 */
  }

  /* "validr/_validator_c.pyx":405
 *                 with mark_key(k):
 *                     result[k] = inner(getter(value, k))
 *         if is_dynamic:             # <<<<<<<<<<<<<<
 *             extra_keys = map(str, set(value) - set(result))
 *             for k in extra_keys:
 */
  __Pyx_TraceLine(405,0,__PYX_ERR(0, 405, __pyx_L1_error))
  __pyx_t_1 = (__pyx_cur_scope->__pyx_v_is_dynamic != 0);
  if (__pyx_t_1) {

    /* "validr/_validator_c.pyx":406
 *                     result[k] = inner(getter(value, k))
 *         if is_dynamic:
 *             extra_keys = map(str, set(value) - set(result))             # <<<<<<<<<<<<<<
 *             for k in extra_keys:
 *                 if validate_extra_key:
 */
    __Pyx_TraceLine(406,0,__PYX_ERR(0, 406, __pyx_L1_error))
    __pyx_t_4 = PySet_New(__pyx_v_value); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 406, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_8 = PySet_New(__pyx_v_result); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 406, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_6 = PyNumber_Subtract(__pyx_t_4, __pyx_t_8); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 406, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 406, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_INCREF(((PyObject *)(&PyUnicode_Type)));
    __Pyx_GIVEREF(((PyObject *)(&PyUnicode_Type)));
    PyTuple_SET_ITEM(__pyx_t_8, 0, ((PyObject *)(&PyUnicode_Type)));
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_6);
    __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_map, __pyx_t_8, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 406, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_v_extra_keys = __pyx_t_6;
    __pyx_t_6 = 0;

    /* "validr/_validator_c.pyx":407
 *         if is_dynamic:
 *             extra_keys = map(str, set(value) - set(result))
 *             for k in extra_keys:             # <<<<<<<<<<<<<<
 *                 if validate_extra_key:
 *                     with mark_key('$self_key'):
 */
    __Pyx_TraceLine(407,0,__PYX_ERR(0, 407, __pyx_L1_error))
    if (likely(PyList_CheckExact(__pyx_v_extra_keys)) || PyTuple_CheckExact(__pyx_v_extra_keys)) {
      __pyx_t_6 = __pyx_v_extra_keys; __Pyx_INCREF(__pyx_t_6); __pyx_t_7 = 0;
      __pyx_t_16 = NULL;
    } else {
      __pyx_t_7 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_v_extra_keys); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 407, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_16 = Py_TYPE(__pyx_t_6)->tp_iternext; if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 407, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_16)) {
        if (likely(PyList_CheckExact(__pyx_t_6))) {
          if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_6)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_8 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_7); __Pyx_INCREF(__pyx_t_8); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 407, __pyx_L1_error)
          #else
          __pyx_t_8 = PySequence_ITEM(__pyx_t_6, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 407, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          #endif
        } else {
          if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_6)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_8 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_7); __Pyx_INCREF(__pyx_t_8); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 407, __pyx_L1_error)
          #else
          __pyx_t_8 = PySequence_ITEM(__pyx_t_6, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 407, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          #endif
        }
      } else {
        __pyx_t_8 = __pyx_t_16(__pyx_t_6);
        if (unlikely(!__pyx_t_8)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 407, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_8);
      }
      if (!(likely(PyUnicode_CheckExact(__pyx_t_8))||((__pyx_t_8) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_8)->tp_name), 0))) __PYX_ERR(0, 407, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_k, ((PyObject*)__pyx_t_8));
      __pyx_t_8 = 0;

      /* "validr/_validator_c.pyx":408
 *             extra_keys = map(str, set(value) - set(result))
 *             for k in extra_keys:
 *                 if validate_extra_key:             # <<<<<<<<<<<<<<
 *                     with mark_key('$self_key'):
 *                         k = str(validate_extra_key(k))
 */
      __Pyx_TraceLine(408,0,__PYX_ERR(0, 408, __pyx_L1_error))
      if (unlikely(!__pyx_cur_scope->__pyx_v_validate_extra_key)) { __Pyx_RaiseClosureNameError("validate_extra_key"); __PYX_ERR(0, 408, __pyx_L1_error) }
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_validate_extra_key); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 408, __pyx_L1_error)
      if (__pyx_t_1) {

        /* "validr/_validator_c.pyx":409
 *             for k in extra_keys:
 *                 if validate_extra_key:
 *                     with mark_key('$self_key'):             # <<<<<<<<<<<<<<
 *                         k = str(validate_extra_key(k))
 *                 with mark_key(k):
 */
        __Pyx_TraceLine(409,0,__PYX_ERR(0, 409, __pyx_L1_error))
        /*with:*/ {
          __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_mark_key); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 409, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_5 = NULL;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          __pyx_t_8 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_kp_u_self_key) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_kp_u_self_key);
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 409, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_11 = __Pyx_PyObject_LookupSpecial(__pyx_t_8, __pyx_n_s_exit); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 409, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_5 = __Pyx_PyObject_LookupSpecial(__pyx_t_8, __pyx_n_s_enter); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 409, __pyx_L44_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_9 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_9)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_9);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
            }
          }
          __pyx_t_4 = (__pyx_t_9) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_9) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 409, __pyx_L44_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          /*try:*/ {
            {
              __Pyx_PyThreadState_declare
              __Pyx_PyThreadState_assign
              __Pyx_ExceptionSave(&__pyx_t_14, &__pyx_t_13, &__pyx_t_12);
              __Pyx_XGOTREF(__pyx_t_14);
              __Pyx_XGOTREF(__pyx_t_13);
              __Pyx_XGOTREF(__pyx_t_12);
              /*try:*/ {

                /* "validr/_validator_c.pyx":410
 *                 if validate_extra_key:
 *                     with mark_key('$self_key'):
 *                         k = str(validate_extra_key(k))             # <<<<<<<<<<<<<<
 *                 with mark_key(k):
 *                     v = getter(value, k)
 */
                __Pyx_TraceLine(410,0,__PYX_ERR(0, 410, __pyx_L50_error))
                if (unlikely(!__pyx_cur_scope->__pyx_v_validate_extra_key)) { __Pyx_RaiseClosureNameError("validate_extra_key"); __PYX_ERR(0, 410, __pyx_L50_error) }
                __Pyx_INCREF(__pyx_cur_scope->__pyx_v_validate_extra_key);
                __pyx_t_4 = __pyx_cur_scope->__pyx_v_validate_extra_key; __pyx_t_5 = NULL;
                if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
                  __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
                  if (likely(__pyx_t_5)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                    __Pyx_INCREF(__pyx_t_5);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_4, function);
                  }
                }
                __pyx_t_8 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_k) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_k);
                __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 410, __pyx_L50_error)
                __Pyx_GOTREF(__pyx_t_8);
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                __pyx_t_4 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyUnicode_Type)), __pyx_t_8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 410, __pyx_L50_error)
                __Pyx_GOTREF(__pyx_t_4);
                __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
                __Pyx_DECREF_SET(__pyx_v_k, ((PyObject*)__pyx_t_4));
                __pyx_t_4 = 0;

                /* "validr/_validator_c.pyx":409
 *             for k in extra_keys:
 *                 if validate_extra_key:
 *                     with mark_key('$self_key'):             # <<<<<<<<<<<<<<
 *                         k = str(validate_extra_key(k))
 *                 with mark_key(k):
 */
              }
              __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
              __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
              __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
              goto __pyx_L57_try_end;
              __pyx_L50_error:;
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
              __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
              /*except:*/ {
                __Pyx_AddTraceback("validr._validator_c.dict_validator.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
                if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_8, &__pyx_t_5) < 0) __PYX_ERR(0, 409, __pyx_L52_except_error)
                __Pyx_GOTREF(__pyx_t_4);
                __Pyx_GOTREF(__pyx_t_8);
                __Pyx_GOTREF(__pyx_t_5);
                __pyx_t_9 = PyTuple_Pack(3, __pyx_t_4, __pyx_t_8, __pyx_t_5); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 409, __pyx_L52_except_error)
                __Pyx_GOTREF(__pyx_t_9);
                __pyx_t_15 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_9, NULL);
                __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
                __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
                if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 409, __pyx_L52_except_error)
                __Pyx_GOTREF(__pyx_t_15);
                __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_15);
                __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
                if (__pyx_t_1 < 0) __PYX_ERR(0, 409, __pyx_L52_except_error)
                __pyx_t_3 = ((!(__pyx_t_1 != 0)) != 0);
                if (__pyx_t_3) {
                  __Pyx_GIVEREF(__pyx_t_4);
                  __Pyx_GIVEREF(__pyx_t_8);
                  __Pyx_XGIVEREF(__pyx_t_5);
                  __Pyx_ErrRestoreWithState(__pyx_t_4, __pyx_t_8, __pyx_t_5);
                  __pyx_t_4 = 0; __pyx_t_8 = 0; __pyx_t_5 = 0; 
                  __PYX_ERR(0, 409, __pyx_L52_except_error)
                }
                __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
                __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                goto __pyx_L51_exception_handled;
              }
              __pyx_L52_except_error:;
              __Pyx_XGIVEREF(__pyx_t_14);
              __Pyx_XGIVEREF(__pyx_t_13);
              __Pyx_XGIVEREF(__pyx_t_12);
              __Pyx_ExceptionReset(__pyx_t_14, __pyx_t_13, __pyx_t_12);
              goto __pyx_L1_error;
              __pyx_L51_exception_handled:;
              __Pyx_XGIVEREF(__pyx_t_14);
              __Pyx_XGIVEREF(__pyx_t_13);
              __Pyx_XGIVEREF(__pyx_t_12);
              __Pyx_ExceptionReset(__pyx_t_14, __pyx_t_13, __pyx_t_12);
              __pyx_L57_try_end:;
            }
          }
          /*finally:*/ {
            /*normal exit:*/{
              if (__pyx_t_11) {
                __pyx_t_12 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_tuple__28, NULL);
                __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
                if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 409, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_12);
                __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
              }
              goto __pyx_L49;
            }
            __pyx_L49:;
          }
          goto __pyx_L61;
          __pyx_L44_error:;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          goto __pyx_L1_error;
          __pyx_L61:;
        }

        /* "validr/_validator_c.pyx":408
 *             extra_keys = map(str, set(value) - set(result))
 *             for k in extra_keys:
 *                 if validate_extra_key:             # <<<<<<<<<<<<<<
 *                     with mark_key('$self_key'):
 *                         k = str(validate_extra_key(k))
 */
      }

      /* "validr/_validator_c.pyx":411
 *                     with mark_key('$self_key'):
 *                         k = str(validate_extra_key(k))
 *                 with mark_key(k):             # <<<<<<<<<<<<<<
 *                     v = getter(value, k)
 *                     if validate_extra_value is not None:
 */
      __Pyx_TraceLine(411,0,__PYX_ERR(0, 411, __pyx_L1_error))
      /*with:*/ {
        __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_mark_key); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 411, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_8);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_8, function);
          }
        }
        __pyx_t_5 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_4, __pyx_v_k) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_v_k);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 411, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_11 = __Pyx_PyObject_LookupSpecial(__pyx_t_5, __pyx_n_s_exit); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 411, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_4 = __Pyx_PyObject_LookupSpecial(__pyx_t_5, __pyx_n_s_enter); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 411, __pyx_L62_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_9 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_9)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_9);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        __pyx_t_8 = (__pyx_t_9) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_9) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 411, __pyx_L62_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        /*try:*/ {
          {
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __Pyx_ExceptionSave(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14);
            __Pyx_XGOTREF(__pyx_t_12);
            __Pyx_XGOTREF(__pyx_t_13);
            __Pyx_XGOTREF(__pyx_t_14);
            /*try:*/ {

              /* "validr/_validator_c.pyx":412
 *                         k = str(validate_extra_key(k))
 *                 with mark_key(k):
 *                     v = getter(value, k)             # <<<<<<<<<<<<<<
 *                     if validate_extra_value is not None:
 *                         result[k] = validate_extra_value(v)
 */
              __Pyx_TraceLine(412,0,__PYX_ERR(0, 412, __pyx_L68_error))
              __pyx_t_5 = __pyx_v_getter(__pyx_v_value, __pyx_v_k, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 412, __pyx_L68_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_XDECREF_SET(__pyx_v_v, __pyx_t_5);
              __pyx_t_5 = 0;

              /* "validr/_validator_c.pyx":413
 *                 with mark_key(k):
 *                     v = getter(value, k)
 *                     if validate_extra_value is not None:             # <<<<<<<<<<<<<<
 *                         result[k] = validate_extra_value(v)
 *                     else:
 */
              __Pyx_TraceLine(413,0,__PYX_ERR(0, 413, __pyx_L68_error))
              if (unlikely(!__pyx_cur_scope->__pyx_v_validate_extra_value)) { __Pyx_RaiseClosureNameError("validate_extra_value"); __PYX_ERR(0, 413, __pyx_L68_error) }
              __pyx_t_3 = (__pyx_cur_scope->__pyx_v_validate_extra_value != Py_None);
              __pyx_t_1 = (__pyx_t_3 != 0);
              if (__pyx_t_1) {

                /* "validr/_validator_c.pyx":414
 *                     v = getter(value, k)
 *                     if validate_extra_value is not None:
 *                         result[k] = validate_extra_value(v)             # <<<<<<<<<<<<<<
 *                     else:
 *                         result[k] = copy(v)
 */
                __Pyx_TraceLine(414,0,__PYX_ERR(0, 414, __pyx_L68_error))
                if (unlikely(!__pyx_cur_scope->__pyx_v_validate_extra_value)) { __Pyx_RaiseClosureNameError("validate_extra_value"); __PYX_ERR(0, 414, __pyx_L68_error) }
                __Pyx_INCREF(__pyx_cur_scope->__pyx_v_validate_extra_value);
                __pyx_t_8 = __pyx_cur_scope->__pyx_v_validate_extra_value; __pyx_t_4 = NULL;
                if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
                  __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_8);
                  if (likely(__pyx_t_4)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
                    __Pyx_INCREF(__pyx_t_4);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_8, function);
                  }
                }
                __pyx_t_5 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_4, __pyx_v_v) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_v_v);
                __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 414, __pyx_L68_error)
                __Pyx_GOTREF(__pyx_t_5);
                __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
                if (unlikely(PyDict_SetItem(__pyx_v_result, __pyx_v_k, __pyx_t_5) < 0)) __PYX_ERR(0, 414, __pyx_L68_error)
                __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

                /* "validr/_validator_c.pyx":413
 *                 with mark_key(k):
 *                     v = getter(value, k)
 *                     if validate_extra_value is not None:             # <<<<<<<<<<<<<<
 *                         result[k] = validate_extra_value(v)
 *                     else:
 */
                goto __pyx_L76;
              }

              /* "validr/_validator_c.pyx":416
 *                         result[k] = validate_extra_value(v)
 *                     else:
 *                         result[k] = copy(v)             # <<<<<<<<<<<<<<
 *         if slim:
 *             result = _slim_dict(result)
 */
              __Pyx_TraceLine(416,0,__PYX_ERR(0, 416, __pyx_L68_error))
              /*else*/ {
                __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_copy); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 416, __pyx_L68_error)
                __Pyx_GOTREF(__pyx_t_8);
                __pyx_t_4 = NULL;
                if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
                  __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_8);
                  if (likely(__pyx_t_4)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
                    __Pyx_INCREF(__pyx_t_4);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_8, function);
                  }
                }
                __pyx_t_5 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_4, __pyx_v_v) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_v_v);
                __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 416, __pyx_L68_error)
                __Pyx_GOTREF(__pyx_t_5);
                __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
                if (unlikely(PyDict_SetItem(__pyx_v_result, __pyx_v_k, __pyx_t_5) < 0)) __PYX_ERR(0, 416, __pyx_L68_error)
                __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              }
              __pyx_L76:;

              /* "validr/_validator_c.pyx":411
 *                     with mark_key('$self_key'):
 *                         k = str(validate_extra_key(k))
 *                 with mark_key(k):             # <<<<<<<<<<<<<<
 *                     v = getter(value, k)
 *                     if validate_extra_value is not None:
 */
            }
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
            goto __pyx_L75_try_end;
            __pyx_L68_error:;
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
            /*except:*/ {
              __Pyx_AddTraceback("validr._validator_c.dict_validator.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
              if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_8, &__pyx_t_4) < 0) __PYX_ERR(0, 411, __pyx_L70_except_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_GOTREF(__pyx_t_8);
              __Pyx_GOTREF(__pyx_t_4);
              __pyx_t_9 = PyTuple_Pack(3, __pyx_t_5, __pyx_t_8, __pyx_t_4); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 411, __pyx_L70_except_error)
              __Pyx_GOTREF(__pyx_t_9);
              __pyx_t_15 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_9, NULL);
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
              __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
              if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 411, __pyx_L70_except_error)
              __Pyx_GOTREF(__pyx_t_15);
              __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_15);
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
              if (__pyx_t_1 < 0) __PYX_ERR(0, 411, __pyx_L70_except_error)
              __pyx_t_3 = ((!(__pyx_t_1 != 0)) != 0);
              if (__pyx_t_3) {
                __Pyx_GIVEREF(__pyx_t_5);
                __Pyx_GIVEREF(__pyx_t_8);
                __Pyx_XGIVEREF(__pyx_t_4);
                __Pyx_ErrRestoreWithState(__pyx_t_5, __pyx_t_8, __pyx_t_4);
                __pyx_t_5 = 0; __pyx_t_8 = 0; __pyx_t_4 = 0; 
                __PYX_ERR(0, 411, __pyx_L70_except_error)
              }
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              goto __pyx_L69_exception_handled;
            }
            __pyx_L70_except_error:;
            __Pyx_XGIVEREF(__pyx_t_12);
            __Pyx_XGIVEREF(__pyx_t_13);
            __Pyx_XGIVEREF(__pyx_t_14);
            __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
            goto __pyx_L1_error;
            __pyx_L69_exception_handled:;
            __Pyx_XGIVEREF(__pyx_t_12);
            __Pyx_XGIVEREF(__pyx_t_13);
            __Pyx_XGIVEREF(__pyx_t_14);
            __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
            __pyx_L75_try_end:;
          }
        }
        /*finally:*/ {
          /*normal exit:*/{
            if (__pyx_t_11) {
              __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_tuple__28, NULL);
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
              if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 411, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_14);
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
            }
            goto __pyx_L67;
          }
          __pyx_L67:;
        }
        goto __pyx_L80;
        __pyx_L62_error:;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        goto __pyx_L1_error;
        __pyx_L80:;
      }

      /* "validr/_validator_c.pyx":407
 *         if is_dynamic:
 *             extra_keys = map(str, set(value) - set(result))
 *             for k in extra_keys:             # <<<<<<<<<<<<<<
 *                 if validate_extra_key:
 *                     with mark_key('$self_key'):
 */
      __Pyx_TraceLine(407,0,__PYX_ERR(0, 407, __pyx_L1_error))
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "validr/_validator_c.pyx":405
 *                 with mark_key(k):
 *                     result[k] = inner(getter(value, k))
 *         if is_dynamic:             # <<<<<<<<<<<<<<
 *             extra_keys = map(str, set(value) - set(result))
 *             for k in extra_keys:
 */
  }

  /* "validr/_validator_c.pyx":417
 *                     else:
 *                         result[k] = copy(v)
 *         if slim:             # <<<<<<<<<<<<<<
 *             result = _slim_dict(result)
 *         return result
 */
  __Pyx_TraceLine(417,0,__PYX_ERR(0, 417, __pyx_L1_error))
  __pyx_t_3 = (__pyx_cur_scope->__pyx_v_slim != 0);
  if (__pyx_t_3) {

    /* "validr/_validator_c.pyx":418
 *                         result[k] = copy(v)
 *         if slim:
 *             result = _slim_dict(result)             # <<<<<<<<<<<<<<
 *         return result
 *     return validate
 */
    __Pyx_TraceLine(418,0,__PYX_ERR(0, 418, __pyx_L1_error))
    __pyx_t_6 = __pyx_f_6validr_12_validator_c__slim_dict(__pyx_v_result, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 418, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF_SET(__pyx_v_result, ((PyObject*)__pyx_t_6));
    __pyx_t_6 = 0;

    /* "validr/_validator_c.pyx":417
 *                     else:
 *                         result[k] = copy(v)
 *         if slim:             # <<<<<<<<<<<<<<
 *             result = _slim_dict(result)
 *         return result
 */
  }

  /* "validr/_validator_c.pyx":419
 *         if slim:
 *             result = _slim_dict(result)
 *         return result             # <<<<<<<<<<<<<<
 *     return validate
 * 
 */
  __Pyx_TraceLine(419,0,__PYX_ERR(0, 419, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "validr/_validator_c.pyx":377
 *     del compiler, items, key, value
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         if inners is None and not is_dynamic:
 *             if not is_dict(value):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("validr._validator_c.dict_validator.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_k);
  __Pyx_XDECREF(__pyx_v_inner);
  __Pyx_XDECREF(__pyx_v_extra_keys);
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":357
 * 
 * @validator(accept=(typing.Mapping, typing.Any), output=dict)
 * def dict_validator(compiler, items=None, key=None, value=None,             # <<<<<<<<<<<<<<
 *                    int minlen=0, int maxlen=1024, bint slim=False):
 *     if items is None:
 */

static PyObject *__pyx_pf_6validr_12_validator_c_24dict_validator(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_compiler, PyObject *__pyx_v_items, PyObject *__pyx_v_key, PyObject *__pyx_v_value, int __pyx_v_minlen, int __pyx_v_maxlen, int __pyx_v_slim) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_9_dict_validator *__pyx_cur_scope;
  PyObject *__pyx_v_k = NULL;
  PyObject *__pyx_v_v = NULL;
  PyObject *__pyx_v_validate = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_t_16;
  PyObject *__pyx_t_17 = NULL;
  __Pyx_TraceFrameInit(__pyx_codeobj__32)
  __Pyx_RefNannySetupContext("dict_validator", 0);
  __pyx_cur_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_9_dict_validator *)__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_9_dict_validator(__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_9_dict_validator, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_9_dict_validator *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 357, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __Pyx_TraceCall("dict_validator", __pyx_f[0], 357, 0, __PYX_ERR(0, 357, __pyx_L1_error));
  __pyx_cur_scope->__pyx_v_minlen = __pyx_v_minlen;
  __pyx_cur_scope->__pyx_v_maxlen = __pyx_v_maxlen;
  __pyx_cur_scope->__pyx_v_slim = __pyx_v_slim;
  __Pyx_INCREF(__pyx_v_compiler);
  __Pyx_INCREF(__pyx_v_items);
  __Pyx_INCREF(__pyx_v_key);
  __Pyx_INCREF(__pyx_v_value);

  /* "validr/_validator_c.pyx":359
 * def dict_validator(compiler, items=None, key=None, value=None,
 *                    int minlen=0, int maxlen=1024, bint slim=False):
 *     if items is None:             # <<<<<<<<<<<<<<
 *         inners = None
 *     else:
 */
  __Pyx_TraceLine(359,0,__PYX_ERR(0, 359, __pyx_L1_error))
  __pyx_t_1 = (__pyx_v_items == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "validr/_validator_c.pyx":360
 *                    int minlen=0, int maxlen=1024, bint slim=False):
 *     if items is None:
 *         inners = None             # <<<<<<<<<<<<<<
 *     else:
 *         inners = []
 */
    __Pyx_TraceLine(360,0,__PYX_ERR(0, 360, __pyx_L1_error))
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __pyx_cur_scope->__pyx_v_inners = ((PyObject*)Py_None);

    /* "validr/_validator_c.pyx":359
 * def dict_validator(compiler, items=None, key=None, value=None,
 *                    int minlen=0, int maxlen=1024, bint slim=False):
 *     if items is None:             # <<<<<<<<<<<<<<
 *         inners = None
 *     else:
 */
    goto __pyx_L3;
  }

  /* "validr/_validator_c.pyx":362
 *         inners = None
 *     else:
 *         inners = []             # <<<<<<<<<<<<<<
 *         for k, v in items.items():
 *             with mark_key(k):
 */
  __Pyx_TraceLine(362,0,__PYX_ERR(0, 362, __pyx_L1_error))
  /*else*/ {
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 362, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_cur_scope->__pyx_v_inners = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "validr/_validator_c.pyx":363
 *     else:
 *         inners = []
 *         for k, v in items.items():             # <<<<<<<<<<<<<<
 *             with mark_key(k):
 *                 inners.append((k, compiler.compile(v)))
 */
    __Pyx_TraceLine(363,0,__PYX_ERR(0, 363, __pyx_L1_error))
    __pyx_t_4 = 0;
    if (unlikely(__pyx_v_items == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
      __PYX_ERR(0, 363, __pyx_L1_error)
    }
    __pyx_t_7 = __Pyx_dict_iterator(__pyx_v_items, 0, __pyx_n_s_items, (&__pyx_t_5), (&__pyx_t_6)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 363, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_3);
    __pyx_t_3 = __pyx_t_7;
    __pyx_t_7 = 0;
    while (1) {
      __pyx_t_9 = __Pyx_dict_iter_next(__pyx_t_3, __pyx_t_5, &__pyx_t_4, &__pyx_t_7, &__pyx_t_8, NULL, __pyx_t_6);
      if (unlikely(__pyx_t_9 == 0)) break;
      if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(0, 363, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_XDECREF_SET(__pyx_v_k, __pyx_t_7);
      __pyx_t_7 = 0;
      __Pyx_XDECREF_SET(__pyx_v_v, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "validr/_validator_c.pyx":364
 *         inners = []
 *         for k, v in items.items():
 *             with mark_key(k):             # <<<<<<<<<<<<<<
 *                 inners.append((k, compiler.compile(v)))
 *     validate_extra_key = validate_extra_value = None
 */
      __Pyx_TraceLine(364,0,__PYX_ERR(0, 364, __pyx_L1_error))
      /*with:*/ {
        __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_mark_key); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 364, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_10 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_10)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_10);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        __pyx_t_8 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_10, __pyx_v_k) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_k);
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 364, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_11 = __Pyx_PyObject_LookupSpecial(__pyx_t_8, __pyx_n_s_exit); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 364, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_10 = __Pyx_PyObject_LookupSpecial(__pyx_t_8, __pyx_n_s_enter); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 364, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_12 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
          __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_10);
          if (likely(__pyx_t_12)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
            __Pyx_INCREF(__pyx_t_12);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_10, function);
          }
        }
        __pyx_t_7 = (__pyx_t_12) ? __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_t_12) : __Pyx_PyObject_CallNoArg(__pyx_t_10);
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 364, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        /*try:*/ {
          {
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __Pyx_ExceptionSave(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_13);
            __Pyx_XGOTREF(__pyx_t_14);
            __Pyx_XGOTREF(__pyx_t_15);
            /*try:*/ {

              /* "validr/_validator_c.pyx":365
 *         for k, v in items.items():
 *             with mark_key(k):
 *                 inners.append((k, compiler.compile(v)))             # <<<<<<<<<<<<<<
 *     validate_extra_key = validate_extra_value = None
 *     if key is not None:
 */
              __Pyx_TraceLine(365,0,__PYX_ERR(0, 365, __pyx_L12_error))
              __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_compiler, __pyx_n_s_compile); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 365, __pyx_L12_error)
              __Pyx_GOTREF(__pyx_t_7);
              __pyx_t_10 = NULL;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
                __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_7);
                if (likely(__pyx_t_10)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
                  __Pyx_INCREF(__pyx_t_10);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_7, function);
                }
              }
              __pyx_t_8 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_10, __pyx_v_v) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_v);
              __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
              if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 365, __pyx_L12_error)
              __Pyx_GOTREF(__pyx_t_8);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 365, __pyx_L12_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_INCREF(__pyx_v_k);
              __Pyx_GIVEREF(__pyx_v_k);
              PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_k);
              __Pyx_GIVEREF(__pyx_t_8);
              PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_8);
              __pyx_t_8 = 0;
              __pyx_t_16 = __Pyx_PyList_Append(__pyx_cur_scope->__pyx_v_inners, __pyx_t_7); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(0, 365, __pyx_L12_error)
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

              /* "validr/_validator_c.pyx":364
 *         inners = []
 *         for k, v in items.items():
 *             with mark_key(k):             # <<<<<<<<<<<<<<
 *                 inners.append((k, compiler.compile(v)))
 *     validate_extra_key = validate_extra_value = None
 */
            }
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
            __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
            goto __pyx_L19_try_end;
            __pyx_L12_error:;
            __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            /*except:*/ {
              __Pyx_AddTraceback("validr._validator_c.dict_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
              if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_8, &__pyx_t_10) < 0) __PYX_ERR(0, 364, __pyx_L14_except_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_GOTREF(__pyx_t_8);
              __Pyx_GOTREF(__pyx_t_10);
              __pyx_t_12 = PyTuple_Pack(3, __pyx_t_7, __pyx_t_8, __pyx_t_10); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 364, __pyx_L14_except_error)
              __Pyx_GOTREF(__pyx_t_12);
              __pyx_t_17 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_12, NULL);
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
              __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
              if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 364, __pyx_L14_except_error)
              __Pyx_GOTREF(__pyx_t_17);
              __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_17);
              __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
              if (__pyx_t_2 < 0) __PYX_ERR(0, 364, __pyx_L14_except_error)
              __pyx_t_1 = ((!(__pyx_t_2 != 0)) != 0);
              if (__pyx_t_1) {
                __Pyx_GIVEREF(__pyx_t_7);
                __Pyx_GIVEREF(__pyx_t_8);
                __Pyx_XGIVEREF(__pyx_t_10);
                __Pyx_ErrRestoreWithState(__pyx_t_7, __pyx_t_8, __pyx_t_10);
                __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_10 = 0; 
                __PYX_ERR(0, 364, __pyx_L14_except_error)
              }
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
              __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
              goto __pyx_L13_exception_handled;
            }
            __pyx_L14_except_error:;
            __Pyx_XGIVEREF(__pyx_t_13);
            __Pyx_XGIVEREF(__pyx_t_14);
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_ExceptionReset(__pyx_t_13, __pyx_t_14, __pyx_t_15);
            goto __pyx_L1_error;
            __pyx_L13_exception_handled:;
            __Pyx_XGIVEREF(__pyx_t_13);
            __Pyx_XGIVEREF(__pyx_t_14);
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_ExceptionReset(__pyx_t_13, __pyx_t_14, __pyx_t_15);
            __pyx_L19_try_end:;
          }
        }
        /*finally:*/ {
          /*normal exit:*/{
            if (__pyx_t_11) {
              __pyx_t_15 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_tuple__28, NULL);
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
              if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 364, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_15);
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            }
            goto __pyx_L11;
          }
          __pyx_L11:;
        }
        goto __pyx_L23;
        __pyx_L6_error:;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        goto __pyx_L1_error;
        __pyx_L23:;
      }
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_L3:;

  /* "validr/_validator_c.pyx":366
 *             with mark_key(k):
 *                 inners.append((k, compiler.compile(v)))
 *     validate_extra_key = validate_extra_value = None             # <<<<<<<<<<<<<<
 *     if key is not None:
 *         with mark_key('$self_key'):
 */
  __Pyx_TraceLine(366,0,__PYX_ERR(0, 366, __pyx_L1_error))
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __pyx_cur_scope->__pyx_v_validate_extra_key = Py_None;
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __pyx_cur_scope->__pyx_v_validate_extra_value = Py_None;

  /* "validr/_validator_c.pyx":367
 *                 inners.append((k, compiler.compile(v)))
 *     validate_extra_key = validate_extra_value = None
 *     if key is not None:             # <<<<<<<<<<<<<<
 *         with mark_key('$self_key'):
 *             validate_extra_key = compiler.compile(key)
 */
  __Pyx_TraceLine(367,0,__PYX_ERR(0, 367, __pyx_L1_error))
  __pyx_t_1 = (__pyx_v_key != Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "validr/_validator_c.pyx":368
 *     validate_extra_key = validate_extra_value = None
 *     if key is not None:
 *         with mark_key('$self_key'):             # <<<<<<<<<<<<<<
 *             validate_extra_key = compiler.compile(key)
 *     if value is not None:
 */
    __Pyx_TraceLine(368,0,__PYX_ERR(0, 368, __pyx_L1_error))
    /*with:*/ {
      __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_mark_key); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 368, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_8 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_10);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_10, function);
        }
      }
      __pyx_t_3 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_8, __pyx_kp_u_self_key) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_kp_u_self_key);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 368, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_11 = __Pyx_PyObject_LookupSpecial(__pyx_t_3, __pyx_n_s_exit); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 368, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_8 = __Pyx_PyObject_LookupSpecial(__pyx_t_3, __pyx_n_s_enter); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 368, __pyx_L25_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_7 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
        }
      }
      __pyx_t_10 = (__pyx_t_7) ? __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_7) : __Pyx_PyObject_CallNoArg(__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 368, __pyx_L25_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      /*try:*/ {
        {
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __Pyx_ExceptionSave(&__pyx_t_15, &__pyx_t_14, &__pyx_t_13);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_13);
          /*try:*/ {

            /* "validr/_validator_c.pyx":369
 *     if key is not None:
 *         with mark_key('$self_key'):
 *             validate_extra_key = compiler.compile(key)             # <<<<<<<<<<<<<<
 *     if value is not None:
 *         with mark_key('$self_value'):
 */
            __Pyx_TraceLine(369,0,__PYX_ERR(0, 369, __pyx_L29_error))
            __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_compiler, __pyx_n_s_compile); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 369, __pyx_L29_error)
            __Pyx_GOTREF(__pyx_t_10);
            __pyx_t_8 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
              __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_10);
              if (likely(__pyx_t_8)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
                __Pyx_INCREF(__pyx_t_8);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_10, function);
              }
            }
            __pyx_t_3 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_8, __pyx_v_key) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_v_key);
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 369, __pyx_L29_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_validate_extra_key);
            __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_validate_extra_key, __pyx_t_3);
            __Pyx_GIVEREF(__pyx_t_3);
            __pyx_t_3 = 0;

            /* "validr/_validator_c.pyx":368
 *     validate_extra_key = validate_extra_value = None
 *     if key is not None:
 *         with mark_key('$self_key'):             # <<<<<<<<<<<<<<
 *             validate_extra_key = compiler.compile(key)
 *     if value is not None:
 */
          }
          __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
          __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          goto __pyx_L34_try_end;
          __pyx_L29_error:;
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          /*except:*/ {
            __Pyx_AddTraceback("validr._validator_c.dict_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
            if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_10, &__pyx_t_8) < 0) __PYX_ERR(0, 368, __pyx_L31_except_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_GOTREF(__pyx_t_10);
            __Pyx_GOTREF(__pyx_t_8);
            __pyx_t_7 = PyTuple_Pack(3, __pyx_t_3, __pyx_t_10, __pyx_t_8); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 368, __pyx_L31_except_error)
            __Pyx_GOTREF(__pyx_t_7);
            __pyx_t_17 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_7, NULL);
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 368, __pyx_L31_except_error)
            __Pyx_GOTREF(__pyx_t_17);
            __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_17);
            __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
            if (__pyx_t_2 < 0) __PYX_ERR(0, 368, __pyx_L31_except_error)
            __pyx_t_1 = ((!(__pyx_t_2 != 0)) != 0);
            if (__pyx_t_1) {
              __Pyx_GIVEREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_10);
              __Pyx_XGIVEREF(__pyx_t_8);
              __Pyx_ErrRestoreWithState(__pyx_t_3, __pyx_t_10, __pyx_t_8);
              __pyx_t_3 = 0; __pyx_t_10 = 0; __pyx_t_8 = 0; 
              __PYX_ERR(0, 368, __pyx_L31_except_error)
            }
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            goto __pyx_L30_exception_handled;
          }
          __pyx_L31_except_error:;
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_XGIVEREF(__pyx_t_13);
          __Pyx_ExceptionReset(__pyx_t_15, __pyx_t_14, __pyx_t_13);
          goto __pyx_L1_error;
          __pyx_L30_exception_handled:;
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_XGIVEREF(__pyx_t_13);
          __Pyx_ExceptionReset(__pyx_t_15, __pyx_t_14, __pyx_t_13);
          __pyx_L34_try_end:;
        }
      }
      /*finally:*/ {
        /*normal exit:*/{
          if (__pyx_t_11) {
            __pyx_t_13 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_tuple__28, NULL);
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 368, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_13);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          goto __pyx_L28;
        }
        __pyx_L28:;
      }
      goto __pyx_L38;
      __pyx_L25_error:;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      goto __pyx_L1_error;
      __pyx_L38:;
    }

    /* "validr/_validator_c.pyx":367
 *                 inners.append((k, compiler.compile(v)))
 *     validate_extra_key = validate_extra_value = None
 *     if key is not None:             # <<<<<<<<<<<<<<
 *         with mark_key('$self_key'):
 *             validate_extra_key = compiler.compile(key)
 */
  }

  /* "validr/_validator_c.pyx":370
 *         with mark_key('$self_key'):
 *             validate_extra_key = compiler.compile(key)
 *     if value is not None:             # <<<<<<<<<<<<<<
 *         with mark_key('$self_value'):
 *             validate_extra_value = compiler.compile(value)
 */
  __Pyx_TraceLine(370,0,__PYX_ERR(0, 370, __pyx_L1_error))
  __pyx_t_1 = (__pyx_v_value != Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "validr/_validator_c.pyx":371
 *             validate_extra_key = compiler.compile(key)
 *     if value is not None:
 *         with mark_key('$self_value'):             # <<<<<<<<<<<<<<
 *             validate_extra_value = compiler.compile(value)
 *     cdef bint is_dynamic
 */
    __Pyx_TraceLine(371,0,__PYX_ERR(0, 371, __pyx_L1_error))
    /*with:*/ {
      __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_mark_key); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 371, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_3 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_10);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_10, function);
        }
      }
      __pyx_t_8 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_3, __pyx_kp_u_self_value) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_kp_u_self_value);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 371, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_11 = __Pyx_PyObject_LookupSpecial(__pyx_t_8, __pyx_n_s_exit); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 371, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_3 = __Pyx_PyObject_LookupSpecial(__pyx_t_8, __pyx_n_s_enter); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 371, __pyx_L40_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_7 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      __pyx_t_10 = (__pyx_t_7) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 371, __pyx_L40_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      /*try:*/ {
        {
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __Pyx_ExceptionSave(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_13);
          __Pyx_XGOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_15);
          /*try:*/ {

            /* "validr/_validator_c.pyx":372
 *     if value is not None:
 *         with mark_key('$self_value'):
 *             validate_extra_value = compiler.compile(value)             # <<<<<<<<<<<<<<
 *     cdef bint is_dynamic
 *     is_dynamic = bool(validate_extra_key or validate_extra_value)
 */
            __Pyx_TraceLine(372,0,__PYX_ERR(0, 372, __pyx_L44_error))
            __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_compiler, __pyx_n_s_compile); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 372, __pyx_L44_error)
            __Pyx_GOTREF(__pyx_t_10);
            __pyx_t_3 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
              __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_10);
              if (likely(__pyx_t_3)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
                __Pyx_INCREF(__pyx_t_3);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_10, function);
              }
            }
            __pyx_t_8 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_3, __pyx_v_value) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_v_value);
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 372, __pyx_L44_error)
            __Pyx_GOTREF(__pyx_t_8);
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_validate_extra_value);
            __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_validate_extra_value, __pyx_t_8);
            __Pyx_GIVEREF(__pyx_t_8);
            __pyx_t_8 = 0;

            /* "validr/_validator_c.pyx":371
 *             validate_extra_key = compiler.compile(key)
 *     if value is not None:
 *         with mark_key('$self_value'):             # <<<<<<<<<<<<<<
 *             validate_extra_value = compiler.compile(value)
 *     cdef bint is_dynamic
 */
          }
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
          __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
          goto __pyx_L49_try_end;
          __pyx_L44_error:;
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          /*except:*/ {
            __Pyx_AddTraceback("validr._validator_c.dict_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
            if (__Pyx_GetException(&__pyx_t_8, &__pyx_t_10, &__pyx_t_3) < 0) __PYX_ERR(0, 371, __pyx_L46_except_error)
            __Pyx_GOTREF(__pyx_t_8);
            __Pyx_GOTREF(__pyx_t_10);
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_7 = PyTuple_Pack(3, __pyx_t_8, __pyx_t_10, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 371, __pyx_L46_except_error)
            __Pyx_GOTREF(__pyx_t_7);
            __pyx_t_17 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_7, NULL);
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 371, __pyx_L46_except_error)
            __Pyx_GOTREF(__pyx_t_17);
            __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_17);
            __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
            if (__pyx_t_2 < 0) __PYX_ERR(0, 371, __pyx_L46_except_error)
            __pyx_t_1 = ((!(__pyx_t_2 != 0)) != 0);
            if (__pyx_t_1) {
              __Pyx_GIVEREF(__pyx_t_8);
              __Pyx_GIVEREF(__pyx_t_10);
              __Pyx_XGIVEREF(__pyx_t_3);
              __Pyx_ErrRestoreWithState(__pyx_t_8, __pyx_t_10, __pyx_t_3);
              __pyx_t_8 = 0; __pyx_t_10 = 0; __pyx_t_3 = 0; 
              __PYX_ERR(0, 371, __pyx_L46_except_error)
            }
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            goto __pyx_L45_exception_handled;
          }
          __pyx_L46_except_error:;
          __Pyx_XGIVEREF(__pyx_t_13);
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_ExceptionReset(__pyx_t_13, __pyx_t_14, __pyx_t_15);
          goto __pyx_L1_error;
          __pyx_L45_exception_handled:;
          __Pyx_XGIVEREF(__pyx_t_13);
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_ExceptionReset(__pyx_t_13, __pyx_t_14, __pyx_t_15);
          __pyx_L49_try_end:;
        }
      }
      /*finally:*/ {
        /*normal exit:*/{
          if (__pyx_t_11) {
            __pyx_t_15 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_tuple__28, NULL);
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 371, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_15);
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          }
          goto __pyx_L43;
        }
        __pyx_L43:;
      }
      goto __pyx_L53;
      __pyx_L40_error:;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      goto __pyx_L1_error;
      __pyx_L53:;
    }

    /* "validr/_validator_c.pyx":370
 *         with mark_key('$self_key'):
 *             validate_extra_key = compiler.compile(key)
 *     if value is not None:             # <<<<<<<<<<<<<<
 *         with mark_key('$self_value'):
 *             validate_extra_value = compiler.compile(value)
 */
  }

  /* "validr/_validator_c.pyx":374
 *             validate_extra_value = compiler.compile(value)
 *     cdef bint is_dynamic
 *     is_dynamic = bool(validate_extra_key or validate_extra_value)             # <<<<<<<<<<<<<<
 *     del compiler, items, key, value
 * 
 */
  __Pyx_TraceLine(374,0,__PYX_ERR(0, 374, __pyx_L1_error))
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_validate_extra_key); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 374, __pyx_L1_error)
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L54_bool_binop_done;
  }
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_validate_extra_value); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 374, __pyx_L1_error)
  __pyx_t_1 = __pyx_t_2;
  __pyx_L54_bool_binop_done:;
  __pyx_cur_scope->__pyx_v_is_dynamic = (!(!__pyx_t_1));

  /* "validr/_validator_c.pyx":375
 *     cdef bint is_dynamic
 *     is_dynamic = bool(validate_extra_key or validate_extra_value)
 *     del compiler, items, key, value             # <<<<<<<<<<<<<<
 * 
 *     def validate(value):
 */
  __Pyx_TraceLine(375,0,__PYX_ERR(0, 375, __pyx_L1_error))
  __Pyx_DECREF(__pyx_v_compiler);
  __pyx_v_compiler = NULL;
  __Pyx_DECREF(__pyx_v_items);
  __pyx_v_items = NULL;
  __Pyx_DECREF(__pyx_v_key);
  __pyx_v_key = NULL;
  __Pyx_DECREF(__pyx_v_value);
  __pyx_v_value = NULL;

  /* "validr/_validator_c.pyx":377
 *     del compiler, items, key, value
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         if inners is None and not is_dynamic:
 *             if not is_dict(value):
 */
  __Pyx_TraceLine(377,0,__PYX_ERR(0, 377, __pyx_L1_error))
  __pyx_t_3 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_14dict_validator_1validate, 0, __pyx_n_s_dict_validator_locals_validate, ((PyObject*)__pyx_cur_scope), __pyx_n_s_validr__validator_c, __pyx_d, ((PyObject *)__pyx_codeobj__34)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 377, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_validate = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "validr/_validator_c.pyx":420
 *             result = _slim_dict(result)
 *         return result
 *     return validate             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(420,0,__PYX_ERR(0, 420, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_validate);
  __pyx_r = __pyx_v_validate;
  goto __pyx_L0;

  /* "validr/_validator_c.pyx":357
 * 
 * @validator(accept=(typing.Mapping, typing.Any), output=dict)
 * def dict_validator(compiler, items=None, key=None, value=None,             # <<<<<<<<<<<<<<
 *                    int minlen=0, int maxlen=1024, bint slim=False):
 *     if items is None:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("validr._validator_c.dict_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_k);
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_XDECREF(__pyx_v_validate);
  __Pyx_XDECREF(__pyx_v_compiler);
  __Pyx_XDECREF(__pyx_v_items);
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":424
 * 
 * @validator(accept=(typing.Mapping, typing.Any), output=object)
 * def model_validator(compiler, items=None):             # <<<<<<<<<<<<<<
 *     if items is None:
 *         raise SchemaError('model class not provided')
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_27model_validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6validr_12_validator_c_27model_validator = {"model_validator", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6validr_12_validator_c_27model_validator, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6validr_12_validator_c_27model_validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_compiler = 0;
  PyObject *__pyx_v_items = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("model_validator (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_compiler,&__pyx_n_s_items,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_compiler)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_items);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "model_validator") < 0)) __PYX_ERR(0, 424, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_compiler = values[0];
    __pyx_v_items = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("model_validator", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 424, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("validr._validator_c.model_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6validr_12_validator_c_26model_validator(__pyx_self, __pyx_v_compiler, __pyx_v_items);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":428
 *         raise SchemaError('model class not provided')
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         return items(value)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_15model_validator_1validate(PyObject *__pyx_self, PyObject *__pyx_v_value); /*proto*/
static PyMethodDef __pyx_mdef_6validr_12_validator_c_15model_validator_1validate = {"validate", (PyCFunction)__pyx_pw_6validr_12_validator_c_15model_validator_1validate, METH_O, 0};
static PyObject *__pyx_pw_6validr_12_validator_c_15model_validator_1validate(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("validate (wrapper)", 0);
  __pyx_r = __pyx_pf_6validr_12_validator_c_15model_validator_validate(__pyx_self, ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6validr_12_validator_c_15model_validator_validate(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_10_model_validator *__pyx_cur_scope;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_10_model_validator *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("validate", 0);
  __pyx_outer_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_10_model_validator *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_TraceCall("validate", __pyx_f[0], 428, 0, __PYX_ERR(0, 428, __pyx_L1_error));

  /* "validr/_validator_c.pyx":429
 * 
 *     def validate(value):
 *         return items(value)             # <<<<<<<<<<<<<<
 * 
 *     return validate
 */
  __Pyx_TraceLine(429,0,__PYX_ERR(0, 429, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_cur_scope->__pyx_v_items)) { __Pyx_RaiseClosureNameError("items"); __PYX_ERR(0, 429, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_items);
  __pyx_t_2 = __pyx_cur_scope->__pyx_v_items; __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_value) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_value);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 429, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "validr/_validator_c.pyx":428
 *         raise SchemaError('model class not provided')
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         return items(value)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("validr._validator_c.model_validator.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":424
 * 
 * @validator(accept=(typing.Mapping, typing.Any), output=object)
 * def model_validator(compiler, items=None):             # <<<<<<<<<<<<<<
 *     if items is None:
 *         raise SchemaError('model class not provided')
 */

static PyObject *__pyx_pf_6validr_12_validator_c_26model_validator(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_compiler, PyObject *__pyx_v_items) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_10_model_validator *__pyx_cur_scope;
  PyObject *__pyx_v_validate = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_TraceFrameInit(__pyx_codeobj__35)
  __Pyx_RefNannySetupContext("model_validator", 0);
  __pyx_cur_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_10_model_validator *)__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_10_model_validator(__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_10_model_validator, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_10_model_validator *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 424, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __Pyx_TraceCall("model_validator", __pyx_f[0], 424, 0, __PYX_ERR(0, 424, __pyx_L1_error));
  __pyx_cur_scope->__pyx_v_items = __pyx_v_items;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_items);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_items);

  /* "validr/_validator_c.pyx":425
 * @validator(accept=(typing.Mapping, typing.Any), output=object)
 * def model_validator(compiler, items=None):
 *     if items is None:             # <<<<<<<<<<<<<<
 *         raise SchemaError('model class not provided')
 * 
 */
  __Pyx_TraceLine(425,0,__PYX_ERR(0, 425, __pyx_L1_error))
  __pyx_t_1 = (__pyx_cur_scope->__pyx_v_items == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (unlikely(__pyx_t_2)) {

    /* "validr/_validator_c.pyx":426
 * def model_validator(compiler, items=None):
 *     if items is None:
 *         raise SchemaError('model class not provided')             # <<<<<<<<<<<<<<
 * 
 *     def validate(value):
 */
    __Pyx_TraceLine(426,0,__PYX_ERR(0, 426, __pyx_L1_error))
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_SchemaError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 426, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_kp_u_model_class_not_provided) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_kp_u_model_class_not_provided);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 426, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 426, __pyx_L1_error)

    /* "validr/_validator_c.pyx":425
 * @validator(accept=(typing.Mapping, typing.Any), output=object)
 * def model_validator(compiler, items=None):
 *     if items is None:             # <<<<<<<<<<<<<<
 *         raise SchemaError('model class not provided')
 * 
 */
  }

  /* "validr/_validator_c.pyx":428
 *         raise SchemaError('model class not provided')
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         return items(value)
 * 
 */
  __Pyx_TraceLine(428,0,__PYX_ERR(0, 428, __pyx_L1_error))
  __pyx_t_3 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_15model_validator_1validate, 0, __pyx_n_s_model_validator_locals_validate, ((PyObject*)__pyx_cur_scope), __pyx_n_s_validr__validator_c, __pyx_d, ((PyObject *)__pyx_codeobj__37)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 428, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_validate = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "validr/_validator_c.pyx":431
 *         return items(value)
 * 
 *     return validate             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(431,0,__PYX_ERR(0, 431, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_validate);
  __pyx_r = __pyx_v_validate;
  goto __pyx_L0;

  /* "validr/_validator_c.pyx":424
 * 
 * @validator(accept=(typing.Mapping, typing.Any), output=object)
 * def model_validator(compiler, items=None):             # <<<<<<<<<<<<<<
 *     if items is None:
 *         raise SchemaError('model class not provided')
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("validr._validator_c.model_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_validate);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":434
 * 
 * 
 * cpdef _dump_enum_value(value):             # <<<<<<<<<<<<<<
 *     if value is None:
 *         return 'null'
 */

static PyObject *__pyx_pw_6validr_12_validator_c_29_dump_enum_value(PyObject *__pyx_self, PyObject *__pyx_v_value); /*proto*/
static PyObject *__pyx_f_6validr_12_validator_c__dump_enum_value(PyObject *__pyx_v_value, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("_dump_enum_value", 0);
  __Pyx_TraceCall("_dump_enum_value", __pyx_f[0], 434, 0, __PYX_ERR(0, 434, __pyx_L1_error));

  /* "validr/_validator_c.pyx":435
 * 
 * cpdef _dump_enum_value(value):
 *     if value is None:             # <<<<<<<<<<<<<<
 *         return 'null'
 *     elif value is False:
 */
  __Pyx_TraceLine(435,0,__PYX_ERR(0, 435, __pyx_L1_error))
  __pyx_t_1 = (__pyx_v_value == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "validr/_validator_c.pyx":436
 * cpdef _dump_enum_value(value):
 *     if value is None:
 *         return 'null'             # <<<<<<<<<<<<<<
 *     elif value is False:
 *         return 'false'
 */
    __Pyx_TraceLine(436,0,__PYX_ERR(0, 436, __pyx_L1_error))
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_n_u_null);
    __pyx_r = __pyx_n_u_null;
    goto __pyx_L0;

    /* "validr/_validator_c.pyx":435
 * 
 * cpdef _dump_enum_value(value):
 *     if value is None:             # <<<<<<<<<<<<<<
 *         return 'null'
 *     elif value is False:
 */
  }

  /* "validr/_validator_c.pyx":437
 *     if value is None:
 *         return 'null'
 *     elif value is False:             # <<<<<<<<<<<<<<
 *         return 'false'
 *     elif value is True:
 */
  __Pyx_TraceLine(437,0,__PYX_ERR(0, 437, __pyx_L1_error))
  __pyx_t_2 = (__pyx_v_value == Py_False);
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "validr/_validator_c.pyx":438
 *         return 'null'
 *     elif value is False:
 *         return 'false'             # <<<<<<<<<<<<<<
 *     elif value is True:
 *         return 'true'
 */
    __Pyx_TraceLine(438,0,__PYX_ERR(0, 438, __pyx_L1_error))
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_n_u_false);
    __pyx_r = __pyx_n_u_false;
    goto __pyx_L0;

    /* "validr/_validator_c.pyx":437
 *     if value is None:
 *         return 'null'
 *     elif value is False:             # <<<<<<<<<<<<<<
 *         return 'false'
 *     elif value is True:
 */
  }

  /* "validr/_validator_c.pyx":439
 *     elif value is False:
 *         return 'false'
 *     elif value is True:             # <<<<<<<<<<<<<<
 *         return 'true'
 *     elif isinstance(value, str):
 */
  __Pyx_TraceLine(439,0,__PYX_ERR(0, 439, __pyx_L1_error))
  __pyx_t_1 = (__pyx_v_value == Py_True);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "validr/_validator_c.pyx":440
 *         return 'false'
 *     elif value is True:
 *         return 'true'             # <<<<<<<<<<<<<<
 *     elif isinstance(value, str):
 *         return repr(value)  # single quotes by default
 */
    __Pyx_TraceLine(440,0,__PYX_ERR(0, 440, __pyx_L1_error))
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_n_u_true);
    __pyx_r = __pyx_n_u_true;
    goto __pyx_L0;

    /* "validr/_validator_c.pyx":439
 *     elif value is False:
 *         return 'false'
 *     elif value is True:             # <<<<<<<<<<<<<<
 *         return 'true'
 *     elif isinstance(value, str):
 */
  }

  /* "validr/_validator_c.pyx":441
 *     elif value is True:
 *         return 'true'
 *     elif isinstance(value, str):             # <<<<<<<<<<<<<<
 *         return repr(value)  # single quotes by default
 *     else:
 */
  __Pyx_TraceLine(441,0,__PYX_ERR(0, 441, __pyx_L1_error))
  __pyx_t_2 = PyUnicode_Check(__pyx_v_value); 
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "validr/_validator_c.pyx":442
 *         return 'true'
 *     elif isinstance(value, str):
 *         return repr(value)  # single quotes by default             # <<<<<<<<<<<<<<
 *     else:
 *         return str(value)  # number
 */
    __Pyx_TraceLine(442,0,__PYX_ERR(0, 442, __pyx_L1_error))
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = PyObject_Repr(__pyx_v_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 442, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "validr/_validator_c.pyx":441
 *     elif value is True:
 *         return 'true'
 *     elif isinstance(value, str):             # <<<<<<<<<<<<<<
 *         return repr(value)  # single quotes by default
 *     else:
 */
  }

  /* "validr/_validator_c.pyx":444
 *         return repr(value)  # single quotes by default
 *     else:
 *         return str(value)  # number             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(444,0,__PYX_ERR(0, 444, __pyx_L1_error))
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyUnicode_Type)), __pyx_v_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 444, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "validr/_validator_c.pyx":434
 * 
 * 
 * cpdef _dump_enum_value(value):             # <<<<<<<<<<<<<<
 *     if value is None:
 *         return 'null'
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("validr._validator_c._dump_enum_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_29_dump_enum_value(PyObject *__pyx_self, PyObject *__pyx_v_value); /*proto*/
static PyObject *__pyx_pw_6validr_12_validator_c_29_dump_enum_value(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_dump_enum_value (wrapper)", 0);
  __pyx_r = __pyx_pf_6validr_12_validator_c_28_dump_enum_value(__pyx_self, ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6validr_12_validator_c_28_dump_enum_value(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_dump_enum_value", 0);
  __Pyx_TraceCall("_dump_enum_value (wrapper)", __pyx_f[0], 434, 0, __PYX_ERR(0, 434, __pyx_L1_error));
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6validr_12_validator_c__dump_enum_value(__pyx_v_value, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("validr._validator_c._dump_enum_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":448
 * 
 * @validator(output=object)
 * def enum_validator(compiler, items):             # <<<<<<<<<<<<<<
 *     if not items:
 *         raise SchemaError("enum items not provided")
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_31enum_validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6validr_12_validator_c_31enum_validator = {"enum_validator", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6validr_12_validator_c_31enum_validator, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6validr_12_validator_c_31enum_validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_compiler = 0;
  PyObject *__pyx_v_items = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enum_validator (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_compiler,&__pyx_n_s_items,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_compiler)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_items)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enum_validator", 1, 2, 2, 1); __PYX_ERR(0, 448, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enum_validator") < 0)) __PYX_ERR(0, 448, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_compiler = values[0];
    __pyx_v_items = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enum_validator", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 448, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("validr._validator_c.enum_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6validr_12_validator_c_30enum_validator(__pyx_self, __pyx_v_compiler, __pyx_v_items);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":454
 *     items = frozenset(items)
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         if value in items:
 *             return value
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_14enum_validator_1validate(PyObject *__pyx_self, PyObject *__pyx_v_value); /*proto*/
static PyMethodDef __pyx_mdef_6validr_12_validator_c_14enum_validator_1validate = {"validate", (PyCFunction)__pyx_pw_6validr_12_validator_c_14enum_validator_1validate, METH_O, 0};
static PyObject *__pyx_pw_6validr_12_validator_c_14enum_validator_1validate(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("validate (wrapper)", 0);
  __pyx_r = __pyx_pf_6validr_12_validator_c_14enum_validator_validate(__pyx_self, ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6validr_12_validator_c_14enum_validator_validate(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_11_enum_validator *__pyx_cur_scope;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_11_enum_validator *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_RefNannySetupContext("validate", 0);
  __pyx_outer_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_11_enum_validator *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_TraceCall("validate", __pyx_f[0], 454, 0, __PYX_ERR(0, 454, __pyx_L1_error));

  /* "validr/_validator_c.pyx":455
 * 
 *     def validate(value):
 *         if value in items:             # <<<<<<<<<<<<<<
 *             return value
 *         raise Invalid("expect one of {}".format(expects))
 */
  __Pyx_TraceLine(455,0,__PYX_ERR(0, 455, __pyx_L1_error))
  if (unlikely(!__pyx_cur_scope->__pyx_v_items)) { __Pyx_RaiseClosureNameError("items"); __PYX_ERR(0, 455, __pyx_L1_error) }
  __pyx_t_1 = (__Pyx_PySequence_ContainsTF(__pyx_v_value, __pyx_cur_scope->__pyx_v_items, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 455, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "validr/_validator_c.pyx":456
 *     def validate(value):
 *         if value in items:
 *             return value             # <<<<<<<<<<<<<<
 *         raise Invalid("expect one of {}".format(expects))
 * 
 */
    __Pyx_TraceLine(456,0,__PYX_ERR(0, 456, __pyx_L1_error))
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_value);
    __pyx_r = __pyx_v_value;
    goto __pyx_L0;

    /* "validr/_validator_c.pyx":455
 * 
 *     def validate(value):
 *         if value in items:             # <<<<<<<<<<<<<<
 *             return value
 *         raise Invalid("expect one of {}".format(expects))
 */
  }

  /* "validr/_validator_c.pyx":457
 *         if value in items:
 *             return value
 *         raise Invalid("expect one of {}".format(expects))             # <<<<<<<<<<<<<<
 * 
 *     return validate
 */
  __Pyx_TraceLine(457,0,__PYX_ERR(0, 457, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 457, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_expect_one_of, __pyx_n_s_format); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 457, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (unlikely(!__pyx_cur_scope->__pyx_v_expects)) { __Pyx_RaiseClosureNameError("expects"); __PYX_ERR(0, 457, __pyx_L1_error) }
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_5 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_7, __pyx_cur_scope->__pyx_v_expects) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_cur_scope->__pyx_v_expects);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 457, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 457, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_Raise(__pyx_t_3, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __PYX_ERR(0, 457, __pyx_L1_error)

  /* "validr/_validator_c.pyx":454
 *     items = frozenset(items)
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         if value in items:
 *             return value
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("validr._validator_c.enum_validator.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":448
 * 
 * @validator(output=object)
 * def enum_validator(compiler, items):             # <<<<<<<<<<<<<<
 *     if not items:
 *         raise SchemaError("enum items not provided")
 */

static PyObject *__pyx_pf_6validr_12_validator_c_30enum_validator(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_compiler, PyObject *__pyx_v_items) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_11_enum_validator *__pyx_cur_scope;
  PyObject *__pyx_v_validate = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_TraceFrameInit(__pyx_codeobj__38)
  __Pyx_RefNannySetupContext("enum_validator", 0);
  __pyx_cur_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_11_enum_validator *)__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_11_enum_validator(__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_11_enum_validator, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_11_enum_validator *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 448, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __Pyx_TraceCall("enum_validator", __pyx_f[0], 448, 0, __PYX_ERR(0, 448, __pyx_L1_error));
  __pyx_cur_scope->__pyx_v_items = __pyx_v_items;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_items);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_items);

  /* "validr/_validator_c.pyx":449
 * @validator(output=object)
 * def enum_validator(compiler, items):
 *     if not items:             # <<<<<<<<<<<<<<
 *         raise SchemaError("enum items not provided")
 *     expects = '{' + ', '.join(map(_dump_enum_value, items)) + '}'
 */
  __Pyx_TraceLine(449,0,__PYX_ERR(0, 449, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_items); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 449, __pyx_L1_error)
  __pyx_t_2 = ((!__pyx_t_1) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "validr/_validator_c.pyx":450
 * def enum_validator(compiler, items):
 *     if not items:
 *         raise SchemaError("enum items not provided")             # <<<<<<<<<<<<<<
 *     expects = '{' + ', '.join(map(_dump_enum_value, items)) + '}'
 *     items = frozenset(items)
 */
    __Pyx_TraceLine(450,0,__PYX_ERR(0, 450, __pyx_L1_error))
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_SchemaError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 450, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_kp_u_enum_items_not_provided) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_kp_u_enum_items_not_provided);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 450, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 450, __pyx_L1_error)

    /* "validr/_validator_c.pyx":449
 * @validator(output=object)
 * def enum_validator(compiler, items):
 *     if not items:             # <<<<<<<<<<<<<<
 *         raise SchemaError("enum items not provided")
 *     expects = '{' + ', '.join(map(_dump_enum_value, items)) + '}'
 */
  }

  /* "validr/_validator_c.pyx":451
 *     if not items:
 *         raise SchemaError("enum items not provided")
 *     expects = '{' + ', '.join(map(_dump_enum_value, items)) + '}'             # <<<<<<<<<<<<<<
 *     items = frozenset(items)
 * 
 */
  __Pyx_TraceLine(451,0,__PYX_ERR(0, 451, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_dump_enum_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_items);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_items);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_cur_scope->__pyx_v_items);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_map, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyUnicode_Join(__pyx_kp_u__40, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyUnicode_Concat(__pyx_kp_u__39, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyUnicode_Concat(__pyx_t_3, __pyx_kp_u__41); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_cur_scope->__pyx_v_expects = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "validr/_validator_c.pyx":452
 *         raise SchemaError("enum items not provided")
 *     expects = '{' + ', '.join(map(_dump_enum_value, items)) + '}'
 *     items = frozenset(items)             # <<<<<<<<<<<<<<
 * 
 *     def validate(value):
 */
  __Pyx_TraceLine(452,0,__PYX_ERR(0, 452, __pyx_L1_error))
  __pyx_t_4 = __Pyx_PyFrozenSet_New(__pyx_cur_scope->__pyx_v_items); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_items);
  __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_items, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "validr/_validator_c.pyx":454
 *     items = frozenset(items)
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         if value in items:
 *             return value
 */
  __Pyx_TraceLine(454,0,__PYX_ERR(0, 454, __pyx_L1_error))
  __pyx_t_4 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_14enum_validator_1validate, 0, __pyx_n_s_enum_validator_locals_validate, ((PyObject*)__pyx_cur_scope), __pyx_n_s_validr__validator_c, __pyx_d, ((PyObject *)__pyx_codeobj__43)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 454, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_validate = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "validr/_validator_c.pyx":459
 *         raise Invalid("expect one of {}".format(expects))
 * 
 *     return validate             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(459,0,__PYX_ERR(0, 459, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_validate);
  __pyx_r = __pyx_v_validate;
  goto __pyx_L0;

  /* "validr/_validator_c.pyx":448
 * 
 * @validator(output=object)
 * def enum_validator(compiler, items):             # <<<<<<<<<<<<<<
 *     if not items:
 *         raise SchemaError("enum items not provided")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("validr._validator_c.enum_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_validate);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":462
 * 
 * 
 * cpdef union_validator(compiler, schema):             # <<<<<<<<<<<<<<
 *     if not schema.items:
 *         raise SchemaError('union schemas not provided')
 */

static PyObject *__pyx_pw_6validr_12_validator_c_33union_validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_6validr_12_validator_c_union_validator(PyObject *__pyx_v_compiler, PyObject *__pyx_v_schema, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_v_default = NULL;
  PyObject *__pyx_v_by = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("union_validator", 0);
  __Pyx_TraceCall("union_validator", __pyx_f[0], 462, 0, __PYX_ERR(0, 462, __pyx_L1_error));

  /* "validr/_validator_c.pyx":463
 * 
 * cpdef union_validator(compiler, schema):
 *     if not schema.items:             # <<<<<<<<<<<<<<
 *         raise SchemaError('union schemas not provided')
 *     default = schema.params.get('default')
 */
  __Pyx_TraceLine(463,0,__PYX_ERR(0, 463, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_schema, __pyx_n_s_items); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 463, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 463, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = ((!__pyx_t_2) != 0);
  if (unlikely(__pyx_t_3)) {

    /* "validr/_validator_c.pyx":464
 * cpdef union_validator(compiler, schema):
 *     if not schema.items:
 *         raise SchemaError('union schemas not provided')             # <<<<<<<<<<<<<<
 *     default = schema.params.get('default')
 *     if default is not None:
 */
    __Pyx_TraceLine(464,0,__PYX_ERR(0, 464, __pyx_L1_error))
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_SchemaError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 464, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_kp_u_union_schemas_not_provided) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_kp_u_union_schemas_not_provided);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 464, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 464, __pyx_L1_error)

    /* "validr/_validator_c.pyx":463
 * 
 * cpdef union_validator(compiler, schema):
 *     if not schema.items:             # <<<<<<<<<<<<<<
 *         raise SchemaError('union schemas not provided')
 *     default = schema.params.get('default')
 */
  }

  /* "validr/_validator_c.pyx":465
 *     if not schema.items:
 *         raise SchemaError('union schemas not provided')
 *     default = schema.params.get('default')             # <<<<<<<<<<<<<<
 *     if default is not None:
 *         raise SchemaError("not allowed default for union schema")
 */
  __Pyx_TraceLine(465,0,__PYX_ERR(0, 465, __pyx_L1_error))
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_schema, __pyx_n_s_params); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 465, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_get); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 465, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_4, __pyx_n_u_default) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_n_u_default);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 465, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_default = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":466
 *         raise SchemaError('union schemas not provided')
 *     default = schema.params.get('default')
 *     if default is not None:             # <<<<<<<<<<<<<<
 *         raise SchemaError("not allowed default for union schema")
 *     by = schema.params.get('by')
 */
  __Pyx_TraceLine(466,0,__PYX_ERR(0, 466, __pyx_L1_error))
  __pyx_t_3 = (__pyx_v_default != Py_None);
  __pyx_t_2 = (__pyx_t_3 != 0);
  if (unlikely(__pyx_t_2)) {

    /* "validr/_validator_c.pyx":467
 *     default = schema.params.get('default')
 *     if default is not None:
 *         raise SchemaError("not allowed default for union schema")             # <<<<<<<<<<<<<<
 *     by = schema.params.get('by')
 *     if isinstance(schema.items, list):
 */
    __Pyx_TraceLine(467,0,__PYX_ERR(0, 467, __pyx_L1_error))
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_SchemaError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 467, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_4, __pyx_kp_u_not_allowed_default_for_union_sc) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_kp_u_not_allowed_default_for_union_sc);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 467, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 467, __pyx_L1_error)

    /* "validr/_validator_c.pyx":466
 *         raise SchemaError('union schemas not provided')
 *     default = schema.params.get('default')
 *     if default is not None:             # <<<<<<<<<<<<<<
 *         raise SchemaError("not allowed default for union schema")
 *     by = schema.params.get('by')
 */
  }

  /* "validr/_validator_c.pyx":468
 *     if default is not None:
 *         raise SchemaError("not allowed default for union schema")
 *     by = schema.params.get('by')             # <<<<<<<<<<<<<<
 *     if isinstance(schema.items, list):
 *         if by is not None:
 */
  __Pyx_TraceLine(468,0,__PYX_ERR(0, 468, __pyx_L1_error))
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_schema, __pyx_n_s_params); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 468, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_get); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 468, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_n_u_by) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_n_u_by);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 468, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_by = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":469
 *         raise SchemaError("not allowed default for union schema")
 *     by = schema.params.get('by')
 *     if isinstance(schema.items, list):             # <<<<<<<<<<<<<<
 *         if by is not None:
 *             raise SchemaError("not allowed 'by' argument for union list schema")
 */
  __Pyx_TraceLine(469,0,__PYX_ERR(0, 469, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_schema, __pyx_n_s_items); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyList_Check(__pyx_t_1); 
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "validr/_validator_c.pyx":470
 *     by = schema.params.get('by')
 *     if isinstance(schema.items, list):
 *         if by is not None:             # <<<<<<<<<<<<<<
 *             raise SchemaError("not allowed 'by' argument for union list schema")
 *         return _union_list_validator(compiler, schema)
 */
    __Pyx_TraceLine(470,0,__PYX_ERR(0, 470, __pyx_L1_error))
    __pyx_t_3 = (__pyx_v_by != Py_None);
    __pyx_t_2 = (__pyx_t_3 != 0);
    if (unlikely(__pyx_t_2)) {

      /* "validr/_validator_c.pyx":471
 *     if isinstance(schema.items, list):
 *         if by is not None:
 *             raise SchemaError("not allowed 'by' argument for union list schema")             # <<<<<<<<<<<<<<
 *         return _union_list_validator(compiler, schema)
 *     elif isinstance(schema.items, dict):
 */
      __Pyx_TraceLine(471,0,__PYX_ERR(0, 471, __pyx_L1_error))
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_SchemaError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 471, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_kp_u_not_allowed_by_argument_for_unio) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_kp_u_not_allowed_by_argument_for_unio);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 471, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(0, 471, __pyx_L1_error)

      /* "validr/_validator_c.pyx":470
 *     by = schema.params.get('by')
 *     if isinstance(schema.items, list):
 *         if by is not None:             # <<<<<<<<<<<<<<
 *             raise SchemaError("not allowed 'by' argument for union list schema")
 *         return _union_list_validator(compiler, schema)
 */
    }

    /* "validr/_validator_c.pyx":472
 *         if by is not None:
 *             raise SchemaError("not allowed 'by' argument for union list schema")
 *         return _union_list_validator(compiler, schema)             # <<<<<<<<<<<<<<
 *     elif isinstance(schema.items, dict):
 *         if by is None or by == "":
 */
    __Pyx_TraceLine(472,0,__PYX_ERR(0, 472, __pyx_L1_error))
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_union_list_validator); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 472, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_compiler, __pyx_v_schema};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 472, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_compiler, __pyx_v_schema};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 472, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 472, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (__pyx_t_5) {
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
      }
      __Pyx_INCREF(__pyx_v_compiler);
      __Pyx_GIVEREF(__pyx_v_compiler);
      PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_compiler);
      __Pyx_INCREF(__pyx_v_schema);
      __Pyx_GIVEREF(__pyx_v_schema);
      PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_schema);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 472, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "validr/_validator_c.pyx":469
 *         raise SchemaError("not allowed default for union schema")
 *     by = schema.params.get('by')
 *     if isinstance(schema.items, list):             # <<<<<<<<<<<<<<
 *         if by is not None:
 *             raise SchemaError("not allowed 'by' argument for union list schema")
 */
  }

  /* "validr/_validator_c.pyx":473
 *             raise SchemaError("not allowed 'by' argument for union list schema")
 *         return _union_list_validator(compiler, schema)
 *     elif isinstance(schema.items, dict):             # <<<<<<<<<<<<<<
 *         if by is None or by == "":
 *             raise SchemaError("required 'by' argument for union dict schema")
 */
  __Pyx_TraceLine(473,0,__PYX_ERR(0, 473, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_schema, __pyx_n_s_items); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 473, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyDict_Check(__pyx_t_1); 
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (likely(__pyx_t_3)) {

    /* "validr/_validator_c.pyx":474
 *         return _union_list_validator(compiler, schema)
 *     elif isinstance(schema.items, dict):
 *         if by is None or by == "":             # <<<<<<<<<<<<<<
 *             raise SchemaError("required 'by' argument for union dict schema")
 *         if not isinstance(by, str):
 */
    __Pyx_TraceLine(474,0,__PYX_ERR(0, 474, __pyx_L1_error))
    __pyx_t_2 = (__pyx_v_by == Py_None);
    __pyx_t_8 = (__pyx_t_2 != 0);
    if (!__pyx_t_8) {
    } else {
      __pyx_t_3 = __pyx_t_8;
      goto __pyx_L8_bool_binop_done;
    }
    __pyx_t_8 = (__Pyx_PyUnicode_Equals(__pyx_v_by, __pyx_kp_u_, Py_EQ)); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 474, __pyx_L1_error)
    __pyx_t_3 = __pyx_t_8;
    __pyx_L8_bool_binop_done:;
    if (unlikely(__pyx_t_3)) {

      /* "validr/_validator_c.pyx":475
 *     elif isinstance(schema.items, dict):
 *         if by is None or by == "":
 *             raise SchemaError("required 'by' argument for union dict schema")             # <<<<<<<<<<<<<<
 *         if not isinstance(by, str):
 *             raise SchemaError("'by' argument must be str type for union schema")
 */
      __Pyx_TraceLine(475,0,__PYX_ERR(0, 475, __pyx_L1_error))
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_SchemaError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 475, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_7 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_7, __pyx_kp_u_required_by_argument_for_union_d) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_kp_u_required_by_argument_for_union_d);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 475, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(0, 475, __pyx_L1_error)

      /* "validr/_validator_c.pyx":474
 *         return _union_list_validator(compiler, schema)
 *     elif isinstance(schema.items, dict):
 *         if by is None or by == "":             # <<<<<<<<<<<<<<
 *             raise SchemaError("required 'by' argument for union dict schema")
 *         if not isinstance(by, str):
 */
    }

    /* "validr/_validator_c.pyx":476
 *         if by is None or by == "":
 *             raise SchemaError("required 'by' argument for union dict schema")
 *         if not isinstance(by, str):             # <<<<<<<<<<<<<<
 *             raise SchemaError("'by' argument must be str type for union schema")
 *         return _union_dict_validator(compiler, schema)
 */
    __Pyx_TraceLine(476,0,__PYX_ERR(0, 476, __pyx_L1_error))
    __pyx_t_3 = PyUnicode_Check(__pyx_v_by); 
    __pyx_t_8 = ((!(__pyx_t_3 != 0)) != 0);
    if (unlikely(__pyx_t_8)) {

      /* "validr/_validator_c.pyx":477
 *             raise SchemaError("required 'by' argument for union dict schema")
 *         if not isinstance(by, str):
 *             raise SchemaError("'by' argument must be str type for union schema")             # <<<<<<<<<<<<<<
 *         return _union_dict_validator(compiler, schema)
 *     else:
 */
      __Pyx_TraceLine(477,0,__PYX_ERR(0, 477, __pyx_L1_error))
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_SchemaError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 477, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_7 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_7, __pyx_kp_u_by_argument_must_be_str_type_fo) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_kp_u_by_argument_must_be_str_type_fo);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 477, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(0, 477, __pyx_L1_error)

      /* "validr/_validator_c.pyx":476
 *         if by is None or by == "":
 *             raise SchemaError("required 'by' argument for union dict schema")
 *         if not isinstance(by, str):             # <<<<<<<<<<<<<<
 *             raise SchemaError("'by' argument must be str type for union schema")
 *         return _union_dict_validator(compiler, schema)
 */
    }

    /* "validr/_validator_c.pyx":478
 *         if not isinstance(by, str):
 *             raise SchemaError("'by' argument must be str type for union schema")
 *         return _union_dict_validator(compiler, schema)             # <<<<<<<<<<<<<<
 *     else:
 *         raise SchemaError('union schemas type invalid')
 */
    __Pyx_TraceLine(478,0,__PYX_ERR(0, 478, __pyx_L1_error))
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_union_dict_validator); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 478, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_compiler, __pyx_v_schema};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 478, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_compiler, __pyx_v_schema};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 478, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 478, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_INCREF(__pyx_v_compiler);
      __Pyx_GIVEREF(__pyx_v_compiler);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_v_compiler);
      __Pyx_INCREF(__pyx_v_schema);
      __Pyx_GIVEREF(__pyx_v_schema);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_v_schema);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 478, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "validr/_validator_c.pyx":473
 *             raise SchemaError("not allowed 'by' argument for union list schema")
 *         return _union_list_validator(compiler, schema)
 *     elif isinstance(schema.items, dict):             # <<<<<<<<<<<<<<
 *         if by is None or by == "":
 *             raise SchemaError("required 'by' argument for union dict schema")
 */
  }

  /* "validr/_validator_c.pyx":480
 *         return _union_dict_validator(compiler, schema)
 *     else:
 *         raise SchemaError('union schemas type invalid')             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(480,0,__PYX_ERR(0, 480, __pyx_L1_error))
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_SchemaError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 480, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_kp_u_union_schemas_type_invalid) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_kp_u_union_schemas_type_invalid);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 480, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 480, __pyx_L1_error)
  }

  /* "validr/_validator_c.pyx":462
 * 
 * 
 * cpdef union_validator(compiler, schema):             # <<<<<<<<<<<<<<
 *     if not schema.items:
 *         raise SchemaError('union schemas not provided')
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("validr._validator_c.union_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_default);
  __Pyx_XDECREF(__pyx_v_by);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_33union_validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6validr_12_validator_c_33union_validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_compiler = 0;
  PyObject *__pyx_v_schema = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("union_validator (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_compiler,&__pyx_n_s_schema_2,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_compiler)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_schema_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("union_validator", 1, 2, 2, 1); __PYX_ERR(0, 462, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "union_validator") < 0)) __PYX_ERR(0, 462, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_compiler = values[0];
    __pyx_v_schema = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("union_validator", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 462, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("validr._validator_c.union_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6validr_12_validator_c_32union_validator(__pyx_self, __pyx_v_compiler, __pyx_v_schema);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6validr_12_validator_c_32union_validator(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_compiler, PyObject *__pyx_v_schema) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("union_validator", 0);
  __Pyx_TraceCall("union_validator (wrapper)", __pyx_f[0], 462, 0, __PYX_ERR(0, 462, __pyx_L1_error));
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6validr_12_validator_c_union_validator(__pyx_v_compiler, __pyx_v_schema, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("validr._validator_c.union_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":483
 * 
 * 
 * cpdef _optional_or_has_default(schema):             # <<<<<<<<<<<<<<
 *     if schema.params.get('optional'):
 *         return True
 */

static PyObject *__pyx_pw_6validr_12_validator_c_35_optional_or_has_default(PyObject *__pyx_self, PyObject *__pyx_v_schema); /*proto*/
static PyObject *__pyx_f_6validr_12_validator_c__optional_or_has_default(PyObject *__pyx_v_schema, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  __Pyx_RefNannySetupContext("_optional_or_has_default", 0);
  __Pyx_TraceCall("_optional_or_has_default", __pyx_f[0], 483, 0, __PYX_ERR(0, 483, __pyx_L1_error));

  /* "validr/_validator_c.pyx":484
 * 
 * cpdef _optional_or_has_default(schema):
 *     if schema.params.get('optional'):             # <<<<<<<<<<<<<<
 *         return True
 *     if schema.params.get('default') is not None:
 */
  __Pyx_TraceLine(484,0,__PYX_ERR(0, 484, __pyx_L1_error))
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_schema, __pyx_n_s_params); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 484, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_get); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 484, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_n_u_optional) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_n_u_optional);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 484, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 484, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_4) {

    /* "validr/_validator_c.pyx":485
 * cpdef _optional_or_has_default(schema):
 *     if schema.params.get('optional'):
 *         return True             # <<<<<<<<<<<<<<
 *     if schema.params.get('default') is not None:
 *         return True
 */
    __Pyx_TraceLine(485,0,__PYX_ERR(0, 485, __pyx_L1_error))
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_True);
    __pyx_r = Py_True;
    goto __pyx_L0;

    /* "validr/_validator_c.pyx":484
 * 
 * cpdef _optional_or_has_default(schema):
 *     if schema.params.get('optional'):             # <<<<<<<<<<<<<<
 *         return True
 *     if schema.params.get('default') is not None:
 */
  }

  /* "validr/_validator_c.pyx":486
 *     if schema.params.get('optional'):
 *         return True
 *     if schema.params.get('default') is not None:             # <<<<<<<<<<<<<<
 *         return True
 *     return False
 */
  __Pyx_TraceLine(486,0,__PYX_ERR(0, 486, __pyx_L1_error))
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_schema, __pyx_n_s_params); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 486, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_get); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 486, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_n_u_default) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_n_u_default);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 486, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = (__pyx_t_1 != Py_None);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {

    /* "validr/_validator_c.pyx":487
 *         return True
 *     if schema.params.get('default') is not None:
 *         return True             # <<<<<<<<<<<<<<
 *     return False
 * 
 */
    __Pyx_TraceLine(487,0,__PYX_ERR(0, 487, __pyx_L1_error))
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_True);
    __pyx_r = Py_True;
    goto __pyx_L0;

    /* "validr/_validator_c.pyx":486
 *     if schema.params.get('optional'):
 *         return True
 *     if schema.params.get('default') is not None:             # <<<<<<<<<<<<<<
 *         return True
 *     return False
 */
  }

  /* "validr/_validator_c.pyx":488
 *     if schema.params.get('default') is not None:
 *         return True
 *     return False             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(488,0,__PYX_ERR(0, 488, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(Py_False);
  __pyx_r = Py_False;
  goto __pyx_L0;

  /* "validr/_validator_c.pyx":483
 * 
 * 
 * cpdef _optional_or_has_default(schema):             # <<<<<<<<<<<<<<
 *     if schema.params.get('optional'):
 *         return True
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("validr._validator_c._optional_or_has_default", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_35_optional_or_has_default(PyObject *__pyx_self, PyObject *__pyx_v_schema); /*proto*/
static PyObject *__pyx_pw_6validr_12_validator_c_35_optional_or_has_default(PyObject *__pyx_self, PyObject *__pyx_v_schema) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_optional_or_has_default (wrapper)", 0);
  __pyx_r = __pyx_pf_6validr_12_validator_c_34_optional_or_has_default(__pyx_self, ((PyObject *)__pyx_v_schema));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6validr_12_validator_c_34_optional_or_has_default(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_schema) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_optional_or_has_default", 0);
  __Pyx_TraceCall("_optional_or_has_default (wrapper)", __pyx_f[0], 483, 0, __PYX_ERR(0, 483, __pyx_L1_error));
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6validr_12_validator_c__optional_or_has_default(__pyx_v_schema, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 483, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("validr._validator_c._optional_or_has_default", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":491
 * 
 * 
 * def _union_list_validator(compiler, schema):             # <<<<<<<<<<<<<<
 *     scalar_inner = None
 *     list_inner = None
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_37_union_list_validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6validr_12_validator_c_37_union_list_validator = {"_union_list_validator", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6validr_12_validator_c_37_union_list_validator, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6validr_12_validator_c_37_union_list_validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_compiler = 0;
  PyObject *__pyx_v_schema = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_union_list_validator (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_compiler,&__pyx_n_s_schema_2,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_compiler)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_schema_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_union_list_validator", 1, 2, 2, 1); __PYX_ERR(0, 491, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_union_list_validator") < 0)) __PYX_ERR(0, 491, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_compiler = values[0];
    __pyx_v_schema = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_union_list_validator", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 491, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("validr._validator_c._union_list_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6validr_12_validator_c_36_union_list_validator(__pyx_self, __pyx_v_compiler, __pyx_v_schema);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":517
 *                 scalar_inner = compiler.compile(inner_schema)
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         if isinstance(value, list):
 *             if list_inner is None:
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_21_union_list_validator_1validate(PyObject *__pyx_self, PyObject *__pyx_v_value); /*proto*/
static PyMethodDef __pyx_mdef_6validr_12_validator_c_21_union_list_validator_1validate = {"validate", (PyCFunction)__pyx_pw_6validr_12_validator_c_21_union_list_validator_1validate, METH_O, 0};
static PyObject *__pyx_pw_6validr_12_validator_c_21_union_list_validator_1validate(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("validate (wrapper)", 0);
  __pyx_r = __pyx_pf_6validr_12_validator_c_21_union_list_validator_validate(__pyx_self, ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6validr_12_validator_c_21_union_list_validator_validate(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_12__union_list_validator *__pyx_cur_scope;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_12__union_list_validator *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("validate", 0);
  __pyx_outer_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_12__union_list_validator *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_TraceCall("validate", __pyx_f[0], 517, 0, __PYX_ERR(0, 517, __pyx_L1_error));

  /* "validr/_validator_c.pyx":518
 * 
 *     def validate(value):
 *         if isinstance(value, list):             # <<<<<<<<<<<<<<
 *             if list_inner is None:
 *                 raise Invalid('not allowed list')
 */
  __Pyx_TraceLine(518,0,__PYX_ERR(0, 518, __pyx_L1_error))
  __pyx_t_1 = PyList_Check(__pyx_v_value); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "validr/_validator_c.pyx":519
 *     def validate(value):
 *         if isinstance(value, list):
 *             if list_inner is None:             # <<<<<<<<<<<<<<
 *                 raise Invalid('not allowed list')
 *             return list_inner(value)
 */
    __Pyx_TraceLine(519,0,__PYX_ERR(0, 519, __pyx_L1_error))
    if (unlikely(!__pyx_cur_scope->__pyx_v_list_inner)) { __Pyx_RaiseClosureNameError("list_inner"); __PYX_ERR(0, 519, __pyx_L1_error) }
    __pyx_t_2 = (__pyx_cur_scope->__pyx_v_list_inner == Py_None);
    __pyx_t_1 = (__pyx_t_2 != 0);
    if (unlikely(__pyx_t_1)) {

      /* "validr/_validator_c.pyx":520
 *         if isinstance(value, list):
 *             if list_inner is None:
 *                 raise Invalid('not allowed list')             # <<<<<<<<<<<<<<
 *             return list_inner(value)
 *         elif is_dict(value) or hasattr(value, '__asdict__'):
 */
      __Pyx_TraceLine(520,0,__PYX_ERR(0, 520, __pyx_L1_error))
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 520, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_kp_u_not_allowed_list) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_kp_u_not_allowed_list);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 520, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 520, __pyx_L1_error)

      /* "validr/_validator_c.pyx":519
 *     def validate(value):
 *         if isinstance(value, list):
 *             if list_inner is None:             # <<<<<<<<<<<<<<
 *                 raise Invalid('not allowed list')
 *             return list_inner(value)
 */
    }

    /* "validr/_validator_c.pyx":521
 *             if list_inner is None:
 *                 raise Invalid('not allowed list')
 *             return list_inner(value)             # <<<<<<<<<<<<<<
 *         elif is_dict(value) or hasattr(value, '__asdict__'):
 *             if dict_inner is None:
 */
    __Pyx_TraceLine(521,0,__PYX_ERR(0, 521, __pyx_L1_error))
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(!__pyx_cur_scope->__pyx_v_list_inner)) { __Pyx_RaiseClosureNameError("list_inner"); __PYX_ERR(0, 521, __pyx_L1_error) }
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_list_inner);
    __pyx_t_4 = __pyx_cur_scope->__pyx_v_list_inner; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_value) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_value);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 521, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "validr/_validator_c.pyx":518
 * 
 *     def validate(value):
 *         if isinstance(value, list):             # <<<<<<<<<<<<<<
 *             if list_inner is None:
 *                 raise Invalid('not allowed list')
 */
  }

  /* "validr/_validator_c.pyx":522
 *                 raise Invalid('not allowed list')
 *             return list_inner(value)
 *         elif is_dict(value) or hasattr(value, '__asdict__'):             # <<<<<<<<<<<<<<
 *             if dict_inner is None:
 *                 raise Invalid('not allowed dict')
 */
  __Pyx_TraceLine(522,0,__PYX_ERR(0, 522, __pyx_L1_error))
  __pyx_t_2 = (__pyx_f_6validr_12_validator_c_is_dict(__pyx_v_value, 0) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_2 = __Pyx_HasAttr(__pyx_v_value, __pyx_n_u_asdict); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 522, __pyx_L1_error)
  __pyx_t_6 = (__pyx_t_2 != 0);
  __pyx_t_1 = __pyx_t_6;
  __pyx_L5_bool_binop_done:;
  if (__pyx_t_1) {

    /* "validr/_validator_c.pyx":523
 *             return list_inner(value)
 *         elif is_dict(value) or hasattr(value, '__asdict__'):
 *             if dict_inner is None:             # <<<<<<<<<<<<<<
 *                 raise Invalid('not allowed dict')
 *             return dict_inner(value)
 */
    __Pyx_TraceLine(523,0,__PYX_ERR(0, 523, __pyx_L1_error))
    if (unlikely(!__pyx_cur_scope->__pyx_v_dict_inner)) { __Pyx_RaiseClosureNameError("dict_inner"); __PYX_ERR(0, 523, __pyx_L1_error) }
    __pyx_t_1 = (__pyx_cur_scope->__pyx_v_dict_inner == Py_None);
    __pyx_t_6 = (__pyx_t_1 != 0);
    if (unlikely(__pyx_t_6)) {

      /* "validr/_validator_c.pyx":524
 *         elif is_dict(value) or hasattr(value, '__asdict__'):
 *             if dict_inner is None:
 *                 raise Invalid('not allowed dict')             # <<<<<<<<<<<<<<
 *             return dict_inner(value)
 *         elif value is None:
 */
      __Pyx_TraceLine(524,0,__PYX_ERR(0, 524, __pyx_L1_error))
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 524, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_kp_u_not_allowed_dict) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_kp_u_not_allowed_dict);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 524, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 524, __pyx_L1_error)

      /* "validr/_validator_c.pyx":523
 *             return list_inner(value)
 *         elif is_dict(value) or hasattr(value, '__asdict__'):
 *             if dict_inner is None:             # <<<<<<<<<<<<<<
 *                 raise Invalid('not allowed dict')
 *             return dict_inner(value)
 */
    }

    /* "validr/_validator_c.pyx":525
 *             if dict_inner is None:
 *                 raise Invalid('not allowed dict')
 *             return dict_inner(value)             # <<<<<<<<<<<<<<
 *         elif value is None:
 *             return (scalar_inner or list_inner or dict_inner)(value)
 */
    __Pyx_TraceLine(525,0,__PYX_ERR(0, 525, __pyx_L1_error))
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(!__pyx_cur_scope->__pyx_v_dict_inner)) { __Pyx_RaiseClosureNameError("dict_inner"); __PYX_ERR(0, 525, __pyx_L1_error) }
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_dict_inner);
    __pyx_t_4 = __pyx_cur_scope->__pyx_v_dict_inner; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_value) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_value);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 525, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "validr/_validator_c.pyx":522
 *                 raise Invalid('not allowed list')
 *             return list_inner(value)
 *         elif is_dict(value) or hasattr(value, '__asdict__'):             # <<<<<<<<<<<<<<
 *             if dict_inner is None:
 *                 raise Invalid('not allowed dict')
 */
  }

  /* "validr/_validator_c.pyx":526
 *                 raise Invalid('not allowed dict')
 *             return dict_inner(value)
 *         elif value is None:             # <<<<<<<<<<<<<<
 *             return (scalar_inner or list_inner or dict_inner)(value)
 *         else:
 */
  __Pyx_TraceLine(526,0,__PYX_ERR(0, 526, __pyx_L1_error))
  __pyx_t_6 = (__pyx_v_value == Py_None);
  __pyx_t_1 = (__pyx_t_6 != 0);
  if (__pyx_t_1) {

    /* "validr/_validator_c.pyx":527
 *             return dict_inner(value)
 *         elif value is None:
 *             return (scalar_inner or list_inner or dict_inner)(value)             # <<<<<<<<<<<<<<
 *         else:
 *             if scalar_inner is None:
 */
    __Pyx_TraceLine(527,0,__PYX_ERR(0, 527, __pyx_L1_error))
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(!__pyx_cur_scope->__pyx_v_scalar_inner)) { __Pyx_RaiseClosureNameError("scalar_inner"); __PYX_ERR(0, 527, __pyx_L1_error) }
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_scalar_inner); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 527, __pyx_L1_error)
    if (!__pyx_t_1) {
    } else {
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_scalar_inner);
      __pyx_t_4 = __pyx_cur_scope->__pyx_v_scalar_inner;
      goto __pyx_L8_bool_binop_done;
    }
    if (unlikely(!__pyx_cur_scope->__pyx_v_list_inner)) { __Pyx_RaiseClosureNameError("list_inner"); __PYX_ERR(0, 527, __pyx_L1_error) }
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_list_inner); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 527, __pyx_L1_error)
    if (!__pyx_t_1) {
    } else {
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_list_inner);
      __pyx_t_4 = __pyx_cur_scope->__pyx_v_list_inner;
      goto __pyx_L8_bool_binop_done;
    }
    if (unlikely(!__pyx_cur_scope->__pyx_v_dict_inner)) { __Pyx_RaiseClosureNameError("dict_inner"); __PYX_ERR(0, 527, __pyx_L1_error) }
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_dict_inner);
    __pyx_t_4 = __pyx_cur_scope->__pyx_v_dict_inner;
    __pyx_L8_bool_binop_done:;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_value) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_value);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 527, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "validr/_validator_c.pyx":526
 *                 raise Invalid('not allowed dict')
 *             return dict_inner(value)
 *         elif value is None:             # <<<<<<<<<<<<<<
 *             return (scalar_inner or list_inner or dict_inner)(value)
 *         else:
 */
  }

  /* "validr/_validator_c.pyx":529
 *             return (scalar_inner or list_inner or dict_inner)(value)
 *         else:
 *             if scalar_inner is None:             # <<<<<<<<<<<<<<
 *                 raise Invalid('not allowed scalar value')
 *             return scalar_inner(value)
 */
  __Pyx_TraceLine(529,0,__PYX_ERR(0, 529, __pyx_L1_error))
  /*else*/ {
    if (unlikely(!__pyx_cur_scope->__pyx_v_scalar_inner)) { __Pyx_RaiseClosureNameError("scalar_inner"); __PYX_ERR(0, 529, __pyx_L1_error) }
    __pyx_t_1 = (__pyx_cur_scope->__pyx_v_scalar_inner == Py_None);
    __pyx_t_6 = (__pyx_t_1 != 0);
    if (unlikely(__pyx_t_6)) {

      /* "validr/_validator_c.pyx":530
 *         else:
 *             if scalar_inner is None:
 *                 raise Invalid('not allowed scalar value')             # <<<<<<<<<<<<<<
 *             return scalar_inner(value)
 * 
 */
      __Pyx_TraceLine(530,0,__PYX_ERR(0, 530, __pyx_L1_error))
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 530, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_kp_u_not_allowed_scalar_value) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_kp_u_not_allowed_scalar_value);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 530, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 530, __pyx_L1_error)

      /* "validr/_validator_c.pyx":529
 *             return (scalar_inner or list_inner or dict_inner)(value)
 *         else:
 *             if scalar_inner is None:             # <<<<<<<<<<<<<<
 *                 raise Invalid('not allowed scalar value')
 *             return scalar_inner(value)
 */
    }

    /* "validr/_validator_c.pyx":531
 *             if scalar_inner is None:
 *                 raise Invalid('not allowed scalar value')
 *             return scalar_inner(value)             # <<<<<<<<<<<<<<
 * 
 *     _update_validate_func_info(validate, union_validator, schema)
 */
    __Pyx_TraceLine(531,0,__PYX_ERR(0, 531, __pyx_L1_error))
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(!__pyx_cur_scope->__pyx_v_scalar_inner)) { __Pyx_RaiseClosureNameError("scalar_inner"); __PYX_ERR(0, 531, __pyx_L1_error) }
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_scalar_inner);
    __pyx_t_4 = __pyx_cur_scope->__pyx_v_scalar_inner; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_value) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_value);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 531, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "validr/_validator_c.pyx":517
 *                 scalar_inner = compiler.compile(inner_schema)
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         if isinstance(value, list):
 *             if list_inner is None:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("validr._validator_c._union_list_validator.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":491
 * 
 * 
 * def _union_list_validator(compiler, schema):             # <<<<<<<<<<<<<<
 *     scalar_inner = None
 *     list_inner = None
 */

static PyObject *__pyx_pf_6validr_12_validator_c_36_union_list_validator(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_compiler, PyObject *__pyx_v_schema) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_12__union_list_validator *__pyx_cur_scope;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_inner_schema = NULL;
  PyObject *__pyx_v_validate = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  int __pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  __Pyx_TraceFrameInit(__pyx_codeobj__44)
  __Pyx_RefNannySetupContext("_union_list_validator", 0);
  __pyx_cur_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_12__union_list_validator *)__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_12__union_list_validator(__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_12__union_list_validator, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_12__union_list_validator *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 491, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __Pyx_TraceCall("_union_list_validator", __pyx_f[0], 491, 0, __PYX_ERR(0, 491, __pyx_L1_error));

  /* "validr/_validator_c.pyx":492
 * 
 * def _union_list_validator(compiler, schema):
 *     scalar_inner = None             # <<<<<<<<<<<<<<
 *     list_inner = None
 *     dict_inner = None
 */
  __Pyx_TraceLine(492,0,__PYX_ERR(0, 492, __pyx_L1_error))
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __pyx_cur_scope->__pyx_v_scalar_inner = Py_None;

  /* "validr/_validator_c.pyx":493
 * def _union_list_validator(compiler, schema):
 *     scalar_inner = None
 *     list_inner = None             # <<<<<<<<<<<<<<
 *     dict_inner = None
 *     for i, inner_schema in enumerate(schema.items):
 */
  __Pyx_TraceLine(493,0,__PYX_ERR(0, 493, __pyx_L1_error))
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __pyx_cur_scope->__pyx_v_list_inner = Py_None;

  /* "validr/_validator_c.pyx":494
 *     scalar_inner = None
 *     list_inner = None
 *     dict_inner = None             # <<<<<<<<<<<<<<
 *     for i, inner_schema in enumerate(schema.items):
 *         with mark_index(i):
 */
  __Pyx_TraceLine(494,0,__PYX_ERR(0, 494, __pyx_L1_error))
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __pyx_cur_scope->__pyx_v_dict_inner = Py_None;

  /* "validr/_validator_c.pyx":495
 *     list_inner = None
 *     dict_inner = None
 *     for i, inner_schema in enumerate(schema.items):             # <<<<<<<<<<<<<<
 *         with mark_index(i):
 *             if inner_schema.validator == 'union':
 */
  __Pyx_TraceLine(495,0,__PYX_ERR(0, 495, __pyx_L1_error))
  __Pyx_INCREF(__pyx_int_0);
  __pyx_t_1 = __pyx_int_0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_schema, __pyx_n_s_items); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_3 = __pyx_t_2; __Pyx_INCREF(__pyx_t_3); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 495, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 495, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 495, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 495, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 495, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 495, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_5(__pyx_t_3);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 495, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_inner_schema, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_1);
    __pyx_t_2 = __Pyx_PyInt_AddObjC(__pyx_t_1, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 495, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1);
    __pyx_t_1 = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "validr/_validator_c.pyx":496
 *     dict_inner = None
 *     for i, inner_schema in enumerate(schema.items):
 *         with mark_index(i):             # <<<<<<<<<<<<<<
 *             if inner_schema.validator == 'union':
 *                 raise SchemaError('ambiguous union schema')
 */
    __Pyx_TraceLine(496,0,__PYX_ERR(0, 496, __pyx_L1_error))
    /*with:*/ {
      __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_mark_index); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 496, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
        }
      }
      __pyx_t_2 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_7, __pyx_v_i) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_i);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 496, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_8 = __Pyx_PyObject_LookupSpecial(__pyx_t_2, __pyx_n_s_exit); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 496, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_7 = __Pyx_PyObject_LookupSpecial(__pyx_t_2, __pyx_n_s_enter); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 496, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_9 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      __pyx_t_6 = (__pyx_t_9) ? __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_9) : __Pyx_PyObject_CallNoArg(__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 496, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      /*try:*/ {
        {
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __Pyx_ExceptionSave(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);
          __Pyx_XGOTREF(__pyx_t_10);
          __Pyx_XGOTREF(__pyx_t_11);
          __Pyx_XGOTREF(__pyx_t_12);
          /*try:*/ {

            /* "validr/_validator_c.pyx":497
 *     for i, inner_schema in enumerate(schema.items):
 *         with mark_index(i):
 *             if inner_schema.validator == 'union':             # <<<<<<<<<<<<<<
 *                 raise SchemaError('ambiguous union schema')
 *             if _optional_or_has_default(inner_schema):
 */
            __Pyx_TraceLine(497,0,__PYX_ERR(0, 497, __pyx_L11_error))
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_inner_schema, __pyx_n_s_validator); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 497, __pyx_L11_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_13 = (__Pyx_PyUnicode_Equals(__pyx_t_2, __pyx_n_u_union, Py_EQ)); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 497, __pyx_L11_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (unlikely(__pyx_t_13)) {

              /* "validr/_validator_c.pyx":498
 *         with mark_index(i):
 *             if inner_schema.validator == 'union':
 *                 raise SchemaError('ambiguous union schema')             # <<<<<<<<<<<<<<
 *             if _optional_or_has_default(inner_schema):
 *                 raise SchemaError('not allowed optional or default for union schemas')
 */
              __Pyx_TraceLine(498,0,__PYX_ERR(0, 498, __pyx_L11_error))
              __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_SchemaError); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 498, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_6);
              __pyx_t_7 = NULL;
              if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
                __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
                if (likely(__pyx_t_7)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
                  __Pyx_INCREF(__pyx_t_7);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_6, function);
                }
              }
              __pyx_t_2 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_7, __pyx_kp_u_ambiguous_union_schema) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_kp_u_ambiguous_union_schema);
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 498, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_Raise(__pyx_t_2, 0, 0, 0);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __PYX_ERR(0, 498, __pyx_L11_error)

              /* "validr/_validator_c.pyx":497
 *     for i, inner_schema in enumerate(schema.items):
 *         with mark_index(i):
 *             if inner_schema.validator == 'union':             # <<<<<<<<<<<<<<
 *                 raise SchemaError('ambiguous union schema')
 *             if _optional_or_has_default(inner_schema):
 */
            }

            /* "validr/_validator_c.pyx":499
 *             if inner_schema.validator == 'union':
 *                 raise SchemaError('ambiguous union schema')
 *             if _optional_or_has_default(inner_schema):             # <<<<<<<<<<<<<<
 *                 raise SchemaError('not allowed optional or default for union schemas')
 *             if schema.params.get('optional'):
 */
            __Pyx_TraceLine(499,0,__PYX_ERR(0, 499, __pyx_L11_error))
            __pyx_t_2 = __pyx_f_6validr_12_validator_c__optional_or_has_default(__pyx_v_inner_schema, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 499, __pyx_L11_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 499, __pyx_L11_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (unlikely(__pyx_t_13)) {

              /* "validr/_validator_c.pyx":500
 *                 raise SchemaError('ambiguous union schema')
 *             if _optional_or_has_default(inner_schema):
 *                 raise SchemaError('not allowed optional or default for union schemas')             # <<<<<<<<<<<<<<
 *             if schema.params.get('optional'):
 *                 inner_schema = inner_schema.copy()
 */
              __Pyx_TraceLine(500,0,__PYX_ERR(0, 500, __pyx_L11_error))
              __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_SchemaError); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 500, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_6);
              __pyx_t_7 = NULL;
              if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
                __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
                if (likely(__pyx_t_7)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
                  __Pyx_INCREF(__pyx_t_7);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_6, function);
                }
              }
              __pyx_t_2 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_7, __pyx_kp_u_not_allowed_optional_or_default) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_kp_u_not_allowed_optional_or_default);
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 500, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_Raise(__pyx_t_2, 0, 0, 0);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __PYX_ERR(0, 500, __pyx_L11_error)

              /* "validr/_validator_c.pyx":499
 *             if inner_schema.validator == 'union':
 *                 raise SchemaError('ambiguous union schema')
 *             if _optional_or_has_default(inner_schema):             # <<<<<<<<<<<<<<
 *                 raise SchemaError('not allowed optional or default for union schemas')
 *             if schema.params.get('optional'):
 */
            }

            /* "validr/_validator_c.pyx":501
 *             if _optional_or_has_default(inner_schema):
 *                 raise SchemaError('not allowed optional or default for union schemas')
 *             if schema.params.get('optional'):             # <<<<<<<<<<<<<<
 *                 inner_schema = inner_schema.copy()
 *                 inner_schema.params['optional'] = True
 */
            __Pyx_TraceLine(501,0,__PYX_ERR(0, 501, __pyx_L11_error))
            __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_schema, __pyx_n_s_params); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 501, __pyx_L11_error)
            __Pyx_GOTREF(__pyx_t_6);
            __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_get); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 501, __pyx_L11_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __pyx_t_6 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
              __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
              if (likely(__pyx_t_6)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
                __Pyx_INCREF(__pyx_t_6);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_7, function);
              }
            }
            __pyx_t_2 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_6, __pyx_n_u_optional) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_n_u_optional);
            __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
            if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 501, __pyx_L11_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 501, __pyx_L11_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (__pyx_t_13) {

              /* "validr/_validator_c.pyx":502
 *                 raise SchemaError('not allowed optional or default for union schemas')
 *             if schema.params.get('optional'):
 *                 inner_schema = inner_schema.copy()             # <<<<<<<<<<<<<<
 *                 inner_schema.params['optional'] = True
 *             if inner_schema.validator == 'list':
 */
              __Pyx_TraceLine(502,0,__PYX_ERR(0, 502, __pyx_L11_error))
              __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_inner_schema, __pyx_n_s_copy); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 502, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_7);
              __pyx_t_6 = NULL;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
                __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
                if (likely(__pyx_t_6)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
                  __Pyx_INCREF(__pyx_t_6);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_7, function);
                }
              }
              __pyx_t_2 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_7);
              __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
              if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 502, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_DECREF_SET(__pyx_v_inner_schema, __pyx_t_2);
              __pyx_t_2 = 0;

              /* "validr/_validator_c.pyx":503
 *             if schema.params.get('optional'):
 *                 inner_schema = inner_schema.copy()
 *                 inner_schema.params['optional'] = True             # <<<<<<<<<<<<<<
 *             if inner_schema.validator == 'list':
 *                 if list_inner is not None:
 */
              __Pyx_TraceLine(503,0,__PYX_ERR(0, 503, __pyx_L11_error))
              __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_inner_schema, __pyx_n_s_params); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 503, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_2);
              if (unlikely(PyObject_SetItem(__pyx_t_2, __pyx_n_u_optional, Py_True) < 0)) __PYX_ERR(0, 503, __pyx_L11_error)
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

              /* "validr/_validator_c.pyx":501
 *             if _optional_or_has_default(inner_schema):
 *                 raise SchemaError('not allowed optional or default for union schemas')
 *             if schema.params.get('optional'):             # <<<<<<<<<<<<<<
 *                 inner_schema = inner_schema.copy()
 *                 inner_schema.params['optional'] = True
 */
            }

            /* "validr/_validator_c.pyx":504
 *                 inner_schema = inner_schema.copy()
 *                 inner_schema.params['optional'] = True
 *             if inner_schema.validator == 'list':             # <<<<<<<<<<<<<<
 *                 if list_inner is not None:
 *                     raise SchemaError('ambiguous union schema')
 */
            __Pyx_TraceLine(504,0,__PYX_ERR(0, 504, __pyx_L11_error))
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_inner_schema, __pyx_n_s_validator); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 504, __pyx_L11_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_13 = (__Pyx_PyUnicode_Equals(__pyx_t_2, __pyx_n_u_list, Py_EQ)); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 504, __pyx_L11_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (__pyx_t_13) {

              /* "validr/_validator_c.pyx":505
 *                 inner_schema.params['optional'] = True
 *             if inner_schema.validator == 'list':
 *                 if list_inner is not None:             # <<<<<<<<<<<<<<
 *                     raise SchemaError('ambiguous union schema')
 *                 list_inner = compiler.compile(inner_schema)
 */
              __Pyx_TraceLine(505,0,__PYX_ERR(0, 505, __pyx_L11_error))
              __pyx_t_13 = (__pyx_cur_scope->__pyx_v_list_inner != Py_None);
              __pyx_t_14 = (__pyx_t_13 != 0);
              if (unlikely(__pyx_t_14)) {

                /* "validr/_validator_c.pyx":506
 *             if inner_schema.validator == 'list':
 *                 if list_inner is not None:
 *                     raise SchemaError('ambiguous union schema')             # <<<<<<<<<<<<<<
 *                 list_inner = compiler.compile(inner_schema)
 *             elif inner_schema.validator in ('dict', 'model'):
 */
                __Pyx_TraceLine(506,0,__PYX_ERR(0, 506, __pyx_L11_error))
                __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_SchemaError); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 506, __pyx_L11_error)
                __Pyx_GOTREF(__pyx_t_7);
                __pyx_t_6 = NULL;
                if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
                  __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
                  if (likely(__pyx_t_6)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
                    __Pyx_INCREF(__pyx_t_6);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_7, function);
                  }
                }
                __pyx_t_2 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_6, __pyx_kp_u_ambiguous_union_schema) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_kp_u_ambiguous_union_schema);
                __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
                if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 506, __pyx_L11_error)
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
                __Pyx_Raise(__pyx_t_2, 0, 0, 0);
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                __PYX_ERR(0, 506, __pyx_L11_error)

                /* "validr/_validator_c.pyx":505
 *                 inner_schema.params['optional'] = True
 *             if inner_schema.validator == 'list':
 *                 if list_inner is not None:             # <<<<<<<<<<<<<<
 *                     raise SchemaError('ambiguous union schema')
 *                 list_inner = compiler.compile(inner_schema)
 */
              }

              /* "validr/_validator_c.pyx":507
 *                 if list_inner is not None:
 *                     raise SchemaError('ambiguous union schema')
 *                 list_inner = compiler.compile(inner_schema)             # <<<<<<<<<<<<<<
 *             elif inner_schema.validator in ('dict', 'model'):
 *                 if dict_inner is not None:
 */
              __Pyx_TraceLine(507,0,__PYX_ERR(0, 507, __pyx_L11_error))
              __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_compiler, __pyx_n_s_compile); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 507, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_7);
              __pyx_t_6 = NULL;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
                __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
                if (likely(__pyx_t_6)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
                  __Pyx_INCREF(__pyx_t_6);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_7, function);
                }
              }
              __pyx_t_2 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_6, __pyx_v_inner_schema) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_inner_schema);
              __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
              if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 507, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_list_inner);
              __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_list_inner, __pyx_t_2);
              __Pyx_GIVEREF(__pyx_t_2);
              __pyx_t_2 = 0;

              /* "validr/_validator_c.pyx":504
 *                 inner_schema = inner_schema.copy()
 *                 inner_schema.params['optional'] = True
 *             if inner_schema.validator == 'list':             # <<<<<<<<<<<<<<
 *                 if list_inner is not None:
 *                     raise SchemaError('ambiguous union schema')
 */
              goto __pyx_L22;
            }

            /* "validr/_validator_c.pyx":508
 *                     raise SchemaError('ambiguous union schema')
 *                 list_inner = compiler.compile(inner_schema)
 *             elif inner_schema.validator in ('dict', 'model'):             # <<<<<<<<<<<<<<
 *                 if dict_inner is not None:
 *                     raise SchemaError('ambiguous union schema')
 */
            __Pyx_TraceLine(508,0,__PYX_ERR(0, 508, __pyx_L11_error))
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_inner_schema, __pyx_n_s_validator); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 508, __pyx_L11_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_13 = (__Pyx_PyUnicode_Equals(__pyx_t_2, __pyx_n_u_dict, Py_EQ)); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 508, __pyx_L11_error)
            if (!__pyx_t_13) {
            } else {
              __pyx_t_14 = __pyx_t_13;
              goto __pyx_L24_bool_binop_done;
            }
            __pyx_t_13 = (__Pyx_PyUnicode_Equals(__pyx_t_2, __pyx_n_u_model, Py_EQ)); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 508, __pyx_L11_error)
            __pyx_t_14 = __pyx_t_13;
            __pyx_L24_bool_binop_done:;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_t_13 = (__pyx_t_14 != 0);
            if (__pyx_t_13) {

              /* "validr/_validator_c.pyx":509
 *                 list_inner = compiler.compile(inner_schema)
 *             elif inner_schema.validator in ('dict', 'model'):
 *                 if dict_inner is not None:             # <<<<<<<<<<<<<<
 *                     raise SchemaError('ambiguous union schema')
 *                 dict_inner = compiler.compile(inner_schema)
 */
              __Pyx_TraceLine(509,0,__PYX_ERR(0, 509, __pyx_L11_error))
              __pyx_t_13 = (__pyx_cur_scope->__pyx_v_dict_inner != Py_None);
              __pyx_t_14 = (__pyx_t_13 != 0);
              if (unlikely(__pyx_t_14)) {

                /* "validr/_validator_c.pyx":510
 *             elif inner_schema.validator in ('dict', 'model'):
 *                 if dict_inner is not None:
 *                     raise SchemaError('ambiguous union schema')             # <<<<<<<<<<<<<<
 *                 dict_inner = compiler.compile(inner_schema)
 *             else:
 */
                __Pyx_TraceLine(510,0,__PYX_ERR(0, 510, __pyx_L11_error))
                __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_SchemaError); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 510, __pyx_L11_error)
                __Pyx_GOTREF(__pyx_t_7);
                __pyx_t_6 = NULL;
                if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
                  __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
                  if (likely(__pyx_t_6)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
                    __Pyx_INCREF(__pyx_t_6);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_7, function);
                  }
                }
                __pyx_t_2 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_6, __pyx_kp_u_ambiguous_union_schema) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_kp_u_ambiguous_union_schema);
                __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
                if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 510, __pyx_L11_error)
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
                __Pyx_Raise(__pyx_t_2, 0, 0, 0);
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                __PYX_ERR(0, 510, __pyx_L11_error)

                /* "validr/_validator_c.pyx":509
 *                 list_inner = compiler.compile(inner_schema)
 *             elif inner_schema.validator in ('dict', 'model'):
 *                 if dict_inner is not None:             # <<<<<<<<<<<<<<
 *                     raise SchemaError('ambiguous union schema')
 *                 dict_inner = compiler.compile(inner_schema)
 */
              }

              /* "validr/_validator_c.pyx":511
 *                 if dict_inner is not None:
 *                     raise SchemaError('ambiguous union schema')
 *                 dict_inner = compiler.compile(inner_schema)             # <<<<<<<<<<<<<<
 *             else:
 *                 if scalar_inner is not None:
 */
              __Pyx_TraceLine(511,0,__PYX_ERR(0, 511, __pyx_L11_error))
              __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_compiler, __pyx_n_s_compile); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 511, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_7);
              __pyx_t_6 = NULL;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
                __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
                if (likely(__pyx_t_6)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
                  __Pyx_INCREF(__pyx_t_6);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_7, function);
                }
              }
              __pyx_t_2 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_6, __pyx_v_inner_schema) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_inner_schema);
              __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
              if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 511, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_dict_inner);
              __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_dict_inner, __pyx_t_2);
              __Pyx_GIVEREF(__pyx_t_2);
              __pyx_t_2 = 0;

              /* "validr/_validator_c.pyx":508
 *                     raise SchemaError('ambiguous union schema')
 *                 list_inner = compiler.compile(inner_schema)
 *             elif inner_schema.validator in ('dict', 'model'):             # <<<<<<<<<<<<<<
 *                 if dict_inner is not None:
 *                     raise SchemaError('ambiguous union schema')
 */
              goto __pyx_L22;
            }

            /* "validr/_validator_c.pyx":513
 *                 dict_inner = compiler.compile(inner_schema)
 *             else:
 *                 if scalar_inner is not None:             # <<<<<<<<<<<<<<
 *                     raise SchemaError('ambiguous union schema')
 *                 scalar_inner = compiler.compile(inner_schema)
 */
            __Pyx_TraceLine(513,0,__PYX_ERR(0, 513, __pyx_L11_error))
            /*else*/ {
              __pyx_t_14 = (__pyx_cur_scope->__pyx_v_scalar_inner != Py_None);
              __pyx_t_13 = (__pyx_t_14 != 0);
              if (unlikely(__pyx_t_13)) {

                /* "validr/_validator_c.pyx":514
 *             else:
 *                 if scalar_inner is not None:
 *                     raise SchemaError('ambiguous union schema')             # <<<<<<<<<<<<<<
 *                 scalar_inner = compiler.compile(inner_schema)
 * 
 */
                __Pyx_TraceLine(514,0,__PYX_ERR(0, 514, __pyx_L11_error))
                __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_SchemaError); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 514, __pyx_L11_error)
                __Pyx_GOTREF(__pyx_t_7);
                __pyx_t_6 = NULL;
                if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
                  __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
                  if (likely(__pyx_t_6)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
                    __Pyx_INCREF(__pyx_t_6);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_7, function);
                  }
                }
                __pyx_t_2 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_6, __pyx_kp_u_ambiguous_union_schema) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_kp_u_ambiguous_union_schema);
                __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
                if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 514, __pyx_L11_error)
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
                __Pyx_Raise(__pyx_t_2, 0, 0, 0);
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                __PYX_ERR(0, 514, __pyx_L11_error)

                /* "validr/_validator_c.pyx":513
 *                 dict_inner = compiler.compile(inner_schema)
 *             else:
 *                 if scalar_inner is not None:             # <<<<<<<<<<<<<<
 *                     raise SchemaError('ambiguous union schema')
 *                 scalar_inner = compiler.compile(inner_schema)
 */
              }

              /* "validr/_validator_c.pyx":515
 *                 if scalar_inner is not None:
 *                     raise SchemaError('ambiguous union schema')
 *                 scalar_inner = compiler.compile(inner_schema)             # <<<<<<<<<<<<<<
 * 
 *     def validate(value):
 */
              __Pyx_TraceLine(515,0,__PYX_ERR(0, 515, __pyx_L11_error))
              __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_compiler, __pyx_n_s_compile); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 515, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_7);
              __pyx_t_6 = NULL;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
                __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
                if (likely(__pyx_t_6)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
                  __Pyx_INCREF(__pyx_t_6);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_7, function);
                }
              }
              __pyx_t_2 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_6, __pyx_v_inner_schema) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_inner_schema);
              __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
              if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 515, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_scalar_inner);
              __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_scalar_inner, __pyx_t_2);
              __Pyx_GIVEREF(__pyx_t_2);
              __pyx_t_2 = 0;
            }
            __pyx_L22:;

            /* "validr/_validator_c.pyx":496
 *     dict_inner = None
 *     for i, inner_schema in enumerate(schema.items):
 *         with mark_index(i):             # <<<<<<<<<<<<<<
 *             if inner_schema.validator == 'union':
 *                 raise SchemaError('ambiguous union schema')
 */
          }
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          goto __pyx_L18_try_end;
          __pyx_L11_error:;
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          /*except:*/ {
            __Pyx_AddTraceback("validr._validator_c._union_list_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
            if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_7, &__pyx_t_6) < 0) __PYX_ERR(0, 496, __pyx_L13_except_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_GOTREF(__pyx_t_6);
            __pyx_t_9 = PyTuple_Pack(3, __pyx_t_2, __pyx_t_7, __pyx_t_6); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 496, __pyx_L13_except_error)
            __Pyx_GOTREF(__pyx_t_9);
            __pyx_t_15 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_9, NULL);
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
            if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 496, __pyx_L13_except_error)
            __Pyx_GOTREF(__pyx_t_15);
            __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_15);
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            if (__pyx_t_13 < 0) __PYX_ERR(0, 496, __pyx_L13_except_error)
            __pyx_t_14 = ((!(__pyx_t_13 != 0)) != 0);
            if (__pyx_t_14) {
              __Pyx_GIVEREF(__pyx_t_2);
              __Pyx_GIVEREF(__pyx_t_7);
              __Pyx_XGIVEREF(__pyx_t_6);
              __Pyx_ErrRestoreWithState(__pyx_t_2, __pyx_t_7, __pyx_t_6);
              __pyx_t_2 = 0; __pyx_t_7 = 0; __pyx_t_6 = 0; 
              __PYX_ERR(0, 496, __pyx_L13_except_error)
            }
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
            goto __pyx_L12_exception_handled;
          }
          __pyx_L13_except_error:;
          __Pyx_XGIVEREF(__pyx_t_10);
          __Pyx_XGIVEREF(__pyx_t_11);
          __Pyx_XGIVEREF(__pyx_t_12);
          __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
          goto __pyx_L1_error;
          __pyx_L12_exception_handled:;
          __Pyx_XGIVEREF(__pyx_t_10);
          __Pyx_XGIVEREF(__pyx_t_11);
          __Pyx_XGIVEREF(__pyx_t_12);
          __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
          __pyx_L18_try_end:;
        }
      }
      /*finally:*/ {
        /*normal exit:*/{
          if (__pyx_t_8) {
            __pyx_t_12 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_tuple__28, NULL);
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 496, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_12);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          }
          goto __pyx_L10;
        }
        __pyx_L10:;
      }
      goto __pyx_L31;
      __pyx_L5_error:;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      goto __pyx_L1_error;
      __pyx_L31:;
    }

    /* "validr/_validator_c.pyx":495
 *     list_inner = None
 *     dict_inner = None
 *     for i, inner_schema in enumerate(schema.items):             # <<<<<<<<<<<<<<
 *         with mark_index(i):
 *             if inner_schema.validator == 'union':
 */
    __Pyx_TraceLine(495,0,__PYX_ERR(0, 495, __pyx_L1_error))
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":517
 *                 scalar_inner = compiler.compile(inner_schema)
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         if isinstance(value, list):
 *             if list_inner is None:
 */
  __Pyx_TraceLine(517,0,__PYX_ERR(0, 517, __pyx_L1_error))
  __pyx_t_1 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_21_union_list_validator_1validate, 0, __pyx_n_s_union_list_validator_locals_val, ((PyObject*)__pyx_cur_scope), __pyx_n_s_validr__validator_c, __pyx_d, ((PyObject *)__pyx_codeobj__46)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 517, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_validate = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":533
 *             return scalar_inner(value)
 * 
 *     _update_validate_func_info(validate, union_validator, schema)             # <<<<<<<<<<<<<<
 * 
 *     return validate
 */
  __Pyx_TraceLine(533,0,__PYX_ERR(0, 533, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_union_validator); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __pyx_f_6validr_12_validator_c__update_validate_func_info(__pyx_v_validate, __pyx_t_1, __pyx_v_schema, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "validr/_validator_c.pyx":535
 *     _update_validate_func_info(validate, union_validator, schema)
 * 
 *     return validate             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(535,0,__PYX_ERR(0, 535, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_validate);
  __pyx_r = __pyx_v_validate;
  goto __pyx_L0;

  /* "validr/_validator_c.pyx":491
 * 
 * 
 * def _union_list_validator(compiler, schema):             # <<<<<<<<<<<<<<
 *     scalar_inner = None
 *     list_inner = None
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("validr._validator_c._union_list_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_inner_schema);
  __Pyx_XDECREF(__pyx_v_validate);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":539
 * 
 * @validator(accept=object, output=object)
 * def _union_dict_validator(compiler, items, str by):             # <<<<<<<<<<<<<<
 *     inners = {}
 *     for key, schema in items.items():
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_39_union_dict_validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6validr_12_validator_c_39_union_dict_validator = {"_union_dict_validator", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6validr_12_validator_c_39_union_dict_validator, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6validr_12_validator_c_39_union_dict_validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_compiler = 0;
  PyObject *__pyx_v_items = 0;
  PyObject *__pyx_v_by = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_union_dict_validator (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_compiler,&__pyx_n_s_items,&__pyx_n_s_by,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_compiler)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_items)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_union_dict_validator", 1, 3, 3, 1); __PYX_ERR(0, 539, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_by)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_union_dict_validator", 1, 3, 3, 2); __PYX_ERR(0, 539, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_union_dict_validator") < 0)) __PYX_ERR(0, 539, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_compiler = values[0];
    __pyx_v_items = values[1];
    __pyx_v_by = ((PyObject*)values[2]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_union_dict_validator", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 539, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("validr._validator_c._union_dict_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_by), (&PyUnicode_Type), 1, "by", 1))) __PYX_ERR(0, 539, __pyx_L1_error)
  __pyx_r = __pyx_pf_6validr_12_validator_c_38_union_dict_validator(__pyx_self, __pyx_v_compiler, __pyx_v_items, __pyx_v_by);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":552
 *     del key, schema, is_model
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         if is_dict(value):
 *             getter = get_dict_value
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_21_union_dict_validator_1validate(PyObject *__pyx_self, PyObject *__pyx_v_value); /*proto*/
static PyMethodDef __pyx_mdef_6validr_12_validator_c_21_union_dict_validator_1validate = {"validate", (PyCFunction)__pyx_pw_6validr_12_validator_c_21_union_dict_validator_1validate, METH_O, 0};
static PyObject *__pyx_pw_6validr_12_validator_c_21_union_dict_validator_1validate(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("validate (wrapper)", 0);
  __pyx_r = __pyx_pf_6validr_12_validator_c_21_union_dict_validator_validate(__pyx_self, ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6validr_12_validator_c_21_union_dict_validator_validate(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_13__union_dict_validator *__pyx_cur_scope;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_13__union_dict_validator *__pyx_outer_scope;
  PyObject *(*__pyx_v_getter)(PyObject *, PyObject *, int __pyx_skip_dispatch);
  PyObject *__pyx_v_by_name = 0;
  PyObject *__pyx_v_inner_info = NULL;
  PyObject *__pyx_v_err_msg = NULL;
  PyObject *__pyx_v_is_model = NULL;
  PyObject *__pyx_v_inner = NULL;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *(*__pyx_t_12)(PyObject *);
  __Pyx_RefNannySetupContext("validate", 0);
  __pyx_outer_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_13__union_dict_validator *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_TraceCall("validate", __pyx_f[0], 552, 0, __PYX_ERR(0, 552, __pyx_L1_error));

  /* "validr/_validator_c.pyx":553
 * 
 *     def validate(value):
 *         if is_dict(value):             # <<<<<<<<<<<<<<
 *             getter = get_dict_value
 *         else:
 */
  __Pyx_TraceLine(553,0,__PYX_ERR(0, 553, __pyx_L1_error))
  __pyx_t_1 = (__pyx_f_6validr_12_validator_c_is_dict(__pyx_v_value, 0) != 0);
  if (__pyx_t_1) {

    /* "validr/_validator_c.pyx":554
 *     def validate(value):
 *         if is_dict(value):
 *             getter = get_dict_value             # <<<<<<<<<<<<<<
 *         else:
 *             getter = get_object_value
 */
    __Pyx_TraceLine(554,0,__PYX_ERR(0, 554, __pyx_L1_error))
    __pyx_v_getter = __pyx_f_6validr_12_validator_c_get_dict_value;

    /* "validr/_validator_c.pyx":553
 * 
 *     def validate(value):
 *         if is_dict(value):             # <<<<<<<<<<<<<<
 *             getter = get_dict_value
 *         else:
 */
    goto __pyx_L3;
  }

  /* "validr/_validator_c.pyx":556
 *             getter = get_dict_value
 *         else:
 *             getter = get_object_value             # <<<<<<<<<<<<<<
 *         cdef str by_name
 *         with mark_key(by):
 */
  __Pyx_TraceLine(556,0,__PYX_ERR(0, 556, __pyx_L1_error))
  /*else*/ {
    __pyx_v_getter = __pyx_f_6validr_12_validator_c_get_object_value;
  }
  __pyx_L3:;

  /* "validr/_validator_c.pyx":558
 *             getter = get_object_value
 *         cdef str by_name
 *         with mark_key(by):             # <<<<<<<<<<<<<<
 *             by_name = getter(value, by)
 *             if not by_name:
 */
  __Pyx_TraceLine(558,0,__PYX_ERR(0, 558, __pyx_L1_error))
  /*with:*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_mark_key); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 558, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (unlikely(!__pyx_cur_scope->__pyx_v_by)) { __Pyx_RaiseClosureNameError("by"); __PYX_ERR(0, 558, __pyx_L1_error) }
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_cur_scope->__pyx_v_by) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_cur_scope->__pyx_v_by);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 558, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_5 = __Pyx_PyObject_LookupSpecial(__pyx_t_2, __pyx_n_s_exit); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 558, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = __Pyx_PyObject_LookupSpecial(__pyx_t_2, __pyx_n_s_enter); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 558, __pyx_L4_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 558, __pyx_L4_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    /*try:*/ {
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_9);
        /*try:*/ {

          /* "validr/_validator_c.pyx":559
 *         cdef str by_name
 *         with mark_key(by):
 *             by_name = getter(value, by)             # <<<<<<<<<<<<<<
 *             if not by_name:
 *                 raise Invalid('required', value=by_name)
 */
          __Pyx_TraceLine(559,0,__PYX_ERR(0, 559, __pyx_L8_error))
          if (unlikely(!__pyx_cur_scope->__pyx_v_by)) { __Pyx_RaiseClosureNameError("by"); __PYX_ERR(0, 559, __pyx_L8_error) }
          __pyx_t_2 = __pyx_cur_scope->__pyx_v_by;
          __Pyx_INCREF(__pyx_t_2);
          __pyx_t_3 = __pyx_v_getter(__pyx_v_value, ((PyObject*)__pyx_t_2), 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 559, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          if (!(likely(PyUnicode_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(0, 559, __pyx_L8_error)
          __pyx_v_by_name = ((PyObject*)__pyx_t_3);
          __pyx_t_3 = 0;

          /* "validr/_validator_c.pyx":560
 *         with mark_key(by):
 *             by_name = getter(value, by)
 *             if not by_name:             # <<<<<<<<<<<<<<
 *                 raise Invalid('required', value=by_name)
 *             inner_info = inners.get(by_name)
 */
          __Pyx_TraceLine(560,0,__PYX_ERR(0, 560, __pyx_L8_error))
          __pyx_t_1 = (__pyx_v_by_name != Py_None)&&(__Pyx_PyUnicode_IS_TRUE(__pyx_v_by_name) != 0);
          __pyx_t_10 = ((!__pyx_t_1) != 0);
          if (unlikely(__pyx_t_10)) {

            /* "validr/_validator_c.pyx":561
 *             by_name = getter(value, by)
 *             if not by_name:
 *                 raise Invalid('required', value=by_name)             # <<<<<<<<<<<<<<
 *             inner_info = inners.get(by_name)
 *             if inner_info is None:
 */
            __Pyx_TraceLine(561,0,__PYX_ERR(0, 561, __pyx_L8_error))
            __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 561, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 561, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_2);
            if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_value, __pyx_v_by_name) < 0) __PYX_ERR(0, 561, __pyx_L8_error)
            __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__47, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 561, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_Raise(__pyx_t_4, 0, 0, 0);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __PYX_ERR(0, 561, __pyx_L8_error)

            /* "validr/_validator_c.pyx":560
 *         with mark_key(by):
 *             by_name = getter(value, by)
 *             if not by_name:             # <<<<<<<<<<<<<<
 *                 raise Invalid('required', value=by_name)
 *             inner_info = inners.get(by_name)
 */
          }

          /* "validr/_validator_c.pyx":562
 *             if not by_name:
 *                 raise Invalid('required', value=by_name)
 *             inner_info = inners.get(by_name)             # <<<<<<<<<<<<<<
 *             if inner_info is None:
 *                 err_msg = 'expect one of {}'.format(expect_bys)
 */
          __Pyx_TraceLine(562,0,__PYX_ERR(0, 562, __pyx_L8_error))
          if (unlikely(!__pyx_cur_scope->__pyx_v_inners)) { __Pyx_RaiseClosureNameError("inners"); __PYX_ERR(0, 562, __pyx_L8_error) }
          if (unlikely(__pyx_cur_scope->__pyx_v_inners == Py_None)) {
            PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "get");
            __PYX_ERR(0, 562, __pyx_L8_error)
          }
          __pyx_t_4 = __Pyx_PyDict_GetItemDefault(__pyx_cur_scope->__pyx_v_inners, __pyx_v_by_name, Py_None); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 562, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_v_inner_info = __pyx_t_4;
          __pyx_t_4 = 0;

          /* "validr/_validator_c.pyx":563
 *                 raise Invalid('required', value=by_name)
 *             inner_info = inners.get(by_name)
 *             if inner_info is None:             # <<<<<<<<<<<<<<
 *                 err_msg = 'expect one of {}'.format(expect_bys)
 *                 raise Invalid(err_msg, value=by_name)
 */
          __Pyx_TraceLine(563,0,__PYX_ERR(0, 563, __pyx_L8_error))
          __pyx_t_10 = (__pyx_v_inner_info == Py_None);
          __pyx_t_1 = (__pyx_t_10 != 0);
          if (unlikely(__pyx_t_1)) {

            /* "validr/_validator_c.pyx":564
 *             inner_info = inners.get(by_name)
 *             if inner_info is None:
 *                 err_msg = 'expect one of {}'.format(expect_bys)             # <<<<<<<<<<<<<<
 *                 raise Invalid(err_msg, value=by_name)
 *         is_model, inner = inner_info
 */
            __Pyx_TraceLine(564,0,__PYX_ERR(0, 564, __pyx_L8_error))
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_expect_one_of, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 564, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_2);
            if (unlikely(!__pyx_cur_scope->__pyx_v_expect_bys)) { __Pyx_RaiseClosureNameError("expect_bys"); __PYX_ERR(0, 564, __pyx_L8_error) }
            __pyx_t_3 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
              __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
              if (likely(__pyx_t_3)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                __Pyx_INCREF(__pyx_t_3);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_2, function);
              }
            }
            __pyx_t_4 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_cur_scope->__pyx_v_expect_bys) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_cur_scope->__pyx_v_expect_bys);
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 564, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_v_err_msg = __pyx_t_4;
            __pyx_t_4 = 0;

            /* "validr/_validator_c.pyx":565
 *             if inner_info is None:
 *                 err_msg = 'expect one of {}'.format(expect_bys)
 *                 raise Invalid(err_msg, value=by_name)             # <<<<<<<<<<<<<<
 *         is_model, inner = inner_info
 *         result = inner(value)
 */
            __Pyx_TraceLine(565,0,__PYX_ERR(0, 565, __pyx_L8_error))
            __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 565, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 565, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_INCREF(__pyx_v_err_msg);
            __Pyx_GIVEREF(__pyx_v_err_msg);
            PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_err_msg);
            __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 565, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_value, __pyx_v_by_name) < 0) __PYX_ERR(0, 565, __pyx_L8_error)
            __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 565, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_6);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_Raise(__pyx_t_6, 0, 0, 0);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __PYX_ERR(0, 565, __pyx_L8_error)

            /* "validr/_validator_c.pyx":563
 *                 raise Invalid('required', value=by_name)
 *             inner_info = inners.get(by_name)
 *             if inner_info is None:             # <<<<<<<<<<<<<<
 *                 err_msg = 'expect one of {}'.format(expect_bys)
 *                 raise Invalid(err_msg, value=by_name)
 */
          }

          /* "validr/_validator_c.pyx":558
 *             getter = get_object_value
 *         cdef str by_name
 *         with mark_key(by):             # <<<<<<<<<<<<<<
 *             by_name = getter(value, by)
 *             if not by_name:
 */
        }
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L13_try_end;
        __pyx_L8_error:;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        /*except:*/ {
          __Pyx_AddTraceback("validr._validator_c._union_dict_validator.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_3, &__pyx_t_2) < 0) __PYX_ERR(0, 558, __pyx_L10_except_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_4 = PyTuple_Pack(3, __pyx_t_6, __pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 558, __pyx_L10_except_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_4, NULL);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 558, __pyx_L10_except_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          if (__pyx_t_1 < 0) __PYX_ERR(0, 558, __pyx_L10_except_error)
          __pyx_t_10 = ((!(__pyx_t_1 != 0)) != 0);
          if (__pyx_t_10) {
            __Pyx_GIVEREF(__pyx_t_6);
            __Pyx_GIVEREF(__pyx_t_3);
            __Pyx_XGIVEREF(__pyx_t_2);
            __Pyx_ErrRestoreWithState(__pyx_t_6, __pyx_t_3, __pyx_t_2);
            __pyx_t_6 = 0; __pyx_t_3 = 0; __pyx_t_2 = 0; 
            __PYX_ERR(0, 558, __pyx_L10_except_error)
          }
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          goto __pyx_L9_exception_handled;
        }
        __pyx_L10_except_error:;
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
        goto __pyx_L1_error;
        __pyx_L9_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
        __pyx_L13_try_end:;
      }
    }
    /*finally:*/ {
      /*normal exit:*/{
        if (__pyx_t_5) {
          __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_tuple__28, NULL);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 558, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        }
        goto __pyx_L7;
      }
      __pyx_L7:;
    }
    goto __pyx_L19;
    __pyx_L4_error:;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L1_error;
    __pyx_L19:;
  }

  /* "validr/_validator_c.pyx":566
 *                 err_msg = 'expect one of {}'.format(expect_bys)
 *                 raise Invalid(err_msg, value=by_name)
 *         is_model, inner = inner_info             # <<<<<<<<<<<<<<
 *         result = inner(value)
 *         if not is_model:
 */
  __Pyx_TraceLine(566,0,__PYX_ERR(0, 566, __pyx_L1_error))
  if (unlikely(!__pyx_v_inner_info)) { __Pyx_RaiseUnboundLocalError("inner_info"); __PYX_ERR(0, 566, __pyx_L1_error) }
  if ((likely(PyTuple_CheckExact(__pyx_v_inner_info))) || (PyList_CheckExact(__pyx_v_inner_info))) {
    PyObject* sequence = __pyx_v_inner_info;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 566, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    #else
    __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 566, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 566, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
  } else {
    Py_ssize_t index = -1;
    __pyx_t_6 = PyObject_GetIter(__pyx_v_inner_info); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 566, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_12 = Py_TYPE(__pyx_t_6)->tp_iternext;
    index = 0; __pyx_t_2 = __pyx_t_12(__pyx_t_6); if (unlikely(!__pyx_t_2)) goto __pyx_L20_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 1; __pyx_t_3 = __pyx_t_12(__pyx_t_6); if (unlikely(!__pyx_t_3)) goto __pyx_L20_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_6), 2) < 0) __PYX_ERR(0, 566, __pyx_L1_error)
    __pyx_t_12 = NULL;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    goto __pyx_L21_unpacking_done;
    __pyx_L20_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_12 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 566, __pyx_L1_error)
    __pyx_L21_unpacking_done:;
  }
  __pyx_v_is_model = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_v_inner = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "validr/_validator_c.pyx":567
 *                 raise Invalid(err_msg, value=by_name)
 *         is_model, inner = inner_info
 *         result = inner(value)             # <<<<<<<<<<<<<<
 *         if not is_model:
 *             result[by] = by_name
 */
  __Pyx_TraceLine(567,0,__PYX_ERR(0, 567, __pyx_L1_error))
  __Pyx_INCREF(__pyx_v_inner);
  __pyx_t_2 = __pyx_v_inner; __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_6, __pyx_v_value) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_value);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "validr/_validator_c.pyx":568
 *         is_model, inner = inner_info
 *         result = inner(value)
 *         if not is_model:             # <<<<<<<<<<<<<<
 *             result[by] = by_name
 *         return result
 */
  __Pyx_TraceLine(568,0,__PYX_ERR(0, 568, __pyx_L1_error))
  __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_v_is_model); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 568, __pyx_L1_error)
  __pyx_t_1 = ((!__pyx_t_10) != 0);
  if (__pyx_t_1) {

    /* "validr/_validator_c.pyx":569
 *         result = inner(value)
 *         if not is_model:
 *             result[by] = by_name             # <<<<<<<<<<<<<<
 *         return result
 * 
 */
    __Pyx_TraceLine(569,0,__PYX_ERR(0, 569, __pyx_L1_error))
    if (unlikely(!__pyx_v_by_name)) { __Pyx_RaiseUnboundLocalError("by_name"); __PYX_ERR(0, 569, __pyx_L1_error) }
    if (unlikely(!__pyx_cur_scope->__pyx_v_by)) { __Pyx_RaiseClosureNameError("by"); __PYX_ERR(0, 569, __pyx_L1_error) }
    if (unlikely(PyObject_SetItem(__pyx_v_result, __pyx_cur_scope->__pyx_v_by, __pyx_v_by_name) < 0)) __PYX_ERR(0, 569, __pyx_L1_error)

    /* "validr/_validator_c.pyx":568
 *         is_model, inner = inner_info
 *         result = inner(value)
 *         if not is_model:             # <<<<<<<<<<<<<<
 *             result[by] = by_name
 *         return result
 */
  }

  /* "validr/_validator_c.pyx":570
 *         if not is_model:
 *             result[by] = by_name
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     return validate
 */
  __Pyx_TraceLine(570,0,__PYX_ERR(0, 570, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "validr/_validator_c.pyx":552
 *     del key, schema, is_model
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         if is_dict(value):
 *             getter = get_dict_value
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("validr._validator_c._union_dict_validator.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_by_name);
  __Pyx_XDECREF(__pyx_v_inner_info);
  __Pyx_XDECREF(__pyx_v_err_msg);
  __Pyx_XDECREF(__pyx_v_is_model);
  __Pyx_XDECREF(__pyx_v_inner);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":539
 * 
 * @validator(accept=object, output=object)
 * def _union_dict_validator(compiler, items, str by):             # <<<<<<<<<<<<<<
 *     inners = {}
 *     for key, schema in items.items():
 */

static PyObject *__pyx_pf_6validr_12_validator_c_38_union_dict_validator(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_compiler, PyObject *__pyx_v_items, PyObject *__pyx_v_by) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_13__union_dict_validator *__pyx_cur_scope;
  PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_schema = NULL;
  PyObject *__pyx_v_is_model = NULL;
  PyObject *__pyx_v_validate = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  int __pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  int __pyx_t_17;
  __Pyx_TraceFrameInit(__pyx_codeobj__48)
  __Pyx_RefNannySetupContext("_union_dict_validator", 0);
  __pyx_cur_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_13__union_dict_validator *)__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_13__union_dict_validator(__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_13__union_dict_validator, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_13__union_dict_validator *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 539, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __Pyx_TraceCall("_union_dict_validator", __pyx_f[0], 539, 0, __PYX_ERR(0, 539, __pyx_L1_error));
  __pyx_cur_scope->__pyx_v_by = __pyx_v_by;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_by);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_by);

  /* "validr/_validator_c.pyx":540
 * @validator(accept=object, output=object)
 * def _union_dict_validator(compiler, items, str by):
 *     inners = {}             # <<<<<<<<<<<<<<
 *     for key, schema in items.items():
 *         with mark_key(key):
 */
  __Pyx_TraceLine(540,0,__PYX_ERR(0, 540, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_inners = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":541
 * def _union_dict_validator(compiler, items, str by):
 *     inners = {}
 *     for key, schema in items.items():             # <<<<<<<<<<<<<<
 *         with mark_key(key):
 *             if schema.validator not in ('dict', 'model'):
 */
  __Pyx_TraceLine(541,0,__PYX_ERR(0, 541, __pyx_L1_error))
  __pyx_t_2 = 0;
  if (unlikely(__pyx_v_items == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
    __PYX_ERR(0, 541, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_dict_iterator(__pyx_v_items, 0, __pyx_n_s_items, (&__pyx_t_3), (&__pyx_t_4)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 541, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_5;
  __pyx_t_5 = 0;
  while (1) {
    __pyx_t_7 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_3, &__pyx_t_2, &__pyx_t_5, &__pyx_t_6, NULL, __pyx_t_4);
    if (unlikely(__pyx_t_7 == 0)) break;
    if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 541, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_XDECREF_SET(__pyx_v_schema, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "validr/_validator_c.pyx":542
 *     inners = {}
 *     for key, schema in items.items():
 *         with mark_key(key):             # <<<<<<<<<<<<<<
 *             if schema.validator not in ('dict', 'model'):
 *                 raise SchemaError('must be dict or model schema')
 */
    __Pyx_TraceLine(542,0,__PYX_ERR(0, 542, __pyx_L1_error))
    /*with:*/ {
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_mark_key); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 542, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_8 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_6 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_8, __pyx_v_key) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_key);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 542, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_9 = __Pyx_PyObject_LookupSpecial(__pyx_t_6, __pyx_n_s_exit); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 542, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_8 = __Pyx_PyObject_LookupSpecial(__pyx_t_6, __pyx_n_s_enter); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 542, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_10 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
        }
      }
      __pyx_t_5 = (__pyx_t_10) ? __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_10) : __Pyx_PyObject_CallNoArg(__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 542, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      /*try:*/ {
        {
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __Pyx_ExceptionSave(&__pyx_t_11, &__pyx_t_12, &__pyx_t_13);
          __Pyx_XGOTREF(__pyx_t_11);
          __Pyx_XGOTREF(__pyx_t_12);
          __Pyx_XGOTREF(__pyx_t_13);
          /*try:*/ {

            /* "validr/_validator_c.pyx":543
 *     for key, schema in items.items():
 *         with mark_key(key):
 *             if schema.validator not in ('dict', 'model'):             # <<<<<<<<<<<<<<
 *                 raise SchemaError('must be dict or model schema')
 *             if _optional_or_has_default(schema):
 */
            __Pyx_TraceLine(543,0,__PYX_ERR(0, 543, __pyx_L11_error))
            __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_schema, __pyx_n_s_validator); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 543, __pyx_L11_error)
            __Pyx_GOTREF(__pyx_t_6);
            __pyx_t_15 = (__Pyx_PyUnicode_Equals(__pyx_t_6, __pyx_n_u_dict, Py_NE)); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 543, __pyx_L11_error)
            if (__pyx_t_15) {
            } else {
              __pyx_t_14 = __pyx_t_15;
              goto __pyx_L20_bool_binop_done;
            }
            __pyx_t_15 = (__Pyx_PyUnicode_Equals(__pyx_t_6, __pyx_n_u_model, Py_NE)); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 543, __pyx_L11_error)
            __pyx_t_14 = __pyx_t_15;
            __pyx_L20_bool_binop_done:;
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __pyx_t_15 = (__pyx_t_14 != 0);
            if (unlikely(__pyx_t_15)) {

              /* "validr/_validator_c.pyx":544
 *         with mark_key(key):
 *             if schema.validator not in ('dict', 'model'):
 *                 raise SchemaError('must be dict or model schema')             # <<<<<<<<<<<<<<
 *             if _optional_or_has_default(schema):
 *                 raise SchemaError('not allowed optional or default for union schemas')
 */
              __Pyx_TraceLine(544,0,__PYX_ERR(0, 544, __pyx_L11_error))
              __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_SchemaError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 544, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_5);
              __pyx_t_8 = NULL;
              if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
                __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
                if (likely(__pyx_t_8)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                  __Pyx_INCREF(__pyx_t_8);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_5, function);
                }
              }
              __pyx_t_6 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_8, __pyx_kp_u_must_be_dict_or_model_schema) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_kp_u_must_be_dict_or_model_schema);
              __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
              if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 544, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_6);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_Raise(__pyx_t_6, 0, 0, 0);
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __PYX_ERR(0, 544, __pyx_L11_error)

              /* "validr/_validator_c.pyx":543
 *     for key, schema in items.items():
 *         with mark_key(key):
 *             if schema.validator not in ('dict', 'model'):             # <<<<<<<<<<<<<<
 *                 raise SchemaError('must be dict or model schema')
 *             if _optional_or_has_default(schema):
 */
            }

            /* "validr/_validator_c.pyx":545
 *             if schema.validator not in ('dict', 'model'):
 *                 raise SchemaError('must be dict or model schema')
 *             if _optional_or_has_default(schema):             # <<<<<<<<<<<<<<
 *                 raise SchemaError('not allowed optional or default for union schemas')
 *             is_model = schema.validator == 'model'
 */
            __Pyx_TraceLine(545,0,__PYX_ERR(0, 545, __pyx_L11_error))
            __pyx_t_6 = __pyx_f_6validr_12_validator_c__optional_or_has_default(__pyx_v_schema, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 545, __pyx_L11_error)
            __Pyx_GOTREF(__pyx_t_6);
            __pyx_t_15 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 545, __pyx_L11_error)
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            if (unlikely(__pyx_t_15)) {

              /* "validr/_validator_c.pyx":546
 *                 raise SchemaError('must be dict or model schema')
 *             if _optional_or_has_default(schema):
 *                 raise SchemaError('not allowed optional or default for union schemas')             # <<<<<<<<<<<<<<
 *             is_model = schema.validator == 'model'
 *             inners[key] = (is_model, compiler.compile(schema))
 */
              __Pyx_TraceLine(546,0,__PYX_ERR(0, 546, __pyx_L11_error))
              __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_SchemaError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 546, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_5);
              __pyx_t_8 = NULL;
              if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
                __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
                if (likely(__pyx_t_8)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                  __Pyx_INCREF(__pyx_t_8);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_5, function);
                }
              }
              __pyx_t_6 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_8, __pyx_kp_u_not_allowed_optional_or_default) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_kp_u_not_allowed_optional_or_default);
              __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
              if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 546, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_6);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_Raise(__pyx_t_6, 0, 0, 0);
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __PYX_ERR(0, 546, __pyx_L11_error)

              /* "validr/_validator_c.pyx":545
 *             if schema.validator not in ('dict', 'model'):
 *                 raise SchemaError('must be dict or model schema')
 *             if _optional_or_has_default(schema):             # <<<<<<<<<<<<<<
 *                 raise SchemaError('not allowed optional or default for union schemas')
 *             is_model = schema.validator == 'model'
 */
            }

            /* "validr/_validator_c.pyx":547
 *             if _optional_or_has_default(schema):
 *                 raise SchemaError('not allowed optional or default for union schemas')
 *             is_model = schema.validator == 'model'             # <<<<<<<<<<<<<<
 *             inners[key] = (is_model, compiler.compile(schema))
 *     expect_bys = '{' + ', '.join(sorted(inners.keys())) + '}'
 */
            __Pyx_TraceLine(547,0,__PYX_ERR(0, 547, __pyx_L11_error))
            __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_schema, __pyx_n_s_validator); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 547, __pyx_L11_error)
            __Pyx_GOTREF(__pyx_t_6);
            __pyx_t_5 = PyObject_RichCompare(__pyx_t_6, __pyx_n_u_model, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 547, __pyx_L11_error)
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_XDECREF_SET(__pyx_v_is_model, __pyx_t_5);
            __pyx_t_5 = 0;

            /* "validr/_validator_c.pyx":548
 *                 raise SchemaError('not allowed optional or default for union schemas')
 *             is_model = schema.validator == 'model'
 *             inners[key] = (is_model, compiler.compile(schema))             # <<<<<<<<<<<<<<
 *     expect_bys = '{' + ', '.join(sorted(inners.keys())) + '}'
 *     del key, schema, is_model
 */
            __Pyx_TraceLine(548,0,__PYX_ERR(0, 548, __pyx_L11_error))
            __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_compiler, __pyx_n_s_compile); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 548, __pyx_L11_error)
            __Pyx_GOTREF(__pyx_t_6);
            __pyx_t_8 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
              __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_6);
              if (likely(__pyx_t_8)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
                __Pyx_INCREF(__pyx_t_8);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_6, function);
              }
            }
            __pyx_t_5 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_8, __pyx_v_schema) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_schema);
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 548, __pyx_L11_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 548, __pyx_L11_error)
            __Pyx_GOTREF(__pyx_t_6);
            __Pyx_INCREF(__pyx_v_is_model);
            __Pyx_GIVEREF(__pyx_v_is_model);
            PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_is_model);
            __Pyx_GIVEREF(__pyx_t_5);
            PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_5);
            __pyx_t_5 = 0;
            if (unlikely(PyDict_SetItem(__pyx_cur_scope->__pyx_v_inners, __pyx_v_key, __pyx_t_6) < 0)) __PYX_ERR(0, 548, __pyx_L11_error)
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

            /* "validr/_validator_c.pyx":542
 *     inners = {}
 *     for key, schema in items.items():
 *         with mark_key(key):             # <<<<<<<<<<<<<<
 *             if schema.validator not in ('dict', 'model'):
 *                 raise SchemaError('must be dict or model schema')
 */
          }
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          goto __pyx_L18_try_end;
          __pyx_L11_error:;
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          /*except:*/ {
            __Pyx_AddTraceback("validr._validator_c._union_dict_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
            if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_5, &__pyx_t_8) < 0) __PYX_ERR(0, 542, __pyx_L13_except_error)
            __Pyx_GOTREF(__pyx_t_6);
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_GOTREF(__pyx_t_8);
            __pyx_t_10 = PyTuple_Pack(3, __pyx_t_6, __pyx_t_5, __pyx_t_8); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 542, __pyx_L13_except_error)
            __Pyx_GOTREF(__pyx_t_10);
            __pyx_t_16 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_10, NULL);
            __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 542, __pyx_L13_except_error)
            __Pyx_GOTREF(__pyx_t_16);
            __pyx_t_15 = __Pyx_PyObject_IsTrue(__pyx_t_16);
            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
            if (__pyx_t_15 < 0) __PYX_ERR(0, 542, __pyx_L13_except_error)
            __pyx_t_14 = ((!(__pyx_t_15 != 0)) != 0);
            if (__pyx_t_14) {
              __Pyx_GIVEREF(__pyx_t_6);
              __Pyx_GIVEREF(__pyx_t_5);
              __Pyx_XGIVEREF(__pyx_t_8);
              __Pyx_ErrRestoreWithState(__pyx_t_6, __pyx_t_5, __pyx_t_8);
              __pyx_t_6 = 0; __pyx_t_5 = 0; __pyx_t_8 = 0; 
              __PYX_ERR(0, 542, __pyx_L13_except_error)
            }
            __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            goto __pyx_L12_exception_handled;
          }
          __pyx_L13_except_error:;
          __Pyx_XGIVEREF(__pyx_t_11);
          __Pyx_XGIVEREF(__pyx_t_12);
          __Pyx_XGIVEREF(__pyx_t_13);
          __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_12, __pyx_t_13);
          goto __pyx_L1_error;
          __pyx_L12_exception_handled:;
          __Pyx_XGIVEREF(__pyx_t_11);
          __Pyx_XGIVEREF(__pyx_t_12);
          __Pyx_XGIVEREF(__pyx_t_13);
          __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_12, __pyx_t_13);
          __pyx_L18_try_end:;
        }
      }
      /*finally:*/ {
        /*normal exit:*/{
          if (__pyx_t_9) {
            __pyx_t_13 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_tuple__28, NULL);
            __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
            if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 542, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_13);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          goto __pyx_L10;
        }
        __pyx_L10:;
      }
      goto __pyx_L26;
      __pyx_L5_error:;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      goto __pyx_L1_error;
      __pyx_L26:;
    }
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":549
 *             is_model = schema.validator == 'model'
 *             inners[key] = (is_model, compiler.compile(schema))
 *     expect_bys = '{' + ', '.join(sorted(inners.keys())) + '}'             # <<<<<<<<<<<<<<
 *     del key, schema, is_model
 * 
 */
  __Pyx_TraceLine(549,0,__PYX_ERR(0, 549, __pyx_L1_error))
  __pyx_t_8 = __Pyx_PyDict_Keys(__pyx_cur_scope->__pyx_v_inners); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 549, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_5 = PySequence_List(__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 549, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_1 = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_17 = PyList_Sort(__pyx_t_1); if (unlikely(__pyx_t_17 == ((int)-1))) __PYX_ERR(0, 549, __pyx_L1_error)
  __pyx_t_5 = PyUnicode_Join(__pyx_kp_u__40, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 549, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyUnicode_Concat(__pyx_kp_u__39, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 549, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyUnicode_Concat(__pyx_t_1, __pyx_kp_u__41); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 549, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GIVEREF(__pyx_t_5);
  __pyx_cur_scope->__pyx_v_expect_bys = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "validr/_validator_c.pyx":550
 *             inners[key] = (is_model, compiler.compile(schema))
 *     expect_bys = '{' + ', '.join(sorted(inners.keys())) + '}'
 *     del key, schema, is_model             # <<<<<<<<<<<<<<
 * 
 *     def validate(value):
 */
  __Pyx_TraceLine(550,0,__PYX_ERR(0, 550, __pyx_L1_error))
  if (unlikely(!__pyx_v_key)) { __Pyx_RaiseUnboundLocalError("key"); __PYX_ERR(0, 550, __pyx_L1_error) }
  __Pyx_DECREF(__pyx_v_key);
  __pyx_v_key = NULL;
  if (unlikely(!__pyx_v_schema)) { __Pyx_RaiseUnboundLocalError("schema"); __PYX_ERR(0, 550, __pyx_L1_error) }
  __Pyx_DECREF(__pyx_v_schema);
  __pyx_v_schema = NULL;
  if (unlikely(!__pyx_v_is_model)) { __Pyx_RaiseUnboundLocalError("is_model"); __PYX_ERR(0, 550, __pyx_L1_error) }
  __Pyx_DECREF(__pyx_v_is_model);
  __pyx_v_is_model = NULL;

  /* "validr/_validator_c.pyx":552
 *     del key, schema, is_model
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         if is_dict(value):
 *             getter = get_dict_value
 */
  __Pyx_TraceLine(552,0,__PYX_ERR(0, 552, __pyx_L1_error))
  __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_21_union_dict_validator_1validate, 0, __pyx_n_s_union_dict_validator_locals_val, ((PyObject*)__pyx_cur_scope), __pyx_n_s_validr__validator_c, __pyx_d, ((PyObject *)__pyx_codeobj__50)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_v_validate = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "validr/_validator_c.pyx":572
 *         return result
 * 
 *     return validate             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(572,0,__PYX_ERR(0, 572, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_validate);
  __pyx_r = __pyx_v_validate;
  goto __pyx_L0;

  /* "validr/_validator_c.pyx":539
 * 
 * @validator(accept=object, output=object)
 * def _union_dict_validator(compiler, items, str by):             # <<<<<<<<<<<<<<
 *     inners = {}
 *     for key, schema in items.items():
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("validr._validator_c._union_dict_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_schema);
  __Pyx_XDECREF(__pyx_v_is_model);
  __Pyx_XDECREF(__pyx_v_validate);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":575
 * 
 * 
 * cpdef any_validate(value):             # <<<<<<<<<<<<<<
 *     return copy(value)
 * 
 */

static PyObject *__pyx_pw_6validr_12_validator_c_41any_validate(PyObject *__pyx_self, PyObject *__pyx_v_value); /*proto*/
static PyObject *__pyx_f_6validr_12_validator_c_any_validate(PyObject *__pyx_v_value, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("any_validate", 0);
  __Pyx_TraceCall("any_validate", __pyx_f[0], 575, 0, __PYX_ERR(0, 575, __pyx_L1_error));

  /* "validr/_validator_c.pyx":576
 * 
 * cpdef any_validate(value):
 *     return copy(value)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(576,0,__PYX_ERR(0, 576, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_copy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_value) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_value);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "validr/_validator_c.pyx":575
 * 
 * 
 * cpdef any_validate(value):             # <<<<<<<<<<<<<<
 *     return copy(value)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("validr._validator_c.any_validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_41any_validate(PyObject *__pyx_self, PyObject *__pyx_v_value); /*proto*/
static PyObject *__pyx_pw_6validr_12_validator_c_41any_validate(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("any_validate (wrapper)", 0);
  __pyx_r = __pyx_pf_6validr_12_validator_c_40any_validate(__pyx_self, ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6validr_12_validator_c_40any_validate(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("any_validate", 0);
  __Pyx_TraceCall("any_validate (wrapper)", __pyx_f[0], 575, 0, __PYX_ERR(0, 575, __pyx_L1_error));
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6validr_12_validator_c_any_validate(__pyx_v_value, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 575, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("validr._validator_c.any_validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":580
 * 
 * @validator(accept=object, output=object)
 * def any_validator(compiler, **ignore_kwargs):             # <<<<<<<<<<<<<<
 *     """Accept any value"""
 *     return any_validate
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_43any_validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6validr_12_validator_c_42any_validator[] = "Accept any value";
static PyMethodDef __pyx_mdef_6validr_12_validator_c_43any_validator = {"any_validator", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6validr_12_validator_c_43any_validator, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6validr_12_validator_c_42any_validator};
static PyObject *__pyx_pw_6validr_12_validator_c_43any_validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_compiler = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ignore_kwargs = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("any_validator (wrapper)", 0);
  __pyx_v_ignore_kwargs = PyDict_New(); if (unlikely(!__pyx_v_ignore_kwargs)) return NULL;
  __Pyx_GOTREF(__pyx_v_ignore_kwargs);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_compiler,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_compiler)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, __pyx_v_ignore_kwargs, values, pos_args, "any_validator") < 0)) __PYX_ERR(0, 580, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_compiler = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("any_validator", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 580, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_DECREF(__pyx_v_ignore_kwargs); __pyx_v_ignore_kwargs = 0;
  __Pyx_AddTraceback("validr._validator_c.any_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6validr_12_validator_c_42any_validator(__pyx_self, __pyx_v_compiler, __pyx_v_ignore_kwargs);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_ignore_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6validr_12_validator_c_42any_validator(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_compiler, CYTHON_UNUSED PyObject *__pyx_v_ignore_kwargs) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_TraceFrameInit(__pyx_codeobj__51)
  __Pyx_RefNannySetupContext("any_validator", 0);
  __Pyx_TraceCall("any_validator", __pyx_f[0], 580, 0, __PYX_ERR(0, 580, __pyx_L1_error));

  /* "validr/_validator_c.pyx":582
 * def any_validator(compiler, **ignore_kwargs):
 *     """Accept any value"""
 *     return any_validate             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(582,0,__PYX_ERR(0, 582, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_any_validate); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "validr/_validator_c.pyx":580
 * 
 * @validator(accept=object, output=object)
 * def any_validator(compiler, **ignore_kwargs):             # <<<<<<<<<<<<<<
 *     """Accept any value"""
 *     return any_validate
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("validr._validator_c.any_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":586
 * 
 * @validator(accept=(int, float, str), output=int)
 * def int_validator(compiler, min=-sys.maxsize, max=sys.maxsize):             # <<<<<<<<<<<<<<
 *     """Validate int or convert string to int
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_45int_validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6validr_12_validator_c_44int_validator[] = "Validate int or convert string to int\n\n    Args:\n        min (int): the min value, default -sys.maxsize\n        max (int): the max value, default sys.maxsize\n    ";
static PyMethodDef __pyx_mdef_6validr_12_validator_c_45int_validator = {"int_validator", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6validr_12_validator_c_45int_validator, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6validr_12_validator_c_44int_validator};
static PyObject *__pyx_pw_6validr_12_validator_c_45int_validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_compiler = 0;
  PyObject *__pyx_v_min = 0;
  PyObject *__pyx_v_max = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("int_validator (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_compiler,&__pyx_n_s_min,&__pyx_n_s_max,0};
    PyObject* values[3] = {0,0,0};
    values[1] = __pyx_k__52;
    values[2] = __pyx_k__53;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_compiler)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_min);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "int_validator") < 0)) __PYX_ERR(0, 586, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_compiler = values[0];
    __pyx_v_min = values[1];
    __pyx_v_max = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("int_validator", 0, 1, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 586, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("validr._validator_c.int_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6validr_12_validator_c_44int_validator(__pyx_self, __pyx_v_compiler, __pyx_v_min, __pyx_v_max);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":595
 *     min, max = int(min), int(max)
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         try:
 *             v = int(value)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_13int_validator_1validate(PyObject *__pyx_self, PyObject *__pyx_v_value); /*proto*/
static PyMethodDef __pyx_mdef_6validr_12_validator_c_13int_validator_1validate = {"validate", (PyCFunction)__pyx_pw_6validr_12_validator_c_13int_validator_1validate, METH_O, 0};
static PyObject *__pyx_pw_6validr_12_validator_c_13int_validator_1validate(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("validate (wrapper)", 0);
  __pyx_r = __pyx_pf_6validr_12_validator_c_13int_validator_validate(__pyx_self, ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6validr_12_validator_c_13int_validator_validate(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_14_int_validator *__pyx_cur_scope;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_14_int_validator *__pyx_outer_scope;
  PyObject *__pyx_v_v = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  __Pyx_RefNannySetupContext("validate", 0);
  __pyx_outer_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_14_int_validator *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_TraceCall("validate", __pyx_f[0], 595, 0, __PYX_ERR(0, 595, __pyx_L1_error));

  /* "validr/_validator_c.pyx":596
 * 
 *     def validate(value):
 *         try:             # <<<<<<<<<<<<<<
 *             v = int(value)
 *         except Exception:
 */
  __Pyx_TraceLine(596,0,__PYX_ERR(0, 596, __pyx_L1_error))
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "validr/_validator_c.pyx":597
 *     def validate(value):
 *         try:
 *             v = int(value)             # <<<<<<<<<<<<<<
 *         except Exception:
 *             raise Invalid('invalid int') from None
 */
      __Pyx_TraceLine(597,0,__PYX_ERR(0, 597, __pyx_L3_error))
      __pyx_t_4 = __Pyx_PyNumber_Int(__pyx_v_value); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 597, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_v_v = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "validr/_validator_c.pyx":596
 * 
 *     def validate(value):
 *         try:             # <<<<<<<<<<<<<<
 *             v = int(value)
 *         except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "validr/_validator_c.pyx":598
 *         try:
 *             v = int(value)
 *         except Exception:             # <<<<<<<<<<<<<<
 *             raise Invalid('invalid int') from None
 *         if v < min:
 */
    __Pyx_TraceLine(598,0,__PYX_ERR(0, 598, __pyx_L5_except_error))
    __pyx_t_5 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_5) {
      __Pyx_AddTraceback("validr._validator_c.int_validator.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(0, 598, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "validr/_validator_c.pyx":599
 *             v = int(value)
 *         except Exception:
 *             raise Invalid('invalid int') from None             # <<<<<<<<<<<<<<
 *         if v < min:
 *             raise Invalid('value must >= %d' % min)
 */
      __Pyx_TraceLine(599,0,__PYX_ERR(0, 599, __pyx_L5_except_error))
      __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 599, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_9, function);
        }
      }
      __pyx_t_8 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_10, __pyx_kp_u_invalid_int) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_kp_u_invalid_int);
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 599, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_Raise(__pyx_t_8, 0, 0, Py_None);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(0, 599, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "validr/_validator_c.pyx":596
 * 
 *     def validate(value):
 *         try:             # <<<<<<<<<<<<<<
 *             v = int(value)
 *         except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "validr/_validator_c.pyx":600
 *         except Exception:
 *             raise Invalid('invalid int') from None
 *         if v < min:             # <<<<<<<<<<<<<<
 *             raise Invalid('value must >= %d' % min)
 *         elif v > max:
 */
  __Pyx_TraceLine(600,0,__PYX_ERR(0, 600, __pyx_L1_error))
  if (unlikely(!__pyx_cur_scope->__pyx_v_min)) { __Pyx_RaiseClosureNameError("min"); __PYX_ERR(0, 600, __pyx_L1_error) }
  __pyx_t_7 = PyObject_RichCompare(__pyx_v_v, __pyx_cur_scope->__pyx_v_min, Py_LT); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 600, __pyx_L1_error)
  __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 600, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(__pyx_t_11)) {

    /* "validr/_validator_c.pyx":601
 *             raise Invalid('invalid int') from None
 *         if v < min:
 *             raise Invalid('value must >= %d' % min)             # <<<<<<<<<<<<<<
 *         elif v > max:
 *             raise Invalid('value must <= %d' % max)
 */
    __Pyx_TraceLine(601,0,__PYX_ERR(0, 601, __pyx_L1_error))
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 601, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (unlikely(!__pyx_cur_scope->__pyx_v_min)) { __Pyx_RaiseClosureNameError("min"); __PYX_ERR(0, 601, __pyx_L1_error) }
    __pyx_t_4 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_value_must_d, __pyx_cur_scope->__pyx_v_min); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 601, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_7 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_8, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 601, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_Raise(__pyx_t_7, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __PYX_ERR(0, 601, __pyx_L1_error)

    /* "validr/_validator_c.pyx":600
 *         except Exception:
 *             raise Invalid('invalid int') from None
 *         if v < min:             # <<<<<<<<<<<<<<
 *             raise Invalid('value must >= %d' % min)
 *         elif v > max:
 */
  }

  /* "validr/_validator_c.pyx":602
 *         if v < min:
 *             raise Invalid('value must >= %d' % min)
 *         elif v > max:             # <<<<<<<<<<<<<<
 *             raise Invalid('value must <= %d' % max)
 *         return v
 */
  __Pyx_TraceLine(602,0,__PYX_ERR(0, 602, __pyx_L1_error))
  if (unlikely(!__pyx_cur_scope->__pyx_v_max)) { __Pyx_RaiseClosureNameError("max"); __PYX_ERR(0, 602, __pyx_L1_error) }
  __pyx_t_7 = PyObject_RichCompare(__pyx_v_v, __pyx_cur_scope->__pyx_v_max, Py_GT); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 602, __pyx_L1_error)
  __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 602, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(__pyx_t_11)) {

    /* "validr/_validator_c.pyx":603
 *             raise Invalid('value must >= %d' % min)
 *         elif v > max:
 *             raise Invalid('value must <= %d' % max)             # <<<<<<<<<<<<<<
 *         return v
 *     return validate
 */
    __Pyx_TraceLine(603,0,__PYX_ERR(0, 603, __pyx_L1_error))
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 603, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (unlikely(!__pyx_cur_scope->__pyx_v_max)) { __Pyx_RaiseClosureNameError("max"); __PYX_ERR(0, 603, __pyx_L1_error) }
    __pyx_t_4 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_value_must_d_2, __pyx_cur_scope->__pyx_v_max); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 603, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_7 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_8, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 603, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_Raise(__pyx_t_7, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __PYX_ERR(0, 603, __pyx_L1_error)

    /* "validr/_validator_c.pyx":602
 *         if v < min:
 *             raise Invalid('value must >= %d' % min)
 *         elif v > max:             # <<<<<<<<<<<<<<
 *             raise Invalid('value must <= %d' % max)
 *         return v
 */
  }

  /* "validr/_validator_c.pyx":604
 *         elif v > max:
 *             raise Invalid('value must <= %d' % max)
 *         return v             # <<<<<<<<<<<<<<
 *     return validate
 * 
 */
  __Pyx_TraceLine(604,0,__PYX_ERR(0, 604, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_v);
  __pyx_r = __pyx_v_v;
  goto __pyx_L0;

  /* "validr/_validator_c.pyx":595
 *     min, max = int(min), int(max)
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         try:
 *             v = int(value)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("validr._validator_c.int_validator.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":586
 * 
 * @validator(accept=(int, float, str), output=int)
 * def int_validator(compiler, min=-sys.maxsize, max=sys.maxsize):             # <<<<<<<<<<<<<<
 *     """Validate int or convert string to int
 * 
 */

static PyObject *__pyx_pf_6validr_12_validator_c_44int_validator(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_compiler, PyObject *__pyx_v_min, PyObject *__pyx_v_max) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_14_int_validator *__pyx_cur_scope;
  PyObject *__pyx_v_validate = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_TraceFrameInit(__pyx_codeobj__54)
  __Pyx_RefNannySetupContext("int_validator", 0);
  __pyx_cur_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_14_int_validator *)__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_14_int_validator(__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_14_int_validator, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_14_int_validator *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 586, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __Pyx_TraceCall("int_validator", __pyx_f[0], 586, 0, __PYX_ERR(0, 586, __pyx_L1_error));
  __pyx_cur_scope->__pyx_v_min = __pyx_v_min;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_min);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_min);
  __pyx_cur_scope->__pyx_v_max = __pyx_v_max;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_max);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_max);

  /* "validr/_validator_c.pyx":593
 *         max (int): the max value, default sys.maxsize
 *     """
 *     min, max = int(min), int(max)             # <<<<<<<<<<<<<<
 * 
 *     def validate(value):
 */
  __Pyx_TraceLine(593,0,__PYX_ERR(0, 593, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_cur_scope->__pyx_v_min); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 593, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyNumber_Int(__pyx_cur_scope->__pyx_v_max); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 593, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_min);
  __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_min, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;
  __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_max);
  __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_max, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "validr/_validator_c.pyx":595
 *     min, max = int(min), int(max)
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         try:
 *             v = int(value)
 */
  __Pyx_TraceLine(595,0,__PYX_ERR(0, 595, __pyx_L1_error))
  __pyx_t_2 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_13int_validator_1validate, 0, __pyx_n_s_int_validator_locals_validate, ((PyObject*)__pyx_cur_scope), __pyx_n_s_validr__validator_c, __pyx_d, ((PyObject *)__pyx_codeobj__56)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 595, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_validate = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "validr/_validator_c.pyx":605
 *             raise Invalid('value must <= %d' % max)
 *         return v
 *     return validate             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(605,0,__PYX_ERR(0, 605, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_validate);
  __pyx_r = __pyx_v_validate;
  goto __pyx_L0;

  /* "validr/_validator_c.pyx":586
 * 
 * @validator(accept=(int, float, str), output=int)
 * def int_validator(compiler, min=-sys.maxsize, max=sys.maxsize):             # <<<<<<<<<<<<<<
 *     """Validate int or convert string to int
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("validr._validator_c.int_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_validate);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":623
 * 
 * @validator(accept=(bool, int, str), output=bool)
 * def bool_validator(compiler):             # <<<<<<<<<<<<<<
 *     """Validate bool"""
 *     def validate(value):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_47bool_validator(PyObject *__pyx_self, PyObject *__pyx_v_compiler); /*proto*/
static char __pyx_doc_6validr_12_validator_c_46bool_validator[] = "Validate bool";
static PyMethodDef __pyx_mdef_6validr_12_validator_c_47bool_validator = {"bool_validator", (PyCFunction)__pyx_pw_6validr_12_validator_c_47bool_validator, METH_O, __pyx_doc_6validr_12_validator_c_46bool_validator};
static PyObject *__pyx_pw_6validr_12_validator_c_47bool_validator(PyObject *__pyx_self, PyObject *__pyx_v_compiler) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("bool_validator (wrapper)", 0);
  __pyx_r = __pyx_pf_6validr_12_validator_c_46bool_validator(__pyx_self, ((PyObject *)__pyx_v_compiler));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":625
 * def bool_validator(compiler):
 *     """Validate bool"""
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         if value in _TRUE_VALUES:
 *             return True
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_14bool_validator_1validate(PyObject *__pyx_self, PyObject *__pyx_v_value); /*proto*/
static PyMethodDef __pyx_mdef_6validr_12_validator_c_14bool_validator_1validate = {"validate", (PyCFunction)__pyx_pw_6validr_12_validator_c_14bool_validator_1validate, METH_O, 0};
static PyObject *__pyx_pw_6validr_12_validator_c_14bool_validator_1validate(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("validate (wrapper)", 0);
  __pyx_r = __pyx_pf_6validr_12_validator_c_14bool_validator_validate(__pyx_self, ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6validr_12_validator_c_14bool_validator_validate(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("validate", 0);
  __Pyx_TraceCall("validate", __pyx_f[0], 625, 0, __PYX_ERR(0, 625, __pyx_L1_error));

  /* "validr/_validator_c.pyx":626
 *     """Validate bool"""
 *     def validate(value):
 *         if value in _TRUE_VALUES:             # <<<<<<<<<<<<<<
 *             return True
 *         elif value in _FALSE_VALUES:
 */
  __Pyx_TraceLine(626,0,__PYX_ERR(0, 626, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_TRUE_VALUES); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 626, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_v_value, __pyx_t_1, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 626, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "validr/_validator_c.pyx":627
 *     def validate(value):
 *         if value in _TRUE_VALUES:
 *             return True             # <<<<<<<<<<<<<<
 *         elif value in _FALSE_VALUES:
 *             return False
 */
    __Pyx_TraceLine(627,0,__PYX_ERR(0, 627, __pyx_L1_error))
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_True);
    __pyx_r = Py_True;
    goto __pyx_L0;

    /* "validr/_validator_c.pyx":626
 *     """Validate bool"""
 *     def validate(value):
 *         if value in _TRUE_VALUES:             # <<<<<<<<<<<<<<
 *             return True
 *         elif value in _FALSE_VALUES:
 */
  }

  /* "validr/_validator_c.pyx":628
 *         if value in _TRUE_VALUES:
 *             return True
 *         elif value in _FALSE_VALUES:             # <<<<<<<<<<<<<<
 *             return False
 *         else:
 */
  __Pyx_TraceLine(628,0,__PYX_ERR(0, 628, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_FALSE_VALUES); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = (__Pyx_PySequence_ContainsTF(__pyx_v_value, __pyx_t_1, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 628, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = (__pyx_t_3 != 0);
  if (likely(__pyx_t_2)) {

    /* "validr/_validator_c.pyx":629
 *             return True
 *         elif value in _FALSE_VALUES:
 *             return False             # <<<<<<<<<<<<<<
 *         else:
 *             raise Invalid('invalid bool')
 */
    __Pyx_TraceLine(629,0,__PYX_ERR(0, 629, __pyx_L1_error))
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;

    /* "validr/_validator_c.pyx":628
 *         if value in _TRUE_VALUES:
 *             return True
 *         elif value in _FALSE_VALUES:             # <<<<<<<<<<<<<<
 *             return False
 *         else:
 */
  }

  /* "validr/_validator_c.pyx":631
 *             return False
 *         else:
 *             raise Invalid('invalid bool')             # <<<<<<<<<<<<<<
 *     return validate
 * 
 */
  __Pyx_TraceLine(631,0,__PYX_ERR(0, 631, __pyx_L1_error))
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 631, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_kp_u_invalid_bool) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_kp_u_invalid_bool);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 631, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 631, __pyx_L1_error)
  }

  /* "validr/_validator_c.pyx":625
 * def bool_validator(compiler):
 *     """Validate bool"""
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         if value in _TRUE_VALUES:
 *             return True
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("validr._validator_c.bool_validator.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":623
 * 
 * @validator(accept=(bool, int, str), output=bool)
 * def bool_validator(compiler):             # <<<<<<<<<<<<<<
 *     """Validate bool"""
 *     def validate(value):
 */

static PyObject *__pyx_pf_6validr_12_validator_c_46bool_validator(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_compiler) {
  PyObject *__pyx_v_validate = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_TraceFrameInit(__pyx_codeobj__57)
  __Pyx_RefNannySetupContext("bool_validator", 0);
  __Pyx_TraceCall("bool_validator", __pyx_f[0], 623, 0, __PYX_ERR(0, 623, __pyx_L1_error));

  /* "validr/_validator_c.pyx":625
 * def bool_validator(compiler):
 *     """Validate bool"""
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         if value in _TRUE_VALUES:
 *             return True
 */
  __Pyx_TraceLine(625,0,__PYX_ERR(0, 625, __pyx_L1_error))
  __pyx_t_1 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_14bool_validator_1validate, 0, __pyx_n_s_bool_validator_locals_validate, NULL, __pyx_n_s_validr__validator_c, __pyx_d, ((PyObject *)__pyx_codeobj__59)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 625, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_validate = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":632
 *         else:
 *             raise Invalid('invalid bool')
 *     return validate             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(632,0,__PYX_ERR(0, 632, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_validate);
  __pyx_r = __pyx_v_validate;
  goto __pyx_L0;

  /* "validr/_validator_c.pyx":623
 * 
 * @validator(accept=(bool, int, str), output=bool)
 * def bool_validator(compiler):             # <<<<<<<<<<<<<<
 *     """Validate bool"""
 *     def validate(value):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("validr._validator_c.bool_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_validate);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":636
 * 
 * @validator(accept=(int, float, str), output=float)
 * def float_validator(compiler, min=-sys.float_info.max, max=sys.float_info.max,             # <<<<<<<<<<<<<<
 *                     exmin=False, exmax=False):
 *     """Validate float string
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_49float_validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6validr_12_validator_c_48float_validator[] = "Validate float string\n\n    Args:\n        min (float): the min value, default -sys.float_info.max\n        max (float): the max value, default sys.float_info.max\n        exmin (bool,float): exclude min value or not, default false\n        exmax (bool,float): exclude max value or not, default false\n    ";
static PyMethodDef __pyx_mdef_6validr_12_validator_c_49float_validator = {"float_validator", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6validr_12_validator_c_49float_validator, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6validr_12_validator_c_48float_validator};
static PyObject *__pyx_pw_6validr_12_validator_c_49float_validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_compiler = 0;
  PyObject *__pyx_v_min = 0;
  PyObject *__pyx_v_max = 0;
  PyObject *__pyx_v_exmin = 0;
  PyObject *__pyx_v_exmax = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("float_validator (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_compiler,&__pyx_n_s_min,&__pyx_n_s_max,&__pyx_n_s_exmin,&__pyx_n_s_exmax,0};
    PyObject* values[5] = {0,0,0,0,0};
    values[1] = __pyx_k__60;
    values[2] = __pyx_k__61;

    /* "validr/_validator_c.pyx":637
 * @validator(accept=(int, float, str), output=float)
 * def float_validator(compiler, min=-sys.float_info.max, max=sys.float_info.max,
 *                     exmin=False, exmax=False):             # <<<<<<<<<<<<<<
 *     """Validate float string
 * 
 */
    values[3] = ((PyObject *)Py_False);
    values[4] = ((PyObject *)Py_False);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_compiler)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_min);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_exmin);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_exmax);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "float_validator") < 0)) __PYX_ERR(0, 636, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_compiler = values[0];
    __pyx_v_min = values[1];
    __pyx_v_max = values[2];
    __pyx_v_exmin = values[3];
    __pyx_v_exmax = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("float_validator", 0, 1, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 636, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("validr._validator_c.float_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6validr_12_validator_c_48float_validator(__pyx_self, __pyx_v_compiler, __pyx_v_min, __pyx_v_max, __pyx_v_exmin, __pyx_v_exmax);

  /* "validr/_validator_c.pyx":636
 * 
 * @validator(accept=(int, float, str), output=float)
 * def float_validator(compiler, min=-sys.float_info.max, max=sys.float_info.max,             # <<<<<<<<<<<<<<
 *                     exmin=False, exmax=False):
 *     """Validate float string
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":658
 *         exmax = bool(exmax)
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         try:
 *             v = float(value)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_15float_validator_1validate(PyObject *__pyx_self, PyObject *__pyx_v_value); /*proto*/
static PyMethodDef __pyx_mdef_6validr_12_validator_c_15float_validator_1validate = {"validate", (PyCFunction)__pyx_pw_6validr_12_validator_c_15float_validator_1validate, METH_O, 0};
static PyObject *__pyx_pw_6validr_12_validator_c_15float_validator_1validate(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("validate (wrapper)", 0);
  __pyx_r = __pyx_pf_6validr_12_validator_c_15float_validator_validate(__pyx_self, ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6validr_12_validator_c_15float_validator_validate(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_15_float_validator *__pyx_cur_scope;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_15_float_validator *__pyx_outer_scope;
  double __pyx_v_v;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  double __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  __Pyx_RefNannySetupContext("validate", 0);
  __pyx_outer_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_15_float_validator *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_TraceCall("validate", __pyx_f[0], 658, 0, __PYX_ERR(0, 658, __pyx_L1_error));

  /* "validr/_validator_c.pyx":659
 * 
 *     def validate(value):
 *         try:             # <<<<<<<<<<<<<<
 *             v = float(value)
 *         except Exception:
 */
  __Pyx_TraceLine(659,0,__PYX_ERR(0, 659, __pyx_L1_error))
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "validr/_validator_c.pyx":660
 *     def validate(value):
 *         try:
 *             v = float(value)             # <<<<<<<<<<<<<<
 *         except Exception:
 *             raise Invalid('invalid float') from None
 */
      __Pyx_TraceLine(660,0,__PYX_ERR(0, 660, __pyx_L3_error))
      __pyx_t_4 = __Pyx_PyObject_AsDouble(__pyx_v_value); if (unlikely(__pyx_t_4 == ((double)((double)-1)) && PyErr_Occurred())) __PYX_ERR(0, 660, __pyx_L3_error)
      __pyx_v_v = __pyx_t_4;

      /* "validr/_validator_c.pyx":659
 * 
 *     def validate(value):
 *         try:             # <<<<<<<<<<<<<<
 *             v = float(value)
 *         except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "validr/_validator_c.pyx":661
 *         try:
 *             v = float(value)
 *         except Exception:             # <<<<<<<<<<<<<<
 *             raise Invalid('invalid float') from None
 *         if exmin:
 */
    __Pyx_TraceLine(661,0,__PYX_ERR(0, 661, __pyx_L5_except_error))
    __pyx_t_5 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_5) {
      __Pyx_AddTraceback("validr._validator_c.float_validator.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8) < 0) __PYX_ERR(0, 661, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GOTREF(__pyx_t_8);

      /* "validr/_validator_c.pyx":662
 *             v = float(value)
 *         except Exception:
 *             raise Invalid('invalid float') from None             # <<<<<<<<<<<<<<
 *         if exmin:
 *             if v <= min:
 */
      __Pyx_TraceLine(662,0,__PYX_ERR(0, 662, __pyx_L5_except_error))
      __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 662, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_11 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
        __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_10);
        if (likely(__pyx_t_11)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
          __Pyx_INCREF(__pyx_t_11);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_10, function);
        }
      }
      __pyx_t_9 = (__pyx_t_11) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_11, __pyx_kp_u_invalid_float) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_kp_u_invalid_float);
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 662, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_Raise(__pyx_t_9, 0, 0, Py_None);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __PYX_ERR(0, 662, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "validr/_validator_c.pyx":659
 * 
 *     def validate(value):
 *         try:             # <<<<<<<<<<<<<<
 *             v = float(value)
 *         except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "validr/_validator_c.pyx":663
 *         except Exception:
 *             raise Invalid('invalid float') from None
 *         if exmin:             # <<<<<<<<<<<<<<
 *             if v <= min:
 *                 raise Invalid('value must > %d' % min)
 */
  __Pyx_TraceLine(663,0,__PYX_ERR(0, 663, __pyx_L1_error))
  if (unlikely(!__pyx_cur_scope->__pyx_v_exmin)) { __Pyx_RaiseClosureNameError("exmin"); __PYX_ERR(0, 663, __pyx_L1_error) }
  __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_exmin); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 663, __pyx_L1_error)
  if (__pyx_t_12) {

    /* "validr/_validator_c.pyx":664
 *             raise Invalid('invalid float') from None
 *         if exmin:
 *             if v <= min:             # <<<<<<<<<<<<<<
 *                 raise Invalid('value must > %d' % min)
 *         else:
 */
    __Pyx_TraceLine(664,0,__PYX_ERR(0, 664, __pyx_L1_error))
    __pyx_t_8 = PyFloat_FromDouble(__pyx_v_v); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 664, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    if (unlikely(!__pyx_cur_scope->__pyx_v_min)) { __Pyx_RaiseClosureNameError("min"); __PYX_ERR(0, 664, __pyx_L1_error) }
    __pyx_t_7 = PyObject_RichCompare(__pyx_t_8, __pyx_cur_scope->__pyx_v_min, Py_LE); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 664, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 664, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(__pyx_t_12)) {

      /* "validr/_validator_c.pyx":665
 *         if exmin:
 *             if v <= min:
 *                 raise Invalid('value must > %d' % min)             # <<<<<<<<<<<<<<
 *         else:
 *             if v < min:
 */
      __Pyx_TraceLine(665,0,__PYX_ERR(0, 665, __pyx_L1_error))
      __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 665, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (unlikely(!__pyx_cur_scope->__pyx_v_min)) { __Pyx_RaiseClosureNameError("min"); __PYX_ERR(0, 665, __pyx_L1_error) }
      __pyx_t_6 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_value_must_d_3, __pyx_cur_scope->__pyx_v_min); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 665, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_9 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
        }
      }
      __pyx_t_7 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_9, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_6);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 665, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_Raise(__pyx_t_7, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __PYX_ERR(0, 665, __pyx_L1_error)

      /* "validr/_validator_c.pyx":664
 *             raise Invalid('invalid float') from None
 *         if exmin:
 *             if v <= min:             # <<<<<<<<<<<<<<
 *                 raise Invalid('value must > %d' % min)
 *         else:
 */
    }

    /* "validr/_validator_c.pyx":663
 *         except Exception:
 *             raise Invalid('invalid float') from None
 *         if exmin:             # <<<<<<<<<<<<<<
 *             if v <= min:
 *                 raise Invalid('value must > %d' % min)
 */
    goto __pyx_L11;
  }

  /* "validr/_validator_c.pyx":667
 *                 raise Invalid('value must > %d' % min)
 *         else:
 *             if v < min:             # <<<<<<<<<<<<<<
 *                 raise Invalid('value must >= %d' % min)
 *         if exmax:
 */
  __Pyx_TraceLine(667,0,__PYX_ERR(0, 667, __pyx_L1_error))
  /*else*/ {
    __pyx_t_7 = PyFloat_FromDouble(__pyx_v_v); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 667, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (unlikely(!__pyx_cur_scope->__pyx_v_min)) { __Pyx_RaiseClosureNameError("min"); __PYX_ERR(0, 667, __pyx_L1_error) }
    __pyx_t_8 = PyObject_RichCompare(__pyx_t_7, __pyx_cur_scope->__pyx_v_min, Py_LT); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 667, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 667, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(__pyx_t_12)) {

      /* "validr/_validator_c.pyx":668
 *         else:
 *             if v < min:
 *                 raise Invalid('value must >= %d' % min)             # <<<<<<<<<<<<<<
 *         if exmax:
 *             if v >= max:
 */
      __Pyx_TraceLine(668,0,__PYX_ERR(0, 668, __pyx_L1_error))
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 668, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (unlikely(!__pyx_cur_scope->__pyx_v_min)) { __Pyx_RaiseClosureNameError("min"); __PYX_ERR(0, 668, __pyx_L1_error) }
      __pyx_t_6 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_value_must_d, __pyx_cur_scope->__pyx_v_min); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 668, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_9 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      __pyx_t_8 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_9, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 668, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(0, 668, __pyx_L1_error)

      /* "validr/_validator_c.pyx":667
 *                 raise Invalid('value must > %d' % min)
 *         else:
 *             if v < min:             # <<<<<<<<<<<<<<
 *                 raise Invalid('value must >= %d' % min)
 *         if exmax:
 */
    }
  }
  __pyx_L11:;

  /* "validr/_validator_c.pyx":669
 *             if v < min:
 *                 raise Invalid('value must >= %d' % min)
 *         if exmax:             # <<<<<<<<<<<<<<
 *             if v >= max:
 *                 raise Invalid('value must < %d' % max)
 */
  __Pyx_TraceLine(669,0,__PYX_ERR(0, 669, __pyx_L1_error))
  if (unlikely(!__pyx_cur_scope->__pyx_v_exmax)) { __Pyx_RaiseClosureNameError("exmax"); __PYX_ERR(0, 669, __pyx_L1_error) }
  __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_exmax); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 669, __pyx_L1_error)
  if (__pyx_t_12) {

    /* "validr/_validator_c.pyx":670
 *                 raise Invalid('value must >= %d' % min)
 *         if exmax:
 *             if v >= max:             # <<<<<<<<<<<<<<
 *                 raise Invalid('value must < %d' % max)
 *         else:
 */
    __Pyx_TraceLine(670,0,__PYX_ERR(0, 670, __pyx_L1_error))
    __pyx_t_8 = PyFloat_FromDouble(__pyx_v_v); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 670, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    if (unlikely(!__pyx_cur_scope->__pyx_v_max)) { __Pyx_RaiseClosureNameError("max"); __PYX_ERR(0, 670, __pyx_L1_error) }
    __pyx_t_7 = PyObject_RichCompare(__pyx_t_8, __pyx_cur_scope->__pyx_v_max, Py_GE); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 670, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 670, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(__pyx_t_12)) {

      /* "validr/_validator_c.pyx":671
 *         if exmax:
 *             if v >= max:
 *                 raise Invalid('value must < %d' % max)             # <<<<<<<<<<<<<<
 *         else:
 *             if v > max:
 */
      __Pyx_TraceLine(671,0,__PYX_ERR(0, 671, __pyx_L1_error))
      __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 671, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (unlikely(!__pyx_cur_scope->__pyx_v_max)) { __Pyx_RaiseClosureNameError("max"); __PYX_ERR(0, 671, __pyx_L1_error) }
      __pyx_t_6 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_value_must_d_4, __pyx_cur_scope->__pyx_v_max); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 671, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_9 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
        }
      }
      __pyx_t_7 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_9, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_6);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 671, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_Raise(__pyx_t_7, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __PYX_ERR(0, 671, __pyx_L1_error)

      /* "validr/_validator_c.pyx":670
 *                 raise Invalid('value must >= %d' % min)
 *         if exmax:
 *             if v >= max:             # <<<<<<<<<<<<<<
 *                 raise Invalid('value must < %d' % max)
 *         else:
 */
    }

    /* "validr/_validator_c.pyx":669
 *             if v < min:
 *                 raise Invalid('value must >= %d' % min)
 *         if exmax:             # <<<<<<<<<<<<<<
 *             if v >= max:
 *                 raise Invalid('value must < %d' % max)
 */
    goto __pyx_L14;
  }

  /* "validr/_validator_c.pyx":673
 *                 raise Invalid('value must < %d' % max)
 *         else:
 *             if v > max:             # <<<<<<<<<<<<<<
 *                 raise Invalid('value must <= %d' % max)
 *         return v
 */
  __Pyx_TraceLine(673,0,__PYX_ERR(0, 673, __pyx_L1_error))
  /*else*/ {
    __pyx_t_7 = PyFloat_FromDouble(__pyx_v_v); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 673, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (unlikely(!__pyx_cur_scope->__pyx_v_max)) { __Pyx_RaiseClosureNameError("max"); __PYX_ERR(0, 673, __pyx_L1_error) }
    __pyx_t_8 = PyObject_RichCompare(__pyx_t_7, __pyx_cur_scope->__pyx_v_max, Py_GT); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 673, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 673, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(__pyx_t_12)) {

      /* "validr/_validator_c.pyx":674
 *         else:
 *             if v > max:
 *                 raise Invalid('value must <= %d' % max)             # <<<<<<<<<<<<<<
 *         return v
 *     return validate
 */
      __Pyx_TraceLine(674,0,__PYX_ERR(0, 674, __pyx_L1_error))
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 674, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (unlikely(!__pyx_cur_scope->__pyx_v_max)) { __Pyx_RaiseClosureNameError("max"); __PYX_ERR(0, 674, __pyx_L1_error) }
      __pyx_t_6 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_value_must_d_2, __pyx_cur_scope->__pyx_v_max); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 674, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_9 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      __pyx_t_8 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_9, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 674, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(0, 674, __pyx_L1_error)

      /* "validr/_validator_c.pyx":673
 *                 raise Invalid('value must < %d' % max)
 *         else:
 *             if v > max:             # <<<<<<<<<<<<<<
 *                 raise Invalid('value must <= %d' % max)
 *         return v
 */
    }
  }
  __pyx_L14:;

  /* "validr/_validator_c.pyx":675
 *             if v > max:
 *                 raise Invalid('value must <= %d' % max)
 *         return v             # <<<<<<<<<<<<<<
 *     return validate
 * 
 */
  __Pyx_TraceLine(675,0,__PYX_ERR(0, 675, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_8 = PyFloat_FromDouble(__pyx_v_v); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 675, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_r = __pyx_t_8;
  __pyx_t_8 = 0;
  goto __pyx_L0;

  /* "validr/_validator_c.pyx":658
 *         exmax = bool(exmax)
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         try:
 *             v = float(value)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("validr._validator_c.float_validator.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":636
 * 
 * @validator(accept=(int, float, str), output=float)
 * def float_validator(compiler, min=-sys.float_info.max, max=sys.float_info.max,             # <<<<<<<<<<<<<<
 *                     exmin=False, exmax=False):
 *     """Validate float string
 */

static PyObject *__pyx_pf_6validr_12_validator_c_48float_validator(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_compiler, PyObject *__pyx_v_min, PyObject *__pyx_v_max, PyObject *__pyx_v_exmin, PyObject *__pyx_v_exmax) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_15_float_validator *__pyx_cur_scope;
  PyObject *__pyx_v_validate = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  __Pyx_TraceFrameInit(__pyx_codeobj__62)
  __Pyx_RefNannySetupContext("float_validator", 0);
  __pyx_cur_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_15_float_validator *)__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_15_float_validator(__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_15_float_validator, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_15_float_validator *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 636, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __Pyx_TraceCall("float_validator", __pyx_f[0], 636, 0, __PYX_ERR(0, 636, __pyx_L1_error));
  __pyx_cur_scope->__pyx_v_min = __pyx_v_min;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_min);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_min);
  __pyx_cur_scope->__pyx_v_max = __pyx_v_max;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_max);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_max);
  __pyx_cur_scope->__pyx_v_exmin = __pyx_v_exmin;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_exmin);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_exmin);
  __pyx_cur_scope->__pyx_v_exmax = __pyx_v_exmax;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_exmax);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_exmax);

  /* "validr/_validator_c.pyx":646
 *         exmax (bool,float): exclude max value or not, default false
 *     """
 *     min, max = float(min), float(max)             # <<<<<<<<<<<<<<
 *     if isinstance(exmin, (int, float)) and not isinstance(exmin, bool):
 *         min = float(exmin)
 */
  __Pyx_TraceLine(646,0,__PYX_ERR(0, 646, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyNumber_Float(__pyx_cur_scope->__pyx_v_min); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyNumber_Float(__pyx_cur_scope->__pyx_v_max); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_min);
  __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_min, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;
  __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_max);
  __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_max, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "validr/_validator_c.pyx":647
 *     """
 *     min, max = float(min), float(max)
 *     if isinstance(exmin, (int, float)) and not isinstance(exmin, bool):             # <<<<<<<<<<<<<<
 *         min = float(exmin)
 *         exmin = True
 */
  __Pyx_TraceLine(647,0,__PYX_ERR(0, 647, __pyx_L1_error))
  __pyx_t_2 = __pyx_cur_scope->__pyx_v_exmin;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = PyInt_Check(__pyx_t_2); 
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_6 = (__pyx_t_5 != 0);
  if (!__pyx_t_6) {
  } else {
    __pyx_t_4 = __pyx_t_6;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_2 = __pyx_cur_scope->__pyx_v_exmin;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_6 = PyFloat_Check(__pyx_t_2); 
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = (__pyx_t_6 != 0);
  __pyx_t_4 = __pyx_t_5;
  __pyx_L6_bool_binop_done:;
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {
  } else {
    __pyx_t_3 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = __pyx_cur_scope->__pyx_v_exmin;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_1 = ((PyObject*)&PyBool_Type);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_5 = PyObject_IsInstance(__pyx_t_2, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 647, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = ((!(__pyx_t_5 != 0)) != 0);
  __pyx_t_3 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "validr/_validator_c.pyx":648
 *     min, max = float(min), float(max)
 *     if isinstance(exmin, (int, float)) and not isinstance(exmin, bool):
 *         min = float(exmin)             # <<<<<<<<<<<<<<
 *         exmin = True
 *     else:
 */
    __Pyx_TraceLine(648,0,__PYX_ERR(0, 648, __pyx_L1_error))
    __pyx_t_1 = __Pyx_PyNumber_Float(__pyx_cur_scope->__pyx_v_exmin); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 648, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_min);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_min, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;

    /* "validr/_validator_c.pyx":649
 *     if isinstance(exmin, (int, float)) and not isinstance(exmin, bool):
 *         min = float(exmin)
 *         exmin = True             # <<<<<<<<<<<<<<
 *     else:
 *         exmin = bool(exmin)
 */
    __Pyx_TraceLine(649,0,__PYX_ERR(0, 649, __pyx_L1_error))
    __Pyx_INCREF(Py_True);
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_exmin);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_exmin, Py_True);
    __Pyx_GIVEREF(Py_True);

    /* "validr/_validator_c.pyx":647
 *     """
 *     min, max = float(min), float(max)
 *     if isinstance(exmin, (int, float)) and not isinstance(exmin, bool):             # <<<<<<<<<<<<<<
 *         min = float(exmin)
 *         exmin = True
 */
    goto __pyx_L3;
  }

  /* "validr/_validator_c.pyx":651
 *         exmin = True
 *     else:
 *         exmin = bool(exmin)             # <<<<<<<<<<<<<<
 *     if isinstance(exmax, (int, float)) and not isinstance(exmax, bool):
 *         max = float(exmax)
 */
  __Pyx_TraceLine(651,0,__PYX_ERR(0, 651, __pyx_L1_error))
  /*else*/ {
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_exmin); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 651, __pyx_L1_error)
    __pyx_t_1 = __Pyx_PyBool_FromLong((!(!__pyx_t_3))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 651, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_exmin);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_exmin, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;
  }
  __pyx_L3:;

  /* "validr/_validator_c.pyx":652
 *     else:
 *         exmin = bool(exmin)
 *     if isinstance(exmax, (int, float)) and not isinstance(exmax, bool):             # <<<<<<<<<<<<<<
 *         max = float(exmax)
 *         exmax = True
 */
  __Pyx_TraceLine(652,0,__PYX_ERR(0, 652, __pyx_L1_error))
  __pyx_t_1 = __pyx_cur_scope->__pyx_v_exmax;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_5 = PyInt_Check(__pyx_t_1); 
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = (__pyx_t_5 != 0);
  if (!__pyx_t_6) {
  } else {
    __pyx_t_4 = __pyx_t_6;
    goto __pyx_L11_bool_binop_done;
  }
  __pyx_t_1 = __pyx_cur_scope->__pyx_v_exmax;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_6 = PyFloat_Check(__pyx_t_1); 
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = (__pyx_t_6 != 0);
  __pyx_t_4 = __pyx_t_5;
  __pyx_L11_bool_binop_done:;
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {
  } else {
    __pyx_t_3 = __pyx_t_5;
    goto __pyx_L9_bool_binop_done;
  }
  __pyx_t_1 = __pyx_cur_scope->__pyx_v_exmax;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = ((PyObject*)&PyBool_Type);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = PyObject_IsInstance(__pyx_t_1, __pyx_t_2); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 652, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = ((!(__pyx_t_5 != 0)) != 0);
  __pyx_t_3 = __pyx_t_4;
  __pyx_L9_bool_binop_done:;
  if (__pyx_t_3) {

    /* "validr/_validator_c.pyx":653
 *         exmin = bool(exmin)
 *     if isinstance(exmax, (int, float)) and not isinstance(exmax, bool):
 *         max = float(exmax)             # <<<<<<<<<<<<<<
 *         exmax = True
 *     else:
 */
    __Pyx_TraceLine(653,0,__PYX_ERR(0, 653, __pyx_L1_error))
    __pyx_t_2 = __Pyx_PyNumber_Float(__pyx_cur_scope->__pyx_v_exmax); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 653, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_max);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_max, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    __pyx_t_2 = 0;

    /* "validr/_validator_c.pyx":654
 *     if isinstance(exmax, (int, float)) and not isinstance(exmax, bool):
 *         max = float(exmax)
 *         exmax = True             # <<<<<<<<<<<<<<
 *     else:
 *         exmax = bool(exmax)
 */
    __Pyx_TraceLine(654,0,__PYX_ERR(0, 654, __pyx_L1_error))
    __Pyx_INCREF(Py_True);
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_exmax);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_exmax, Py_True);
    __Pyx_GIVEREF(Py_True);

    /* "validr/_validator_c.pyx":652
 *     else:
 *         exmin = bool(exmin)
 *     if isinstance(exmax, (int, float)) and not isinstance(exmax, bool):             # <<<<<<<<<<<<<<
 *         max = float(exmax)
 *         exmax = True
 */
    goto __pyx_L8;
  }

  /* "validr/_validator_c.pyx":656
 *         exmax = True
 *     else:
 *         exmax = bool(exmax)             # <<<<<<<<<<<<<<
 * 
 *     def validate(value):
 */
  __Pyx_TraceLine(656,0,__PYX_ERR(0, 656, __pyx_L1_error))
  /*else*/ {
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_exmax); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 656, __pyx_L1_error)
    __pyx_t_2 = __Pyx_PyBool_FromLong((!(!__pyx_t_3))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 656, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_exmax);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_exmax, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    __pyx_t_2 = 0;
  }
  __pyx_L8:;

  /* "validr/_validator_c.pyx":658
 *         exmax = bool(exmax)
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         try:
 *             v = float(value)
 */
  __Pyx_TraceLine(658,0,__PYX_ERR(0, 658, __pyx_L1_error))
  __pyx_t_2 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_15float_validator_1validate, 0, __pyx_n_s_float_validator_locals_validate, ((PyObject*)__pyx_cur_scope), __pyx_n_s_validr__validator_c, __pyx_d, ((PyObject *)__pyx_codeobj__64)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_validate = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "validr/_validator_c.pyx":676
 *                 raise Invalid('value must <= %d' % max)
 *         return v
 *     return validate             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(676,0,__PYX_ERR(0, 676, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_validate);
  __pyx_r = __pyx_v_validate;
  goto __pyx_L0;

  /* "validr/_validator_c.pyx":636
 * 
 * @validator(accept=(int, float, str), output=float)
 * def float_validator(compiler, min=-sys.float_info.max, max=sys.float_info.max,             # <<<<<<<<<<<<<<
 *                     exmin=False, exmax=False):
 *     """Validate float string
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("validr._validator_c.float_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_validate);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":680
 * 
 * @validator(accept=(str, object), output=str)
 * def str_validator(compiler, int minlen=0, int maxlen=1024 * 1024,             # <<<<<<<<<<<<<<
 *                   bint strip=False, bint escape=False, str match=None,
 *                   bint accept_object=False):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_51str_validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6validr_12_validator_c_50str_validator[] = "Validate string\n\n    Args:\n        minlen (int): min length of string, default 0\n        maxlen (int): max length of string, default 1024*1024\n        strip (bool): strip white space or not, default false\n        escape (bool): escape to html safe string or not, default false\n        match (str): regex to match, default None\n    ";
static PyMethodDef __pyx_mdef_6validr_12_validator_c_51str_validator = {"str_validator", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6validr_12_validator_c_51str_validator, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6validr_12_validator_c_50str_validator};
static PyObject *__pyx_pw_6validr_12_validator_c_51str_validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_compiler = 0;
  int __pyx_v_minlen;
  int __pyx_v_maxlen;
  int __pyx_v_strip;
  int __pyx_v_escape;
  PyObject *__pyx_v_match = 0;
  int __pyx_v_accept_object;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("str_validator (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_compiler,&__pyx_n_s_minlen,&__pyx_n_s_maxlen,&__pyx_n_s_strip,&__pyx_n_s_escape,&__pyx_n_s_match,&__pyx_n_s_accept_object,0};
    PyObject* values[7] = {0,0,0,0,0,0,0};

    /* "validr/_validator_c.pyx":681
 * @validator(accept=(str, object), output=str)
 * def str_validator(compiler, int minlen=0, int maxlen=1024 * 1024,
 *                   bint strip=False, bint escape=False, str match=None,             # <<<<<<<<<<<<<<
 *                   bint accept_object=False):
 *     """Validate string
 */
    values[5] = ((PyObject*)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_compiler)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_minlen);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_maxlen);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_strip);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_escape);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_match);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_accept_object);
          if (value) { values[6] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "str_validator") < 0)) __PYX_ERR(0, 680, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_compiler = values[0];
    if (values[1]) {
      __pyx_v_minlen = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_minlen == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 680, __pyx_L3_error)
    } else {
      __pyx_v_minlen = ((int)0);
    }
    if (values[2]) {
      __pyx_v_maxlen = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_maxlen == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 680, __pyx_L3_error)
    } else {
      __pyx_v_maxlen = ((int)0x100000);
    }
    if (values[3]) {
      __pyx_v_strip = __Pyx_PyObject_IsTrue(values[3]); if (unlikely((__pyx_v_strip == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 681, __pyx_L3_error)
    } else {
      __pyx_v_strip = ((int)0);
    }
    if (values[4]) {
      __pyx_v_escape = __Pyx_PyObject_IsTrue(values[4]); if (unlikely((__pyx_v_escape == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 681, __pyx_L3_error)
    } else {
      __pyx_v_escape = ((int)0);
    }
    __pyx_v_match = ((PyObject*)values[5]);
    if (values[6]) {
      __pyx_v_accept_object = __Pyx_PyObject_IsTrue(values[6]); if (unlikely((__pyx_v_accept_object == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 682, __pyx_L3_error)
    } else {

      /* "validr/_validator_c.pyx":682
 * def str_validator(compiler, int minlen=0, int maxlen=1024 * 1024,
 *                   bint strip=False, bint escape=False, str match=None,
 *                   bint accept_object=False):             # <<<<<<<<<<<<<<
 *     """Validate string
 * 
 */
      __pyx_v_accept_object = ((int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("str_validator", 0, 1, 7, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 680, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("validr._validator_c.str_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_match), (&PyUnicode_Type), 1, "match", 1))) __PYX_ERR(0, 681, __pyx_L1_error)
  __pyx_r = __pyx_pf_6validr_12_validator_c_50str_validator(__pyx_self, __pyx_v_compiler, __pyx_v_minlen, __pyx_v_maxlen, __pyx_v_strip, __pyx_v_escape, __pyx_v_match, __pyx_v_accept_object);

  /* "validr/_validator_c.pyx":680
 * 
 * @validator(accept=(str, object), output=str)
 * def str_validator(compiler, int minlen=0, int maxlen=1024 * 1024,             # <<<<<<<<<<<<<<
 *                   bint strip=False, bint escape=False, str match=None,
 *                   bint accept_object=False):
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":702
 *         re_match = None
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         if not isinstance(value, str):
 *             if accept_object or isinstance(value, int):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_13str_validator_1validate(PyObject *__pyx_self, PyObject *__pyx_v_value); /*proto*/
static PyMethodDef __pyx_mdef_6validr_12_validator_c_13str_validator_1validate = {"validate", (PyCFunction)__pyx_pw_6validr_12_validator_c_13str_validator_1validate, METH_O, 0};
static PyObject *__pyx_pw_6validr_12_validator_c_13str_validator_1validate(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("validate (wrapper)", 0);
  __pyx_r = __pyx_pf_6validr_12_validator_c_13str_validator_validate(__pyx_self, ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6validr_12_validator_c_13str_validator_validate(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_16_str_validator *__pyx_cur_scope;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_16_str_validator *__pyx_outer_scope;
  int __pyx_v_length;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("validate", 0);
  __pyx_outer_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_16_str_validator *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_TraceCall("validate", __pyx_f[0], 702, 0, __PYX_ERR(0, 702, __pyx_L1_error));
  __Pyx_INCREF(__pyx_v_value);

  /* "validr/_validator_c.pyx":703
 * 
 *     def validate(value):
 *         if not isinstance(value, str):             # <<<<<<<<<<<<<<
 *             if accept_object or isinstance(value, int):
 *                 value = str(value)
 */
  __Pyx_TraceLine(703,0,__PYX_ERR(0, 703, __pyx_L1_error))
  __pyx_t_1 = PyUnicode_Check(__pyx_v_value); 
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (__pyx_t_2) {

    /* "validr/_validator_c.pyx":704
 *     def validate(value):
 *         if not isinstance(value, str):
 *             if accept_object or isinstance(value, int):             # <<<<<<<<<<<<<<
 *                 value = str(value)
 *             else:
 */
    __Pyx_TraceLine(704,0,__PYX_ERR(0, 704, __pyx_L1_error))
    __pyx_t_1 = (__pyx_cur_scope->__pyx_v_accept_object != 0);
    if (!__pyx_t_1) {
    } else {
      __pyx_t_2 = __pyx_t_1;
      goto __pyx_L5_bool_binop_done;
    }
    __pyx_t_1 = PyInt_Check(__pyx_v_value); 
    __pyx_t_3 = (__pyx_t_1 != 0);
    __pyx_t_2 = __pyx_t_3;
    __pyx_L5_bool_binop_done:;
    if (likely(__pyx_t_2)) {

      /* "validr/_validator_c.pyx":705
 *         if not isinstance(value, str):
 *             if accept_object or isinstance(value, int):
 *                 value = str(value)             # <<<<<<<<<<<<<<
 *             else:
 *                 raise Invalid('invalid string')
 */
      __Pyx_TraceLine(705,0,__PYX_ERR(0, 705, __pyx_L1_error))
      __pyx_t_4 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyUnicode_Type)), __pyx_v_value); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 705, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "validr/_validator_c.pyx":704
 *     def validate(value):
 *         if not isinstance(value, str):
 *             if accept_object or isinstance(value, int):             # <<<<<<<<<<<<<<
 *                 value = str(value)
 *             else:
 */
      goto __pyx_L4;
    }

    /* "validr/_validator_c.pyx":707
 *                 value = str(value)
 *             else:
 *                 raise Invalid('invalid string')             # <<<<<<<<<<<<<<
 *         if strip:
 *             value = value.strip()
 */
    __Pyx_TraceLine(707,0,__PYX_ERR(0, 707, __pyx_L1_error))
    /*else*/ {
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 707, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_kp_u_invalid_string) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_kp_u_invalid_string);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 707, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_Raise(__pyx_t_4, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __PYX_ERR(0, 707, __pyx_L1_error)
    }
    __pyx_L4:;

    /* "validr/_validator_c.pyx":703
 * 
 *     def validate(value):
 *         if not isinstance(value, str):             # <<<<<<<<<<<<<<
 *             if accept_object or isinstance(value, int):
 *                 value = str(value)
 */
  }

  /* "validr/_validator_c.pyx":708
 *             else:
 *                 raise Invalid('invalid string')
 *         if strip:             # <<<<<<<<<<<<<<
 *             value = value.strip()
 *         cdef int length = len(value)
 */
  __Pyx_TraceLine(708,0,__PYX_ERR(0, 708, __pyx_L1_error))
  __pyx_t_2 = (__pyx_cur_scope->__pyx_v_strip != 0);
  if (__pyx_t_2) {

    /* "validr/_validator_c.pyx":709
 *                 raise Invalid('invalid string')
 *         if strip:
 *             value = value.strip()             # <<<<<<<<<<<<<<
 *         cdef int length = len(value)
 *         if length < minlen:
 */
    __Pyx_TraceLine(709,0,__PYX_ERR(0, 709, __pyx_L1_error))
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_strip); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 709, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 709, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "validr/_validator_c.pyx":708
 *             else:
 *                 raise Invalid('invalid string')
 *         if strip:             # <<<<<<<<<<<<<<
 *             value = value.strip()
 *         cdef int length = len(value)
 */
  }

  /* "validr/_validator_c.pyx":710
 *         if strip:
 *             value = value.strip()
 *         cdef int length = len(value)             # <<<<<<<<<<<<<<
 *         if length < minlen:
 *             raise Invalid('string length must >= %d' % minlen)
 */
  __Pyx_TraceLine(710,0,__PYX_ERR(0, 710, __pyx_L1_error))
  __pyx_t_7 = PyObject_Length(__pyx_v_value); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 710, __pyx_L1_error)
  __pyx_v_length = __pyx_t_7;

  /* "validr/_validator_c.pyx":711
 *             value = value.strip()
 *         cdef int length = len(value)
 *         if length < minlen:             # <<<<<<<<<<<<<<
 *             raise Invalid('string length must >= %d' % minlen)
 *         elif length > maxlen:
 */
  __Pyx_TraceLine(711,0,__PYX_ERR(0, 711, __pyx_L1_error))
  __pyx_t_2 = ((__pyx_v_length < __pyx_cur_scope->__pyx_v_minlen) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "validr/_validator_c.pyx":712
 *         cdef int length = len(value)
 *         if length < minlen:
 *             raise Invalid('string length must >= %d' % minlen)             # <<<<<<<<<<<<<<
 *         elif length > maxlen:
 *             raise Invalid('string length must <= %d' % maxlen)
 */
    __Pyx_TraceLine(712,0,__PYX_ERR(0, 712, __pyx_L1_error))
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 712, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_cur_scope->__pyx_v_minlen); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 712, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = PyUnicode_Format(__pyx_kp_u_string_length_must_d, __pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 712, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_8);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 712, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 712, __pyx_L1_error)

    /* "validr/_validator_c.pyx":711
 *             value = value.strip()
 *         cdef int length = len(value)
 *         if length < minlen:             # <<<<<<<<<<<<<<
 *             raise Invalid('string length must >= %d' % minlen)
 *         elif length > maxlen:
 */
  }

  /* "validr/_validator_c.pyx":713
 *         if length < minlen:
 *             raise Invalid('string length must >= %d' % minlen)
 *         elif length > maxlen:             # <<<<<<<<<<<<<<
 *             raise Invalid('string length must <= %d' % maxlen)
 *         if escape:
 */
  __Pyx_TraceLine(713,0,__PYX_ERR(0, 713, __pyx_L1_error))
  __pyx_t_2 = ((__pyx_v_length > __pyx_cur_scope->__pyx_v_maxlen) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "validr/_validator_c.pyx":714
 *             raise Invalid('string length must >= %d' % minlen)
 *         elif length > maxlen:
 *             raise Invalid('string length must <= %d' % maxlen)             # <<<<<<<<<<<<<<
 *         if escape:
 *             value = (value.replace('&', '&amp;')
 */
    __Pyx_TraceLine(714,0,__PYX_ERR(0, 714, __pyx_L1_error))
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 714, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_cur_scope->__pyx_v_maxlen); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 714, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_6 = PyUnicode_Format(__pyx_kp_u_string_length_must_d_2, __pyx_t_8); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 714, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_4 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 714, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 714, __pyx_L1_error)

    /* "validr/_validator_c.pyx":713
 *         if length < minlen:
 *             raise Invalid('string length must >= %d' % minlen)
 *         elif length > maxlen:             # <<<<<<<<<<<<<<
 *             raise Invalid('string length must <= %d' % maxlen)
 *         if escape:
 */
  }

  /* "validr/_validator_c.pyx":715
 *         elif length > maxlen:
 *             raise Invalid('string length must <= %d' % maxlen)
 *         if escape:             # <<<<<<<<<<<<<<
 *             value = (value.replace('&', '&amp;')
 *                      .replace('>', '&gt;')
 */
  __Pyx_TraceLine(715,0,__PYX_ERR(0, 715, __pyx_L1_error))
  __pyx_t_2 = (__pyx_cur_scope->__pyx_v_escape != 0);
  if (__pyx_t_2) {

    /* "validr/_validator_c.pyx":716
 *             raise Invalid('string length must <= %d' % maxlen)
 *         if escape:
 *             value = (value.replace('&', '&amp;')             # <<<<<<<<<<<<<<
 *                      .replace('>', '&gt;')
 *                      .replace('<', '&lt;')
 */
    __Pyx_TraceLine(716,0,__PYX_ERR(0, 716, __pyx_L1_error))
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_replace); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 716, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__66, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 716, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "validr/_validator_c.pyx":717
 *         if escape:
 *             value = (value.replace('&', '&amp;')
 *                      .replace('>', '&gt;')             # <<<<<<<<<<<<<<
 *                      .replace('<', '&lt;')
 *                      .replace("'", '&#39;')
 */
    __Pyx_TraceLine(717,0,__PYX_ERR(0, 717, __pyx_L1_error))
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_replace); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 717, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__68, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 717, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "validr/_validator_c.pyx":718
 *             value = (value.replace('&', '&amp;')
 *                      .replace('>', '&gt;')
 *                      .replace('<', '&lt;')             # <<<<<<<<<<<<<<
 *                      .replace("'", '&#39;')
 *                      .replace('"', '&#34;'))
 */
    __Pyx_TraceLine(718,0,__PYX_ERR(0, 718, __pyx_L1_error))
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_replace); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 718, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__70, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 718, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "validr/_validator_c.pyx":719
 *                      .replace('>', '&gt;')
 *                      .replace('<', '&lt;')
 *                      .replace("'", '&#39;')             # <<<<<<<<<<<<<<
 *                      .replace('"', '&#34;'))
 *         if re_match is not None and not re_match(value):
 */
    __Pyx_TraceLine(719,0,__PYX_ERR(0, 719, __pyx_L1_error))
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_replace); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 719, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__72, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 719, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "validr/_validator_c.pyx":720
 *                      .replace('<', '&lt;')
 *                      .replace("'", '&#39;')
 *                      .replace('"', '&#34;'))             # <<<<<<<<<<<<<<
 *         if re_match is not None and not re_match(value):
 *             raise Invalid("string not match regex %s" % match)
 */
    __Pyx_TraceLine(720,0,__PYX_ERR(0, 720, __pyx_L1_error))
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_replace); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 720, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__74, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 720, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "validr/_validator_c.pyx":715
 *         elif length > maxlen:
 *             raise Invalid('string length must <= %d' % maxlen)
 *         if escape:             # <<<<<<<<<<<<<<
 *             value = (value.replace('&', '&amp;')
 *                      .replace('>', '&gt;')
 */
  }

  /* "validr/_validator_c.pyx":721
 *                      .replace("'", '&#39;')
 *                      .replace('"', '&#34;'))
 *         if re_match is not None and not re_match(value):             # <<<<<<<<<<<<<<
 *             raise Invalid("string not match regex %s" % match)
 *         return value
 */
  __Pyx_TraceLine(721,0,__PYX_ERR(0, 721, __pyx_L1_error))
  if (unlikely(!__pyx_cur_scope->__pyx_v_re_match)) { __Pyx_RaiseClosureNameError("re_match"); __PYX_ERR(0, 721, __pyx_L1_error) }
  __pyx_t_3 = (__pyx_cur_scope->__pyx_v_re_match != Py_None);
  __pyx_t_1 = (__pyx_t_3 != 0);
  if (__pyx_t_1) {
  } else {
    __pyx_t_2 = __pyx_t_1;
    goto __pyx_L11_bool_binop_done;
  }
  if (unlikely(!__pyx_cur_scope->__pyx_v_re_match)) { __Pyx_RaiseClosureNameError("re_match"); __PYX_ERR(0, 721, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_re_match);
  __pyx_t_4 = __pyx_cur_scope->__pyx_v_re_match; __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_5 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_6, __pyx_v_value) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_value);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 721, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 721, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_3 = ((!__pyx_t_1) != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L11_bool_binop_done:;
  if (unlikely(__pyx_t_2)) {

    /* "validr/_validator_c.pyx":722
 *                      .replace('"', '&#34;'))
 *         if re_match is not None and not re_match(value):
 *             raise Invalid("string not match regex %s" % match)             # <<<<<<<<<<<<<<
 *         return value
 *     return validate
 */
    __Pyx_TraceLine(722,0,__PYX_ERR(0, 722, __pyx_L1_error))
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 722, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (unlikely(!__pyx_cur_scope->__pyx_v_match)) { __Pyx_RaiseClosureNameError("match"); __PYX_ERR(0, 722, __pyx_L1_error) }
    __pyx_t_6 = PyUnicode_Format(__pyx_kp_u_string_not_match_regex_s, __pyx_cur_scope->__pyx_v_match); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 722, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_5 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 722, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(0, 722, __pyx_L1_error)

    /* "validr/_validator_c.pyx":721
 *                      .replace("'", '&#39;')
 *                      .replace('"', '&#34;'))
 *         if re_match is not None and not re_match(value):             # <<<<<<<<<<<<<<
 *             raise Invalid("string not match regex %s" % match)
 *         return value
 */
  }

  /* "validr/_validator_c.pyx":723
 *         if re_match is not None and not re_match(value):
 *             raise Invalid("string not match regex %s" % match)
 *         return value             # <<<<<<<<<<<<<<
 *     return validate
 * 
 */
  __Pyx_TraceLine(723,0,__PYX_ERR(0, 723, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_value);
  __pyx_r = __pyx_v_value;
  goto __pyx_L0;

  /* "validr/_validator_c.pyx":702
 *         re_match = None
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         if not isinstance(value, str):
 *             if accept_object or isinstance(value, int):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("validr._validator_c.str_validator.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":680
 * 
 * @validator(accept=(str, object), output=str)
 * def str_validator(compiler, int minlen=0, int maxlen=1024 * 1024,             # <<<<<<<<<<<<<<
 *                   bint strip=False, bint escape=False, str match=None,
 *                   bint accept_object=False):
 */

static PyObject *__pyx_pf_6validr_12_validator_c_50str_validator(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_compiler, int __pyx_v_minlen, int __pyx_v_maxlen, int __pyx_v_strip, int __pyx_v_escape, PyObject *__pyx_v_match, int __pyx_v_accept_object) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_16_str_validator *__pyx_cur_scope;
  PyObject *__pyx_v_ex = NULL;
  PyObject *__pyx_v_validate = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  char const *__pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  __Pyx_TraceFrameInit(__pyx_codeobj__75)
  __Pyx_RefNannySetupContext("str_validator", 0);
  __pyx_cur_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_16_str_validator *)__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_16_str_validator(__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_16_str_validator, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_16_str_validator *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 680, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __Pyx_TraceCall("str_validator", __pyx_f[0], 680, 0, __PYX_ERR(0, 680, __pyx_L1_error));
  __pyx_cur_scope->__pyx_v_minlen = __pyx_v_minlen;
  __pyx_cur_scope->__pyx_v_maxlen = __pyx_v_maxlen;
  __pyx_cur_scope->__pyx_v_strip = __pyx_v_strip;
  __pyx_cur_scope->__pyx_v_escape = __pyx_v_escape;
  __pyx_cur_scope->__pyx_v_match = __pyx_v_match;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_match);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_match);
  __pyx_cur_scope->__pyx_v_accept_object = __pyx_v_accept_object;

  /* "validr/_validator_c.pyx":694
 * 
 *     # To make sure that the entire string matches
 *     if match:             # <<<<<<<<<<<<<<
 *         try:
 *             re_match = re.compile(r'(?:%s)\Z' % match).match
 */
  __Pyx_TraceLine(694,0,__PYX_ERR(0, 694, __pyx_L1_error))
  __pyx_t_1 = (__pyx_cur_scope->__pyx_v_match != Py_None)&&(__Pyx_PyUnicode_IS_TRUE(__pyx_cur_scope->__pyx_v_match) != 0);
  if (__pyx_t_1) {

    /* "validr/_validator_c.pyx":695
 *     # To make sure that the entire string matches
 *     if match:
 *         try:             # <<<<<<<<<<<<<<
 *             re_match = re.compile(r'(?:%s)\Z' % match).match
 *         except Exception as ex:
 */
    __Pyx_TraceLine(695,0,__PYX_ERR(0, 695, __pyx_L1_error))
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      /*try:*/ {

        /* "validr/_validator_c.pyx":696
 *     if match:
 *         try:
 *             re_match = re.compile(r'(?:%s)\Z' % match).match             # <<<<<<<<<<<<<<
 *         except Exception as ex:
 *             raise SchemaError('match regex %s compile failed' % match) from ex
 */
        __Pyx_TraceLine(696,0,__PYX_ERR(0, 696, __pyx_L4_error))
        __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_re); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 696, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_compile); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 696, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_6 = PyUnicode_Format(__pyx_kp_u_s_Z, __pyx_cur_scope->__pyx_v_match); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 696, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_8 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_8)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_8);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        __pyx_t_5 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6);
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 696, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 696, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GIVEREF(__pyx_t_7);
        __pyx_cur_scope->__pyx_v_re_match = __pyx_t_7;
        __pyx_t_7 = 0;

        /* "validr/_validator_c.pyx":695
 *     # To make sure that the entire string matches
 *     if match:
 *         try:             # <<<<<<<<<<<<<<
 *             re_match = re.compile(r'(?:%s)\Z' % match).match
 *         except Exception as ex:
 */
      }
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L9_try_end;
      __pyx_L4_error:;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "validr/_validator_c.pyx":697
 *         try:
 *             re_match = re.compile(r'(?:%s)\Z' % match).match
 *         except Exception as ex:             # <<<<<<<<<<<<<<
 *             raise SchemaError('match regex %s compile failed' % match) from ex
 *     else:
 */
      __Pyx_TraceLine(697,0,__PYX_ERR(0, 697, __pyx_L6_except_error))
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
      if (__pyx_t_9) {
        __Pyx_AddTraceback("validr._validator_c.str_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_5, &__pyx_t_6) < 0) __PYX_ERR(0, 697, __pyx_L6_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_5);
        __pyx_v_ex = __pyx_t_5;
        /*try:*/ {

          /* "validr/_validator_c.pyx":698
 *             re_match = re.compile(r'(?:%s)\Z' % match).match
 *         except Exception as ex:
 *             raise SchemaError('match regex %s compile failed' % match) from ex             # <<<<<<<<<<<<<<
 *     else:
 *         re_match = None
 */
          __Pyx_TraceLine(698,0,__PYX_ERR(0, 698, __pyx_L15_error))
          __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_SchemaError); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 698, __pyx_L15_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_11 = PyUnicode_Format(__pyx_kp_u_match_regex_s_compile_failed, __pyx_cur_scope->__pyx_v_match); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 698, __pyx_L15_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_12 = NULL;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
            __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_10);
            if (likely(__pyx_t_12)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
              __Pyx_INCREF(__pyx_t_12);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_10, function);
            }
          }
          __pyx_t_8 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_12, __pyx_t_11) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_t_11);
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 698, __pyx_L15_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_Raise(__pyx_t_8, 0, 0, __pyx_v_ex);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __PYX_ERR(0, 698, __pyx_L15_error)
        }

        /* "validr/_validator_c.pyx":697
 *         try:
 *             re_match = re.compile(r'(?:%s)\Z' % match).match
 *         except Exception as ex:             # <<<<<<<<<<<<<<
 *             raise SchemaError('match regex %s compile failed' % match) from ex
 *     else:
 */
        __Pyx_TraceLine(697,0,__PYX_ERR(0, 697, __pyx_L15_error))
        /*finally:*/ {
          __pyx_L15_error:;
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17) < 0)) __Pyx_ErrFetch(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __pyx_t_9 = __pyx_lineno; __pyx_t_13 = __pyx_clineno; __pyx_t_14 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_ex);
              __pyx_v_ex = NULL;
            }
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
            }
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_ErrRestore(__pyx_t_15, __pyx_t_16, __pyx_t_17);
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __pyx_lineno = __pyx_t_9; __pyx_clineno = __pyx_t_13; __pyx_filename = __pyx_t_14;
            goto __pyx_L6_except_error;
          }
        }
      }
      goto __pyx_L6_except_error;
      __pyx_L6_except_error:;

      /* "validr/_validator_c.pyx":695
 *     # To make sure that the entire string matches
 *     if match:
 *         try:             # <<<<<<<<<<<<<<
 *             re_match = re.compile(r'(?:%s)\Z' % match).match
 *         except Exception as ex:
 */
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      goto __pyx_L1_error;
      __pyx_L9_try_end:;
    }

    /* "validr/_validator_c.pyx":694
 * 
 *     # To make sure that the entire string matches
 *     if match:             # <<<<<<<<<<<<<<
 *         try:
 *             re_match = re.compile(r'(?:%s)\Z' % match).match
 */
    goto __pyx_L3;
  }

  /* "validr/_validator_c.pyx":700
 *             raise SchemaError('match regex %s compile failed' % match) from ex
 *     else:
 *         re_match = None             # <<<<<<<<<<<<<<
 * 
 *     def validate(value):
 */
  __Pyx_TraceLine(700,0,__PYX_ERR(0, 700, __pyx_L1_error))
  /*else*/ {
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __pyx_cur_scope->__pyx_v_re_match = Py_None;
  }
  __pyx_L3:;

  /* "validr/_validator_c.pyx":702
 *         re_match = None
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         if not isinstance(value, str):
 *             if accept_object or isinstance(value, int):
 */
  __Pyx_TraceLine(702,0,__PYX_ERR(0, 702, __pyx_L1_error))
  __pyx_t_6 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_13str_validator_1validate, 0, __pyx_n_s_str_validator_locals_validate, ((PyObject*)__pyx_cur_scope), __pyx_n_s_validr__validator_c, __pyx_d, ((PyObject *)__pyx_codeobj__77)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 702, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_validate = __pyx_t_6;
  __pyx_t_6 = 0;

  /* "validr/_validator_c.pyx":724
 *             raise Invalid("string not match regex %s" % match)
 *         return value
 *     return validate             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(724,0,__PYX_ERR(0, 724, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_validate);
  __pyx_r = __pyx_v_validate;
  goto __pyx_L0;

  /* "validr/_validator_c.pyx":680
 * 
 * @validator(accept=(str, object), output=str)
 * def str_validator(compiler, int minlen=0, int maxlen=1024 * 1024,             # <<<<<<<<<<<<<<
 *                   bint strip=False, bint escape=False, str match=None,
 *                   bint accept_object=False):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("validr._validator_c.str_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ex);
  __Pyx_XDECREF(__pyx_v_validate);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":728
 * 
 * @validator(accept=bytes, output=bytes)
 * def bytes_validator(compiler, int minlen=0, int maxlen=-1):             # <<<<<<<<<<<<<<
 *     """Validate bytes
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_53bytes_validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6validr_12_validator_c_52bytes_validator[] = "Validate bytes\n\n    Args:\n        minlen (int): min length of bytes, default 0\n        maxlen (int): max length of bytes, default unlimited\n    ";
static PyMethodDef __pyx_mdef_6validr_12_validator_c_53bytes_validator = {"bytes_validator", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6validr_12_validator_c_53bytes_validator, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6validr_12_validator_c_52bytes_validator};
static PyObject *__pyx_pw_6validr_12_validator_c_53bytes_validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_compiler = 0;
  int __pyx_v_minlen;
  int __pyx_v_maxlen;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("bytes_validator (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_compiler,&__pyx_n_s_minlen,&__pyx_n_s_maxlen,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_compiler)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_minlen);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_maxlen);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "bytes_validator") < 0)) __PYX_ERR(0, 728, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_compiler = values[0];
    if (values[1]) {
      __pyx_v_minlen = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_minlen == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 728, __pyx_L3_error)
    } else {
      __pyx_v_minlen = ((int)0);
    }
    if (values[2]) {
      __pyx_v_maxlen = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_maxlen == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 728, __pyx_L3_error)
    } else {
      __pyx_v_maxlen = ((int)-1);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("bytes_validator", 0, 1, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 728, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("validr._validator_c.bytes_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6validr_12_validator_c_52bytes_validator(__pyx_self, __pyx_v_compiler, __pyx_v_minlen, __pyx_v_maxlen);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":736
 *     """
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         if not isinstance(value, bytes):
 *             raise Invalid('invalid bytes')
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_15bytes_validator_1validate(PyObject *__pyx_self, PyObject *__pyx_v_value); /*proto*/
static PyMethodDef __pyx_mdef_6validr_12_validator_c_15bytes_validator_1validate = {"validate", (PyCFunction)__pyx_pw_6validr_12_validator_c_15bytes_validator_1validate, METH_O, 0};
static PyObject *__pyx_pw_6validr_12_validator_c_15bytes_validator_1validate(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("validate (wrapper)", 0);
  __pyx_r = __pyx_pf_6validr_12_validator_c_15bytes_validator_validate(__pyx_self, ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6validr_12_validator_c_15bytes_validator_validate(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_17_bytes_validator *__pyx_cur_scope;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_17_bytes_validator *__pyx_outer_scope;
  int __pyx_v_length;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_RefNannySetupContext("validate", 0);
  __pyx_outer_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_17_bytes_validator *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_TraceCall("validate", __pyx_f[0], 736, 0, __PYX_ERR(0, 736, __pyx_L1_error));

  /* "validr/_validator_c.pyx":737
 * 
 *     def validate(value):
 *         if not isinstance(value, bytes):             # <<<<<<<<<<<<<<
 *             raise Invalid('invalid bytes')
 *         cdef int length = len(value)
 */
  __Pyx_TraceLine(737,0,__PYX_ERR(0, 737, __pyx_L1_error))
  __pyx_t_1 = PyBytes_Check(__pyx_v_value); 
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "validr/_validator_c.pyx":738
 *     def validate(value):
 *         if not isinstance(value, bytes):
 *             raise Invalid('invalid bytes')             # <<<<<<<<<<<<<<
 *         cdef int length = len(value)
 *         if length < minlen:
 */
    __Pyx_TraceLine(738,0,__PYX_ERR(0, 738, __pyx_L1_error))
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 738, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_kp_u_invalid_bytes) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_kp_u_invalid_bytes);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 738, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 738, __pyx_L1_error)

    /* "validr/_validator_c.pyx":737
 * 
 *     def validate(value):
 *         if not isinstance(value, bytes):             # <<<<<<<<<<<<<<
 *             raise Invalid('invalid bytes')
 *         cdef int length = len(value)
 */
  }

  /* "validr/_validator_c.pyx":739
 *         if not isinstance(value, bytes):
 *             raise Invalid('invalid bytes')
 *         cdef int length = len(value)             # <<<<<<<<<<<<<<
 *         if length < minlen:
 *             raise Invalid('bytes length must >= %d' % minlen)
 */
  __Pyx_TraceLine(739,0,__PYX_ERR(0, 739, __pyx_L1_error))
  __pyx_t_6 = PyObject_Length(__pyx_v_value); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 739, __pyx_L1_error)
  __pyx_v_length = __pyx_t_6;

  /* "validr/_validator_c.pyx":740
 *             raise Invalid('invalid bytes')
 *         cdef int length = len(value)
 *         if length < minlen:             # <<<<<<<<<<<<<<
 *             raise Invalid('bytes length must >= %d' % minlen)
 *         elif maxlen > -1 and length > maxlen:
 */
  __Pyx_TraceLine(740,0,__PYX_ERR(0, 740, __pyx_L1_error))
  __pyx_t_2 = ((__pyx_v_length < __pyx_cur_scope->__pyx_v_minlen) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "validr/_validator_c.pyx":741
 *         cdef int length = len(value)
 *         if length < minlen:
 *             raise Invalid('bytes length must >= %d' % minlen)             # <<<<<<<<<<<<<<
 *         elif maxlen > -1 and length > maxlen:
 *             raise Invalid('bytes length must <= %d' % maxlen)
 */
    __Pyx_TraceLine(741,0,__PYX_ERR(0, 741, __pyx_L1_error))
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 741, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_cur_scope->__pyx_v_minlen); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 741, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = PyUnicode_Format(__pyx_kp_u_bytes_length_must_d, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 741, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_7);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 741, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 741, __pyx_L1_error)

    /* "validr/_validator_c.pyx":740
 *             raise Invalid('invalid bytes')
 *         cdef int length = len(value)
 *         if length < minlen:             # <<<<<<<<<<<<<<
 *             raise Invalid('bytes length must >= %d' % minlen)
 *         elif maxlen > -1 and length > maxlen:
 */
  }

  /* "validr/_validator_c.pyx":742
 *         if length < minlen:
 *             raise Invalid('bytes length must >= %d' % minlen)
 *         elif maxlen > -1 and length > maxlen:             # <<<<<<<<<<<<<<
 *             raise Invalid('bytes length must <= %d' % maxlen)
 *         return value
 */
  __Pyx_TraceLine(742,0,__PYX_ERR(0, 742, __pyx_L1_error))
  __pyx_t_1 = ((__pyx_cur_scope->__pyx_v_maxlen > -1L) != 0);
  if (__pyx_t_1) {
  } else {
    __pyx_t_2 = __pyx_t_1;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_1 = ((__pyx_v_length > __pyx_cur_scope->__pyx_v_maxlen) != 0);
  __pyx_t_2 = __pyx_t_1;
  __pyx_L5_bool_binop_done:;
  if (unlikely(__pyx_t_2)) {

    /* "validr/_validator_c.pyx":743
 *             raise Invalid('bytes length must >= %d' % minlen)
 *         elif maxlen > -1 and length > maxlen:
 *             raise Invalid('bytes length must <= %d' % maxlen)             # <<<<<<<<<<<<<<
 *         return value
 * 
 */
    __Pyx_TraceLine(743,0,__PYX_ERR(0, 743, __pyx_L1_error))
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 743, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_cur_scope->__pyx_v_maxlen); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 743, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_5 = PyUnicode_Format(__pyx_kp_u_bytes_length_must_d_2, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 743, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_7, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 743, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 743, __pyx_L1_error)

    /* "validr/_validator_c.pyx":742
 *         if length < minlen:
 *             raise Invalid('bytes length must >= %d' % minlen)
 *         elif maxlen > -1 and length > maxlen:             # <<<<<<<<<<<<<<
 *             raise Invalid('bytes length must <= %d' % maxlen)
 *         return value
 */
  }

  /* "validr/_validator_c.pyx":744
 *         elif maxlen > -1 and length > maxlen:
 *             raise Invalid('bytes length must <= %d' % maxlen)
 *         return value             # <<<<<<<<<<<<<<
 * 
 *     return validate
 */
  __Pyx_TraceLine(744,0,__PYX_ERR(0, 744, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_value);
  __pyx_r = __pyx_v_value;
  goto __pyx_L0;

  /* "validr/_validator_c.pyx":736
 *     """
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         if not isinstance(value, bytes):
 *             raise Invalid('invalid bytes')
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("validr._validator_c.bytes_validator.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":728
 * 
 * @validator(accept=bytes, output=bytes)
 * def bytes_validator(compiler, int minlen=0, int maxlen=-1):             # <<<<<<<<<<<<<<
 *     """Validate bytes
 * 
 */

static PyObject *__pyx_pf_6validr_12_validator_c_52bytes_validator(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_compiler, int __pyx_v_minlen, int __pyx_v_maxlen) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_17_bytes_validator *__pyx_cur_scope;
  PyObject *__pyx_v_validate = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_TraceFrameInit(__pyx_codeobj__78)
  __Pyx_RefNannySetupContext("bytes_validator", 0);
  __pyx_cur_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_17_bytes_validator *)__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_17_bytes_validator(__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_17_bytes_validator, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_17_bytes_validator *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 728, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __Pyx_TraceCall("bytes_validator", __pyx_f[0], 728, 0, __PYX_ERR(0, 728, __pyx_L1_error));
  __pyx_cur_scope->__pyx_v_minlen = __pyx_v_minlen;
  __pyx_cur_scope->__pyx_v_maxlen = __pyx_v_maxlen;

  /* "validr/_validator_c.pyx":736
 *     """
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         if not isinstance(value, bytes):
 *             raise Invalid('invalid bytes')
 */
  __Pyx_TraceLine(736,0,__PYX_ERR(0, 736, __pyx_L1_error))
  __pyx_t_1 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_15bytes_validator_1validate, 0, __pyx_n_s_bytes_validator_locals_validate, ((PyObject*)__pyx_cur_scope), __pyx_n_s_validr__validator_c, __pyx_d, ((PyObject *)__pyx_codeobj__80)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 736, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_validate = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":746
 *         return value
 * 
 *     return validate             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(746,0,__PYX_ERR(0, 746, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_validate);
  __pyx_r = __pyx_v_validate;
  goto __pyx_L0;

  /* "validr/_validator_c.pyx":728
 * 
 * @validator(accept=bytes, output=bytes)
 * def bytes_validator(compiler, int minlen=0, int maxlen=-1):             # <<<<<<<<<<<<<<
 *     """Validate bytes
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("validr._validator_c.bytes_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_validate);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":750
 * 
 * @validator(accept=(str, datetime.date), output=(str, datetime.date))
 * def date_validator(compiler, str format='%Y-%m-%d', bint output_object=False):             # <<<<<<<<<<<<<<
 *     """Validate date string or convert date to string
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_55date_validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6validr_12_validator_c_54date_validator[] = "Validate date string or convert date to string\n\n    Args:\n        format (str): date format, default ISO8601 format\n    ";
static PyMethodDef __pyx_mdef_6validr_12_validator_c_55date_validator = {"date_validator", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6validr_12_validator_c_55date_validator, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6validr_12_validator_c_54date_validator};
static PyObject *__pyx_pw_6validr_12_validator_c_55date_validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_compiler = 0;
  PyObject *__pyx_v_format = 0;
  int __pyx_v_output_object;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("date_validator (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_compiler,&__pyx_n_s_format,&__pyx_n_s_output_object,0};
    PyObject* values[3] = {0,0,0};
    values[1] = ((PyObject*)__pyx_kp_u_Y_m_d);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_compiler)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_format);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_output_object);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "date_validator") < 0)) __PYX_ERR(0, 750, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_compiler = values[0];
    __pyx_v_format = ((PyObject*)values[1]);
    if (values[2]) {
      __pyx_v_output_object = __Pyx_PyObject_IsTrue(values[2]); if (unlikely((__pyx_v_output_object == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 750, __pyx_L3_error)
    } else {
      __pyx_v_output_object = ((int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("date_validator", 0, 1, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 750, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("validr._validator_c.date_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_format), (&PyUnicode_Type), 1, "format", 1))) __PYX_ERR(0, 750, __pyx_L1_error)
  __pyx_r = __pyx_pf_6validr_12_validator_c_54date_validator(__pyx_self, __pyx_v_compiler, __pyx_v_format, __pyx_v_output_object);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":756
 *         format (str): date format, default ISO8601 format
 *     """
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         try:
 *             if not isinstance(value, (datetime.datetime, datetime.date)):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_14date_validator_1validate(PyObject *__pyx_self, PyObject *__pyx_v_value); /*proto*/
static PyMethodDef __pyx_mdef_6validr_12_validator_c_14date_validator_1validate = {"validate", (PyCFunction)__pyx_pw_6validr_12_validator_c_14date_validator_1validate, METH_O, 0};
static PyObject *__pyx_pw_6validr_12_validator_c_14date_validator_1validate(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("validate (wrapper)", 0);
  __pyx_r = __pyx_pf_6validr_12_validator_c_14date_validator_validate(__pyx_self, ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6validr_12_validator_c_14date_validator_validate(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_18_date_validator *__pyx_cur_scope;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_18_date_validator *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  __Pyx_RefNannySetupContext("validate", 0);
  __pyx_outer_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_18_date_validator *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_TraceCall("validate", __pyx_f[0], 756, 0, __PYX_ERR(0, 756, __pyx_L1_error));
  __Pyx_INCREF(__pyx_v_value);

  /* "validr/_validator_c.pyx":757
 *     """
 *     def validate(value):
 *         try:             # <<<<<<<<<<<<<<
 *             if not isinstance(value, (datetime.datetime, datetime.date)):
 *                 value = datetime.datetime.strptime(value, format)
 */
  __Pyx_TraceLine(757,0,__PYX_ERR(0, 757, __pyx_L1_error))
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "validr/_validator_c.pyx":758
 *     def validate(value):
 *         try:
 *             if not isinstance(value, (datetime.datetime, datetime.date)):             # <<<<<<<<<<<<<<
 *                 value = datetime.datetime.strptime(value, format)
 *             if isinstance(value, datetime.datetime):
 */
      __Pyx_TraceLine(758,0,__PYX_ERR(0, 758, __pyx_L3_error))
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_datetime); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 758, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_datetime); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 758, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_datetime); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 758, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_date); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 758, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_8 = PyObject_IsInstance(__pyx_v_value, __pyx_t_5); 
      __pyx_t_9 = (__pyx_t_8 != 0);
      if (!__pyx_t_9) {
      } else {
        __pyx_t_7 = __pyx_t_9;
        goto __pyx_L10_bool_binop_done;
      }
      __pyx_t_9 = PyObject_IsInstance(__pyx_v_value, __pyx_t_6); 
      __pyx_t_8 = (__pyx_t_9 != 0);
      __pyx_t_7 = __pyx_t_8;
      __pyx_L10_bool_binop_done:;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_8 = ((!(__pyx_t_7 != 0)) != 0);
      if (__pyx_t_8) {

        /* "validr/_validator_c.pyx":759
 *         try:
 *             if not isinstance(value, (datetime.datetime, datetime.date)):
 *                 value = datetime.datetime.strptime(value, format)             # <<<<<<<<<<<<<<
 *             if isinstance(value, datetime.datetime):
 *                 value = value.date()
 */
        __Pyx_TraceLine(759,0,__PYX_ERR(0, 759, __pyx_L3_error))
        __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_datetime); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 759, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_datetime); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 759, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_strptime); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 759, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_cur_scope->__pyx_v_format)) { __Pyx_RaiseClosureNameError("format"); __PYX_ERR(0, 759, __pyx_L3_error) }
        __pyx_t_4 = NULL;
        __pyx_t_10 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
            __pyx_t_10 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_6)) {
          PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_value, __pyx_cur_scope->__pyx_v_format};
          __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 759, __pyx_L3_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_5);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
          PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_value, __pyx_cur_scope->__pyx_v_format};
          __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 759, __pyx_L3_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_5);
        } else
        #endif
        {
          __pyx_t_11 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 759, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_11);
          if (__pyx_t_4) {
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_4); __pyx_t_4 = NULL;
          }
          __Pyx_INCREF(__pyx_v_value);
          __Pyx_GIVEREF(__pyx_v_value);
          PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_10, __pyx_v_value);
          __Pyx_INCREF(__pyx_cur_scope->__pyx_v_format);
          __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_format);
          PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_10, __pyx_cur_scope->__pyx_v_format);
          __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_11, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 759, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        }
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_5);
        __pyx_t_5 = 0;

        /* "validr/_validator_c.pyx":758
 *     def validate(value):
 *         try:
 *             if not isinstance(value, (datetime.datetime, datetime.date)):             # <<<<<<<<<<<<<<
 *                 value = datetime.datetime.strptime(value, format)
 *             if isinstance(value, datetime.datetime):
 */
      }

      /* "validr/_validator_c.pyx":760
 *             if not isinstance(value, (datetime.datetime, datetime.date)):
 *                 value = datetime.datetime.strptime(value, format)
 *             if isinstance(value, datetime.datetime):             # <<<<<<<<<<<<<<
 *                 value = value.date()
 *             if output_object:
 */
      __Pyx_TraceLine(760,0,__PYX_ERR(0, 760, __pyx_L3_error))
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_datetime); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 760, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_datetime); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 760, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_8 = PyObject_IsInstance(__pyx_v_value, __pyx_t_6); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 760, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_7 = (__pyx_t_8 != 0);
      if (__pyx_t_7) {

        /* "validr/_validator_c.pyx":761
 *                 value = datetime.datetime.strptime(value, format)
 *             if isinstance(value, datetime.datetime):
 *                 value = value.date()             # <<<<<<<<<<<<<<
 *             if output_object:
 *                 return value
 */
        __Pyx_TraceLine(761,0,__PYX_ERR(0, 761, __pyx_L3_error))
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_date); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 761, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_11 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_11)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_11);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
          }
        }
        __pyx_t_6 = (__pyx_t_11) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_11) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 761, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_6);
        __pyx_t_6 = 0;

        /* "validr/_validator_c.pyx":760
 *             if not isinstance(value, (datetime.datetime, datetime.date)):
 *                 value = datetime.datetime.strptime(value, format)
 *             if isinstance(value, datetime.datetime):             # <<<<<<<<<<<<<<
 *                 value = value.date()
 *             if output_object:
 */
      }

      /* "validr/_validator_c.pyx":762
 *             if isinstance(value, datetime.datetime):
 *                 value = value.date()
 *             if output_object:             # <<<<<<<<<<<<<<
 *                 return value
 *             else:
 */
      __Pyx_TraceLine(762,0,__PYX_ERR(0, 762, __pyx_L3_error))
      __pyx_t_7 = (__pyx_cur_scope->__pyx_v_output_object != 0);
      if (__pyx_t_7) {

        /* "validr/_validator_c.pyx":763
 *                 value = value.date()
 *             if output_object:
 *                 return value             # <<<<<<<<<<<<<<
 *             else:
 *                 return value.strftime(format)
 */
        __Pyx_TraceLine(763,0,__PYX_ERR(0, 763, __pyx_L3_error))
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_v_value);
        __pyx_r = __pyx_v_value;
        goto __pyx_L7_try_return;

        /* "validr/_validator_c.pyx":762
 *             if isinstance(value, datetime.datetime):
 *                 value = value.date()
 *             if output_object:             # <<<<<<<<<<<<<<
 *                 return value
 *             else:
 */
      }

      /* "validr/_validator_c.pyx":765
 *                 return value
 *             else:
 *                 return value.strftime(format)             # <<<<<<<<<<<<<<
 *         except Exception:
 *             raise Invalid('invalid date') from None
 */
      __Pyx_TraceLine(765,0,__PYX_ERR(0, 765, __pyx_L3_error))
      /*else*/ {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_strftime); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 765, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_5);
        if (unlikely(!__pyx_cur_scope->__pyx_v_format)) { __Pyx_RaiseClosureNameError("format"); __PYX_ERR(0, 765, __pyx_L3_error) }
        __pyx_t_11 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_11)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_11);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
          }
        }
        __pyx_t_6 = (__pyx_t_11) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_11, __pyx_cur_scope->__pyx_v_format) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_cur_scope->__pyx_v_format);
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 765, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_r = __pyx_t_6;
        __pyx_t_6 = 0;
        goto __pyx_L7_try_return;
      }

      /* "validr/_validator_c.pyx":757
 *     """
 *     def validate(value):
 *         try:             # <<<<<<<<<<<<<<
 *             if not isinstance(value, (datetime.datetime, datetime.date)):
 *                 value = datetime.datetime.strptime(value, format)
 */
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "validr/_validator_c.pyx":766
 *             else:
 *                 return value.strftime(format)
 *         except Exception:             # <<<<<<<<<<<<<<
 *             raise Invalid('invalid date') from None
 *     return validate
 */
    __Pyx_TraceLine(766,0,__PYX_ERR(0, 766, __pyx_L5_except_error))
    __pyx_t_10 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_10) {
      __Pyx_AddTraceback("validr._validator_c.date_validator.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_5, &__pyx_t_11) < 0) __PYX_ERR(0, 766, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_11);

      /* "validr/_validator_c.pyx":767
 *                 return value.strftime(format)
 *         except Exception:
 *             raise Invalid('invalid date') from None             # <<<<<<<<<<<<<<
 *     return validate
 * 
 */
      __Pyx_TraceLine(767,0,__PYX_ERR(0, 767, __pyx_L5_except_error))
      __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 767, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_12);
      __pyx_t_13 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_12))) {
        __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_12);
        if (likely(__pyx_t_13)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
          __Pyx_INCREF(__pyx_t_13);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_12, function);
        }
      }
      __pyx_t_4 = (__pyx_t_13) ? __Pyx_PyObject_Call2Args(__pyx_t_12, __pyx_t_13, __pyx_kp_u_invalid_date) : __Pyx_PyObject_CallOneArg(__pyx_t_12, __pyx_kp_u_invalid_date);
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 767, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_Raise(__pyx_t_4, 0, 0, Py_None);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __PYX_ERR(0, 767, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "validr/_validator_c.pyx":757
 *     """
 *     def validate(value):
 *         try:             # <<<<<<<<<<<<<<
 *             if not isinstance(value, (datetime.datetime, datetime.date)):
 *                 value = datetime.datetime.strptime(value, format)
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L7_try_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
  }

  /* "validr/_validator_c.pyx":756
 *         format (str): date format, default ISO8601 format
 *     """
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         try:
 *             if not isinstance(value, (datetime.datetime, datetime.date)):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("validr._validator_c.date_validator.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":750
 * 
 * @validator(accept=(str, datetime.date), output=(str, datetime.date))
 * def date_validator(compiler, str format='%Y-%m-%d', bint output_object=False):             # <<<<<<<<<<<<<<
 *     """Validate date string or convert date to string
 * 
 */

static PyObject *__pyx_pf_6validr_12_validator_c_54date_validator(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_compiler, PyObject *__pyx_v_format, int __pyx_v_output_object) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_18_date_validator *__pyx_cur_scope;
  PyObject *__pyx_v_validate = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_TraceFrameInit(__pyx_codeobj__81)
  __Pyx_RefNannySetupContext("date_validator", 0);
  __pyx_cur_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_18_date_validator *)__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_18_date_validator(__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_18_date_validator, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_18_date_validator *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 750, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __Pyx_TraceCall("date_validator", __pyx_f[0], 750, 0, __PYX_ERR(0, 750, __pyx_L1_error));
  __pyx_cur_scope->__pyx_v_format = __pyx_v_format;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_format);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_format);
  __pyx_cur_scope->__pyx_v_output_object = __pyx_v_output_object;

  /* "validr/_validator_c.pyx":756
 *         format (str): date format, default ISO8601 format
 *     """
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         try:
 *             if not isinstance(value, (datetime.datetime, datetime.date)):
 */
  __Pyx_TraceLine(756,0,__PYX_ERR(0, 756, __pyx_L1_error))
  __pyx_t_1 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_14date_validator_1validate, 0, __pyx_n_s_date_validator_locals_validate, ((PyObject*)__pyx_cur_scope), __pyx_n_s_validr__validator_c, __pyx_d, ((PyObject *)__pyx_codeobj__83)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 756, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_validate = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":768
 *         except Exception:
 *             raise Invalid('invalid date') from None
 *     return validate             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(768,0,__PYX_ERR(0, 768, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_validate);
  __pyx_r = __pyx_v_validate;
  goto __pyx_L0;

  /* "validr/_validator_c.pyx":750
 * 
 * @validator(accept=(str, datetime.date), output=(str, datetime.date))
 * def date_validator(compiler, str format='%Y-%m-%d', bint output_object=False):             # <<<<<<<<<<<<<<
 *     """Validate date string or convert date to string
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("validr._validator_c.date_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_validate);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":772
 * 
 * @validator(accept=(str, datetime.time), output=(str, datetime.time))
 * def time_validator(compiler, str format='%H:%M:%S', bint output_object=False):             # <<<<<<<<<<<<<<
 *     """Validate time string or convert time to string
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_57time_validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6validr_12_validator_c_56time_validator[] = "Validate time string or convert time to string\n\n    Args:\n        format (str): time format, default ISO8601 format\n    ";
static PyMethodDef __pyx_mdef_6validr_12_validator_c_57time_validator = {"time_validator", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6validr_12_validator_c_57time_validator, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6validr_12_validator_c_56time_validator};
static PyObject *__pyx_pw_6validr_12_validator_c_57time_validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_compiler = 0;
  PyObject *__pyx_v_format = 0;
  int __pyx_v_output_object;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("time_validator (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_compiler,&__pyx_n_s_format,&__pyx_n_s_output_object,0};
    PyObject* values[3] = {0,0,0};
    values[1] = ((PyObject*)__pyx_kp_u_H_M_S);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_compiler)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_format);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_output_object);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "time_validator") < 0)) __PYX_ERR(0, 772, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_compiler = values[0];
    __pyx_v_format = ((PyObject*)values[1]);
    if (values[2]) {
      __pyx_v_output_object = __Pyx_PyObject_IsTrue(values[2]); if (unlikely((__pyx_v_output_object == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 772, __pyx_L3_error)
    } else {
      __pyx_v_output_object = ((int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("time_validator", 0, 1, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 772, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("validr._validator_c.time_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_format), (&PyUnicode_Type), 1, "format", 1))) __PYX_ERR(0, 772, __pyx_L1_error)
  __pyx_r = __pyx_pf_6validr_12_validator_c_56time_validator(__pyx_self, __pyx_v_compiler, __pyx_v_format, __pyx_v_output_object);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":778
 *         format (str): time format, default ISO8601 format
 *     """
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         try:
 *             if not isinstance(value, (datetime.datetime, datetime.time)):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_14time_validator_1validate(PyObject *__pyx_self, PyObject *__pyx_v_value); /*proto*/
static PyMethodDef __pyx_mdef_6validr_12_validator_c_14time_validator_1validate = {"validate", (PyCFunction)__pyx_pw_6validr_12_validator_c_14time_validator_1validate, METH_O, 0};
static PyObject *__pyx_pw_6validr_12_validator_c_14time_validator_1validate(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("validate (wrapper)", 0);
  __pyx_r = __pyx_pf_6validr_12_validator_c_14time_validator_validate(__pyx_self, ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6validr_12_validator_c_14time_validator_validate(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_19_time_validator *__pyx_cur_scope;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_19_time_validator *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  __Pyx_RefNannySetupContext("validate", 0);
  __pyx_outer_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_19_time_validator *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_TraceCall("validate", __pyx_f[0], 778, 0, __PYX_ERR(0, 778, __pyx_L1_error));
  __Pyx_INCREF(__pyx_v_value);

  /* "validr/_validator_c.pyx":779
 *     """
 *     def validate(value):
 *         try:             # <<<<<<<<<<<<<<
 *             if not isinstance(value, (datetime.datetime, datetime.time)):
 *                 value = datetime.datetime.strptime(value, format)
 */
  __Pyx_TraceLine(779,0,__PYX_ERR(0, 779, __pyx_L1_error))
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "validr/_validator_c.pyx":780
 *     def validate(value):
 *         try:
 *             if not isinstance(value, (datetime.datetime, datetime.time)):             # <<<<<<<<<<<<<<
 *                 value = datetime.datetime.strptime(value, format)
 *             if isinstance(value, datetime.datetime):
 */
      __Pyx_TraceLine(780,0,__PYX_ERR(0, 780, __pyx_L3_error))
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_datetime); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 780, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_datetime); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 780, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_datetime); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 780, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_time); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 780, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_8 = PyObject_IsInstance(__pyx_v_value, __pyx_t_5); 
      __pyx_t_9 = (__pyx_t_8 != 0);
      if (!__pyx_t_9) {
      } else {
        __pyx_t_7 = __pyx_t_9;
        goto __pyx_L10_bool_binop_done;
      }
      __pyx_t_9 = PyObject_IsInstance(__pyx_v_value, __pyx_t_6); 
      __pyx_t_8 = (__pyx_t_9 != 0);
      __pyx_t_7 = __pyx_t_8;
      __pyx_L10_bool_binop_done:;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_8 = ((!(__pyx_t_7 != 0)) != 0);
      if (__pyx_t_8) {

        /* "validr/_validator_c.pyx":781
 *         try:
 *             if not isinstance(value, (datetime.datetime, datetime.time)):
 *                 value = datetime.datetime.strptime(value, format)             # <<<<<<<<<<<<<<
 *             if isinstance(value, datetime.datetime):
 *                 value = value.time()
 */
        __Pyx_TraceLine(781,0,__PYX_ERR(0, 781, __pyx_L3_error))
        __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_datetime); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 781, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_datetime); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 781, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_strptime); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 781, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_cur_scope->__pyx_v_format)) { __Pyx_RaiseClosureNameError("format"); __PYX_ERR(0, 781, __pyx_L3_error) }
        __pyx_t_4 = NULL;
        __pyx_t_10 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
            __pyx_t_10 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_6)) {
          PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_value, __pyx_cur_scope->__pyx_v_format};
          __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 781, __pyx_L3_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_5);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
          PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_value, __pyx_cur_scope->__pyx_v_format};
          __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 781, __pyx_L3_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_5);
        } else
        #endif
        {
          __pyx_t_11 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 781, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_11);
          if (__pyx_t_4) {
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_4); __pyx_t_4 = NULL;
          }
          __Pyx_INCREF(__pyx_v_value);
          __Pyx_GIVEREF(__pyx_v_value);
          PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_10, __pyx_v_value);
          __Pyx_INCREF(__pyx_cur_scope->__pyx_v_format);
          __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_format);
          PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_10, __pyx_cur_scope->__pyx_v_format);
          __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_11, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 781, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        }
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_5);
        __pyx_t_5 = 0;

        /* "validr/_validator_c.pyx":780
 *     def validate(value):
 *         try:
 *             if not isinstance(value, (datetime.datetime, datetime.time)):             # <<<<<<<<<<<<<<
 *                 value = datetime.datetime.strptime(value, format)
 *             if isinstance(value, datetime.datetime):
 */
      }

      /* "validr/_validator_c.pyx":782
 *             if not isinstance(value, (datetime.datetime, datetime.time)):
 *                 value = datetime.datetime.strptime(value, format)
 *             if isinstance(value, datetime.datetime):             # <<<<<<<<<<<<<<
 *                 value = value.time()
 *             if output_object:
 */
      __Pyx_TraceLine(782,0,__PYX_ERR(0, 782, __pyx_L3_error))
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_datetime); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 782, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_datetime); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 782, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_8 = PyObject_IsInstance(__pyx_v_value, __pyx_t_6); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 782, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_7 = (__pyx_t_8 != 0);
      if (__pyx_t_7) {

        /* "validr/_validator_c.pyx":783
 *                 value = datetime.datetime.strptime(value, format)
 *             if isinstance(value, datetime.datetime):
 *                 value = value.time()             # <<<<<<<<<<<<<<
 *             if output_object:
 *                 return value
 */
        __Pyx_TraceLine(783,0,__PYX_ERR(0, 783, __pyx_L3_error))
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_time); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 783, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_11 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_11)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_11);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
          }
        }
        __pyx_t_6 = (__pyx_t_11) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_11) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 783, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_6);
        __pyx_t_6 = 0;

        /* "validr/_validator_c.pyx":782
 *             if not isinstance(value, (datetime.datetime, datetime.time)):
 *                 value = datetime.datetime.strptime(value, format)
 *             if isinstance(value, datetime.datetime):             # <<<<<<<<<<<<<<
 *                 value = value.time()
 *             if output_object:
 */
      }

      /* "validr/_validator_c.pyx":784
 *             if isinstance(value, datetime.datetime):
 *                 value = value.time()
 *             if output_object:             # <<<<<<<<<<<<<<
 *                 return value
 *             else:
 */
      __Pyx_TraceLine(784,0,__PYX_ERR(0, 784, __pyx_L3_error))
      __pyx_t_7 = (__pyx_cur_scope->__pyx_v_output_object != 0);
      if (__pyx_t_7) {

        /* "validr/_validator_c.pyx":785
 *                 value = value.time()
 *             if output_object:
 *                 return value             # <<<<<<<<<<<<<<
 *             else:
 *                 return value.strftime(format)
 */
        __Pyx_TraceLine(785,0,__PYX_ERR(0, 785, __pyx_L3_error))
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_v_value);
        __pyx_r = __pyx_v_value;
        goto __pyx_L7_try_return;

        /* "validr/_validator_c.pyx":784
 *             if isinstance(value, datetime.datetime):
 *                 value = value.time()
 *             if output_object:             # <<<<<<<<<<<<<<
 *                 return value
 *             else:
 */
      }

      /* "validr/_validator_c.pyx":787
 *                 return value
 *             else:
 *                 return value.strftime(format)             # <<<<<<<<<<<<<<
 *         except Exception:
 *             raise Invalid('invalid time') from None
 */
      __Pyx_TraceLine(787,0,__PYX_ERR(0, 787, __pyx_L3_error))
      /*else*/ {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_strftime); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 787, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_5);
        if (unlikely(!__pyx_cur_scope->__pyx_v_format)) { __Pyx_RaiseClosureNameError("format"); __PYX_ERR(0, 787, __pyx_L3_error) }
        __pyx_t_11 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_11)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_11);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
          }
        }
        __pyx_t_6 = (__pyx_t_11) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_11, __pyx_cur_scope->__pyx_v_format) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_cur_scope->__pyx_v_format);
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 787, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_r = __pyx_t_6;
        __pyx_t_6 = 0;
        goto __pyx_L7_try_return;
      }

      /* "validr/_validator_c.pyx":779
 *     """
 *     def validate(value):
 *         try:             # <<<<<<<<<<<<<<
 *             if not isinstance(value, (datetime.datetime, datetime.time)):
 *                 value = datetime.datetime.strptime(value, format)
 */
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "validr/_validator_c.pyx":788
 *             else:
 *                 return value.strftime(format)
 *         except Exception:             # <<<<<<<<<<<<<<
 *             raise Invalid('invalid time') from None
 *     return validate
 */
    __Pyx_TraceLine(788,0,__PYX_ERR(0, 788, __pyx_L5_except_error))
    __pyx_t_10 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_10) {
      __Pyx_AddTraceback("validr._validator_c.time_validator.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_5, &__pyx_t_11) < 0) __PYX_ERR(0, 788, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_11);

      /* "validr/_validator_c.pyx":789
 *                 return value.strftime(format)
 *         except Exception:
 *             raise Invalid('invalid time') from None             # <<<<<<<<<<<<<<
 *     return validate
 * 
 */
      __Pyx_TraceLine(789,0,__PYX_ERR(0, 789, __pyx_L5_except_error))
      __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 789, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_12);
      __pyx_t_13 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_12))) {
        __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_12);
        if (likely(__pyx_t_13)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
          __Pyx_INCREF(__pyx_t_13);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_12, function);
        }
      }
      __pyx_t_4 = (__pyx_t_13) ? __Pyx_PyObject_Call2Args(__pyx_t_12, __pyx_t_13, __pyx_kp_u_invalid_time) : __Pyx_PyObject_CallOneArg(__pyx_t_12, __pyx_kp_u_invalid_time);
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 789, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_Raise(__pyx_t_4, 0, 0, Py_None);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __PYX_ERR(0, 789, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "validr/_validator_c.pyx":779
 *     """
 *     def validate(value):
 *         try:             # <<<<<<<<<<<<<<
 *             if not isinstance(value, (datetime.datetime, datetime.time)):
 *                 value = datetime.datetime.strptime(value, format)
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L7_try_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
  }

  /* "validr/_validator_c.pyx":778
 *         format (str): time format, default ISO8601 format
 *     """
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         try:
 *             if not isinstance(value, (datetime.datetime, datetime.time)):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("validr._validator_c.time_validator.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":772
 * 
 * @validator(accept=(str, datetime.time), output=(str, datetime.time))
 * def time_validator(compiler, str format='%H:%M:%S', bint output_object=False):             # <<<<<<<<<<<<<<
 *     """Validate time string or convert time to string
 * 
 */

static PyObject *__pyx_pf_6validr_12_validator_c_56time_validator(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_compiler, PyObject *__pyx_v_format, int __pyx_v_output_object) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_19_time_validator *__pyx_cur_scope;
  PyObject *__pyx_v_validate = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_TraceFrameInit(__pyx_codeobj__84)
  __Pyx_RefNannySetupContext("time_validator", 0);
  __pyx_cur_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_19_time_validator *)__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_19_time_validator(__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_19_time_validator, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_19_time_validator *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 772, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __Pyx_TraceCall("time_validator", __pyx_f[0], 772, 0, __PYX_ERR(0, 772, __pyx_L1_error));
  __pyx_cur_scope->__pyx_v_format = __pyx_v_format;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_format);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_format);
  __pyx_cur_scope->__pyx_v_output_object = __pyx_v_output_object;

  /* "validr/_validator_c.pyx":778
 *         format (str): time format, default ISO8601 format
 *     """
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         try:
 *             if not isinstance(value, (datetime.datetime, datetime.time)):
 */
  __Pyx_TraceLine(778,0,__PYX_ERR(0, 778, __pyx_L1_error))
  __pyx_t_1 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_14time_validator_1validate, 0, __pyx_n_s_time_validator_locals_validate, ((PyObject*)__pyx_cur_scope), __pyx_n_s_validr__validator_c, __pyx_d, ((PyObject *)__pyx_codeobj__86)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 778, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_validate = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":790
 *         except Exception:
 *             raise Invalid('invalid time') from None
 *     return validate             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(790,0,__PYX_ERR(0, 790, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_validate);
  __pyx_r = __pyx_v_validate;
  goto __pyx_L0;

  /* "validr/_validator_c.pyx":772
 * 
 * @validator(accept=(str, datetime.time), output=(str, datetime.time))
 * def time_validator(compiler, str format='%H:%M:%S', bint output_object=False):             # <<<<<<<<<<<<<<
 *     """Validate time string or convert time to string
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("validr._validator_c.time_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_validate);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":794
 * 
 * @validator(accept=(str, datetime.datetime), output=(str, datetime.datetime))
 * def datetime_validator(compiler, str format='%Y-%m-%dT%H:%M:%S.%fZ', bint output_object=False):             # <<<<<<<<<<<<<<
 *     """Validate datetime string or convert datetime to string
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_59datetime_validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6validr_12_validator_c_58datetime_validator[] = "Validate datetime string or convert datetime to string\n\n    Args:\n        format (str): datetime format, default ISO8601 format\n    ";
static PyMethodDef __pyx_mdef_6validr_12_validator_c_59datetime_validator = {"datetime_validator", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6validr_12_validator_c_59datetime_validator, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6validr_12_validator_c_58datetime_validator};
static PyObject *__pyx_pw_6validr_12_validator_c_59datetime_validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_compiler = 0;
  PyObject *__pyx_v_format = 0;
  int __pyx_v_output_object;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("datetime_validator (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_compiler,&__pyx_n_s_format,&__pyx_n_s_output_object,0};
    PyObject* values[3] = {0,0,0};
    values[1] = ((PyObject*)__pyx_kp_u_Y_m_dT_H_M_S_fZ);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_compiler)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_format);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_output_object);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "datetime_validator") < 0)) __PYX_ERR(0, 794, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_compiler = values[0];
    __pyx_v_format = ((PyObject*)values[1]);
    if (values[2]) {
      __pyx_v_output_object = __Pyx_PyObject_IsTrue(values[2]); if (unlikely((__pyx_v_output_object == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 794, __pyx_L3_error)
    } else {
      __pyx_v_output_object = ((int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("datetime_validator", 0, 1, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 794, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("validr._validator_c.datetime_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_format), (&PyUnicode_Type), 1, "format", 1))) __PYX_ERR(0, 794, __pyx_L1_error)
  __pyx_r = __pyx_pf_6validr_12_validator_c_58datetime_validator(__pyx_self, __pyx_v_compiler, __pyx_v_format, __pyx_v_output_object);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":800
 *         format (str): datetime format, default ISO8601 format
 *     """
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         try:
 *             if isinstance(value, tuple):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_18datetime_validator_1validate(PyObject *__pyx_self, PyObject *__pyx_v_value); /*proto*/
static PyMethodDef __pyx_mdef_6validr_12_validator_c_18datetime_validator_1validate = {"validate", (PyCFunction)__pyx_pw_6validr_12_validator_c_18datetime_validator_1validate, METH_O, 0};
static PyObject *__pyx_pw_6validr_12_validator_c_18datetime_validator_1validate(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("validate (wrapper)", 0);
  __pyx_r = __pyx_pf_6validr_12_validator_c_18datetime_validator_validate(__pyx_self, ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6validr_12_validator_c_18datetime_validator_validate(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_20_datetime_validator *__pyx_cur_scope;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_20_datetime_validator *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  __Pyx_RefNannySetupContext("validate", 0);
  __pyx_outer_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_20_datetime_validator *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_TraceCall("validate", __pyx_f[0], 800, 0, __PYX_ERR(0, 800, __pyx_L1_error));
  __Pyx_INCREF(__pyx_v_value);

  /* "validr/_validator_c.pyx":801
 *     """
 *     def validate(value):
 *         try:             # <<<<<<<<<<<<<<
 *             if isinstance(value, tuple):
 *                 value = datetime.datetime.fromtimestamp(time.mktime(value))
 */
  __Pyx_TraceLine(801,0,__PYX_ERR(0, 801, __pyx_L1_error))
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "validr/_validator_c.pyx":802
 *     def validate(value):
 *         try:
 *             if isinstance(value, tuple):             # <<<<<<<<<<<<<<
 *                 value = datetime.datetime.fromtimestamp(time.mktime(value))
 *             elif not isinstance(value, datetime.datetime):
 */
      __Pyx_TraceLine(802,0,__PYX_ERR(0, 802, __pyx_L3_error))
      __pyx_t_4 = PyTuple_Check(__pyx_v_value); 
      __pyx_t_5 = (__pyx_t_4 != 0);
      if (__pyx_t_5) {

        /* "validr/_validator_c.pyx":803
 *         try:
 *             if isinstance(value, tuple):
 *                 value = datetime.datetime.fromtimestamp(time.mktime(value))             # <<<<<<<<<<<<<<
 *             elif not isinstance(value, datetime.datetime):
 *                 value = datetime.datetime.strptime(value, format)
 */
        __Pyx_TraceLine(803,0,__PYX_ERR(0, 803, __pyx_L3_error))
        __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_datetime); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 803, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_datetime); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 803, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_fromtimestamp); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 803, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_time); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 803, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_mktime); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 803, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_9 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
          __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_10);
          if (likely(__pyx_t_9)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
            __Pyx_INCREF(__pyx_t_9);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_10, function);
          }
        }
        __pyx_t_8 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_9, __pyx_v_value) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_v_value);
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 803, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_t_10 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_10)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_10);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        __pyx_t_6 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_10, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_8);
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 803, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_6);
        __pyx_t_6 = 0;

        /* "validr/_validator_c.pyx":802
 *     def validate(value):
 *         try:
 *             if isinstance(value, tuple):             # <<<<<<<<<<<<<<
 *                 value = datetime.datetime.fromtimestamp(time.mktime(value))
 *             elif not isinstance(value, datetime.datetime):
 */
        goto __pyx_L9;
      }

      /* "validr/_validator_c.pyx":804
 *             if isinstance(value, tuple):
 *                 value = datetime.datetime.fromtimestamp(time.mktime(value))
 *             elif not isinstance(value, datetime.datetime):             # <<<<<<<<<<<<<<
 *                 value = datetime.datetime.strptime(value, format)
 *             if output_object:
 */
      __Pyx_TraceLine(804,0,__PYX_ERR(0, 804, __pyx_L3_error))
      __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_datetime); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 804, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_datetime); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 804, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_5 = PyObject_IsInstance(__pyx_v_value, __pyx_t_7); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 804, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_4 = ((!(__pyx_t_5 != 0)) != 0);
      if (__pyx_t_4) {

        /* "validr/_validator_c.pyx":805
 *                 value = datetime.datetime.fromtimestamp(time.mktime(value))
 *             elif not isinstance(value, datetime.datetime):
 *                 value = datetime.datetime.strptime(value, format)             # <<<<<<<<<<<<<<
 *             if output_object:
 *                 return value
 */
        __Pyx_TraceLine(805,0,__PYX_ERR(0, 805, __pyx_L3_error))
        __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_datetime); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 805, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_datetime); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 805, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_strptime); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 805, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (unlikely(!__pyx_cur_scope->__pyx_v_format)) { __Pyx_RaiseClosureNameError("format"); __PYX_ERR(0, 805, __pyx_L3_error) }
        __pyx_t_8 = NULL;
        __pyx_t_11 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_8)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_8);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
            __pyx_t_11 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_6)) {
          PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_v_value, __pyx_cur_scope->__pyx_v_format};
          __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_11, 2+__pyx_t_11); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 805, __pyx_L3_error)
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_GOTREF(__pyx_t_7);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
          PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_v_value, __pyx_cur_scope->__pyx_v_format};
          __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_11, 2+__pyx_t_11); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 805, __pyx_L3_error)
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_GOTREF(__pyx_t_7);
        } else
        #endif
        {
          __pyx_t_10 = PyTuple_New(2+__pyx_t_11); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 805, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_10);
          if (__pyx_t_8) {
            __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
          }
          __Pyx_INCREF(__pyx_v_value);
          __Pyx_GIVEREF(__pyx_v_value);
          PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_11, __pyx_v_value);
          __Pyx_INCREF(__pyx_cur_scope->__pyx_v_format);
          __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_format);
          PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_11, __pyx_cur_scope->__pyx_v_format);
          __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_10, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 805, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        }
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_7);
        __pyx_t_7 = 0;

        /* "validr/_validator_c.pyx":804
 *             if isinstance(value, tuple):
 *                 value = datetime.datetime.fromtimestamp(time.mktime(value))
 *             elif not isinstance(value, datetime.datetime):             # <<<<<<<<<<<<<<
 *                 value = datetime.datetime.strptime(value, format)
 *             if output_object:
 */
      }
      __pyx_L9:;

      /* "validr/_validator_c.pyx":806
 *             elif not isinstance(value, datetime.datetime):
 *                 value = datetime.datetime.strptime(value, format)
 *             if output_object:             # <<<<<<<<<<<<<<
 *                 return value
 *             else:
 */
      __Pyx_TraceLine(806,0,__PYX_ERR(0, 806, __pyx_L3_error))
      __pyx_t_4 = (__pyx_cur_scope->__pyx_v_output_object != 0);
      if (__pyx_t_4) {

        /* "validr/_validator_c.pyx":807
 *                 value = datetime.datetime.strptime(value, format)
 *             if output_object:
 *                 return value             # <<<<<<<<<<<<<<
 *             else:
 *                 return value.strftime(format)
 */
        __Pyx_TraceLine(807,0,__PYX_ERR(0, 807, __pyx_L3_error))
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_v_value);
        __pyx_r = __pyx_v_value;
        goto __pyx_L7_try_return;

        /* "validr/_validator_c.pyx":806
 *             elif not isinstance(value, datetime.datetime):
 *                 value = datetime.datetime.strptime(value, format)
 *             if output_object:             # <<<<<<<<<<<<<<
 *                 return value
 *             else:
 */
      }

      /* "validr/_validator_c.pyx":809
 *                 return value
 *             else:
 *                 return value.strftime(format)             # <<<<<<<<<<<<<<
 *         except Exception:
 *             raise Invalid('invalid datetime') from None
 */
      __Pyx_TraceLine(809,0,__PYX_ERR(0, 809, __pyx_L3_error))
      /*else*/ {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_strftime); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 809, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_6);
        if (unlikely(!__pyx_cur_scope->__pyx_v_format)) { __Pyx_RaiseClosureNameError("format"); __PYX_ERR(0, 809, __pyx_L3_error) }
        __pyx_t_10 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_10)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_10);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
          }
        }
        __pyx_t_7 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_10, __pyx_cur_scope->__pyx_v_format) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_cur_scope->__pyx_v_format);
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 809, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_r = __pyx_t_7;
        __pyx_t_7 = 0;
        goto __pyx_L7_try_return;
      }

      /* "validr/_validator_c.pyx":801
 *     """
 *     def validate(value):
 *         try:             # <<<<<<<<<<<<<<
 *             if isinstance(value, tuple):
 *                 value = datetime.datetime.fromtimestamp(time.mktime(value))
 */
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "validr/_validator_c.pyx":810
 *             else:
 *                 return value.strftime(format)
 *         except Exception:             # <<<<<<<<<<<<<<
 *             raise Invalid('invalid datetime') from None
 *     return validate
 */
    __Pyx_TraceLine(810,0,__PYX_ERR(0, 810, __pyx_L5_except_error))
    __pyx_t_11 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_11) {
      __Pyx_AddTraceback("validr._validator_c.datetime_validator.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_6, &__pyx_t_10) < 0) __PYX_ERR(0, 810, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_10);

      /* "validr/_validator_c.pyx":811
 *                 return value.strftime(format)
 *         except Exception:
 *             raise Invalid('invalid datetime') from None             # <<<<<<<<<<<<<<
 *     return validate
 * 
 */
      __Pyx_TraceLine(811,0,__PYX_ERR(0, 811, __pyx_L5_except_error))
      __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 811, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_12 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
        __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_9);
        if (likely(__pyx_t_12)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
          __Pyx_INCREF(__pyx_t_12);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_9, function);
        }
      }
      __pyx_t_8 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_12, __pyx_kp_u_invalid_datetime) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_kp_u_invalid_datetime);
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 811, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_Raise(__pyx_t_8, 0, 0, Py_None);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(0, 811, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "validr/_validator_c.pyx":801
 *     """
 *     def validate(value):
 *         try:             # <<<<<<<<<<<<<<
 *             if isinstance(value, tuple):
 *                 value = datetime.datetime.fromtimestamp(time.mktime(value))
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L7_try_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
  }

  /* "validr/_validator_c.pyx":800
 *         format (str): datetime format, default ISO8601 format
 *     """
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         try:
 *             if isinstance(value, tuple):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("validr._validator_c.datetime_validator.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":794
 * 
 * @validator(accept=(str, datetime.datetime), output=(str, datetime.datetime))
 * def datetime_validator(compiler, str format='%Y-%m-%dT%H:%M:%S.%fZ', bint output_object=False):             # <<<<<<<<<<<<<<
 *     """Validate datetime string or convert datetime to string
 * 
 */

static PyObject *__pyx_pf_6validr_12_validator_c_58datetime_validator(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_compiler, PyObject *__pyx_v_format, int __pyx_v_output_object) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_20_datetime_validator *__pyx_cur_scope;
  PyObject *__pyx_v_validate = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_TraceFrameInit(__pyx_codeobj__87)
  __Pyx_RefNannySetupContext("datetime_validator", 0);
  __pyx_cur_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_20_datetime_validator *)__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_20_datetime_validator(__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_20_datetime_validator, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_20_datetime_validator *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 794, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __Pyx_TraceCall("datetime_validator", __pyx_f[0], 794, 0, __PYX_ERR(0, 794, __pyx_L1_error));
  __pyx_cur_scope->__pyx_v_format = __pyx_v_format;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_format);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_format);
  __pyx_cur_scope->__pyx_v_output_object = __pyx_v_output_object;

  /* "validr/_validator_c.pyx":800
 *         format (str): datetime format, default ISO8601 format
 *     """
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         try:
 *             if isinstance(value, tuple):
 */
  __Pyx_TraceLine(800,0,__PYX_ERR(0, 800, __pyx_L1_error))
  __pyx_t_1 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_18datetime_validator_1validate, 0, __pyx_n_s_datetime_validator_locals_valida, ((PyObject*)__pyx_cur_scope), __pyx_n_s_validr__validator_c, __pyx_d, ((PyObject *)__pyx_codeobj__89)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 800, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_validate = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":812
 *         except Exception:
 *             raise Invalid('invalid datetime') from None
 *     return validate             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(812,0,__PYX_ERR(0, 812, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_validate);
  __pyx_r = __pyx_v_validate;
  goto __pyx_L0;

  /* "validr/_validator_c.pyx":794
 * 
 * @validator(accept=(str, datetime.datetime), output=(str, datetime.datetime))
 * def datetime_validator(compiler, str format='%Y-%m-%dT%H:%M:%S.%fZ', bint output_object=False):             # <<<<<<<<<<<<<<
 *     """Validate datetime string or convert datetime to string
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("validr._validator_c.datetime_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_validate);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":815
 * 
 * 
 * cpdef _parse_timedelta(value):             # <<<<<<<<<<<<<<
 *     if isinstance(value, (int, float)):
 *         value = datetime.timedelta(seconds=value)
 */

static PyObject *__pyx_pw_6validr_12_validator_c_61_parse_timedelta(PyObject *__pyx_self, PyObject *__pyx_v_value); /*proto*/
static PyObject *__pyx_f_6validr_12_validator_c__parse_timedelta(PyObject *__pyx_v_value, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("_parse_timedelta", 0);
  __Pyx_TraceCall("_parse_timedelta", __pyx_f[0], 815, 0, __PYX_ERR(0, 815, __pyx_L1_error));
  __Pyx_INCREF(__pyx_v_value);

  /* "validr/_validator_c.pyx":816
 * 
 * cpdef _parse_timedelta(value):
 *     if isinstance(value, (int, float)):             # <<<<<<<<<<<<<<
 *         value = datetime.timedelta(seconds=value)
 *     elif isinstance(value, str):
 */
  __Pyx_TraceLine(816,0,__PYX_ERR(0, 816, __pyx_L1_error))
  __pyx_t_2 = PyInt_Check(__pyx_v_value); 
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (!__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = PyFloat_Check(__pyx_v_value); 
  __pyx_t_2 = (__pyx_t_3 != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "validr/_validator_c.pyx":817
 * cpdef _parse_timedelta(value):
 *     if isinstance(value, (int, float)):
 *         value = datetime.timedelta(seconds=value)             # <<<<<<<<<<<<<<
 *     elif isinstance(value, str):
 *         value = durationpy.from_str(value)
 */
    __Pyx_TraceLine(817,0,__PYX_ERR(0, 817, __pyx_L1_error))
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_datetime); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 817, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_timedelta); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 817, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 817, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_seconds, __pyx_v_value) < 0) __PYX_ERR(0, 817, __pyx_L1_error)
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_empty_tuple, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 817, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "validr/_validator_c.pyx":816
 * 
 * cpdef _parse_timedelta(value):
 *     if isinstance(value, (int, float)):             # <<<<<<<<<<<<<<
 *         value = datetime.timedelta(seconds=value)
 *     elif isinstance(value, str):
 */
    goto __pyx_L3;
  }

  /* "validr/_validator_c.pyx":818
 *     if isinstance(value, (int, float)):
 *         value = datetime.timedelta(seconds=value)
 *     elif isinstance(value, str):             # <<<<<<<<<<<<<<
 *         value = durationpy.from_str(value)
 *     else:
 */
  __Pyx_TraceLine(818,0,__PYX_ERR(0, 818, __pyx_L1_error))
  __pyx_t_2 = PyUnicode_Check(__pyx_v_value); 
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "validr/_validator_c.pyx":819
 *         value = datetime.timedelta(seconds=value)
 *     elif isinstance(value, str):
 *         value = durationpy.from_str(value)             # <<<<<<<<<<<<<<
 *     else:
 *         if not isinstance(value, datetime.timedelta):
 */
    __Pyx_TraceLine(819,0,__PYX_ERR(0, 819, __pyx_L1_error))
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_durationpy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 819, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_from_str); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 819, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_6 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_4, __pyx_v_value) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_value);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 819, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "validr/_validator_c.pyx":818
 *     if isinstance(value, (int, float)):
 *         value = datetime.timedelta(seconds=value)
 *     elif isinstance(value, str):             # <<<<<<<<<<<<<<
 *         value = durationpy.from_str(value)
 *     else:
 */
    goto __pyx_L3;
  }

  /* "validr/_validator_c.pyx":821
 *         value = durationpy.from_str(value)
 *     else:
 *         if not isinstance(value, datetime.timedelta):             # <<<<<<<<<<<<<<
 *             raise ValueError("invalid timedelta")
 *     return value
 */
  __Pyx_TraceLine(821,0,__PYX_ERR(0, 821, __pyx_L1_error))
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_datetime); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 821, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_timedelta); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 821, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_1 = PyObject_IsInstance(__pyx_v_value, __pyx_t_5); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 821, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
    if (unlikely(__pyx_t_2)) {

      /* "validr/_validator_c.pyx":822
 *     else:
 *         if not isinstance(value, datetime.timedelta):
 *             raise ValueError("invalid timedelta")             # <<<<<<<<<<<<<<
 *     return value
 * 
 */
      __Pyx_TraceLine(822,0,__PYX_ERR(0, 822, __pyx_L1_error))
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__90, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 822, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __PYX_ERR(0, 822, __pyx_L1_error)

      /* "validr/_validator_c.pyx":821
 *         value = durationpy.from_str(value)
 *     else:
 *         if not isinstance(value, datetime.timedelta):             # <<<<<<<<<<<<<<
 *             raise ValueError("invalid timedelta")
 *     return value
 */
    }
  }
  __pyx_L3:;

  /* "validr/_validator_c.pyx":823
 *         if not isinstance(value, datetime.timedelta):
 *             raise ValueError("invalid timedelta")
 *     return value             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(823,0,__PYX_ERR(0, 823, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_value);
  __pyx_r = __pyx_v_value;
  goto __pyx_L0;

  /* "validr/_validator_c.pyx":815
 * 
 * 
 * cpdef _parse_timedelta(value):             # <<<<<<<<<<<<<<
 *     if isinstance(value, (int, float)):
 *         value = datetime.timedelta(seconds=value)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("validr._validator_c._parse_timedelta", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_61_parse_timedelta(PyObject *__pyx_self, PyObject *__pyx_v_value); /*proto*/
static PyObject *__pyx_pw_6validr_12_validator_c_61_parse_timedelta(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_parse_timedelta (wrapper)", 0);
  __pyx_r = __pyx_pf_6validr_12_validator_c_60_parse_timedelta(__pyx_self, ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6validr_12_validator_c_60_parse_timedelta(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_parse_timedelta", 0);
  __Pyx_TraceCall("_parse_timedelta (wrapper)", __pyx_f[0], 815, 0, __PYX_ERR(0, 815, __pyx_L1_error));
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6validr_12_validator_c__parse_timedelta(__pyx_v_value, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 815, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("validr._validator_c._parse_timedelta", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":827
 * 
 * @validator(accept=(int, float, str, datetime.timedelta), output=(str, float, datetime.timedelta))
 * def timedelta_validator(compiler, min=None, max=None, bint string=False,             # <<<<<<<<<<<<<<
 *                         bint extended=False, bint output_object=False):
 *     """Validate timedelta string or convert timedelta to string
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_63timedelta_validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6validr_12_validator_c_62timedelta_validator[] = "Validate timedelta string or convert timedelta to string\n\n    Format (Go's Duration strings):\n        ns - nanoseconds\n        us - microseconds\n        ms - milliseconds\n        s - seconds\n        m - minutes\n        h - hours\n        d - days\n        mo - months\n        y - years\n    ";
static PyMethodDef __pyx_mdef_6validr_12_validator_c_63timedelta_validator = {"timedelta_validator", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6validr_12_validator_c_63timedelta_validator, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6validr_12_validator_c_62timedelta_validator};
static PyObject *__pyx_pw_6validr_12_validator_c_63timedelta_validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_compiler = 0;
  PyObject *__pyx_v_min = 0;
  PyObject *__pyx_v_max = 0;
  int __pyx_v_string;
  int __pyx_v_extended;
  int __pyx_v_output_object;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("timedelta_validator (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_compiler,&__pyx_n_s_min,&__pyx_n_s_max,&__pyx_n_s_string,&__pyx_n_s_extended,&__pyx_n_s_output_object,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    values[1] = ((PyObject *)Py_None);
    values[2] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_compiler)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_min);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_string);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_extended);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_output_object);
          if (value) { values[5] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "timedelta_validator") < 0)) __PYX_ERR(0, 827, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_compiler = values[0];
    __pyx_v_min = values[1];
    __pyx_v_max = values[2];
    if (values[3]) {
      __pyx_v_string = __Pyx_PyObject_IsTrue(values[3]); if (unlikely((__pyx_v_string == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 827, __pyx_L3_error)
    } else {
      __pyx_v_string = ((int)0);
    }
    if (values[4]) {
      __pyx_v_extended = __Pyx_PyObject_IsTrue(values[4]); if (unlikely((__pyx_v_extended == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 828, __pyx_L3_error)
    } else {

      /* "validr/_validator_c.pyx":828
 * @validator(accept=(int, float, str, datetime.timedelta), output=(str, float, datetime.timedelta))
 * def timedelta_validator(compiler, min=None, max=None, bint string=False,
 *                         bint extended=False, bint output_object=False):             # <<<<<<<<<<<<<<
 *     """Validate timedelta string or convert timedelta to string
 * 
 */
      __pyx_v_extended = ((int)0);
    }
    if (values[5]) {
      __pyx_v_output_object = __Pyx_PyObject_IsTrue(values[5]); if (unlikely((__pyx_v_output_object == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 828, __pyx_L3_error)
    } else {
      __pyx_v_output_object = ((int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("timedelta_validator", 0, 1, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 827, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("validr._validator_c.timedelta_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6validr_12_validator_c_62timedelta_validator(__pyx_self, __pyx_v_compiler, __pyx_v_min, __pyx_v_max, __pyx_v_string, __pyx_v_extended, __pyx_v_output_object);

  /* "validr/_validator_c.pyx":827
 * 
 * @validator(accept=(int, float, str, datetime.timedelta), output=(str, float, datetime.timedelta))
 * def timedelta_validator(compiler, min=None, max=None, bint string=False,             # <<<<<<<<<<<<<<
 *                         bint extended=False, bint output_object=False):
 *     """Validate timedelta string or convert timedelta to string
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":859
 *         max_repr = durationpy.to_str(max_value, extended=True)
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         try:
 *             value = _parse_timedelta(value)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_19timedelta_validator_1validate(PyObject *__pyx_self, PyObject *__pyx_v_value); /*proto*/
static PyMethodDef __pyx_mdef_6validr_12_validator_c_19timedelta_validator_1validate = {"validate", (PyCFunction)__pyx_pw_6validr_12_validator_c_19timedelta_validator_1validate, METH_O, 0};
static PyObject *__pyx_pw_6validr_12_validator_c_19timedelta_validator_1validate(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("validate (wrapper)", 0);
  __pyx_r = __pyx_pf_6validr_12_validator_c_19timedelta_validator_validate(__pyx_self, ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6validr_12_validator_c_19timedelta_validator_validate(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_21_timedelta_validator *__pyx_cur_scope;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_21_timedelta_validator *__pyx_outer_scope;
  PyObject *__pyx_v_ex = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  char const *__pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  int __pyx_t_19;
  int __pyx_t_20;
  __Pyx_RefNannySetupContext("validate", 0);
  __pyx_outer_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_21_timedelta_validator *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_TraceCall("validate", __pyx_f[0], 859, 0, __PYX_ERR(0, 859, __pyx_L1_error));
  __Pyx_INCREF(__pyx_v_value);

  /* "validr/_validator_c.pyx":860
 * 
 *     def validate(value):
 *         try:             # <<<<<<<<<<<<<<
 *             value = _parse_timedelta(value)
 *         except (durationpy.DurationError, ValueError, TypeError) as ex:
 */
  __Pyx_TraceLine(860,0,__PYX_ERR(0, 860, __pyx_L1_error))
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "validr/_validator_c.pyx":861
 *     def validate(value):
 *         try:
 *             value = _parse_timedelta(value)             # <<<<<<<<<<<<<<
 *         except (durationpy.DurationError, ValueError, TypeError) as ex:
 *             raise Invalid('invalid timedelta') from ex
 */
      __Pyx_TraceLine(861,0,__PYX_ERR(0, 861, __pyx_L3_error))
      __pyx_t_4 = __pyx_f_6validr_12_validator_c__parse_timedelta(__pyx_v_value, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 861, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "validr/_validator_c.pyx":860
 * 
 *     def validate(value):
 *         try:             # <<<<<<<<<<<<<<
 *             value = _parse_timedelta(value)
 *         except (durationpy.DurationError, ValueError, TypeError) as ex:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "validr/_validator_c.pyx":862
 *         try:
 *             value = _parse_timedelta(value)
 *         except (durationpy.DurationError, ValueError, TypeError) as ex:             # <<<<<<<<<<<<<<
 *             raise Invalid('invalid timedelta') from ex
 *         if min_value is not None:
 */
    __Pyx_TraceLine(862,0,__PYX_ERR(0, 862, __pyx_L5_except_error))
    __Pyx_ErrFetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_durationpy); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 862, __pyx_L5_except_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_DurationError); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 862, __pyx_L5_except_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_9 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_4, __pyx_t_8) || __Pyx_PyErr_GivenExceptionMatches(__pyx_t_4, __pyx_builtin_ValueError) || __Pyx_PyErr_GivenExceptionMatches(__pyx_t_4, __pyx_builtin_TypeError);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_ErrRestore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
    __pyx_t_4 = 0; __pyx_t_5 = 0; __pyx_t_6 = 0;
    if (__pyx_t_9) {
      __Pyx_AddTraceback("validr._validator_c.timedelta_validator.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_5, &__pyx_t_4) < 0) __PYX_ERR(0, 862, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __pyx_v_ex = __pyx_t_5;
      /*try:*/ {

        /* "validr/_validator_c.pyx":863
 *             value = _parse_timedelta(value)
 *         except (durationpy.DurationError, ValueError, TypeError) as ex:
 *             raise Invalid('invalid timedelta') from ex             # <<<<<<<<<<<<<<
 *         if min_value is not None:
 *             if value < min_value:
 */
        __Pyx_TraceLine(863,0,__PYX_ERR(0, 863, __pyx_L14_error))
        __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 863, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_10 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_10)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_10);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        __pyx_t_8 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_10, __pyx_kp_u_invalid_timedelta) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_kp_u_invalid_timedelta);
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 863, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_Raise(__pyx_t_8, 0, 0, __pyx_v_ex);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __PYX_ERR(0, 863, __pyx_L14_error)
      }

      /* "validr/_validator_c.pyx":862
 *         try:
 *             value = _parse_timedelta(value)
 *         except (durationpy.DurationError, ValueError, TypeError) as ex:             # <<<<<<<<<<<<<<
 *             raise Invalid('invalid timedelta') from ex
 *         if min_value is not None:
 */
      __Pyx_TraceLine(862,0,__PYX_ERR(0, 862, __pyx_L14_error))
      /*finally:*/ {
        __pyx_L14_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0;
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15) < 0)) __Pyx_ErrFetch(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_13);
          __Pyx_XGOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_18);
          __pyx_t_9 = __pyx_lineno; __pyx_t_11 = __pyx_clineno; __pyx_t_12 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_ex);
            __pyx_v_ex = NULL;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ExceptionReset(__pyx_t_16, __pyx_t_17, __pyx_t_18);
          }
          __Pyx_XGIVEREF(__pyx_t_13);
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_ErrRestore(__pyx_t_13, __pyx_t_14, __pyx_t_15);
          __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0;
          __pyx_lineno = __pyx_t_9; __pyx_clineno = __pyx_t_11; __pyx_filename = __pyx_t_12;
          goto __pyx_L5_except_error;
        }
      }
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "validr/_validator_c.pyx":860
 * 
 *     def validate(value):
 *         try:             # <<<<<<<<<<<<<<
 *             value = _parse_timedelta(value)
 *         except (durationpy.DurationError, ValueError, TypeError) as ex:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "validr/_validator_c.pyx":864
 *         except (durationpy.DurationError, ValueError, TypeError) as ex:
 *             raise Invalid('invalid timedelta') from ex
 *         if min_value is not None:             # <<<<<<<<<<<<<<
 *             if value < min_value:
 *                 raise Invalid('value must >= {}'.format(min_repr))
 */
  __Pyx_TraceLine(864,0,__PYX_ERR(0, 864, __pyx_L1_error))
  if (unlikely(!__pyx_cur_scope->__pyx_v_min_value)) { __Pyx_RaiseClosureNameError("min_value"); __PYX_ERR(0, 864, __pyx_L1_error) }
  __pyx_t_19 = (__pyx_cur_scope->__pyx_v_min_value != Py_None);
  __pyx_t_20 = (__pyx_t_19 != 0);
  if (__pyx_t_20) {

    /* "validr/_validator_c.pyx":865
 *             raise Invalid('invalid timedelta') from ex
 *         if min_value is not None:
 *             if value < min_value:             # <<<<<<<<<<<<<<
 *                 raise Invalid('value must >= {}'.format(min_repr))
 *         if max_value is not None:
 */
    __Pyx_TraceLine(865,0,__PYX_ERR(0, 865, __pyx_L1_error))
    if (unlikely(!__pyx_cur_scope->__pyx_v_min_value)) { __Pyx_RaiseClosureNameError("min_value"); __PYX_ERR(0, 865, __pyx_L1_error) }
    __pyx_t_4 = PyObject_RichCompare(__pyx_v_value, __pyx_cur_scope->__pyx_v_min_value, Py_LT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 865, __pyx_L1_error)
    __pyx_t_20 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_20 < 0)) __PYX_ERR(0, 865, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(__pyx_t_20)) {

      /* "validr/_validator_c.pyx":866
 *         if min_value is not None:
 *             if value < min_value:
 *                 raise Invalid('value must >= {}'.format(min_repr))             # <<<<<<<<<<<<<<
 *         if max_value is not None:
 *             if value > max_value:
 */
      __Pyx_TraceLine(866,0,__PYX_ERR(0, 866, __pyx_L1_error))
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 866, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_value_must, __pyx_n_s_format); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 866, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (unlikely(!__pyx_cur_scope->__pyx_v_min_repr)) { __Pyx_RaiseClosureNameError("min_repr"); __PYX_ERR(0, 866, __pyx_L1_error) }
      __pyx_t_7 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
        }
      }
      __pyx_t_6 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_7, __pyx_cur_scope->__pyx_v_min_repr) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_cur_scope->__pyx_v_min_repr);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 866, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_4 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 866, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_Raise(__pyx_t_4, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __PYX_ERR(0, 866, __pyx_L1_error)

      /* "validr/_validator_c.pyx":865
 *             raise Invalid('invalid timedelta') from ex
 *         if min_value is not None:
 *             if value < min_value:             # <<<<<<<<<<<<<<
 *                 raise Invalid('value must >= {}'.format(min_repr))
 *         if max_value is not None:
 */
    }

    /* "validr/_validator_c.pyx":864
 *         except (durationpy.DurationError, ValueError, TypeError) as ex:
 *             raise Invalid('invalid timedelta') from ex
 *         if min_value is not None:             # <<<<<<<<<<<<<<
 *             if value < min_value:
 *                 raise Invalid('value must >= {}'.format(min_repr))
 */
  }

  /* "validr/_validator_c.pyx":867
 *             if value < min_value:
 *                 raise Invalid('value must >= {}'.format(min_repr))
 *         if max_value is not None:             # <<<<<<<<<<<<<<
 *             if value > max_value:
 *                 raise Invalid('value must <= {}'.format(max_repr))
 */
  __Pyx_TraceLine(867,0,__PYX_ERR(0, 867, __pyx_L1_error))
  if (unlikely(!__pyx_cur_scope->__pyx_v_max_value)) { __Pyx_RaiseClosureNameError("max_value"); __PYX_ERR(0, 867, __pyx_L1_error) }
  __pyx_t_20 = (__pyx_cur_scope->__pyx_v_max_value != Py_None);
  __pyx_t_19 = (__pyx_t_20 != 0);
  if (__pyx_t_19) {

    /* "validr/_validator_c.pyx":868
 *                 raise Invalid('value must >= {}'.format(min_repr))
 *         if max_value is not None:
 *             if value > max_value:             # <<<<<<<<<<<<<<
 *                 raise Invalid('value must <= {}'.format(max_repr))
 *         if output_object:
 */
    __Pyx_TraceLine(868,0,__PYX_ERR(0, 868, __pyx_L1_error))
    if (unlikely(!__pyx_cur_scope->__pyx_v_max_value)) { __Pyx_RaiseClosureNameError("max_value"); __PYX_ERR(0, 868, __pyx_L1_error) }
    __pyx_t_4 = PyObject_RichCompare(__pyx_v_value, __pyx_cur_scope->__pyx_v_max_value, Py_GT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 868, __pyx_L1_error)
    __pyx_t_19 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_19 < 0)) __PYX_ERR(0, 868, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(__pyx_t_19)) {

      /* "validr/_validator_c.pyx":869
 *         if max_value is not None:
 *             if value > max_value:
 *                 raise Invalid('value must <= {}'.format(max_repr))             # <<<<<<<<<<<<<<
 *         if output_object:
 *             return value
 */
      __Pyx_TraceLine(869,0,__PYX_ERR(0, 869, __pyx_L1_error))
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 869, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_value_must_2, __pyx_n_s_format); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 869, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (unlikely(!__pyx_cur_scope->__pyx_v_max_repr)) { __Pyx_RaiseClosureNameError("max_repr"); __PYX_ERR(0, 869, __pyx_L1_error) }
      __pyx_t_7 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
        }
      }
      __pyx_t_6 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_7, __pyx_cur_scope->__pyx_v_max_repr) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_cur_scope->__pyx_v_max_repr);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 869, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_4 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 869, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_Raise(__pyx_t_4, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __PYX_ERR(0, 869, __pyx_L1_error)

      /* "validr/_validator_c.pyx":868
 *                 raise Invalid('value must >= {}'.format(min_repr))
 *         if max_value is not None:
 *             if value > max_value:             # <<<<<<<<<<<<<<
 *                 raise Invalid('value must <= {}'.format(max_repr))
 *         if output_object:
 */
    }

    /* "validr/_validator_c.pyx":867
 *             if value < min_value:
 *                 raise Invalid('value must >= {}'.format(min_repr))
 *         if max_value is not None:             # <<<<<<<<<<<<<<
 *             if value > max_value:
 *                 raise Invalid('value must <= {}'.format(max_repr))
 */
  }

  /* "validr/_validator_c.pyx":870
 *             if value > max_value:
 *                 raise Invalid('value must <= {}'.format(max_repr))
 *         if output_object:             # <<<<<<<<<<<<<<
 *             return value
 *         if string:
 */
  __Pyx_TraceLine(870,0,__PYX_ERR(0, 870, __pyx_L1_error))
  __pyx_t_19 = (__pyx_cur_scope->__pyx_v_output_object != 0);
  if (__pyx_t_19) {

    /* "validr/_validator_c.pyx":871
 *                 raise Invalid('value must <= {}'.format(max_repr))
 *         if output_object:
 *             return value             # <<<<<<<<<<<<<<
 *         if string:
 *             value = durationpy.to_str(value, extended=extended)
 */
    __Pyx_TraceLine(871,0,__PYX_ERR(0, 871, __pyx_L1_error))
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_value);
    __pyx_r = __pyx_v_value;
    goto __pyx_L0;

    /* "validr/_validator_c.pyx":870
 *             if value > max_value:
 *                 raise Invalid('value must <= {}'.format(max_repr))
 *         if output_object:             # <<<<<<<<<<<<<<
 *             return value
 *         if string:
 */
  }

  /* "validr/_validator_c.pyx":872
 *         if output_object:
 *             return value
 *         if string:             # <<<<<<<<<<<<<<
 *             value = durationpy.to_str(value, extended=extended)
 *         else:
 */
  __Pyx_TraceLine(872,0,__PYX_ERR(0, 872, __pyx_L1_error))
  __pyx_t_19 = (__pyx_cur_scope->__pyx_v_string != 0);
  if (__pyx_t_19) {

    /* "validr/_validator_c.pyx":873
 *             return value
 *         if string:
 *             value = durationpy.to_str(value, extended=extended)             # <<<<<<<<<<<<<<
 *         else:
 *             value = value.total_seconds()
 */
    __Pyx_TraceLine(873,0,__PYX_ERR(0, 873, __pyx_L1_error))
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_durationpy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 873, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_to_str); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 873, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 873, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_value);
    __Pyx_GIVEREF(__pyx_v_value);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_value);
    __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 873, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = __Pyx_PyBool_FromLong(__pyx_cur_scope->__pyx_v_extended); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 873, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_extended, __pyx_t_8) < 0) __PYX_ERR(0, 873, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 873, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_8);
    __pyx_t_8 = 0;

    /* "validr/_validator_c.pyx":872
 *         if output_object:
 *             return value
 *         if string:             # <<<<<<<<<<<<<<
 *             value = durationpy.to_str(value, extended=extended)
 *         else:
 */
    goto __pyx_L25;
  }

  /* "validr/_validator_c.pyx":875
 *             value = durationpy.to_str(value, extended=extended)
 *         else:
 *             value = value.total_seconds()             # <<<<<<<<<<<<<<
 *         return value
 *     return validate
 */
  __Pyx_TraceLine(875,0,__PYX_ERR(0, 875, __pyx_L1_error))
  /*else*/ {
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_total_seconds); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 875, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_8 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 875, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_8);
    __pyx_t_8 = 0;
  }
  __pyx_L25:;

  /* "validr/_validator_c.pyx":876
 *         else:
 *             value = value.total_seconds()
 *         return value             # <<<<<<<<<<<<<<
 *     return validate
 * 
 */
  __Pyx_TraceLine(876,0,__PYX_ERR(0, 876, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_value);
  __pyx_r = __pyx_v_value;
  goto __pyx_L0;

  /* "validr/_validator_c.pyx":859
 *         max_repr = durationpy.to_str(max_value, extended=True)
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         try:
 *             value = _parse_timedelta(value)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("validr._validator_c.timedelta_validator.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ex);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":827
 * 
 * @validator(accept=(int, float, str, datetime.timedelta), output=(str, float, datetime.timedelta))
 * def timedelta_validator(compiler, min=None, max=None, bint string=False,             # <<<<<<<<<<<<<<
 *                         bint extended=False, bint output_object=False):
 *     """Validate timedelta string or convert timedelta to string
 */

static PyObject *__pyx_pf_6validr_12_validator_c_62timedelta_validator(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_compiler, PyObject *__pyx_v_min, PyObject *__pyx_v_max, int __pyx_v_string, int __pyx_v_extended, int __pyx_v_output_object) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_21_timedelta_validator *__pyx_cur_scope;
  PyObject *__pyx_v_ex = NULL;
  PyObject *__pyx_v_validate = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  char const *__pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  char const *__pyx_t_21;
  __Pyx_TraceFrameInit(__pyx_codeobj__91)
  __Pyx_RefNannySetupContext("timedelta_validator", 0);
  __pyx_cur_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_21_timedelta_validator *)__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_21_timedelta_validator(__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_21_timedelta_validator, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_21_timedelta_validator *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 827, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __Pyx_TraceCall("timedelta_validator", __pyx_f[0], 827, 0, __PYX_ERR(0, 827, __pyx_L1_error));
  __pyx_cur_scope->__pyx_v_string = __pyx_v_string;
  __pyx_cur_scope->__pyx_v_extended = __pyx_v_extended;
  __pyx_cur_scope->__pyx_v_output_object = __pyx_v_output_object;
  __Pyx_INCREF(__pyx_v_min);
  __Pyx_INCREF(__pyx_v_max);

  /* "validr/_validator_c.pyx":842
 *         y - years
 *     """
 *     if string and output_object:             # <<<<<<<<<<<<<<
 *         raise SchemaError('can not output both string and object')
 *     try:
 */
  __Pyx_TraceLine(842,0,__PYX_ERR(0, 842, __pyx_L1_error))
  __pyx_t_2 = (__pyx_cur_scope->__pyx_v_string != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_cur_scope->__pyx_v_output_object != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "validr/_validator_c.pyx":843
 *     """
 *     if string and output_object:
 *         raise SchemaError('can not output both string and object')             # <<<<<<<<<<<<<<
 *     try:
 *         min_value = _parse_timedelta(min) if min is not None else None
 */
    __Pyx_TraceLine(843,0,__PYX_ERR(0, 843, __pyx_L1_error))
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_SchemaError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 843, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_kp_u_can_not_output_both_string_and_o) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_kp_u_can_not_output_both_string_and_o);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 843, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 843, __pyx_L1_error)

    /* "validr/_validator_c.pyx":842
 *         y - years
 *     """
 *     if string and output_object:             # <<<<<<<<<<<<<<
 *         raise SchemaError('can not output both string and object')
 *     try:
 */
  }

  /* "validr/_validator_c.pyx":844
 *     if string and output_object:
 *         raise SchemaError('can not output both string and object')
 *     try:             # <<<<<<<<<<<<<<
 *         min_value = _parse_timedelta(min) if min is not None else None
 *     except (durationpy.DurationError, ValueError, TypeError) as ex:
 */
  __Pyx_TraceLine(844,0,__PYX_ERR(0, 844, __pyx_L1_error))
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
    __Pyx_XGOTREF(__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_7);
    __Pyx_XGOTREF(__pyx_t_8);
    /*try:*/ {

      /* "validr/_validator_c.pyx":845
 *         raise SchemaError('can not output both string and object')
 *     try:
 *         min_value = _parse_timedelta(min) if min is not None else None             # <<<<<<<<<<<<<<
 *     except (durationpy.DurationError, ValueError, TypeError) as ex:
 *         raise SchemaError('invalid min timedelta') from ex
 */
      __Pyx_TraceLine(845,0,__PYX_ERR(0, 845, __pyx_L6_error))
      __pyx_t_1 = (__pyx_v_min != Py_None);
      if ((__pyx_t_1 != 0)) {
        __pyx_t_4 = __pyx_f_6validr_12_validator_c__parse_timedelta(__pyx_v_min, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 845, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_3 = __pyx_t_4;
        __pyx_t_4 = 0;
      } else {
        __Pyx_INCREF(Py_None);
        __pyx_t_3 = Py_None;
      }
      __Pyx_GIVEREF(__pyx_t_3);
      __pyx_cur_scope->__pyx_v_min_value = __pyx_t_3;
      __pyx_t_3 = 0;

      /* "validr/_validator_c.pyx":844
 *     if string and output_object:
 *         raise SchemaError('can not output both string and object')
 *     try:             # <<<<<<<<<<<<<<
 *         min_value = _parse_timedelta(min) if min is not None else None
 *     except (durationpy.DurationError, ValueError, TypeError) as ex:
 */
    }
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    goto __pyx_L11_try_end;
    __pyx_L6_error:;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "validr/_validator_c.pyx":846
 *     try:
 *         min_value = _parse_timedelta(min) if min is not None else None
 *     except (durationpy.DurationError, ValueError, TypeError) as ex:             # <<<<<<<<<<<<<<
 *         raise SchemaError('invalid min timedelta') from ex
 *     try:
 */
    __Pyx_TraceLine(846,0,__PYX_ERR(0, 846, __pyx_L8_except_error))
    __Pyx_ErrFetch(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_durationpy); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 846, __pyx_L8_except_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_DurationError); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 846, __pyx_L8_except_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_11 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_3, __pyx_t_10) || __Pyx_PyErr_GivenExceptionMatches(__pyx_t_3, __pyx_builtin_ValueError) || __Pyx_PyErr_GivenExceptionMatches(__pyx_t_3, __pyx_builtin_TypeError);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_ErrRestore(__pyx_t_3, __pyx_t_4, __pyx_t_5);
    __pyx_t_3 = 0; __pyx_t_4 = 0; __pyx_t_5 = 0;
    if (__pyx_t_11) {
      __Pyx_AddTraceback("validr._validator_c.timedelta_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_4, &__pyx_t_3) < 0) __PYX_ERR(0, 846, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __pyx_v_ex = __pyx_t_4;
      /*try:*/ {

        /* "validr/_validator_c.pyx":847
 *         min_value = _parse_timedelta(min) if min is not None else None
 *     except (durationpy.DurationError, ValueError, TypeError) as ex:
 *         raise SchemaError('invalid min timedelta') from ex             # <<<<<<<<<<<<<<
 *     try:
 *         max_value = _parse_timedelta(max) if max is not None else None
 */
        __Pyx_TraceLine(847,0,__PYX_ERR(0, 847, __pyx_L17_error))
        __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_SchemaError); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 847, __pyx_L17_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_12 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
          __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_9);
          if (likely(__pyx_t_12)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
            __Pyx_INCREF(__pyx_t_12);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_9, function);
          }
        }
        __pyx_t_10 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_12, __pyx_kp_u_invalid_min_timedelta) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_kp_u_invalid_min_timedelta);
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 847, __pyx_L17_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_Raise(__pyx_t_10, 0, 0, __pyx_v_ex);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __PYX_ERR(0, 847, __pyx_L17_error)
      }

      /* "validr/_validator_c.pyx":846
 *     try:
 *         min_value = _parse_timedelta(min) if min is not None else None
 *     except (durationpy.DurationError, ValueError, TypeError) as ex:             # <<<<<<<<<<<<<<
 *         raise SchemaError('invalid min timedelta') from ex
 *     try:
 */
      __Pyx_TraceLine(846,0,__PYX_ERR(0, 846, __pyx_L17_error))
      /*finally:*/ {
        __pyx_L17_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17) < 0)) __Pyx_ErrFetch(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_18);
          __Pyx_XGOTREF(__pyx_t_19);
          __Pyx_XGOTREF(__pyx_t_20);
          __pyx_t_11 = __pyx_lineno; __pyx_t_13 = __pyx_clineno; __pyx_t_14 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_ex);
            __pyx_v_ex = NULL;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_XGIVEREF(__pyx_t_19);
            __Pyx_XGIVEREF(__pyx_t_20);
            __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
          }
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_XGIVEREF(__pyx_t_16);
          __Pyx_XGIVEREF(__pyx_t_17);
          __Pyx_ErrRestore(__pyx_t_15, __pyx_t_16, __pyx_t_17);
          __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
          __pyx_lineno = __pyx_t_11; __pyx_clineno = __pyx_t_13; __pyx_filename = __pyx_t_14;
          goto __pyx_L8_except_error;
        }
      }
    }
    goto __pyx_L8_except_error;
    __pyx_L8_except_error:;

    /* "validr/_validator_c.pyx":844
 *     if string and output_object:
 *         raise SchemaError('can not output both string and object')
 *     try:             # <<<<<<<<<<<<<<
 *         min_value = _parse_timedelta(min) if min is not None else None
 *     except (durationpy.DurationError, ValueError, TypeError) as ex:
 */
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
    goto __pyx_L1_error;
    __pyx_L11_try_end:;
  }

  /* "validr/_validator_c.pyx":848
 *     except (durationpy.DurationError, ValueError, TypeError) as ex:
 *         raise SchemaError('invalid min timedelta') from ex
 *     try:             # <<<<<<<<<<<<<<
 *         max_value = _parse_timedelta(max) if max is not None else None
 *     except (durationpy.DurationError, ValueError, TypeError) as ex:
 */
  __Pyx_TraceLine(848,0,__PYX_ERR(0, 848, __pyx_L1_error))
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_7, &__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_8);
    __Pyx_XGOTREF(__pyx_t_7);
    __Pyx_XGOTREF(__pyx_t_6);
    /*try:*/ {

      /* "validr/_validator_c.pyx":849
 *         raise SchemaError('invalid min timedelta') from ex
 *     try:
 *         max_value = _parse_timedelta(max) if max is not None else None             # <<<<<<<<<<<<<<
 *     except (durationpy.DurationError, ValueError, TypeError) as ex:
 *         raise SchemaError('invalid max timedelta') from ex
 */
      __Pyx_TraceLine(849,0,__PYX_ERR(0, 849, __pyx_L23_error))
      __pyx_t_1 = (__pyx_v_max != Py_None);
      if ((__pyx_t_1 != 0)) {
        __pyx_t_4 = __pyx_f_6validr_12_validator_c__parse_timedelta(__pyx_v_max, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 849, __pyx_L23_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_3 = __pyx_t_4;
        __pyx_t_4 = 0;
      } else {
        __Pyx_INCREF(Py_None);
        __pyx_t_3 = Py_None;
      }
      __Pyx_GIVEREF(__pyx_t_3);
      __pyx_cur_scope->__pyx_v_max_value = __pyx_t_3;
      __pyx_t_3 = 0;

      /* "validr/_validator_c.pyx":848
 *     except (durationpy.DurationError, ValueError, TypeError) as ex:
 *         raise SchemaError('invalid min timedelta') from ex
 *     try:             # <<<<<<<<<<<<<<
 *         max_value = _parse_timedelta(max) if max is not None else None
 *     except (durationpy.DurationError, ValueError, TypeError) as ex:
 */
    }
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    goto __pyx_L28_try_end;
    __pyx_L23_error:;
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "validr/_validator_c.pyx":850
 *     try:
 *         max_value = _parse_timedelta(max) if max is not None else None
 *     except (durationpy.DurationError, ValueError, TypeError) as ex:             # <<<<<<<<<<<<<<
 *         raise SchemaError('invalid max timedelta') from ex
 *     del min, max
 */
    __Pyx_TraceLine(850,0,__PYX_ERR(0, 850, __pyx_L25_except_error))
    __Pyx_ErrFetch(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
    __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_durationpy); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 850, __pyx_L25_except_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_DurationError); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 850, __pyx_L25_except_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_13 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_3, __pyx_t_9) || __Pyx_PyErr_GivenExceptionMatches(__pyx_t_3, __pyx_builtin_ValueError) || __Pyx_PyErr_GivenExceptionMatches(__pyx_t_3, __pyx_builtin_TypeError);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_ErrRestore(__pyx_t_3, __pyx_t_4, __pyx_t_5);
    __pyx_t_3 = 0; __pyx_t_4 = 0; __pyx_t_5 = 0;
    if (__pyx_t_13) {
      __Pyx_AddTraceback("validr._validator_c.timedelta_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_4, &__pyx_t_3) < 0) __PYX_ERR(0, 850, __pyx_L25_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __pyx_v_ex = __pyx_t_4;
      /*try:*/ {

        /* "validr/_validator_c.pyx":851
 *         max_value = _parse_timedelta(max) if max is not None else None
 *     except (durationpy.DurationError, ValueError, TypeError) as ex:
 *         raise SchemaError('invalid max timedelta') from ex             # <<<<<<<<<<<<<<
 *     del min, max
 *     min_repr = max_repr = None
 */
        __Pyx_TraceLine(851,0,__PYX_ERR(0, 851, __pyx_L34_error))
        __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_SchemaError); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 851, __pyx_L34_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_12 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
          __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_10);
          if (likely(__pyx_t_12)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
            __Pyx_INCREF(__pyx_t_12);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_10, function);
          }
        }
        __pyx_t_9 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_12, __pyx_kp_u_invalid_max_timedelta) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_kp_u_invalid_max_timedelta);
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 851, __pyx_L34_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_Raise(__pyx_t_9, 0, 0, __pyx_v_ex);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __PYX_ERR(0, 851, __pyx_L34_error)
      }

      /* "validr/_validator_c.pyx":850
 *     try:
 *         max_value = _parse_timedelta(max) if max is not None else None
 *     except (durationpy.DurationError, ValueError, TypeError) as ex:             # <<<<<<<<<<<<<<
 *         raise SchemaError('invalid max timedelta') from ex
 *     del min, max
 */
      __Pyx_TraceLine(850,0,__PYX_ERR(0, 850, __pyx_L34_error))
      /*finally:*/ {
        __pyx_L34_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0; __pyx_t_16 = 0; __pyx_t_15 = 0;
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_17, &__pyx_t_16, &__pyx_t_15);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_20, &__pyx_t_19, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_20, &__pyx_t_19, &__pyx_t_18);
          __Pyx_XGOTREF(__pyx_t_20);
          __Pyx_XGOTREF(__pyx_t_19);
          __Pyx_XGOTREF(__pyx_t_18);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_15);
          __pyx_t_13 = __pyx_lineno; __pyx_t_11 = __pyx_clineno; __pyx_t_21 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_ex);
            __pyx_v_ex = NULL;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_16, __pyx_t_15);
          }
          __Pyx_XGIVEREF(__pyx_t_20);
          __Pyx_XGIVEREF(__pyx_t_19);
          __Pyx_XGIVEREF(__pyx_t_18);
          __Pyx_ErrRestore(__pyx_t_20, __pyx_t_19, __pyx_t_18);
          __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0; __pyx_t_16 = 0; __pyx_t_15 = 0;
          __pyx_lineno = __pyx_t_13; __pyx_clineno = __pyx_t_11; __pyx_filename = __pyx_t_21;
          goto __pyx_L25_except_error;
        }
      }
    }
    goto __pyx_L25_except_error;
    __pyx_L25_except_error:;

    /* "validr/_validator_c.pyx":848
 *     except (durationpy.DurationError, ValueError, TypeError) as ex:
 *         raise SchemaError('invalid min timedelta') from ex
 *     try:             # <<<<<<<<<<<<<<
 *         max_value = _parse_timedelta(max) if max is not None else None
 *     except (durationpy.DurationError, ValueError, TypeError) as ex:
 */
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_7, __pyx_t_6);
    goto __pyx_L1_error;
    __pyx_L28_try_end:;
  }

  /* "validr/_validator_c.pyx":852
 *     except (durationpy.DurationError, ValueError, TypeError) as ex:
 *         raise SchemaError('invalid max timedelta') from ex
 *     del min, max             # <<<<<<<<<<<<<<
 *     min_repr = max_repr = None
 *     if min_value is not None:
 */
  __Pyx_TraceLine(852,0,__PYX_ERR(0, 852, __pyx_L1_error))
  __Pyx_DECREF(__pyx_v_min);
  __pyx_v_min = NULL;
  __Pyx_DECREF(__pyx_v_max);
  __pyx_v_max = NULL;

  /* "validr/_validator_c.pyx":853
 *         raise SchemaError('invalid max timedelta') from ex
 *     del min, max
 *     min_repr = max_repr = None             # <<<<<<<<<<<<<<
 *     if min_value is not None:
 *         min_repr = durationpy.to_str(min_value, extended=True)
 */
  __Pyx_TraceLine(853,0,__PYX_ERR(0, 853, __pyx_L1_error))
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __pyx_cur_scope->__pyx_v_min_repr = Py_None;
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __pyx_cur_scope->__pyx_v_max_repr = Py_None;

  /* "validr/_validator_c.pyx":854
 *     del min, max
 *     min_repr = max_repr = None
 *     if min_value is not None:             # <<<<<<<<<<<<<<
 *         min_repr = durationpy.to_str(min_value, extended=True)
 *     if max_value is not None:
 */
  __Pyx_TraceLine(854,0,__PYX_ERR(0, 854, __pyx_L1_error))
  __pyx_t_1 = (__pyx_cur_scope->__pyx_v_min_value != Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "validr/_validator_c.pyx":855
 *     min_repr = max_repr = None
 *     if min_value is not None:
 *         min_repr = durationpy.to_str(min_value, extended=True)             # <<<<<<<<<<<<<<
 *     if max_value is not None:
 *         max_repr = durationpy.to_str(max_value, extended=True)
 */
    __Pyx_TraceLine(855,0,__PYX_ERR(0, 855, __pyx_L1_error))
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_durationpy); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 855, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_to_str); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 855, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 855, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_min_value);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_min_value);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_cur_scope->__pyx_v_min_value);
    __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 855, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_extended, Py_True) < 0) __PYX_ERR(0, 855, __pyx_L1_error)
    __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, __pyx_t_5); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 855, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_min_repr);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_min_repr, __pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_9);
    __pyx_t_9 = 0;

    /* "validr/_validator_c.pyx":854
 *     del min, max
 *     min_repr = max_repr = None
 *     if min_value is not None:             # <<<<<<<<<<<<<<
 *         min_repr = durationpy.to_str(min_value, extended=True)
 *     if max_value is not None:
 */
  }

  /* "validr/_validator_c.pyx":856
 *     if min_value is not None:
 *         min_repr = durationpy.to_str(min_value, extended=True)
 *     if max_value is not None:             # <<<<<<<<<<<<<<
 *         max_repr = durationpy.to_str(max_value, extended=True)
 * 
 */
  __Pyx_TraceLine(856,0,__PYX_ERR(0, 856, __pyx_L1_error))
  __pyx_t_2 = (__pyx_cur_scope->__pyx_v_max_value != Py_None);
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "validr/_validator_c.pyx":857
 *         min_repr = durationpy.to_str(min_value, extended=True)
 *     if max_value is not None:
 *         max_repr = durationpy.to_str(max_value, extended=True)             # <<<<<<<<<<<<<<
 * 
 *     def validate(value):
 */
    __Pyx_TraceLine(857,0,__PYX_ERR(0, 857, __pyx_L1_error))
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_durationpy); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 857, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_to_str); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 857, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = PyTuple_New(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 857, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_max_value);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_max_value);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_cur_scope->__pyx_v_max_value);
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 857, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_extended, Py_True) < 0) __PYX_ERR(0, 857, __pyx_L1_error)
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_9, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 857, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_max_repr);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_max_repr, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;

    /* "validr/_validator_c.pyx":856
 *     if min_value is not None:
 *         min_repr = durationpy.to_str(min_value, extended=True)
 *     if max_value is not None:             # <<<<<<<<<<<<<<
 *         max_repr = durationpy.to_str(max_value, extended=True)
 * 
 */
  }

  /* "validr/_validator_c.pyx":859
 *         max_repr = durationpy.to_str(max_value, extended=True)
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         try:
 *             value = _parse_timedelta(value)
 */
  __Pyx_TraceLine(859,0,__PYX_ERR(0, 859, __pyx_L1_error))
  __pyx_t_4 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_19timedelta_validator_1validate, 0, __pyx_n_s_timedelta_validator_locals_valid, ((PyObject*)__pyx_cur_scope), __pyx_n_s_validr__validator_c, __pyx_d, ((PyObject *)__pyx_codeobj__93)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_validate = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "validr/_validator_c.pyx":877
 *             value = value.total_seconds()
 *         return value
 *     return validate             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(877,0,__PYX_ERR(0, 877, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_validate);
  __pyx_r = __pyx_v_validate;
  goto __pyx_L0;

  /* "validr/_validator_c.pyx":827
 * 
 * @validator(accept=(int, float, str, datetime.timedelta), output=(str, float, datetime.timedelta))
 * def timedelta_validator(compiler, min=None, max=None, bint string=False,             # <<<<<<<<<<<<<<
 *                         bint extended=False, bint output_object=False):
 *     """Validate timedelta string or convert timedelta to string
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("validr._validator_c.timedelta_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ex);
  __Pyx_XDECREF(__pyx_v_validate);
  __Pyx_XDECREF(__pyx_v_min);
  __Pyx_XDECREF(__pyx_v_max);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":881
 * 
 * @validator(accept=(str, ipaddress.IPv4Address), output=(str, ipaddress.IPv4Address))
 * def ipv4_validator(compiler, bint output_object=False):             # <<<<<<<<<<<<<<
 *     def validate(value):
 *         try:
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_65ipv4_validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6validr_12_validator_c_65ipv4_validator = {"ipv4_validator", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6validr_12_validator_c_65ipv4_validator, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6validr_12_validator_c_65ipv4_validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_compiler = 0;
  int __pyx_v_output_object;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ipv4_validator (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_compiler,&__pyx_n_s_output_object,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_compiler)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_output_object);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "ipv4_validator") < 0)) __PYX_ERR(0, 881, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_compiler = values[0];
    if (values[1]) {
      __pyx_v_output_object = __Pyx_PyObject_IsTrue(values[1]); if (unlikely((__pyx_v_output_object == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 881, __pyx_L3_error)
    } else {
      __pyx_v_output_object = ((int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ipv4_validator", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 881, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("validr._validator_c.ipv4_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6validr_12_validator_c_64ipv4_validator(__pyx_self, __pyx_v_compiler, __pyx_v_output_object);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":882
 * @validator(accept=(str, ipaddress.IPv4Address), output=(str, ipaddress.IPv4Address))
 * def ipv4_validator(compiler, bint output_object=False):
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         try:
 *             value = ipaddress.IPv4Address(value.strip())
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_14ipv4_validator_1validate(PyObject *__pyx_self, PyObject *__pyx_v_value); /*proto*/
static PyMethodDef __pyx_mdef_6validr_12_validator_c_14ipv4_validator_1validate = {"validate", (PyCFunction)__pyx_pw_6validr_12_validator_c_14ipv4_validator_1validate, METH_O, 0};
static PyObject *__pyx_pw_6validr_12_validator_c_14ipv4_validator_1validate(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("validate (wrapper)", 0);
  __pyx_r = __pyx_pf_6validr_12_validator_c_14ipv4_validator_validate(__pyx_self, ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6validr_12_validator_c_14ipv4_validator_validate(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_22_ipv4_validator *__pyx_cur_scope;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_22_ipv4_validator *__pyx_outer_scope;
  PyObject *__pyx_v_ex = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  char const *__pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  int __pyx_t_20;
  __Pyx_RefNannySetupContext("validate", 0);
  __pyx_outer_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_22_ipv4_validator *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_TraceCall("validate", __pyx_f[0], 882, 0, __PYX_ERR(0, 882, __pyx_L1_error));
  __Pyx_INCREF(__pyx_v_value);

  /* "validr/_validator_c.pyx":883
 * def ipv4_validator(compiler, bint output_object=False):
 *     def validate(value):
 *         try:             # <<<<<<<<<<<<<<
 *             value = ipaddress.IPv4Address(value.strip())
 *         except ipaddress.AddressValueError as ex:
 */
  __Pyx_TraceLine(883,0,__PYX_ERR(0, 883, __pyx_L1_error))
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "validr/_validator_c.pyx":884
 *     def validate(value):
 *         try:
 *             value = ipaddress.IPv4Address(value.strip())             # <<<<<<<<<<<<<<
 *         except ipaddress.AddressValueError as ex:
 *             raise Invalid(str(ex)) from None
 */
      __Pyx_TraceLine(884,0,__PYX_ERR(0, 884, __pyx_L3_error))
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_ipaddress); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 884, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_IPv4Address); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 884, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_strip); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 884, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      __pyx_t_5 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 884, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
        }
      }
      __pyx_t_4 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_7, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_5);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 884, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "validr/_validator_c.pyx":883
 * def ipv4_validator(compiler, bint output_object=False):
 *     def validate(value):
 *         try:             # <<<<<<<<<<<<<<
 *             value = ipaddress.IPv4Address(value.strip())
 *         except ipaddress.AddressValueError as ex:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "validr/_validator_c.pyx":885
 *         try:
 *             value = ipaddress.IPv4Address(value.strip())
 *         except ipaddress.AddressValueError as ex:             # <<<<<<<<<<<<<<
 *             raise Invalid(str(ex)) from None
 *         except Exception:
 */
    __Pyx_TraceLine(885,0,__PYX_ERR(0, 885, __pyx_L5_except_error))
    __Pyx_ErrFetch(&__pyx_t_4, &__pyx_t_6, &__pyx_t_5);
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_ipaddress); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 885, __pyx_L5_except_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_AddressValueError); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 885, __pyx_L5_except_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_9 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_4, __pyx_t_8);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_ErrRestore(__pyx_t_4, __pyx_t_6, __pyx_t_5);
    __pyx_t_4 = 0; __pyx_t_6 = 0; __pyx_t_5 = 0;
    if (__pyx_t_9) {
      __Pyx_AddTraceback("validr._validator_c.ipv4_validator.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_4) < 0) __PYX_ERR(0, 885, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_6);
      __pyx_v_ex = __pyx_t_6;
      /*try:*/ {

        /* "validr/_validator_c.pyx":886
 *             value = ipaddress.IPv4Address(value.strip())
 *         except ipaddress.AddressValueError as ex:
 *             raise Invalid(str(ex)) from None             # <<<<<<<<<<<<<<
 *         except Exception:
 *             raise Invalid('invalid ipv4 address') from None
 */
        __Pyx_TraceLine(886,0,__PYX_ERR(0, 886, __pyx_L14_error))
        __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 886, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_10 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyUnicode_Type)), __pyx_v_ex); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 886, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_11 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_11)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_11);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        __pyx_t_8 = (__pyx_t_11) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_11, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_10);
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 886, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_Raise(__pyx_t_8, 0, 0, Py_None);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __PYX_ERR(0, 886, __pyx_L14_error)
      }

      /* "validr/_validator_c.pyx":885
 *         try:
 *             value = ipaddress.IPv4Address(value.strip())
 *         except ipaddress.AddressValueError as ex:             # <<<<<<<<<<<<<<
 *             raise Invalid(str(ex)) from None
 *         except Exception:
 */
      __Pyx_TraceLine(885,0,__PYX_ERR(0, 885, __pyx_L14_error))
      /*finally:*/ {
        __pyx_L14_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16) < 0)) __Pyx_ErrFetch(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_18);
          __Pyx_XGOTREF(__pyx_t_19);
          __pyx_t_9 = __pyx_lineno; __pyx_t_12 = __pyx_clineno; __pyx_t_13 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_ex);
            __pyx_v_ex = NULL;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_XGIVEREF(__pyx_t_19);
            __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
          }
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_XGIVEREF(__pyx_t_16);
          __Pyx_ErrRestore(__pyx_t_14, __pyx_t_15, __pyx_t_16);
          __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
          __pyx_lineno = __pyx_t_9; __pyx_clineno = __pyx_t_12; __pyx_filename = __pyx_t_13;
          goto __pyx_L5_except_error;
        }
      }
    }

    /* "validr/_validator_c.pyx":887
 *         except ipaddress.AddressValueError as ex:
 *             raise Invalid(str(ex)) from None
 *         except Exception:             # <<<<<<<<<<<<<<
 *             raise Invalid('invalid ipv4 address') from None
 *         if output_object:
 */
    __Pyx_TraceLine(887,0,__PYX_ERR(0, 887, __pyx_L5_except_error))
    __pyx_t_12 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_12) {
      __Pyx_AddTraceback("validr._validator_c.ipv4_validator.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_6, &__pyx_t_5) < 0) __PYX_ERR(0, 887, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_5);

      /* "validr/_validator_c.pyx":888
 *             raise Invalid(str(ex)) from None
 *         except Exception:
 *             raise Invalid('invalid ipv4 address') from None             # <<<<<<<<<<<<<<
 *         if output_object:
 *             return value
 */
      __Pyx_TraceLine(888,0,__PYX_ERR(0, 888, __pyx_L5_except_error))
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 888, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_10 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      __pyx_t_8 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_10, __pyx_kp_u_invalid_ipv4_address) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_kp_u_invalid_ipv4_address);
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 888, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_Raise(__pyx_t_8, 0, 0, Py_None);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(0, 888, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "validr/_validator_c.pyx":883
 * def ipv4_validator(compiler, bint output_object=False):
 *     def validate(value):
 *         try:             # <<<<<<<<<<<<<<
 *             value = ipaddress.IPv4Address(value.strip())
 *         except ipaddress.AddressValueError as ex:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "validr/_validator_c.pyx":889
 *         except Exception:
 *             raise Invalid('invalid ipv4 address') from None
 *         if output_object:             # <<<<<<<<<<<<<<
 *             return value
 *         else:
 */
  __Pyx_TraceLine(889,0,__PYX_ERR(0, 889, __pyx_L1_error))
  __pyx_t_20 = (__pyx_cur_scope->__pyx_v_output_object != 0);
  if (__pyx_t_20) {

    /* "validr/_validator_c.pyx":890
 *             raise Invalid('invalid ipv4 address') from None
 *         if output_object:
 *             return value             # <<<<<<<<<<<<<<
 *         else:
 *             return value.compressed
 */
    __Pyx_TraceLine(890,0,__PYX_ERR(0, 890, __pyx_L1_error))
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_value);
    __pyx_r = __pyx_v_value;
    goto __pyx_L0;

    /* "validr/_validator_c.pyx":889
 *         except Exception:
 *             raise Invalid('invalid ipv4 address') from None
 *         if output_object:             # <<<<<<<<<<<<<<
 *             return value
 *         else:
 */
  }

  /* "validr/_validator_c.pyx":892
 *             return value
 *         else:
 *             return value.compressed             # <<<<<<<<<<<<<<
 *     return validate
 * 
 */
  __Pyx_TraceLine(892,0,__PYX_ERR(0, 892, __pyx_L1_error))
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_compressed); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 892, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;
  }

  /* "validr/_validator_c.pyx":882
 * @validator(accept=(str, ipaddress.IPv4Address), output=(str, ipaddress.IPv4Address))
 * def ipv4_validator(compiler, bint output_object=False):
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         try:
 *             value = ipaddress.IPv4Address(value.strip())
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("validr._validator_c.ipv4_validator.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ex);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":881
 * 
 * @validator(accept=(str, ipaddress.IPv4Address), output=(str, ipaddress.IPv4Address))
 * def ipv4_validator(compiler, bint output_object=False):             # <<<<<<<<<<<<<<
 *     def validate(value):
 *         try:
 */

static PyObject *__pyx_pf_6validr_12_validator_c_64ipv4_validator(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_compiler, int __pyx_v_output_object) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_22_ipv4_validator *__pyx_cur_scope;
  PyObject *__pyx_v_validate = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_TraceFrameInit(__pyx_codeobj__94)
  __Pyx_RefNannySetupContext("ipv4_validator", 0);
  __pyx_cur_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_22_ipv4_validator *)__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_22_ipv4_validator(__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_22_ipv4_validator, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_22_ipv4_validator *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 881, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __Pyx_TraceCall("ipv4_validator", __pyx_f[0], 881, 0, __PYX_ERR(0, 881, __pyx_L1_error));
  __pyx_cur_scope->__pyx_v_output_object = __pyx_v_output_object;

  /* "validr/_validator_c.pyx":882
 * @validator(accept=(str, ipaddress.IPv4Address), output=(str, ipaddress.IPv4Address))
 * def ipv4_validator(compiler, bint output_object=False):
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         try:
 *             value = ipaddress.IPv4Address(value.strip())
 */
  __Pyx_TraceLine(882,0,__PYX_ERR(0, 882, __pyx_L1_error))
  __pyx_t_1 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_14ipv4_validator_1validate, 0, __pyx_n_s_ipv4_validator_locals_validate, ((PyObject*)__pyx_cur_scope), __pyx_n_s_validr__validator_c, __pyx_d, ((PyObject *)__pyx_codeobj__96)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 882, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_validate = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":893
 *         else:
 *             return value.compressed
 *     return validate             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(893,0,__PYX_ERR(0, 893, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_validate);
  __pyx_r = __pyx_v_validate;
  goto __pyx_L0;

  /* "validr/_validator_c.pyx":881
 * 
 * @validator(accept=(str, ipaddress.IPv4Address), output=(str, ipaddress.IPv4Address))
 * def ipv4_validator(compiler, bint output_object=False):             # <<<<<<<<<<<<<<
 *     def validate(value):
 *         try:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("validr._validator_c.ipv4_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_validate);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":897
 * 
 * @validator(accept=(str, ipaddress.IPv6Address), output=(str, ipaddress.IPv6Address))
 * def ipv6_validator(compiler, bint output_object=False):             # <<<<<<<<<<<<<<
 *     def validate(value):
 *         try:
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_67ipv6_validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6validr_12_validator_c_67ipv6_validator = {"ipv6_validator", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6validr_12_validator_c_67ipv6_validator, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6validr_12_validator_c_67ipv6_validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_compiler = 0;
  int __pyx_v_output_object;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ipv6_validator (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_compiler,&__pyx_n_s_output_object,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_compiler)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_output_object);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "ipv6_validator") < 0)) __PYX_ERR(0, 897, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_compiler = values[0];
    if (values[1]) {
      __pyx_v_output_object = __Pyx_PyObject_IsTrue(values[1]); if (unlikely((__pyx_v_output_object == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 897, __pyx_L3_error)
    } else {
      __pyx_v_output_object = ((int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ipv6_validator", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 897, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("validr._validator_c.ipv6_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6validr_12_validator_c_66ipv6_validator(__pyx_self, __pyx_v_compiler, __pyx_v_output_object);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":898
 * @validator(accept=(str, ipaddress.IPv6Address), output=(str, ipaddress.IPv6Address))
 * def ipv6_validator(compiler, bint output_object=False):
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         try:
 *             value = ipaddress.IPv6Address(value.strip())
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_14ipv6_validator_1validate(PyObject *__pyx_self, PyObject *__pyx_v_value); /*proto*/
static PyMethodDef __pyx_mdef_6validr_12_validator_c_14ipv6_validator_1validate = {"validate", (PyCFunction)__pyx_pw_6validr_12_validator_c_14ipv6_validator_1validate, METH_O, 0};
static PyObject *__pyx_pw_6validr_12_validator_c_14ipv6_validator_1validate(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("validate (wrapper)", 0);
  __pyx_r = __pyx_pf_6validr_12_validator_c_14ipv6_validator_validate(__pyx_self, ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6validr_12_validator_c_14ipv6_validator_validate(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_23_ipv6_validator *__pyx_cur_scope;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_23_ipv6_validator *__pyx_outer_scope;
  PyObject *__pyx_v_ex = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  char const *__pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  int __pyx_t_20;
  __Pyx_RefNannySetupContext("validate", 0);
  __pyx_outer_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_23_ipv6_validator *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_TraceCall("validate", __pyx_f[0], 898, 0, __PYX_ERR(0, 898, __pyx_L1_error));
  __Pyx_INCREF(__pyx_v_value);

  /* "validr/_validator_c.pyx":899
 * def ipv6_validator(compiler, bint output_object=False):
 *     def validate(value):
 *         try:             # <<<<<<<<<<<<<<
 *             value = ipaddress.IPv6Address(value.strip())
 *         except ipaddress.AddressValueError as ex:
 */
  __Pyx_TraceLine(899,0,__PYX_ERR(0, 899, __pyx_L1_error))
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "validr/_validator_c.pyx":900
 *     def validate(value):
 *         try:
 *             value = ipaddress.IPv6Address(value.strip())             # <<<<<<<<<<<<<<
 *         except ipaddress.AddressValueError as ex:
 *             raise Invalid(str(ex)) from None
 */
      __Pyx_TraceLine(900,0,__PYX_ERR(0, 900, __pyx_L3_error))
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_ipaddress); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 900, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_IPv6Address); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 900, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_strip); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 900, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      __pyx_t_5 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 900, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
        }
      }
      __pyx_t_4 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_7, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_5);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 900, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "validr/_validator_c.pyx":899
 * def ipv6_validator(compiler, bint output_object=False):
 *     def validate(value):
 *         try:             # <<<<<<<<<<<<<<
 *             value = ipaddress.IPv6Address(value.strip())
 *         except ipaddress.AddressValueError as ex:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "validr/_validator_c.pyx":901
 *         try:
 *             value = ipaddress.IPv6Address(value.strip())
 *         except ipaddress.AddressValueError as ex:             # <<<<<<<<<<<<<<
 *             raise Invalid(str(ex)) from None
 *         except Exception:
 */
    __Pyx_TraceLine(901,0,__PYX_ERR(0, 901, __pyx_L5_except_error))
    __Pyx_ErrFetch(&__pyx_t_4, &__pyx_t_6, &__pyx_t_5);
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_ipaddress); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 901, __pyx_L5_except_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_AddressValueError); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 901, __pyx_L5_except_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_9 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_4, __pyx_t_8);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_ErrRestore(__pyx_t_4, __pyx_t_6, __pyx_t_5);
    __pyx_t_4 = 0; __pyx_t_6 = 0; __pyx_t_5 = 0;
    if (__pyx_t_9) {
      __Pyx_AddTraceback("validr._validator_c.ipv6_validator.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_4) < 0) __PYX_ERR(0, 901, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_6);
      __pyx_v_ex = __pyx_t_6;
      /*try:*/ {

        /* "validr/_validator_c.pyx":902
 *             value = ipaddress.IPv6Address(value.strip())
 *         except ipaddress.AddressValueError as ex:
 *             raise Invalid(str(ex)) from None             # <<<<<<<<<<<<<<
 *         except Exception:
 *             raise Invalid('invalid ipv6 address') from None
 */
        __Pyx_TraceLine(902,0,__PYX_ERR(0, 902, __pyx_L14_error))
        __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 902, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_10 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyUnicode_Type)), __pyx_v_ex); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 902, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_11 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_11)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_11);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        __pyx_t_8 = (__pyx_t_11) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_11, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_10);
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 902, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_Raise(__pyx_t_8, 0, 0, Py_None);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __PYX_ERR(0, 902, __pyx_L14_error)
      }

      /* "validr/_validator_c.pyx":901
 *         try:
 *             value = ipaddress.IPv6Address(value.strip())
 *         except ipaddress.AddressValueError as ex:             # <<<<<<<<<<<<<<
 *             raise Invalid(str(ex)) from None
 *         except Exception:
 */
      __Pyx_TraceLine(901,0,__PYX_ERR(0, 901, __pyx_L14_error))
      /*finally:*/ {
        __pyx_L14_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16) < 0)) __Pyx_ErrFetch(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_18);
          __Pyx_XGOTREF(__pyx_t_19);
          __pyx_t_9 = __pyx_lineno; __pyx_t_12 = __pyx_clineno; __pyx_t_13 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_ex);
            __pyx_v_ex = NULL;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_XGIVEREF(__pyx_t_19);
            __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
          }
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_XGIVEREF(__pyx_t_16);
          __Pyx_ErrRestore(__pyx_t_14, __pyx_t_15, __pyx_t_16);
          __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
          __pyx_lineno = __pyx_t_9; __pyx_clineno = __pyx_t_12; __pyx_filename = __pyx_t_13;
          goto __pyx_L5_except_error;
        }
      }
    }

    /* "validr/_validator_c.pyx":903
 *         except ipaddress.AddressValueError as ex:
 *             raise Invalid(str(ex)) from None
 *         except Exception:             # <<<<<<<<<<<<<<
 *             raise Invalid('invalid ipv6 address') from None
 *         if output_object:
 */
    __Pyx_TraceLine(903,0,__PYX_ERR(0, 903, __pyx_L5_except_error))
    __pyx_t_12 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_12) {
      __Pyx_AddTraceback("validr._validator_c.ipv6_validator.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_6, &__pyx_t_5) < 0) __PYX_ERR(0, 903, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_5);

      /* "validr/_validator_c.pyx":904
 *             raise Invalid(str(ex)) from None
 *         except Exception:
 *             raise Invalid('invalid ipv6 address') from None             # <<<<<<<<<<<<<<
 *         if output_object:
 *             return value
 */
      __Pyx_TraceLine(904,0,__PYX_ERR(0, 904, __pyx_L5_except_error))
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 904, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_10 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      __pyx_t_8 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_10, __pyx_kp_u_invalid_ipv6_address) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_kp_u_invalid_ipv6_address);
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 904, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_Raise(__pyx_t_8, 0, 0, Py_None);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(0, 904, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "validr/_validator_c.pyx":899
 * def ipv6_validator(compiler, bint output_object=False):
 *     def validate(value):
 *         try:             # <<<<<<<<<<<<<<
 *             value = ipaddress.IPv6Address(value.strip())
 *         except ipaddress.AddressValueError as ex:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "validr/_validator_c.pyx":905
 *         except Exception:
 *             raise Invalid('invalid ipv6 address') from None
 *         if output_object:             # <<<<<<<<<<<<<<
 *             return value
 *         else:
 */
  __Pyx_TraceLine(905,0,__PYX_ERR(0, 905, __pyx_L1_error))
  __pyx_t_20 = (__pyx_cur_scope->__pyx_v_output_object != 0);
  if (__pyx_t_20) {

    /* "validr/_validator_c.pyx":906
 *             raise Invalid('invalid ipv6 address') from None
 *         if output_object:
 *             return value             # <<<<<<<<<<<<<<
 *         else:
 *             return value.compressed
 */
    __Pyx_TraceLine(906,0,__PYX_ERR(0, 906, __pyx_L1_error))
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_value);
    __pyx_r = __pyx_v_value;
    goto __pyx_L0;

    /* "validr/_validator_c.pyx":905
 *         except Exception:
 *             raise Invalid('invalid ipv6 address') from None
 *         if output_object:             # <<<<<<<<<<<<<<
 *             return value
 *         else:
 */
  }

  /* "validr/_validator_c.pyx":908
 *             return value
 *         else:
 *             return value.compressed             # <<<<<<<<<<<<<<
 *     return validate
 * 
 */
  __Pyx_TraceLine(908,0,__PYX_ERR(0, 908, __pyx_L1_error))
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_compressed); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 908, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;
  }

  /* "validr/_validator_c.pyx":898
 * @validator(accept=(str, ipaddress.IPv6Address), output=(str, ipaddress.IPv6Address))
 * def ipv6_validator(compiler, bint output_object=False):
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         try:
 *             value = ipaddress.IPv6Address(value.strip())
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("validr._validator_c.ipv6_validator.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ex);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":897
 * 
 * @validator(accept=(str, ipaddress.IPv6Address), output=(str, ipaddress.IPv6Address))
 * def ipv6_validator(compiler, bint output_object=False):             # <<<<<<<<<<<<<<
 *     def validate(value):
 *         try:
 */

static PyObject *__pyx_pf_6validr_12_validator_c_66ipv6_validator(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_compiler, int __pyx_v_output_object) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_23_ipv6_validator *__pyx_cur_scope;
  PyObject *__pyx_v_validate = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_TraceFrameInit(__pyx_codeobj__97)
  __Pyx_RefNannySetupContext("ipv6_validator", 0);
  __pyx_cur_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_23_ipv6_validator *)__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_23_ipv6_validator(__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_23_ipv6_validator, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_23_ipv6_validator *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 897, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __Pyx_TraceCall("ipv6_validator", __pyx_f[0], 897, 0, __PYX_ERR(0, 897, __pyx_L1_error));
  __pyx_cur_scope->__pyx_v_output_object = __pyx_v_output_object;

  /* "validr/_validator_c.pyx":898
 * @validator(accept=(str, ipaddress.IPv6Address), output=(str, ipaddress.IPv6Address))
 * def ipv6_validator(compiler, bint output_object=False):
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         try:
 *             value = ipaddress.IPv6Address(value.strip())
 */
  __Pyx_TraceLine(898,0,__PYX_ERR(0, 898, __pyx_L1_error))
  __pyx_t_1 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_14ipv6_validator_1validate, 0, __pyx_n_s_ipv6_validator_locals_validate, ((PyObject*)__pyx_cur_scope), __pyx_n_s_validr__validator_c, __pyx_d, ((PyObject *)__pyx_codeobj__99)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 898, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_validate = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":909
 *         else:
 *             return value.compressed
 *     return validate             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(909,0,__PYX_ERR(0, 909, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_validate);
  __pyx_r = __pyx_v_validate;
  goto __pyx_L0;

  /* "validr/_validator_c.pyx":897
 * 
 * @validator(accept=(str, ipaddress.IPv6Address), output=(str, ipaddress.IPv6Address))
 * def ipv6_validator(compiler, bint output_object=False):             # <<<<<<<<<<<<<<
 *     def validate(value):
 *         try:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("validr._validator_c.ipv6_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_validate);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":913
 * 
 * @validator(accept=str, output=str)
 * def email_validator(compiler):             # <<<<<<<<<<<<<<
 *     # https://stackoverflow.com/questions/201323/using-a-regular-expression-to-validate-an-email-address
 *     # http://emailregex.com/
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_69email_validator(PyObject *__pyx_self, PyObject *__pyx_v_compiler); /*proto*/
static PyMethodDef __pyx_mdef_6validr_12_validator_c_69email_validator = {"email_validator", (PyCFunction)__pyx_pw_6validr_12_validator_c_69email_validator, METH_O, 0};
static PyObject *__pyx_pw_6validr_12_validator_c_69email_validator(PyObject *__pyx_self, PyObject *__pyx_v_compiler) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("email_validator (wrapper)", 0);
  __pyx_r = __pyx_pf_6validr_12_validator_c_68email_validator(__pyx_self, ((PyObject *)__pyx_v_compiler));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":924
 *     )
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         try:
 *             value = _validate(value.strip())
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_15email_validator_1validate(PyObject *__pyx_self, PyObject *__pyx_v_value); /*proto*/
static PyMethodDef __pyx_mdef_6validr_12_validator_c_15email_validator_1validate = {"validate", (PyCFunction)__pyx_pw_6validr_12_validator_c_15email_validator_1validate, METH_O, 0};
static PyObject *__pyx_pw_6validr_12_validator_c_15email_validator_1validate(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("validate (wrapper)", 0);
  __pyx_r = __pyx_pf_6validr_12_validator_c_15email_validator_validate(__pyx_self, ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6validr_12_validator_c_15email_validator_validate(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_24_email_validator *__pyx_cur_scope;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_24_email_validator *__pyx_outer_scope;
  PyObject *__pyx_v_ex = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  char const *__pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  __Pyx_RefNannySetupContext("validate", 0);
  __pyx_outer_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_24_email_validator *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_TraceCall("validate", __pyx_f[0], 924, 0, __PYX_ERR(0, 924, __pyx_L1_error));
  __Pyx_INCREF(__pyx_v_value);

  /* "validr/_validator_c.pyx":925
 * 
 *     def validate(value):
 *         try:             # <<<<<<<<<<<<<<
 *             value = _validate(value.strip())
 *         except EmailNotValidError as ex:
 */
  __Pyx_TraceLine(925,0,__PYX_ERR(0, 925, __pyx_L1_error))
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "validr/_validator_c.pyx":926
 *     def validate(value):
 *         try:
 *             value = _validate(value.strip())             # <<<<<<<<<<<<<<
 *         except EmailNotValidError as ex:
 *             raise Invalid(str(ex)) from None
 */
      __Pyx_TraceLine(926,0,__PYX_ERR(0, 926, __pyx_L3_error))
      if (unlikely(!__pyx_cur_scope->__pyx_v__validate)) { __Pyx_RaiseClosureNameError("_validate"); __PYX_ERR(0, 926, __pyx_L3_error) }
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_strip); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 926, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
        }
      }
      __pyx_t_5 = (__pyx_t_7) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_7) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 926, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v__validate);
      __pyx_t_6 = __pyx_cur_scope->__pyx_v__validate; __pyx_t_7 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
        }
      }
      __pyx_t_4 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_7, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_5);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 926, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "validr/_validator_c.pyx":925
 * 
 *     def validate(value):
 *         try:             # <<<<<<<<<<<<<<
 *             value = _validate(value.strip())
 *         except EmailNotValidError as ex:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "validr/_validator_c.pyx":927
 *         try:
 *             value = _validate(value.strip())
 *         except EmailNotValidError as ex:             # <<<<<<<<<<<<<<
 *             raise Invalid(str(ex)) from None
 *         except Exception:
 */
    __Pyx_TraceLine(927,0,__PYX_ERR(0, 927, __pyx_L5_except_error))
    __Pyx_ErrFetch(&__pyx_t_4, &__pyx_t_6, &__pyx_t_5);
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_EmailNotValidError); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 927, __pyx_L5_except_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_4, __pyx_t_7);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_ErrRestore(__pyx_t_4, __pyx_t_6, __pyx_t_5);
    __pyx_t_4 = 0; __pyx_t_6 = 0; __pyx_t_5 = 0;
    if (__pyx_t_8) {
      __Pyx_AddTraceback("validr._validator_c.email_validator.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_4) < 0) __PYX_ERR(0, 927, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_6);
      __pyx_v_ex = __pyx_t_6;
      /*try:*/ {

        /* "validr/_validator_c.pyx":928
 *             value = _validate(value.strip())
 *         except EmailNotValidError as ex:
 *             raise Invalid(str(ex)) from None             # <<<<<<<<<<<<<<
 *         except Exception:
 *             raise Invalid('invalid email address') from None
 */
        __Pyx_TraceLine(928,0,__PYX_ERR(0, 928, __pyx_L14_error))
        __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 928, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_10 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyUnicode_Type)), __pyx_v_ex); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 928, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_11 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
          __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_9);
          if (likely(__pyx_t_11)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
            __Pyx_INCREF(__pyx_t_11);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_9, function);
          }
        }
        __pyx_t_7 = (__pyx_t_11) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_11, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_10);
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 928, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_Raise(__pyx_t_7, 0, 0, Py_None);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __PYX_ERR(0, 928, __pyx_L14_error)
      }

      /* "validr/_validator_c.pyx":927
 *         try:
 *             value = _validate(value.strip())
 *         except EmailNotValidError as ex:             # <<<<<<<<<<<<<<
 *             raise Invalid(str(ex)) from None
 *         except Exception:
 */
      __Pyx_TraceLine(927,0,__PYX_ERR(0, 927, __pyx_L14_error))
      /*finally:*/ {
        __pyx_L14_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16) < 0)) __Pyx_ErrFetch(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_18);
          __Pyx_XGOTREF(__pyx_t_19);
          __pyx_t_8 = __pyx_lineno; __pyx_t_12 = __pyx_clineno; __pyx_t_13 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_ex);
            __pyx_v_ex = NULL;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_XGIVEREF(__pyx_t_19);
            __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
          }
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_XGIVEREF(__pyx_t_16);
          __Pyx_ErrRestore(__pyx_t_14, __pyx_t_15, __pyx_t_16);
          __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
          __pyx_lineno = __pyx_t_8; __pyx_clineno = __pyx_t_12; __pyx_filename = __pyx_t_13;
          goto __pyx_L5_except_error;
        }
      }
    }

    /* "validr/_validator_c.pyx":929
 *         except EmailNotValidError as ex:
 *             raise Invalid(str(ex)) from None
 *         except Exception:             # <<<<<<<<<<<<<<
 *             raise Invalid('invalid email address') from None
 *         return value['email']
 */
    __Pyx_TraceLine(929,0,__PYX_ERR(0, 929, __pyx_L5_except_error))
    __pyx_t_12 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_12) {
      __Pyx_AddTraceback("validr._validator_c.email_validator.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_6, &__pyx_t_5) < 0) __PYX_ERR(0, 929, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_5);

      /* "validr/_validator_c.pyx":930
 *             raise Invalid(str(ex)) from None
 *         except Exception:
 *             raise Invalid('invalid email address') from None             # <<<<<<<<<<<<<<
 *         return value['email']
 *     return validate
 */
      __Pyx_TraceLine(930,0,__PYX_ERR(0, 930, __pyx_L5_except_error))
      __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 930, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_9, function);
        }
      }
      __pyx_t_7 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_10, __pyx_kp_u_invalid_email_address) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_kp_u_invalid_email_address);
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 930, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_Raise(__pyx_t_7, 0, 0, Py_None);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __PYX_ERR(0, 930, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "validr/_validator_c.pyx":925
 * 
 *     def validate(value):
 *         try:             # <<<<<<<<<<<<<<
 *             value = _validate(value.strip())
 *         except EmailNotValidError as ex:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "validr/_validator_c.pyx":931
 *         except Exception:
 *             raise Invalid('invalid email address') from None
 *         return value['email']             # <<<<<<<<<<<<<<
 *     return validate
 * 
 */
  __Pyx_TraceLine(931,0,__PYX_ERR(0, 931, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_v_value, __pyx_n_u_email); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 931, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "validr/_validator_c.pyx":924
 *     )
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         try:
 *             value = _validate(value.strip())
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("validr._validator_c.email_validator.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ex);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":913
 * 
 * @validator(accept=str, output=str)
 * def email_validator(compiler):             # <<<<<<<<<<<<<<
 *     # https://stackoverflow.com/questions/201323/using-a-regular-expression-to-validate-an-email-address
 *     # http://emailregex.com/
 */

static PyObject *__pyx_pf_6validr_12_validator_c_68email_validator(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_compiler) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_24_email_validator *__pyx_cur_scope;
  PyObject *__pyx_v_validate = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_TraceFrameInit(__pyx_codeobj__100)
  __Pyx_RefNannySetupContext("email_validator", 0);
  __pyx_cur_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_24_email_validator *)__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_24_email_validator(__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_24_email_validator, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_24_email_validator *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 913, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __Pyx_TraceCall("email_validator", __pyx_f[0], 913, 0, __PYX_ERR(0, 913, __pyx_L1_error));

  /* "validr/_validator_c.pyx":917
 *     # http://emailregex.com/
 *     # https://github.com/JoshData/python-email-validator
 *     _validate = partial(             # <<<<<<<<<<<<<<
 *         validate_email,
 *         allow_smtputf8=False,
 */
  __Pyx_TraceLine(917,0,__PYX_ERR(0, 917, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_partial); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 917, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "validr/_validator_c.pyx":918
 *     # https://github.com/JoshData/python-email-validator
 *     _validate = partial(
 *         validate_email,             # <<<<<<<<<<<<<<
 *         allow_smtputf8=False,
 *         check_deliverability=False,
 */
  __Pyx_TraceLine(918,0,__PYX_ERR(0, 918, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_validate_email); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 918, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "validr/_validator_c.pyx":917
 *     # http://emailregex.com/
 *     # https://github.com/JoshData/python-email-validator
 *     _validate = partial(             # <<<<<<<<<<<<<<
 *         validate_email,
 *         allow_smtputf8=False,
 */
  __Pyx_TraceLine(917,0,__PYX_ERR(0, 917, __pyx_L1_error))
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 917, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __pyx_t_2 = 0;

  /* "validr/_validator_c.pyx":919
 *     _validate = partial(
 *         validate_email,
 *         allow_smtputf8=False,             # <<<<<<<<<<<<<<
 *         check_deliverability=False,
 *         allow_empty_local=False,
 */
  __Pyx_TraceLine(919,0,__PYX_ERR(0, 919, __pyx_L1_error))
  __pyx_t_2 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 919, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_allow_smtputf8, Py_False) < 0) __PYX_ERR(0, 919, __pyx_L1_error)

  /* "validr/_validator_c.pyx":920
 *         validate_email,
 *         allow_smtputf8=False,
 *         check_deliverability=False,             # <<<<<<<<<<<<<<
 *         allow_empty_local=False,
 *     )
 */
  __Pyx_TraceLine(920,0,__PYX_ERR(0, 920, __pyx_L1_error))
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_check_deliverability, Py_False) < 0) __PYX_ERR(0, 919, __pyx_L1_error)

  /* "validr/_validator_c.pyx":921
 *         allow_smtputf8=False,
 *         check_deliverability=False,
 *         allow_empty_local=False,             # <<<<<<<<<<<<<<
 *     )
 * 
 */
  __Pyx_TraceLine(921,0,__PYX_ERR(0, 921, __pyx_L1_error))
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_allow_empty_local, Py_False) < 0) __PYX_ERR(0, 919, __pyx_L1_error)

  /* "validr/_validator_c.pyx":917
 *     # http://emailregex.com/
 *     # https://github.com/JoshData/python-email-validator
 *     _validate = partial(             # <<<<<<<<<<<<<<
 *         validate_email,
 *         allow_smtputf8=False,
 */
  __Pyx_TraceLine(917,0,__PYX_ERR(0, 917, __pyx_L1_error))
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 917, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_cur_scope->__pyx_v__validate = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "validr/_validator_c.pyx":924
 *     )
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         try:
 *             value = _validate(value.strip())
 */
  __Pyx_TraceLine(924,0,__PYX_ERR(0, 924, __pyx_L1_error))
  __pyx_t_4 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_15email_validator_1validate, 0, __pyx_n_s_email_validator_locals_validate, ((PyObject*)__pyx_cur_scope), __pyx_n_s_validr__validator_c, __pyx_d, ((PyObject *)__pyx_codeobj__102)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 924, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_validate = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "validr/_validator_c.pyx":932
 *             raise Invalid('invalid email address') from None
 *         return value['email']
 *     return validate             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(932,0,__PYX_ERR(0, 932, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_validate);
  __pyx_r = __pyx_v_validate;
  goto __pyx_L0;

  /* "validr/_validator_c.pyx":913
 * 
 * @validator(accept=str, output=str)
 * def email_validator(compiler):             # <<<<<<<<<<<<<<
 *     # https://stackoverflow.com/questions/201323/using-a-regular-expression-to-validate-an-email-address
 *     # http://emailregex.com/
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("validr._validator_c.email_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_validate);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":936
 * 
 * @validator(output=(str, object))
 * def url_validator(compiler, str scheme='http https', int maxlen=255, bint output_object=False):             # <<<<<<<<<<<<<<
 *     # https://stackoverflow.com/questions/7160737/python-how-to-validate-a-url-in-python-malformed-or-not
 *     # https://stackoverflow.com/questions/827557/how-do-you-validate-a-url-with-a-regular-expression-in-python
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_71url_validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6validr_12_validator_c_71url_validator = {"url_validator", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6validr_12_validator_c_71url_validator, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6validr_12_validator_c_71url_validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_compiler = 0;
  PyObject *__pyx_v_scheme = 0;
  int __pyx_v_maxlen;
  int __pyx_v_output_object;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("url_validator (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_compiler,&__pyx_n_s_scheme,&__pyx_n_s_maxlen,&__pyx_n_s_output_object,0};
    PyObject* values[4] = {0,0,0,0};
    values[1] = ((PyObject*)__pyx_kp_u_http_https);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_compiler)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_scheme);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_maxlen);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_output_object);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "url_validator") < 0)) __PYX_ERR(0, 936, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_compiler = values[0];
    __pyx_v_scheme = ((PyObject*)values[1]);
    if (values[2]) {
      __pyx_v_maxlen = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_maxlen == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 936, __pyx_L3_error)
    } else {
      __pyx_v_maxlen = ((int)0xFF);
    }
    if (values[3]) {
      __pyx_v_output_object = __Pyx_PyObject_IsTrue(values[3]); if (unlikely((__pyx_v_output_object == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 936, __pyx_L3_error)
    } else {
      __pyx_v_output_object = ((int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("url_validator", 0, 1, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 936, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("validr._validator_c.url_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_scheme), (&PyUnicode_Type), 1, "scheme", 1))) __PYX_ERR(0, 936, __pyx_L1_error)
  __pyx_r = __pyx_pf_6validr_12_validator_c_70url_validator(__pyx_self, __pyx_v_compiler, __pyx_v_scheme, __pyx_v_maxlen, __pyx_v_output_object);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":944
 *     allow_scheme = set(scheme.split())
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         try:
 *             value = value.strip()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_13url_validator_1validate(PyObject *__pyx_self, PyObject *__pyx_v_value); /*proto*/
static PyMethodDef __pyx_mdef_6validr_12_validator_c_13url_validator_1validate = {"validate", (PyCFunction)__pyx_pw_6validr_12_validator_c_13url_validator_1validate, METH_O, 0};
static PyObject *__pyx_pw_6validr_12_validator_c_13url_validator_1validate(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("validate (wrapper)", 0);
  __pyx_r = __pyx_pf_6validr_12_validator_c_13url_validator_validate(__pyx_self, ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6validr_12_validator_c_13url_validator_validate(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_25_url_validator *__pyx_cur_scope;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_25_url_validator *__pyx_outer_scope;
  PyObject *__pyx_v_parsed = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  Py_ssize_t __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  int __pyx_t_14;
  __Pyx_RefNannySetupContext("validate", 0);
  __pyx_outer_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_25_url_validator *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_TraceCall("validate", __pyx_f[0], 944, 0, __PYX_ERR(0, 944, __pyx_L1_error));
  __Pyx_INCREF(__pyx_v_value);

  /* "validr/_validator_c.pyx":945
 * 
 *     def validate(value):
 *         try:             # <<<<<<<<<<<<<<
 *             value = value.strip()
 *         except Exception:
 */
  __Pyx_TraceLine(945,0,__PYX_ERR(0, 945, __pyx_L1_error))
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "validr/_validator_c.pyx":946
 *     def validate(value):
 *         try:
 *             value = value.strip()             # <<<<<<<<<<<<<<
 *         except Exception:
 *             raise Invalid('invalid url') from None
 */
      __Pyx_TraceLine(946,0,__PYX_ERR(0, 946, __pyx_L3_error))
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_strip); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 946, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 946, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "validr/_validator_c.pyx":945
 * 
 *     def validate(value):
 *         try:             # <<<<<<<<<<<<<<
 *             value = value.strip()
 *         except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "validr/_validator_c.pyx":947
 *         try:
 *             value = value.strip()
 *         except Exception:             # <<<<<<<<<<<<<<
 *             raise Invalid('invalid url') from None
 *         if len(value) > maxlen:
 */
    __Pyx_TraceLine(947,0,__PYX_ERR(0, 947, __pyx_L5_except_error))
    __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_7) {
      __Pyx_AddTraceback("validr._validator_c.url_validator.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6) < 0) __PYX_ERR(0, 947, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);

      /* "validr/_validator_c.pyx":948
 *             value = value.strip()
 *         except Exception:
 *             raise Invalid('invalid url') from None             # <<<<<<<<<<<<<<
 *         if len(value) > maxlen:
 *             raise Invalid('url length must <= {}'.format(maxlen))
 */
      __Pyx_TraceLine(948,0,__PYX_ERR(0, 948, __pyx_L5_except_error))
      __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 948, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_9, function);
        }
      }
      __pyx_t_8 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_10, __pyx_kp_u_invalid_url) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_kp_u_invalid_url);
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 948, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_Raise(__pyx_t_8, 0, 0, Py_None);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(0, 948, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "validr/_validator_c.pyx":945
 * 
 *     def validate(value):
 *         try:             # <<<<<<<<<<<<<<
 *             value = value.strip()
 *         except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "validr/_validator_c.pyx":949
 *         except Exception:
 *             raise Invalid('invalid url') from None
 *         if len(value) > maxlen:             # <<<<<<<<<<<<<<
 *             raise Invalid('url length must <= {}'.format(maxlen))
 *         try:
 */
  __Pyx_TraceLine(949,0,__PYX_ERR(0, 949, __pyx_L1_error))
  __pyx_t_11 = PyObject_Length(__pyx_v_value); if (unlikely(__pyx_t_11 == ((Py_ssize_t)-1))) __PYX_ERR(0, 949, __pyx_L1_error)
  __pyx_t_12 = ((__pyx_t_11 > __pyx_cur_scope->__pyx_v_maxlen) != 0);
  if (unlikely(__pyx_t_12)) {

    /* "validr/_validator_c.pyx":950
 *             raise Invalid('invalid url') from None
 *         if len(value) > maxlen:
 *             raise Invalid('url length must <= {}'.format(maxlen))             # <<<<<<<<<<<<<<
 *         try:
 *             parsed = urlparse(value)
 */
    __Pyx_TraceLine(950,0,__PYX_ERR(0, 950, __pyx_L1_error))
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 950, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_url_length_must, __pyx_n_s_format); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 950, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyInt_From_int(__pyx_cur_scope->__pyx_v_maxlen); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 950, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_4 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_10, __pyx_t_9) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_9);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 950, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_6 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_8, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 950, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 950, __pyx_L1_error)

    /* "validr/_validator_c.pyx":949
 *         except Exception:
 *             raise Invalid('invalid url') from None
 *         if len(value) > maxlen:             # <<<<<<<<<<<<<<
 *             raise Invalid('url length must <= {}'.format(maxlen))
 *         try:
 */
  }

  /* "validr/_validator_c.pyx":951
 *         if len(value) > maxlen:
 *             raise Invalid('url length must <= {}'.format(maxlen))
 *         try:             # <<<<<<<<<<<<<<
 *             parsed = urlparse(value)
 *         except Exception:
 */
  __Pyx_TraceLine(951,0,__PYX_ERR(0, 951, __pyx_L1_error))
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_2, &__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_1);
    /*try:*/ {

      /* "validr/_validator_c.pyx":952
 *             raise Invalid('url length must <= {}'.format(maxlen))
 *         try:
 *             parsed = urlparse(value)             # <<<<<<<<<<<<<<
 *         except Exception:
 *             raise Invalid('invalid url') from None
 */
      __Pyx_TraceLine(952,0,__PYX_ERR(0, 952, __pyx_L12_error))
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_urlparse); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 952, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_6 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_4, __pyx_v_value) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_value);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 952, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_v_parsed = __pyx_t_6;
      __pyx_t_6 = 0;

      /* "validr/_validator_c.pyx":951
 *         if len(value) > maxlen:
 *             raise Invalid('url length must <= {}'.format(maxlen))
 *         try:             # <<<<<<<<<<<<<<
 *             parsed = urlparse(value)
 *         except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    goto __pyx_L17_try_end;
    __pyx_L12_error:;
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "validr/_validator_c.pyx":953
 *         try:
 *             parsed = urlparse(value)
 *         except Exception:             # <<<<<<<<<<<<<<
 *             raise Invalid('invalid url') from None
 *         if not parsed.scheme or parsed.scheme not in allow_scheme:
 */
    __Pyx_TraceLine(953,0,__PYX_ERR(0, 953, __pyx_L14_except_error))
    __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_7) {
      __Pyx_AddTraceback("validr._validator_c.url_validator.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_5, &__pyx_t_4) < 0) __PYX_ERR(0, 953, __pyx_L14_except_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_4);

      /* "validr/_validator_c.pyx":954
 *             parsed = urlparse(value)
 *         except Exception:
 *             raise Invalid('invalid url') from None             # <<<<<<<<<<<<<<
 *         if not parsed.scheme or parsed.scheme not in allow_scheme:
 *             raise Invalid('invalid url scheme, expect {}'.format(allow_scheme))
 */
      __Pyx_TraceLine(954,0,__PYX_ERR(0, 954, __pyx_L14_except_error))
      __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 954, __pyx_L14_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_9, function);
        }
      }
      __pyx_t_8 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_10, __pyx_kp_u_invalid_url) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_kp_u_invalid_url);
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 954, __pyx_L14_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_Raise(__pyx_t_8, 0, 0, Py_None);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(0, 954, __pyx_L14_except_error)
    }
    goto __pyx_L14_except_error;
    __pyx_L14_except_error:;

    /* "validr/_validator_c.pyx":951
 *         if len(value) > maxlen:
 *             raise Invalid('url length must <= {}'.format(maxlen))
 *         try:             # <<<<<<<<<<<<<<
 *             parsed = urlparse(value)
 *         except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_2, __pyx_t_1);
    goto __pyx_L1_error;
    __pyx_L17_try_end:;
  }

  /* "validr/_validator_c.pyx":955
 *         except Exception:
 *             raise Invalid('invalid url') from None
 *         if not parsed.scheme or parsed.scheme not in allow_scheme:             # <<<<<<<<<<<<<<
 *             raise Invalid('invalid url scheme, expect {}'.format(allow_scheme))
 *         if output_object:
 */
  __Pyx_TraceLine(955,0,__PYX_ERR(0, 955, __pyx_L1_error))
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_parsed, __pyx_n_s_scheme); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 955, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 955, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_14 = ((!__pyx_t_13) != 0);
  if (!__pyx_t_14) {
  } else {
    __pyx_t_12 = __pyx_t_14;
    goto __pyx_L21_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_parsed, __pyx_n_s_scheme); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 955, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (unlikely(!__pyx_cur_scope->__pyx_v_allow_scheme)) { __Pyx_RaiseClosureNameError("allow_scheme"); __PYX_ERR(0, 955, __pyx_L1_error) }
  if (unlikely(__pyx_cur_scope->__pyx_v_allow_scheme == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 955, __pyx_L1_error)
  }
  __pyx_t_14 = (__Pyx_PySet_ContainsTF(__pyx_t_4, __pyx_cur_scope->__pyx_v_allow_scheme, Py_NE)); if (unlikely(__pyx_t_14 < 0)) __PYX_ERR(0, 955, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_13 = (__pyx_t_14 != 0);
  __pyx_t_12 = __pyx_t_13;
  __pyx_L21_bool_binop_done:;
  if (unlikely(__pyx_t_12)) {

    /* "validr/_validator_c.pyx":956
 *             raise Invalid('invalid url') from None
 *         if not parsed.scheme or parsed.scheme not in allow_scheme:
 *             raise Invalid('invalid url scheme, expect {}'.format(allow_scheme))             # <<<<<<<<<<<<<<
 *         if output_object:
 *             return parsed
 */
    __Pyx_TraceLine(956,0,__PYX_ERR(0, 956, __pyx_L1_error))
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 956, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_invalid_url_scheme_expect, __pyx_n_s_format); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 956, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    if (unlikely(!__pyx_cur_scope->__pyx_v_allow_scheme)) { __Pyx_RaiseClosureNameError("allow_scheme"); __PYX_ERR(0, 956, __pyx_L1_error) }
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_6 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_9, __pyx_cur_scope->__pyx_v_allow_scheme) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_cur_scope->__pyx_v_allow_scheme);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 956, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_4 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 956, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 956, __pyx_L1_error)

    /* "validr/_validator_c.pyx":955
 *         except Exception:
 *             raise Invalid('invalid url') from None
 *         if not parsed.scheme or parsed.scheme not in allow_scheme:             # <<<<<<<<<<<<<<
 *             raise Invalid('invalid url scheme, expect {}'.format(allow_scheme))
 *         if output_object:
 */
  }

  /* "validr/_validator_c.pyx":957
 *         if not parsed.scheme or parsed.scheme not in allow_scheme:
 *             raise Invalid('invalid url scheme, expect {}'.format(allow_scheme))
 *         if output_object:             # <<<<<<<<<<<<<<
 *             return parsed
 *         else:
 */
  __Pyx_TraceLine(957,0,__PYX_ERR(0, 957, __pyx_L1_error))
  __pyx_t_12 = (__pyx_cur_scope->__pyx_v_output_object != 0);
  if (__pyx_t_12) {

    /* "validr/_validator_c.pyx":958
 *             raise Invalid('invalid url scheme, expect {}'.format(allow_scheme))
 *         if output_object:
 *             return parsed             # <<<<<<<<<<<<<<
 *         else:
 *             return urlunparse(parsed)
 */
    __Pyx_TraceLine(958,0,__PYX_ERR(0, 958, __pyx_L1_error))
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_parsed);
    __pyx_r = __pyx_v_parsed;
    goto __pyx_L0;

    /* "validr/_validator_c.pyx":957
 *         if not parsed.scheme or parsed.scheme not in allow_scheme:
 *             raise Invalid('invalid url scheme, expect {}'.format(allow_scheme))
 *         if output_object:             # <<<<<<<<<<<<<<
 *             return parsed
 *         else:
 */
  }

  /* "validr/_validator_c.pyx":960
 *             return parsed
 *         else:
 *             return urlunparse(parsed)             # <<<<<<<<<<<<<<
 *     return validate
 * 
 */
  __Pyx_TraceLine(960,0,__PYX_ERR(0, 960, __pyx_L1_error))
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_urlunparse); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 960, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_v_parsed) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_parsed);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 960, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "validr/_validator_c.pyx":944
 *     allow_scheme = set(scheme.split())
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         try:
 *             value = value.strip()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("validr._validator_c.url_validator.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_parsed);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":936
 * 
 * @validator(output=(str, object))
 * def url_validator(compiler, str scheme='http https', int maxlen=255, bint output_object=False):             # <<<<<<<<<<<<<<
 *     # https://stackoverflow.com/questions/7160737/python-how-to-validate-a-url-in-python-malformed-or-not
 *     # https://stackoverflow.com/questions/827557/how-do-you-validate-a-url-with-a-regular-expression-in-python
 */

static PyObject *__pyx_pf_6validr_12_validator_c_70url_validator(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_compiler, PyObject *__pyx_v_scheme, int __pyx_v_maxlen, int __pyx_v_output_object) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_25_url_validator *__pyx_cur_scope;
  PyObject *__pyx_v_validate = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_TraceFrameInit(__pyx_codeobj__103)
  __Pyx_RefNannySetupContext("url_validator", 0);
  __pyx_cur_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_25_url_validator *)__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_25_url_validator(__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_25_url_validator, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_25_url_validator *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 936, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __Pyx_TraceCall("url_validator", __pyx_f[0], 936, 0, __PYX_ERR(0, 936, __pyx_L1_error));
  __pyx_cur_scope->__pyx_v_maxlen = __pyx_v_maxlen;
  __pyx_cur_scope->__pyx_v_output_object = __pyx_v_output_object;

  /* "validr/_validator_c.pyx":942
 *     # https://github.com/dgerber/rfc3987
 *     # https://github.com/tkem/uritools
 *     allow_scheme = set(scheme.split())             # <<<<<<<<<<<<<<
 * 
 *     def validate(value):
 */
  __Pyx_TraceLine(942,0,__PYX_ERR(0, 942, __pyx_L1_error))
  if (unlikely(__pyx_v_scheme == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "split");
    __PYX_ERR(0, 942, __pyx_L1_error)
  }
  __pyx_t_1 = PyUnicode_Split(__pyx_v_scheme, ((PyObject *)NULL), -1L); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 942, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PySet_New(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 942, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_cur_scope->__pyx_v_allow_scheme = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "validr/_validator_c.pyx":944
 *     allow_scheme = set(scheme.split())
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         try:
 *             value = value.strip()
 */
  __Pyx_TraceLine(944,0,__PYX_ERR(0, 944, __pyx_L1_error))
  __pyx_t_2 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_13url_validator_1validate, 0, __pyx_n_s_url_validator_locals_validate, ((PyObject*)__pyx_cur_scope), __pyx_n_s_validr__validator_c, __pyx_d, ((PyObject *)__pyx_codeobj__105)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 944, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_validate = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "validr/_validator_c.pyx":961
 *         else:
 *             return urlunparse(parsed)
 *     return validate             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(961,0,__PYX_ERR(0, 961, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_validate);
  __pyx_r = __pyx_v_validate;
  goto __pyx_L0;

  /* "validr/_validator_c.pyx":936
 * 
 * @validator(output=(str, object))
 * def url_validator(compiler, str scheme='http https', int maxlen=255, bint output_object=False):             # <<<<<<<<<<<<<<
 *     # https://stackoverflow.com/questions/7160737/python-how-to-validate-a-url-in-python-malformed-or-not
 *     # https://stackoverflow.com/questions/827557/how-do-you-validate-a-url-with-a-regular-expression-in-python
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("validr._validator_c.url_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_validate);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":965
 * 
 * @validator(output=str)
 * def fqdn_validator(compiler):             # <<<<<<<<<<<<<<
 *     def validate(value):
 *         try:
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_73fqdn_validator(PyObject *__pyx_self, PyObject *__pyx_v_compiler); /*proto*/
static PyMethodDef __pyx_mdef_6validr_12_validator_c_73fqdn_validator = {"fqdn_validator", (PyCFunction)__pyx_pw_6validr_12_validator_c_73fqdn_validator, METH_O, 0};
static PyObject *__pyx_pw_6validr_12_validator_c_73fqdn_validator(PyObject *__pyx_self, PyObject *__pyx_v_compiler) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("fqdn_validator (wrapper)", 0);
  __pyx_r = __pyx_pf_6validr_12_validator_c_72fqdn_validator(__pyx_self, ((PyObject *)__pyx_v_compiler));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":966
 * @validator(output=str)
 * def fqdn_validator(compiler):
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         try:
 *             value = value.strip()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_14fqdn_validator_1validate(PyObject *__pyx_self, PyObject *__pyx_v_value); /*proto*/
static PyMethodDef __pyx_mdef_6validr_12_validator_c_14fqdn_validator_1validate = {"validate", (PyCFunction)__pyx_pw_6validr_12_validator_c_14fqdn_validator_1validate, METH_O, 0};
static PyObject *__pyx_pw_6validr_12_validator_c_14fqdn_validator_1validate(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("validate (wrapper)", 0);
  __pyx_r = __pyx_pf_6validr_12_validator_c_14fqdn_validator_validate(__pyx_self, ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6validr_12_validator_c_14fqdn_validator_validate(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_fqdn_obj = NULL;
  PyObject *__pyx_v_ex = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  char const *__pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  __Pyx_RefNannySetupContext("validate", 0);
  __Pyx_TraceCall("validate", __pyx_f[0], 966, 0, __PYX_ERR(0, 966, __pyx_L1_error));
  __Pyx_INCREF(__pyx_v_value);

  /* "validr/_validator_c.pyx":967
 * def fqdn_validator(compiler):
 *     def validate(value):
 *         try:             # <<<<<<<<<<<<<<
 *             value = value.strip()
 *             fqdn_obj = FQDN(value)
 */
  __Pyx_TraceLine(967,0,__PYX_ERR(0, 967, __pyx_L1_error))
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "validr/_validator_c.pyx":968
 *     def validate(value):
 *         try:
 *             value = value.strip()             # <<<<<<<<<<<<<<
 *             fqdn_obj = FQDN(value)
 *             if fqdn_obj.is_valid:
 */
      __Pyx_TraceLine(968,0,__PYX_ERR(0, 968, __pyx_L3_error))
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_strip); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 968, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 968, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "validr/_validator_c.pyx":969
 *         try:
 *             value = value.strip()
 *             fqdn_obj = FQDN(value)             # <<<<<<<<<<<<<<
 *             if fqdn_obj.is_valid:
 *                 return fqdn_obj.relative
 */
      __Pyx_TraceLine(969,0,__PYX_ERR(0, 969, __pyx_L3_error))
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_FQDN); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 969, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_v_value) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_value);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 969, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_v_fqdn_obj = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "validr/_validator_c.pyx":970
 *             value = value.strip()
 *             fqdn_obj = FQDN(value)
 *             if fqdn_obj.is_valid:             # <<<<<<<<<<<<<<
 *                 return fqdn_obj.relative
 *         except (ValueError, TypeError, AttributeError) as ex:
 */
      __Pyx_TraceLine(970,0,__PYX_ERR(0, 970, __pyx_L3_error))
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_fqdn_obj, __pyx_n_s_is_valid); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 970, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 970, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_7) {

        /* "validr/_validator_c.pyx":971
 *             fqdn_obj = FQDN(value)
 *             if fqdn_obj.is_valid:
 *                 return fqdn_obj.relative             # <<<<<<<<<<<<<<
 *         except (ValueError, TypeError, AttributeError) as ex:
 *             raise Invalid("invalid fqdn") from ex
 */
        __Pyx_TraceLine(971,0,__PYX_ERR(0, 971, __pyx_L3_error))
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_fqdn_obj, __pyx_n_s_relative); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 971, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_r = __pyx_t_4;
        __pyx_t_4 = 0;
        goto __pyx_L7_try_return;

        /* "validr/_validator_c.pyx":970
 *             value = value.strip()
 *             fqdn_obj = FQDN(value)
 *             if fqdn_obj.is_valid:             # <<<<<<<<<<<<<<
 *                 return fqdn_obj.relative
 *         except (ValueError, TypeError, AttributeError) as ex:
 */
      }

      /* "validr/_validator_c.pyx":967
 * def fqdn_validator(compiler):
 *     def validate(value):
 *         try:             # <<<<<<<<<<<<<<
 *             value = value.strip()
 *             fqdn_obj = FQDN(value)
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "validr/_validator_c.pyx":972
 *             if fqdn_obj.is_valid:
 *                 return fqdn_obj.relative
 *         except (ValueError, TypeError, AttributeError) as ex:             # <<<<<<<<<<<<<<
 *             raise Invalid("invalid fqdn") from ex
 *         raise Invalid("invalid fqdn")
 */
    __Pyx_TraceLine(972,0,__PYX_ERR(0, 972, __pyx_L5_except_error))
    __pyx_t_8 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ValueError) || __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError) || __Pyx_PyErr_ExceptionMatches(__pyx_builtin_AttributeError);
    if (__pyx_t_8) {
      __Pyx_AddTraceback("validr._validator_c.fqdn_validator.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6) < 0) __PYX_ERR(0, 972, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_5);
      __pyx_v_ex = __pyx_t_5;
      /*try:*/ {

        /* "validr/_validator_c.pyx":973
 *                 return fqdn_obj.relative
 *         except (ValueError, TypeError, AttributeError) as ex:
 *             raise Invalid("invalid fqdn") from ex             # <<<<<<<<<<<<<<
 *         raise Invalid("invalid fqdn")
 *     return validate
 */
        __Pyx_TraceLine(973,0,__PYX_ERR(0, 973, __pyx_L15_error))
        __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 973, __pyx_L15_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_11 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
          __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_10);
          if (likely(__pyx_t_11)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
            __Pyx_INCREF(__pyx_t_11);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_10, function);
          }
        }
        __pyx_t_9 = (__pyx_t_11) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_11, __pyx_kp_u_invalid_fqdn) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_kp_u_invalid_fqdn);
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 973, __pyx_L15_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_Raise(__pyx_t_9, 0, 0, __pyx_v_ex);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __PYX_ERR(0, 973, __pyx_L15_error)
      }

      /* "validr/_validator_c.pyx":972
 *             if fqdn_obj.is_valid:
 *                 return fqdn_obj.relative
 *         except (ValueError, TypeError, AttributeError) as ex:             # <<<<<<<<<<<<<<
 *             raise Invalid("invalid fqdn") from ex
 *         raise Invalid("invalid fqdn")
 */
      __Pyx_TraceLine(972,0,__PYX_ERR(0, 972, __pyx_L15_error))
      /*finally:*/ {
        __pyx_L15_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16) < 0)) __Pyx_ErrFetch(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_18);
          __Pyx_XGOTREF(__pyx_t_19);
          __pyx_t_8 = __pyx_lineno; __pyx_t_12 = __pyx_clineno; __pyx_t_13 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_ex);
            __pyx_v_ex = NULL;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_XGIVEREF(__pyx_t_19);
            __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
          }
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_XGIVEREF(__pyx_t_16);
          __Pyx_ErrRestore(__pyx_t_14, __pyx_t_15, __pyx_t_16);
          __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
          __pyx_lineno = __pyx_t_8; __pyx_clineno = __pyx_t_12; __pyx_filename = __pyx_t_13;
          goto __pyx_L5_except_error;
        }
      }
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "validr/_validator_c.pyx":967
 * def fqdn_validator(compiler):
 *     def validate(value):
 *         try:             # <<<<<<<<<<<<<<
 *             value = value.strip()
 *             fqdn_obj = FQDN(value)
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L7_try_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
    __pyx_L8_try_end:;
  }

  /* "validr/_validator_c.pyx":974
 *         except (ValueError, TypeError, AttributeError) as ex:
 *             raise Invalid("invalid fqdn") from ex
 *         raise Invalid("invalid fqdn")             # <<<<<<<<<<<<<<
 *     return validate
 * 
 */
  __Pyx_TraceLine(974,0,__PYX_ERR(0, 974, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 974, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_6 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_4, __pyx_kp_u_invalid_fqdn) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_kp_u_invalid_fqdn);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 974, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_Raise(__pyx_t_6, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __PYX_ERR(0, 974, __pyx_L1_error)

  /* "validr/_validator_c.pyx":966
 * @validator(output=str)
 * def fqdn_validator(compiler):
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         try:
 *             value = value.strip()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("validr._validator_c.fqdn_validator.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_fqdn_obj);
  __Pyx_XDECREF(__pyx_v_ex);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":965
 * 
 * @validator(output=str)
 * def fqdn_validator(compiler):             # <<<<<<<<<<<<<<
 *     def validate(value):
 *         try:
 */

static PyObject *__pyx_pf_6validr_12_validator_c_72fqdn_validator(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_compiler) {
  PyObject *__pyx_v_validate = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_TraceFrameInit(__pyx_codeobj__106)
  __Pyx_RefNannySetupContext("fqdn_validator", 0);
  __Pyx_TraceCall("fqdn_validator", __pyx_f[0], 965, 0, __PYX_ERR(0, 965, __pyx_L1_error));

  /* "validr/_validator_c.pyx":966
 * @validator(output=str)
 * def fqdn_validator(compiler):
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         try:
 *             value = value.strip()
 */
  __Pyx_TraceLine(966,0,__PYX_ERR(0, 966, __pyx_L1_error))
  __pyx_t_1 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_14fqdn_validator_1validate, 0, __pyx_n_s_fqdn_validator_locals_validate, NULL, __pyx_n_s_validr__validator_c, __pyx_d, ((PyObject *)__pyx_codeobj__108)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 966, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_validate = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":975
 *             raise Invalid("invalid fqdn") from ex
 *         raise Invalid("invalid fqdn")
 *     return validate             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(975,0,__PYX_ERR(0, 975, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_validate);
  __pyx_r = __pyx_v_validate;
  goto __pyx_L0;

  /* "validr/_validator_c.pyx":965
 * 
 * @validator(output=str)
 * def fqdn_validator(compiler):             # <<<<<<<<<<<<<<
 *     def validate(value):
 *         try:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("validr._validator_c.fqdn_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_validate);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":979
 * 
 * @validator(output=(str, uuid.UUID))
 * def uuid_validator(compiler, version=None, bint output_object=False):             # <<<<<<<<<<<<<<
 *     if version is None:
 *         msg = 'invalid uuid'
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_75uuid_validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6validr_12_validator_c_75uuid_validator = {"uuid_validator", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6validr_12_validator_c_75uuid_validator, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6validr_12_validator_c_75uuid_validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_compiler = 0;
  PyObject *__pyx_v_version = 0;
  int __pyx_v_output_object;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("uuid_validator (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_compiler,&__pyx_n_s_version,&__pyx_n_s_output_object,0};
    PyObject* values[3] = {0,0,0};
    values[1] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_compiler)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_version);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_output_object);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "uuid_validator") < 0)) __PYX_ERR(0, 979, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_compiler = values[0];
    __pyx_v_version = values[1];
    if (values[2]) {
      __pyx_v_output_object = __Pyx_PyObject_IsTrue(values[2]); if (unlikely((__pyx_v_output_object == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 979, __pyx_L3_error)
    } else {
      __pyx_v_output_object = ((int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("uuid_validator", 0, 1, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 979, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("validr._validator_c.uuid_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6validr_12_validator_c_74uuid_validator(__pyx_self, __pyx_v_compiler, __pyx_v_version, __pyx_v_output_object);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":987
 *         msg = 'invalid uuid{}'.format(version)
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         if not isinstance(value, uuid.UUID):
 *             try:
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_14uuid_validator_1validate(PyObject *__pyx_self, PyObject *__pyx_v_value); /*proto*/
static PyMethodDef __pyx_mdef_6validr_12_validator_c_14uuid_validator_1validate = {"validate", (PyCFunction)__pyx_pw_6validr_12_validator_c_14uuid_validator_1validate, METH_O, 0};
static PyObject *__pyx_pw_6validr_12_validator_c_14uuid_validator_1validate(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("validate (wrapper)", 0);
  __pyx_r = __pyx_pf_6validr_12_validator_c_14uuid_validator_validate(__pyx_self, ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6validr_12_validator_c_14uuid_validator_validate(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_26_uuid_validator *__pyx_cur_scope;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_26_uuid_validator *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  __Pyx_RefNannySetupContext("validate", 0);
  __pyx_outer_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_26_uuid_validator *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_TraceCall("validate", __pyx_f[0], 987, 0, __PYX_ERR(0, 987, __pyx_L1_error));
  __Pyx_INCREF(__pyx_v_value);

  /* "validr/_validator_c.pyx":988
 * 
 *     def validate(value):
 *         if not isinstance(value, uuid.UUID):             # <<<<<<<<<<<<<<
 *             try:
 *                 value = uuid.UUID(value.strip())
 */
  __Pyx_TraceLine(988,0,__PYX_ERR(0, 988, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_uuid); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 988, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_UUID); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 988, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = PyObject_IsInstance(__pyx_v_value, __pyx_t_2); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 988, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = ((!(__pyx_t_3 != 0)) != 0);
  if (__pyx_t_4) {

    /* "validr/_validator_c.pyx":989
 *     def validate(value):
 *         if not isinstance(value, uuid.UUID):
 *             try:             # <<<<<<<<<<<<<<
 *                 value = uuid.UUID(value.strip())
 *             except Exception:
 */
    __Pyx_TraceLine(989,0,__PYX_ERR(0, 989, __pyx_L1_error))
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      /*try:*/ {

        /* "validr/_validator_c.pyx":990
 *         if not isinstance(value, uuid.UUID):
 *             try:
 *                 value = uuid.UUID(value.strip())             # <<<<<<<<<<<<<<
 *             except Exception:
 *                 raise Invalid(msg) from None
 */
        __Pyx_TraceLine(990,0,__PYX_ERR(0, 990, __pyx_L4_error))
        __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_uuid); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 990, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_UUID); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 990, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_strip); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 990, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_10 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
          __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
          if (likely(__pyx_t_10)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
            __Pyx_INCREF(__pyx_t_10);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_9, function);
          }
        }
        __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_10) : __Pyx_PyObject_CallNoArg(__pyx_t_9);
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 990, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_9 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
          __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
          if (likely(__pyx_t_9)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
            __Pyx_INCREF(__pyx_t_9);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_8, function);
          }
        }
        __pyx_t_2 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_9, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_1);
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 990, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "validr/_validator_c.pyx":989
 *     def validate(value):
 *         if not isinstance(value, uuid.UUID):
 *             try:             # <<<<<<<<<<<<<<
 *                 value = uuid.UUID(value.strip())
 *             except Exception:
 */
      }
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L9_try_end;
      __pyx_L4_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "validr/_validator_c.pyx":991
 *             try:
 *                 value = uuid.UUID(value.strip())
 *             except Exception:             # <<<<<<<<<<<<<<
 *                 raise Invalid(msg) from None
 *         if version is not None and value.version != version:
 */
      __Pyx_TraceLine(991,0,__PYX_ERR(0, 991, __pyx_L6_except_error))
      __pyx_t_11 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
      if (__pyx_t_11) {
        __Pyx_AddTraceback("validr._validator_c.uuid_validator.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_8, &__pyx_t_1) < 0) __PYX_ERR(0, 991, __pyx_L6_except_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_GOTREF(__pyx_t_1);

        /* "validr/_validator_c.pyx":992
 *                 value = uuid.UUID(value.strip())
 *             except Exception:
 *                 raise Invalid(msg) from None             # <<<<<<<<<<<<<<
 *         if version is not None and value.version != version:
 *             raise Invalid(msg)
 */
        __Pyx_TraceLine(992,0,__PYX_ERR(0, 992, __pyx_L6_except_error))
        __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 992, __pyx_L6_except_error)
        __Pyx_GOTREF(__pyx_t_10);
        if (unlikely(!__pyx_cur_scope->__pyx_v_msg)) { __Pyx_RaiseClosureNameError("msg"); __PYX_ERR(0, 992, __pyx_L6_except_error) }
        __pyx_t_12 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
          __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_10);
          if (likely(__pyx_t_12)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
            __Pyx_INCREF(__pyx_t_12);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_10, function);
          }
        }
        __pyx_t_9 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_12, __pyx_cur_scope->__pyx_v_msg) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_cur_scope->__pyx_v_msg);
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 992, __pyx_L6_except_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_Raise(__pyx_t_9, 0, 0, Py_None);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __PYX_ERR(0, 992, __pyx_L6_except_error)
      }
      goto __pyx_L6_except_error;
      __pyx_L6_except_error:;

      /* "validr/_validator_c.pyx":989
 *     def validate(value):
 *         if not isinstance(value, uuid.UUID):
 *             try:             # <<<<<<<<<<<<<<
 *                 value = uuid.UUID(value.strip())
 *             except Exception:
 */
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
      goto __pyx_L1_error;
      __pyx_L9_try_end:;
    }

    /* "validr/_validator_c.pyx":988
 * 
 *     def validate(value):
 *         if not isinstance(value, uuid.UUID):             # <<<<<<<<<<<<<<
 *             try:
 *                 value = uuid.UUID(value.strip())
 */
  }

  /* "validr/_validator_c.pyx":993
 *             except Exception:
 *                 raise Invalid(msg) from None
 *         if version is not None and value.version != version:             # <<<<<<<<<<<<<<
 *             raise Invalid(msg)
 *         if output_object:
 */
  __Pyx_TraceLine(993,0,__PYX_ERR(0, 993, __pyx_L1_error))
  if (unlikely(!__pyx_cur_scope->__pyx_v_version)) { __Pyx_RaiseClosureNameError("version"); __PYX_ERR(0, 993, __pyx_L1_error) }
  __pyx_t_3 = (__pyx_cur_scope->__pyx_v_version != Py_None);
  __pyx_t_13 = (__pyx_t_3 != 0);
  if (__pyx_t_13) {
  } else {
    __pyx_t_4 = __pyx_t_13;
    goto __pyx_L13_bool_binop_done;
  }
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_version); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 993, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(!__pyx_cur_scope->__pyx_v_version)) { __Pyx_RaiseClosureNameError("version"); __PYX_ERR(0, 993, __pyx_L1_error) }
  __pyx_t_8 = PyObject_RichCompare(__pyx_t_1, __pyx_cur_scope->__pyx_v_version, Py_NE); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 993, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 993, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_4 = __pyx_t_13;
  __pyx_L13_bool_binop_done:;
  if (unlikely(__pyx_t_4)) {

    /* "validr/_validator_c.pyx":994
 *                 raise Invalid(msg) from None
 *         if version is not None and value.version != version:
 *             raise Invalid(msg)             # <<<<<<<<<<<<<<
 *         if output_object:
 *             return value
 */
    __Pyx_TraceLine(994,0,__PYX_ERR(0, 994, __pyx_L1_error))
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 994, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(!__pyx_cur_scope->__pyx_v_msg)) { __Pyx_RaiseClosureNameError("msg"); __PYX_ERR(0, 994, __pyx_L1_error) }
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_8 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, __pyx_cur_scope->__pyx_v_msg) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_cur_scope->__pyx_v_msg);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 994, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_8, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __PYX_ERR(0, 994, __pyx_L1_error)

    /* "validr/_validator_c.pyx":993
 *             except Exception:
 *                 raise Invalid(msg) from None
 *         if version is not None and value.version != version:             # <<<<<<<<<<<<<<
 *             raise Invalid(msg)
 *         if output_object:
 */
  }

  /* "validr/_validator_c.pyx":995
 *         if version is not None and value.version != version:
 *             raise Invalid(msg)
 *         if output_object:             # <<<<<<<<<<<<<<
 *             return value
 *         else:
 */
  __Pyx_TraceLine(995,0,__PYX_ERR(0, 995, __pyx_L1_error))
  __pyx_t_4 = (__pyx_cur_scope->__pyx_v_output_object != 0);
  if (__pyx_t_4) {

    /* "validr/_validator_c.pyx":996
 *             raise Invalid(msg)
 *         if output_object:
 *             return value             # <<<<<<<<<<<<<<
 *         else:
 *             return str(value)
 */
    __Pyx_TraceLine(996,0,__PYX_ERR(0, 996, __pyx_L1_error))
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_value);
    __pyx_r = __pyx_v_value;
    goto __pyx_L0;

    /* "validr/_validator_c.pyx":995
 *         if version is not None and value.version != version:
 *             raise Invalid(msg)
 *         if output_object:             # <<<<<<<<<<<<<<
 *             return value
 *         else:
 */
  }

  /* "validr/_validator_c.pyx":998
 *             return value
 *         else:
 *             return str(value)             # <<<<<<<<<<<<<<
 *     return validate
 * 
 */
  __Pyx_TraceLine(998,0,__PYX_ERR(0, 998, __pyx_L1_error))
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_8 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyUnicode_Type)), __pyx_v_value); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 998, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_r = __pyx_t_8;
    __pyx_t_8 = 0;
    goto __pyx_L0;
  }

  /* "validr/_validator_c.pyx":987
 *         msg = 'invalid uuid{}'.format(version)
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         if not isinstance(value, uuid.UUID):
 *             try:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("validr._validator_c.uuid_validator.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":979
 * 
 * @validator(output=(str, uuid.UUID))
 * def uuid_validator(compiler, version=None, bint output_object=False):             # <<<<<<<<<<<<<<
 *     if version is None:
 *         msg = 'invalid uuid'
 */

static PyObject *__pyx_pf_6validr_12_validator_c_74uuid_validator(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_compiler, PyObject *__pyx_v_version, int __pyx_v_output_object) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_26_uuid_validator *__pyx_cur_scope;
  PyObject *__pyx_v_validate = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_TraceFrameInit(__pyx_codeobj__109)
  __Pyx_RefNannySetupContext("uuid_validator", 0);
  __pyx_cur_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_26_uuid_validator *)__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_26_uuid_validator(__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_26_uuid_validator, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_26_uuid_validator *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 979, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __Pyx_TraceCall("uuid_validator", __pyx_f[0], 979, 0, __PYX_ERR(0, 979, __pyx_L1_error));
  __pyx_cur_scope->__pyx_v_version = __pyx_v_version;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_version);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_version);
  __pyx_cur_scope->__pyx_v_output_object = __pyx_v_output_object;

  /* "validr/_validator_c.pyx":980
 * @validator(output=(str, uuid.UUID))
 * def uuid_validator(compiler, version=None, bint output_object=False):
 *     if version is None:             # <<<<<<<<<<<<<<
 *         msg = 'invalid uuid'
 *     else:
 */
  __Pyx_TraceLine(980,0,__PYX_ERR(0, 980, __pyx_L1_error))
  __pyx_t_1 = (__pyx_cur_scope->__pyx_v_version == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "validr/_validator_c.pyx":981
 * def uuid_validator(compiler, version=None, bint output_object=False):
 *     if version is None:
 *         msg = 'invalid uuid'             # <<<<<<<<<<<<<<
 *     else:
 *         if version not in {1, 3, 4, 5}:
 */
    __Pyx_TraceLine(981,0,__PYX_ERR(0, 981, __pyx_L1_error))
    __Pyx_INCREF(__pyx_kp_u_invalid_uuid);
    __Pyx_GIVEREF(__pyx_kp_u_invalid_uuid);
    __pyx_cur_scope->__pyx_v_msg = __pyx_kp_u_invalid_uuid;

    /* "validr/_validator_c.pyx":980
 * @validator(output=(str, uuid.UUID))
 * def uuid_validator(compiler, version=None, bint output_object=False):
 *     if version is None:             # <<<<<<<<<<<<<<
 *         msg = 'invalid uuid'
 *     else:
 */
    goto __pyx_L3;
  }

  /* "validr/_validator_c.pyx":983
 *         msg = 'invalid uuid'
 *     else:
 *         if version not in {1, 3, 4, 5}:             # <<<<<<<<<<<<<<
 *             raise SchemaError('illegal version number')
 *         msg = 'invalid uuid{}'.format(version)
 */
  __Pyx_TraceLine(983,0,__PYX_ERR(0, 983, __pyx_L1_error))
  /*else*/ {
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_version);
    __pyx_t_3 = __pyx_cur_scope->__pyx_v_version;
    __pyx_t_4 = __Pyx_PyInt_NeObjC(__pyx_t_3, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 983, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 983, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_1) {
    } else {
      __pyx_t_2 = __pyx_t_1;
      goto __pyx_L5_bool_binop_done;
    }
    __pyx_t_4 = __Pyx_PyInt_NeObjC(__pyx_t_3, __pyx_int_3, 3, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 983, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 983, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_1) {
    } else {
      __pyx_t_2 = __pyx_t_1;
      goto __pyx_L5_bool_binop_done;
    }
    __pyx_t_4 = __Pyx_PyInt_NeObjC(__pyx_t_3, __pyx_int_4, 4, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 983, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 983, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_1) {
    } else {
      __pyx_t_2 = __pyx_t_1;
      goto __pyx_L5_bool_binop_done;
    }
    __pyx_t_4 = __Pyx_PyInt_NeObjC(__pyx_t_3, __pyx_int_5, 5, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 983, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 983, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_2 = __pyx_t_1;
    __pyx_L5_bool_binop_done:;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_1 = (__pyx_t_2 != 0);
    if (unlikely(__pyx_t_1)) {

      /* "validr/_validator_c.pyx":984
 *     else:
 *         if version not in {1, 3, 4, 5}:
 *             raise SchemaError('illegal version number')             # <<<<<<<<<<<<<<
 *         msg = 'invalid uuid{}'.format(version)
 * 
 */
      __Pyx_TraceLine(984,0,__PYX_ERR(0, 984, __pyx_L1_error))
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_SchemaError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 984, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_kp_u_illegal_version_number) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_kp_u_illegal_version_number);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 984, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 984, __pyx_L1_error)

      /* "validr/_validator_c.pyx":983
 *         msg = 'invalid uuid'
 *     else:
 *         if version not in {1, 3, 4, 5}:             # <<<<<<<<<<<<<<
 *             raise SchemaError('illegal version number')
 *         msg = 'invalid uuid{}'.format(version)
 */
    }

    /* "validr/_validator_c.pyx":985
 *         if version not in {1, 3, 4, 5}:
 *             raise SchemaError('illegal version number')
 *         msg = 'invalid uuid{}'.format(version)             # <<<<<<<<<<<<<<
 * 
 *     def validate(value):
 */
    __Pyx_TraceLine(985,0,__PYX_ERR(0, 985, __pyx_L1_error))
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_invalid_uuid_2, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 985, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_cur_scope->__pyx_v_version) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_cur_scope->__pyx_v_version);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 985, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_cur_scope->__pyx_v_msg = __pyx_t_3;
    __pyx_t_3 = 0;
  }
  __pyx_L3:;

  /* "validr/_validator_c.pyx":987
 *         msg = 'invalid uuid{}'.format(version)
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         if not isinstance(value, uuid.UUID):
 *             try:
 */
  __Pyx_TraceLine(987,0,__PYX_ERR(0, 987, __pyx_L1_error))
  __pyx_t_3 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_14uuid_validator_1validate, 0, __pyx_n_s_uuid_validator_locals_validate, ((PyObject*)__pyx_cur_scope), __pyx_n_s_validr__validator_c, __pyx_d, ((PyObject *)__pyx_codeobj__111)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 987, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_validate = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "validr/_validator_c.pyx":999
 *         else:
 *             return str(value)
 *     return validate             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(999,0,__PYX_ERR(0, 999, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_validate);
  __pyx_r = __pyx_v_validate;
  goto __pyx_L0;

  /* "validr/_validator_c.pyx":979
 * 
 * @validator(output=(str, uuid.UUID))
 * def uuid_validator(compiler, version=None, bint output_object=False):             # <<<<<<<<<<<<<<
 *     if version is None:
 *         msg = 'invalid uuid'
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("validr._validator_c.uuid_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_validate);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":1002
 * 
 * 
 * def create_re_validator(str name, str r, int default_maxlen=255):             # <<<<<<<<<<<<<<
 *     """Create validator by regex string
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_77create_re_validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6validr_12_validator_c_76create_re_validator[] = "Create validator by regex string\n\n    It will make sure that the entire string matches, so needn't\n    add `^`,`$` to regex string.\n\n    Args:\n        name (str): validator name, used in error message\n        r (str): regex string\n    ";
static PyMethodDef __pyx_mdef_6validr_12_validator_c_77create_re_validator = {"create_re_validator", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6validr_12_validator_c_77create_re_validator, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6validr_12_validator_c_76create_re_validator};
static PyObject *__pyx_pw_6validr_12_validator_c_77create_re_validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_r = 0;
  int __pyx_v_default_maxlen;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create_re_validator (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_name_2,&__pyx_n_s_r,&__pyx_n_s_default_maxlen,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name_2)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_r)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("create_re_validator", 0, 2, 3, 1); __PYX_ERR(0, 1002, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_default_maxlen);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "create_re_validator") < 0)) __PYX_ERR(0, 1002, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_name = ((PyObject*)values[0]);
    __pyx_v_r = ((PyObject*)values[1]);
    if (values[2]) {
      __pyx_v_default_maxlen = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_default_maxlen == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1002, __pyx_L3_error)
    } else {
      __pyx_v_default_maxlen = ((int)0xFF);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("create_re_validator", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1002, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("validr._validator_c.create_re_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_name), (&PyUnicode_Type), 1, "name", 1))) __PYX_ERR(0, 1002, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_r), (&PyUnicode_Type), 1, "r", 1))) __PYX_ERR(0, 1002, __pyx_L1_error)
  __pyx_r = __pyx_pf_6validr_12_validator_c_76create_re_validator(__pyx_self, __pyx_v_name, __pyx_v_r, __pyx_v_default_maxlen);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":1016
 *     message = 'invalid %s' % name
 * 
 *     def re_validator(compiler, int minlen=0, int maxlen=default_maxlen, bint strip=False):             # <<<<<<<<<<<<<<
 *         def validate(value):
 *             if not isinstance(value, str):
 */

static PyObject *__pyx_pf_6validr_12_validator_c_80__defaults__(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__defaults__", 0);
  __Pyx_TraceCall("__defaults__", __pyx_f[0], 1016, 0, __PYX_ERR(0, 1016, __pyx_L1_error));
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(((int)0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1016, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_int(__Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_maxlen); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1016, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyBool_FromLong(((int)0)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1016, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1016, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1016, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_3, 1, Py_None);
  __pyx_t_4 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("validr._validator_c.__defaults__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_19create_re_validator_1re_validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6validr_12_validator_c_19create_re_validator_1re_validator = {"re_validator", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6validr_12_validator_c_19create_re_validator_1re_validator, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6validr_12_validator_c_19create_re_validator_1re_validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_compiler = 0;
  int __pyx_v_minlen;
  int __pyx_v_maxlen;
  int __pyx_v_strip;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("re_validator (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_compiler,&__pyx_n_s_minlen,&__pyx_n_s_maxlen,&__pyx_n_s_strip,0};
    PyObject* values[4] = {0,0,0,0};
    __pyx_defaults *__pyx_dynamic_args = __Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_compiler)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_minlen);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_maxlen);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_strip);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "re_validator") < 0)) __PYX_ERR(0, 1016, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_compiler = values[0];
    if (values[1]) {
      __pyx_v_minlen = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_minlen == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1016, __pyx_L3_error)
    } else {
      __pyx_v_minlen = ((int)((int)0));
    }
    if (values[2]) {
      __pyx_v_maxlen = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_maxlen == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1016, __pyx_L3_error)
    } else {
      __pyx_v_maxlen = __pyx_dynamic_args->__pyx_arg_maxlen;
    }
    if (values[3]) {
      __pyx_v_strip = __Pyx_PyObject_IsTrue(values[3]); if (unlikely((__pyx_v_strip == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1016, __pyx_L3_error)
    } else {
      __pyx_v_strip = ((int)((int)0));
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("re_validator", 0, 1, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1016, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("validr._validator_c.create_re_validator.re_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6validr_12_validator_c_19create_re_validator_re_validator(__pyx_self, __pyx_v_compiler, __pyx_v_minlen, __pyx_v_maxlen, __pyx_v_strip);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":1017
 * 
 *     def re_validator(compiler, int minlen=0, int maxlen=default_maxlen, bint strip=False):
 *         def validate(value):             # <<<<<<<<<<<<<<
 *             if not isinstance(value, str):
 *                 raise Invalid('value must be string')
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_19create_re_validator_12re_validator_1validate(PyObject *__pyx_self, PyObject *__pyx_v_value); /*proto*/
static PyMethodDef __pyx_mdef_6validr_12_validator_c_19create_re_validator_12re_validator_1validate = {"validate", (PyCFunction)__pyx_pw_6validr_12_validator_c_19create_re_validator_12re_validator_1validate, METH_O, 0};
static PyObject *__pyx_pw_6validr_12_validator_c_19create_re_validator_12re_validator_1validate(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("validate (wrapper)", 0);
  __pyx_r = __pyx_pf_6validr_12_validator_c_19create_re_validator_12re_validator_validate(__pyx_self, ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6validr_12_validator_c_19create_re_validator_12re_validator_validate(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_28_re_validator *__pyx_cur_scope;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_28_re_validator *__pyx_outer_scope;
  int __pyx_v_length;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  Py_UCS4 __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("validate", 0);
  __pyx_outer_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_28_re_validator *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_TraceCall("validate", __pyx_f[0], 1017, 0, __PYX_ERR(0, 1017, __pyx_L1_error));
  __Pyx_INCREF(__pyx_v_value);

  /* "validr/_validator_c.pyx":1018
 *     def re_validator(compiler, int minlen=0, int maxlen=default_maxlen, bint strip=False):
 *         def validate(value):
 *             if not isinstance(value, str):             # <<<<<<<<<<<<<<
 *                 raise Invalid('value must be string')
 *             if strip:
 */
  __Pyx_TraceLine(1018,0,__PYX_ERR(0, 1018, __pyx_L1_error))
  __pyx_t_1 = PyUnicode_Check(__pyx_v_value); 
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "validr/_validator_c.pyx":1019
 *         def validate(value):
 *             if not isinstance(value, str):
 *                 raise Invalid('value must be string')             # <<<<<<<<<<<<<<
 *             if strip:
 *                 value = value.strip()
 */
    __Pyx_TraceLine(1019,0,__PYX_ERR(0, 1019, __pyx_L1_error))
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1019, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_kp_u_value_must_be_string) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_kp_u_value_must_be_string);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1019, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 1019, __pyx_L1_error)

    /* "validr/_validator_c.pyx":1018
 *     def re_validator(compiler, int minlen=0, int maxlen=default_maxlen, bint strip=False):
 *         def validate(value):
 *             if not isinstance(value, str):             # <<<<<<<<<<<<<<
 *                 raise Invalid('value must be string')
 *             if strip:
 */
  }

  /* "validr/_validator_c.pyx":1020
 *             if not isinstance(value, str):
 *                 raise Invalid('value must be string')
 *             if strip:             # <<<<<<<<<<<<<<
 *                 value = value.strip()
 *             cdef int length = len(value)
 */
  __Pyx_TraceLine(1020,0,__PYX_ERR(0, 1020, __pyx_L1_error))
  __pyx_t_2 = (__pyx_cur_scope->__pyx_v_strip != 0);
  if (__pyx_t_2) {

    /* "validr/_validator_c.pyx":1021
 *                 raise Invalid('value must be string')
 *             if strip:
 *                 value = value.strip()             # <<<<<<<<<<<<<<
 *             cdef int length = len(value)
 *             if length < minlen:
 */
    __Pyx_TraceLine(1021,0,__PYX_ERR(0, 1021, __pyx_L1_error))
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_strip); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1021, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1021, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "validr/_validator_c.pyx":1020
 *             if not isinstance(value, str):
 *                 raise Invalid('value must be string')
 *             if strip:             # <<<<<<<<<<<<<<
 *                 value = value.strip()
 *             cdef int length = len(value)
 */
  }

  /* "validr/_validator_c.pyx":1022
 *             if strip:
 *                 value = value.strip()
 *             cdef int length = len(value)             # <<<<<<<<<<<<<<
 *             if length < minlen:
 *                 raise Invalid('%s length must >= %d' % (name, minlen))
 */
  __Pyx_TraceLine(1022,0,__PYX_ERR(0, 1022, __pyx_L1_error))
  __pyx_t_6 = PyObject_Length(__pyx_v_value); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1022, __pyx_L1_error)
  __pyx_v_length = __pyx_t_6;

  /* "validr/_validator_c.pyx":1023
 *                 value = value.strip()
 *             cdef int length = len(value)
 *             if length < minlen:             # <<<<<<<<<<<<<<
 *                 raise Invalid('%s length must >= %d' % (name, minlen))
 *             elif length > maxlen:
 */
  __Pyx_TraceLine(1023,0,__PYX_ERR(0, 1023, __pyx_L1_error))
  __pyx_t_2 = ((__pyx_v_length < __pyx_cur_scope->__pyx_v_minlen) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "validr/_validator_c.pyx":1024
 *             cdef int length = len(value)
 *             if length < minlen:
 *                 raise Invalid('%s length must >= %d' % (name, minlen))             # <<<<<<<<<<<<<<
 *             elif length > maxlen:
 *                 raise Invalid('%s length must <= %d' % (name, maxlen))
 */
    __Pyx_TraceLine(1024,0,__PYX_ERR(0, 1024, __pyx_L1_error))
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1024, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1024, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = 0;
    __pyx_t_7 = 127;
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_name)) { __Pyx_RaiseClosureNameError("name"); __PYX_ERR(0, 1024, __pyx_L1_error) }
    __pyx_t_8 = __Pyx_PyUnicode_Unicode(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_name); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1024, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) > __pyx_t_7) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) : __pyx_t_7;
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_8);
    __pyx_t_8 = 0;
    __Pyx_INCREF(__pyx_kp_u_length_must);
    __pyx_t_6 += 16;
    __Pyx_GIVEREF(__pyx_kp_u_length_must);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_kp_u_length_must);
    __pyx_t_8 = __Pyx_PyUnicode_From_int(__pyx_cur_scope->__pyx_v_minlen, 0, ' ', 'd'); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1024, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_8);
    __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_PyUnicode_Join(__pyx_t_5, 3, __pyx_t_6, __pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1024, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_8);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1024, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 1024, __pyx_L1_error)

    /* "validr/_validator_c.pyx":1023
 *                 value = value.strip()
 *             cdef int length = len(value)
 *             if length < minlen:             # <<<<<<<<<<<<<<
 *                 raise Invalid('%s length must >= %d' % (name, minlen))
 *             elif length > maxlen:
 */
  }

  /* "validr/_validator_c.pyx":1025
 *             if length < minlen:
 *                 raise Invalid('%s length must >= %d' % (name, minlen))
 *             elif length > maxlen:             # <<<<<<<<<<<<<<
 *                 raise Invalid('%s length must <= %d' % (name, maxlen))
 *             if match(value):
 */
  __Pyx_TraceLine(1025,0,__PYX_ERR(0, 1025, __pyx_L1_error))
  __pyx_t_2 = ((__pyx_v_length > __pyx_cur_scope->__pyx_v_maxlen) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "validr/_validator_c.pyx":1026
 *                 raise Invalid('%s length must >= %d' % (name, minlen))
 *             elif length > maxlen:
 *                 raise Invalid('%s length must <= %d' % (name, maxlen))             # <<<<<<<<<<<<<<
 *             if match(value):
 *                 return value
 */
    __Pyx_TraceLine(1026,0,__PYX_ERR(0, 1026, __pyx_L1_error))
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1026, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_8 = PyTuple_New(3); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1026, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_6 = 0;
    __pyx_t_7 = 127;
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_name)) { __Pyx_RaiseClosureNameError("name"); __PYX_ERR(0, 1026, __pyx_L1_error) }
    __pyx_t_5 = __Pyx_PyUnicode_Unicode(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1026, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_7) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_7;
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_INCREF(__pyx_kp_u_length_must_2);
    __pyx_t_6 += 16;
    __Pyx_GIVEREF(__pyx_kp_u_length_must_2);
    PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_kp_u_length_must_2);
    __pyx_t_5 = __Pyx_PyUnicode_From_int(__pyx_cur_scope->__pyx_v_maxlen, 0, ' ', 'd'); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1026, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_t_5);
    __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyUnicode_Join(__pyx_t_8, 3, __pyx_t_6, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1026, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_8, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1026, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 1026, __pyx_L1_error)

    /* "validr/_validator_c.pyx":1025
 *             if length < minlen:
 *                 raise Invalid('%s length must >= %d' % (name, minlen))
 *             elif length > maxlen:             # <<<<<<<<<<<<<<
 *                 raise Invalid('%s length must <= %d' % (name, maxlen))
 *             if match(value):
 */
  }

  /* "validr/_validator_c.pyx":1027
 *             elif length > maxlen:
 *                 raise Invalid('%s length must <= %d' % (name, maxlen))
 *             if match(value):             # <<<<<<<<<<<<<<
 *                 return value
 *             else:
 */
  __Pyx_TraceLine(1027,0,__PYX_ERR(0, 1027, __pyx_L1_error))
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_match)) { __Pyx_RaiseClosureNameError("match"); __PYX_ERR(0, 1027, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_match);
  __pyx_t_4 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_match; __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_value) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_value);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1027, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1027, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (likely(__pyx_t_2)) {

    /* "validr/_validator_c.pyx":1028
 *                 raise Invalid('%s length must <= %d' % (name, maxlen))
 *             if match(value):
 *                 return value             # <<<<<<<<<<<<<<
 *             else:
 *                 raise Invalid(message)
 */
    __Pyx_TraceLine(1028,0,__PYX_ERR(0, 1028, __pyx_L1_error))
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_value);
    __pyx_r = __pyx_v_value;
    goto __pyx_L0;

    /* "validr/_validator_c.pyx":1027
 *             elif length > maxlen:
 *                 raise Invalid('%s length must <= %d' % (name, maxlen))
 *             if match(value):             # <<<<<<<<<<<<<<
 *                 return value
 *             else:
 */
  }

  /* "validr/_validator_c.pyx":1030
 *                 return value
 *             else:
 *                 raise Invalid(message)             # <<<<<<<<<<<<<<
 *         return validate
 *     re_validator.__name__ = name + '_validator'
 */
  __Pyx_TraceLine(1030,0,__PYX_ERR(0, 1030, __pyx_L1_error))
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1030, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_message)) { __Pyx_RaiseClosureNameError("message"); __PYX_ERR(0, 1030, __pyx_L1_error) }
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_message) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_message);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1030, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 1030, __pyx_L1_error)
  }

  /* "validr/_validator_c.pyx":1017
 * 
 *     def re_validator(compiler, int minlen=0, int maxlen=default_maxlen, bint strip=False):
 *         def validate(value):             # <<<<<<<<<<<<<<
 *             if not isinstance(value, str):
 *                 raise Invalid('value must be string')
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("validr._validator_c.create_re_validator.re_validator.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":1016
 *     message = 'invalid %s' % name
 * 
 *     def re_validator(compiler, int minlen=0, int maxlen=default_maxlen, bint strip=False):             # <<<<<<<<<<<<<<
 *         def validate(value):
 *             if not isinstance(value, str):
 */

static PyObject *__pyx_pf_6validr_12_validator_c_19create_re_validator_re_validator(PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_compiler, int __pyx_v_minlen, int __pyx_v_maxlen, int __pyx_v_strip) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_28_re_validator *__pyx_cur_scope;
  PyObject *__pyx_v_validate = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("re_validator", 0);
  __pyx_cur_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_28_re_validator *)__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_28_re_validator(__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_28_re_validator, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_28_re_validator *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1016, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_27_create_re_validator *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  __Pyx_TraceCall("re_validator", __pyx_f[0], 1016, 0, __PYX_ERR(0, 1016, __pyx_L1_error));
  __pyx_cur_scope->__pyx_v_minlen = __pyx_v_minlen;
  __pyx_cur_scope->__pyx_v_maxlen = __pyx_v_maxlen;
  __pyx_cur_scope->__pyx_v_strip = __pyx_v_strip;

  /* "validr/_validator_c.pyx":1017
 * 
 *     def re_validator(compiler, int minlen=0, int maxlen=default_maxlen, bint strip=False):
 *         def validate(value):             # <<<<<<<<<<<<<<
 *             if not isinstance(value, str):
 *                 raise Invalid('value must be string')
 */
  __Pyx_TraceLine(1017,0,__PYX_ERR(0, 1017, __pyx_L1_error))
  __pyx_t_1 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_19create_re_validator_12re_validator_1validate, 0, __pyx_n_s_create_re_validator_locals_re_va, ((PyObject*)__pyx_cur_scope), __pyx_n_s_validr__validator_c, __pyx_d, ((PyObject *)__pyx_codeobj__113)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1017, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_validate = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":1031
 *             else:
 *                 raise Invalid(message)
 *         return validate             # <<<<<<<<<<<<<<
 *     re_validator.__name__ = name + '_validator'
 *     re_validator.__qualname__ = name + '_validator'
 */
  __Pyx_TraceLine(1031,0,__PYX_ERR(0, 1031, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_validate);
  __pyx_r = __pyx_v_validate;
  goto __pyx_L0;

  /* "validr/_validator_c.pyx":1016
 *     message = 'invalid %s' % name
 * 
 *     def re_validator(compiler, int minlen=0, int maxlen=default_maxlen, bint strip=False):             # <<<<<<<<<<<<<<
 *         def validate(value):
 *             if not isinstance(value, str):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("validr._validator_c.create_re_validator.re_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_validate);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":1002
 * 
 * 
 * def create_re_validator(str name, str r, int default_maxlen=255):             # <<<<<<<<<<<<<<
 *     """Create validator by regex string
 * 
 */

static PyObject *__pyx_pf_6validr_12_validator_c_76create_re_validator(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_name, PyObject *__pyx_v_r, int __pyx_v_default_maxlen) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_27_create_re_validator *__pyx_cur_scope;
  PyObject *__pyx_v_re_validator = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_TraceFrameInit(__pyx_codeobj__114)
  __Pyx_RefNannySetupContext("create_re_validator", 0);
  __pyx_cur_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_27_create_re_validator *)__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_27_create_re_validator(__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_27_create_re_validator, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_27_create_re_validator *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1002, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __Pyx_TraceCall("create_re_validator", __pyx_f[0], 1002, 0, __PYX_ERR(0, 1002, __pyx_L1_error));
  __pyx_cur_scope->__pyx_v_name = __pyx_v_name;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_name);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_name);

  /* "validr/_validator_c.pyx":1013
 *     """
 *     # To make sure that the entire string matches
 *     match = re.compile(r'(?:%s)\Z' % r).match             # <<<<<<<<<<<<<<
 *     message = 'invalid %s' % name
 * 
 */
  __Pyx_TraceLine(1013,0,__PYX_ERR(0, 1013, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_re); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1013, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_compile); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1013, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyUnicode_Format(__pyx_kp_u_s_Z, __pyx_v_r); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1013, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1013, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1013, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_cur_scope->__pyx_v_match = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "validr/_validator_c.pyx":1014
 *     # To make sure that the entire string matches
 *     match = re.compile(r'(?:%s)\Z' % r).match
 *     message = 'invalid %s' % name             # <<<<<<<<<<<<<<
 * 
 *     def re_validator(compiler, int minlen=0, int maxlen=default_maxlen, bint strip=False):
 */
  __Pyx_TraceLine(1014,0,__PYX_ERR(0, 1014, __pyx_L1_error))
  __pyx_t_3 = PyUnicode_Format(__pyx_kp_u_invalid_s, __pyx_cur_scope->__pyx_v_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_cur_scope->__pyx_v_message = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "validr/_validator_c.pyx":1016
 *     message = 'invalid %s' % name
 * 
 *     def re_validator(compiler, int minlen=0, int maxlen=default_maxlen, bint strip=False):             # <<<<<<<<<<<<<<
 *         def validate(value):
 *             if not isinstance(value, str):
 */
  __Pyx_TraceLine(1016,0,__PYX_ERR(0, 1016, __pyx_L1_error))
  __pyx_t_3 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_19create_re_validator_1re_validator, 0, __pyx_n_s_create_re_validator_locals_re_va_2, ((PyObject*)__pyx_cur_scope), __pyx_n_s_validr__validator_c, __pyx_d, ((PyObject *)__pyx_codeobj__116)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1016, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (!__Pyx_CyFunction_InitDefaults(__pyx_t_3, sizeof(__pyx_defaults), 0)) __PYX_ERR(0, 1016, __pyx_L1_error)
  __Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_t_3)->__pyx_arg_maxlen = __pyx_v_default_maxlen;
  __Pyx_CyFunction_SetDefaultsGetter(__pyx_t_3, __pyx_pf_6validr_12_validator_c_80__defaults__);
  __pyx_v_re_validator = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "validr/_validator_c.pyx":1032
 *                 raise Invalid(message)
 *         return validate
 *     re_validator.__name__ = name + '_validator'             # <<<<<<<<<<<<<<
 *     re_validator.__qualname__ = name + '_validator'
 *     return validator(accept=str, output=str)(re_validator)
 */
  __Pyx_TraceLine(1032,0,__PYX_ERR(0, 1032, __pyx_L1_error))
  __pyx_t_3 = __Pyx_PyUnicode_ConcatSafe(__pyx_cur_scope->__pyx_v_name, __pyx_n_u_validator_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1032, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_re_validator, __pyx_n_s_name, __pyx_t_3) < 0) __PYX_ERR(0, 1032, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "validr/_validator_c.pyx":1033
 *         return validate
 *     re_validator.__name__ = name + '_validator'
 *     re_validator.__qualname__ = name + '_validator'             # <<<<<<<<<<<<<<
 *     return validator(accept=str, output=str)(re_validator)
 * 
 */
  __Pyx_TraceLine(1033,0,__PYX_ERR(0, 1033, __pyx_L1_error))
  __pyx_t_3 = __Pyx_PyUnicode_ConcatSafe(__pyx_cur_scope->__pyx_v_name, __pyx_n_u_validator_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1033, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_re_validator, __pyx_n_s_qualname, __pyx_t_3) < 0) __PYX_ERR(0, 1033, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "validr/_validator_c.pyx":1034
 *     re_validator.__name__ = name + '_validator'
 *     re_validator.__qualname__ = name + '_validator'
 *     return validator(accept=str, output=str)(re_validator)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(1034,0,__PYX_ERR(0, 1034, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_validator); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_accept, ((PyObject *)(&PyUnicode_Type))) < 0) __PYX_ERR(0, 1034, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_output, ((PyObject *)(&PyUnicode_Type))) < 0) __PYX_ERR(0, 1034, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_2, __pyx_v_re_validator) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_re_validator);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "validr/_validator_c.pyx":1002
 * 
 * 
 * def create_re_validator(str name, str r, int default_maxlen=255):             # <<<<<<<<<<<<<<
 *     """Create validator by regex string
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("validr._validator_c.create_re_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_re_validator);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":1070
 * 
 * 
 * def create_enum_validator(str name, items, bint string=True):             # <<<<<<<<<<<<<<
 *     """Create validator by enum items
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_79create_enum_validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6validr_12_validator_c_78create_enum_validator[] = "Create validator by enum items\n\n    Args:\n        name (str): validator name, used in error message\n        items (iterable): enum items\n        string (bool): is string like or not\n\n    Deprecated since v1.2.0, use enum validator instead.\n    ";
static PyMethodDef __pyx_mdef_6validr_12_validator_c_79create_enum_validator = {"create_enum_validator", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6validr_12_validator_c_79create_enum_validator, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6validr_12_validator_c_78create_enum_validator};
static PyObject *__pyx_pw_6validr_12_validator_c_79create_enum_validator(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_items = 0;
  int __pyx_v_string;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create_enum_validator (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_name_2,&__pyx_n_s_items,&__pyx_n_s_string,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name_2)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_items)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("create_enum_validator", 0, 2, 3, 1); __PYX_ERR(0, 1070, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_string);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "create_enum_validator") < 0)) __PYX_ERR(0, 1070, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_name = ((PyObject*)values[0]);
    __pyx_v_items = values[1];
    if (values[2]) {
      __pyx_v_string = __Pyx_PyObject_IsTrue(values[2]); if (unlikely((__pyx_v_string == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1070, __pyx_L3_error)
    } else {
      __pyx_v_string = ((int)1);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("create_enum_validator", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1070, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("validr._validator_c.create_enum_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_name), (&PyUnicode_Type), 1, "name", 1))) __PYX_ERR(0, 1070, __pyx_L1_error)
  __pyx_r = __pyx_pf_6validr_12_validator_c_78create_enum_validator(__pyx_self, __pyx_v_name, __pyx_v_items, __pyx_v_string);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":1083
 *     message = 'invalid {}, expect one of {}'.format(name, list(sorted(items)))
 * 
 *     def enum_validator(compiler):             # <<<<<<<<<<<<<<
 *         def validate(value):
 *             if value in items:
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_21create_enum_validator_1enum_validator(PyObject *__pyx_self, PyObject *__pyx_v_compiler); /*proto*/
static PyMethodDef __pyx_mdef_6validr_12_validator_c_21create_enum_validator_1enum_validator = {"enum_validator", (PyCFunction)__pyx_pw_6validr_12_validator_c_21create_enum_validator_1enum_validator, METH_O, 0};
static PyObject *__pyx_pw_6validr_12_validator_c_21create_enum_validator_1enum_validator(PyObject *__pyx_self, PyObject *__pyx_v_compiler) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enum_validator (wrapper)", 0);
  __pyx_r = __pyx_pf_6validr_12_validator_c_21create_enum_validator_enum_validator(__pyx_self, ((PyObject *)__pyx_v_compiler));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":1084
 * 
 *     def enum_validator(compiler):
 *         def validate(value):             # <<<<<<<<<<<<<<
 *             if value in items:
 *                 return value
 */

/* Python wrapper */
static PyObject *__pyx_pw_6validr_12_validator_c_21create_enum_validator_14enum_validator_1validate(PyObject *__pyx_self, PyObject *__pyx_v_value); /*proto*/
static PyMethodDef __pyx_mdef_6validr_12_validator_c_21create_enum_validator_14enum_validator_1validate = {"validate", (PyCFunction)__pyx_pw_6validr_12_validator_c_21create_enum_validator_14enum_validator_1validate, METH_O, 0};
static PyObject *__pyx_pw_6validr_12_validator_c_21create_enum_validator_14enum_validator_1validate(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("validate (wrapper)", 0);
  __pyx_r = __pyx_pf_6validr_12_validator_c_21create_enum_validator_14enum_validator_validate(__pyx_self, ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6validr_12_validator_c_21create_enum_validator_14enum_validator_validate(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_29_create_enum_validator *__pyx_cur_scope;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_29_create_enum_validator *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("validate", 0);
  __pyx_outer_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_29_create_enum_validator *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_TraceCall("validate", __pyx_f[0], 1084, 0, __PYX_ERR(0, 1084, __pyx_L1_error));

  /* "validr/_validator_c.pyx":1085
 *     def enum_validator(compiler):
 *         def validate(value):
 *             if value in items:             # <<<<<<<<<<<<<<
 *                 return value
 *             raise Invalid(message)
 */
  __Pyx_TraceLine(1085,0,__PYX_ERR(0, 1085, __pyx_L1_error))
  if (unlikely(!__pyx_cur_scope->__pyx_v_items)) { __Pyx_RaiseClosureNameError("items"); __PYX_ERR(0, 1085, __pyx_L1_error) }
  __pyx_t_1 = (__Pyx_PySequence_ContainsTF(__pyx_v_value, __pyx_cur_scope->__pyx_v_items, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1085, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "validr/_validator_c.pyx":1086
 *         def validate(value):
 *             if value in items:
 *                 return value             # <<<<<<<<<<<<<<
 *             raise Invalid(message)
 *         return validate
 */
    __Pyx_TraceLine(1086,0,__PYX_ERR(0, 1086, __pyx_L1_error))
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_value);
    __pyx_r = __pyx_v_value;
    goto __pyx_L0;

    /* "validr/_validator_c.pyx":1085
 *     def enum_validator(compiler):
 *         def validate(value):
 *             if value in items:             # <<<<<<<<<<<<<<
 *                 return value
 *             raise Invalid(message)
 */
  }

  /* "validr/_validator_c.pyx":1087
 *             if value in items:
 *                 return value
 *             raise Invalid(message)             # <<<<<<<<<<<<<<
 *         return validate
 *     enum_validator.__name__ = name + '_validator'
 */
  __Pyx_TraceLine(1087,0,__PYX_ERR(0, 1087, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1087, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (unlikely(!__pyx_cur_scope->__pyx_v_message)) { __Pyx_RaiseClosureNameError("message"); __PYX_ERR(0, 1087, __pyx_L1_error) }
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_cur_scope->__pyx_v_message) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_cur_scope->__pyx_v_message);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1087, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_Raise(__pyx_t_3, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __PYX_ERR(0, 1087, __pyx_L1_error)

  /* "validr/_validator_c.pyx":1084
 * 
 *     def enum_validator(compiler):
 *         def validate(value):             # <<<<<<<<<<<<<<
 *             if value in items:
 *                 return value
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("validr._validator_c.create_enum_validator.enum_validator.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":1083
 *     message = 'invalid {}, expect one of {}'.format(name, list(sorted(items)))
 * 
 *     def enum_validator(compiler):             # <<<<<<<<<<<<<<
 *         def validate(value):
 *             if value in items:
 */

static PyObject *__pyx_pf_6validr_12_validator_c_21create_enum_validator_enum_validator(PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_compiler) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_29_create_enum_validator *__pyx_cur_scope;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_29_create_enum_validator *__pyx_outer_scope;
  PyObject *__pyx_v_validate = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("enum_validator", 0);
  __pyx_outer_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_29_create_enum_validator *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_TraceCall("enum_validator", __pyx_f[0], 1083, 0, __PYX_ERR(0, 1083, __pyx_L1_error));

  /* "validr/_validator_c.pyx":1084
 * 
 *     def enum_validator(compiler):
 *         def validate(value):             # <<<<<<<<<<<<<<
 *             if value in items:
 *                 return value
 */
  __Pyx_TraceLine(1084,0,__PYX_ERR(0, 1084, __pyx_L1_error))
  __pyx_t_1 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_21create_enum_validator_14enum_validator_1validate, 0, __pyx_n_s_create_enum_validator_locals_enu, ((PyObject*)__pyx_cur_scope), __pyx_n_s_validr__validator_c, __pyx_d, ((PyObject *)__pyx_codeobj__118)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1084, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_validate = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":1088
 *                 return value
 *             raise Invalid(message)
 *         return validate             # <<<<<<<<<<<<<<
 *     enum_validator.__name__ = name + '_validator'
 *     enum_validator.__qualname__ = name + '_validator'
 */
  __Pyx_TraceLine(1088,0,__PYX_ERR(0, 1088, __pyx_L1_error))
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_validate);
  __pyx_r = __pyx_v_validate;
  goto __pyx_L0;

  /* "validr/_validator_c.pyx":1083
 *     message = 'invalid {}, expect one of {}'.format(name, list(sorted(items)))
 * 
 *     def enum_validator(compiler):             # <<<<<<<<<<<<<<
 *         def validate(value):
 *             if value in items:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("validr._validator_c.create_enum_validator.enum_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_validate);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "validr/_validator_c.pyx":1070
 * 
 * 
 * def create_enum_validator(str name, items, bint string=True):             # <<<<<<<<<<<<<<
 *     """Create validator by enum items
 * 
 */

static PyObject *__pyx_pf_6validr_12_validator_c_78create_enum_validator(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_name, PyObject *__pyx_v_items, int __pyx_v_string) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_29_create_enum_validator *__pyx_cur_scope;
  PyObject *__pyx_v_enum_validator = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  __Pyx_TraceFrameInit(__pyx_codeobj__119)
  __Pyx_RefNannySetupContext("create_enum_validator", 0);
  __pyx_cur_scope = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_29_create_enum_validator *)__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_29_create_enum_validator(__pyx_ptype_6validr_12_validator_c___pyx_scope_struct_29_create_enum_validator, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_29_create_enum_validator *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1070, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __Pyx_TraceCall("create_enum_validator", __pyx_f[0], 1070, 0, __PYX_ERR(0, 1070, __pyx_L1_error));
  __pyx_cur_scope->__pyx_v_items = __pyx_v_items;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_items);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_items);

  /* "validr/_validator_c.pyx":1080
 *     Deprecated since v1.2.0, use enum validator instead.
 *     """
 *     items = set(items)             # <<<<<<<<<<<<<<
 *     message = 'invalid {}, expect one of {}'.format(name, list(sorted(items)))
 * 
 */
  __Pyx_TraceLine(1080,0,__PYX_ERR(0, 1080, __pyx_L1_error))
  __pyx_t_1 = PySet_New(__pyx_cur_scope->__pyx_v_items); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1080, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_items);
  __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_items, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":1081
 *     """
 *     items = set(items)
 *     message = 'invalid {}, expect one of {}'.format(name, list(sorted(items)))             # <<<<<<<<<<<<<<
 * 
 *     def enum_validator(compiler):
 */
  __Pyx_TraceLine(1081,0,__PYX_ERR(0, 1081, __pyx_L1_error))
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_invalid_expect_one_of, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1081, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __pyx_cur_scope->__pyx_v_items;
  __Pyx_INCREF(__pyx_t_4);
  __pyx_t_5 = PySequence_List(__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1081, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_3 = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_6 = PyList_Sort(__pyx_t_3); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1081, __pyx_L1_error)
  __pyx_t_5 = PySequence_List(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1081, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_7 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_7 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_name, __pyx_t_5};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1081, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_name, __pyx_t_5};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1081, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  } else
  #endif
  {
    __pyx_t_4 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1081, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_INCREF(__pyx_v_name);
    __Pyx_GIVEREF(__pyx_v_name);
    PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_7, __pyx_v_name);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_7, __pyx_t_5);
    __pyx_t_5 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1081, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_message = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":1083
 *     message = 'invalid {}, expect one of {}'.format(name, list(sorted(items)))
 * 
 *     def enum_validator(compiler):             # <<<<<<<<<<<<<<
 *         def validate(value):
 *             if value in items:
 */
  __Pyx_TraceLine(1083,0,__PYX_ERR(0, 1083, __pyx_L1_error))
  __pyx_t_1 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_21create_enum_validator_1enum_validator, 0, __pyx_n_s_create_enum_validator_locals_enu_2, ((PyObject*)__pyx_cur_scope), __pyx_n_s_validr__validator_c, __pyx_d, ((PyObject *)__pyx_codeobj__121)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1083, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_enum_validator = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":1089
 *             raise Invalid(message)
 *         return validate
 *     enum_validator.__name__ = name + '_validator'             # <<<<<<<<<<<<<<
 *     enum_validator.__qualname__ = name + '_validator'
 *     if string:
 */
  __Pyx_TraceLine(1089,0,__PYX_ERR(0, 1089, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyUnicode_ConcatSafe(__pyx_v_name, __pyx_n_u_validator_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1089, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_enum_validator, __pyx_n_s_name, __pyx_t_1) < 0) __PYX_ERR(0, 1089, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":1090
 *         return validate
 *     enum_validator.__name__ = name + '_validator'
 *     enum_validator.__qualname__ = name + '_validator'             # <<<<<<<<<<<<<<
 *     if string:
 *         return validator(accept=str, output=str)(enum_validator)
 */
  __Pyx_TraceLine(1090,0,__PYX_ERR(0, 1090, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyUnicode_ConcatSafe(__pyx_v_name, __pyx_n_u_validator_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1090, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_enum_validator, __pyx_n_s_qualname, __pyx_t_1) < 0) __PYX_ERR(0, 1090, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":1091
 *     enum_validator.__name__ = name + '_validator'
 *     enum_validator.__qualname__ = name + '_validator'
 *     if string:             # <<<<<<<<<<<<<<
 *         return validator(accept=str, output=str)(enum_validator)
 *     else:
 */
  __Pyx_TraceLine(1091,0,__PYX_ERR(0, 1091, __pyx_L1_error))
  __pyx_t_8 = (__pyx_v_string != 0);
  if (__pyx_t_8) {

    /* "validr/_validator_c.pyx":1092
 *     enum_validator.__qualname__ = name + '_validator'
 *     if string:
 *         return validator(accept=str, output=str)(enum_validator)             # <<<<<<<<<<<<<<
 *     else:
 *         return validator(accept=object, output=object)(enum_validator)
 */
    __Pyx_TraceLine(1092,0,__PYX_ERR(0, 1092, __pyx_L1_error))
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_validator); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1092, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1092, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_accept, ((PyObject *)(&PyUnicode_Type))) < 0) __PYX_ERR(0, 1092, __pyx_L1_error)
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_output, ((PyObject *)(&PyUnicode_Type))) < 0) __PYX_ERR(0, 1092, __pyx_L1_error)
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1092, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_4, __pyx_v_enum_validator) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_enum_validator);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1092, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "validr/_validator_c.pyx":1091
 *     enum_validator.__name__ = name + '_validator'
 *     enum_validator.__qualname__ = name + '_validator'
 *     if string:             # <<<<<<<<<<<<<<
 *         return validator(accept=str, output=str)(enum_validator)
 *     else:
 */
  }

  /* "validr/_validator_c.pyx":1094
 *         return validator(accept=str, output=str)(enum_validator)
 *     else:
 *         return validator(accept=object, output=object)(enum_validator)             # <<<<<<<<<<<<<<
 */
  __Pyx_TraceLine(1094,0,__PYX_ERR(0, 1094, __pyx_L1_error))
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_validator); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1094, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1094, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_accept, __pyx_builtin_object) < 0) __PYX_ERR(0, 1094, __pyx_L1_error)
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_output, __pyx_builtin_object) < 0) __PYX_ERR(0, 1094, __pyx_L1_error)
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_empty_tuple, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1094, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v_enum_validator) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_enum_validator);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1094, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "validr/_validator_c.pyx":1070
 * 
 * 
 * def create_enum_validator(str name, items, bint string=True):             # <<<<<<<<<<<<<<
 *     """Create validator by enum items
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("validr._validator_c.create_enum_validator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_enum_validator);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct__validator *__pyx_freelist_6validr_12_validator_c___pyx_scope_struct__validator[8];
static int __pyx_freecount_6validr_12_validator_c___pyx_scope_struct__validator = 0;

static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct__validator(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct__validator > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct__validator)))) {
    o = (PyObject*)__pyx_freelist_6validr_12_validator_c___pyx_scope_struct__validator[--__pyx_freecount_6validr_12_validator_c___pyx_scope_struct__validator];
    memset(o, 0, sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct__validator));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct__validator(PyObject *o) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct__validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct__validator *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_accept_hints);
  Py_CLEAR(p->__pyx_v_output_hints);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct__validator < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct__validator)))) {
    __pyx_freelist_6validr_12_validator_c___pyx_scope_struct__validator[__pyx_freecount_6validr_12_validator_c___pyx_scope_struct__validator++] = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct__validator *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6validr_12_validator_c___pyx_scope_struct__validator(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct__validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct__validator *)o;
  if (p->__pyx_v_accept_hints) {
    e = (*v)(p->__pyx_v_accept_hints, a); if (e) return e;
  }
  if (p->__pyx_v_output_hints) {
    e = (*v)(p->__pyx_v_output_hints, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6validr_12_validator_c___pyx_scope_struct__validator(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct__validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct__validator *)o;
  tmp = ((PyObject*)p->__pyx_v_accept_hints);
  p->__pyx_v_accept_hints = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_output_hints);
  p->__pyx_v_output_hints = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_6validr_12_validator_c___pyx_scope_struct__validator = {
  PyVarObject_HEAD_INIT(0, 0)
  "validr._validator_c.__pyx_scope_struct__validator", /*tp_name*/
  sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct__validator), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct__validator, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6validr_12_validator_c___pyx_scope_struct__validator, /*tp_traverse*/
  __pyx_tp_clear_6validr_12_validator_c___pyx_scope_struct__validator, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6validr_12_validator_c___pyx_scope_struct__validator, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_1_decorator *__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_1_decorator[8];
static int __pyx_freecount_6validr_12_validator_c___pyx_scope_struct_1_decorator = 0;

static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_1_decorator(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_1_decorator > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_1_decorator)))) {
    o = (PyObject*)__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_1_decorator[--__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_1_decorator];
    memset(o, 0, sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_1_decorator));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_1_decorator(PyObject *o) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_1_decorator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_1_decorator *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v__m_validator);
  Py_CLEAR(p->__pyx_v_f);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_1_decorator < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_1_decorator)))) {
    __pyx_freelist_6validr_12_validator_c___pyx_scope_struct_1_decorator[__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_1_decorator++] = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_1_decorator *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6validr_12_validator_c___pyx_scope_struct_1_decorator(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_1_decorator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_1_decorator *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v__m_validator) {
    e = (*v)(p->__pyx_v__m_validator, a); if (e) return e;
  }
  if (p->__pyx_v_f) {
    e = (*v)(p->__pyx_v_f, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6validr_12_validator_c___pyx_scope_struct_1_decorator(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_1_decorator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_1_decorator *)o;
  tmp = ((PyObject*)p->__pyx_outer_scope);
  p->__pyx_outer_scope = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct__validator *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v__m_validator);
  p->__pyx_v__m_validator = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_f);
  p->__pyx_v_f = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_6validr_12_validator_c___pyx_scope_struct_1_decorator = {
  PyVarObject_HEAD_INIT(0, 0)
  "validr._validator_c.__pyx_scope_struct_1_decorator", /*tp_name*/
  sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_1_decorator), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_1_decorator, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6validr_12_validator_c___pyx_scope_struct_1_decorator, /*tp_traverse*/
  __pyx_tp_clear_6validr_12_validator_c___pyx_scope_struct_1_decorator, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_1_decorator, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_2__m_validator *__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_2__m_validator[8];
static int __pyx_freecount_6validr_12_validator_c___pyx_scope_struct_2__m_validator = 0;

static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_2__m_validator(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_2__m_validator > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_2__m_validator)))) {
    o = (PyObject*)__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_2__m_validator[--__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_2__m_validator];
    memset(o, 0, sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_2__m_validator));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_2__m_validator(PyObject *o) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_2__m_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_2__m_validator *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v__m_validate);
  Py_CLEAR(p->__pyx_v_default);
  Py_CLEAR(p->__pyx_v_invalid_to);
  Py_CLEAR(p->__pyx_v_null_output);
  Py_CLEAR(p->__pyx_v_validate);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_2__m_validator < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_2__m_validator)))) {
    __pyx_freelist_6validr_12_validator_c___pyx_scope_struct_2__m_validator[__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_2__m_validator++] = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_2__m_validator *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6validr_12_validator_c___pyx_scope_struct_2__m_validator(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_2__m_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_2__m_validator *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v__m_validate) {
    e = (*v)(p->__pyx_v__m_validate, a); if (e) return e;
  }
  if (p->__pyx_v_default) {
    e = (*v)(p->__pyx_v_default, a); if (e) return e;
  }
  if (p->__pyx_v_invalid_to) {
    e = (*v)(p->__pyx_v_invalid_to, a); if (e) return e;
  }
  if (p->__pyx_v_validate) {
    e = (*v)(p->__pyx_v_validate, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6validr_12_validator_c___pyx_scope_struct_2__m_validator(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_2__m_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_2__m_validator *)o;
  tmp = ((PyObject*)p->__pyx_outer_scope);
  p->__pyx_outer_scope = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_1_decorator *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v__m_validate);
  p->__pyx_v__m_validate = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_default);
  p->__pyx_v_default = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_invalid_to);
  p->__pyx_v_invalid_to = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_validate);
  p->__pyx_v_validate = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_6validr_12_validator_c___pyx_scope_struct_2__m_validator = {
  PyVarObject_HEAD_INIT(0, 0)
  "validr._validator_c.__pyx_scope_struct_2__m_validator", /*tp_name*/
  sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_2__m_validator), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_2__m_validator, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6validr_12_validator_c___pyx_scope_struct_2__m_validator, /*tp_traverse*/
  __pyx_tp_clear_6validr_12_validator_c___pyx_scope_struct_2__m_validator, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_2__m_validator, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_3__key_func_of_schema *__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_3__key_func_of_schema[8];
static int __pyx_freecount_6validr_12_validator_c___pyx_scope_struct_3__key_func_of_schema = 0;

static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_3__key_func_of_schema(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_3__key_func_of_schema > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_3__key_func_of_schema)))) {
    o = (PyObject*)__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_3__key_func_of_schema[--__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_3__key_func_of_schema];
    memset(o, 0, sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_3__key_func_of_schema));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_3__key_func_of_schema(PyObject *o) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_3__key_func_of_schema *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_3__key_func_of_schema *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_key_of_value);
  Py_CLEAR(p->__pyx_v_keys);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_3__key_func_of_schema < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_3__key_func_of_schema)))) {
    __pyx_freelist_6validr_12_validator_c___pyx_scope_struct_3__key_func_of_schema[__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_3__key_func_of_schema++] = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_3__key_func_of_schema *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6validr_12_validator_c___pyx_scope_struct_3__key_func_of_schema(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_3__key_func_of_schema *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_3__key_func_of_schema *)o;
  if (p->__pyx_v_key_of_value) {
    e = (*v)(p->__pyx_v_key_of_value, a); if (e) return e;
  }
  if (p->__pyx_v_keys) {
    e = (*v)(p->__pyx_v_keys, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6validr_12_validator_c___pyx_scope_struct_3__key_func_of_schema(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_3__key_func_of_schema *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_3__key_func_of_schema *)o;
  tmp = ((PyObject*)p->__pyx_v_key_of_value);
  p->__pyx_v_key_of_value = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_keys);
  p->__pyx_v_keys = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_6validr_12_validator_c___pyx_scope_struct_3__key_func_of_schema = {
  PyVarObject_HEAD_INIT(0, 0)
  "validr._validator_c.__pyx_scope_struct_3__key_func_of_schema", /*tp_name*/
  sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_3__key_func_of_schema), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_3__key_func_of_schema, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6validr_12_validator_c___pyx_scope_struct_3__key_func_of_schema, /*tp_traverse*/
  __pyx_tp_clear_6validr_12_validator_c___pyx_scope_struct_3__key_func_of_schema, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_3__key_func_of_schema, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_4_key_of *__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_4_key_of[8];
static int __pyx_freecount_6validr_12_validator_c___pyx_scope_struct_4_key_of = 0;

static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_4_key_of(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_4_key_of > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_4_key_of)))) {
    o = (PyObject*)__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_4_key_of[--__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_4_key_of];
    memset(o, 0, sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_4_key_of));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_4_key_of(PyObject *o) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_4_key_of *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_4_key_of *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_v);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_4_key_of < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_4_key_of)))) {
    __pyx_freelist_6validr_12_validator_c___pyx_scope_struct_4_key_of[__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_4_key_of++] = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_4_key_of *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6validr_12_validator_c___pyx_scope_struct_4_key_of(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_4_key_of *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_4_key_of *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_v) {
    e = (*v)(p->__pyx_v_v, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6validr_12_validator_c___pyx_scope_struct_4_key_of(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_4_key_of *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_4_key_of *)o;
  tmp = ((PyObject*)p->__pyx_outer_scope);
  p->__pyx_outer_scope = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_3__key_func_of_schema *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_v);
  p->__pyx_v_v = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_6validr_12_validator_c___pyx_scope_struct_4_key_of = {
  PyVarObject_HEAD_INIT(0, 0)
  "validr._validator_c.__pyx_scope_struct_4_key_of", /*tp_name*/
  sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_4_key_of), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_4_key_of, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6validr_12_validator_c___pyx_scope_struct_4_key_of, /*tp_traverse*/
  __pyx_tp_clear_6validr_12_validator_c___pyx_scope_struct_4_key_of, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_4_key_of, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_5_genexpr *__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_5_genexpr[8];
static int __pyx_freecount_6validr_12_validator_c___pyx_scope_struct_5_genexpr = 0;

static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_5_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_5_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_5_genexpr)))) {
    o = (PyObject*)__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_5_genexpr[--__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_5_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_5_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_5_genexpr(PyObject *o) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_5_genexpr *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_5_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_k);
  Py_CLEAR(p->__pyx_v_key_of_value);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_5_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_5_genexpr)))) {
    __pyx_freelist_6validr_12_validator_c___pyx_scope_struct_5_genexpr[__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_5_genexpr++] = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_5_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6validr_12_validator_c___pyx_scope_struct_5_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_5_genexpr *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_5_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_k) {
    e = (*v)(p->__pyx_v_k, a); if (e) return e;
  }
  if (p->__pyx_v_key_of_value) {
    e = (*v)(p->__pyx_v_key_of_value, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_6validr_12_validator_c___pyx_scope_struct_5_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "validr._validator_c.__pyx_scope_struct_5_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_5_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_5_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6validr_12_validator_c___pyx_scope_struct_5_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_5_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_6_key_of *__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_6_key_of[8];
static int __pyx_freecount_6validr_12_validator_c___pyx_scope_struct_6_key_of = 0;

static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_6_key_of(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_6_key_of > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_6_key_of)))) {
    o = (PyObject*)__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_6_key_of[--__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_6_key_of];
    memset(o, 0, sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_6_key_of));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_6_key_of(PyObject *o) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_6_key_of *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_6_key_of *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_v);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_6_key_of < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_6_key_of)))) {
    __pyx_freelist_6validr_12_validator_c___pyx_scope_struct_6_key_of[__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_6_key_of++] = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_6_key_of *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6validr_12_validator_c___pyx_scope_struct_6_key_of(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_6_key_of *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_6_key_of *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_v) {
    e = (*v)(p->__pyx_v_v, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6validr_12_validator_c___pyx_scope_struct_6_key_of(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_6_key_of *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_6_key_of *)o;
  tmp = ((PyObject*)p->__pyx_outer_scope);
  p->__pyx_outer_scope = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_3__key_func_of_schema *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_v);
  p->__pyx_v_v = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_6validr_12_validator_c___pyx_scope_struct_6_key_of = {
  PyVarObject_HEAD_INIT(0, 0)
  "validr._validator_c.__pyx_scope_struct_6_key_of", /*tp_name*/
  sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_6_key_of), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_6_key_of, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6validr_12_validator_c___pyx_scope_struct_6_key_of, /*tp_traverse*/
  __pyx_tp_clear_6validr_12_validator_c___pyx_scope_struct_6_key_of, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_6_key_of, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_7_genexpr *__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_7_genexpr[8];
static int __pyx_freecount_6validr_12_validator_c___pyx_scope_struct_7_genexpr = 0;

static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_7_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_7_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_7_genexpr)))) {
    o = (PyObject*)__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_7_genexpr[--__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_7_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_7_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_7_genexpr(PyObject *o) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_7_genexpr *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_7_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_x);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_7_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_7_genexpr)))) {
    __pyx_freelist_6validr_12_validator_c___pyx_scope_struct_7_genexpr[__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_7_genexpr++] = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_7_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6validr_12_validator_c___pyx_scope_struct_7_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_7_genexpr *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_7_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_x) {
    e = (*v)(p->__pyx_v_x, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_6validr_12_validator_c___pyx_scope_struct_7_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "validr._validator_c.__pyx_scope_struct_7_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_7_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_7_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6validr_12_validator_c___pyx_scope_struct_7_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_7_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_8_list_validator *__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_8_list_validator[8];
static int __pyx_freecount_6validr_12_validator_c___pyx_scope_struct_8_list_validator = 0;

static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_8_list_validator(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_8_list_validator > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_8_list_validator)))) {
    o = (PyObject*)__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_8_list_validator[--__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_8_list_validator];
    memset(o, 0, sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_8_list_validator));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_8_list_validator(PyObject *o) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_8_list_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_8_list_validator *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_inner);
  Py_CLEAR(p->__pyx_v_key_of);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_8_list_validator < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_8_list_validator)))) {
    __pyx_freelist_6validr_12_validator_c___pyx_scope_struct_8_list_validator[__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_8_list_validator++] = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_8_list_validator *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6validr_12_validator_c___pyx_scope_struct_8_list_validator(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_8_list_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_8_list_validator *)o;
  if (p->__pyx_v_inner) {
    e = (*v)(p->__pyx_v_inner, a); if (e) return e;
  }
  if (p->__pyx_v_key_of) {
    e = (*v)(p->__pyx_v_key_of, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6validr_12_validator_c___pyx_scope_struct_8_list_validator(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_8_list_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_8_list_validator *)o;
  tmp = ((PyObject*)p->__pyx_v_inner);
  p->__pyx_v_inner = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_key_of);
  p->__pyx_v_key_of = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_6validr_12_validator_c___pyx_scope_struct_8_list_validator = {
  PyVarObject_HEAD_INIT(0, 0)
  "validr._validator_c.__pyx_scope_struct_8_list_validator", /*tp_name*/
  sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_8_list_validator), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_8_list_validator, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6validr_12_validator_c___pyx_scope_struct_8_list_validator, /*tp_traverse*/
  __pyx_tp_clear_6validr_12_validator_c___pyx_scope_struct_8_list_validator, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_8_list_validator, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_9_dict_validator *__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_9_dict_validator[8];
static int __pyx_freecount_6validr_12_validator_c___pyx_scope_struct_9_dict_validator = 0;

static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_9_dict_validator(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_9_dict_validator > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_9_dict_validator)))) {
    o = (PyObject*)__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_9_dict_validator[--__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_9_dict_validator];
    memset(o, 0, sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_9_dict_validator));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_9_dict_validator(PyObject *o) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_9_dict_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_9_dict_validator *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_inners);
  Py_CLEAR(p->__pyx_v_validate_extra_key);
  Py_CLEAR(p->__pyx_v_validate_extra_value);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_9_dict_validator < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_9_dict_validator)))) {
    __pyx_freelist_6validr_12_validator_c___pyx_scope_struct_9_dict_validator[__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_9_dict_validator++] = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_9_dict_validator *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6validr_12_validator_c___pyx_scope_struct_9_dict_validator(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_9_dict_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_9_dict_validator *)o;
  if (p->__pyx_v_inners) {
    e = (*v)(p->__pyx_v_inners, a); if (e) return e;
  }
  if (p->__pyx_v_validate_extra_key) {
    e = (*v)(p->__pyx_v_validate_extra_key, a); if (e) return e;
  }
  if (p->__pyx_v_validate_extra_value) {
    e = (*v)(p->__pyx_v_validate_extra_value, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6validr_12_validator_c___pyx_scope_struct_9_dict_validator(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_9_dict_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_9_dict_validator *)o;
  tmp = ((PyObject*)p->__pyx_v_inners);
  p->__pyx_v_inners = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_validate_extra_key);
  p->__pyx_v_validate_extra_key = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_validate_extra_value);
  p->__pyx_v_validate_extra_value = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_6validr_12_validator_c___pyx_scope_struct_9_dict_validator = {
  PyVarObject_HEAD_INIT(0, 0)
  "validr._validator_c.__pyx_scope_struct_9_dict_validator", /*tp_name*/
  sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_9_dict_validator), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_9_dict_validator, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6validr_12_validator_c___pyx_scope_struct_9_dict_validator, /*tp_traverse*/
  __pyx_tp_clear_6validr_12_validator_c___pyx_scope_struct_9_dict_validator, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_9_dict_validator, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_10_model_validator *__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_10_model_validator[8];
static int __pyx_freecount_6validr_12_validator_c___pyx_scope_struct_10_model_validator = 0;

static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_10_model_validator(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_10_model_validator > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_10_model_validator)))) {
    o = (PyObject*)__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_10_model_validator[--__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_10_model_validator];
    memset(o, 0, sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_10_model_validator));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_10_model_validator(PyObject *o) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_10_model_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_10_model_validator *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_items);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_10_model_validator < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_10_model_validator)))) {
    __pyx_freelist_6validr_12_validator_c___pyx_scope_struct_10_model_validator[__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_10_model_validator++] = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_10_model_validator *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6validr_12_validator_c___pyx_scope_struct_10_model_validator(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_10_model_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_10_model_validator *)o;
  if (p->__pyx_v_items) {
    e = (*v)(p->__pyx_v_items, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6validr_12_validator_c___pyx_scope_struct_10_model_validator(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_10_model_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_10_model_validator *)o;
  tmp = ((PyObject*)p->__pyx_v_items);
  p->__pyx_v_items = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_6validr_12_validator_c___pyx_scope_struct_10_model_validator = {
  PyVarObject_HEAD_INIT(0, 0)
  "validr._validator_c.__pyx_scope_struct_10_model_validator", /*tp_name*/
  sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_10_model_validator), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_10_model_validator, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6validr_12_validator_c___pyx_scope_struct_10_model_validator, /*tp_traverse*/
  __pyx_tp_clear_6validr_12_validator_c___pyx_scope_struct_10_model_validator, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_10_model_validator, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_11_enum_validator *__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_11_enum_validator[8];
static int __pyx_freecount_6validr_12_validator_c___pyx_scope_struct_11_enum_validator = 0;

static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_11_enum_validator(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_11_enum_validator > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_11_enum_validator)))) {
    o = (PyObject*)__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_11_enum_validator[--__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_11_enum_validator];
    memset(o, 0, sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_11_enum_validator));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_11_enum_validator(PyObject *o) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_11_enum_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_11_enum_validator *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_expects);
  Py_CLEAR(p->__pyx_v_items);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_11_enum_validator < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_11_enum_validator)))) {
    __pyx_freelist_6validr_12_validator_c___pyx_scope_struct_11_enum_validator[__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_11_enum_validator++] = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_11_enum_validator *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6validr_12_validator_c___pyx_scope_struct_11_enum_validator(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_11_enum_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_11_enum_validator *)o;
  if (p->__pyx_v_items) {
    e = (*v)(p->__pyx_v_items, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6validr_12_validator_c___pyx_scope_struct_11_enum_validator(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_11_enum_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_11_enum_validator *)o;
  tmp = ((PyObject*)p->__pyx_v_items);
  p->__pyx_v_items = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_6validr_12_validator_c___pyx_scope_struct_11_enum_validator = {
  PyVarObject_HEAD_INIT(0, 0)
  "validr._validator_c.__pyx_scope_struct_11_enum_validator", /*tp_name*/
  sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_11_enum_validator), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_11_enum_validator, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6validr_12_validator_c___pyx_scope_struct_11_enum_validator, /*tp_traverse*/
  __pyx_tp_clear_6validr_12_validator_c___pyx_scope_struct_11_enum_validator, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_11_enum_validator, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_12__union_list_validator *__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_12__union_list_validator[8];
static int __pyx_freecount_6validr_12_validator_c___pyx_scope_struct_12__union_list_validator = 0;

static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_12__union_list_validator(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_12__union_list_validator > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_12__union_list_validator)))) {
    o = (PyObject*)__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_12__union_list_validator[--__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_12__union_list_validator];
    memset(o, 0, sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_12__union_list_validator));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_12__union_list_validator(PyObject *o) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_12__union_list_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_12__union_list_validator *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_dict_inner);
  Py_CLEAR(p->__pyx_v_list_inner);
  Py_CLEAR(p->__pyx_v_scalar_inner);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_12__union_list_validator < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_12__union_list_validator)))) {
    __pyx_freelist_6validr_12_validator_c___pyx_scope_struct_12__union_list_validator[__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_12__union_list_validator++] = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_12__union_list_validator *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6validr_12_validator_c___pyx_scope_struct_12__union_list_validator(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_12__union_list_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_12__union_list_validator *)o;
  if (p->__pyx_v_dict_inner) {
    e = (*v)(p->__pyx_v_dict_inner, a); if (e) return e;
  }
  if (p->__pyx_v_list_inner) {
    e = (*v)(p->__pyx_v_list_inner, a); if (e) return e;
  }
  if (p->__pyx_v_scalar_inner) {
    e = (*v)(p->__pyx_v_scalar_inner, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6validr_12_validator_c___pyx_scope_struct_12__union_list_validator(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_12__union_list_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_12__union_list_validator *)o;
  tmp = ((PyObject*)p->__pyx_v_dict_inner);
  p->__pyx_v_dict_inner = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_list_inner);
  p->__pyx_v_list_inner = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_scalar_inner);
  p->__pyx_v_scalar_inner = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_6validr_12_validator_c___pyx_scope_struct_12__union_list_validator = {
  PyVarObject_HEAD_INIT(0, 0)
  "validr._validator_c.__pyx_scope_struct_12__union_list_validator", /*tp_name*/
  sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_12__union_list_validator), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_12__union_list_validator, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6validr_12_validator_c___pyx_scope_struct_12__union_list_validator, /*tp_traverse*/
  __pyx_tp_clear_6validr_12_validator_c___pyx_scope_struct_12__union_list_validator, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_12__union_list_validator, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_13__union_dict_validator *__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_13__union_dict_validator[8];
static int __pyx_freecount_6validr_12_validator_c___pyx_scope_struct_13__union_dict_validator = 0;

static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_13__union_dict_validator(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_13__union_dict_validator > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_13__union_dict_validator)))) {
    o = (PyObject*)__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_13__union_dict_validator[--__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_13__union_dict_validator];
    memset(o, 0, sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_13__union_dict_validator));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_13__union_dict_validator(PyObject *o) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_13__union_dict_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_13__union_dict_validator *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_by);
  Py_CLEAR(p->__pyx_v_expect_bys);
  Py_CLEAR(p->__pyx_v_inners);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_13__union_dict_validator < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_13__union_dict_validator)))) {
    __pyx_freelist_6validr_12_validator_c___pyx_scope_struct_13__union_dict_validator[__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_13__union_dict_validator++] = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_13__union_dict_validator *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6validr_12_validator_c___pyx_scope_struct_13__union_dict_validator(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_13__union_dict_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_13__union_dict_validator *)o;
  if (p->__pyx_v_inners) {
    e = (*v)(p->__pyx_v_inners, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6validr_12_validator_c___pyx_scope_struct_13__union_dict_validator(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_13__union_dict_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_13__union_dict_validator *)o;
  tmp = ((PyObject*)p->__pyx_v_inners);
  p->__pyx_v_inners = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_6validr_12_validator_c___pyx_scope_struct_13__union_dict_validator = {
  PyVarObject_HEAD_INIT(0, 0)
  "validr._validator_c.__pyx_scope_struct_13__union_dict_validator", /*tp_name*/
  sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_13__union_dict_validator), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_13__union_dict_validator, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6validr_12_validator_c___pyx_scope_struct_13__union_dict_validator, /*tp_traverse*/
  __pyx_tp_clear_6validr_12_validator_c___pyx_scope_struct_13__union_dict_validator, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_13__union_dict_validator, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_14_int_validator *__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_14_int_validator[8];
static int __pyx_freecount_6validr_12_validator_c___pyx_scope_struct_14_int_validator = 0;

static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_14_int_validator(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_14_int_validator > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_14_int_validator)))) {
    o = (PyObject*)__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_14_int_validator[--__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_14_int_validator];
    memset(o, 0, sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_14_int_validator));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_14_int_validator(PyObject *o) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_14_int_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_14_int_validator *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_max);
  Py_CLEAR(p->__pyx_v_min);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_14_int_validator < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_14_int_validator)))) {
    __pyx_freelist_6validr_12_validator_c___pyx_scope_struct_14_int_validator[__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_14_int_validator++] = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_14_int_validator *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6validr_12_validator_c___pyx_scope_struct_14_int_validator(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_14_int_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_14_int_validator *)o;
  if (p->__pyx_v_max) {
    e = (*v)(p->__pyx_v_max, a); if (e) return e;
  }
  if (p->__pyx_v_min) {
    e = (*v)(p->__pyx_v_min, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6validr_12_validator_c___pyx_scope_struct_14_int_validator(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_14_int_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_14_int_validator *)o;
  tmp = ((PyObject*)p->__pyx_v_max);
  p->__pyx_v_max = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_min);
  p->__pyx_v_min = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_6validr_12_validator_c___pyx_scope_struct_14_int_validator = {
  PyVarObject_HEAD_INIT(0, 0)
  "validr._validator_c.__pyx_scope_struct_14_int_validator", /*tp_name*/
  sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_14_int_validator), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_14_int_validator, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6validr_12_validator_c___pyx_scope_struct_14_int_validator, /*tp_traverse*/
  __pyx_tp_clear_6validr_12_validator_c___pyx_scope_struct_14_int_validator, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_14_int_validator, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_15_float_validator *__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_15_float_validator[8];
static int __pyx_freecount_6validr_12_validator_c___pyx_scope_struct_15_float_validator = 0;

static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_15_float_validator(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_15_float_validator > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_15_float_validator)))) {
    o = (PyObject*)__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_15_float_validator[--__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_15_float_validator];
    memset(o, 0, sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_15_float_validator));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_15_float_validator(PyObject *o) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_15_float_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_15_float_validator *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_exmax);
  Py_CLEAR(p->__pyx_v_exmin);
  Py_CLEAR(p->__pyx_v_max);
  Py_CLEAR(p->__pyx_v_min);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_15_float_validator < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_15_float_validator)))) {
    __pyx_freelist_6validr_12_validator_c___pyx_scope_struct_15_float_validator[__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_15_float_validator++] = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_15_float_validator *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6validr_12_validator_c___pyx_scope_struct_15_float_validator(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_15_float_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_15_float_validator *)o;
  if (p->__pyx_v_exmax) {
    e = (*v)(p->__pyx_v_exmax, a); if (e) return e;
  }
  if (p->__pyx_v_exmin) {
    e = (*v)(p->__pyx_v_exmin, a); if (e) return e;
  }
  if (p->__pyx_v_max) {
    e = (*v)(p->__pyx_v_max, a); if (e) return e;
  }
  if (p->__pyx_v_min) {
    e = (*v)(p->__pyx_v_min, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6validr_12_validator_c___pyx_scope_struct_15_float_validator(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_15_float_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_15_float_validator *)o;
  tmp = ((PyObject*)p->__pyx_v_exmax);
  p->__pyx_v_exmax = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_exmin);
  p->__pyx_v_exmin = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_max);
  p->__pyx_v_max = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_min);
  p->__pyx_v_min = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_6validr_12_validator_c___pyx_scope_struct_15_float_validator = {
  PyVarObject_HEAD_INIT(0, 0)
  "validr._validator_c.__pyx_scope_struct_15_float_validator", /*tp_name*/
  sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_15_float_validator), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_15_float_validator, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6validr_12_validator_c___pyx_scope_struct_15_float_validator, /*tp_traverse*/
  __pyx_tp_clear_6validr_12_validator_c___pyx_scope_struct_15_float_validator, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_15_float_validator, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_16_str_validator *__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_16_str_validator[8];
static int __pyx_freecount_6validr_12_validator_c___pyx_scope_struct_16_str_validator = 0;

static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_16_str_validator(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_16_str_validator > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_16_str_validator)))) {
    o = (PyObject*)__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_16_str_validator[--__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_16_str_validator];
    memset(o, 0, sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_16_str_validator));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_16_str_validator(PyObject *o) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_16_str_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_16_str_validator *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_match);
  Py_CLEAR(p->__pyx_v_re_match);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_16_str_validator < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_16_str_validator)))) {
    __pyx_freelist_6validr_12_validator_c___pyx_scope_struct_16_str_validator[__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_16_str_validator++] = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_16_str_validator *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6validr_12_validator_c___pyx_scope_struct_16_str_validator(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_16_str_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_16_str_validator *)o;
  if (p->__pyx_v_re_match) {
    e = (*v)(p->__pyx_v_re_match, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6validr_12_validator_c___pyx_scope_struct_16_str_validator(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_16_str_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_16_str_validator *)o;
  tmp = ((PyObject*)p->__pyx_v_re_match);
  p->__pyx_v_re_match = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_6validr_12_validator_c___pyx_scope_struct_16_str_validator = {
  PyVarObject_HEAD_INIT(0, 0)
  "validr._validator_c.__pyx_scope_struct_16_str_validator", /*tp_name*/
  sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_16_str_validator), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_16_str_validator, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6validr_12_validator_c___pyx_scope_struct_16_str_validator, /*tp_traverse*/
  __pyx_tp_clear_6validr_12_validator_c___pyx_scope_struct_16_str_validator, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_16_str_validator, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_17_bytes_validator *__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_17_bytes_validator[8];
static int __pyx_freecount_6validr_12_validator_c___pyx_scope_struct_17_bytes_validator = 0;

static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_17_bytes_validator(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_17_bytes_validator > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_17_bytes_validator)))) {
    o = (PyObject*)__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_17_bytes_validator[--__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_17_bytes_validator];
    memset(o, 0, sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_17_bytes_validator));
    (void) PyObject_INIT(o, t);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_17_bytes_validator(PyObject *o) {
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_17_bytes_validator < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_17_bytes_validator)))) {
    __pyx_freelist_6validr_12_validator_c___pyx_scope_struct_17_bytes_validator[__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_17_bytes_validator++] = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_17_bytes_validator *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static PyTypeObject __pyx_type_6validr_12_validator_c___pyx_scope_struct_17_bytes_validator = {
  PyVarObject_HEAD_INIT(0, 0)
  "validr._validator_c.__pyx_scope_struct_17_bytes_validator", /*tp_name*/
  sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_17_bytes_validator), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_17_bytes_validator, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_17_bytes_validator, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_18_date_validator *__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_18_date_validator[8];
static int __pyx_freecount_6validr_12_validator_c___pyx_scope_struct_18_date_validator = 0;

static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_18_date_validator(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_18_date_validator > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_18_date_validator)))) {
    o = (PyObject*)__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_18_date_validator[--__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_18_date_validator];
    memset(o, 0, sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_18_date_validator));
    (void) PyObject_INIT(o, t);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_18_date_validator(PyObject *o) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_18_date_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_18_date_validator *)o;
  Py_CLEAR(p->__pyx_v_format);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_18_date_validator < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_18_date_validator)))) {
    __pyx_freelist_6validr_12_validator_c___pyx_scope_struct_18_date_validator[__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_18_date_validator++] = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_18_date_validator *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static PyTypeObject __pyx_type_6validr_12_validator_c___pyx_scope_struct_18_date_validator = {
  PyVarObject_HEAD_INIT(0, 0)
  "validr._validator_c.__pyx_scope_struct_18_date_validator", /*tp_name*/
  sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_18_date_validator), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_18_date_validator, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_18_date_validator, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_19_time_validator *__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_19_time_validator[8];
static int __pyx_freecount_6validr_12_validator_c___pyx_scope_struct_19_time_validator = 0;

static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_19_time_validator(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_19_time_validator > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_19_time_validator)))) {
    o = (PyObject*)__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_19_time_validator[--__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_19_time_validator];
    memset(o, 0, sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_19_time_validator));
    (void) PyObject_INIT(o, t);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_19_time_validator(PyObject *o) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_19_time_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_19_time_validator *)o;
  Py_CLEAR(p->__pyx_v_format);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_19_time_validator < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_19_time_validator)))) {
    __pyx_freelist_6validr_12_validator_c___pyx_scope_struct_19_time_validator[__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_19_time_validator++] = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_19_time_validator *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static PyTypeObject __pyx_type_6validr_12_validator_c___pyx_scope_struct_19_time_validator = {
  PyVarObject_HEAD_INIT(0, 0)
  "validr._validator_c.__pyx_scope_struct_19_time_validator", /*tp_name*/
  sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_19_time_validator), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_19_time_validator, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_19_time_validator, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_20_datetime_validator *__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_20_datetime_validator[8];
static int __pyx_freecount_6validr_12_validator_c___pyx_scope_struct_20_datetime_validator = 0;

static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_20_datetime_validator(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_20_datetime_validator > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_20_datetime_validator)))) {
    o = (PyObject*)__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_20_datetime_validator[--__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_20_datetime_validator];
    memset(o, 0, sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_20_datetime_validator));
    (void) PyObject_INIT(o, t);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_20_datetime_validator(PyObject *o) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_20_datetime_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_20_datetime_validator *)o;
  Py_CLEAR(p->__pyx_v_format);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_20_datetime_validator < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_20_datetime_validator)))) {
    __pyx_freelist_6validr_12_validator_c___pyx_scope_struct_20_datetime_validator[__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_20_datetime_validator++] = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_20_datetime_validator *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static PyTypeObject __pyx_type_6validr_12_validator_c___pyx_scope_struct_20_datetime_validator = {
  PyVarObject_HEAD_INIT(0, 0)
  "validr._validator_c.__pyx_scope_struct_20_datetime_validator", /*tp_name*/
  sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_20_datetime_validator), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_20_datetime_validator, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_20_datetime_validator, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_21_timedelta_validator *__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_21_timedelta_validator[8];
static int __pyx_freecount_6validr_12_validator_c___pyx_scope_struct_21_timedelta_validator = 0;

static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_21_timedelta_validator(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_21_timedelta_validator > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_21_timedelta_validator)))) {
    o = (PyObject*)__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_21_timedelta_validator[--__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_21_timedelta_validator];
    memset(o, 0, sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_21_timedelta_validator));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_21_timedelta_validator(PyObject *o) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_21_timedelta_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_21_timedelta_validator *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_max_repr);
  Py_CLEAR(p->__pyx_v_max_value);
  Py_CLEAR(p->__pyx_v_min_repr);
  Py_CLEAR(p->__pyx_v_min_value);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_21_timedelta_validator < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_21_timedelta_validator)))) {
    __pyx_freelist_6validr_12_validator_c___pyx_scope_struct_21_timedelta_validator[__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_21_timedelta_validator++] = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_21_timedelta_validator *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6validr_12_validator_c___pyx_scope_struct_21_timedelta_validator(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_21_timedelta_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_21_timedelta_validator *)o;
  if (p->__pyx_v_max_repr) {
    e = (*v)(p->__pyx_v_max_repr, a); if (e) return e;
  }
  if (p->__pyx_v_max_value) {
    e = (*v)(p->__pyx_v_max_value, a); if (e) return e;
  }
  if (p->__pyx_v_min_repr) {
    e = (*v)(p->__pyx_v_min_repr, a); if (e) return e;
  }
  if (p->__pyx_v_min_value) {
    e = (*v)(p->__pyx_v_min_value, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6validr_12_validator_c___pyx_scope_struct_21_timedelta_validator(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_21_timedelta_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_21_timedelta_validator *)o;
  tmp = ((PyObject*)p->__pyx_v_max_repr);
  p->__pyx_v_max_repr = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_max_value);
  p->__pyx_v_max_value = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_min_repr);
  p->__pyx_v_min_repr = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_min_value);
  p->__pyx_v_min_value = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_6validr_12_validator_c___pyx_scope_struct_21_timedelta_validator = {
  PyVarObject_HEAD_INIT(0, 0)
  "validr._validator_c.__pyx_scope_struct_21_timedelta_validator", /*tp_name*/
  sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_21_timedelta_validator), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_21_timedelta_validator, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6validr_12_validator_c___pyx_scope_struct_21_timedelta_validator, /*tp_traverse*/
  __pyx_tp_clear_6validr_12_validator_c___pyx_scope_struct_21_timedelta_validator, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_21_timedelta_validator, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_22_ipv4_validator *__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_22_ipv4_validator[8];
static int __pyx_freecount_6validr_12_validator_c___pyx_scope_struct_22_ipv4_validator = 0;

static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_22_ipv4_validator(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_22_ipv4_validator > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_22_ipv4_validator)))) {
    o = (PyObject*)__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_22_ipv4_validator[--__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_22_ipv4_validator];
    memset(o, 0, sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_22_ipv4_validator));
    (void) PyObject_INIT(o, t);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_22_ipv4_validator(PyObject *o) {
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_22_ipv4_validator < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_22_ipv4_validator)))) {
    __pyx_freelist_6validr_12_validator_c___pyx_scope_struct_22_ipv4_validator[__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_22_ipv4_validator++] = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_22_ipv4_validator *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static PyTypeObject __pyx_type_6validr_12_validator_c___pyx_scope_struct_22_ipv4_validator = {
  PyVarObject_HEAD_INIT(0, 0)
  "validr._validator_c.__pyx_scope_struct_22_ipv4_validator", /*tp_name*/
  sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_22_ipv4_validator), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_22_ipv4_validator, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_22_ipv4_validator, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_23_ipv6_validator *__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_23_ipv6_validator[8];
static int __pyx_freecount_6validr_12_validator_c___pyx_scope_struct_23_ipv6_validator = 0;

static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_23_ipv6_validator(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_23_ipv6_validator > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_23_ipv6_validator)))) {
    o = (PyObject*)__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_23_ipv6_validator[--__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_23_ipv6_validator];
    memset(o, 0, sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_23_ipv6_validator));
    (void) PyObject_INIT(o, t);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_23_ipv6_validator(PyObject *o) {
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_23_ipv6_validator < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_23_ipv6_validator)))) {
    __pyx_freelist_6validr_12_validator_c___pyx_scope_struct_23_ipv6_validator[__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_23_ipv6_validator++] = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_23_ipv6_validator *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static PyTypeObject __pyx_type_6validr_12_validator_c___pyx_scope_struct_23_ipv6_validator = {
  PyVarObject_HEAD_INIT(0, 0)
  "validr._validator_c.__pyx_scope_struct_23_ipv6_validator", /*tp_name*/
  sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_23_ipv6_validator), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_23_ipv6_validator, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_23_ipv6_validator, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_24_email_validator *__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_24_email_validator[8];
static int __pyx_freecount_6validr_12_validator_c___pyx_scope_struct_24_email_validator = 0;

static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_24_email_validator(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_24_email_validator > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_24_email_validator)))) {
    o = (PyObject*)__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_24_email_validator[--__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_24_email_validator];
    memset(o, 0, sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_24_email_validator));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_24_email_validator(PyObject *o) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_24_email_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_24_email_validator *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v__validate);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_24_email_validator < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_24_email_validator)))) {
    __pyx_freelist_6validr_12_validator_c___pyx_scope_struct_24_email_validator[__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_24_email_validator++] = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_24_email_validator *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6validr_12_validator_c___pyx_scope_struct_24_email_validator(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_24_email_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_24_email_validator *)o;
  if (p->__pyx_v__validate) {
    e = (*v)(p->__pyx_v__validate, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6validr_12_validator_c___pyx_scope_struct_24_email_validator(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_24_email_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_24_email_validator *)o;
  tmp = ((PyObject*)p->__pyx_v__validate);
  p->__pyx_v__validate = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_6validr_12_validator_c___pyx_scope_struct_24_email_validator = {
  PyVarObject_HEAD_INIT(0, 0)
  "validr._validator_c.__pyx_scope_struct_24_email_validator", /*tp_name*/
  sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_24_email_validator), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_24_email_validator, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6validr_12_validator_c___pyx_scope_struct_24_email_validator, /*tp_traverse*/
  __pyx_tp_clear_6validr_12_validator_c___pyx_scope_struct_24_email_validator, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_24_email_validator, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_25_url_validator *__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_25_url_validator[8];
static int __pyx_freecount_6validr_12_validator_c___pyx_scope_struct_25_url_validator = 0;

static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_25_url_validator(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_25_url_validator > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_25_url_validator)))) {
    o = (PyObject*)__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_25_url_validator[--__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_25_url_validator];
    memset(o, 0, sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_25_url_validator));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_25_url_validator(PyObject *o) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_25_url_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_25_url_validator *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_allow_scheme);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_25_url_validator < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_25_url_validator)))) {
    __pyx_freelist_6validr_12_validator_c___pyx_scope_struct_25_url_validator[__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_25_url_validator++] = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_25_url_validator *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6validr_12_validator_c___pyx_scope_struct_25_url_validator(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_25_url_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_25_url_validator *)o;
  if (p->__pyx_v_allow_scheme) {
    e = (*v)(p->__pyx_v_allow_scheme, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6validr_12_validator_c___pyx_scope_struct_25_url_validator(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_25_url_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_25_url_validator *)o;
  tmp = ((PyObject*)p->__pyx_v_allow_scheme);
  p->__pyx_v_allow_scheme = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_6validr_12_validator_c___pyx_scope_struct_25_url_validator = {
  PyVarObject_HEAD_INIT(0, 0)
  "validr._validator_c.__pyx_scope_struct_25_url_validator", /*tp_name*/
  sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_25_url_validator), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_25_url_validator, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6validr_12_validator_c___pyx_scope_struct_25_url_validator, /*tp_traverse*/
  __pyx_tp_clear_6validr_12_validator_c___pyx_scope_struct_25_url_validator, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_25_url_validator, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_26_uuid_validator *__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_26_uuid_validator[8];
static int __pyx_freecount_6validr_12_validator_c___pyx_scope_struct_26_uuid_validator = 0;

static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_26_uuid_validator(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_26_uuid_validator > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_26_uuid_validator)))) {
    o = (PyObject*)__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_26_uuid_validator[--__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_26_uuid_validator];
    memset(o, 0, sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_26_uuid_validator));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_26_uuid_validator(PyObject *o) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_26_uuid_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_26_uuid_validator *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_msg);
  Py_CLEAR(p->__pyx_v_version);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_26_uuid_validator < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_26_uuid_validator)))) {
    __pyx_freelist_6validr_12_validator_c___pyx_scope_struct_26_uuid_validator[__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_26_uuid_validator++] = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_26_uuid_validator *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6validr_12_validator_c___pyx_scope_struct_26_uuid_validator(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_26_uuid_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_26_uuid_validator *)o;
  if (p->__pyx_v_msg) {
    e = (*v)(p->__pyx_v_msg, a); if (e) return e;
  }
  if (p->__pyx_v_version) {
    e = (*v)(p->__pyx_v_version, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6validr_12_validator_c___pyx_scope_struct_26_uuid_validator(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_26_uuid_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_26_uuid_validator *)o;
  tmp = ((PyObject*)p->__pyx_v_msg);
  p->__pyx_v_msg = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_version);
  p->__pyx_v_version = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_6validr_12_validator_c___pyx_scope_struct_26_uuid_validator = {
  PyVarObject_HEAD_INIT(0, 0)
  "validr._validator_c.__pyx_scope_struct_26_uuid_validator", /*tp_name*/
  sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_26_uuid_validator), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_26_uuid_validator, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6validr_12_validator_c___pyx_scope_struct_26_uuid_validator, /*tp_traverse*/
  __pyx_tp_clear_6validr_12_validator_c___pyx_scope_struct_26_uuid_validator, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_26_uuid_validator, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_27_create_re_validator *__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_27_create_re_validator[8];
static int __pyx_freecount_6validr_12_validator_c___pyx_scope_struct_27_create_re_validator = 0;

static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_27_create_re_validator(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_27_create_re_validator > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_27_create_re_validator)))) {
    o = (PyObject*)__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_27_create_re_validator[--__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_27_create_re_validator];
    memset(o, 0, sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_27_create_re_validator));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_27_create_re_validator(PyObject *o) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_27_create_re_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_27_create_re_validator *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_match);
  Py_CLEAR(p->__pyx_v_message);
  Py_CLEAR(p->__pyx_v_name);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_27_create_re_validator < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_27_create_re_validator)))) {
    __pyx_freelist_6validr_12_validator_c___pyx_scope_struct_27_create_re_validator[__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_27_create_re_validator++] = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_27_create_re_validator *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6validr_12_validator_c___pyx_scope_struct_27_create_re_validator(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_27_create_re_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_27_create_re_validator *)o;
  if (p->__pyx_v_match) {
    e = (*v)(p->__pyx_v_match, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6validr_12_validator_c___pyx_scope_struct_27_create_re_validator(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_27_create_re_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_27_create_re_validator *)o;
  tmp = ((PyObject*)p->__pyx_v_match);
  p->__pyx_v_match = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_6validr_12_validator_c___pyx_scope_struct_27_create_re_validator = {
  PyVarObject_HEAD_INIT(0, 0)
  "validr._validator_c.__pyx_scope_struct_27_create_re_validator", /*tp_name*/
  sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_27_create_re_validator), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_27_create_re_validator, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6validr_12_validator_c___pyx_scope_struct_27_create_re_validator, /*tp_traverse*/
  __pyx_tp_clear_6validr_12_validator_c___pyx_scope_struct_27_create_re_validator, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_27_create_re_validator, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_28_re_validator *__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_28_re_validator[8];
static int __pyx_freecount_6validr_12_validator_c___pyx_scope_struct_28_re_validator = 0;

static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_28_re_validator(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_28_re_validator > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_28_re_validator)))) {
    o = (PyObject*)__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_28_re_validator[--__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_28_re_validator];
    memset(o, 0, sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_28_re_validator));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_28_re_validator(PyObject *o) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_28_re_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_28_re_validator *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_28_re_validator < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_28_re_validator)))) {
    __pyx_freelist_6validr_12_validator_c___pyx_scope_struct_28_re_validator[__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_28_re_validator++] = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_28_re_validator *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6validr_12_validator_c___pyx_scope_struct_28_re_validator(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_28_re_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_28_re_validator *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6validr_12_validator_c___pyx_scope_struct_28_re_validator(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_28_re_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_28_re_validator *)o;
  tmp = ((PyObject*)p->__pyx_outer_scope);
  p->__pyx_outer_scope = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_27_create_re_validator *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_6validr_12_validator_c___pyx_scope_struct_28_re_validator = {
  PyVarObject_HEAD_INIT(0, 0)
  "validr._validator_c.__pyx_scope_struct_28_re_validator", /*tp_name*/
  sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_28_re_validator), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_28_re_validator, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6validr_12_validator_c___pyx_scope_struct_28_re_validator, /*tp_traverse*/
  __pyx_tp_clear_6validr_12_validator_c___pyx_scope_struct_28_re_validator, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_28_re_validator, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_29_create_enum_validator *__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_29_create_enum_validator[8];
static int __pyx_freecount_6validr_12_validator_c___pyx_scope_struct_29_create_enum_validator = 0;

static PyObject *__pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_29_create_enum_validator(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_29_create_enum_validator > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_29_create_enum_validator)))) {
    o = (PyObject*)__pyx_freelist_6validr_12_validator_c___pyx_scope_struct_29_create_enum_validator[--__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_29_create_enum_validator];
    memset(o, 0, sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_29_create_enum_validator));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_29_create_enum_validator(PyObject *o) {
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_29_create_enum_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_29_create_enum_validator *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_items);
  Py_CLEAR(p->__pyx_v_message);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_29_create_enum_validator < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_29_create_enum_validator)))) {
    __pyx_freelist_6validr_12_validator_c___pyx_scope_struct_29_create_enum_validator[__pyx_freecount_6validr_12_validator_c___pyx_scope_struct_29_create_enum_validator++] = ((struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_29_create_enum_validator *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6validr_12_validator_c___pyx_scope_struct_29_create_enum_validator(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_29_create_enum_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_29_create_enum_validator *)o;
  if (p->__pyx_v_items) {
    e = (*v)(p->__pyx_v_items, a); if (e) return e;
  }
  if (p->__pyx_v_message) {
    e = (*v)(p->__pyx_v_message, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6validr_12_validator_c___pyx_scope_struct_29_create_enum_validator(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_29_create_enum_validator *p = (struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_29_create_enum_validator *)o;
  tmp = ((PyObject*)p->__pyx_v_items);
  p->__pyx_v_items = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_message);
  p->__pyx_v_message = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_6validr_12_validator_c___pyx_scope_struct_29_create_enum_validator = {
  PyVarObject_HEAD_INIT(0, 0)
  "validr._validator_c.__pyx_scope_struct_29_create_enum_validator", /*tp_name*/
  sizeof(struct __pyx_obj_6validr_12_validator_c___pyx_scope_struct_29_create_enum_validator), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6validr_12_validator_c___pyx_scope_struct_29_create_enum_validator, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6validr_12_validator_c___pyx_scope_struct_29_create_enum_validator, /*tp_traverse*/
  __pyx_tp_clear_6validr_12_validator_c___pyx_scope_struct_29_create_enum_validator, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6validr_12_validator_c___pyx_scope_struct_29_create_enum_validator, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {"is_dict", (PyCFunction)__pyx_pw_6validr_12_validator_c_1is_dict, METH_O, 0},
  {"get_dict_value", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6validr_12_validator_c_3get_dict_value, METH_VARARGS|METH_KEYWORDS, 0},
  {"get_object_value", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6validr_12_validator_c_5get_object_value, METH_VARARGS|METH_KEYWORDS, 0},
  {"_is_empty", (PyCFunction)__pyx_pw_6validr_12_validator_c_7_is_empty, METH_O, 0},
  {"_update_validate_func_info", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6validr_12_validator_c_9_update_validate_func_info, METH_VARARGS|METH_KEYWORDS, 0},
  {"_key_of_scalar", (PyCFunction)__pyx_pw_6validr_12_validator_c_17_key_of_scalar, METH_O, 0},
  {"_slim_dict", (PyCFunction)__pyx_pw_6validr_12_validator_c_23_slim_dict, METH_O, 0},
  {"_dump_enum_value", (PyCFunction)__pyx_pw_6validr_12_validator_c_29_dump_enum_value, METH_O, 0},
  {"union_validator", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6validr_12_validator_c_33union_validator, METH_VARARGS|METH_KEYWORDS, 0},
  {"_optional_or_has_default", (PyCFunction)__pyx_pw_6validr_12_validator_c_35_optional_or_has_default, METH_O, 0},
  {"any_validate", (PyCFunction)__pyx_pw_6validr_12_validator_c_41any_validate, METH_O, 0},
  {"_parse_timedelta", (PyCFunction)__pyx_pw_6validr_12_validator_c_61_parse_timedelta, METH_O, 0},
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec__validator_c(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec__validator_c},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "_validator_c",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_u_, __pyx_k_, sizeof(__pyx_k_), 0, 1, 0, 0},
  {&__pyx_kp_u_0, __pyx_k_0, sizeof(__pyx_k_0), 0, 1, 0, 0},
  {&__pyx_kp_u_1, __pyx_k_1, sizeof(__pyx_k_1), 0, 1, 0, 0},
  {&__pyx_kp_u_34, __pyx_k_34, sizeof(__pyx_k_34), 0, 1, 0, 0},
  {&__pyx_kp_u_39, __pyx_k_39, sizeof(__pyx_k_39), 0, 1, 0, 0},
  {&__pyx_n_s_AddressValueError, __pyx_k_AddressValueError, sizeof(__pyx_k_AddressValueError), 0, 0, 1, 1},
  {&__pyx_n_s_Any, __pyx_k_Any, sizeof(__pyx_k_Any), 0, 0, 1, 1},
  {&__pyx_n_s_AttributeError, __pyx_k_AttributeError, sizeof(__pyx_k_AttributeError), 0, 0, 1, 1},
  {&__pyx_n_s_DurationError, __pyx_k_DurationError, sizeof(__pyx_k_DurationError), 0, 0, 1, 1},
  {&__pyx_n_s_EmailNotValidError, __pyx_k_EmailNotValidError, sizeof(__pyx_k_EmailNotValidError), 0, 0, 1, 1},
  {&__pyx_n_u_FALSE, __pyx_k_FALSE, sizeof(__pyx_k_FALSE), 0, 1, 0, 1},
  {&__pyx_n_s_FALSE_VALUES, __pyx_k_FALSE_VALUES, sizeof(__pyx_k_FALSE_VALUES), 0, 0, 1, 1},
  {&__pyx_n_s_FQDN, __pyx_k_FQDN, sizeof(__pyx_k_FQDN), 0, 0, 1, 1},
  {&__pyx_n_u_False, __pyx_k_False, sizeof(__pyx_k_False), 0, 1, 0, 1},
  {&__pyx_kp_u_H_M_S, __pyx_k_H_M_S, sizeof(__pyx_k_H_M_S), 0, 1, 0, 0},
  {&__pyx_n_s_IPv4Address, __pyx_k_IPv4Address, sizeof(__pyx_k_IPv4Address), 0, 0, 1, 1},
  {&__pyx_n_s_IPv6Address, __pyx_k_IPv6Address, sizeof(__pyx_k_IPv6Address), 0, 0, 1, 1},
  {&__pyx_n_s_Invalid, __pyx_k_Invalid, sizeof(__pyx_k_Invalid), 0, 0, 1, 1},
  {&__pyx_n_s_Iterable, __pyx_k_Iterable, sizeof(__pyx_k_Iterable), 0, 0, 1, 1},
  {&__pyx_n_s_List, __pyx_k_List, sizeof(__pyx_k_List), 0, 0, 1, 1},
  {&__pyx_n_s_Mapping, __pyx_k_Mapping, sizeof(__pyx_k_Mapping), 0, 0, 1, 1},
  {&__pyx_n_u_N, __pyx_k_N, sizeof(__pyx_k_N), 0, 1, 0, 1},
  {&__pyx_n_u_NO, __pyx_k_NO, sizeof(__pyx_k_NO), 0, 1, 0, 1},
  {&__pyx_n_u_No, __pyx_k_No, sizeof(__pyx_k_No), 0, 1, 0, 1},
  {&__pyx_kp_u_None, __pyx_k_None, sizeof(__pyx_k_None), 0, 1, 0, 0},
  {&__pyx_n_u_OFF, __pyx_k_OFF, sizeof(__pyx_k_OFF), 0, 1, 0, 1},
  {&__pyx_n_u_ON, __pyx_k_ON, sizeof(__pyx_k_ON), 0, 1, 0, 1},
  {&__pyx_n_u_Off, __pyx_k_Off, sizeof(__pyx_k_Off), 0, 1, 0, 1},
  {&__pyx_n_u_On, __pyx_k_On, sizeof(__pyx_k_On), 0, 1, 0, 1},
  {&__pyx_n_s_Optional, __pyx_k_Optional, sizeof(__pyx_k_Optional), 0, 0, 1, 1},
  {&__pyx_n_s_SchemaError, __pyx_k_SchemaError, sizeof(__pyx_k_SchemaError), 0, 0, 1, 1},
  {&__pyx_n_u_TRUE, __pyx_k_TRUE, sizeof(__pyx_k_TRUE), 0, 1, 0, 1},
  {&__pyx_n_s_TRUE_VALUES, __pyx_k_TRUE_VALUES, sizeof(__pyx_k_TRUE_VALUES), 0, 0, 1, 1},
  {&__pyx_n_u_True, __pyx_k_True, sizeof(__pyx_k_True), 0, 1, 0, 1},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_n_s_UUID, __pyx_k_UUID, sizeof(__pyx_k_UUID), 0, 0, 1, 1},
  {&__pyx_n_s_Union, __pyx_k_Union, sizeof(__pyx_k_Union), 0, 0, 1, 1},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_n_u_Y, __pyx_k_Y, sizeof(__pyx_k_Y), 0, 1, 0, 1},
  {&__pyx_n_u_YES, __pyx_k_YES, sizeof(__pyx_k_YES), 0, 1, 0, 1},
  {&__pyx_kp_u_Y_m_d, __pyx_k_Y_m_d, sizeof(__pyx_k_Y_m_d), 0, 1, 0, 0},
  {&__pyx_kp_u_Y_m_dT_H_M_S_fZ, __pyx_k_Y_m_dT_H_M_S_fZ, sizeof(__pyx_k_Y_m_dT_H_M_S_fZ), 0, 1, 0, 0},
  {&__pyx_n_u_Yes, __pyx_k_Yes, sizeof(__pyx_k_Yes), 0, 1, 0, 1},
  {&__pyx_kp_u__2, __pyx_k__2, sizeof(__pyx_k__2), 0, 1, 0, 0},
  {&__pyx_kp_u__39, __pyx_k__39, sizeof(__pyx_k__39), 0, 1, 0, 0},
  {&__pyx_kp_u__40, __pyx_k__40, sizeof(__pyx_k__40), 0, 1, 0, 0},
  {&__pyx_kp_u__41, __pyx_k__41, sizeof(__pyx_k__41), 0, 1, 0, 0},
  {&__pyx_kp_u__65, __pyx_k__65, sizeof(__pyx_k__65), 0, 1, 0, 0},
  {&__pyx_kp_u__67, __pyx_k__67, sizeof(__pyx_k__67), 0, 1, 0, 0},
  {&__pyx_kp_u__69, __pyx_k__69, sizeof(__pyx_k__69), 0, 1, 0, 0},
  {&__pyx_kp_u__71, __pyx_k__71, sizeof(__pyx_k__71), 0, 1, 0, 0},
  {&__pyx_kp_u__73, __pyx_k__73, sizeof(__pyx_k__73), 0, 1, 0, 0},
  {&__pyx_kp_u_a_z0_9_a_z0_9, __pyx_k_a_z0_9_a_z0_9, sizeof(__pyx_k_a_z0_9_a_z0_9), 0, 1, 0, 0},
  {&__pyx_n_s_accept, __pyx_k_accept, sizeof(__pyx_k_accept), 0, 0, 1, 1},
  {&__pyx_n_s_accept_hints, __pyx_k_accept_hints, sizeof(__pyx_k_accept_hints), 0, 0, 1, 1},
  {&__pyx_n_s_accept_object, __pyx_k_accept_object, sizeof(__pyx_k_accept_object), 0, 0, 1, 1},
  {&__pyx_n_s_accept_string, __pyx_k_accept_string, sizeof(__pyx_k_accept_string), 0, 0, 1, 1},
  {&__pyx_n_s_allow_empty_local, __pyx_k_allow_empty_local, sizeof(__pyx_k_allow_empty_local), 0, 0, 1, 1},
  {&__pyx_n_s_allow_scheme, __pyx_k_allow_scheme, sizeof(__pyx_k_allow_scheme), 0, 0, 1, 1},
  {&__pyx_n_s_allow_smtputf8, __pyx_k_allow_smtputf8, sizeof(__pyx_k_allow_smtputf8), 0, 0, 1, 1},
  {&__pyx_kp_u_ambiguous_union_schema, __pyx_k_ambiguous_union_schema, sizeof(__pyx_k_ambiguous_union_schema), 0, 1, 0, 0},
  {&__pyx_kp_u_amp, __pyx_k_amp, sizeof(__pyx_k_amp), 0, 1, 0, 0},
  {&__pyx_n_s_annotations, __pyx_k_annotations, sizeof(__pyx_k_annotations), 0, 0, 1, 1},
  {&__pyx_n_s_annotations_2, __pyx_k_annotations_2, sizeof(__pyx_k_annotations_2), 0, 0, 1, 1},
  {&__pyx_n_u_any, __pyx_k_any, sizeof(__pyx_k_any), 0, 1, 0, 1},
  {&__pyx_n_s_any_validate, __pyx_k_any_validate, sizeof(__pyx_k_any_validate), 0, 0, 1, 1},
  {&__pyx_n_s_any_validator, __pyx_k_any_validator, sizeof(__pyx_k_any_validator), 0, 0, 1, 1},
  {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
  {&__pyx_n_u_asdict, __pyx_k_asdict, sizeof(__pyx_k_asdict), 0, 1, 0, 1},
  {&__pyx_n_u_bool, __pyx_k_bool, sizeof(__pyx_k_bool), 0, 1, 0, 1},
  {&__pyx_n_s_bool_validator, __pyx_k_bool_validator, sizeof(__pyx_k_bool_validator), 0, 0, 1, 1},
  {&__pyx_n_s_bool_validator_locals_validate, __pyx_k_bool_validator_locals_validate, sizeof(__pyx_k_bool_validator_locals_validate), 0, 0, 1, 1},
  {&__pyx_n_s_builtin_validators, __pyx_k_builtin_validators, sizeof(__pyx_k_builtin_validators), 0, 0, 1, 1},
  {&__pyx_n_s_by, __pyx_k_by, sizeof(__pyx_k_by), 0, 0, 1, 1},
  {&__pyx_n_u_by, __pyx_k_by, sizeof(__pyx_k_by), 0, 1, 0, 1},
  {&__pyx_kp_u_by_argument_must_be_str_type_fo, __pyx_k_by_argument_must_be_str_type_fo, sizeof(__pyx_k_by_argument_must_be_str_type_fo), 0, 1, 0, 0},
  {&__pyx_n_s_by_name, __pyx_k_by_name, sizeof(__pyx_k_by_name), 0, 0, 1, 1},
  {&__pyx_n_u_bytes, __pyx_k_bytes, sizeof(__pyx_k_bytes), 0, 1, 0, 1},
  {&__pyx_kp_u_bytes_length_must_d, __pyx_k_bytes_length_must_d, sizeof(__pyx_k_bytes_length_must_d), 0, 1, 0, 0},
  {&__pyx_kp_u_bytes_length_must_d_2, __pyx_k_bytes_length_must_d_2, sizeof(__pyx_k_bytes_length_must_d_2), 0, 1, 0, 0},
  {&__pyx_n_s_bytes_validator, __pyx_k_bytes_validator, sizeof(__pyx_k_bytes_validator), 0, 0, 1, 1},
  {&__pyx_n_s_bytes_validator_locals_validate, __pyx_k_bytes_validator_locals_validate, sizeof(__pyx_k_bytes_validator_locals_validate), 0, 0, 1, 1},
  {&__pyx_kp_u_can_not_output_both_string_and_o, __pyx_k_can_not_output_both_string_and_o, sizeof(__pyx_k_can_not_output_both_string_and_o), 0, 1, 0, 0},
  {&__pyx_kp_u_can_not_set_both_invalid_to_and, __pyx_k_can_not_set_both_invalid_to_and, sizeof(__pyx_k_can_not_set_both_invalid_to_and), 0, 1, 0, 0},
  {&__pyx_n_s_check_deliverability, __pyx_k_check_deliverability, sizeof(__pyx_k_check_deliverability), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
  {&__pyx_n_s_compile, __pyx_k_compile, sizeof(__pyx_k_compile), 0, 0, 1, 1},
  {&__pyx_n_s_compiler, __pyx_k_compiler, sizeof(__pyx_k_compiler), 0, 0, 1, 1},
  {&__pyx_n_s_compressed, __pyx_k_compressed, sizeof(__pyx_k_compressed), 0, 0, 1, 1},
  {&__pyx_n_s_copy, __pyx_k_copy, sizeof(__pyx_k_copy), 0, 0, 1, 1},
  {&__pyx_n_s_create_enum_validator, __pyx_k_create_enum_validator, sizeof(__pyx_k_create_enum_validator), 0, 0, 1, 1},
  {&__pyx_n_s_create_enum_validator_locals_enu, __pyx_k_create_enum_validator_locals_enu, sizeof(__pyx_k_create_enum_validator_locals_enu), 0, 0, 1, 1},
  {&__pyx_n_s_create_enum_validator_locals_enu_2, __pyx_k_create_enum_validator_locals_enu_2, sizeof(__pyx_k_create_enum_validator_locals_enu_2), 0, 0, 1, 1},
  {&__pyx_n_s_create_re_validator, __pyx_k_create_re_validator, sizeof(__pyx_k_create_re_validator), 0, 0, 1, 1},
  {&__pyx_n_s_create_re_validator_locals_re_va, __pyx_k_create_re_validator_locals_re_va, sizeof(__pyx_k_create_re_validator_locals_re_va), 0, 0, 1, 1},
  {&__pyx_n_s_create_re_validator_locals_re_va_2, __pyx_k_create_re_validator_locals_re_va_2, sizeof(__pyx_k_create_re_validator_locals_re_va_2), 0, 0, 1, 1},
  {&__pyx_n_u_d, __pyx_k_d, sizeof(__pyx_k_d), 0, 1, 0, 1},
  {&__pyx_kp_u_d_17_d_x_X_d_15, __pyx_k_d_17_d_x_X_d_15, sizeof(__pyx_k_d_17_d_x_X_d_15), 0, 1, 0, 0},
  {&__pyx_kp_u_d_2_s_d_2_s_1_d_10, __pyx_k_d_2_s_d_2_s_1_d_10, sizeof(__pyx_k_d_2_s_d_2_s_1_d_10), 0, 1, 0, 0},
  {&__pyx_n_s_date, __pyx_k_date, sizeof(__pyx_k_date), 0, 0, 1, 1},
  {&__pyx_n_u_date, __pyx_k_date, sizeof(__pyx_k_date), 0, 1, 0, 1},
  {&__pyx_n_s_date_validator, __pyx_k_date_validator, sizeof(__pyx_k_date_validator), 0, 0, 1, 1},
  {&__pyx_n_s_date_validator_locals_validate, __pyx_k_date_validator_locals_validate, sizeof(__pyx_k_date_validator_locals_validate), 0, 0, 1, 1},
  {&__pyx_n_s_datetime, __pyx_k_datetime, sizeof(__pyx_k_datetime), 0, 0, 1, 1},
  {&__pyx_n_u_datetime, __pyx_k_datetime, sizeof(__pyx_k_datetime), 0, 1, 0, 1},
  {&__pyx_n_s_datetime_validator, __pyx_k_datetime_validator, sizeof(__pyx_k_datetime_validator), 0, 0, 1, 1},
  {&__pyx_n_s_datetime_validator_locals_valida, __pyx_k_datetime_validator_locals_valida, sizeof(__pyx_k_datetime_validator_locals_valida), 0, 0, 1, 1},
  {&__pyx_n_s_decorator, __pyx_k_decorator, sizeof(__pyx_k_decorator), 0, 0, 1, 1},
  {&__pyx_n_s_default, __pyx_k_default, sizeof(__pyx_k_default), 0, 0, 1, 1},
  {&__pyx_n_u_default, __pyx_k_default, sizeof(__pyx_k_default), 0, 1, 0, 1},
  {&__pyx_n_s_default_maxlen, __pyx_k_default_maxlen, sizeof(__pyx_k_default_maxlen), 0, 0, 1, 1},
  {&__pyx_kp_u_default_or_optional_must_be_set, __pyx_k_default_or_optional_must_be_set, sizeof(__pyx_k_default_or_optional_must_be_set), 0, 1, 0, 0},
  {&__pyx_n_s_desc, __pyx_k_desc, sizeof(__pyx_k_desc), 0, 0, 1, 1},
  {&__pyx_n_u_desc, __pyx_k_desc, sizeof(__pyx_k_desc), 0, 1, 0, 1},
  {&__pyx_n_u_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 1, 0, 1},
  {&__pyx_n_s_dict_inner, __pyx_k_dict_inner, sizeof(__pyx_k_dict_inner), 0, 0, 1, 1},
  {&__pyx_kp_u_dict_length_must_d, __pyx_k_dict_length_must_d, sizeof(__pyx_k_dict_length_must_d), 0, 1, 0, 0},
  {&__pyx_kp_u_dict_length_must_d_2, __pyx_k_dict_length_must_d_2, sizeof(__pyx_k_dict_length_must_d_2), 0, 1, 0, 0},
  {&__pyx_n_s_dict_validator, __pyx_k_dict_validator, sizeof(__pyx_k_dict_validator), 0, 0, 1, 1},
  {&__pyx_n_s_dict_validator_locals_validate, __pyx_k_dict_validator_locals_validate, sizeof(__pyx_k_dict_validator_locals_validate), 0, 0, 1, 1},
  {&__pyx_n_s_doc, __pyx_k_doc, sizeof(__pyx_k_doc), 0, 0, 1, 1},
  {&__pyx_n_s_dump_enum_value, __pyx_k_dump_enum_value, sizeof(__pyx_k_dump_enum_value), 0, 0, 1, 1},
  {&__pyx_n_s_durationpy, __pyx_k_durationpy, sizeof(__pyx_k_durationpy), 0, 0, 1, 1},
  {&__pyx_kp_u_dynamic_dict_not_allowed_non_dic, __pyx_k_dynamic_dict_not_allowed_non_dic, sizeof(__pyx_k_dynamic_dict_not_allowed_non_dic), 0, 1, 0, 0},
  {&__pyx_n_s_e, __pyx_k_e, sizeof(__pyx_k_e), 0, 0, 1, 1},
  {&__pyx_n_u_email, __pyx_k_email, sizeof(__pyx_k_email), 0, 1, 0, 1},
  {&__pyx_n_s_email_validator, __pyx_k_email_validator, sizeof(__pyx_k_email_validator), 0, 0, 1, 1},
  {&__pyx_n_s_email_validator_locals_validate, __pyx_k_email_validator_locals_validate, sizeof(__pyx_k_email_validator_locals_validate), 0, 0, 1, 1},
  {&__pyx_n_s_enter, __pyx_k_enter, sizeof(__pyx_k_enter), 0, 0, 1, 1},
  {&__pyx_n_u_enum, __pyx_k_enum, sizeof(__pyx_k_enum), 0, 1, 0, 1},
  {&__pyx_kp_u_enum_items_not_provided, __pyx_k_enum_items_not_provided, sizeof(__pyx_k_enum_items_not_provided), 0, 1, 0, 0},
  {&__pyx_n_s_enum_validator, __pyx_k_enum_validator, sizeof(__pyx_k_enum_validator), 0, 0, 1, 1},
  {&__pyx_n_s_enum_validator_locals_validate, __pyx_k_enum_validator_locals_validate, sizeof(__pyx_k_enum_validator_locals_validate), 0, 0, 1, 1},
  {&__pyx_n_s_enumerate, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
  {&__pyx_n_s_err_msg, __pyx_k_err_msg, sizeof(__pyx_k_err_msg), 0, 0, 1, 1},
  {&__pyx_n_s_escape, __pyx_k_escape, sizeof(__pyx_k_escape), 0, 0, 1, 1},
  {&__pyx_n_s_ex, __pyx_k_ex, sizeof(__pyx_k_ex), 0, 0, 1, 1},
  {&__pyx_n_s_exception, __pyx_k_exception, sizeof(__pyx_k_exception), 0, 0, 1, 1},
  {&__pyx_n_s_exit, __pyx_k_exit, sizeof(__pyx_k_exit), 0, 0, 1, 1},
  {&__pyx_n_s_exmax, __pyx_k_exmax, sizeof(__pyx_k_exmax), 0, 0, 1, 1},
  {&__pyx_n_s_exmin, __pyx_k_exmin, sizeof(__pyx_k_exmin), 0, 0, 1, 1},
  {&__pyx_n_s_expect_bys, __pyx_k_expect_bys, sizeof(__pyx_k_expect_bys), 0, 0, 1, 1},
  {&__pyx_kp_u_expect_one_of, __pyx_k_expect_one_of, sizeof(__pyx_k_expect_one_of), 0, 1, 0, 0},
  {&__pyx_n_s_expects, __pyx_k_expects, sizeof(__pyx_k_expects), 0, 0, 1, 1},
  {&__pyx_n_s_extended, __pyx_k_extended, sizeof(__pyx_k_extended), 0, 0, 1, 1},
  {&__pyx_n_s_extra_keys, __pyx_k_extra_keys, sizeof(__pyx_k_extra_keys), 0, 0, 1, 1},
  {&__pyx_n_s_f, __pyx_k_f, sizeof(__pyx_k_f), 0, 0, 1, 1},
  {&__pyx_n_u_false, __pyx_k_false, sizeof(__pyx_k_false), 0, 1, 0, 1},
  {&__pyx_n_u_float, __pyx_k_float, sizeof(__pyx_k_float), 0, 1, 0, 1},
  {&__pyx_n_s_float_info, __pyx_k_float_info, sizeof(__pyx_k_float_info), 0, 0, 1, 1},
  {&__pyx_n_s_float_validator, __pyx_k_float_validator, sizeof(__pyx_k_float_validator), 0, 0, 1, 1},
  {&__pyx_n_s_float_validator_locals_validate, __pyx_k_float_validator_locals_validate, sizeof(__pyx_k_float_validator_locals_validate), 0, 0, 1, 1},
  {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
  {&__pyx_n_u_fqdn, __pyx_k_fqdn, sizeof(__pyx_k_fqdn), 0, 1, 0, 1},
  {&__pyx_n_s_fqdn_obj, __pyx_k_fqdn_obj, sizeof(__pyx_k_fqdn_obj), 0, 0, 1, 1},
  {&__pyx_n_s_fqdn_validator, __pyx_k_fqdn_validator, sizeof(__pyx_k_fqdn_validator), 0, 0, 1, 1},
  {&__pyx_n_s_fqdn_validator_locals_validate, __pyx_k_fqdn_validator_locals_validate, sizeof(__pyx_k_fqdn_validator_locals_validate), 0, 0, 1, 1},
  {&__pyx_n_s_from_str, __pyx_k_from_str, sizeof(__pyx_k_from_str), 0, 0, 1, 1},
  {&__pyx_n_s_fromtimestamp, __pyx_k_fromtimestamp, sizeof(__pyx_k_fromtimestamp), 0, 0, 1, 1},
  {&__pyx_n_s_functools, __pyx_k_functools, sizeof(__pyx_k_functools), 0, 0, 1, 1},
  {&__pyx_n_s_genexpr, __pyx_k_genexpr, sizeof(__pyx_k_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_get, __pyx_k_get, sizeof(__pyx_k_get), 0, 0, 1, 1},
  {&__pyx_n_u_get, __pyx_k_get, sizeof(__pyx_k_get), 0, 1, 0, 1},
  {&__pyx_n_u_getitem, __pyx_k_getitem, sizeof(__pyx_k_getitem), 0, 1, 0, 1},
  {&__pyx_n_s_getter, __pyx_k_getter, sizeof(__pyx_k_getter), 0, 0, 1, 1},
  {&__pyx_kp_u_gt, __pyx_k_gt, sizeof(__pyx_k_gt), 0, 1, 0, 0},
  {&__pyx_n_s_has_default, __pyx_k_has_default, sizeof(__pyx_k_has_default), 0, 0, 1, 1},
  {&__pyx_n_s_has_invalid_to, __pyx_k_has_invalid_to, sizeof(__pyx_k_has_invalid_to), 0, 0, 1, 1},
  {&__pyx_n_s_hint, __pyx_k_hint, sizeof(__pyx_k_hint), 0, 0, 1, 1},
  {&__pyx_n_s_hints, __pyx_k_hints, sizeof(__pyx_k_hints), 0, 0, 1, 1},
  {&__pyx_kp_u_http_https, __pyx_k_http_https, sizeof(__pyx_k_http_https), 0, 1, 0, 0},
  {&__pyx_n_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
  {&__pyx_n_u_idcard, __pyx_k_idcard, sizeof(__pyx_k_idcard), 0, 1, 0, 1},
  {&__pyx_n_s_ignore_kwargs, __pyx_k_ignore_kwargs, sizeof(__pyx_k_ignore_kwargs), 0, 0, 1, 1},
  {&__pyx_kp_u_illegal_version_number, __pyx_k_illegal_version_number, sizeof(__pyx_k_illegal_version_number), 0, 1, 0, 0},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_inner, __pyx_k_inner, sizeof(__pyx_k_inner), 0, 0, 1, 1},
  {&__pyx_n_s_inner_info, __pyx_k_inner_info, sizeof(__pyx_k_inner_info), 0, 0, 1, 1},
  {&__pyx_n_s_inner_schema, __pyx_k_inner_schema, sizeof(__pyx_k_inner_schema), 0, 0, 1, 1},
  {&__pyx_n_s_inners, __pyx_k_inners, sizeof(__pyx_k_inners), 0, 0, 1, 1},
  {&__pyx_n_u_int, __pyx_k_int, sizeof(__pyx_k_int), 0, 1, 0, 1},
  {&__pyx_n_s_int_validator, __pyx_k_int_validator, sizeof(__pyx_k_int_validator), 0, 0, 1, 1},
  {&__pyx_n_s_int_validator_locals_validate, __pyx_k_int_validator_locals_validate, sizeof(__pyx_k_int_validator_locals_validate), 0, 0, 1, 1},
  {&__pyx_kp_u_invalid_accept_argument, __pyx_k_invalid_accept_argument, sizeof(__pyx_k_invalid_accept_argument), 0, 1, 0, 0},
  {&__pyx_kp_u_invalid_bool, __pyx_k_invalid_bool, sizeof(__pyx_k_invalid_bool), 0, 1, 0, 0},
  {&__pyx_kp_u_invalid_bytes, __pyx_k_invalid_bytes, sizeof(__pyx_k_invalid_bytes), 0, 1, 0, 0},
  {&__pyx_kp_u_invalid_date, __pyx_k_invalid_date, sizeof(__pyx_k_invalid_date), 0, 1, 0, 0},
  {&__pyx_kp_u_invalid_datetime, __pyx_k_invalid_datetime, sizeof(__pyx_k_invalid_datetime), 0, 1, 0, 0},
  {&__pyx_kp_u_invalid_default_value_r, __pyx_k_invalid_default_value_r, sizeof(__pyx_k_invalid_default_value_r), 0, 1, 0, 0},
  {&__pyx_kp_u_invalid_email_address, __pyx_k_invalid_email_address, sizeof(__pyx_k_invalid_email_address), 0, 1, 0, 0},
  {&__pyx_kp_u_invalid_expect_one_of, __pyx_k_invalid_expect_one_of, sizeof(__pyx_k_invalid_expect_one_of), 0, 1, 0, 0},
  {&__pyx_kp_u_invalid_float, __pyx_k_invalid_float, sizeof(__pyx_k_invalid_float), 0, 1, 0, 0},
  {&__pyx_kp_u_invalid_fqdn, __pyx_k_invalid_fqdn, sizeof(__pyx_k_invalid_fqdn), 0, 1, 0, 0},
  {&__pyx_kp_u_invalid_int, __pyx_k_invalid_int, sizeof(__pyx_k_invalid_int), 0, 1, 0, 0},
  {&__pyx_kp_u_invalid_invalid_to_value_r, __pyx_k_invalid_invalid_to_value_r, sizeof(__pyx_k_invalid_invalid_to_value_r), 0, 1, 0, 0},
  {&__pyx_kp_u_invalid_ipv4_address, __pyx_k_invalid_ipv4_address, sizeof(__pyx_k_invalid_ipv4_address), 0, 1, 0, 0},
  {&__pyx_kp_u_invalid_ipv6_address, __pyx_k_invalid_ipv6_address, sizeof(__pyx_k_invalid_ipv6_address), 0, 1, 0, 0},
  {&__pyx_kp_u_invalid_max_timedelta, __pyx_k_invalid_max_timedelta, sizeof(__pyx_k_invalid_max_timedelta), 0, 1, 0, 0},
  {&__pyx_kp_u_invalid_min_timedelta, __pyx_k_invalid_min_timedelta, sizeof(__pyx_k_invalid_min_timedelta), 0, 1, 0, 0},
  {&__pyx_kp_u_invalid_output_argument, __pyx_k_invalid_output_argument, sizeof(__pyx_k_invalid_output_argument), 0, 1, 0, 0},
  {&__pyx_kp_u_invalid_s, __pyx_k_invalid_s, sizeof(__pyx_k_invalid_s), 0, 1, 0, 0},
  {&__pyx_kp_u_invalid_string, __pyx_k_invalid_string, sizeof(__pyx_k_invalid_string), 0, 1, 0, 0},
  {&__pyx_kp_u_invalid_time, __pyx_k_invalid_time, sizeof(__pyx_k_invalid_time), 0, 1, 0, 0},
  {&__pyx_kp_u_invalid_timedelta, __pyx_k_invalid_timedelta, sizeof(__pyx_k_invalid_timedelta), 0, 1, 0, 0},
  {&__pyx_n_s_invalid_to, __pyx_k_invalid_to, sizeof(__pyx_k_invalid_to), 0, 0, 1, 1},
  {&__pyx_n_u_invalid_to, __pyx_k_invalid_to, sizeof(__pyx_k_invalid_to), 0, 1, 0, 1},
  {&__pyx_n_s_invalid_to_default, __pyx_k_invalid_to_default, sizeof(__pyx_k_invalid_to_default), 0, 0, 1, 1},
  {&__pyx_n_u_invalid_to_default, __pyx_k_invalid_to_default, sizeof(__pyx_k_invalid_to_default), 0, 1, 0, 1},
  {&__pyx_kp_u_invalid_url, __pyx_k_invalid_url, sizeof(__pyx_k_invalid_url), 0, 1, 0, 0},
  {&__pyx_kp_u_invalid_url_scheme_expect, __pyx_k_invalid_url_scheme_expect, sizeof(__pyx_k_invalid_url_scheme_expect), 0, 1, 0, 0},
  {&__pyx_kp_u_invalid_uuid, __pyx_k_invalid_uuid, sizeof(__pyx_k_invalid_uuid), 0, 1, 0, 0},
  {&__pyx_kp_u_invalid_uuid_2, __pyx_k_invalid_uuid_2, sizeof(__pyx_k_invalid_uuid_2), 0, 1, 0, 0},
  {&__pyx_n_s_ipaddress, __pyx_k_ipaddress, sizeof(__pyx_k_ipaddress), 0, 0, 1, 1},
  {&__pyx_n_u_ipv4, __pyx_k_ipv4, sizeof(__pyx_k_ipv4), 0, 1, 0, 1},
  {&__pyx_n_s_ipv4_validator, __pyx_k_ipv4_validator, sizeof(__pyx_k_ipv4_validator), 0, 0, 1, 1},
  {&__pyx_n_s_ipv4_validator_locals_validate, __pyx_k_ipv4_validator_locals_validate, sizeof(__pyx_k_ipv4_validator_locals_validate), 0, 0, 1, 1},
  {&__pyx_n_u_ipv6, __pyx_k_ipv6, sizeof(__pyx_k_ipv6), 0, 1, 0, 1},
  {&__pyx_n_s_ipv6_validator, __pyx_k_ipv6_validator, sizeof(__pyx_k_ipv6_validator), 0, 0, 1, 1},
  {&__pyx_n_s_ipv6_validator_locals_validate, __pyx_k_ipv6_validator_locals_validate, sizeof(__pyx_k_ipv6_validator_locals_validate), 0, 0, 1, 1},
  {&__pyx_n_s_is_dynamic, __pyx_k_is_dynamic, sizeof(__pyx_k_is_dynamic), 0, 0, 1, 1},
  {&__pyx_n_s_is_model, __pyx_k_is_model, sizeof(__pyx_k_is_model), 0, 0, 1, 1},
  {&__pyx_n_s_is_null, __pyx_k_is_null, sizeof(__pyx_k_is_null), 0, 0, 1, 1},
  {&__pyx_n_s_is_object, __pyx_k_is_object, sizeof(__pyx_k_is_object), 0, 0, 1, 1},
  {&__pyx_n_s_is_string, __pyx_k_is_string, sizeof(__pyx_k_is_string), 0, 0, 1, 1},
  {&__pyx_n_s_is_valid, __pyx_k_is_valid, sizeof(__pyx_k_is_valid), 0, 0, 1, 1},
  {&__pyx_n_s_items, __pyx_k_items, sizeof(__pyx_k_items), 0, 0, 1, 1},
  {&__pyx_n_u_items, __pyx_k_items, sizeof(__pyx_k_items), 0, 1, 0, 1},
  {&__pyx_n_s_k, __pyx_k_k, sizeof(__pyx_k_k), 0, 0, 1, 1},
  {&__pyx_n_s_key, __pyx_k_key, sizeof(__pyx_k_key), 0, 0, 1, 1},
  {&__pyx_n_s_key_func_of_schema, __pyx_k_key_func_of_schema, sizeof(__pyx_k_key_func_of_schema), 0, 0, 1, 1},
  {&__pyx_n_s_key_func_of_schema_locals_key_o, __pyx_k_key_func_of_schema_locals_key_o, sizeof(__pyx_k_key_func_of_schema_locals_key_o), 0, 0, 1, 1},
  {&__pyx_n_s_key_func_of_schema_locals_key_o_2, __pyx_k_key_func_of_schema_locals_key_o_2, sizeof(__pyx_k_key_func_of_schema_locals_key_o_2), 0, 0, 1, 1},
  {&__pyx_n_s_key_of, __pyx_k_key_of, sizeof(__pyx_k_key_of), 0, 0, 1, 1},
  {&__pyx_n_s_key_of_scalar, __pyx_k_key_of_scalar, sizeof(__pyx_k_key_of_scalar), 0, 0, 1, 1},
  {&__pyx_n_s_key_of_value, __pyx_k_key_of_value, sizeof(__pyx_k_key_of_value), 0, 0, 1, 1},
  {&__pyx_n_s_keys, __pyx_k_keys, sizeof(__pyx_k_keys), 0, 0, 1, 1},
  {&__pyx_n_s_length, __pyx_k_length, sizeof(__pyx_k_length), 0, 0, 1, 1},
  {&__pyx_kp_u_length_must, __pyx_k_length_must, sizeof(__pyx_k_length_must), 0, 1, 0, 0},
  {&__pyx_kp_u_length_must_2, __pyx_k_length_must_2, sizeof(__pyx_k_length_must_2), 0, 1, 0, 0},
  {&__pyx_n_u_list, __pyx_k_list, sizeof(__pyx_k_list), 0, 1, 0, 1},
  {&__pyx_n_s_list_inner, __pyx_k_list_inner, sizeof(__pyx_k_list_inner), 0, 0, 1, 1},
  {&__pyx_kp_u_list_length_must_d, __pyx_k_list_length_must_d, sizeof(__pyx_k_list_length_must_d), 0, 1, 0, 0},
  {&__pyx_kp_u_list_length_must_d_2, __pyx_k_list_length_must_d_2, sizeof(__pyx_k_list_length_must_d_2), 0, 1, 0, 0},
  {&__pyx_n_s_list_validator, __pyx_k_list_validator, sizeof(__pyx_k_list_validator), 0, 0, 1, 1},
  {&__pyx_n_s_list_validator_locals_validate, __pyx_k_list_validator_locals_validate, sizeof(__pyx_k_list_validator_locals_validate), 0, 0, 1, 1},
  {&__pyx_n_s_local_output_object, __pyx_k_local_output_object, sizeof(__pyx_k_local_output_object), 0, 0, 1, 1},
  {&__pyx_kp_u_lt, __pyx_k_lt, sizeof(__pyx_k_lt), 0, 1, 0, 0},
  {&__pyx_n_s_m_validate, __pyx_k_m_validate, sizeof(__pyx_k_m_validate), 0, 0, 1, 1},
  {&__pyx_n_s_m_validate_2, __pyx_k_m_validate_2, sizeof(__pyx_k_m_validate_2), 0, 0, 1, 1},
  {&__pyx_n_s_m_validator, __pyx_k_m_validator, sizeof(__pyx_k_m_validator), 0, 0, 1, 1},
  {&__pyx_n_s_m_validator_2, __pyx_k_m_validator_2, sizeof(__pyx_k_m_validator_2), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_map, __pyx_k_map, sizeof(__pyx_k_map), 0, 0, 1, 1},
  {&__pyx_n_s_mark_index, __pyx_k_mark_index, sizeof(__pyx_k_mark_index), 0, 0, 1, 1},
  {&__pyx_n_s_mark_key, __pyx_k_mark_key, sizeof(__pyx_k_mark_key), 0, 0, 1, 1},
  {&__pyx_n_s_match, __pyx_k_match, sizeof(__pyx_k_match), 0, 0, 1, 1},
  {&__pyx_kp_u_match_regex_s_compile_failed, __pyx_k_match_regex_s_compile_failed, sizeof(__pyx_k_match_regex_s_compile_failed), 0, 1, 0, 0},
  {&__pyx_n_s_max, __pyx_k_max, sizeof(__pyx_k_max), 0, 0, 1, 1},
  {&__pyx_n_s_max_repr, __pyx_k_max_repr, sizeof(__pyx_k_max_repr), 0, 0, 1, 1},
  {&__pyx_n_s_max_value, __pyx_k_max_value, sizeof(__pyx_k_max_value), 0, 0, 1, 1},
  {&__pyx_n_s_maxlen, __pyx_k_maxlen, sizeof(__pyx_k_maxlen), 0, 0, 1, 1},
  {&__pyx_n_s_maxsize, __pyx_k_maxsize, sizeof(__pyx_k_maxsize), 0, 0, 1, 1},
  {&__pyx_n_s_message, __pyx_k_message, sizeof(__pyx_k_message), 0, 0, 1, 1},
  {&__pyx_n_s_min, __pyx_k_min, sizeof(__pyx_k_min), 0, 0, 1, 1},
  {&__pyx_n_s_min_repr, __pyx_k_min_repr, sizeof(__pyx_k_min_repr), 0, 0, 1, 1},
  {&__pyx_n_s_min_value, __pyx_k_min_value, sizeof(__pyx_k_min_value), 0, 0, 1, 1},
  {&__pyx_n_s_minlen, __pyx_k_minlen, sizeof(__pyx_k_minlen), 0, 0, 1, 1},
  {&__pyx_n_s_mktime, __pyx_k_mktime, sizeof(__pyx_k_mktime), 0, 0, 1, 1},
  {&__pyx_n_u_model, __pyx_k_model, sizeof(__pyx_k_model), 0, 1, 0, 1},
  {&__pyx_kp_u_model_class_not_provided, __pyx_k_model_class_not_provided, sizeof(__pyx_k_model_class_not_provided), 0, 1, 0, 0},
  {&__pyx_n_s_model_validator, __pyx_k_model_validator, sizeof(__pyx_k_model_validator), 0, 0, 1, 1},
  {&__pyx_n_s_model_validator_locals_validate, __pyx_k_model_validator_locals_validate, sizeof(__pyx_k_model_validator_locals_validate), 0, 0, 1, 1},
  {&__pyx_n_s_module, __pyx_k_module, sizeof(__pyx_k_module), 0, 0, 1, 1},
  {&__pyx_n_s_msg, __pyx_k_msg, sizeof(__pyx_k_msg), 0, 0, 1, 1},
  {&__pyx_kp_u_must_be_dict, __pyx_k_must_be_dict, sizeof(__pyx_k_must_be_dict), 0, 1, 0, 0},
  {&__pyx_kp_u_must_be_dict_or_model_schema, __pyx_k_must_be_dict_or_model_schema, sizeof(__pyx_k_must_be_dict_or_model_schema), 0, 1, 0, 0},
  {&__pyx_n_u_n, __pyx_k_n, sizeof(__pyx_k_n), 0, 1, 0, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
  {&__pyx_n_u_no, __pyx_k_no, sizeof(__pyx_k_no), 0, 1, 0, 1},
  {&__pyx_kp_u_no_accept_hints, __pyx_k_no_accept_hints, sizeof(__pyx_k_no_accept_hints), 0, 1, 0, 0},
  {&__pyx_kp_u_no_output_hints, __pyx_k_no_output_hints, sizeof(__pyx_k_no_output_hints), 0, 1, 0, 0},
  {&__pyx_kp_u_not_allowed_by_argument_for_unio, __pyx_k_not_allowed_by_argument_for_unio, sizeof(__pyx_k_not_allowed_by_argument_for_unio), 0, 1, 0, 0},
  {&__pyx_kp_u_not_allowed_default_for_union_sc, __pyx_k_not_allowed_default_for_union_sc, sizeof(__pyx_k_not_allowed_default_for_union_sc), 0, 1, 0, 0},
  {&__pyx_kp_u_not_allowed_dict, __pyx_k_not_allowed_dict, sizeof(__pyx_k_not_allowed_dict), 0, 1, 0, 0},
  {&__pyx_kp_u_not_allowed_list, __pyx_k_not_allowed_list, sizeof(__pyx_k_not_allowed_list), 0, 1, 0, 0},
  {&__pyx_kp_u_not_allowed_optional_or_default, __pyx_k_not_allowed_optional_or_default, sizeof(__pyx_k_not_allowed_optional_or_default), 0, 1, 0, 0},
  {&__pyx_kp_u_not_allowed_scalar_value, __pyx_k_not_allowed_scalar_value, sizeof(__pyx_k_not_allowed_scalar_value), 0, 1, 0, 0},
  {&__pyx_kp_u_not_list, __pyx_k_not_list, sizeof(__pyx_k_not_list), 0, 1, 0, 0},
  {&__pyx_kp_u_not_unique, __pyx_k_not_unique, sizeof(__pyx_k_not_unique), 0, 1, 0, 0},
  {&__pyx_n_u_null, __pyx_k_null, sizeof(__pyx_k_null), 0, 1, 0, 1},
  {&__pyx_n_s_null_output, __pyx_k_null_output, sizeof(__pyx_k_null_output), 0, 0, 1, 1},
  {&__pyx_n_s_obj, __pyx_k_obj, sizeof(__pyx_k_obj), 0, 0, 1, 1},
  {&__pyx_n_s_object, __pyx_k_object, sizeof(__pyx_k_object), 0, 0, 1, 1},
  {&__pyx_n_u_object, __pyx_k_object, sizeof(__pyx_k_object), 0, 1, 0, 1},
  {&__pyx_n_u_off, __pyx_k_off, sizeof(__pyx_k_off), 0, 1, 0, 1},
  {&__pyx_n_u_on, __pyx_k_on, sizeof(__pyx_k_on), 0, 1, 0, 1},
  {&__pyx_n_s_optional, __pyx_k_optional, sizeof(__pyx_k_optional), 0, 0, 1, 1},
  {&__pyx_n_u_optional, __pyx_k_optional, sizeof(__pyx_k_optional), 0, 1, 0, 1},
  {&__pyx_n_s_options, __pyx_k_options, sizeof(__pyx_k_options), 0, 0, 1, 1},
  {&__pyx_n_s_origin_func, __pyx_k_origin_func, sizeof(__pyx_k_origin_func), 0, 0, 1, 1},
  {&__pyx_n_s_output, __pyx_k_output, sizeof(__pyx_k_output), 0, 0, 1, 1},
  {&__pyx_n_s_output_hints, __pyx_k_output_hints, sizeof(__pyx_k_output_hints), 0, 0, 1, 1},
  {&__pyx_n_s_output_object, __pyx_k_output_object, sizeof(__pyx_k_output_object), 0, 0, 1, 1},
  {&__pyx_n_u_output_object, __pyx_k_output_object, sizeof(__pyx_k_output_object), 0, 1, 0, 1},
  {&__pyx_n_s_output_string, __pyx_k_output_string, sizeof(__pyx_k_output_string), 0, 0, 1, 1},
  {&__pyx_n_s_params, __pyx_k_params, sizeof(__pyx_k_params), 0, 0, 1, 1},
  {&__pyx_n_s_parse_hints, __pyx_k_parse_hints, sizeof(__pyx_k_parse_hints), 0, 0, 1, 1},
  {&__pyx_n_s_parsed, __pyx_k_parsed, sizeof(__pyx_k_parsed), 0, 0, 1, 1},
  {&__pyx_n_s_partial, __pyx_k_partial, sizeof(__pyx_k_partial), 0, 0, 1, 1},
  {&__pyx_n_u_phone, __pyx_k_phone, sizeof(__pyx_k_phone), 0, 1, 0, 1},
  {&__pyx_n_s_pop, __pyx_k_pop, sizeof(__pyx_k_pop), 0, 0, 1, 1},
  {&__pyx_n_s_prefix, __pyx_k_prefix, sizeof(__pyx_k_prefix), 0, 0, 1, 1},
  {&__pyx_n_s_qualname, __pyx_k_qualname, sizeof(__pyx_k_qualname), 0, 0, 1, 1},
  {&__pyx_n_u_qualname, __pyx_k_qualname, sizeof(__pyx_k_qualname), 0, 1, 0, 1},
  {&__pyx_n_s_r, __pyx_k_r, sizeof(__pyx_k_r), 0, 0, 1, 1},
  {&__pyx_n_s_re, __pyx_k_re, sizeof(__pyx_k_re), 0, 0, 1, 1},
  {&__pyx_n_s_re_match, __pyx_k_re_match, sizeof(__pyx_k_re_match), 0, 0, 1, 1},
  {&__pyx_n_s_re_validator, __pyx_k_re_validator, sizeof(__pyx_k_re_validator), 0, 0, 1, 1},
  {&__pyx_n_s_regexs, __pyx_k_regexs, sizeof(__pyx_k_regexs), 0, 0, 1, 1},
  {&__pyx_n_s_relative, __pyx_k_relative, sizeof(__pyx_k_relative), 0, 0, 1, 1},
  {&__pyx_n_s_replace, __pyx_k_replace, sizeof(__pyx_k_replace), 0, 0, 1, 1},
  {&__pyx_n_s_repr, __pyx_k_repr, sizeof(__pyx_k_repr), 0, 0, 1, 1},
  {&__pyx_kp_u_require_string_value, __pyx_k_require_string_value, sizeof(__pyx_k_require_string_value), 0, 1, 0, 0},
  {&__pyx_n_u_required, __pyx_k_required, sizeof(__pyx_k_required), 0, 1, 0, 1},
  {&__pyx_kp_u_required_by_argument_for_union_d, __pyx_k_required_by_argument_for_union_d, sizeof(__pyx_k_required_by_argument_for_union_d), 0, 1, 0, 0},
  {&__pyx_n_s_result, __pyx_k_result, sizeof(__pyx_k_result), 0, 0, 1, 1},
  {&__pyx_n_u_return, __pyx_k_return, sizeof(__pyx_k_return), 0, 1, 0, 1},
  {&__pyx_n_s_return_typing, __pyx_k_return_typing, sizeof(__pyx_k_return_typing), 0, 0, 1, 1},
  {&__pyx_kp_u_s_Z, __pyx_k_s_Z, sizeof(__pyx_k_s_Z), 0, 1, 0, 0},
  {&__pyx_n_s_scalar_inner, __pyx_k_scalar_inner, sizeof(__pyx_k_scalar_inner), 0, 0, 1, 1},
  {&__pyx_n_s_schema, __pyx_k_schema, sizeof(__pyx_k_schema), 0, 0, 1, 1},
  {&__pyx_n_s_schema_2, __pyx_k_schema_2, sizeof(__pyx_k_schema_2), 0, 0, 1, 1},
  {&__pyx_n_s_scheme, __pyx_k_scheme, sizeof(__pyx_k_scheme), 0, 0, 1, 1},
  {&__pyx_n_s_seconds, __pyx_k_seconds, sizeof(__pyx_k_seconds), 0, 0, 1, 1},
  {&__pyx_kp_u_self_key, __pyx_k_self_key, sizeof(__pyx_k_self_key), 0, 1, 0, 0},
  {&__pyx_kp_u_self_value, __pyx_k_self_value, sizeof(__pyx_k_self_value), 0, 1, 0, 0},
  {&__pyx_n_s_send, __pyx_k_send, sizeof(__pyx_k_send), 0, 0, 1, 1},
  {&__pyx_n_s_set_value, __pyx_k_set_value, sizeof(__pyx_k_set_value), 0, 0, 1, 1},
  {&__pyx_n_s_slim, __pyx_k_slim, sizeof(__pyx_k_slim), 0, 0, 1, 1},
  {&__pyx_n_u_slug, __pyx_k_slug, sizeof(__pyx_k_slug), 0, 1, 0, 1},
  {&__pyx_kp_s_src_validr__validator_c_pyx, __pyx_k_src_validr__validator_c_pyx, sizeof(__pyx_k_src_validr__validator_c_pyx), 0, 0, 1, 0},
  {&__pyx_n_u_str, __pyx_k_str, sizeof(__pyx_k_str), 0, 1, 0, 1},
  {&__pyx_n_s_str_validator, __pyx_k_str_validator, sizeof(__pyx_k_str_validator), 0, 0, 1, 1},
  {&__pyx_n_s_str_validator_locals_validate, __pyx_k_str_validator_locals_validate, sizeof(__pyx_k_str_validator_locals_validate), 0, 0, 1, 1},
  {&__pyx_n_s_strftime, __pyx_k_strftime, sizeof(__pyx_k_strftime), 0, 0, 1, 1},
  {&__pyx_n_s_string, __pyx_k_string, sizeof(__pyx_k_string), 0, 0, 1, 1},
  {&__pyx_n_u_string, __pyx_k_string, sizeof(__pyx_k_string), 0, 1, 0, 1},
  {&__pyx_kp_u_string_length_must_d, __pyx_k_string_length_must_d, sizeof(__pyx_k_string_length_must_d), 0, 1, 0, 0},
  {&__pyx_kp_u_string_length_must_d_2, __pyx_k_string_length_must_d_2, sizeof(__pyx_k_string_length_must_d_2), 0, 1, 0, 0},
  {&__pyx_kp_u_string_not_match_regex_s, __pyx_k_string_not_match_regex_s, sizeof(__pyx_k_string_not_match_regex_s), 0, 1, 0, 0},
  {&__pyx_n_s_strip, __pyx_k_strip, sizeof(__pyx_k_strip), 0, 0, 1, 1},
  {&__pyx_n_s_strptime, __pyx_k_strptime, sizeof(__pyx_k_strptime), 0, 0, 1, 1},
  {&__pyx_n_s_supress_invalid, __pyx_k_supress_invalid, sizeof(__pyx_k_supress_invalid), 0, 0, 1, 1},
  {&__pyx_n_s_sys, __pyx_k_sys, sizeof(__pyx_k_sys), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_throw, __pyx_k_throw, sizeof(__pyx_k_throw), 0, 0, 1, 1},
  {&__pyx_n_s_time, __pyx_k_time, sizeof(__pyx_k_time), 0, 0, 1, 1},
  {&__pyx_n_u_time, __pyx_k_time, sizeof(__pyx_k_time), 0, 1, 0, 1},
  {&__pyx_n_s_time_validator, __pyx_k_time_validator, sizeof(__pyx_k_time_validator), 0, 0, 1, 1},
  {&__pyx_n_s_time_validator_locals_validate, __pyx_k_time_validator_locals_validate, sizeof(__pyx_k_time_validator_locals_validate), 0, 0, 1, 1},
  {&__pyx_n_s_timedelta, __pyx_k_timedelta, sizeof(__pyx_k_timedelta), 0, 0, 1, 1},
  {&__pyx_n_u_timedelta, __pyx_k_timedelta, sizeof(__pyx_k_timedelta), 0, 1, 0, 1},
  {&__pyx_n_s_timedelta_validator, __pyx_k_timedelta_validator, sizeof(__pyx_k_timedelta_validator), 0, 0, 1, 1},
  {&__pyx_n_s_timedelta_validator_locals_valid, __pyx_k_timedelta_validator_locals_valid, sizeof(__pyx_k_timedelta_validator_locals_valid), 0, 0, 1, 1},
  {&__pyx_n_s_to_str, __pyx_k_to_str, sizeof(__pyx_k_to_str), 0, 0, 1, 1},
  {&__pyx_n_s_total_seconds, __pyx_k_total_seconds, sizeof(__pyx_k_total_seconds), 0, 0, 1, 1},
  {&__pyx_n_u_true, __pyx_k_true, sizeof(__pyx_k_true), 0, 1, 0, 1},
  {&__pyx_n_s_types, __pyx_k_types, sizeof(__pyx_k_types), 0, 0, 1, 1},
  {&__pyx_n_s_typing, __pyx_k_typing, sizeof(__pyx_k_typing), 0, 0, 1, 1},
  {&__pyx_kp_u_unable_to_check_unique_for_non_h, __pyx_k_unable_to_check_unique_for_non_h, sizeof(__pyx_k_unable_to_check_unique_for_non_h), 0, 1, 0, 0},
  {&__pyx_n_u_union, __pyx_k_union, sizeof(__pyx_k_union), 0, 1, 0, 1},
  {&__pyx_n_s_union_dict_validator, __pyx_k_union_dict_validator, sizeof(__pyx_k_union_dict_validator), 0, 0, 1, 1},
  {&__pyx_n_s_union_dict_validator_locals_val, __pyx_k_union_dict_validator_locals_val, sizeof(__pyx_k_union_dict_validator_locals_val), 0, 0, 1, 1},
  {&__pyx_n_s_union_list_validator, __pyx_k_union_list_validator, sizeof(__pyx_k_union_list_validator), 0, 0, 1, 1},
  {&__pyx_n_s_union_list_validator_locals_val, __pyx_k_union_list_validator_locals_val, sizeof(__pyx_k_union_list_validator_locals_val), 0, 0, 1, 1},
  {&__pyx_kp_u_union_schemas_not_provided, __pyx_k_union_schemas_not_provided, sizeof(__pyx_k_union_schemas_not_provided), 0, 1, 0, 0},
  {&__pyx_kp_u_union_schemas_type_invalid, __pyx_k_union_schemas_type_invalid, sizeof(__pyx_k_union_schemas_type_invalid), 0, 1, 0, 0},
  {&__pyx_n_s_union_validator, __pyx_k_union_validator, sizeof(__pyx_k_union_validator), 0, 0, 1, 1},
  {&__pyx_n_s_unique, __pyx_k_unique, sizeof(__pyx_k_unique), 0, 0, 1, 1},
  {&__pyx_n_s_update_validate_func_type_hints, __pyx_k_update_validate_func_type_hints, sizeof(__pyx_k_update_validate_func_type_hints), 0, 0, 1, 1},
  {&__pyx_n_u_url, __pyx_k_url, sizeof(__pyx_k_url), 0, 1, 0, 1},
  {&__pyx_kp_u_url_length_must, __pyx_k_url_length_must, sizeof(__pyx_k_url_length_must), 0, 1, 0, 0},
  {&__pyx_n_s_url_validator, __pyx_k_url_validator, sizeof(__pyx_k_url_validator), 0, 0, 1, 1},
  {&__pyx_n_s_url_validator_locals_validate, __pyx_k_url_validator_locals_validate, sizeof(__pyx_k_url_validator_locals_validate), 0, 0, 1, 1},
  {&__pyx_n_s_urllib_parse, __pyx_k_urllib_parse, sizeof(__pyx_k_urllib_parse), 0, 0, 1, 1},
  {&__pyx_n_s_urlparse, __pyx_k_urlparse, sizeof(__pyx_k_urlparse), 0, 0, 1, 1},
  {&__pyx_n_s_urlunparse, __pyx_k_urlunparse, sizeof(__pyx_k_urlunparse), 0, 0, 1, 1},
  {&__pyx_n_s_uuid, __pyx_k_uuid, sizeof(__pyx_k_uuid), 0, 0, 1, 1},
  {&__pyx_n_u_uuid, __pyx_k_uuid, sizeof(__pyx_k_uuid), 0, 1, 0, 1},
  {&__pyx_n_s_uuid_validator, __pyx_k_uuid_validator, sizeof(__pyx_k_uuid_validator), 0, 0, 1, 1},
  {&__pyx_n_s_uuid_validator_locals_validate, __pyx_k_uuid_validator_locals_validate, sizeof(__pyx_k_uuid_validator_locals_validate), 0, 0, 1, 1},
  {&__pyx_n_s_v, __pyx_k_v, sizeof(__pyx_k_v), 0, 0, 1, 1},
  {&__pyx_n_s_validate, __pyx_k_validate, sizeof(__pyx_k_validate), 0, 0, 1, 1},
  {&__pyx_n_s_validate_2, __pyx_k_validate_2, sizeof(__pyx_k_validate_2), 0, 0, 1, 1},
  {&__pyx_n_s_validate_email, __pyx_k_validate_email, sizeof(__pyx_k_validate_email), 0, 0, 1, 1},
  {&__pyx_n_s_validate_extra_key, __pyx_k_validate_extra_key, sizeof(__pyx_k_validate_extra_key), 0, 0, 1, 1},
  {&__pyx_n_s_validate_extra_value, __pyx_k_validate_extra_value, sizeof(__pyx_k_validate_extra_value), 0, 0, 1, 1},
  {&__pyx_n_s_validate_func, __pyx_k_validate_func, sizeof(__pyx_k_validate_func), 0, 0, 1, 1},
  {&__pyx_n_s_validator, __pyx_k_validator, sizeof(__pyx_k_validator), 0, 0, 1, 1},
  {&__pyx_n_u_validator_2, __pyx_k_validator_2, sizeof(__pyx_k_validator_2), 0, 1, 0, 1},
  {&__pyx_n_s_validator_locals_decorator, __pyx_k_validator_locals_decorator, sizeof(__pyx_k_validator_locals_decorator), 0, 0, 1, 1},
  {&__pyx_n_s_validator_locals_decorator_local, __pyx_k_validator_locals_decorator_local, sizeof(__pyx_k_validator_locals_decorator_local), 0, 0, 1, 1},
  {&__pyx_n_s_validator_locals_decorator_local_2, __pyx_k_validator_locals_decorator_local_2, sizeof(__pyx_k_validator_locals_decorator_local_2), 0, 0, 1, 1},
  {&__pyx_n_s_validator_locals_decorator_local_3, __pyx_k_validator_locals_decorator_local_3, sizeof(__pyx_k_validator_locals_decorator_local_3), 0, 0, 1, 1},
  {&__pyx_n_s_validator_locals_decorator_local_4, __pyx_k_validator_locals_decorator_local_4, sizeof(__pyx_k_validator_locals_decorator_local_4), 0, 0, 1, 1},
  {&__pyx_n_s_validr__validator_c, __pyx_k_validr__validator_c, sizeof(__pyx_k_validr__validator_c), 0, 0, 1, 1},
  {&__pyx_n_s_value, __pyx_k_value, sizeof(__pyx_k_value), 0, 0, 1, 1},
  {&__pyx_n_u_value, __pyx_k_value, sizeof(__pyx_k_value), 0, 1, 0, 1},
  {&__pyx_kp_u_value_must, __pyx_k_value_must, sizeof(__pyx_k_value_must), 0, 1, 0, 0},
  {&__pyx_kp_u_value_must_2, __pyx_k_value_must_2, sizeof(__pyx_k_value_must_2), 0, 1, 0, 0},
  {&__pyx_kp_u_value_must_be_string, __pyx_k_value_must_be_string, sizeof(__pyx_k_value_must_be_string), 0, 1, 0, 0},
  {&__pyx_kp_u_value_must_d, __pyx_k_value_must_d, sizeof(__pyx_k_value_must_d), 0, 1, 0, 0},
  {&__pyx_kp_u_value_must_d_2, __pyx_k_value_must_d_2, sizeof(__pyx_k_value_must_d_2), 0, 1, 0, 0},
  {&__pyx_kp_u_value_must_d_3, __pyx_k_value_must_d_3, sizeof(__pyx_k_value_must_d_3), 0, 1, 0, 0},
  {&__pyx_kp_u_value_must_d_4, __pyx_k_value_must_d_4, sizeof(__pyx_k_value_must_d_4), 0, 1, 0, 0},
  {&__pyx_n_s_value_typing, __pyx_k_value_typing, sizeof(__pyx_k_value_typing), 0, 0, 1, 1},
  {&__pyx_n_s_vendor, __pyx_k_vendor, sizeof(__pyx_k_vendor), 0, 0, 1, 1},
  {&__pyx_n_s_vendor_email_validator, __pyx_k_vendor_email_validator, sizeof(__pyx_k_vendor_email_validator), 0, 0, 1, 1},
  {&__pyx_n_s_vendor_fqdn, __pyx_k_vendor_fqdn, sizeof(__pyx_k_vendor_fqdn), 0, 0, 1, 1},
  {&__pyx_n_s_version, __pyx_k_version, sizeof(__pyx_k_version), 0, 0, 1, 1},
  {&__pyx_n_s_x, __pyx_k_x, sizeof(__pyx_k_x), 0, 0, 1, 1},
  {&__pyx_n_u_y, __pyx_k_y, sizeof(__pyx_k_y), 0, 1, 0, 1},
  {&__pyx_n_u_yes, __pyx_k_yes, sizeof(__pyx_k_yes), 0, 1, 0, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_object = __Pyx_GetBuiltinName(__pyx_n_s_object); if (!__pyx_builtin_object) __PYX_ERR(0, 423, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 134, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(0, 180, __pyx_L1_error)
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_n_s_enumerate); if (!__pyx_builtin_enumerate) __PYX_ERR(0, 328, __pyx_L1_error)
  __pyx_builtin_map = __Pyx_GetBuiltinName(__pyx_n_s_map); if (!__pyx_builtin_map) __PYX_ERR(0, 406, __pyx_L1_error)
  __pyx_builtin_AttributeError = __Pyx_GetBuiltinName(__pyx_n_s_AttributeError); if (!__pyx_builtin_AttributeError) __PYX_ERR(0, 972, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "validr/_validator_c.pyx":153
 *             cdef bint local_output_object = output_object
 *             if output_string and output_object:
 *                 local_output_object = bool(params.get('object', None))             # <<<<<<<<<<<<<<
 *                 # TODO: fix special case of timedelta
 *                 if schema.validator == 'timedelta':
 */
  __pyx_tuple__5 = PyTuple_Pack(2, __pyx_n_u_object, Py_None); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(0, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "validr/_validator_c.pyx":156
 *                 # TODO: fix special case of timedelta
 *                 if schema.validator == 'timedelta':
 *                     local_output_object = not bool(params.get('string', None))             # <<<<<<<<<<<<<<
 *             if output_object and 'object' in params:
 *                 params['output_object'] = bool(params.pop('object', None))
 */
  __pyx_tuple__6 = PyTuple_Pack(2, __pyx_n_u_string, Py_None); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(0, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "validr/_validator_c.pyx":163
 *             else:
 *                 null_output = ''
 *             cdef bint optional = params.pop('optional', False)             # <<<<<<<<<<<<<<
 *             default = params.pop('default', None)
 *             params.pop('desc', None)
 */
  __pyx_tuple__7 = PyTuple_Pack(2, __pyx_n_u_optional, Py_False); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(0, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "validr/_validator_c.pyx":164
 *                 null_output = ''
 *             cdef bint optional = params.pop('optional', False)
 *             default = params.pop('default', None)             # <<<<<<<<<<<<<<
 *             params.pop('desc', None)
 *             cdef bint invalid_to_default = params.pop('invalid_to_default', False)
 */
  __pyx_tuple__8 = PyTuple_Pack(2, __pyx_n_u_default, Py_None); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(0, 164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "validr/_validator_c.pyx":165
 *             cdef bint optional = params.pop('optional', False)
 *             default = params.pop('default', None)
 *             params.pop('desc', None)             # <<<<<<<<<<<<<<
 *             cdef bint invalid_to_default = params.pop('invalid_to_default', False)
 *             cdef bint has_invalid_to = 'invalid_to' in params
 */
  __pyx_tuple__9 = PyTuple_Pack(2, __pyx_n_u_desc, Py_None); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "validr/_validator_c.pyx":166
 *             default = params.pop('default', None)
 *             params.pop('desc', None)
 *             cdef bint invalid_to_default = params.pop('invalid_to_default', False)             # <<<<<<<<<<<<<<
 *             cdef bint has_invalid_to = 'invalid_to' in params
 *             invalid_to = params.pop('invalid_to', None)
 */
  __pyx_tuple__10 = PyTuple_Pack(2, __pyx_n_u_invalid_to_default, Py_False); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(0, 166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "validr/_validator_c.pyx":168
 *             cdef bint invalid_to_default = params.pop('invalid_to_default', False)
 *             cdef bint has_invalid_to = 'invalid_to' in params
 *             invalid_to = params.pop('invalid_to', None)             # <<<<<<<<<<<<<<
 *             cdef bint has_default
 *             if accept_string:
 */
  __pyx_tuple__11 = PyTuple_Pack(2, __pyx_n_u_invalid_to, Py_None); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(0, 168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);

  /* "validr/_validator_c.pyx":203
 * 
 *             # check null, empty string and default
 *             def _m_validate(value):             # <<<<<<<<<<<<<<
 *                 cdef bint is_null
 *                 if accept_string:
 */
  __pyx_tuple__12 = PyTuple_Pack(2, __pyx_n_s_value, __pyx_n_s_is_null); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(0, 203, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);
  __pyx_codeobj__13 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__12, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_m_validate, 203, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__13)) __PYX_ERR(0, 203, __pyx_L1_error)

  /* "validr/_validator_c.pyx":235
 *             supress_invalid = has_invalid_to or invalid_to_default
 * 
 *             def m_validate(value):             # <<<<<<<<<<<<<<
 *                 try:
 *                     return _m_validate(value)
 */
  __pyx_tuple__14 = PyTuple_Pack(2, __pyx_n_s_value, __pyx_n_s_ex); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(0, 235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);
  __pyx_codeobj__15 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__14, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_m_validate_2, 235, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__15)) __PYX_ERR(0, 235, __pyx_L1_error)

  /* "validr/_validator_c.pyx":147
 *     def decorator(f):
 * 
 *         def _m_validator(compiler, schema):             # <<<<<<<<<<<<<<
 *             params = schema.params.copy()
 *             if schema.items is not None:
 */
  __pyx_tuple__16 = PyTuple_Pack(19, __pyx_n_s_compiler, __pyx_n_s_schema_2, __pyx_n_s_params, __pyx_n_s_local_output_object, __pyx_n_s_null_output, __pyx_n_s_optional, __pyx_n_s_default, __pyx_n_s_invalid_to_default, __pyx_n_s_has_invalid_to, __pyx_n_s_invalid_to, __pyx_n_s_has_default, __pyx_n_s_validate, __pyx_n_s_e, __pyx_n_s_msg, __pyx_n_s_m_validate, __pyx_n_s_m_validate, __pyx_n_s_supress_invalid, __pyx_n_s_m_validate_2, __pyx_n_s_m_validate_2); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);
  __pyx_codeobj__17 = (PyObject*)__Pyx_PyCode_New(2, 0, 19, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__16, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_m_validator, 147, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__17)) __PYX_ERR(0, 147, __pyx_L1_error)

  /* "validr/_validator_c.pyx":253
 *             return m_validate
 * 
 *         def m_validator(compiler, schema):             # <<<<<<<<<<<<<<
 *             try:
 *                 return _m_validator(compiler, schema)
 */
  __pyx_tuple__18 = PyTuple_Pack(3, __pyx_n_s_compiler, __pyx_n_s_schema_2, __pyx_n_s_ex); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);
  __pyx_codeobj__19 = (PyObject*)__Pyx_PyCode_New(2, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__18, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_m_validator_2, 253, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__19)) __PYX_ERR(0, 253, __pyx_L1_error)

  /* "validr/_validator_c.pyx":145
 *     del string, accept, output
 * 
 *     def decorator(f):             # <<<<<<<<<<<<<<
 * 
 *         def _m_validator(compiler, schema):
 */
  __pyx_tuple__21 = PyTuple_Pack(5, __pyx_n_s_f, __pyx_n_s_m_validator, __pyx_n_s_m_validator, __pyx_n_s_m_validator_2, __pyx_n_s_m_validator_2); if (unlikely(!__pyx_tuple__21)) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__21);
  __Pyx_GIVEREF(__pyx_tuple__21);
  __pyx_codeobj__22 = (PyObject*)__Pyx_PyCode_New(1, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__21, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_decorator, 145, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__22)) __PYX_ERR(0, 145, __pyx_L1_error)

  /* "validr/_validator_c.pyx":296
 *             keys.append((k, _key_func_of_schema(v)))
 * 
 *         def key_of(dict v):             # <<<<<<<<<<<<<<
 *             cdef str k
 *             return tuple(key_of_value(v[k]) for k, key_of_value in keys)
 */
  __pyx_tuple__24 = PyTuple_Pack(4, __pyx_n_s_v, __pyx_n_s_k, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__24)) __PYX_ERR(0, 296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__24);
  __Pyx_GIVEREF(__pyx_tuple__24);
  __pyx_codeobj__25 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__24, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_key_of, 296, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__25)) __PYX_ERR(0, 296, __pyx_L1_error)

  /* "validr/_validator_c.pyx":305
 *         key_of_value = _key_func_of_schema(schema.items)
 * 
 *         def key_of(list v):             # <<<<<<<<<<<<<<
 *             return tuple(key_of_value(x) for x in v)
 * 
 */
  __pyx_tuple__26 = PyTuple_Pack(3, __pyx_n_s_v, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(0, 305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__26);
  __Pyx_GIVEREF(__pyx_tuple__26);
  __pyx_codeobj__27 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_key_of, 305, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__27)) __PYX_ERR(0, 305, __pyx_L1_error)

  /* "validr/_validator_c.pyx":338
 *             if i >= maxlen:
 *                 raise Invalid('list length must <= %d' % maxlen)
 *             with mark_index(i):             # <<<<<<<<<<<<<<
 *                 v = inner(x) if inner is not None else copy(x)
 *                 if unique:
 */
  __pyx_tuple__28 = PyTuple_Pack(3, Py_None, Py_None, Py_None); if (unlikely(!__pyx_tuple__28)) __PYX_ERR(0, 338, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__28);
  __Pyx_GIVEREF(__pyx_tuple__28);

  /* "validr/_validator_c.pyx":326
 *     del compiler, items
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         try:
 *             value = enumerate(value)
 */
  __pyx_tuple__30 = PyTuple_Pack(7, __pyx_n_s_value, __pyx_n_s_result, __pyx_n_s_keys, __pyx_n_s_i, __pyx_n_s_x, __pyx_n_s_v, __pyx_n_s_k); if (unlikely(!__pyx_tuple__30)) __PYX_ERR(0, 326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__30);
  __Pyx_GIVEREF(__pyx_tuple__30);
  __pyx_codeobj__31 = (PyObject*)__Pyx_PyCode_New(1, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__30, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_validate, 326, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__31)) __PYX_ERR(0, 326, __pyx_L1_error)

  /* "validr/_validator_c.pyx":377
 *     del compiler, items, key, value
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         if inners is None and not is_dynamic:
 *             if not is_dict(value):
 */
  __pyx_tuple__33 = PyTuple_Pack(7, __pyx_n_s_value, __pyx_n_s_getter, __pyx_n_s_result, __pyx_n_s_k, __pyx_n_s_inner, __pyx_n_s_extra_keys, __pyx_n_s_v); if (unlikely(!__pyx_tuple__33)) __PYX_ERR(0, 377, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__33);
  __Pyx_GIVEREF(__pyx_tuple__33);
  __pyx_codeobj__34 = (PyObject*)__Pyx_PyCode_New(1, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__33, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_validate, 377, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__34)) __PYX_ERR(0, 377, __pyx_L1_error)

  /* "validr/_validator_c.pyx":428
 *         raise SchemaError('model class not provided')
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         return items(value)
 * 
 */
  __pyx_tuple__36 = PyTuple_Pack(1, __pyx_n_s_value); if (unlikely(!__pyx_tuple__36)) __PYX_ERR(0, 428, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__36);
  __Pyx_GIVEREF(__pyx_tuple__36);
  __pyx_codeobj__37 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__36, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_validate, 428, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__37)) __PYX_ERR(0, 428, __pyx_L1_error)

  /* "validr/_validator_c.pyx":454
 *     items = frozenset(items)
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         if value in items:
 *             return value
 */
  __pyx_tuple__42 = PyTuple_Pack(1, __pyx_n_s_value); if (unlikely(!__pyx_tuple__42)) __PYX_ERR(0, 454, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__42);
  __Pyx_GIVEREF(__pyx_tuple__42);
  __pyx_codeobj__43 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__42, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_validate, 454, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__43)) __PYX_ERR(0, 454, __pyx_L1_error)

  /* "validr/_validator_c.pyx":517
 *                 scalar_inner = compiler.compile(inner_schema)
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         if isinstance(value, list):
 *             if list_inner is None:
 */
  __pyx_tuple__45 = PyTuple_Pack(1, __pyx_n_s_value); if (unlikely(!__pyx_tuple__45)) __PYX_ERR(0, 517, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__45);
  __Pyx_GIVEREF(__pyx_tuple__45);
  __pyx_codeobj__46 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__45, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_validate, 517, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__46)) __PYX_ERR(0, 517, __pyx_L1_error)

  /* "validr/_validator_c.pyx":561
 *             by_name = getter(value, by)
 *             if not by_name:
 *                 raise Invalid('required', value=by_name)             # <<<<<<<<<<<<<<
 *             inner_info = inners.get(by_name)
 *             if inner_info is None:
 */
  __pyx_tuple__47 = PyTuple_Pack(1, __pyx_n_u_required); if (unlikely(!__pyx_tuple__47)) __PYX_ERR(0, 561, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__47);
  __Pyx_GIVEREF(__pyx_tuple__47);

  /* "validr/_validator_c.pyx":552
 *     del key, schema, is_model
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         if is_dict(value):
 *             getter = get_dict_value
 */
  __pyx_tuple__49 = PyTuple_Pack(8, __pyx_n_s_value, __pyx_n_s_getter, __pyx_n_s_by_name, __pyx_n_s_inner_info, __pyx_n_s_err_msg, __pyx_n_s_is_model, __pyx_n_s_inner, __pyx_n_s_result); if (unlikely(!__pyx_tuple__49)) __PYX_ERR(0, 552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__49);
  __Pyx_GIVEREF(__pyx_tuple__49);
  __pyx_codeobj__50 = (PyObject*)__Pyx_PyCode_New(1, 0, 8, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__49, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_validate, 552, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__50)) __PYX_ERR(0, 552, __pyx_L1_error)

  /* "validr/_validator_c.pyx":595
 *     min, max = int(min), int(max)
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         try:
 *             v = int(value)
 */
  __pyx_tuple__55 = PyTuple_Pack(2, __pyx_n_s_value, __pyx_n_s_v); if (unlikely(!__pyx_tuple__55)) __PYX_ERR(0, 595, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__55);
  __Pyx_GIVEREF(__pyx_tuple__55);
  __pyx_codeobj__56 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__55, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_validate, 595, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__56)) __PYX_ERR(0, 595, __pyx_L1_error)

  /* "validr/_validator_c.pyx":625
 * def bool_validator(compiler):
 *     """Validate bool"""
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         if value in _TRUE_VALUES:
 *             return True
 */
  __pyx_tuple__58 = PyTuple_Pack(1, __pyx_n_s_value); if (unlikely(!__pyx_tuple__58)) __PYX_ERR(0, 625, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__58);
  __Pyx_GIVEREF(__pyx_tuple__58);
  __pyx_codeobj__59 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__58, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_validate, 625, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__59)) __PYX_ERR(0, 625, __pyx_L1_error)

  /* "validr/_validator_c.pyx":658
 *         exmax = bool(exmax)
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         try:
 *             v = float(value)
 */
  __pyx_tuple__63 = PyTuple_Pack(2, __pyx_n_s_value, __pyx_n_s_v); if (unlikely(!__pyx_tuple__63)) __PYX_ERR(0, 658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__63);
  __Pyx_GIVEREF(__pyx_tuple__63);
  __pyx_codeobj__64 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__63, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_validate, 658, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__64)) __PYX_ERR(0, 658, __pyx_L1_error)

  /* "validr/_validator_c.pyx":716
 *             raise Invalid('string length must <= %d' % maxlen)
 *         if escape:
 *             value = (value.replace('&', '&amp;')             # <<<<<<<<<<<<<<
 *                      .replace('>', '&gt;')
 *                      .replace('<', '&lt;')
 */
  __pyx_tuple__66 = PyTuple_Pack(2, __pyx_kp_u__65, __pyx_kp_u_amp); if (unlikely(!__pyx_tuple__66)) __PYX_ERR(0, 716, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__66);
  __Pyx_GIVEREF(__pyx_tuple__66);

  /* "validr/_validator_c.pyx":717
 *         if escape:
 *             value = (value.replace('&', '&amp;')
 *                      .replace('>', '&gt;')             # <<<<<<<<<<<<<<
 *                      .replace('<', '&lt;')
 *                      .replace("'", '&#39;')
 */
  __pyx_tuple__68 = PyTuple_Pack(2, __pyx_kp_u__67, __pyx_kp_u_gt); if (unlikely(!__pyx_tuple__68)) __PYX_ERR(0, 717, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__68);
  __Pyx_GIVEREF(__pyx_tuple__68);

  /* "validr/_validator_c.pyx":718
 *             value = (value.replace('&', '&amp;')
 *                      .replace('>', '&gt;')
 *                      .replace('<', '&lt;')             # <<<<<<<<<<<<<<
 *                      .replace("'", '&#39;')
 *                      .replace('"', '&#34;'))
 */
  __pyx_tuple__70 = PyTuple_Pack(2, __pyx_kp_u__69, __pyx_kp_u_lt); if (unlikely(!__pyx_tuple__70)) __PYX_ERR(0, 718, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__70);
  __Pyx_GIVEREF(__pyx_tuple__70);

  /* "validr/_validator_c.pyx":719
 *                      .replace('>', '&gt;')
 *                      .replace('<', '&lt;')
 *                      .replace("'", '&#39;')             # <<<<<<<<<<<<<<
 *                      .replace('"', '&#34;'))
 *         if re_match is not None and not re_match(value):
 */
  __pyx_tuple__72 = PyTuple_Pack(2, __pyx_kp_u__71, __pyx_kp_u_39); if (unlikely(!__pyx_tuple__72)) __PYX_ERR(0, 719, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__72);
  __Pyx_GIVEREF(__pyx_tuple__72);

  /* "validr/_validator_c.pyx":720
 *                      .replace('<', '&lt;')
 *                      .replace("'", '&#39;')
 *                      .replace('"', '&#34;'))             # <<<<<<<<<<<<<<
 *         if re_match is not None and not re_match(value):
 *             raise Invalid("string not match regex %s" % match)
 */
  __pyx_tuple__74 = PyTuple_Pack(2, __pyx_kp_u__73, __pyx_kp_u_34); if (unlikely(!__pyx_tuple__74)) __PYX_ERR(0, 720, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__74);
  __Pyx_GIVEREF(__pyx_tuple__74);

  /* "validr/_validator_c.pyx":702
 *         re_match = None
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         if not isinstance(value, str):
 *             if accept_object or isinstance(value, int):
 */
  __pyx_tuple__76 = PyTuple_Pack(2, __pyx_n_s_value, __pyx_n_s_length); if (unlikely(!__pyx_tuple__76)) __PYX_ERR(0, 702, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__76);
  __Pyx_GIVEREF(__pyx_tuple__76);
  __pyx_codeobj__77 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__76, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_validate, 702, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__77)) __PYX_ERR(0, 702, __pyx_L1_error)

  /* "validr/_validator_c.pyx":736
 *     """
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         if not isinstance(value, bytes):
 *             raise Invalid('invalid bytes')
 */
  __pyx_tuple__79 = PyTuple_Pack(2, __pyx_n_s_value, __pyx_n_s_length); if (unlikely(!__pyx_tuple__79)) __PYX_ERR(0, 736, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__79);
  __Pyx_GIVEREF(__pyx_tuple__79);
  __pyx_codeobj__80 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__79, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_validate, 736, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__80)) __PYX_ERR(0, 736, __pyx_L1_error)

  /* "validr/_validator_c.pyx":756
 *         format (str): date format, default ISO8601 format
 *     """
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         try:
 *             if not isinstance(value, (datetime.datetime, datetime.date)):
 */
  __pyx_tuple__82 = PyTuple_Pack(1, __pyx_n_s_value); if (unlikely(!__pyx_tuple__82)) __PYX_ERR(0, 756, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__82);
  __Pyx_GIVEREF(__pyx_tuple__82);
  __pyx_codeobj__83 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__82, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_validate, 756, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__83)) __PYX_ERR(0, 756, __pyx_L1_error)

  /* "validr/_validator_c.pyx":778
 *         format (str): time format, default ISO8601 format
 *     """
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         try:
 *             if not isinstance(value, (datetime.datetime, datetime.time)):
 */
  __pyx_tuple__85 = PyTuple_Pack(1, __pyx_n_s_value); if (unlikely(!__pyx_tuple__85)) __PYX_ERR(0, 778, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__85);
  __Pyx_GIVEREF(__pyx_tuple__85);
  __pyx_codeobj__86 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__85, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_validate, 778, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__86)) __PYX_ERR(0, 778, __pyx_L1_error)

  /* "validr/_validator_c.pyx":800
 *         format (str): datetime format, default ISO8601 format
 *     """
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         try:
 *             if isinstance(value, tuple):
 */
  __pyx_tuple__88 = PyTuple_Pack(1, __pyx_n_s_value); if (unlikely(!__pyx_tuple__88)) __PYX_ERR(0, 800, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__88);
  __Pyx_GIVEREF(__pyx_tuple__88);
  __pyx_codeobj__89 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__88, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_validate, 800, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__89)) __PYX_ERR(0, 800, __pyx_L1_error)

  /* "validr/_validator_c.pyx":822
 *     else:
 *         if not isinstance(value, datetime.timedelta):
 *             raise ValueError("invalid timedelta")             # <<<<<<<<<<<<<<
 *     return value
 * 
 */
  __pyx_tuple__90 = PyTuple_Pack(1, __pyx_kp_u_invalid_timedelta); if (unlikely(!__pyx_tuple__90)) __PYX_ERR(0, 822, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__90);
  __Pyx_GIVEREF(__pyx_tuple__90);

  /* "validr/_validator_c.pyx":859
 *         max_repr = durationpy.to_str(max_value, extended=True)
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         try:
 *             value = _parse_timedelta(value)
 */
  __pyx_tuple__92 = PyTuple_Pack(2, __pyx_n_s_value, __pyx_n_s_ex); if (unlikely(!__pyx_tuple__92)) __PYX_ERR(0, 859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__92);
  __Pyx_GIVEREF(__pyx_tuple__92);
  __pyx_codeobj__93 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__92, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_validate, 859, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__93)) __PYX_ERR(0, 859, __pyx_L1_error)

  /* "validr/_validator_c.pyx":882
 * @validator(accept=(str, ipaddress.IPv4Address), output=(str, ipaddress.IPv4Address))
 * def ipv4_validator(compiler, bint output_object=False):
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         try:
 *             value = ipaddress.IPv4Address(value.strip())
 */
  __pyx_tuple__95 = PyTuple_Pack(2, __pyx_n_s_value, __pyx_n_s_ex); if (unlikely(!__pyx_tuple__95)) __PYX_ERR(0, 882, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__95);
  __Pyx_GIVEREF(__pyx_tuple__95);
  __pyx_codeobj__96 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__95, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_validate, 882, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__96)) __PYX_ERR(0, 882, __pyx_L1_error)

  /* "validr/_validator_c.pyx":898
 * @validator(accept=(str, ipaddress.IPv6Address), output=(str, ipaddress.IPv6Address))
 * def ipv6_validator(compiler, bint output_object=False):
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         try:
 *             value = ipaddress.IPv6Address(value.strip())
 */
  __pyx_tuple__98 = PyTuple_Pack(2, __pyx_n_s_value, __pyx_n_s_ex); if (unlikely(!__pyx_tuple__98)) __PYX_ERR(0, 898, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__98);
  __Pyx_GIVEREF(__pyx_tuple__98);
  __pyx_codeobj__99 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__98, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_validate, 898, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__99)) __PYX_ERR(0, 898, __pyx_L1_error)

  /* "validr/_validator_c.pyx":924
 *     )
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         try:
 *             value = _validate(value.strip())
 */
  __pyx_tuple__101 = PyTuple_Pack(2, __pyx_n_s_value, __pyx_n_s_ex); if (unlikely(!__pyx_tuple__101)) __PYX_ERR(0, 924, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__101);
  __Pyx_GIVEREF(__pyx_tuple__101);
  __pyx_codeobj__102 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__101, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_validate, 924, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__102)) __PYX_ERR(0, 924, __pyx_L1_error)

  /* "validr/_validator_c.pyx":944
 *     allow_scheme = set(scheme.split())
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         try:
 *             value = value.strip()
 */
  __pyx_tuple__104 = PyTuple_Pack(2, __pyx_n_s_value, __pyx_n_s_parsed); if (unlikely(!__pyx_tuple__104)) __PYX_ERR(0, 944, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__104);
  __Pyx_GIVEREF(__pyx_tuple__104);
  __pyx_codeobj__105 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__104, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_validate, 944, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__105)) __PYX_ERR(0, 944, __pyx_L1_error)

  /* "validr/_validator_c.pyx":966
 * @validator(output=str)
 * def fqdn_validator(compiler):
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         try:
 *             value = value.strip()
 */
  __pyx_tuple__107 = PyTuple_Pack(3, __pyx_n_s_value, __pyx_n_s_fqdn_obj, __pyx_n_s_ex); if (unlikely(!__pyx_tuple__107)) __PYX_ERR(0, 966, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__107);
  __Pyx_GIVEREF(__pyx_tuple__107);
  __pyx_codeobj__108 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__107, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_validate, 966, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__108)) __PYX_ERR(0, 966, __pyx_L1_error)

  /* "validr/_validator_c.pyx":987
 *         msg = 'invalid uuid{}'.format(version)
 * 
 *     def validate(value):             # <<<<<<<<<<<<<<
 *         if not isinstance(value, uuid.UUID):
 *             try:
 */
  __pyx_tuple__110 = PyTuple_Pack(1, __pyx_n_s_value); if (unlikely(!__pyx_tuple__110)) __PYX_ERR(0, 987, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__110);
  __Pyx_GIVEREF(__pyx_tuple__110);
  __pyx_codeobj__111 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__110, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_validate, 987, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__111)) __PYX_ERR(0, 987, __pyx_L1_error)

  /* "validr/_validator_c.pyx":1017
 * 
 *     def re_validator(compiler, int minlen=0, int maxlen=default_maxlen, bint strip=False):
 *         def validate(value):             # <<<<<<<<<<<<<<
 *             if not isinstance(value, str):
 *                 raise Invalid('value must be string')
 */
  __pyx_tuple__112 = PyTuple_Pack(2, __pyx_n_s_value, __pyx_n_s_length); if (unlikely(!__pyx_tuple__112)) __PYX_ERR(0, 1017, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__112);
  __Pyx_GIVEREF(__pyx_tuple__112);
  __pyx_codeobj__113 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__112, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_validate, 1017, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__113)) __PYX_ERR(0, 1017, __pyx_L1_error)

  /* "validr/_validator_c.pyx":1016
 *     message = 'invalid %s' % name
 * 
 *     def re_validator(compiler, int minlen=0, int maxlen=default_maxlen, bint strip=False):             # <<<<<<<<<<<<<<
 *         def validate(value):
 *             if not isinstance(value, str):
 */
  __pyx_tuple__115 = PyTuple_Pack(6, __pyx_n_s_compiler, __pyx_n_s_minlen, __pyx_n_s_maxlen, __pyx_n_s_strip, __pyx_n_s_validate, __pyx_n_s_validate); if (unlikely(!__pyx_tuple__115)) __PYX_ERR(0, 1016, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__115);
  __Pyx_GIVEREF(__pyx_tuple__115);
  __pyx_codeobj__116 = (PyObject*)__Pyx_PyCode_New(4, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__115, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_re_validator, 1016, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__116)) __PYX_ERR(0, 1016, __pyx_L1_error)

  /* "validr/_validator_c.pyx":1084
 * 
 *     def enum_validator(compiler):
 *         def validate(value):             # <<<<<<<<<<<<<<
 *             if value in items:
 *                 return value
 */
  __pyx_tuple__117 = PyTuple_Pack(1, __pyx_n_s_value); if (unlikely(!__pyx_tuple__117)) __PYX_ERR(0, 1084, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__117);
  __Pyx_GIVEREF(__pyx_tuple__117);
  __pyx_codeobj__118 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__117, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_validate, 1084, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__118)) __PYX_ERR(0, 1084, __pyx_L1_error)

  /* "validr/_validator_c.pyx":1083
 *     message = 'invalid {}, expect one of {}'.format(name, list(sorted(items)))
 * 
 *     def enum_validator(compiler):             # <<<<<<<<<<<<<<
 *         def validate(value):
 *             if value in items:
 */
  __pyx_tuple__120 = PyTuple_Pack(3, __pyx_n_s_compiler, __pyx_n_s_validate, __pyx_n_s_validate); if (unlikely(!__pyx_tuple__120)) __PYX_ERR(0, 1083, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__120);
  __Pyx_GIVEREF(__pyx_tuple__120);
  __pyx_codeobj__121 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__120, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_enum_validator, 1083, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__121)) __PYX_ERR(0, 1083, __pyx_L1_error)

  /* "validr/_validator_c.pyx":51
 * 
 * 
 * def _parse_hints(hints):             # <<<<<<<<<<<<<<
 *     if not isinstance(hints, (tuple, set, list)):
 *         hints = [hints]
 */
  __pyx_tuple__122 = PyTuple_Pack(5, __pyx_n_s_hints, __pyx_n_s_is_string, __pyx_n_s_is_object, __pyx_n_s_types, __pyx_n_s_hint); if (unlikely(!__pyx_tuple__122)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__122);
  __Pyx_GIVEREF(__pyx_tuple__122);
  __pyx_codeobj__3 = (PyObject*)__Pyx_PyCode_New(1, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__122, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_parse_hints, 51, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__3)) __PYX_ERR(0, 51, __pyx_L1_error)

  /* "validr/_validator_c.pyx":68
 * 
 * 
 * def _update_validate_func_type_hints(             # <<<<<<<<<<<<<<
 *         validate_func,
 *         optional, has_default,
 */
  __pyx_tuple__123 = PyTuple_Pack(8, __pyx_n_s_validate_func, __pyx_n_s_optional, __pyx_n_s_has_default, __pyx_n_s_accept_hints, __pyx_n_s_output_hints, __pyx_n_s_value_typing, __pyx_n_s_return_typing, __pyx_n_s_annotations_2); if (unlikely(!__pyx_tuple__123)) __PYX_ERR(0, 68, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__123);
  __Pyx_GIVEREF(__pyx_tuple__123);
  __pyx_codeobj__4 = (PyObject*)__Pyx_PyCode_New(5, 0, 8, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__123, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_update_validate_func_type_hints, 68, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__4)) __PYX_ERR(0, 68, __pyx_L1_error)

  /* "validr/_validator_c.pyx":91
 * 
 * 
 * def validator(string=None, *, accept=None, output=None):             # <<<<<<<<<<<<<<
 *     """Decorator for create validator
 * 
 */
  __pyx_tuple__124 = PyTuple_Pack(11, __pyx_n_s_string, __pyx_n_s_accept, __pyx_n_s_output, __pyx_n_s_accept_string, __pyx_n_s_accept_object, __pyx_n_s_output_string, __pyx_n_s_output_object, __pyx_n_s_accept_hints, __pyx_n_s_output_hints, __pyx_n_s_decorator, __pyx_n_s_decorator); if (unlikely(!__pyx_tuple__124)) __PYX_ERR(0, 91, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__124);
  __Pyx_GIVEREF(__pyx_tuple__124);
  __pyx_codeobj__20 = (PyObject*)__Pyx_PyCode_New(1, 2, 11, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__124, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_validator, 91, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__20)) __PYX_ERR(0, 91, __pyx_L1_error)

  /* "validr/_validator_c.pyx":285
 * 
 * 
 * def _key_func_of_schema(schema):             # <<<<<<<<<<<<<<
 *     if schema is None:
 *         raise SchemaError(_UNIQUE_CHECK_ERROR_MESSAGE)
 */
  __pyx_tuple__125 = PyTuple_Pack(8, __pyx_n_s_schema_2, __pyx_n_s_keys, __pyx_n_s_k, __pyx_n_s_v, __pyx_n_s_key_of, __pyx_n_s_key_of, __pyx_n_s_key_of_value, __pyx_n_s_key_of); if (unlikely(!__pyx_tuple__125)) __PYX_ERR(0, 285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__125);
  __Pyx_GIVEREF(__pyx_tuple__125);
  __pyx_codeobj__23 = (PyObject*)__Pyx_PyCode_New(1, 0, 8, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__125, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_key_func_of_schema, 285, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__23)) __PYX_ERR(0, 285, __pyx_L1_error)

  /* "validr/_validator_c.pyx":315
 * 
 * @validator(accept=typing.Iterable, output=typing.List)
 * def list_validator(compiler, items=None, int minlen=0, int maxlen=1024,             # <<<<<<<<<<<<<<
 *                    bint unique=False):
 *     if items is None:
 */
  __pyx_tuple__126 = PyTuple_Pack(9, __pyx_n_s_compiler, __pyx_n_s_items, __pyx_n_s_minlen, __pyx_n_s_maxlen, __pyx_n_s_unique, __pyx_n_s_inner, __pyx_n_s_key_of, __pyx_n_s_validate, __pyx_n_s_validate); if (unlikely(!__pyx_tuple__126)) __PYX_ERR(0, 315, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__126);
  __Pyx_GIVEREF(__pyx_tuple__126);
  __pyx_codeobj__29 = (PyObject*)__Pyx_PyCode_New(5, 0, 9, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__126, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_list_validator, 315, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__29)) __PYX_ERR(0, 315, __pyx_L1_error)

  /* "validr/_validator_c.pyx":357
 * 
 * @validator(accept=(typing.Mapping, typing.Any), output=dict)
 * def dict_validator(compiler, items=None, key=None, value=None,             # <<<<<<<<<<<<<<
 *                    int minlen=0, int maxlen=1024, bint slim=False):
 *     if items is None:
 */
  __pyx_tuple__127 = PyTuple_Pack(15, __pyx_n_s_compiler, __pyx_n_s_items, __pyx_n_s_key, __pyx_n_s_value, __pyx_n_s_minlen, __pyx_n_s_maxlen, __pyx_n_s_slim, __pyx_n_s_inners, __pyx_n_s_k, __pyx_n_s_v, __pyx_n_s_validate_extra_key, __pyx_n_s_validate_extra_value, __pyx_n_s_is_dynamic, __pyx_n_s_validate, __pyx_n_s_validate); if (unlikely(!__pyx_tuple__127)) __PYX_ERR(0, 357, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__127);
  __Pyx_GIVEREF(__pyx_tuple__127);
  __pyx_codeobj__32 = (PyObject*)__Pyx_PyCode_New(7, 0, 15, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__127, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_dict_validator, 357, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__32)) __PYX_ERR(0, 357, __pyx_L1_error)

  /* "validr/_validator_c.pyx":424
 * 
 * @validator(accept=(typing.Mapping, typing.Any), output=object)
 * def model_validator(compiler, items=None):             # <<<<<<<<<<<<<<
 *     if items is None:
 *         raise SchemaError('model class not provided')
 */
  __pyx_tuple__128 = PyTuple_Pack(4, __pyx_n_s_compiler, __pyx_n_s_items, __pyx_n_s_validate, __pyx_n_s_validate); if (unlikely(!__pyx_tuple__128)) __PYX_ERR(0, 424, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__128);
  __Pyx_GIVEREF(__pyx_tuple__128);
  __pyx_codeobj__35 = (PyObject*)__Pyx_PyCode_New(2, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__128, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_model_validator, 424, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__35)) __PYX_ERR(0, 424, __pyx_L1_error)

  /* "validr/_validator_c.pyx":448
 * 
 * @validator(output=object)
 * def enum_validator(compiler, items):             # <<<<<<<<<<<<<<
 *     if not items:
 *         raise SchemaError("enum items not provided")
 */
  __pyx_tuple__129 = PyTuple_Pack(5, __pyx_n_s_compiler, __pyx_n_s_items, __pyx_n_s_expects, __pyx_n_s_validate, __pyx_n_s_validate); if (unlikely(!__pyx_tuple__129)) __PYX_ERR(0, 448, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__129);
  __Pyx_GIVEREF(__pyx_tuple__129);
  __pyx_codeobj__38 = (PyObject*)__Pyx_PyCode_New(2, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__129, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_enum_validator, 448, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__38)) __PYX_ERR(0, 448, __pyx_L1_error)

  /* "validr/_validator_c.pyx":491
 * 
 * 
 * def _union_list_validator(compiler, schema):             # <<<<<<<<<<<<<<
 *     scalar_inner = None
 *     list_inner = None
 */
  __pyx_tuple__130 = PyTuple_Pack(9, __pyx_n_s_compiler, __pyx_n_s_schema_2, __pyx_n_s_scalar_inner, __pyx_n_s_list_inner, __pyx_n_s_dict_inner, __pyx_n_s_i, __pyx_n_s_inner_schema, __pyx_n_s_validate, __pyx_n_s_validate); if (unlikely(!__pyx_tuple__130)) __PYX_ERR(0, 491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__130);
  __Pyx_GIVEREF(__pyx_tuple__130);
  __pyx_codeobj__44 = (PyObject*)__Pyx_PyCode_New(2, 0, 9, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__130, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_union_list_validator, 491, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__44)) __PYX_ERR(0, 491, __pyx_L1_error)

  /* "validr/_validator_c.pyx":539
 * 
 * @validator(accept=object, output=object)
 * def _union_dict_validator(compiler, items, str by):             # <<<<<<<<<<<<<<
 *     inners = {}
 *     for key, schema in items.items():
 */
  __pyx_tuple__131 = PyTuple_Pack(10, __pyx_n_s_compiler, __pyx_n_s_items, __pyx_n_s_by, __pyx_n_s_inners, __pyx_n_s_key, __pyx_n_s_schema_2, __pyx_n_s_is_model, __pyx_n_s_expect_bys, __pyx_n_s_validate, __pyx_n_s_validate); if (unlikely(!__pyx_tuple__131)) __PYX_ERR(0, 539, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__131);
  __Pyx_GIVEREF(__pyx_tuple__131);
  __pyx_codeobj__48 = (PyObject*)__Pyx_PyCode_New(3, 0, 10, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__131, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_union_dict_validator, 539, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__48)) __PYX_ERR(0, 539, __pyx_L1_error)

  /* "validr/_validator_c.pyx":580
 * 
 * @validator(accept=object, output=object)
 * def any_validator(compiler, **ignore_kwargs):             # <<<<<<<<<<<<<<
 *     """Accept any value"""
 *     return any_validate
 */
  __pyx_tuple__132 = PyTuple_Pack(2, __pyx_n_s_compiler, __pyx_n_s_ignore_kwargs); if (unlikely(!__pyx_tuple__132)) __PYX_ERR(0, 580, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__132);
  __Pyx_GIVEREF(__pyx_tuple__132);
  __pyx_codeobj__51 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_any_validator, 580, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__51)) __PYX_ERR(0, 580, __pyx_L1_error)

  /* "validr/_validator_c.pyx":586
 * 
 * @validator(accept=(int, float, str), output=int)
 * def int_validator(compiler, min=-sys.maxsize, max=sys.maxsize):             # <<<<<<<<<<<<<<
 *     """Validate int or convert string to int
 * 
 */
  __pyx_tuple__133 = PyTuple_Pack(5, __pyx_n_s_compiler, __pyx_n_s_min, __pyx_n_s_max, __pyx_n_s_validate, __pyx_n_s_validate); if (unlikely(!__pyx_tuple__133)) __PYX_ERR(0, 586, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__133);
  __Pyx_GIVEREF(__pyx_tuple__133);
  __pyx_codeobj__54 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__133, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_int_validator, 586, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__54)) __PYX_ERR(0, 586, __pyx_L1_error)

  /* "validr/_validator_c.pyx":623
 * 
 * @validator(accept=(bool, int, str), output=bool)
 * def bool_validator(compiler):             # <<<<<<<<<<<<<<
 *     """Validate bool"""
 *     def validate(value):
 */
  __pyx_tuple__134 = PyTuple_Pack(3, __pyx_n_s_compiler, __pyx_n_s_validate, __pyx_n_s_validate); if (unlikely(!__pyx_tuple__134)) __PYX_ERR(0, 623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__134);
  __Pyx_GIVEREF(__pyx_tuple__134);
  __pyx_codeobj__57 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__134, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_bool_validator, 623, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__57)) __PYX_ERR(0, 623, __pyx_L1_error)

  /* "validr/_validator_c.pyx":636
 * 
 * @validator(accept=(int, float, str), output=float)
 * def float_validator(compiler, min=-sys.float_info.max, max=sys.float_info.max,             # <<<<<<<<<<<<<<
 *                     exmin=False, exmax=False):
 *     """Validate float string
 */
  __pyx_tuple__135 = PyTuple_Pack(7, __pyx_n_s_compiler, __pyx_n_s_min, __pyx_n_s_max, __pyx_n_s_exmin, __pyx_n_s_exmax, __pyx_n_s_validate, __pyx_n_s_validate); if (unlikely(!__pyx_tuple__135)) __PYX_ERR(0, 636, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__135);
  __Pyx_GIVEREF(__pyx_tuple__135);
  __pyx_codeobj__62 = (PyObject*)__Pyx_PyCode_New(5, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__135, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_float_validator, 636, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__62)) __PYX_ERR(0, 636, __pyx_L1_error)

  /* "validr/_validator_c.pyx":680
 * 
 * @validator(accept=(str, object), output=str)
 * def str_validator(compiler, int minlen=0, int maxlen=1024 * 1024,             # <<<<<<<<<<<<<<
 *                   bint strip=False, bint escape=False, str match=None,
 *                   bint accept_object=False):
 */
  __pyx_tuple__136 = PyTuple_Pack(11, __pyx_n_s_compiler, __pyx_n_s_minlen, __pyx_n_s_maxlen, __pyx_n_s_strip, __pyx_n_s_escape, __pyx_n_s_match, __pyx_n_s_accept_object, __pyx_n_s_re_match, __pyx_n_s_ex, __pyx_n_s_validate, __pyx_n_s_validate); if (unlikely(!__pyx_tuple__136)) __PYX_ERR(0, 680, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__136);
  __Pyx_GIVEREF(__pyx_tuple__136);
  __pyx_codeobj__75 = (PyObject*)__Pyx_PyCode_New(7, 0, 11, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__136, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_str_validator, 680, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__75)) __PYX_ERR(0, 680, __pyx_L1_error)

  /* "validr/_validator_c.pyx":728
 * 
 * @validator(accept=bytes, output=bytes)
 * def bytes_validator(compiler, int minlen=0, int maxlen=-1):             # <<<<<<<<<<<<<<
 *     """Validate bytes
 * 
 */
  __pyx_tuple__137 = PyTuple_Pack(5, __pyx_n_s_compiler, __pyx_n_s_minlen, __pyx_n_s_maxlen, __pyx_n_s_validate, __pyx_n_s_validate); if (unlikely(!__pyx_tuple__137)) __PYX_ERR(0, 728, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__137);
  __Pyx_GIVEREF(__pyx_tuple__137);
  __pyx_codeobj__78 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_bytes_validator, 728, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__78)) __PYX_ERR(0, 728, __pyx_L1_error)

  /* "validr/_validator_c.pyx":750
 * 
 * @validator(accept=(str, datetime.date), output=(str, datetime.date))
 * def date_validator(compiler, str format='%Y-%m-%d', bint output_object=False):             # <<<<<<<<<<<<<<
 *     """Validate date string or convert date to string
 * 
 */
  __pyx_tuple__138 = PyTuple_Pack(5, __pyx_n_s_compiler, __pyx_n_s_format, __pyx_n_s_output_object, __pyx_n_s_validate, __pyx_n_s_validate); if (unlikely(!__pyx_tuple__138)) __PYX_ERR(0, 750, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__138);
  __Pyx_GIVEREF(__pyx_tuple__138);
  __pyx_codeobj__81 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__138, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_date_validator, 750, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__81)) __PYX_ERR(0, 750, __pyx_L1_error)

  /* "validr/_validator_c.pyx":772
 * 
 * @validator(accept=(str, datetime.time), output=(str, datetime.time))
 * def time_validator(compiler, str format='%H:%M:%S', bint output_object=False):             # <<<<<<<<<<<<<<
 *     """Validate time string or convert time to string
 * 
 */
  __pyx_tuple__139 = PyTuple_Pack(5, __pyx_n_s_compiler, __pyx_n_s_format, __pyx_n_s_output_object, __pyx_n_s_validate, __pyx_n_s_validate); if (unlikely(!__pyx_tuple__139)) __PYX_ERR(0, 772, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__139);
  __Pyx_GIVEREF(__pyx_tuple__139);
  __pyx_codeobj__84 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__139, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_time_validator, 772, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__84)) __PYX_ERR(0, 772, __pyx_L1_error)

  /* "validr/_validator_c.pyx":794
 * 
 * @validator(accept=(str, datetime.datetime), output=(str, datetime.datetime))
 * def datetime_validator(compiler, str format='%Y-%m-%dT%H:%M:%S.%fZ', bint output_object=False):             # <<<<<<<<<<<<<<
 *     """Validate datetime string or convert datetime to string
 * 
 */
  __pyx_tuple__140 = PyTuple_Pack(5, __pyx_n_s_compiler, __pyx_n_s_format, __pyx_n_s_output_object, __pyx_n_s_validate, __pyx_n_s_validate); if (unlikely(!__pyx_tuple__140)) __PYX_ERR(0, 794, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__140);
  __Pyx_GIVEREF(__pyx_tuple__140);
  __pyx_codeobj__87 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__140, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_datetime_validator, 794, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__87)) __PYX_ERR(0, 794, __pyx_L1_error)

  /* "validr/_validator_c.pyx":827
 * 
 * @validator(accept=(int, float, str, datetime.timedelta), output=(str, float, datetime.timedelta))
 * def timedelta_validator(compiler, min=None, max=None, bint string=False,             # <<<<<<<<<<<<<<
 *                         bint extended=False, bint output_object=False):
 *     """Validate timedelta string or convert timedelta to string
 */
  __pyx_tuple__141 = PyTuple_Pack(13, __pyx_n_s_compiler, __pyx_n_s_min, __pyx_n_s_max, __pyx_n_s_string, __pyx_n_s_extended, __pyx_n_s_output_object, __pyx_n_s_min_value, __pyx_n_s_ex, __pyx_n_s_max_value, __pyx_n_s_min_repr, __pyx_n_s_max_repr, __pyx_n_s_validate, __pyx_n_s_validate); if (unlikely(!__pyx_tuple__141)) __PYX_ERR(0, 827, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__141);
  __Pyx_GIVEREF(__pyx_tuple__141);
  __pyx_codeobj__91 = (PyObject*)__Pyx_PyCode_New(6, 0, 13, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__141, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_timedelta_validator, 827, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__91)) __PYX_ERR(0, 827, __pyx_L1_error)

  /* "validr/_validator_c.pyx":881
 * 
 * @validator(accept=(str, ipaddress.IPv4Address), output=(str, ipaddress.IPv4Address))
 * def ipv4_validator(compiler, bint output_object=False):             # <<<<<<<<<<<<<<
 *     def validate(value):
 *         try:
 */
  __pyx_tuple__142 = PyTuple_Pack(4, __pyx_n_s_compiler, __pyx_n_s_output_object, __pyx_n_s_validate, __pyx_n_s_validate); if (unlikely(!__pyx_tuple__142)) __PYX_ERR(0, 881, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__142);
  __Pyx_GIVEREF(__pyx_tuple__142);
  __pyx_codeobj__94 = (PyObject*)__Pyx_PyCode_New(2, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__142, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_ipv4_validator, 881, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__94)) __PYX_ERR(0, 881, __pyx_L1_error)

  /* "validr/_validator_c.pyx":897
 * 
 * @validator(accept=(str, ipaddress.IPv6Address), output=(str, ipaddress.IPv6Address))
 * def ipv6_validator(compiler, bint output_object=False):             # <<<<<<<<<<<<<<
 *     def validate(value):
 *         try:
 */
  __pyx_tuple__143 = PyTuple_Pack(4, __pyx_n_s_compiler, __pyx_n_s_output_object, __pyx_n_s_validate, __pyx_n_s_validate); if (unlikely(!__pyx_tuple__143)) __PYX_ERR(0, 897, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__143);
  __Pyx_GIVEREF(__pyx_tuple__143);
  __pyx_codeobj__97 = (PyObject*)__Pyx_PyCode_New(2, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__143, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_ipv6_validator, 897, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__97)) __PYX_ERR(0, 897, __pyx_L1_error)

  /* "validr/_validator_c.pyx":913
 * 
 * @validator(accept=str, output=str)
 * def email_validator(compiler):             # <<<<<<<<<<<<<<
 *     # https://stackoverflow.com/questions/201323/using-a-regular-expression-to-validate-an-email-address
 *     # http://emailregex.com/
 */
  __pyx_tuple__144 = PyTuple_Pack(4, __pyx_n_s_compiler, __pyx_n_s_validate_2, __pyx_n_s_validate, __pyx_n_s_validate); if (unlikely(!__pyx_tuple__144)) __PYX_ERR(0, 913, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__144);
  __Pyx_GIVEREF(__pyx_tuple__144);
  __pyx_codeobj__100 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__144, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_email_validator, 913, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__100)) __PYX_ERR(0, 913, __pyx_L1_error)

  /* "validr/_validator_c.pyx":936
 * 
 * @validator(output=(str, object))
 * def url_validator(compiler, str scheme='http https', int maxlen=255, bint output_object=False):             # <<<<<<<<<<<<<<
 *     # https://stackoverflow.com/questions/7160737/python-how-to-validate-a-url-in-python-malformed-or-not
 *     # https://stackoverflow.com/questions/827557/how-do-you-validate-a-url-with-a-regular-expression-in-python
 */
  __pyx_tuple__145 = PyTuple_Pack(7, __pyx_n_s_compiler, __pyx_n_s_scheme, __pyx_n_s_maxlen, __pyx_n_s_output_object, __pyx_n_s_allow_scheme, __pyx_n_s_validate, __pyx_n_s_validate); if (unlikely(!__pyx_tuple__145)) __PYX_ERR(0, 936, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__145);
  __Pyx_GIVEREF(__pyx_tuple__145);
  __pyx_codeobj__103 = (PyObject*)__Pyx_PyCode_New(4, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__145, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_url_validator, 936, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__103)) __PYX_ERR(0, 936, __pyx_L1_error)

  /* "validr/_validator_c.pyx":965
 * 
 * @validator(output=str)
 * def fqdn_validator(compiler):             # <<<<<<<<<<<<<<
 *     def validate(value):
 *         try:
 */
  __pyx_tuple__146 = PyTuple_Pack(3, __pyx_n_s_compiler, __pyx_n_s_validate, __pyx_n_s_validate); if (unlikely(!__pyx_tuple__146)) __PYX_ERR(0, 965, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__146);
  __Pyx_GIVEREF(__pyx_tuple__146);
  __pyx_codeobj__106 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__146, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_fqdn_validator, 965, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__106)) __PYX_ERR(0, 965, __pyx_L1_error)

  /* "validr/_validator_c.pyx":979
 * 
 * @validator(output=(str, uuid.UUID))
 * def uuid_validator(compiler, version=None, bint output_object=False):             # <<<<<<<<<<<<<<
 *     if version is None:
 *         msg = 'invalid uuid'
 */
  __pyx_tuple__147 = PyTuple_Pack(6, __pyx_n_s_compiler, __pyx_n_s_version, __pyx_n_s_output_object, __pyx_n_s_msg, __pyx_n_s_validate, __pyx_n_s_validate); if (unlikely(!__pyx_tuple__147)) __PYX_ERR(0, 979, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__147);
  __Pyx_GIVEREF(__pyx_tuple__147);
  __pyx_codeobj__109 = (PyObject*)__Pyx_PyCode_New(3, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__147, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_uuid_validator, 979, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__109)) __PYX_ERR(0, 979, __pyx_L1_error)

  /* "validr/_validator_c.pyx":1002
 * 
 * 
 * def create_re_validator(str name, str r, int default_maxlen=255):             # <<<<<<<<<<<<<<
 *     """Create validator by regex string
 * 
 */
  __pyx_tuple__148 = PyTuple_Pack(7, __pyx_n_s_name_2, __pyx_n_s_r, __pyx_n_s_default_maxlen, __pyx_n_s_match, __pyx_n_s_message, __pyx_n_s_re_validator, __pyx_n_s_re_validator); if (unlikely(!__pyx_tuple__148)) __PYX_ERR(0, 1002, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__148);
  __Pyx_GIVEREF(__pyx_tuple__148);
  __pyx_codeobj__114 = (PyObject*)__Pyx_PyCode_New(3, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__148, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_create_re_validator, 1002, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__114)) __PYX_ERR(0, 1002, __pyx_L1_error)

  /* "validr/_validator_c.pyx":1062
 * 
 * regexs = {
 *     'phone': (r'((\+\d{2}\s?)|(\d{2}\s?))?1\d{10}', 15),             # <<<<<<<<<<<<<<
 *     'idcard': (r'(\d{17}[\d|x|X])|(\d{15})', 18),
 *     'slug': (r'[a-z0-9]+(?:-[a-z0-9]+)*', 255),
 */
  __pyx_tuple__149 = PyTuple_Pack(2, __pyx_kp_u_d_2_s_d_2_s_1_d_10, __pyx_int_15); if (unlikely(!__pyx_tuple__149)) __PYX_ERR(0, 1062, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__149);
  __Pyx_GIVEREF(__pyx_tuple__149);

  /* "validr/_validator_c.pyx":1063
 * regexs = {
 *     'phone': (r'((\+\d{2}\s?)|(\d{2}\s?))?1\d{10}', 15),
 *     'idcard': (r'(\d{17}[\d|x|X])|(\d{15})', 18),             # <<<<<<<<<<<<<<
 *     'slug': (r'[a-z0-9]+(?:-[a-z0-9]+)*', 255),
 * }
 */
  __pyx_tuple__150 = PyTuple_Pack(2, __pyx_kp_u_d_17_d_x_X_d_15, __pyx_int_18); if (unlikely(!__pyx_tuple__150)) __PYX_ERR(0, 1063, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__150);
  __Pyx_GIVEREF(__pyx_tuple__150);

  /* "validr/_validator_c.pyx":1064
 *     'phone': (r'((\+\d{2}\s?)|(\d{2}\s?))?1\d{10}', 15),
 *     'idcard': (r'(\d{17}[\d|x|X])|(\d{15})', 18),
 *     'slug': (r'[a-z0-9]+(?:-[a-z0-9]+)*', 255),             # <<<<<<<<<<<<<<
 * }
 * for name, options in regexs.items():
 */
  __pyx_tuple__151 = PyTuple_Pack(2, __pyx_kp_u_a_z0_9_a_z0_9, __pyx_int_255); if (unlikely(!__pyx_tuple__151)) __PYX_ERR(0, 1064, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__151);
  __Pyx_GIVEREF(__pyx_tuple__151);

  /* "validr/_validator_c.pyx":1070
 * 
 * 
 * def create_enum_validator(str name, items, bint string=True):             # <<<<<<<<<<<<<<
 *     """Create validator by enum items
 * 
 */
  __pyx_tuple__152 = PyTuple_Pack(6, __pyx_n_s_name_2, __pyx_n_s_items, __pyx_n_s_string, __pyx_n_s_message, __pyx_n_s_enum_validator, __pyx_n_s_enum_validator); if (unlikely(!__pyx_tuple__152)) __PYX_ERR(0, 1070, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__152);
  __Pyx_GIVEREF(__pyx_tuple__152);
  __pyx_codeobj__119 = (PyObject*)__Pyx_PyCode_New(3, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__152, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_validr__validator_c_pyx, __pyx_n_s_create_enum_validator, 1070, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__119)) __PYX_ERR(0, 1070, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  __pyx_umethod_PyDict_Type_get.type = (PyObject*)&PyDict_Type;
  __pyx_umethod_PyDict_Type_keys.type = (PyObject*)&PyDict_Type;
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_3 = PyInt_FromLong(3); if (unlikely(!__pyx_int_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_4 = PyInt_FromLong(4); if (unlikely(!__pyx_int_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_5 = PyInt_FromLong(5); if (unlikely(!__pyx_int_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_15 = PyInt_FromLong(15); if (unlikely(!__pyx_int_15)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_18 = PyInt_FromLong(18); if (unlikely(!__pyx_int_18)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_255 = PyInt_FromLong(255); if (unlikely(!__pyx_int_255)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __pyx_v_6validr_12_validator_c__UNIQUE_CHECK_ERROR_MESSAGE = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  if (PyType_Ready(&__pyx_type_6validr_12_validator_c___pyx_scope_struct__validator) < 0) __PYX_ERR(0, 91, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6validr_12_validator_c___pyx_scope_struct__validator.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6validr_12_validator_c___pyx_scope_struct__validator.tp_dictoffset && __pyx_type_6validr_12_validator_c___pyx_scope_struct__validator.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6validr_12_validator_c___pyx_scope_struct__validator.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6validr_12_validator_c___pyx_scope_struct__validator = &__pyx_type_6validr_12_validator_c___pyx_scope_struct__validator;
  if (PyType_Ready(&__pyx_type_6validr_12_validator_c___pyx_scope_struct_1_decorator) < 0) __PYX_ERR(0, 145, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6validr_12_validator_c___pyx_scope_struct_1_decorator.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6validr_12_validator_c___pyx_scope_struct_1_decorator.tp_dictoffset && __pyx_type_6validr_12_validator_c___pyx_scope_struct_1_decorator.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6validr_12_validator_c___pyx_scope_struct_1_decorator.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6validr_12_validator_c___pyx_scope_struct_1_decorator = &__pyx_type_6validr_12_validator_c___pyx_scope_struct_1_decorator;
  if (PyType_Ready(&__pyx_type_6validr_12_validator_c___pyx_scope_struct_2__m_validator) < 0) __PYX_ERR(0, 147, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6validr_12_validator_c___pyx_scope_struct_2__m_validator.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6validr_12_validator_c___pyx_scope_struct_2__m_validator.tp_dictoffset && __pyx_type_6validr_12_validator_c___pyx_scope_struct_2__m_validator.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6validr_12_validator_c___pyx_scope_struct_2__m_validator.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6validr_12_validator_c___pyx_scope_struct_2__m_validator = &__pyx_type_6validr_12_validator_c___pyx_scope_struct_2__m_validator;
  if (PyType_Ready(&__pyx_type_6validr_12_validator_c___pyx_scope_struct_3__key_func_of_schema) < 0) __PYX_ERR(0, 285, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6validr_12_validator_c___pyx_scope_struct_3__key_func_of_schema.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6validr_12_validator_c___pyx_scope_struct_3__key_func_of_schema.tp_dictoffset && __pyx_type_6validr_12_validator_c___pyx_scope_struct_3__key_func_of_schema.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6validr_12_validator_c___pyx_scope_struct_3__key_func_of_schema.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6validr_12_validator_c___pyx_scope_struct_3__key_func_of_schema = &__pyx_type_6validr_12_validator_c___pyx_scope_struct_3__key_func_of_schema;
  if (PyType_Ready(&__pyx_type_6validr_12_validator_c___pyx_scope_struct_4_key_of) < 0) __PYX_ERR(0, 296, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6validr_12_validator_c___pyx_scope_struct_4_key_of.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6validr_12_validator_c___pyx_scope_struct_4_key_of.tp_dictoffset && __pyx_type_6validr_12_validator_c___pyx_scope_struct_4_key_of.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6validr_12_validator_c___pyx_scope_struct_4_key_of.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6validr_12_validator_c___pyx_scope_struct_4_key_of = &__pyx_type_6validr_12_validator_c___pyx_scope_struct_4_key_of;
  if (PyType_Ready(&__pyx_type_6validr_12_validator_c___pyx_scope_struct_5_genexpr) < 0) __PYX_ERR(0, 298, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6validr_12_validator_c___pyx_scope_struct_5_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6validr_12_validator_c___pyx_scope_struct_5_genexpr.tp_dictoffset && __pyx_type_6validr_12_validator_c___pyx_scope_struct_5_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6validr_12_validator_c___pyx_scope_struct_5_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6validr_12_validator_c___pyx_scope_struct_5_genexpr = &__pyx_type_6validr_12_validator_c___pyx_scope_struct_5_genexpr;
  if (PyType_Ready(&__pyx_type_6validr_12_validator_c___pyx_scope_struct_6_key_of) < 0) __PYX_ERR(0, 305, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6validr_12_validator_c___pyx_scope_struct_6_key_of.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6validr_12_validator_c___pyx_scope_struct_6_key_of.tp_dictoffset && __pyx_type_6validr_12_validator_c___pyx_scope_struct_6_key_of.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6validr_12_validator_c___pyx_scope_struct_6_key_of.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6validr_12_validator_c___pyx_scope_struct_6_key_of = &__pyx_type_6validr_12_validator_c___pyx_scope_struct_6_key_of;
  if (PyType_Ready(&__pyx_type_6validr_12_validator_c___pyx_scope_struct_7_genexpr) < 0) __PYX_ERR(0, 306, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6validr_12_validator_c___pyx_scope_struct_7_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6validr_12_validator_c___pyx_scope_struct_7_genexpr.tp_dictoffset && __pyx_type_6validr_12_validator_c___pyx_scope_struct_7_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6validr_12_validator_c___pyx_scope_struct_7_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6validr_12_validator_c___pyx_scope_struct_7_genexpr = &__pyx_type_6validr_12_validator_c___pyx_scope_struct_7_genexpr;
  if (PyType_Ready(&__pyx_type_6validr_12_validator_c___pyx_scope_struct_8_list_validator) < 0) __PYX_ERR(0, 315, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6validr_12_validator_c___pyx_scope_struct_8_list_validator.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6validr_12_validator_c___pyx_scope_struct_8_list_validator.tp_dictoffset && __pyx_type_6validr_12_validator_c___pyx_scope_struct_8_list_validator.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6validr_12_validator_c___pyx_scope_struct_8_list_validator.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6validr_12_validator_c___pyx_scope_struct_8_list_validator = &__pyx_type_6validr_12_validator_c___pyx_scope_struct_8_list_validator;
  if (PyType_Ready(&__pyx_type_6validr_12_validator_c___pyx_scope_struct_9_dict_validator) < 0) __PYX_ERR(0, 357, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6validr_12_validator_c___pyx_scope_struct_9_dict_validator.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6validr_12_validator_c___pyx_scope_struct_9_dict_validator.tp_dictoffset && __pyx_type_6validr_12_validator_c___pyx_scope_struct_9_dict_validator.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6validr_12_validator_c___pyx_scope_struct_9_dict_validator.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6validr_12_validator_c___pyx_scope_struct_9_dict_validator = &__pyx_type_6validr_12_validator_c___pyx_scope_struct_9_dict_validator;
  if (PyType_Ready(&__pyx_type_6validr_12_validator_c___pyx_scope_struct_10_model_validator) < 0) __PYX_ERR(0, 424, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6validr_12_validator_c___pyx_scope_struct_10_model_validator.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6validr_12_validator_c___pyx_scope_struct_10_model_validator.tp_dictoffset && __pyx_type_6validr_12_validator_c___pyx_scope_struct_10_model_validator.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6validr_12_validator_c___pyx_scope_struct_10_model_validator.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6validr_12_validator_c___pyx_scope_struct_10_model_validator = &__pyx_type_6validr_12_validator_c___pyx_scope_struct_10_model_validator;
  if (PyType_Ready(&__pyx_type_6validr_12_validator_c___pyx_scope_struct_11_enum_validator) < 0) __PYX_ERR(0, 448, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6validr_12_validator_c___pyx_scope_struct_11_enum_validator.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6validr_12_validator_c___pyx_scope_struct_11_enum_validator.tp_dictoffset && __pyx_type_6validr_12_validator_c___pyx_scope_struct_11_enum_validator.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6validr_12_validator_c___pyx_scope_struct_11_enum_validator.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6validr_12_validator_c___pyx_scope_struct_11_enum_validator = &__pyx_type_6validr_12_validator_c___pyx_scope_struct_11_enum_validator;
  if (PyType_Ready(&__pyx_type_6validr_12_validator_c___pyx_scope_struct_12__union_list_validator) < 0) __PYX_ERR(0, 491, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6validr_12_validator_c___pyx_scope_struct_12__union_list_validator.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6validr_12_validator_c___pyx_scope_struct_12__union_list_validator.tp_dictoffset && __pyx_type_6validr_12_validator_c___pyx_scope_struct_12__union_list_validator.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6validr_12_validator_c___pyx_scope_struct_12__union_list_validator.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6validr_12_validator_c___pyx_scope_struct_12__union_list_validator = &__pyx_type_6validr_12_validator_c___pyx_scope_struct_12__union_list_validator;
  if (PyType_Ready(&__pyx_type_6validr_12_validator_c___pyx_scope_struct_13__union_dict_validator) < 0) __PYX_ERR(0, 539, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6validr_12_validator_c___pyx_scope_struct_13__union_dict_validator.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6validr_12_validator_c___pyx_scope_struct_13__union_dict_validator.tp_dictoffset && __pyx_type_6validr_12_validator_c___pyx_scope_struct_13__union_dict_validator.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6validr_12_validator_c___pyx_scope_struct_13__union_dict_validator.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6validr_12_validator_c___pyx_scope_struct_13__union_dict_validator = &__pyx_type_6validr_12_validator_c___pyx_scope_struct_13__union_dict_validator;
  if (PyType_Ready(&__pyx_type_6validr_12_validator_c___pyx_scope_struct_14_int_validator) < 0) __PYX_ERR(0, 586, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6validr_12_validator_c___pyx_scope_struct_14_int_validator.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6validr_12_validator_c___pyx_scope_struct_14_int_validator.tp_dictoffset && __pyx_type_6validr_12_validator_c___pyx_scope_struct_14_int_validator.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6validr_12_validator_c___pyx_scope_struct_14_int_validator.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6validr_12_validator_c___pyx_scope_struct_14_int_validator = &__pyx_type_6validr_12_validator_c___pyx_scope_struct_14_int_validator;
  if (PyType_Ready(&__pyx_type_6validr_12_validator_c___pyx_scope_struct_15_float_validator) < 0) __PYX_ERR(0, 636, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6validr_12_validator_c___pyx_scope_struct_15_float_validator.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6validr_12_validator_c___pyx_scope_struct_15_float_validator.tp_dictoffset && __pyx_type_6validr_12_validator_c___pyx_scope_struct_15_float_validator.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6validr_12_validator_c___pyx_scope_struct_15_float_validator.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6validr_12_validator_c___pyx_scope_struct_15_float_validator = &__pyx_type_6validr_12_validator_c___pyx_scope_struct_15_float_validator;
  if (PyType_Ready(&__pyx_type_6validr_12_validator_c___pyx_scope_struct_16_str_validator) < 0) __PYX_ERR(0, 680, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6validr_12_validator_c___pyx_scope_struct_16_str_validator.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6validr_12_validator_c___pyx_scope_struct_16_str_validator.tp_dictoffset && __pyx_type_6validr_12_validator_c___pyx_scope_struct_16_str_validator.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6validr_12_validator_c___pyx_scope_struct_16_str_validator.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6validr_12_validator_c___pyx_scope_struct_16_str_validator = &__pyx_type_6validr_12_validator_c___pyx_scope_struct_16_str_validator;
  if (PyType_Ready(&__pyx_type_6validr_12_validator_c___pyx_scope_struct_17_bytes_validator) < 0) __PYX_ERR(0, 728, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6validr_12_validator_c___pyx_scope_struct_17_bytes_validator.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6validr_12_validator_c___pyx_scope_struct_17_bytes_validator.tp_dictoffset && __pyx_type_6validr_12_validator_c___pyx_scope_struct_17_bytes_validator.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6validr_12_validator_c___pyx_scope_struct_17_bytes_validator.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6validr_12_validator_c___pyx_scope_struct_17_bytes_validator = &__pyx_type_6validr_12_validator_c___pyx_scope_struct_17_bytes_validator;
  if (PyType_Ready(&__pyx_type_6validr_12_validator_c___pyx_scope_struct_18_date_validator) < 0) __PYX_ERR(0, 750, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6validr_12_validator_c___pyx_scope_struct_18_date_validator.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6validr_12_validator_c___pyx_scope_struct_18_date_validator.tp_dictoffset && __pyx_type_6validr_12_validator_c___pyx_scope_struct_18_date_validator.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6validr_12_validator_c___pyx_scope_struct_18_date_validator.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6validr_12_validator_c___pyx_scope_struct_18_date_validator = &__pyx_type_6validr_12_validator_c___pyx_scope_struct_18_date_validator;
  if (PyType_Ready(&__pyx_type_6validr_12_validator_c___pyx_scope_struct_19_time_validator) < 0) __PYX_ERR(0, 772, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6validr_12_validator_c___pyx_scope_struct_19_time_validator.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6validr_12_validator_c___pyx_scope_struct_19_time_validator.tp_dictoffset && __pyx_type_6validr_12_validator_c___pyx_scope_struct_19_time_validator.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6validr_12_validator_c___pyx_scope_struct_19_time_validator.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6validr_12_validator_c___pyx_scope_struct_19_time_validator = &__pyx_type_6validr_12_validator_c___pyx_scope_struct_19_time_validator;
  if (PyType_Ready(&__pyx_type_6validr_12_validator_c___pyx_scope_struct_20_datetime_validator) < 0) __PYX_ERR(0, 794, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6validr_12_validator_c___pyx_scope_struct_20_datetime_validator.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6validr_12_validator_c___pyx_scope_struct_20_datetime_validator.tp_dictoffset && __pyx_type_6validr_12_validator_c___pyx_scope_struct_20_datetime_validator.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6validr_12_validator_c___pyx_scope_struct_20_datetime_validator.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6validr_12_validator_c___pyx_scope_struct_20_datetime_validator = &__pyx_type_6validr_12_validator_c___pyx_scope_struct_20_datetime_validator;
  if (PyType_Ready(&__pyx_type_6validr_12_validator_c___pyx_scope_struct_21_timedelta_validator) < 0) __PYX_ERR(0, 827, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6validr_12_validator_c___pyx_scope_struct_21_timedelta_validator.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6validr_12_validator_c___pyx_scope_struct_21_timedelta_validator.tp_dictoffset && __pyx_type_6validr_12_validator_c___pyx_scope_struct_21_timedelta_validator.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6validr_12_validator_c___pyx_scope_struct_21_timedelta_validator.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6validr_12_validator_c___pyx_scope_struct_21_timedelta_validator = &__pyx_type_6validr_12_validator_c___pyx_scope_struct_21_timedelta_validator;
  if (PyType_Ready(&__pyx_type_6validr_12_validator_c___pyx_scope_struct_22_ipv4_validator) < 0) __PYX_ERR(0, 881, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6validr_12_validator_c___pyx_scope_struct_22_ipv4_validator.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6validr_12_validator_c___pyx_scope_struct_22_ipv4_validator.tp_dictoffset && __pyx_type_6validr_12_validator_c___pyx_scope_struct_22_ipv4_validator.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6validr_12_validator_c___pyx_scope_struct_22_ipv4_validator.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6validr_12_validator_c___pyx_scope_struct_22_ipv4_validator = &__pyx_type_6validr_12_validator_c___pyx_scope_struct_22_ipv4_validator;
  if (PyType_Ready(&__pyx_type_6validr_12_validator_c___pyx_scope_struct_23_ipv6_validator) < 0) __PYX_ERR(0, 897, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6validr_12_validator_c___pyx_scope_struct_23_ipv6_validator.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6validr_12_validator_c___pyx_scope_struct_23_ipv6_validator.tp_dictoffset && __pyx_type_6validr_12_validator_c___pyx_scope_struct_23_ipv6_validator.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6validr_12_validator_c___pyx_scope_struct_23_ipv6_validator.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6validr_12_validator_c___pyx_scope_struct_23_ipv6_validator = &__pyx_type_6validr_12_validator_c___pyx_scope_struct_23_ipv6_validator;
  if (PyType_Ready(&__pyx_type_6validr_12_validator_c___pyx_scope_struct_24_email_validator) < 0) __PYX_ERR(0, 913, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6validr_12_validator_c___pyx_scope_struct_24_email_validator.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6validr_12_validator_c___pyx_scope_struct_24_email_validator.tp_dictoffset && __pyx_type_6validr_12_validator_c___pyx_scope_struct_24_email_validator.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6validr_12_validator_c___pyx_scope_struct_24_email_validator.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6validr_12_validator_c___pyx_scope_struct_24_email_validator = &__pyx_type_6validr_12_validator_c___pyx_scope_struct_24_email_validator;
  if (PyType_Ready(&__pyx_type_6validr_12_validator_c___pyx_scope_struct_25_url_validator) < 0) __PYX_ERR(0, 936, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6validr_12_validator_c___pyx_scope_struct_25_url_validator.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6validr_12_validator_c___pyx_scope_struct_25_url_validator.tp_dictoffset && __pyx_type_6validr_12_validator_c___pyx_scope_struct_25_url_validator.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6validr_12_validator_c___pyx_scope_struct_25_url_validator.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6validr_12_validator_c___pyx_scope_struct_25_url_validator = &__pyx_type_6validr_12_validator_c___pyx_scope_struct_25_url_validator;
  if (PyType_Ready(&__pyx_type_6validr_12_validator_c___pyx_scope_struct_26_uuid_validator) < 0) __PYX_ERR(0, 979, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6validr_12_validator_c___pyx_scope_struct_26_uuid_validator.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6validr_12_validator_c___pyx_scope_struct_26_uuid_validator.tp_dictoffset && __pyx_type_6validr_12_validator_c___pyx_scope_struct_26_uuid_validator.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6validr_12_validator_c___pyx_scope_struct_26_uuid_validator.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6validr_12_validator_c___pyx_scope_struct_26_uuid_validator = &__pyx_type_6validr_12_validator_c___pyx_scope_struct_26_uuid_validator;
  if (PyType_Ready(&__pyx_type_6validr_12_validator_c___pyx_scope_struct_27_create_re_validator) < 0) __PYX_ERR(0, 1002, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6validr_12_validator_c___pyx_scope_struct_27_create_re_validator.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6validr_12_validator_c___pyx_scope_struct_27_create_re_validator.tp_dictoffset && __pyx_type_6validr_12_validator_c___pyx_scope_struct_27_create_re_validator.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6validr_12_validator_c___pyx_scope_struct_27_create_re_validator.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6validr_12_validator_c___pyx_scope_struct_27_create_re_validator = &__pyx_type_6validr_12_validator_c___pyx_scope_struct_27_create_re_validator;
  if (PyType_Ready(&__pyx_type_6validr_12_validator_c___pyx_scope_struct_28_re_validator) < 0) __PYX_ERR(0, 1016, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6validr_12_validator_c___pyx_scope_struct_28_re_validator.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6validr_12_validator_c___pyx_scope_struct_28_re_validator.tp_dictoffset && __pyx_type_6validr_12_validator_c___pyx_scope_struct_28_re_validator.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6validr_12_validator_c___pyx_scope_struct_28_re_validator.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6validr_12_validator_c___pyx_scope_struct_28_re_validator = &__pyx_type_6validr_12_validator_c___pyx_scope_struct_28_re_validator;
  if (PyType_Ready(&__pyx_type_6validr_12_validator_c___pyx_scope_struct_29_create_enum_validator) < 0) __PYX_ERR(0, 1070, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6validr_12_validator_c___pyx_scope_struct_29_create_enum_validator.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6validr_12_validator_c___pyx_scope_struct_29_create_enum_validator.tp_dictoffset && __pyx_type_6validr_12_validator_c___pyx_scope_struct_29_create_enum_validator.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6validr_12_validator_c___pyx_scope_struct_29_create_enum_validator.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6validr_12_validator_c___pyx_scope_struct_29_create_enum_validator = &__pyx_type_6validr_12_validator_c___pyx_scope_struct_29_create_enum_validator;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#if PY_MAJOR_VERSION < 3
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC void
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#else
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC init_validator_c(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC init_validator_c(void)
#else
__Pyx_PyMODINIT_FUNC PyInit__validator_c(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit__validator_c(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec__validator_c(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  __Pyx_TraceDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module '_validator_c' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit__validator_c(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("_validator_c", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_validr___validator_c) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "validr._validator_c")) {
      if (unlikely(PyDict_SetItemString(modules, "validr._validator_c", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) goto __pyx_L1_error;
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) goto __pyx_L1_error;
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() != 0)) goto __pyx_L1_error;
  (void)__Pyx_modinit_type_import_code();
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __Pyx_TraceCall("__Pyx_PyMODINIT_FUNC PyInit__validator_c(void)", __pyx_f[0], 1, 0, __PYX_ERR(0, 1, __pyx_L1_error));

  /* "validr/_validator_c.pyx":1
 * import re             # <<<<<<<<<<<<<<
 * import sys
 * import uuid
 */
  __Pyx_TraceLine(1,0,__PYX_ERR(0, 1, __pyx_L1_error))
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_re, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_re, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":2
 * import re
 * import sys             # <<<<<<<<<<<<<<
 * import uuid
 * import time
 */
  __Pyx_TraceLine(2,0,__PYX_ERR(0, 2, __pyx_L1_error))
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_sys, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_sys, __pyx_t_1) < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":3
 * import re
 * import sys
 * import uuid             # <<<<<<<<<<<<<<
 * import time
 * import datetime
 */
  __Pyx_TraceLine(3,0,__PYX_ERR(0, 3, __pyx_L1_error))
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_uuid, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_uuid, __pyx_t_1) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":4
 * import sys
 * import uuid
 * import time             # <<<<<<<<<<<<<<
 * import datetime
 * import ipaddress
 */
  __Pyx_TraceLine(4,0,__PYX_ERR(0, 4, __pyx_L1_error))
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_time, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_time, __pyx_t_1) < 0) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":5
 * import uuid
 * import time
 * import datetime             # <<<<<<<<<<<<<<
 * import ipaddress
 * import typing
 */
  __Pyx_TraceLine(5,0,__PYX_ERR(0, 5, __pyx_L1_error))
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_datetime, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_datetime, __pyx_t_1) < 0) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":6
 * import time
 * import datetime
 * import ipaddress             # <<<<<<<<<<<<<<
 * import typing
 * from copy import copy
 */
  __Pyx_TraceLine(6,0,__PYX_ERR(0, 6, __pyx_L1_error))
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_ipaddress, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ipaddress, __pyx_t_1) < 0) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":7
 * import datetime
 * import ipaddress
 * import typing             # <<<<<<<<<<<<<<
 * from copy import copy
 * from functools import partial
 */
  __Pyx_TraceLine(7,0,__PYX_ERR(0, 7, __pyx_L1_error))
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_typing, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_typing, __pyx_t_1) < 0) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":8
 * import ipaddress
 * import typing
 * from copy import copy             # <<<<<<<<<<<<<<
 * from functools import partial
 * from urllib.parse import urlparse, urlunparse
 */
  __Pyx_TraceLine(8,0,__PYX_ERR(0, 8, __pyx_L1_error))
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_copy);
  __Pyx_GIVEREF(__pyx_n_s_copy);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_copy);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_copy, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_copy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_copy, __pyx_t_1) < 0) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "validr/_validator_c.pyx":9
 * import typing
 * from copy import copy
 * from functools import partial             # <<<<<<<<<<<<<<
 * from urllib.parse import urlparse, urlunparse
 * 
 */
  __Pyx_TraceLine(9,0,__PYX_ERR(0, 9, __pyx_L1_error))
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_partial);
  __Pyx_GIVEREF(__pyx_n_s_partial);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_partial);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_functools, __pyx_t_2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_partial); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_partial, __pyx_t_2) < 0) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":10
 * from copy import copy
 * from functools import partial
 * from urllib.parse import urlparse, urlunparse             # <<<<<<<<<<<<<<
 * 
 * from .exception import Invalid, SchemaError, mark_key, mark_index
 */
  __Pyx_TraceLine(10,0,__PYX_ERR(0, 10, __pyx_L1_error))
  __pyx_t_1 = PyList_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_urlparse);
  __Pyx_GIVEREF(__pyx_n_s_urlparse);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_urlparse);
  __Pyx_INCREF(__pyx_n_s_urlunparse);
  __Pyx_GIVEREF(__pyx_n_s_urlunparse);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_s_urlunparse);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_urllib_parse, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_urlparse); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_urlparse, __pyx_t_1) < 0) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_urlunparse); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_urlunparse, __pyx_t_1) < 0) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "validr/_validator_c.pyx":12
 * from urllib.parse import urlparse, urlunparse
 * 
 * from .exception import Invalid, SchemaError, mark_key, mark_index             # <<<<<<<<<<<<<<
 * from ._vendor import durationpy
 * from ._vendor.email_validator import validate_email, EmailNotValidError
 */
  __Pyx_TraceLine(12,0,__PYX_ERR(0, 12, __pyx_L1_error))
  __pyx_t_2 = PyList_New(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_Invalid);
  __Pyx_GIVEREF(__pyx_n_s_Invalid);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_Invalid);
  __Pyx_INCREF(__pyx_n_s_SchemaError);
  __Pyx_GIVEREF(__pyx_n_s_SchemaError);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_SchemaError);
  __Pyx_INCREF(__pyx_n_s_mark_key);
  __Pyx_GIVEREF(__pyx_n_s_mark_key);
  PyList_SET_ITEM(__pyx_t_2, 2, __pyx_n_s_mark_key);
  __Pyx_INCREF(__pyx_n_s_mark_index);
  __Pyx_GIVEREF(__pyx_n_s_mark_index);
  PyList_SET_ITEM(__pyx_t_2, 3, __pyx_n_s_mark_index);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_exception, __pyx_t_2, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_Invalid); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Invalid, __pyx_t_2) < 0) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_SchemaError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_SchemaError, __pyx_t_2) < 0) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_mark_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_mark_key, __pyx_t_2) < 0) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_mark_index); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_mark_index, __pyx_t_2) < 0) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":13
 * 
 * from .exception import Invalid, SchemaError, mark_key, mark_index
 * from ._vendor import durationpy             # <<<<<<<<<<<<<<
 * from ._vendor.email_validator import validate_email, EmailNotValidError
 * from ._vendor.fqdn import FQDN
 */
  __Pyx_TraceLine(13,0,__PYX_ERR(0, 13, __pyx_L1_error))
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_durationpy);
  __Pyx_GIVEREF(__pyx_n_s_durationpy);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_durationpy);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_vendor, __pyx_t_1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_durationpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_durationpy, __pyx_t_1) < 0) __PYX_ERR(0, 13, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "validr/_validator_c.pyx":14
 * from .exception import Invalid, SchemaError, mark_key, mark_index
 * from ._vendor import durationpy
 * from ._vendor.email_validator import validate_email, EmailNotValidError             # <<<<<<<<<<<<<<
 * from ._vendor.fqdn import FQDN
 * 
 */
  __Pyx_TraceLine(14,0,__PYX_ERR(0, 14, __pyx_L1_error))
  __pyx_t_2 = PyList_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_validate_email);
  __Pyx_GIVEREF(__pyx_n_s_validate_email);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_validate_email);
  __Pyx_INCREF(__pyx_n_s_EmailNotValidError);
  __Pyx_GIVEREF(__pyx_n_s_EmailNotValidError);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_EmailNotValidError);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_vendor_email_validator, __pyx_t_2, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_validate_email); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_validate_email, __pyx_t_2) < 0) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_EmailNotValidError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_EmailNotValidError, __pyx_t_2) < 0) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":15
 * from ._vendor import durationpy
 * from ._vendor.email_validator import validate_email, EmailNotValidError
 * from ._vendor.fqdn import FQDN             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(15,0,__PYX_ERR(0, 15, __pyx_L1_error))
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_FQDN);
  __Pyx_GIVEREF(__pyx_n_s_FQDN);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_FQDN);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_vendor_fqdn, __pyx_t_1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_FQDN); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_FQDN, __pyx_t_1) < 0) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "validr/_validator_c.pyx":18
 * 
 * 
 * cpdef bint is_dict(obj):             # <<<<<<<<<<<<<<
 *     # use isinstance(obj, Mapping) is slow,
 *     # hasattr check can speed up about 30%
 */
  __Pyx_TraceLine(18,0,__PYX_ERR(0, 18, __pyx_L1_error))


  /* "validr/_validator_c.pyx":24
 * 
 * 
 * cpdef inline get_dict_value(obj, str key):             # <<<<<<<<<<<<<<
 *     return obj.get(key, None)
 * 
 */
  __Pyx_TraceLine(24,0,__PYX_ERR(0, 24, __pyx_L1_error))


  /* "validr/_validator_c.pyx":28
 * 
 * 
 * cpdef inline get_object_value(obj, str key):             # <<<<<<<<<<<<<<
 *     return getattr(obj, key, None)
 * 
 */
  __Pyx_TraceLine(28,0,__PYX_ERR(0, 28, __pyx_L1_error))


  /* "validr/_validator_c.pyx":32
 * 
 * 
 * cpdef inline bint _is_empty(value):             # <<<<<<<<<<<<<<
 *     return value is None or value == ''
 * 
 */
  __Pyx_TraceLine(32,0,__PYX_ERR(0, 32, __pyx_L1_error))


  /* "validr/_validator_c.pyx":36
 * 
 * 
 * cpdef _update_validate_func_info(validate_func, origin_func, schema):             # <<<<<<<<<<<<<<
 *     # make friendly validate func representation
 *     m_repr = schema.repr(prefix=False, desc=False)
 */
  __Pyx_TraceLine(36,0,__PYX_ERR(0, 36, __pyx_L1_error))


  /* "validr/_validator_c.pyx":51
 * 
 * 
 * def _parse_hints(hints):             # <<<<<<<<<<<<<<
 *     if not isinstance(hints, (tuple, set, list)):
 *         hints = [hints]
 */
  __Pyx_TraceLine(51,0,__PYX_ERR(0, 51, __pyx_L1_error))
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_11_parse_hints, NULL, __pyx_n_s_validr__validator_c); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_parse_hints, __pyx_t_2) < 0) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "validr/_validator_c.pyx":68
 * 
 * 
 * def _update_validate_func_type_hints(             # <<<<<<<<<<<<<<
 *         validate_func,
 *         optional, has_default,
 */
  __Pyx_TraceLine(68,0,__PYX_ERR(0, 68, __pyx_L1_error))
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_13_update_validate_func_type_hints, NULL, __pyx_n_s_validr__validator_c); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 68, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_update_validate_func_type_hints, __pyx_t_2) < 0) __PYX_ERR(0, 68, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "validr/_validator_c.pyx":91
 * 
 * 
 * def validator(string=None, *, accept=None, output=None):             # <<<<<<<<<<<<<<
 *     """Decorator for create validator
 * 
 */
  __Pyx_TraceLine(91,0,__PYX_ERR(0, 91, __pyx_L1_error))
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_15validator, NULL, __pyx_n_s_validr__validator_c); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 91, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_validator, __pyx_t_2) < 0) __PYX_ERR(0, 91, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "validr/_validator_c.pyx":278
 * 
 * 
 * cpdef str _UNIQUE_CHECK_ERROR_MESSAGE = "unable to check unique for non-hashable types"             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_TraceLine(278,0,__PYX_ERR(0, 278, __pyx_L1_error))
  __Pyx_INCREF(__pyx_kp_u_unable_to_check_unique_for_non_h);
  __Pyx_XGOTREF(__pyx_v_6validr_12_validator_c__UNIQUE_CHECK_ERROR_MESSAGE);
  __Pyx_DECREF_SET(__pyx_v_6validr_12_validator_c__UNIQUE_CHECK_ERROR_MESSAGE, __pyx_kp_u_unable_to_check_unique_for_non_h);
  __Pyx_GIVEREF(__pyx_kp_u_unable_to_check_unique_for_non_h);

  /* "validr/_validator_c.pyx":281
 * 
 * 
 * cpdef inline _key_of_scalar(v):             # <<<<<<<<<<<<<<
 *     return v
 * 
 */
  __Pyx_TraceLine(281,0,__PYX_ERR(0, 281, __pyx_L1_error))


  /* "validr/_validator_c.pyx":285
 * 
 * 
 * def _key_func_of_schema(schema):             # <<<<<<<<<<<<<<
 *     if schema is None:
 *         raise SchemaError(_UNIQUE_CHECK_ERROR_MESSAGE)
 */
  __Pyx_TraceLine(285,0,__PYX_ERR(0, 285, __pyx_L1_error))
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_19_key_func_of_schema, NULL, __pyx_n_s_validr__validator_c); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_key_func_of_schema, __pyx_t_2) < 0) __PYX_ERR(0, 285, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "validr/_validator_c.pyx":314
 * 
 * 
 * @validator(accept=typing.Iterable, output=typing.List)             # <<<<<<<<<<<<<<
 * def list_validator(compiler, items=None, int minlen=0, int maxlen=1024,
 *                    bint unique=False):
 */
  __Pyx_TraceLine(314,0,__PYX_ERR(0, 314, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_validator); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_typing); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Iterable); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_accept, __pyx_t_4) < 0) __PYX_ERR(0, 314, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_typing); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_List); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_output, __pyx_t_3) < 0) __PYX_ERR(0, 314, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":315
 * 
 * @validator(accept=typing.Iterable, output=typing.List)
 * def list_validator(compiler, items=None, int minlen=0, int maxlen=1024,             # <<<<<<<<<<<<<<
 *                    bint unique=False):
 *     if items is None:
 */
  __Pyx_TraceLine(315,0,__PYX_ERR(0, 315, __pyx_L1_error))
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_21list_validator, NULL, __pyx_n_s_validr__validator_c); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 315, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "validr/_validator_c.pyx":314
 * 
 * 
 * @validator(accept=typing.Iterable, output=typing.List)             # <<<<<<<<<<<<<<
 * def list_validator(compiler, items=None, int minlen=0, int maxlen=1024,
 *                    bint unique=False):
 */
  __Pyx_TraceLine(314,0,__PYX_ERR(0, 314, __pyx_L1_error))
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_list_validator, __pyx_t_2) < 0) __PYX_ERR(0, 315, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "validr/_validator_c.pyx":352
 * 
 * 
 * cpdef inline dict _slim_dict(dict value):             # <<<<<<<<<<<<<<
 *     return {k: v for k, v in value.items() if not _is_empty(v)}
 * 
 */
  __Pyx_TraceLine(352,0,__PYX_ERR(0, 352, __pyx_L1_error))


  /* "validr/_validator_c.pyx":356
 * 
 * 
 * @validator(accept=(typing.Mapping, typing.Any), output=dict)             # <<<<<<<<<<<<<<
 * def dict_validator(compiler, items=None, key=None, value=None,
 *                    int minlen=0, int maxlen=1024, bint slim=False):
 */
  __Pyx_TraceLine(356,0,__PYX_ERR(0, 356, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_validator); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 356, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 356, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_typing); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 356, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Mapping); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 356, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_typing); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 356, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Any); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 356, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 356, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_5);
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_accept, __pyx_t_3) < 0) __PYX_ERR(0, 356, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_output, ((PyObject *)(&PyDict_Type))) < 0) __PYX_ERR(0, 356, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 356, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":357
 * 
 * @validator(accept=(typing.Mapping, typing.Any), output=dict)
 * def dict_validator(compiler, items=None, key=None, value=None,             # <<<<<<<<<<<<<<
 *                    int minlen=0, int maxlen=1024, bint slim=False):
 *     if items is None:
 */
  __Pyx_TraceLine(357,0,__PYX_ERR(0, 357, __pyx_L1_error))
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_25dict_validator, NULL, __pyx_n_s_validr__validator_c); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 357, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "validr/_validator_c.pyx":356
 * 
 * 
 * @validator(accept=(typing.Mapping, typing.Any), output=dict)             # <<<<<<<<<<<<<<
 * def dict_validator(compiler, items=None, key=None, value=None,
 *                    int minlen=0, int maxlen=1024, bint slim=False):
 */
  __Pyx_TraceLine(356,0,__PYX_ERR(0, 356, __pyx_L1_error))
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 356, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_dict_validator, __pyx_t_2) < 0) __PYX_ERR(0, 357, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "validr/_validator_c.pyx":423
 * 
 * 
 * @validator(accept=(typing.Mapping, typing.Any), output=object)             # <<<<<<<<<<<<<<
 * def model_validator(compiler, items=None):
 *     if items is None:
 */
  __Pyx_TraceLine(423,0,__PYX_ERR(0, 423, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_validator); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_typing); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Mapping); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_typing); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Any); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4);
  __pyx_t_5 = 0;
  __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_accept, __pyx_t_3) < 0) __PYX_ERR(0, 423, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_output, __pyx_builtin_object) < 0) __PYX_ERR(0, 423, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":424
 * 
 * @validator(accept=(typing.Mapping, typing.Any), output=object)
 * def model_validator(compiler, items=None):             # <<<<<<<<<<<<<<
 *     if items is None:
 *         raise SchemaError('model class not provided')
 */
  __Pyx_TraceLine(424,0,__PYX_ERR(0, 424, __pyx_L1_error))
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_27model_validator, NULL, __pyx_n_s_validr__validator_c); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 424, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "validr/_validator_c.pyx":423
 * 
 * 
 * @validator(accept=(typing.Mapping, typing.Any), output=object)             # <<<<<<<<<<<<<<
 * def model_validator(compiler, items=None):
 *     if items is None:
 */
  __Pyx_TraceLine(423,0,__PYX_ERR(0, 423, __pyx_L1_error))
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_model_validator, __pyx_t_2) < 0) __PYX_ERR(0, 424, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "validr/_validator_c.pyx":434
 * 
 * 
 * cpdef _dump_enum_value(value):             # <<<<<<<<<<<<<<
 *     if value is None:
 *         return 'null'
 */
  __Pyx_TraceLine(434,0,__PYX_ERR(0, 434, __pyx_L1_error))


  /* "validr/_validator_c.pyx":447
 * 
 * 
 * @validator(output=object)             # <<<<<<<<<<<<<<
 * def enum_validator(compiler, items):
 *     if not items:
 */
  __Pyx_TraceLine(447,0,__PYX_ERR(0, 447, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_validator); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_output, __pyx_builtin_object) < 0) __PYX_ERR(0, 447, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":448
 * 
 * @validator(output=object)
 * def enum_validator(compiler, items):             # <<<<<<<<<<<<<<
 *     if not items:
 *         raise SchemaError("enum items not provided")
 */
  __Pyx_TraceLine(448,0,__PYX_ERR(0, 448, __pyx_L1_error))
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_31enum_validator, NULL, __pyx_n_s_validr__validator_c); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 448, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "validr/_validator_c.pyx":447
 * 
 * 
 * @validator(output=object)             # <<<<<<<<<<<<<<
 * def enum_validator(compiler, items):
 *     if not items:
 */
  __Pyx_TraceLine(447,0,__PYX_ERR(0, 447, __pyx_L1_error))
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_enum_validator, __pyx_t_2) < 0) __PYX_ERR(0, 448, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "validr/_validator_c.pyx":462
 * 
 * 
 * cpdef union_validator(compiler, schema):             # <<<<<<<<<<<<<<
 *     if not schema.items:
 *         raise SchemaError('union schemas not provided')
 */
  __Pyx_TraceLine(462,0,__PYX_ERR(0, 462, __pyx_L1_error))


  /* "validr/_validator_c.pyx":483
 * 
 * 
 * cpdef _optional_or_has_default(schema):             # <<<<<<<<<<<<<<
 *     if schema.params.get('optional'):
 *         return True
 */
  __Pyx_TraceLine(483,0,__PYX_ERR(0, 483, __pyx_L1_error))


  /* "validr/_validator_c.pyx":491
 * 
 * 
 * def _union_list_validator(compiler, schema):             # <<<<<<<<<<<<<<
 *     scalar_inner = None
 *     list_inner = None
 */
  __Pyx_TraceLine(491,0,__PYX_ERR(0, 491, __pyx_L1_error))
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_37_union_list_validator, NULL, __pyx_n_s_validr__validator_c); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_union_list_validator, __pyx_t_2) < 0) __PYX_ERR(0, 491, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "validr/_validator_c.pyx":538
 * 
 * 
 * @validator(accept=object, output=object)             # <<<<<<<<<<<<<<
 * def _union_dict_validator(compiler, items, str by):
 *     inners = {}
 */
  __Pyx_TraceLine(538,0,__PYX_ERR(0, 538, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_validator); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_accept, __pyx_builtin_object) < 0) __PYX_ERR(0, 538, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_output, __pyx_builtin_object) < 0) __PYX_ERR(0, 538, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":539
 * 
 * @validator(accept=object, output=object)
 * def _union_dict_validator(compiler, items, str by):             # <<<<<<<<<<<<<<
 *     inners = {}
 *     for key, schema in items.items():
 */
  __Pyx_TraceLine(539,0,__PYX_ERR(0, 539, __pyx_L1_error))
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_39_union_dict_validator, NULL, __pyx_n_s_validr__validator_c); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 539, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "validr/_validator_c.pyx":538
 * 
 * 
 * @validator(accept=object, output=object)             # <<<<<<<<<<<<<<
 * def _union_dict_validator(compiler, items, str by):
 *     inners = {}
 */
  __Pyx_TraceLine(538,0,__PYX_ERR(0, 538, __pyx_L1_error))
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_union_dict_validator, __pyx_t_2) < 0) __PYX_ERR(0, 539, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "validr/_validator_c.pyx":575
 * 
 * 
 * cpdef any_validate(value):             # <<<<<<<<<<<<<<
 *     return copy(value)
 * 
 */
  __Pyx_TraceLine(575,0,__PYX_ERR(0, 575, __pyx_L1_error))


  /* "validr/_validator_c.pyx":579
 * 
 * 
 * @validator(accept=object, output=object)             # <<<<<<<<<<<<<<
 * def any_validator(compiler, **ignore_kwargs):
 *     """Accept any value"""
 */
  __Pyx_TraceLine(579,0,__PYX_ERR(0, 579, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_validator); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_accept, __pyx_builtin_object) < 0) __PYX_ERR(0, 579, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_output, __pyx_builtin_object) < 0) __PYX_ERR(0, 579, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":580
 * 
 * @validator(accept=object, output=object)
 * def any_validator(compiler, **ignore_kwargs):             # <<<<<<<<<<<<<<
 *     """Accept any value"""
 *     return any_validate
 */
  __Pyx_TraceLine(580,0,__PYX_ERR(0, 580, __pyx_L1_error))
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_43any_validator, NULL, __pyx_n_s_validr__validator_c); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 580, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "validr/_validator_c.pyx":579
 * 
 * 
 * @validator(accept=object, output=object)             # <<<<<<<<<<<<<<
 * def any_validator(compiler, **ignore_kwargs):
 *     """Accept any value"""
 */
  __Pyx_TraceLine(579,0,__PYX_ERR(0, 579, __pyx_L1_error))
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_any_validator, __pyx_t_2) < 0) __PYX_ERR(0, 580, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "validr/_validator_c.pyx":586
 * 
 * @validator(accept=(int, float, str), output=int)
 * def int_validator(compiler, min=-sys.maxsize, max=sys.maxsize):             # <<<<<<<<<<<<<<
 *     """Validate int or convert string to int
 * 
 */
  __Pyx_TraceLine(586,0,__PYX_ERR(0, 586, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_sys); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 586, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_maxsize); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 586, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyNumber_Negative(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 586, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_k__52 = __pyx_t_2;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_sys); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 586, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_maxsize); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 586, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_k__53 = __pyx_t_1;
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":585
 * 
 * 
 * @validator(accept=(int, float, str), output=int)             # <<<<<<<<<<<<<<
 * def int_validator(compiler, min=-sys.maxsize, max=sys.maxsize):
 *     """Validate int or convert string to int
 */
  __Pyx_TraceLine(585,0,__PYX_ERR(0, 585, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_validator); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 585, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 585, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 585, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)(&PyInt_Type)));
  __Pyx_GIVEREF(((PyObject *)(&PyInt_Type)));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)(&PyInt_Type)));
  __Pyx_INCREF(((PyObject *)(&PyFloat_Type)));
  __Pyx_GIVEREF(((PyObject *)(&PyFloat_Type)));
  PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)(&PyFloat_Type)));
  __Pyx_INCREF(((PyObject *)(&PyUnicode_Type)));
  __Pyx_GIVEREF(((PyObject *)(&PyUnicode_Type)));
  PyTuple_SET_ITEM(__pyx_t_3, 2, ((PyObject *)(&PyUnicode_Type)));
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_accept, __pyx_t_3) < 0) __PYX_ERR(0, 585, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_output, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 585, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 585, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "validr/_validator_c.pyx":586
 * 
 * @validator(accept=(int, float, str), output=int)
 * def int_validator(compiler, min=-sys.maxsize, max=sys.maxsize):             # <<<<<<<<<<<<<<
 *     """Validate int or convert string to int
 * 
 */
  __Pyx_TraceLine(586,0,__PYX_ERR(0, 586, __pyx_L1_error))
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_45int_validator, NULL, __pyx_n_s_validr__validator_c); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 586, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "validr/_validator_c.pyx":585
 * 
 * 
 * @validator(accept=(int, float, str), output=int)             # <<<<<<<<<<<<<<
 * def int_validator(compiler, min=-sys.maxsize, max=sys.maxsize):
 *     """Validate int or convert string to int
 */
  __Pyx_TraceLine(585,0,__PYX_ERR(0, 585, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 585, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_int_validator, __pyx_t_1) < 0) __PYX_ERR(0, 586, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":609
 * 
 * _TRUE_VALUES = {
 *     True, 1, '1',             # <<<<<<<<<<<<<<
 *     'True', 'true', 'TRUE',
 *     'Yes', 'yes', 'YES', 'y', 'Y',
 */
  __Pyx_TraceLine(609,0,__PYX_ERR(0, 609, __pyx_L1_error))
  __pyx_t_1 = PySet_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 609, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PySet_Add(__pyx_t_1, Py_True) < 0) __PYX_ERR(0, 609, __pyx_L1_error)
  if (PySet_Add(__pyx_t_1, __pyx_int_1) < 0) __PYX_ERR(0, 609, __pyx_L1_error)
  if (PySet_Add(__pyx_t_1, __pyx_kp_u_1) < 0) __PYX_ERR(0, 609, __pyx_L1_error)
  if (PySet_Add(__pyx_t_1, __pyx_n_u_True) < 0) __PYX_ERR(0, 609, __pyx_L1_error)
  if (PySet_Add(__pyx_t_1, __pyx_n_u_true) < 0) __PYX_ERR(0, 609, __pyx_L1_error)
  if (PySet_Add(__pyx_t_1, __pyx_n_u_TRUE) < 0) __PYX_ERR(0, 609, __pyx_L1_error)
  if (PySet_Add(__pyx_t_1, __pyx_n_u_Yes) < 0) __PYX_ERR(0, 609, __pyx_L1_error)
  if (PySet_Add(__pyx_t_1, __pyx_n_u_yes) < 0) __PYX_ERR(0, 609, __pyx_L1_error)
  if (PySet_Add(__pyx_t_1, __pyx_n_u_YES) < 0) __PYX_ERR(0, 609, __pyx_L1_error)
  if (PySet_Add(__pyx_t_1, __pyx_n_u_y) < 0) __PYX_ERR(0, 609, __pyx_L1_error)
  if (PySet_Add(__pyx_t_1, __pyx_n_u_Y) < 0) __PYX_ERR(0, 609, __pyx_L1_error)
  if (PySet_Add(__pyx_t_1, __pyx_n_u_On) < 0) __PYX_ERR(0, 609, __pyx_L1_error)
  if (PySet_Add(__pyx_t_1, __pyx_n_u_on) < 0) __PYX_ERR(0, 609, __pyx_L1_error)
  if (PySet_Add(__pyx_t_1, __pyx_n_u_ON) < 0) __PYX_ERR(0, 609, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TRUE_VALUES, __pyx_t_1) < 0) __PYX_ERR(0, 608, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":615
 * }
 * _FALSE_VALUES = {
 *     False, 0, '0',             # <<<<<<<<<<<<<<
 *     'False', 'false', 'FALSE',
 *     'No', 'no', 'NO', 'n', 'N',
 */
  __Pyx_TraceLine(615,0,__PYX_ERR(0, 615, __pyx_L1_error))
  __pyx_t_1 = PySet_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 615, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PySet_Add(__pyx_t_1, Py_False) < 0) __PYX_ERR(0, 615, __pyx_L1_error)
  if (PySet_Add(__pyx_t_1, __pyx_int_0) < 0) __PYX_ERR(0, 615, __pyx_L1_error)
  if (PySet_Add(__pyx_t_1, __pyx_kp_u_0) < 0) __PYX_ERR(0, 615, __pyx_L1_error)
  if (PySet_Add(__pyx_t_1, __pyx_n_u_False) < 0) __PYX_ERR(0, 615, __pyx_L1_error)
  if (PySet_Add(__pyx_t_1, __pyx_n_u_false) < 0) __PYX_ERR(0, 615, __pyx_L1_error)
  if (PySet_Add(__pyx_t_1, __pyx_n_u_FALSE) < 0) __PYX_ERR(0, 615, __pyx_L1_error)
  if (PySet_Add(__pyx_t_1, __pyx_n_u_No) < 0) __PYX_ERR(0, 615, __pyx_L1_error)
  if (PySet_Add(__pyx_t_1, __pyx_n_u_no) < 0) __PYX_ERR(0, 615, __pyx_L1_error)
  if (PySet_Add(__pyx_t_1, __pyx_n_u_NO) < 0) __PYX_ERR(0, 615, __pyx_L1_error)
  if (PySet_Add(__pyx_t_1, __pyx_n_u_n) < 0) __PYX_ERR(0, 615, __pyx_L1_error)
  if (PySet_Add(__pyx_t_1, __pyx_n_u_N) < 0) __PYX_ERR(0, 615, __pyx_L1_error)
  if (PySet_Add(__pyx_t_1, __pyx_n_u_Off) < 0) __PYX_ERR(0, 615, __pyx_L1_error)
  if (PySet_Add(__pyx_t_1, __pyx_n_u_off) < 0) __PYX_ERR(0, 615, __pyx_L1_error)
  if (PySet_Add(__pyx_t_1, __pyx_n_u_OFF) < 0) __PYX_ERR(0, 615, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_FALSE_VALUES, __pyx_t_1) < 0) __PYX_ERR(0, 614, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":622
 * 
 * 
 * @validator(accept=(bool, int, str), output=bool)             # <<<<<<<<<<<<<<
 * def bool_validator(compiler):
 *     """Validate bool"""
 */
  __Pyx_TraceLine(622,0,__PYX_ERR(0, 622, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_validator); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject*)&PyBool_Type));
  __Pyx_GIVEREF(((PyObject*)&PyBool_Type));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject*)&PyBool_Type));
  __Pyx_INCREF(((PyObject *)(&PyInt_Type)));
  __Pyx_GIVEREF(((PyObject *)(&PyInt_Type)));
  PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)(&PyInt_Type)));
  __Pyx_INCREF(((PyObject *)(&PyUnicode_Type)));
  __Pyx_GIVEREF(((PyObject *)(&PyUnicode_Type)));
  PyTuple_SET_ITEM(__pyx_t_3, 2, ((PyObject *)(&PyUnicode_Type)));
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_accept, __pyx_t_3) < 0) __PYX_ERR(0, 622, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_output, ((PyObject*)&PyBool_Type)) < 0) __PYX_ERR(0, 622, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "validr/_validator_c.pyx":623
 * 
 * @validator(accept=(bool, int, str), output=bool)
 * def bool_validator(compiler):             # <<<<<<<<<<<<<<
 *     """Validate bool"""
 *     def validate(value):
 */
  __Pyx_TraceLine(623,0,__PYX_ERR(0, 623, __pyx_L1_error))
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_47bool_validator, NULL, __pyx_n_s_validr__validator_c); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "validr/_validator_c.pyx":622
 * 
 * 
 * @validator(accept=(bool, int, str), output=bool)             # <<<<<<<<<<<<<<
 * def bool_validator(compiler):
 *     """Validate bool"""
 */
  __Pyx_TraceLine(622,0,__PYX_ERR(0, 622, __pyx_L1_error))
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_bool_validator, __pyx_t_1) < 0) __PYX_ERR(0, 623, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":636
 * 
 * @validator(accept=(int, float, str), output=float)
 * def float_validator(compiler, min=-sys.float_info.max, max=sys.float_info.max,             # <<<<<<<<<<<<<<
 *                     exmin=False, exmax=False):
 *     """Validate float string
 */
  __Pyx_TraceLine(636,0,__PYX_ERR(0, 636, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_sys); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 636, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_float_info); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 636, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_max); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 636, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyNumber_Negative(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 636, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_k__60 = __pyx_t_2;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_sys); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 636, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_float_info); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 636, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_max); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 636, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_k__61 = __pyx_t_2;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "validr/_validator_c.pyx":635
 * 
 * 
 * @validator(accept=(int, float, str), output=float)             # <<<<<<<<<<<<<<
 * def float_validator(compiler, min=-sys.float_info.max, max=sys.float_info.max,
 *                     exmin=False, exmax=False):
 */
  __Pyx_TraceLine(635,0,__PYX_ERR(0, 635, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_validator); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)(&PyInt_Type)));
  __Pyx_GIVEREF(((PyObject *)(&PyInt_Type)));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)(&PyInt_Type)));
  __Pyx_INCREF(((PyObject *)(&PyFloat_Type)));
  __Pyx_GIVEREF(((PyObject *)(&PyFloat_Type)));
  PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)(&PyFloat_Type)));
  __Pyx_INCREF(((PyObject *)(&PyUnicode_Type)));
  __Pyx_GIVEREF(((PyObject *)(&PyUnicode_Type)));
  PyTuple_SET_ITEM(__pyx_t_3, 2, ((PyObject *)(&PyUnicode_Type)));
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_accept, __pyx_t_3) < 0) __PYX_ERR(0, 635, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_output, ((PyObject *)(&PyFloat_Type))) < 0) __PYX_ERR(0, 635, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":636
 * 
 * @validator(accept=(int, float, str), output=float)
 * def float_validator(compiler, min=-sys.float_info.max, max=sys.float_info.max,             # <<<<<<<<<<<<<<
 *                     exmin=False, exmax=False):
 *     """Validate float string
 */
  __Pyx_TraceLine(636,0,__PYX_ERR(0, 636, __pyx_L1_error))
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_49float_validator, NULL, __pyx_n_s_validr__validator_c); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 636, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "validr/_validator_c.pyx":635
 * 
 * 
 * @validator(accept=(int, float, str), output=float)             # <<<<<<<<<<<<<<
 * def float_validator(compiler, min=-sys.float_info.max, max=sys.float_info.max,
 *                     exmin=False, exmax=False):
 */
  __Pyx_TraceLine(635,0,__PYX_ERR(0, 635, __pyx_L1_error))
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_float_validator, __pyx_t_2) < 0) __PYX_ERR(0, 636, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "validr/_validator_c.pyx":679
 * 
 * 
 * @validator(accept=(str, object), output=str)             # <<<<<<<<<<<<<<
 * def str_validator(compiler, int minlen=0, int maxlen=1024 * 1024,
 *                   bint strip=False, bint escape=False, str match=None,
 */
  __Pyx_TraceLine(679,0,__PYX_ERR(0, 679, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_validator); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 679, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 679, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 679, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)(&PyUnicode_Type)));
  __Pyx_GIVEREF(((PyObject *)(&PyUnicode_Type)));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)(&PyUnicode_Type)));
  __Pyx_INCREF(__pyx_builtin_object);
  __Pyx_GIVEREF(__pyx_builtin_object);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_builtin_object);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_accept, __pyx_t_3) < 0) __PYX_ERR(0, 679, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_output, ((PyObject *)(&PyUnicode_Type))) < 0) __PYX_ERR(0, 679, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 679, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":680
 * 
 * @validator(accept=(str, object), output=str)
 * def str_validator(compiler, int minlen=0, int maxlen=1024 * 1024,             # <<<<<<<<<<<<<<
 *                   bint strip=False, bint escape=False, str match=None,
 *                   bint accept_object=False):
 */
  __Pyx_TraceLine(680,0,__PYX_ERR(0, 680, __pyx_L1_error))
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_51str_validator, NULL, __pyx_n_s_validr__validator_c); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 680, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "validr/_validator_c.pyx":679
 * 
 * 
 * @validator(accept=(str, object), output=str)             # <<<<<<<<<<<<<<
 * def str_validator(compiler, int minlen=0, int maxlen=1024 * 1024,
 *                   bint strip=False, bint escape=False, str match=None,
 */
  __Pyx_TraceLine(679,0,__PYX_ERR(0, 679, __pyx_L1_error))
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 679, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_str_validator, __pyx_t_2) < 0) __PYX_ERR(0, 680, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "validr/_validator_c.pyx":727
 * 
 * 
 * @validator(accept=bytes, output=bytes)             # <<<<<<<<<<<<<<
 * def bytes_validator(compiler, int minlen=0, int maxlen=-1):
 *     """Validate bytes
 */
  __Pyx_TraceLine(727,0,__PYX_ERR(0, 727, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_validator); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 727, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 727, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_accept, ((PyObject *)(&PyBytes_Type))) < 0) __PYX_ERR(0, 727, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_output, ((PyObject *)(&PyBytes_Type))) < 0) __PYX_ERR(0, 727, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 727, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":728
 * 
 * @validator(accept=bytes, output=bytes)
 * def bytes_validator(compiler, int minlen=0, int maxlen=-1):             # <<<<<<<<<<<<<<
 *     """Validate bytes
 * 
 */
  __Pyx_TraceLine(728,0,__PYX_ERR(0, 728, __pyx_L1_error))
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_53bytes_validator, NULL, __pyx_n_s_validr__validator_c); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 728, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "validr/_validator_c.pyx":727
 * 
 * 
 * @validator(accept=bytes, output=bytes)             # <<<<<<<<<<<<<<
 * def bytes_validator(compiler, int minlen=0, int maxlen=-1):
 *     """Validate bytes
 */
  __Pyx_TraceLine(727,0,__PYX_ERR(0, 727, __pyx_L1_error))
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 727, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_bytes_validator, __pyx_t_2) < 0) __PYX_ERR(0, 728, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "validr/_validator_c.pyx":749
 * 
 * 
 * @validator(accept=(str, datetime.date), output=(str, datetime.date))             # <<<<<<<<<<<<<<
 * def date_validator(compiler, str format='%Y-%m-%d', bint output_object=False):
 *     """Validate date string or convert date to string
 */
  __Pyx_TraceLine(749,0,__PYX_ERR(0, 749, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_validator); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 749, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 749, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_datetime); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 749, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_date); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 749, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 749, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)(&PyUnicode_Type)));
  __Pyx_GIVEREF(((PyObject *)(&PyUnicode_Type)));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)(&PyUnicode_Type)));
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_accept, __pyx_t_3) < 0) __PYX_ERR(0, 749, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_datetime); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 749, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_date); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 749, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 749, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)(&PyUnicode_Type)));
  __Pyx_GIVEREF(((PyObject *)(&PyUnicode_Type)));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)(&PyUnicode_Type)));
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_output, __pyx_t_3) < 0) __PYX_ERR(0, 749, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 749, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":750
 * 
 * @validator(accept=(str, datetime.date), output=(str, datetime.date))
 * def date_validator(compiler, str format='%Y-%m-%d', bint output_object=False):             # <<<<<<<<<<<<<<
 *     """Validate date string or convert date to string
 * 
 */
  __Pyx_TraceLine(750,0,__PYX_ERR(0, 750, __pyx_L1_error))
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_55date_validator, NULL, __pyx_n_s_validr__validator_c); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 750, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "validr/_validator_c.pyx":749
 * 
 * 
 * @validator(accept=(str, datetime.date), output=(str, datetime.date))             # <<<<<<<<<<<<<<
 * def date_validator(compiler, str format='%Y-%m-%d', bint output_object=False):
 *     """Validate date string or convert date to string
 */
  __Pyx_TraceLine(749,0,__PYX_ERR(0, 749, __pyx_L1_error))
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 749, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_date_validator, __pyx_t_2) < 0) __PYX_ERR(0, 750, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "validr/_validator_c.pyx":771
 * 
 * 
 * @validator(accept=(str, datetime.time), output=(str, datetime.time))             # <<<<<<<<<<<<<<
 * def time_validator(compiler, str format='%H:%M:%S', bint output_object=False):
 *     """Validate time string or convert time to string
 */
  __Pyx_TraceLine(771,0,__PYX_ERR(0, 771, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_validator); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 771, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 771, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_datetime); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 771, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_time); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 771, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 771, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)(&PyUnicode_Type)));
  __Pyx_GIVEREF(((PyObject *)(&PyUnicode_Type)));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)(&PyUnicode_Type)));
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_accept, __pyx_t_3) < 0) __PYX_ERR(0, 771, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_datetime); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 771, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_time); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 771, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 771, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)(&PyUnicode_Type)));
  __Pyx_GIVEREF(((PyObject *)(&PyUnicode_Type)));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)(&PyUnicode_Type)));
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_output, __pyx_t_3) < 0) __PYX_ERR(0, 771, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 771, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":772
 * 
 * @validator(accept=(str, datetime.time), output=(str, datetime.time))
 * def time_validator(compiler, str format='%H:%M:%S', bint output_object=False):             # <<<<<<<<<<<<<<
 *     """Validate time string or convert time to string
 * 
 */
  __Pyx_TraceLine(772,0,__PYX_ERR(0, 772, __pyx_L1_error))
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_57time_validator, NULL, __pyx_n_s_validr__validator_c); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 772, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "validr/_validator_c.pyx":771
 * 
 * 
 * @validator(accept=(str, datetime.time), output=(str, datetime.time))             # <<<<<<<<<<<<<<
 * def time_validator(compiler, str format='%H:%M:%S', bint output_object=False):
 *     """Validate time string or convert time to string
 */
  __Pyx_TraceLine(771,0,__PYX_ERR(0, 771, __pyx_L1_error))
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 771, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_time_validator, __pyx_t_2) < 0) __PYX_ERR(0, 772, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "validr/_validator_c.pyx":793
 * 
 * 
 * @validator(accept=(str, datetime.datetime), output=(str, datetime.datetime))             # <<<<<<<<<<<<<<
 * def datetime_validator(compiler, str format='%Y-%m-%dT%H:%M:%S.%fZ', bint output_object=False):
 *     """Validate datetime string or convert datetime to string
 */
  __Pyx_TraceLine(793,0,__PYX_ERR(0, 793, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_validator); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 793, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 793, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_datetime); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 793, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_datetime); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 793, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 793, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)(&PyUnicode_Type)));
  __Pyx_GIVEREF(((PyObject *)(&PyUnicode_Type)));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)(&PyUnicode_Type)));
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_accept, __pyx_t_3) < 0) __PYX_ERR(0, 793, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_datetime); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 793, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_datetime); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 793, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 793, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)(&PyUnicode_Type)));
  __Pyx_GIVEREF(((PyObject *)(&PyUnicode_Type)));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)(&PyUnicode_Type)));
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_output, __pyx_t_3) < 0) __PYX_ERR(0, 793, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 793, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":794
 * 
 * @validator(accept=(str, datetime.datetime), output=(str, datetime.datetime))
 * def datetime_validator(compiler, str format='%Y-%m-%dT%H:%M:%S.%fZ', bint output_object=False):             # <<<<<<<<<<<<<<
 *     """Validate datetime string or convert datetime to string
 * 
 */
  __Pyx_TraceLine(794,0,__PYX_ERR(0, 794, __pyx_L1_error))
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_59datetime_validator, NULL, __pyx_n_s_validr__validator_c); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 794, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "validr/_validator_c.pyx":793
 * 
 * 
 * @validator(accept=(str, datetime.datetime), output=(str, datetime.datetime))             # <<<<<<<<<<<<<<
 * def datetime_validator(compiler, str format='%Y-%m-%dT%H:%M:%S.%fZ', bint output_object=False):
 *     """Validate datetime string or convert datetime to string
 */
  __Pyx_TraceLine(793,0,__PYX_ERR(0, 793, __pyx_L1_error))
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 793, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_datetime_validator, __pyx_t_2) < 0) __PYX_ERR(0, 794, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "validr/_validator_c.pyx":815
 * 
 * 
 * cpdef _parse_timedelta(value):             # <<<<<<<<<<<<<<
 *     if isinstance(value, (int, float)):
 *         value = datetime.timedelta(seconds=value)
 */
  __Pyx_TraceLine(815,0,__PYX_ERR(0, 815, __pyx_L1_error))


  /* "validr/_validator_c.pyx":826
 * 
 * 
 * @validator(accept=(int, float, str, datetime.timedelta), output=(str, float, datetime.timedelta))             # <<<<<<<<<<<<<<
 * def timedelta_validator(compiler, min=None, max=None, bint string=False,
 *                         bint extended=False, bint output_object=False):
 */
  __Pyx_TraceLine(826,0,__PYX_ERR(0, 826, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_validator); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_datetime); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_timedelta); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)(&PyInt_Type)));
  __Pyx_GIVEREF(((PyObject *)(&PyInt_Type)));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)(&PyInt_Type)));
  __Pyx_INCREF(((PyObject *)(&PyFloat_Type)));
  __Pyx_GIVEREF(((PyObject *)(&PyFloat_Type)));
  PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)(&PyFloat_Type)));
  __Pyx_INCREF(((PyObject *)(&PyUnicode_Type)));
  __Pyx_GIVEREF(((PyObject *)(&PyUnicode_Type)));
  PyTuple_SET_ITEM(__pyx_t_3, 2, ((PyObject *)(&PyUnicode_Type)));
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_4);
  __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_accept, __pyx_t_3) < 0) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_datetime); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_timedelta); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)(&PyUnicode_Type)));
  __Pyx_GIVEREF(((PyObject *)(&PyUnicode_Type)));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)(&PyUnicode_Type)));
  __Pyx_INCREF(((PyObject *)(&PyFloat_Type)));
  __Pyx_GIVEREF(((PyObject *)(&PyFloat_Type)));
  PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)(&PyFloat_Type)));
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_4);
  __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_output, __pyx_t_3) < 0) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":827
 * 
 * @validator(accept=(int, float, str, datetime.timedelta), output=(str, float, datetime.timedelta))
 * def timedelta_validator(compiler, min=None, max=None, bint string=False,             # <<<<<<<<<<<<<<
 *                         bint extended=False, bint output_object=False):
 *     """Validate timedelta string or convert timedelta to string
 */
  __Pyx_TraceLine(827,0,__PYX_ERR(0, 827, __pyx_L1_error))
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_63timedelta_validator, NULL, __pyx_n_s_validr__validator_c); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 827, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "validr/_validator_c.pyx":826
 * 
 * 
 * @validator(accept=(int, float, str, datetime.timedelta), output=(str, float, datetime.timedelta))             # <<<<<<<<<<<<<<
 * def timedelta_validator(compiler, min=None, max=None, bint string=False,
 *                         bint extended=False, bint output_object=False):
 */
  __Pyx_TraceLine(826,0,__PYX_ERR(0, 826, __pyx_L1_error))
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_timedelta_validator, __pyx_t_2) < 0) __PYX_ERR(0, 827, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "validr/_validator_c.pyx":880
 * 
 * 
 * @validator(accept=(str, ipaddress.IPv4Address), output=(str, ipaddress.IPv4Address))             # <<<<<<<<<<<<<<
 * def ipv4_validator(compiler, bint output_object=False):
 *     def validate(value):
 */
  __Pyx_TraceLine(880,0,__PYX_ERR(0, 880, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_validator); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_ipaddress); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_IPv4Address); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)(&PyUnicode_Type)));
  __Pyx_GIVEREF(((PyObject *)(&PyUnicode_Type)));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)(&PyUnicode_Type)));
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_accept, __pyx_t_3) < 0) __PYX_ERR(0, 880, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_ipaddress); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_IPv4Address); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)(&PyUnicode_Type)));
  __Pyx_GIVEREF(((PyObject *)(&PyUnicode_Type)));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)(&PyUnicode_Type)));
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_output, __pyx_t_3) < 0) __PYX_ERR(0, 880, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":881
 * 
 * @validator(accept=(str, ipaddress.IPv4Address), output=(str, ipaddress.IPv4Address))
 * def ipv4_validator(compiler, bint output_object=False):             # <<<<<<<<<<<<<<
 *     def validate(value):
 *         try:
 */
  __Pyx_TraceLine(881,0,__PYX_ERR(0, 881, __pyx_L1_error))
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_65ipv4_validator, NULL, __pyx_n_s_validr__validator_c); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 881, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "validr/_validator_c.pyx":880
 * 
 * 
 * @validator(accept=(str, ipaddress.IPv4Address), output=(str, ipaddress.IPv4Address))             # <<<<<<<<<<<<<<
 * def ipv4_validator(compiler, bint output_object=False):
 *     def validate(value):
 */
  __Pyx_TraceLine(880,0,__PYX_ERR(0, 880, __pyx_L1_error))
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ipv4_validator, __pyx_t_2) < 0) __PYX_ERR(0, 881, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "validr/_validator_c.pyx":896
 * 
 * 
 * @validator(accept=(str, ipaddress.IPv6Address), output=(str, ipaddress.IPv6Address))             # <<<<<<<<<<<<<<
 * def ipv6_validator(compiler, bint output_object=False):
 *     def validate(value):
 */
  __Pyx_TraceLine(896,0,__PYX_ERR(0, 896, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_validator); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 896, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 896, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_ipaddress); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 896, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_IPv6Address); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 896, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 896, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)(&PyUnicode_Type)));
  __Pyx_GIVEREF(((PyObject *)(&PyUnicode_Type)));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)(&PyUnicode_Type)));
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_accept, __pyx_t_3) < 0) __PYX_ERR(0, 896, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_ipaddress); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 896, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_IPv6Address); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 896, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 896, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)(&PyUnicode_Type)));
  __Pyx_GIVEREF(((PyObject *)(&PyUnicode_Type)));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)(&PyUnicode_Type)));
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_output, __pyx_t_3) < 0) __PYX_ERR(0, 896, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 896, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":897
 * 
 * @validator(accept=(str, ipaddress.IPv6Address), output=(str, ipaddress.IPv6Address))
 * def ipv6_validator(compiler, bint output_object=False):             # <<<<<<<<<<<<<<
 *     def validate(value):
 *         try:
 */
  __Pyx_TraceLine(897,0,__PYX_ERR(0, 897, __pyx_L1_error))
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_67ipv6_validator, NULL, __pyx_n_s_validr__validator_c); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 897, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "validr/_validator_c.pyx":896
 * 
 * 
 * @validator(accept=(str, ipaddress.IPv6Address), output=(str, ipaddress.IPv6Address))             # <<<<<<<<<<<<<<
 * def ipv6_validator(compiler, bint output_object=False):
 *     def validate(value):
 */
  __Pyx_TraceLine(896,0,__PYX_ERR(0, 896, __pyx_L1_error))
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 896, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ipv6_validator, __pyx_t_2) < 0) __PYX_ERR(0, 897, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "validr/_validator_c.pyx":912
 * 
 * 
 * @validator(accept=str, output=str)             # <<<<<<<<<<<<<<
 * def email_validator(compiler):
 *     # https://stackoverflow.com/questions/201323/using-a-regular-expression-to-validate-an-email-address
 */
  __Pyx_TraceLine(912,0,__PYX_ERR(0, 912, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_validator); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 912, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 912, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_accept, ((PyObject *)(&PyUnicode_Type))) < 0) __PYX_ERR(0, 912, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_output, ((PyObject *)(&PyUnicode_Type))) < 0) __PYX_ERR(0, 912, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 912, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":913
 * 
 * @validator(accept=str, output=str)
 * def email_validator(compiler):             # <<<<<<<<<<<<<<
 *     # https://stackoverflow.com/questions/201323/using-a-regular-expression-to-validate-an-email-address
 *     # http://emailregex.com/
 */
  __Pyx_TraceLine(913,0,__PYX_ERR(0, 913, __pyx_L1_error))
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_69email_validator, NULL, __pyx_n_s_validr__validator_c); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 913, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "validr/_validator_c.pyx":912
 * 
 * 
 * @validator(accept=str, output=str)             # <<<<<<<<<<<<<<
 * def email_validator(compiler):
 *     # https://stackoverflow.com/questions/201323/using-a-regular-expression-to-validate-an-email-address
 */
  __Pyx_TraceLine(912,0,__PYX_ERR(0, 912, __pyx_L1_error))
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 912, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_email_validator, __pyx_t_2) < 0) __PYX_ERR(0, 913, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "validr/_validator_c.pyx":935
 * 
 * 
 * @validator(output=(str, object))             # <<<<<<<<<<<<<<
 * def url_validator(compiler, str scheme='http https', int maxlen=255, bint output_object=False):
 *     # https://stackoverflow.com/questions/7160737/python-how-to-validate-a-url-in-python-malformed-or-not
 */
  __Pyx_TraceLine(935,0,__PYX_ERR(0, 935, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_validator); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 935, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 935, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 935, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)(&PyUnicode_Type)));
  __Pyx_GIVEREF(((PyObject *)(&PyUnicode_Type)));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)(&PyUnicode_Type)));
  __Pyx_INCREF(__pyx_builtin_object);
  __Pyx_GIVEREF(__pyx_builtin_object);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_builtin_object);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_output, __pyx_t_3) < 0) __PYX_ERR(0, 935, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 935, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":936
 * 
 * @validator(output=(str, object))
 * def url_validator(compiler, str scheme='http https', int maxlen=255, bint output_object=False):             # <<<<<<<<<<<<<<
 *     # https://stackoverflow.com/questions/7160737/python-how-to-validate-a-url-in-python-malformed-or-not
 *     # https://stackoverflow.com/questions/827557/how-do-you-validate-a-url-with-a-regular-expression-in-python
 */
  __Pyx_TraceLine(936,0,__PYX_ERR(0, 936, __pyx_L1_error))
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_71url_validator, NULL, __pyx_n_s_validr__validator_c); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 936, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "validr/_validator_c.pyx":935
 * 
 * 
 * @validator(output=(str, object))             # <<<<<<<<<<<<<<
 * def url_validator(compiler, str scheme='http https', int maxlen=255, bint output_object=False):
 *     # https://stackoverflow.com/questions/7160737/python-how-to-validate-a-url-in-python-malformed-or-not
 */
  __Pyx_TraceLine(935,0,__PYX_ERR(0, 935, __pyx_L1_error))
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 935, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_url_validator, __pyx_t_2) < 0) __PYX_ERR(0, 936, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "validr/_validator_c.pyx":964
 * 
 * 
 * @validator(output=str)             # <<<<<<<<<<<<<<
 * def fqdn_validator(compiler):
 *     def validate(value):
 */
  __Pyx_TraceLine(964,0,__PYX_ERR(0, 964, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_validator); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 964, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 964, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_output, ((PyObject *)(&PyUnicode_Type))) < 0) __PYX_ERR(0, 964, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 964, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":965
 * 
 * @validator(output=str)
 * def fqdn_validator(compiler):             # <<<<<<<<<<<<<<
 *     def validate(value):
 *         try:
 */
  __Pyx_TraceLine(965,0,__PYX_ERR(0, 965, __pyx_L1_error))
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_73fqdn_validator, NULL, __pyx_n_s_validr__validator_c); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 965, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "validr/_validator_c.pyx":964
 * 
 * 
 * @validator(output=str)             # <<<<<<<<<<<<<<
 * def fqdn_validator(compiler):
 *     def validate(value):
 */
  __Pyx_TraceLine(964,0,__PYX_ERR(0, 964, __pyx_L1_error))
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 964, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_fqdn_validator, __pyx_t_2) < 0) __PYX_ERR(0, 965, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "validr/_validator_c.pyx":978
 * 
 * 
 * @validator(output=(str, uuid.UUID))             # <<<<<<<<<<<<<<
 * def uuid_validator(compiler, version=None, bint output_object=False):
 *     if version is None:
 */
  __Pyx_TraceLine(978,0,__PYX_ERR(0, 978, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_validator); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 978, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 978, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_uuid); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 978, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_UUID); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 978, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 978, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)(&PyUnicode_Type)));
  __Pyx_GIVEREF(((PyObject *)(&PyUnicode_Type)));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)(&PyUnicode_Type)));
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_output, __pyx_t_3) < 0) __PYX_ERR(0, 978, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 978, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":979
 * 
 * @validator(output=(str, uuid.UUID))
 * def uuid_validator(compiler, version=None, bint output_object=False):             # <<<<<<<<<<<<<<
 *     if version is None:
 *         msg = 'invalid uuid'
 */
  __Pyx_TraceLine(979,0,__PYX_ERR(0, 979, __pyx_L1_error))
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_75uuid_validator, NULL, __pyx_n_s_validr__validator_c); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 979, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "validr/_validator_c.pyx":978
 * 
 * 
 * @validator(output=(str, uuid.UUID))             # <<<<<<<<<<<<<<
 * def uuid_validator(compiler, version=None, bint output_object=False):
 *     if version is None:
 */
  __Pyx_TraceLine(978,0,__PYX_ERR(0, 978, __pyx_L1_error))
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 978, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_uuid_validator, __pyx_t_2) < 0) __PYX_ERR(0, 979, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "validr/_validator_c.pyx":1002
 * 
 * 
 * def create_re_validator(str name, str r, int default_maxlen=255):             # <<<<<<<<<<<<<<
 *     """Create validator by regex string
 * 
 */
  __Pyx_TraceLine(1002,0,__PYX_ERR(0, 1002, __pyx_L1_error))
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_77create_re_validator, NULL, __pyx_n_s_validr__validator_c); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1002, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_create_re_validator, __pyx_t_2) < 0) __PYX_ERR(0, 1002, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "validr/_validator_c.pyx":1038
 * 
 * builtin_validators = {
 *     'list': list_validator,             # <<<<<<<<<<<<<<
 *     'dict': dict_validator,
 *     'model': model_validator,
 */
  __Pyx_TraceLine(1038,0,__PYX_ERR(0, 1038, __pyx_L1_error))
  __pyx_t_2 = __Pyx_PyDict_NewPresized(21); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1038, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_list_validator); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1038, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_list, __pyx_t_1) < 0) __PYX_ERR(0, 1038, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":1039
 * builtin_validators = {
 *     'list': list_validator,
 *     'dict': dict_validator,             # <<<<<<<<<<<<<<
 *     'model': model_validator,
 *     'union': union_validator,
 */
  __Pyx_TraceLine(1039,0,__PYX_ERR(0, 1039, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_dict_validator); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1039, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_dict, __pyx_t_1) < 0) __PYX_ERR(0, 1038, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":1040
 *     'list': list_validator,
 *     'dict': dict_validator,
 *     'model': model_validator,             # <<<<<<<<<<<<<<
 *     'union': union_validator,
 *     'enum': enum_validator,
 */
  __Pyx_TraceLine(1040,0,__PYX_ERR(0, 1040, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_model_validator); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1040, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_model, __pyx_t_1) < 0) __PYX_ERR(0, 1038, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":1041
 *     'dict': dict_validator,
 *     'model': model_validator,
 *     'union': union_validator,             # <<<<<<<<<<<<<<
 *     'enum': enum_validator,
 *     'any': any_validator,
 */
  __Pyx_TraceLine(1041,0,__PYX_ERR(0, 1041, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_union_validator); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1041, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_union, __pyx_t_1) < 0) __PYX_ERR(0, 1038, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":1042
 *     'model': model_validator,
 *     'union': union_validator,
 *     'enum': enum_validator,             # <<<<<<<<<<<<<<
 *     'any': any_validator,
 *     'int': int_validator,
 */
  __Pyx_TraceLine(1042,0,__PYX_ERR(0, 1042, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_enum_validator); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1042, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_enum, __pyx_t_1) < 0) __PYX_ERR(0, 1038, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":1043
 *     'union': union_validator,
 *     'enum': enum_validator,
 *     'any': any_validator,             # <<<<<<<<<<<<<<
 *     'int': int_validator,
 *     'bool': bool_validator,
 */
  __Pyx_TraceLine(1043,0,__PYX_ERR(0, 1043, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_any_validator); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1043, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_any, __pyx_t_1) < 0) __PYX_ERR(0, 1038, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":1044
 *     'enum': enum_validator,
 *     'any': any_validator,
 *     'int': int_validator,             # <<<<<<<<<<<<<<
 *     'bool': bool_validator,
 *     'float': float_validator,
 */
  __Pyx_TraceLine(1044,0,__PYX_ERR(0, 1044, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_int_validator); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1044, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_int, __pyx_t_1) < 0) __PYX_ERR(0, 1038, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":1045
 *     'any': any_validator,
 *     'int': int_validator,
 *     'bool': bool_validator,             # <<<<<<<<<<<<<<
 *     'float': float_validator,
 *     'str': str_validator,
 */
  __Pyx_TraceLine(1045,0,__PYX_ERR(0, 1045, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_bool_validator); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1045, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_bool, __pyx_t_1) < 0) __PYX_ERR(0, 1038, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":1046
 *     'int': int_validator,
 *     'bool': bool_validator,
 *     'float': float_validator,             # <<<<<<<<<<<<<<
 *     'str': str_validator,
 *     'bytes': bytes_validator,
 */
  __Pyx_TraceLine(1046,0,__PYX_ERR(0, 1046, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_float_validator); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1046, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_float, __pyx_t_1) < 0) __PYX_ERR(0, 1038, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":1047
 *     'bool': bool_validator,
 *     'float': float_validator,
 *     'str': str_validator,             # <<<<<<<<<<<<<<
 *     'bytes': bytes_validator,
 *     'date': date_validator,
 */
  __Pyx_TraceLine(1047,0,__PYX_ERR(0, 1047, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_str_validator); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1047, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_str, __pyx_t_1) < 0) __PYX_ERR(0, 1038, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":1048
 *     'float': float_validator,
 *     'str': str_validator,
 *     'bytes': bytes_validator,             # <<<<<<<<<<<<<<
 *     'date': date_validator,
 *     'time': time_validator,
 */
  __Pyx_TraceLine(1048,0,__PYX_ERR(0, 1048, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_bytes_validator); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1048, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_bytes, __pyx_t_1) < 0) __PYX_ERR(0, 1038, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":1049
 *     'str': str_validator,
 *     'bytes': bytes_validator,
 *     'date': date_validator,             # <<<<<<<<<<<<<<
 *     'time': time_validator,
 *     'datetime': datetime_validator,
 */
  __Pyx_TraceLine(1049,0,__PYX_ERR(0, 1049, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_date_validator); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1049, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_date, __pyx_t_1) < 0) __PYX_ERR(0, 1038, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":1050
 *     'bytes': bytes_validator,
 *     'date': date_validator,
 *     'time': time_validator,             # <<<<<<<<<<<<<<
 *     'datetime': datetime_validator,
 *     'timedelta': timedelta_validator,
 */
  __Pyx_TraceLine(1050,0,__PYX_ERR(0, 1050, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_time_validator); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1050, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_time, __pyx_t_1) < 0) __PYX_ERR(0, 1038, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":1051
 *     'date': date_validator,
 *     'time': time_validator,
 *     'datetime': datetime_validator,             # <<<<<<<<<<<<<<
 *     'timedelta': timedelta_validator,
 *     'ipv4': ipv4_validator,
 */
  __Pyx_TraceLine(1051,0,__PYX_ERR(0, 1051, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_datetime_validator); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1051, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_datetime, __pyx_t_1) < 0) __PYX_ERR(0, 1038, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":1052
 *     'time': time_validator,
 *     'datetime': datetime_validator,
 *     'timedelta': timedelta_validator,             # <<<<<<<<<<<<<<
 *     'ipv4': ipv4_validator,
 *     'ipv6': ipv6_validator,
 */
  __Pyx_TraceLine(1052,0,__PYX_ERR(0, 1052, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_timedelta_validator); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1052, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_timedelta, __pyx_t_1) < 0) __PYX_ERR(0, 1038, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":1053
 *     'datetime': datetime_validator,
 *     'timedelta': timedelta_validator,
 *     'ipv4': ipv4_validator,             # <<<<<<<<<<<<<<
 *     'ipv6': ipv6_validator,
 *     'email': email_validator,
 */
  __Pyx_TraceLine(1053,0,__PYX_ERR(0, 1053, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ipv4_validator); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1053, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_ipv4, __pyx_t_1) < 0) __PYX_ERR(0, 1038, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":1054
 *     'timedelta': timedelta_validator,
 *     'ipv4': ipv4_validator,
 *     'ipv6': ipv6_validator,             # <<<<<<<<<<<<<<
 *     'email': email_validator,
 *     'url': url_validator,
 */
  __Pyx_TraceLine(1054,0,__PYX_ERR(0, 1054, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ipv6_validator); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1054, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_ipv6, __pyx_t_1) < 0) __PYX_ERR(0, 1038, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":1055
 *     'ipv4': ipv4_validator,
 *     'ipv6': ipv6_validator,
 *     'email': email_validator,             # <<<<<<<<<<<<<<
 *     'url': url_validator,
 *     'fqdn': fqdn_validator,
 */
  __Pyx_TraceLine(1055,0,__PYX_ERR(0, 1055, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_email_validator); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1055, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_email, __pyx_t_1) < 0) __PYX_ERR(0, 1038, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":1056
 *     'ipv6': ipv6_validator,
 *     'email': email_validator,
 *     'url': url_validator,             # <<<<<<<<<<<<<<
 *     'fqdn': fqdn_validator,
 *     'uuid': uuid_validator,
 */
  __Pyx_TraceLine(1056,0,__PYX_ERR(0, 1056, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_url_validator); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1056, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_url, __pyx_t_1) < 0) __PYX_ERR(0, 1038, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":1057
 *     'email': email_validator,
 *     'url': url_validator,
 *     'fqdn': fqdn_validator,             # <<<<<<<<<<<<<<
 *     'uuid': uuid_validator,
 * }
 */
  __Pyx_TraceLine(1057,0,__PYX_ERR(0, 1057, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_fqdn_validator); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1057, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_fqdn, __pyx_t_1) < 0) __PYX_ERR(0, 1038, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "validr/_validator_c.pyx":1058
 *     'url': url_validator,
 *     'fqdn': fqdn_validator,
 *     'uuid': uuid_validator,             # <<<<<<<<<<<<<<
 * }
 * 
 */
  __Pyx_TraceLine(1058,0,__PYX_ERR(0, 1058, __pyx_L1_error))
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_uuid_validator); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1058, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_uuid, __pyx_t_1) < 0) __PYX_ERR(0, 1038, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_builtin_validators, __pyx_t_2) < 0) __PYX_ERR(0, 1037, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "validr/_validator_c.pyx":1062
 * 
 * regexs = {
 *     'phone': (r'((\+\d{2}\s?)|(\d{2}\s?))?1\d{10}', 15),             # <<<<<<<<<<<<<<
 *     'idcard': (r'(\d{17}[\d|x|X])|(\d{15})', 18),
 *     'slug': (r'[a-z0-9]+(?:-[a-z0-9]+)*', 255),
 */
  __Pyx_TraceLine(1062,0,__PYX_ERR(0, 1062, __pyx_L1_error))
  __pyx_t_2 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1062, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_phone, __pyx_tuple__149) < 0) __PYX_ERR(0, 1062, __pyx_L1_error)

  /* "validr/_validator_c.pyx":1063
 * regexs = {
 *     'phone': (r'((\+\d{2}\s?)|(\d{2}\s?))?1\d{10}', 15),
 *     'idcard': (r'(\d{17}[\d|x|X])|(\d{15})', 18),             # <<<<<<<<<<<<<<
 *     'slug': (r'[a-z0-9]+(?:-[a-z0-9]+)*', 255),
 * }
 */
  __Pyx_TraceLine(1063,0,__PYX_ERR(0, 1063, __pyx_L1_error))
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_idcard, __pyx_tuple__150) < 0) __PYX_ERR(0, 1062, __pyx_L1_error)

  /* "validr/_validator_c.pyx":1064
 *     'phone': (r'((\+\d{2}\s?)|(\d{2}\s?))?1\d{10}', 15),
 *     'idcard': (r'(\d{17}[\d|x|X])|(\d{15})', 18),
 *     'slug': (r'[a-z0-9]+(?:-[a-z0-9]+)*', 255),             # <<<<<<<<<<<<<<
 * }
 * for name, options in regexs.items():
 */
  __Pyx_TraceLine(1064,0,__PYX_ERR(0, 1064, __pyx_L1_error))
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_slug, __pyx_tuple__151) < 0) __PYX_ERR(0, 1062, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_regexs, __pyx_t_2) < 0) __PYX_ERR(0, 1061, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "validr/_validator_c.pyx":1066
 *     'slug': (r'[a-z0-9]+(?:-[a-z0-9]+)*', 255),
 * }
 * for name, options in regexs.items():             # <<<<<<<<<<<<<<
 *     builtin_validators[name] = create_re_validator(name, *options)
 * 
 */
  __Pyx_TraceLine(1066,0,__PYX_ERR(0, 1066, __pyx_L1_error))
  __pyx_t_6 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_regexs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1066, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
    __PYX_ERR(0, 1066, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_dict_iterator(__pyx_t_1, 0, __pyx_n_s_items, (&__pyx_t_7), (&__pyx_t_8)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1066, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_XDECREF(__pyx_t_2);
  __pyx_t_2 = __pyx_t_3;
  __pyx_t_3 = 0;
  while (1) {
    __pyx_t_9 = __Pyx_dict_iter_next(__pyx_t_2, __pyx_t_7, &__pyx_t_6, &__pyx_t_3, &__pyx_t_1, NULL, __pyx_t_8);
    if (unlikely(__pyx_t_9 == 0)) break;
    if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(0, 1066, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_name_2, __pyx_t_3) < 0) __PYX_ERR(0, 1066, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_options, __pyx_t_1) < 0) __PYX_ERR(0, 1066, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "validr/_validator_c.pyx":1067
 * }
 * for name, options in regexs.items():
 *     builtin_validators[name] = create_re_validator(name, *options)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_TraceLine(1067,0,__PYX_ERR(0, 1067, __pyx_L1_error))
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_create_re_validator); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1067, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_name_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1067, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1067, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
    __pyx_t_3 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_options); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1067, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PySequence_Tuple(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1067, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Add(__pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1067, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1067, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_builtin_validators); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1067, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1067, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(PyObject_SetItem(__pyx_t_3, __pyx_t_1, __pyx_t_5) < 0)) __PYX_ERR(0, 1067, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "validr/_validator_c.pyx":1070
 * 
 * 
 * def create_enum_validator(str name, items, bint string=True):             # <<<<<<<<<<<<<<
 *     """Create validator by enum items
 * 
 */
  __Pyx_TraceLine(1070,0,__PYX_ERR(0, 1070, __pyx_L1_error))
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_6validr_12_validator_c_79create_enum_validator, NULL, __pyx_n_s_validr__validator_c); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1070, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_create_enum_validator, __pyx_t_2) < 0) __PYX_ERR(0, 1070, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "validr/_validator_c.pyx":1
 * import re             # <<<<<<<<<<<<<<
 * import sys
 * import uuid
 */
  __Pyx_TraceLine(1,0,__PYX_ERR(0, 1, __pyx_L1_error))
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_TraceReturn(Py_None, 0);

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init validr._validator_c", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init validr._validator_c");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* Profile */
#if CYTHON_PROFILE
static int __Pyx_TraceSetupAndCall(PyCodeObject** code,
                                   PyFrameObject** frame,
                                   PyThreadState* tstate,
                                   const char *funcname,
                                   const char *srcfile,
                                   int firstlineno) {
    PyObject *type, *value, *traceback;
    int retval;
    if (*frame == NULL || !CYTHON_PROFILE_REUSE_FRAME) {
        if (*code == NULL) {
            *code = __Pyx_createFrameCodeObject(funcname, srcfile, firstlineno);
            if (*code == NULL) return 0;
        }
        *frame = PyFrame_New(
            tstate,                          /*PyThreadState *tstate*/
            *code,                           /*PyCodeObject *code*/
            __pyx_d,                  /*PyObject *globals*/
            0                                /*PyObject *locals*/
        );
        if (*frame == NULL) return 0;
        if (CYTHON_TRACE && (*frame)->f_trace == NULL) {
            Py_INCREF(Py_None);
            (*frame)->f_trace = Py_None;
        }
#if PY_VERSION_HEX < 0x030400B1
    } else {
        (*frame)->f_tstate = tstate;
#endif
    }
      __Pyx_PyFrame_SetLineNumber(*frame, firstlineno);
    retval = 1;
    tstate->tracing++;
    tstate->use_tracing = 0;
    __Pyx_ErrFetchInState(tstate, &type, &value, &traceback);
    #if CYTHON_TRACE
    if (tstate->c_tracefunc)
        retval = tstate->c_tracefunc(tstate->c_traceobj, *frame, PyTrace_CALL, NULL) == 0;
    if (retval && tstate->c_profilefunc)
    #endif
        retval = tstate->c_profilefunc(tstate->c_profileobj, *frame, PyTrace_CALL, NULL) == 0;
    tstate->use_tracing = (tstate->c_profilefunc ||
                           (CYTHON_TRACE && tstate->c_tracefunc));
    tstate->tracing--;
    if (retval) {
        __Pyx_ErrRestoreInState(tstate, type, value, traceback);
        return tstate->use_tracing && retval;
    } else {
        Py_XDECREF(type);
        Py_XDECREF(value);
        Py_XDECREF(traceback);
        return -1;
    }
}
static PyCodeObject *__Pyx_createFrameCodeObject(const char *funcname, const char *srcfile, int firstlineno) {
    PyCodeObject *py_code = 0;
#if PY_MAJOR_VERSION >= 3
    py_code = PyCode_NewEmpty(srcfile, funcname, firstlineno);
    if (likely(py_code)) {
        py_code->co_flags |= CO_OPTIMIZED | CO_NEWLOCALS;
    }
#else
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    py_funcname = PyString_FromString(funcname);
    if (unlikely(!py_funcname)) goto bad;
    py_srcfile = PyString_FromString(srcfile);
    if (unlikely(!py_srcfile)) goto bad;
    py_code = PyCode_New(
        0,
        0,
        0,
        CO_OPTIMIZED | CO_NEWLOCALS,
        __pyx_empty_bytes,     /*PyObject *code,*/
        __pyx_empty_tuple,     /*PyObject *consts,*/
        __pyx_empty_tuple,     /*PyObject *names,*/
        __pyx_empty_tuple,     /*PyObject *varnames,*/
        __pyx_empty_tuple,     /*PyObject *freevars,*/
        __pyx_empty_tuple,     /*PyObject *cellvars,*/
        py_srcfile,       /*PyObject *filename,*/
        py_funcname,      /*PyObject *name,*/
        firstlineno,
        __pyx_empty_bytes      /*PyObject *lnotab*/
    );
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
#endif
    return py_code;
}
#endif

/* GetAttr */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* HasAttr */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (unlikely(!r)) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
#ifdef _MSC_VER
    else state = (PyGILState_STATE)-1;
#endif
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_CheckExact(key)) || likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (PyUnicode_GET_SIZE(**name) != PyUnicode_GET_SIZE(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (PyUnicode_GET_SIZE(**argname) != PyUnicode_GET_SIZE(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetAttr3 */
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r = __Pyx_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
}

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (PyCFunction_GET_FLAGS(func) & METH_FASTCALL) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* None */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname) {
    PyErr_Format(PyExc_NameError, "free variable '%s' referenced before assignment in enclosing scope", varname);
}

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* FetchCommonType */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CythonFunction */
#include <structmember.h>
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *closure)
{
    if (unlikely(op->func_doc == NULL)) {
        if (op->func.m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(op->func.m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(op->func.m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp = op->func_doc;
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    op->func_doc = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_name == NULL)) {
#if PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(op->func.m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(op->func.m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = op->func_name;
    Py_INCREF(value);
    op->func_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = op->func_qualname;
    Py_INCREF(value);
    op->func_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_self(__pyx_CyFunctionObject *m, CYTHON_UNUSED void *closure)
{
    PyObject *self;
    self = m->func_closure;
    if (self == NULL)
        self = Py_None;
    Py_INCREF(self);
    return self;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    tmp = op->func_dict;
    Py_INCREF(value);
    op->func_dict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(CYTHON_UNUSED __pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyTuple_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_tuple;
    op->defaults_tuple = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_kwdict;
    op->defaults_kwdict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value || value == Py_None) {
        value = NULL;
    } else if (!PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    tmp = op->func_annotations;
    op->func_annotations = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "__self__", (getter)__Pyx_CyFunction_get_self, 0, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
    {(char *) "__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), PY_WRITE_RESTRICTED, 0},
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, CYTHON_UNUSED PyObject *args)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromString(m->func.m_ml->ml_name);
#else
    return PyString_FromString(m->func.m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func.m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_New(PyTypeObject *type, PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    __pyx_CyFunctionObject *op = PyObject_GC_New(__pyx_CyFunctionObject, type);
    if (op == NULL)
        return NULL;
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
    op->func.m_ml = ml;
    op->func.m_self = (PyObject *) op;
    Py_XINCREF(closure);
    op->func_closure = closure;
    Py_XINCREF(module);
    op->func.m_module = module;
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
    op->func_classobj = NULL;
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    PyObject_GC_Track(op);
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
    Py_CLEAR(m->func.m_module);
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
    Py_CLEAR(m->func_classobj);
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    PyObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
    Py_VISIT(m->func.m_module);
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
    Py_VISIT(m->func_classobj);
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject *__Pyx_CyFunction_descr_get(PyObject *func, PyObject *obj, PyObject *type)
{
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    if (m->flags & __Pyx_CYFUNCTION_STATICMETHOD) {
        Py_INCREF(func);
        return func;
    }
    if (m->flags & __Pyx_CYFUNCTION_CLASSMETHOD) {
        if (type == NULL)
            type = (PyObject *)(Py_TYPE(obj));
        return __Pyx_PyMethod_New(func, type, (PyObject *)(Py_TYPE(type)));
    }
    if (obj == Py_None)
        obj = NULL;
    return __Pyx_PyMethod_New(func, obj, type);
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    Py_ssize_t size;
    switch (f->m_ml->ml_flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void*)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 0))
                return (*meth)(self, NULL);
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags in "
                        "__Pyx_CyFunction_Call. METH_OLDARGS is no "
                        "longer supported!");
        return NULL;
    }
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    return __Pyx_CyFunction_CallMethod(func, ((PyCFunctionObject*)func)->m_self, arg, kw);
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
        argc = PyTuple_GET_SIZE(args);
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
    0,
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_CyFunction_descr_get,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
};
static int __pyx_CyFunction_init(void) {
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
    if (unlikely(__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (unlikely(!m->defaults))
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* DictGetItem */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (descr != NULL) {
        *method = descr;
        return 0;
    }
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(name));
#endif
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod0 */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
}

/* RaiseNoneIterError */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* UnpackTupleError */
static void __Pyx_UnpackTupleError(PyObject *t, Py_ssize_t index) {
    if (t == Py_None) {
      __Pyx_RaiseNoneNotIterableError();
    } else if (PyTuple_GET_SIZE(t) < index) {
      __Pyx_RaiseNeedMoreValuesError(PyTuple_GET_SIZE(t));
    } else {
      __Pyx_RaiseTooManyValuesError(index);
    }
}

/* UnpackTuple2 */
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
        PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2, int decref_tuple) {
    PyObject *value1 = NULL, *value2 = NULL;
#if CYTHON_COMPILING_IN_PYPY
    value1 = PySequence_ITEM(tuple, 0);  if (unlikely(!value1)) goto bad;
    value2 = PySequence_ITEM(tuple, 1);  if (unlikely(!value2)) goto bad;
#else
    value1 = PyTuple_GET_ITEM(tuple, 0);  Py_INCREF(value1);
    value2 = PyTuple_GET_ITEM(tuple, 1);  Py_INCREF(value2);
#endif
    if (decref_tuple) {
        Py_DECREF(tuple);
    }
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
#if CYTHON_COMPILING_IN_PYPY
bad:
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
#endif
}
static int __Pyx_unpack_tuple2_generic(PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2,
                                       int has_known_size, int decref_tuple) {
    Py_ssize_t index;
    PyObject *value1 = NULL, *value2 = NULL, *iter = NULL;
    iternextfunc iternext;
    iter = PyObject_GetIter(tuple);
    if (unlikely(!iter)) goto bad;
    if (decref_tuple) { Py_DECREF(tuple); tuple = NULL; }
    iternext = Py_TYPE(iter)->tp_iternext;
    value1 = iternext(iter); if (unlikely(!value1)) { index = 0; goto unpacking_failed; }
    value2 = iternext(iter); if (unlikely(!value2)) { index = 1; goto unpacking_failed; }
    if (!has_known_size && unlikely(__Pyx_IternextUnpackEndCheck(iternext(iter), 2))) goto bad;
    Py_DECREF(iter);
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
unpacking_failed:
    if (!has_known_size && __Pyx_IterFinish() == 0)
        __Pyx_RaiseNeedMoreValuesError(index);
bad:
    Py_XDECREF(iter);
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
}

/* dict_iter */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* iterable, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_source_is_dict) {
    is_dict = is_dict || likely(PyDict_CheckExact(iterable));
    *p_source_is_dict = is_dict;
    if (is_dict) {
#if !CYTHON_COMPILING_IN_PYPY
        *p_orig_length = PyDict_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
#elif PY_MAJOR_VERSION >= 3
        static PyObject *py_items = NULL, *py_keys = NULL, *py_values = NULL;
        PyObject **pp = NULL;
        if (method_name) {
            const char *name = PyUnicode_AsUTF8(method_name);
            if (strcmp(name, "iteritems") == 0) pp = &py_items;
            else if (strcmp(name, "iterkeys") == 0) pp = &py_keys;
            else if (strcmp(name, "itervalues") == 0) pp = &py_values;
            if (pp) {
                if (!*pp) {
                    *pp = PyUnicode_FromString(name + 4);
                    if (!*pp)
                        return NULL;
                }
                method_name = *pp;
            }
        }
#endif
    }
    *p_orig_length = 0;
    if (method_name) {
        PyObject* iter;
        iterable = __Pyx_PyObject_CallMethod0(iterable, method_name);
        if (!iterable)
            return NULL;
#if !CYTHON_COMPILING_IN_PYPY
        if (PyTuple_CheckExact(iterable) || PyList_CheckExact(iterable))
            return iterable;
#endif
        iter = PyObject_GetIter(iterable);
        Py_DECREF(iterable);
        return iter;
    }
    return PyObject_GetIter(iterable);
}
static CYTHON_INLINE int __Pyx_dict_iter_next(
        PyObject* iter_obj, CYTHON_NCP_UNUSED Py_ssize_t orig_length, CYTHON_NCP_UNUSED Py_ssize_t* ppos,
        PyObject** pkey, PyObject** pvalue, PyObject** pitem, int source_is_dict) {
    PyObject* next_item;
#if !CYTHON_COMPILING_IN_PYPY
    if (source_is_dict) {
        PyObject *key, *value;
        if (unlikely(orig_length != PyDict_Size(iter_obj))) {
            PyErr_SetString(PyExc_RuntimeError, "dictionary changed size during iteration");
            return -1;
        }
        if (unlikely(!PyDict_Next(iter_obj, ppos, &key, &value))) {
            return 0;
        }
        if (pitem) {
            PyObject* tuple = PyTuple_New(2);
            if (unlikely(!tuple)) {
                return -1;
            }
            Py_INCREF(key);
            Py_INCREF(value);
            PyTuple_SET_ITEM(tuple, 0, key);
            PyTuple_SET_ITEM(tuple, 1, value);
            *pitem = tuple;
        } else {
            if (pkey) {
                Py_INCREF(key);
                *pkey = key;
            }
            if (pvalue) {
                Py_INCREF(value);
                *pvalue = value;
            }
        }
        return 1;
    } else if (PyTuple_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyTuple_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyTuple_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else if (PyList_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyList_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyList_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else
#endif
    {
        next_item = PyIter_Next(iter_obj);
        if (unlikely(!next_item)) {
            return __Pyx_IterFinish();
        }
    }
    if (pitem) {
        *pitem = next_item;
    } else if (pkey && pvalue) {
        if (__Pyx_unpack_tuple2(next_item, pkey, pvalue, source_is_dict, source_is_dict, 1))
            return -1;
    } else if (pkey) {
        *pkey = next_item;
    } else {
        *pvalue = next_item;
    }
    return 1;
}

/* None */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname) {
    PyErr_Format(PyExc_UnboundLocalError, "local variable '%s' referenced before assignment", varname);
}

/* pyfrozenset_new */
static CYTHON_INLINE PyObject* __Pyx_PyFrozenSet_New(PyObject* it) {
    if (it) {
        PyObject* result;
#if CYTHON_COMPILING_IN_PYPY
        PyObject* args;
        args = PyTuple_Pack(1, it);
        if (unlikely(!args))
            return NULL;
        result = PyObject_Call((PyObject*)&PyFrozenSet_Type, args, NULL);
        Py_DECREF(args);
        return result;
#else
        if (PyFrozenSet_CheckExact(it)) {
            Py_INCREF(it);
            return it;
        }
        result = PyFrozenSet_New(it);
        if (unlikely(!result))
            return NULL;
        if (likely(PySet_GET_SIZE(result)))
            return result;
        Py_DECREF(result);
#endif
    }
#if CYTHON_USE_TYPE_SLOTS
    return PyFrozenSet_Type.tp_new(&PyFrozenSet_Type, __pyx_empty_tuple, NULL);
#else
    return PyObject_Call((PyObject*)&PyFrozenSet_Type, __pyx_empty_tuple, NULL);
#endif
}

/* PySetContains */
static int __Pyx_PySet_ContainsUnhashable(PyObject *set, PyObject *key) {
    int result = -1;
    if (PySet_Check(key) && PyErr_ExceptionMatches(PyExc_TypeError)) {
        PyObject *tmpkey;
        PyErr_Clear();
        tmpkey = __Pyx_PyFrozenSet_New(key);
        if (tmpkey != NULL) {
            result = PySet_Contains(set, tmpkey);
            Py_DECREF(tmpkey);
        }
    }
    return result;
}
static CYTHON_INLINE int __Pyx_PySet_ContainsTF(PyObject* key, PyObject* set, int eq) {
    int result = PySet_Contains(set, key);
    if (unlikely(result < 0)) {
        result = __Pyx_PySet_ContainsUnhashable(set, key);
    }
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a + b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* UnpackUnboundCMethod */
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    target->method = method;
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION >= 3
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    #endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    }
#endif
    return 0;
}

/* CallUnboundCMethod1 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg) {
    if (likely(cfunc->func)) {
        int flag = cfunc->flag;
        if (flag == METH_O) {
            return (*(cfunc->func))(self, arg);
        } else if (PY_VERSION_HEX >= 0x030600B1 && flag == METH_FASTCALL) {
            if (PY_VERSION_HEX >= 0x030700A0) {
                return (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)cfunc->func)(self, &arg, 1);
            } else {
                return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, &arg, 1, NULL);
            }
        } else if (PY_VERSION_HEX >= 0x030700A0 && flag == (METH_FASTCALL | METH_KEYWORDS)) {
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, &arg, 1, NULL);
        }
    }
    return __Pyx__CallUnboundCMethod1(cfunc, self, arg);
}
#endif
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg){
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        args = PyTuple_New(1);
        if (unlikely(!args)) goto bad;
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 0, arg);
        if (cfunc->flag & METH_KEYWORDS)
            result = (*(PyCFunctionWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, NULL);
        else
            result = (*cfunc->func)(self, args);
    } else {
        args = PyTuple_New(2);
        if (unlikely(!args)) goto bad;
        Py_INCREF(self);
        PyTuple_SET_ITEM(args, 0, self);
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 1, arg);
        result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    }
#else
    args = PyTuple_Pack(2, self, arg);
    if (unlikely(!args)) goto bad;
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
#endif
bad:
    Py_XDECREF(args);
    return result;
}

/* CallUnboundCMethod2 */
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030600B1
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2) {
    if (likely(cfunc->func)) {
        PyObject *args[2] = {arg1, arg2};
        if (cfunc->flag == METH_FASTCALL) {
            #if PY_VERSION_HEX >= 0x030700A0
            return (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)cfunc->func)(self, args, 2);
            #else
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, 2, NULL);
            #endif
        }
        #if PY_VERSION_HEX >= 0x030700A0
        if (cfunc->flag == (METH_FASTCALL | METH_KEYWORDS))
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, 2, NULL);
        #endif
    }
    return __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2);
}
#endif
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2){
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        args = PyTuple_New(2);
        if (unlikely(!args)) goto bad;
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 0, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 1, arg2);
        if (cfunc->flag & METH_KEYWORDS)
            result = (*(PyCFunctionWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, NULL);
        else
            result = (*cfunc->func)(self, args);
    } else {
        args = PyTuple_New(3);
        if (unlikely(!args)) goto bad;
        Py_INCREF(self);
        PyTuple_SET_ITEM(args, 0, self);
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 1, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 2, arg2);
        result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    }
#else
    args = PyTuple_Pack(3, self, arg1, arg2);
    if (unlikely(!args)) goto bad;
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
#endif
bad:
    Py_XDECREF(args);
    return result;
}

/* dict_getitem_default */
static PyObject* __Pyx_PyDict_GetItemDefault(PyObject* d, PyObject* key, PyObject* default_value) {
    PyObject* value;
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (unlikely(PyErr_Occurred()))
            return NULL;
        value = default_value;
    }
    Py_INCREF(value);
    if ((1));
#else
    if (PyString_CheckExact(key) || PyUnicode_CheckExact(key) || PyInt_CheckExact(key)) {
        value = PyDict_GetItem(d, key);
        if (unlikely(!value)) {
            value = default_value;
        }
        Py_INCREF(value);
    }
#endif
    else {
        if (default_value == Py_None)
            value = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyDict_Type_get, d, key);
        else
            value = __Pyx_CallUnboundCMethod2(&__pyx_umethod_PyDict_Type_get, d, key, default_value);
    }
    return value;
}

/* CallUnboundCMethod0 */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_ASSUME_SAFE_MACROS
    args = PyTuple_New(1);
    if (unlikely(!args)) goto bad;
    Py_INCREF(self);
    PyTuple_SET_ITEM(args, 0, self);
#else
    args = PyTuple_Pack(1, self);
    if (unlikely(!args)) goto bad;
#endif
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    Py_DECREF(args);
bad:
    return result;
}

/* py_dict_keys */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Keys(PyObject* d) {
    if (PY_MAJOR_VERSION >= 3)
        return __Pyx_CallUnboundCMethod0(&__pyx_umethod_PyDict_Type_keys, d);
    else
        return PyDict_Keys(d);
}

/* pyobject_as_double */
static double __Pyx__PyObject_AsDouble(PyObject* obj) {
    PyObject* float_value;
#if !CYTHON_USE_TYPE_SLOTS
    float_value = PyNumber_Float(obj);  if ((0)) goto bad;
#else
    PyNumberMethods *nb = Py_TYPE(obj)->tp_as_number;
    if (likely(nb) && likely(nb->nb_float)) {
        float_value = nb->nb_float(obj);
        if (likely(float_value) && unlikely(!PyFloat_Check(float_value))) {
            PyErr_Format(PyExc_TypeError,
                "__float__ returned non-float (type %.200s)",
                Py_TYPE(float_value)->tp_name);
            Py_DECREF(float_value);
            goto bad;
        }
    } else if (PyUnicode_CheckExact(obj) || PyBytes_CheckExact(obj)) {
#if PY_MAJOR_VERSION >= 3
        float_value = PyFloat_FromString(obj);
#else
        float_value = PyFloat_FromString(obj, 0);
#endif
    } else {
        PyObject* args = PyTuple_New(1);
        if (unlikely(!args)) goto bad;
        PyTuple_SET_ITEM(args, 0, obj);
        float_value = PyObject_Call((PyObject*)&PyFloat_Type, args, 0);
        PyTuple_SET_ITEM(args, 0, 0);
        Py_DECREF(args);
    }
#endif
    if (likely(float_value)) {
        double value = PyFloat_AS_DOUBLE(float_value);
        Py_DECREF(float_value);
        return value;
    }
bad:
    return (double)-1;
}

/* PyIntCompare */
static CYTHON_INLINE PyObject* __Pyx_PyInt_NeObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, CYTHON_UNUSED long inplace) {
    if (op1 == op2) {
        Py_RETURN_FALSE;
    }
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        if (a != b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = Py_SIZE(op1);
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        if (intval == 0) {
            if (size != 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
        } else if (intval < 0) {
            if (size >= 0)
                Py_RETURN_TRUE;
            intval = -intval;
            size = -size;
        } else {
            if (size <= 0)
                Py_RETURN_TRUE;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        if (unequal != 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
        if ((double)a != (double)b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    return (
        PyObject_RichCompare(op1, op2, Py_NE));
}

/* PyUnicode_Unicode */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_Unicode(PyObject *obj) {
    if (unlikely(obj == Py_None))
        obj = __pyx_kp_u_None;
    return __Pyx_NewRef(obj);
}

/* CIntToDigits */
static const char DIGIT_PAIRS_10[2*10*10+1] = {
    "00010203040506070809"
    "10111213141516171819"
    "20212223242526272829"
    "30313233343536373839"
    "40414243444546474849"
    "50515253545556575859"
    "60616263646566676869"
    "70717273747576777879"
    "80818283848586878889"
    "90919293949596979899"
};
static const char DIGIT_PAIRS_8[2*8*8+1] = {
    "0001020304050607"
    "1011121314151617"
    "2021222324252627"
    "3031323334353637"
    "4041424344454647"
    "5051525354555657"
    "6061626364656667"
    "7071727374757677"
};
static const char DIGITS_HEX[2*16+1] = {
    "0123456789abcdef"
    "0123456789ABCDEF"
};

/* BuildPyUnicode */
static PyObject* __Pyx_PyUnicode_BuildFromAscii(Py_ssize_t ulength, char* chars, int clength,
                                                int prepend_sign, char padding_char) {
    PyObject *uval;
    Py_ssize_t uoffset = ulength - clength;
#if CYTHON_USE_UNICODE_INTERNALS
    Py_ssize_t i;
#if CYTHON_PEP393_ENABLED
    void *udata;
    uval = PyUnicode_New(ulength, 127);
    if (unlikely(!uval)) return NULL;
    udata = PyUnicode_DATA(uval);
#else
    Py_UNICODE *udata;
    uval = PyUnicode_FromUnicode(NULL, ulength);
    if (unlikely(!uval)) return NULL;
    udata = PyUnicode_AS_UNICODE(uval);
#endif
    if (uoffset > 0) {
        i = 0;
        if (prepend_sign) {
            __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, 0, '-');
            i++;
        }
        for (; i < uoffset; i++) {
            __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, i, padding_char);
        }
    }
    for (i=0; i < clength; i++) {
        __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, uoffset+i, chars[i]);
    }
#else
    {
        PyObject *sign = NULL, *padding = NULL;
        uval = NULL;
        if (uoffset > 0) {
            prepend_sign = !!prepend_sign;
            if (uoffset > prepend_sign) {
                padding = PyUnicode_FromOrdinal(padding_char);
                if (likely(padding) && uoffset > prepend_sign + 1) {
                    PyObject *tmp;
                    PyObject *repeat = PyInt_FromSize_t(uoffset - prepend_sign);
                    if (unlikely(!repeat)) goto done_or_error;
                    tmp = PyNumber_Multiply(padding, repeat);
                    Py_DECREF(repeat);
                    Py_DECREF(padding);
                    padding = tmp;
                }
                if (unlikely(!padding)) goto done_or_error;
            }
            if (prepend_sign) {
                sign = PyUnicode_FromOrdinal('-');
                if (unlikely(!sign)) goto done_or_error;
            }
        }
        uval = PyUnicode_DecodeASCII(chars, clength, NULL);
        if (likely(uval) && padding) {
            PyObject *tmp = PyNumber_Add(padding, uval);
            Py_DECREF(uval);
            uval = tmp;
        }
        if (likely(uval) && sign) {
            PyObject *tmp = PyNumber_Add(sign, uval);
            Py_DECREF(uval);
            uval = tmp;
        }
done_or_error:
        Py_XDECREF(padding);
        Py_XDECREF(sign);
    }
#endif
    return uval;
}

/* CIntToPyUnicode */
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned short    uint16_t;
        #else
           typedef unsigned __int16  uint16_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#if defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define GCC_DIAGNOSTIC
#endif
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_From_int(int value, Py_ssize_t width, char padding_char, char format_char) {
    char digits[sizeof(int)*3+2];
    char *dpos, *end = digits + sizeof(int)*3+2;
    const char *hex_digits = DIGITS_HEX;
    Py_ssize_t length, ulength;
    int prepend_sign, last_one_off;
    int remaining;
#ifdef GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (format_char == 'X') {
        hex_digits += 16;
        format_char = 'x';
    }
    remaining = value;
    last_one_off = 0;
    dpos = end;
    do {
        int digit_pos;
        switch (format_char) {
        case 'o':
            digit_pos = abs((int)(remaining % (8*8)));
            remaining = (int) (remaining / (8*8));
            dpos -= 2;
            *(uint16_t*)dpos = ((const uint16_t*)DIGIT_PAIRS_8)[digit_pos];
            last_one_off = (digit_pos < 8);
            break;
        case 'd':
            digit_pos = abs((int)(remaining % (10*10)));
            remaining = (int) (remaining / (10*10));
            dpos -= 2;
            *(uint16_t*)dpos = ((const uint16_t*)DIGIT_PAIRS_10)[digit_pos];
            last_one_off = (digit_pos < 10);
            break;
        case 'x':
            *(--dpos) = hex_digits[abs((int)(remaining % 16))];
            remaining = (int) (remaining / 16);
            break;
        default:
            assert(0);
            break;
        }
    } while (unlikely(remaining != 0));
    if (last_one_off) {
        assert(*dpos == '0');
        dpos++;
    }
    length = end - dpos;
    ulength = length;
    prepend_sign = 0;
    if (!is_unsigned && value <= neg_one) {
        if (padding_char == ' ' || width <= length + 1) {
            *(--dpos) = '-';
            ++length;
        } else {
            prepend_sign = 1;
        }
        ++ulength;
    }
    if (width > ulength) {
        ulength = width;
    }
    if (ulength == 1) {
        return PyUnicode_FromOrdinal(*dpos);
    }
    return __Pyx_PyUnicode_BuildFromAscii(ulength, dpos, (int) length, prepend_sign, padding_char);
}

/* JoinPyUnicode */
static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      CYTHON_UNUSED Py_UCS4 max_char) {
#if CYTHON_USE_UNICODE_INTERNALS && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    PyObject *result_uval;
    int result_ukind;
    Py_ssize_t i, char_pos;
    void *result_udata;
#if CYTHON_PEP393_ENABLED
    result_uval = PyUnicode_New(result_ulength, max_char);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = (max_char <= 255) ? PyUnicode_1BYTE_KIND : (max_char <= 65535) ? PyUnicode_2BYTE_KIND : PyUnicode_4BYTE_KIND;
    result_udata = PyUnicode_DATA(result_uval);
#else
    result_uval = PyUnicode_FromUnicode(NULL, result_ulength);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = sizeof(Py_UNICODE);
    result_udata = PyUnicode_AS_UNICODE(result_uval);
#endif
    char_pos = 0;
    for (i=0; i < value_count; i++) {
        int ukind;
        Py_ssize_t ulength;
        void *udata;
        PyObject *uval = PyTuple_GET_ITEM(value_tuple, i);
        if (unlikely(__Pyx_PyUnicode_READY(uval)))
            goto bad;
        ulength = __Pyx_PyUnicode_GET_LENGTH(uval);
        if (unlikely(!ulength))
            continue;
        if (unlikely(char_pos + ulength < 0))
            goto overflow;
        ukind = __Pyx_PyUnicode_KIND(uval);
        udata = __Pyx_PyUnicode_DATA(uval);
        if (!CYTHON_PEP393_ENABLED || ukind == result_ukind) {
            memcpy((char *)result_udata + char_pos * result_ukind, udata, (size_t) (ulength * result_ukind));
        } else {
            #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030300F0 || defined(_PyUnicode_FastCopyCharacters)
            _PyUnicode_FastCopyCharacters(result_uval, char_pos, uval, 0, ulength);
            #else
            Py_ssize_t j;
            for (j=0; j < ulength; j++) {
                Py_UCS4 uchar = __Pyx_PyUnicode_READ(ukind, udata, j);
                __Pyx_PyUnicode_WRITE(result_ukind, result_udata, char_pos+j, uchar);
            }
            #endif
        }
        char_pos += ulength;
    }
    return result_uval;
overflow:
    PyErr_SetString(PyExc_OverflowError, "join() result is too long for a Python string");
bad:
    Py_DECREF(result_uval);
    return NULL;
#else
    result_ulength++;
    value_count++;
    return PyUnicode_Join(__pyx_empty_unicode, value_tuple);
#endif
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.')) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, (size_t)new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* PyObjectCallMethod1 */
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
}

/* CoroutineBase */
#include <structmember.h>
#include <frameobject.h>
#define __Pyx_Coroutine_Undelegate(gen) Py_CLEAR((gen)->yieldfrom)
static int __Pyx_PyGen__FetchStopIterationValue(CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
        if (!ev) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#if PY_VERSION_HEX >= 0x030300A0
        else if (Py_TYPE(ev) == (PyTypeObject*)PyExc_StopIteration) {
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            Py_DECREF(ev);
        }
#endif
        else if (unlikely(PyTuple_Check(ev))) {
            if (PyTuple_GET_SIZE(ev) >= 1) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                value = PyTuple_GET_ITEM(ev, 0);
                Py_INCREF(value);
#else
                value = PySequence_ITEM(ev, 0);
#endif
            } else {
                Py_INCREF(Py_None);
                value = Py_None;
            }
            Py_DECREF(ev);
        }
        else if (!__Pyx_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            value = ev;
        }
        if (likely(value)) {
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
    } else if (!__Pyx_PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if PY_VERSION_HEX >= 0x030300A0
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
    Py_DECREF(ev);
#else
    {
        PyObject* args = __Pyx_PyObject_GetAttrStr(ev, __pyx_n_s_args);
        Py_DECREF(ev);
        if (likely(args)) {
            value = PySequence_GetItem(args, 0);
            Py_DECREF(args);
        }
        if (unlikely(!value)) {
            __Pyx_ErrRestore(NULL, NULL, NULL);
            Py_INCREF(Py_None);
            value = Py_None;
        }
    }
#endif
    *pvalue = value;
    return 0;
}
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *t, *v, *tb;
    t = exc_state->exc_type;
    v = exc_state->exc_value;
    tb = exc_state->exc_traceback;
    exc_state->exc_type = NULL;
    exc_state->exc_value = NULL;
    exc_state->exc_traceback = NULL;
    Py_XDECREF(t);
    Py_XDECREF(v);
    Py_XDECREF(tb);
}
#define __Pyx_Coroutine_AlreadyRunningError(gen)  (__Pyx__Coroutine_AlreadyRunningError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyRunningError(CYTHON_UNUSED __pyx_CoroutineObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check((PyObject*)gen)) {
        msg = "coroutine already executing";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact((PyObject*)gen)) {
        msg = "async generator already executing";
    #endif
    } else {
        msg = "generator already executing";
    }
    PyErr_SetString(PyExc_ValueError, msg);
}
#define __Pyx_Coroutine_NotStartedError(gen)  (__Pyx__Coroutine_NotStartedError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_NotStartedError(CYTHON_UNUSED PyObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check(gen)) {
        msg = "can't send non-None value to a just-started coroutine";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact(gen)) {
        msg = "can't send non-None value to a just-started async generator";
    #endif
    } else {
        msg = "can't send non-None value to a just-started generator";
    }
    PyErr_SetString(PyExc_TypeError, msg);
}
#define __Pyx_Coroutine_AlreadyTerminatedError(gen, value, closing)  (__Pyx__Coroutine_AlreadyTerminatedError(gen, value, closing), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyTerminatedError(CYTHON_UNUSED PyObject *gen, PyObject *value, CYTHON_UNUSED int closing) {
    #ifdef __Pyx_Coroutine_USED
    if (!closing && __Pyx_Coroutine_Check(gen)) {
        PyErr_SetString(PyExc_RuntimeError, "cannot reuse already awaited coroutine");
    } else
    #endif
    if (value) {
        #ifdef __Pyx_AsyncGen_USED
        if (__Pyx_AsyncGen_CheckExact(gen))
            PyErr_SetNone(__Pyx_PyExc_StopAsyncIteration);
        else
        #endif
        PyErr_SetNone(PyExc_StopIteration);
    }
}
static
PyObject *__Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value, int closing) {
    __Pyx_PyThreadState_declare
    PyThreadState *tstate;
    __Pyx_ExcInfoStruct *exc_state;
    PyObject *retval;
    assert(!self->is_running);
    if (unlikely(self->resume_label == 0)) {
        if (unlikely(value && value != Py_None)) {
            return __Pyx_Coroutine_NotStartedError((PyObject*)self);
        }
    }
    if (unlikely(self->resume_label == -1)) {
        return __Pyx_Coroutine_AlreadyTerminatedError((PyObject*)self, value, closing);
    }
#if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    tstate = __pyx_tstate;
#else
    tstate = __Pyx_PyThreadState_Current;
#endif
    exc_state = &self->gi_exc_state;
    if (exc_state->exc_type) {
        #if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
        #else
        if (exc_state->exc_traceback) {
            PyTracebackObject *tb = (PyTracebackObject *) exc_state->exc_traceback;
            PyFrameObject *f = tb->tb_frame;
            Py_XINCREF(tstate->frame);
            assert(f->f_back == NULL);
            f->f_back = tstate->frame;
        }
        #endif
    }
#if CYTHON_USE_EXC_INFO_STACK
    exc_state->previous_item = tstate->exc_info;
    tstate->exc_info = exc_state;
#else
    if (exc_state->exc_type) {
        __Pyx_ExceptionSwap(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(exc_state);
        __Pyx_ExceptionSave(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    }
#endif
    self->is_running = 1;
    retval = self->body((PyObject *) self, tstate, value);
    self->is_running = 0;
#if CYTHON_USE_EXC_INFO_STACK
    exc_state = &self->gi_exc_state;
    tstate->exc_info = exc_state->previous_item;
    exc_state->previous_item = NULL;
    __Pyx_Coroutine_ResetFrameBackpointer(exc_state);
#endif
    return retval;
}
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *exc_tb = exc_state->exc_traceback;
    if (likely(exc_tb)) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
#else
        PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
        PyFrameObject *f = tb->tb_frame;
        Py_CLEAR(f->f_back);
#endif
    }
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_MethodReturn(CYTHON_UNUSED PyObject* gen, PyObject *retval) {
    if (unlikely(!retval)) {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        if (!__Pyx_PyErr_Occurred()) {
            PyObject *exc = PyExc_StopIteration;
            #ifdef __Pyx_AsyncGen_USED
            if (__Pyx_AsyncGen_CheckExact(gen))
                exc = __Pyx_PyExc_StopAsyncIteration;
            #endif
            __Pyx_PyErr_SetNone(exc);
        }
    }
    return retval;
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen) {
    PyObject *ret;
    PyObject *val = NULL;
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, &val);
    ret = __Pyx_Coroutine_SendEx(gen, val, 0);
    Py_XDECREF(val);
    return ret;
}
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_AsyncGen_USED
        if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
            ret = __Pyx_async_gen_asend_send(yf, value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = _PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03050000 && defined(PyCoro_CheckExact) && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyCoro_CheckExact(yf)) {
            ret = _PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        {
            if (value == Py_None)
                ret = Py_TYPE(yf)->tp_iternext(yf);
            else
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_n_s_send, value);
        }
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        retval = __Pyx_Coroutine_FinishDelegation(gen);
    } else {
        retval = __Pyx_Coroutine_SendEx(gen, value, 0);
    }
    return __Pyx_Coroutine_MethodReturn(self, retval);
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    PyObject *retval = NULL;
    int err = 0;
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    if (__Pyx_CoroutineAwait_CheckExact(yf)) {
        retval = __Pyx_CoroutineAwait_Close((__pyx_CoroutineAwaitObject*)yf, NULL);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_AsyncGen_USED
    if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
        retval = __Pyx_async_gen_asend_close(yf, NULL);
    } else
    if (__pyx_PyAsyncGenAThrow_CheckExact(yf)) {
        retval = __Pyx_async_gen_athrow_close(yf, NULL);
    } else
    #endif
    {
        PyObject *meth;
        gen->is_running = 1;
        meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_close);
        if (unlikely(!meth)) {
            if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                PyErr_WriteUnraisable(yf);
            }
            PyErr_Clear();
        } else {
            retval = PyObject_CallFunction(meth, NULL);
            Py_DECREF(meth);
            if (!retval)
                err = -1;
        }
        gen->is_running = 0;
    }
    Py_XDECREF(retval);
    return err;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Generator_Next(yf);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = _PyGen_Send((PyGenObject*)yf, NULL);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, Py_None);
        } else
        #endif
            ret = Py_TYPE(yf)->tp_iternext(yf);
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        return __Pyx_Coroutine_FinishDelegation(gen);
    }
    return __Pyx_Coroutine_SendEx(gen, Py_None, 0);
}
static PyObject *__Pyx_Coroutine_Close_Method(PyObject *self, CYTHON_UNUSED PyObject *arg) {
    return __Pyx_Coroutine_Close(self);
}
static PyObject *__Pyx_Coroutine_Close(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *retval, *raised_exception;
    PyObject *yf = gen->yieldfrom;
    int err = 0;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    retval = __Pyx_Coroutine_SendEx(gen, NULL, 1);
    if (unlikely(retval)) {
        const char *msg;
        Py_DECREF(retval);
        if ((0)) {
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_Coroutine_Check(self)) {
            msg = "coroutine ignored GeneratorExit";
        #endif
        #ifdef __Pyx_AsyncGen_USED
        } else if (__Pyx_AsyncGen_CheckExact(self)) {
#if PY_VERSION_HEX < 0x03060000
            msg = "async generator ignored GeneratorExit - might require Python 3.6+ finalisation (PEP 525)";
#else
            msg = "async generator ignored GeneratorExit";
#endif
        #endif
        } else {
            msg = "generator ignored GeneratorExit";
        }
        PyErr_SetString(PyExc_RuntimeError, msg);
        return NULL;
    }
    raised_exception = PyErr_Occurred();
    if (likely(!raised_exception || __Pyx_PyErr_GivenExceptionMatches2(raised_exception, PyExc_GeneratorExit, PyExc_StopIteration))) {
        if (raised_exception) PyErr_Clear();
        Py_INCREF(Py_None);
        return Py_None;
    }
    return NULL;
}
static PyObject *__Pyx__Coroutine_Throw(PyObject *self, PyObject *typ, PyObject *val, PyObject *tb,
                                        PyObject *args, int close_on_genexit) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        Py_INCREF(yf);
        if (__Pyx_PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit) && close_on_genexit) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
            goto throw_here;
        }
        gen->is_running = 1;
        if (0
        #ifdef __Pyx_Generator_USED
            || __Pyx_Generator_CheckExact(yf)
        #endif
        #ifdef __Pyx_Coroutine_USED
            || __Pyx_Coroutine_Check(yf)
        #endif
            ) {
            ret = __Pyx__Coroutine_Throw(yf, typ, val, tb, args, close_on_genexit);
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_CoroutineAwait_CheckExact(yf)) {
            ret = __Pyx__Coroutine_Throw(((__pyx_CoroutineAwaitObject*)yf)->coroutine, typ, val, tb, args, close_on_genexit);
        #endif
        } else {
            PyObject *meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                    gen->is_running = 0;
                    return NULL;
                }
                PyErr_Clear();
                __Pyx_Coroutine_Undelegate(gen);
                gen->is_running = 0;
                goto throw_here;
            }
            if (likely(args)) {
                ret = PyObject_CallObject(meth, args);
            } else {
                ret = PyObject_CallFunctionObjArgs(meth, typ, val, tb, NULL);
            }
            Py_DECREF(meth);
        }
        gen->is_running = 0;
        Py_DECREF(yf);
        if (!ret) {
            ret = __Pyx_Coroutine_FinishDelegation(gen);
        }
        return __Pyx_Coroutine_MethodReturn(self, ret);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
    return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    PyObject *typ;
    PyObject *val = NULL;
    PyObject *tb = NULL;
    if (!PyArg_UnpackTuple(args, (char *)"throw", 1, 3, &typ, &val, &tb))
        return NULL;
    return __Pyx__Coroutine_Throw(self, typ, val, tb, args, 1);
}
static CYTHON_INLINE int __Pyx_Coroutine_traverse_excstate(__Pyx_ExcInfoStruct *exc_state, visitproc visit, void *arg) {
    Py_VISIT(exc_state->exc_type);
    Py_VISIT(exc_state->exc_value);
    Py_VISIT(exc_state->exc_traceback);
    return 0;
}
static int __Pyx_Coroutine_traverse(__pyx_CoroutineObject *gen, visitproc visit, void *arg) {
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    return __Pyx_Coroutine_traverse_excstate(&gen->gi_exc_state, visit, arg);
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    Py_CLEAR(gen->yieldfrom);
    __Pyx_Coroutine_ExceptionClear(&gen->gi_exc_state);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        Py_CLEAR(((__pyx_PyAsyncGenObject*)gen)->ag_finalizer);
    }
#endif
    Py_CLEAR(gen->gi_code);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    Py_CLEAR(gen->gi_modulename);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    if (gen->gi_weakreflist != NULL)
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label >= 0) {
        PyObject_GC_Track(self);
#if PY_VERSION_HEX >= 0x030400a1 && CYTHON_USE_TP_FINALIZE
        if (PyObject_CallFinalizerFromDealloc(self))
#else
        Py_TYPE(gen)->tp_del(self);
        if (self->ob_refcnt > 0)
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        /* We have to handle this case for asynchronous generators
           right here, because this code has to be between UNTRACK
           and GC_Del. */
        Py_CLEAR(((__pyx_PyAsyncGenObject*)self)->ag_finalizer);
    }
#endif
    __Pyx_Coroutine_clear(self);
    PyObject_GC_Del(gen);
}
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label < 0) {
        return;
    }
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt == 0);
    self->ob_refcnt = 1;
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        __pyx_PyAsyncGenObject *agen = (__pyx_PyAsyncGenObject*)self;
        PyObject *finalizer = agen->ag_finalizer;
        if (finalizer && !agen->ag_closed) {
            PyObject *res = __Pyx_PyObject_CallOneArg(finalizer, self);
            if (unlikely(!res)) {
                PyErr_WriteUnraisable(self);
            } else {
                Py_DECREF(res);
            }
            __Pyx_ErrRestore(error_type, error_value, error_traceback);
            return;
        }
    }
#endif
    if (unlikely(gen->resume_label == 0 && !error_value)) {
#ifdef __Pyx_Coroutine_USED
#ifdef __Pyx_Generator_USED
    if (!__Pyx_Generator_CheckExact(self))
#endif
        {
        PyObject_GC_UnTrack(self);
#if PY_MAJOR_VERSION >= 3  || defined(PyErr_WarnFormat)
        if (unlikely(PyErr_WarnFormat(PyExc_RuntimeWarning, 1, "coroutine '%.50S' was never awaited", gen->gi_qualname) < 0))
            PyErr_WriteUnraisable(self);
#else
        {PyObject *msg;
        char *cmsg;
        #if CYTHON_COMPILING_IN_PYPY
        msg = NULL;
        cmsg = (char*) "coroutine was never awaited";
        #else
        char *cname;
        PyObject *qualname;
        qualname = gen->gi_qualname;
        cname = PyString_AS_STRING(qualname);
        msg = PyString_FromFormat("coroutine '%.50s' was never awaited", cname);
        if (unlikely(!msg)) {
            PyErr_Clear();
            cmsg = (char*) "coroutine was never awaited";
        } else {
            cmsg = PyString_AS_STRING(msg);
        }
        #endif
        if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, cmsg, 1) < 0))
            PyErr_WriteUnraisable(self);
        Py_XDECREF(msg);}
#endif
        PyObject_GC_Track(self);
        }
#endif
    } else {
        PyObject *res = __Pyx_Coroutine_Close(self);
        if (unlikely(!res)) {
            if (PyErr_Occurred())
                PyErr_WriteUnraisable(self);
        } else {
            Py_DECREF(res);
        }
    }
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt > 0);
    if (--self->ob_refcnt == 0) {
        return;
    }
    {
        Py_ssize_t refcnt = self->ob_refcnt;
        _Py_NewReference(self);
        self->ob_refcnt = refcnt;
    }
#if CYTHON_COMPILING_IN_CPYTHON
    assert(PyType_IS_GC(self->ob_type) &&
           _Py_AS_GC(self)->gc.gc_refs != _PyGC_REFS_UNTRACKED);
    _Py_DEC_REFTOTAL;
#endif
#ifdef COUNT_ALLOCS
    --Py_TYPE(self)->tp_frees;
    --Py_TYPE(self)->tp_allocs;
#endif
#endif
}
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_name;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_name;
    Py_INCREF(value);
    self->gi_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_qualname;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_qualname;
    Py_INCREF(value);
    self->gi_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (unlikely(!gen))
        return NULL;
    return __Pyx__Coroutine_NewInit(gen, body, code, closure, name, qualname, module_name);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    gen->gi_exc_state.exc_type = NULL;
    gen->gi_exc_state.exc_value = NULL;
    gen->gi_exc_state.exc_traceback = NULL;
#if CYTHON_USE_EXC_INFO_STACK
    gen->gi_exc_state.previous_item = NULL;
#endif
    gen->gi_weakreflist = NULL;
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    Py_XINCREF(module_name);
    gen->gi_modulename = module_name;
    Py_XINCREF(code);
    gen->gi_code = code;
    PyObject_GC_Track(gen);
    return gen;
}

/* PatchModuleWithCoroutine */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    int result;
    PyObject *globals, *result_obj;
    globals = PyDict_New();  if (unlikely(!globals)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_coroutine_type",
    #ifdef __Pyx_Coroutine_USED
        (PyObject*)__pyx_CoroutineType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_generator_type",
    #ifdef __Pyx_Generator_USED
        (PyObject*)__pyx_GeneratorType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "_module", module) < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "__builtins__", __pyx_b) < 0)) goto ignore;
    result_obj = PyRun_String(py_code, Py_file_input, globals, globals);
    if (unlikely(!result_obj)) goto ignore;
    Py_DECREF(result_obj);
    Py_DECREF(globals);
    return module;
ignore:
    Py_XDECREF(globals);
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch module with custom type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    py_code++;
#endif
    return module;
}

/* PatchGeneratorABC */
#ifndef CYTHON_REGISTER_ABCS
#define CYTHON_REGISTER_ABCS 1
#endif
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
static PyObject* __Pyx_patch_abc_module(PyObject *module);
static PyObject* __Pyx_patch_abc_module(PyObject *module) {
    module = __Pyx_Coroutine_patch_module(
        module, ""
"if _cython_generator_type is not None:\n"
"    try: Generator = _module.Generator\n"
"    except AttributeError: pass\n"
"    else: Generator.register(_cython_generator_type)\n"
"if _cython_coroutine_type is not None:\n"
"    try: Coroutine = _module.Coroutine\n"
"    except AttributeError: pass\n"
"    else: Coroutine.register(_cython_coroutine_type)\n"
    );
    return module;
}
#endif
static int __Pyx_patch_abc(void) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    static int abc_patched = 0;
    if (CYTHON_REGISTER_ABCS && !abc_patched) {
        PyObject *module;
        module = PyImport_ImportModule((PY_MAJOR_VERSION >= 3) ? "collections.abc" : "collections");
        if (!module) {
            PyErr_WriteUnraisable(NULL);
            if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning,
                    ((PY_MAJOR_VERSION >= 3) ?
                        "Cython module failed to register with collections.abc module" :
                        "Cython module failed to register with collections module"), 1) < 0)) {
                return -1;
            }
        } else {
            module = __Pyx_patch_abc_module(module);
            abc_patched = 1;
            if (unlikely(!module))
                return -1;
            Py_DECREF(module);
        }
        module = PyImport_ImportModule("backports_abc");
        if (module) {
            module = __Pyx_patch_abc_module(module);
            Py_XDECREF(module);
        }
        if (!module) {
            PyErr_Clear();
        }
    }
#else
    if ((0)) __Pyx_Coroutine_patch_module(NULL, NULL);
#endif
    return 0;
}

/* Generator */
static PyMethodDef __pyx_Generator_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into generator,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in generator,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close_Method, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside generator.")},
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Generator_memberlist[] = {
    {(char *) "gi_running", T_BOOL, offsetof(__pyx_CoroutineObject, is_running), READONLY, NULL},
    {(char*) "gi_yieldfrom", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     (char*) PyDoc_STR("object being iterated by 'yield from', or None")},
    {(char*) "gi_code", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_code), READONLY, NULL},
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Generator_getsets[] = {
    {(char *) "__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     (char*) PyDoc_STR("name of the generator"), 0},
    {(char *) "__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     (char*) PyDoc_STR("qualified name of the generator"), 0},
    {0, 0, 0, 0, 0}
};
static PyTypeObject __pyx_GeneratorType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
    0,
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    0,
    (iternextfunc) __Pyx_Generator_Next,
    __pyx_Generator_methods,
    __pyx_Generator_memberlist,
    __pyx_Generator_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if CYTHON_USE_TP_FINALIZE
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if CYTHON_USE_TP_FINALIZE
    __Pyx_Coroutine_del,
#elif PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
};
static int __pyx_Generator_init(void) {
    __pyx_GeneratorType_type.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
    __pyx_GeneratorType_type.tp_iter = PyObject_SelfIter;
    __pyx_GeneratorType = __Pyx_FetchCommonType(&__pyx_GeneratorType_type);
    if (unlikely(!__pyx_GeneratorType)) {
        return -1;
    }
    return 0;
}

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
